#ifndef __DEFORM_GRAPH_ALL_H__
#define __DEFORM_GRAPH_ALL_H__

#include <QMutex>
#include <memory>
#include <hash_map>
#include "../Util/BoostGraphDef.h"
#include "deform.h"

class ShapeModel;
class Mesh3_real;
class ShapeAnimation;
class CorrespondenceTrack;
class SE3;
class GCoptimizationGeneralGraph;
class SpinImage;
class SpinImageStack;
class FileSequence;
class Scene;
class CorrespondenceList;
class MatchCandidates;
class CurvInfo;
class ColormapRange;

typedef double taucsType;
typedef std::map<int3_type,idxList,int3Comp> int3mapList;
typedef std::vector<SE3> RigidDeform;
typedef std::vector<float> FloatArray;
typedef std::vector< std::vector<CurvInfo> > CurvInfoList;
typedef std::vector<std::string> StringArray;

namespace Deform
{
	class ErrorTerm;
	typedef std::vector<ErrorTerm*> TermList;

	namespace DeformAll_Utility		/**< Forward declarations */
	{
		class SE3_6DOF;
		class TfVarTranslate;
		class ShapeDeformed;
		class ManagedFrameSequence;
		class AllSampleGraph;
		class RefWeight;
		typedef std::vector<SE3_6DOF> TransformList;
		typedef std::map<int,const TransformList*> SparseTLArray;
		typedef std::deque<RefWeight> RefWeightList;

		// Predefined constants for DeformAll
		static const bool USING_FORWARD_TF = false;
		static const double ERROR_POINT_WEIGHT = 0.2;
		static const double ERROR_PLANE_WEIGHT = 0.8;

		// Saving DeformAll state: version #
		static const float VER_NUM = 0.4f;
		static const std::string version_msg("DeformAll_VER");

		// Solver options for DeformAll
		class Options
		{
		public:
			Options() : stopEpsilon(1e-12), stopEpsilon_sqRt(1e-6), stopEpsilon_cubeRt(1e-4), ptr_scene(NULL), colormap(NULL) { }

			std::string getSIFileName() const {
				std::string siFile = autoSaveSolverState;
				siFile.insert(siFile.length() - 5, "_SI");
				//cout << DeformGraphAll_Utility::printHeader << "Spin image save file is: " << siFile << endl;
				return siFile;
			}
			std::string getCurvFileName() const {
				std::string curvFile = autoSaveSolverState;
				curvFile.insert(curvFile.length() - 5, "_CV");
				//cout << DeformGraphAll_Utility::printHeader << "Curv info save file is: " << curvFile << endl;
				return curvFile;
			}
			std::string getGCFileName() const {
				std::string GCFile = autoSaveSolverState;
				GCFile.insert(GCFile.length() - 5, "_GC");
				return GCFile;
			}
			void setFromGUI( const Options& options );

			int numBones;
			real_type sceneGridSpacing;
			taucsType stopEpsilon, stopEpsilon_sqRt, stopEpsilon_cubeRt;
			int maxSolverIterations, maxTStepIterations;
			real_type maxCorrepDist, bdyMaxCorrepDist, minNormalCosAngle;			/**< Correspondence outlier thresholds */
			real_type refSampleDist, closestPt_normalWeight;
			FloatArray TWeightsIK, TWeights;						/**< Weights for each error term in each step */
			FloatArray WWeights;							/**< Weights for each error term in each step */
			real_type maxJointDist;									/**< Maximum distance a joint can stretch apart */
			int refFrameIndex;														/**< Index of the reference frame */
			real_type hingeExtent;										/**< Extent of a hinge (half the length) */
			int hingeJointSamples;													/**< Number of samples for a hinge joint */
			int hingeJointNumFrames;												/**< Number of frames required before estimating a hinge */
			real_type jointIsHingeThreshold;										/**< Threshold that determines whether a joint is a hinge.  The ratio of the lowest singular value over the sum of the singular values.  A conservative estimate is 1.0. */
			int slidingWindowSize;
			real_type edgeLenStretchThreshold_Sq;							/**< Multiplier of the edge length for removing artifacts when deforming the mesh */
			real_type deformCosAngleThreshold;										/**< A skinny deformed vertex will be removed when the angle is close to 0 or 180 degrees using this threshold. */

			int excludeLabelThreshold_num;
			double excludeLabelThreshold_ratio;

			// Discrete W-Step Related
			real_type splitRegionMinError;
			real_type occludedLabel_ValidCorrespRatioThreshold;						/**< The minimum ratio of valid_corresp/num_corresp in each region */
			real_type occludedLabel_ValidCorrespNumThreshold;						/**< The minimum number of valid_corresp in each region */

			// Saving current state
			std::string autoSaveSolverState, autoLoadSolverState;
			std::string autoSaveSampleSpinImages, autoLoadSampleSpinImages;
			std::string writeFrameToRef, writeRefToFrame;
			std::string saveTaucsErrorLog;
			bool loadGraphOnly;

			// Loading ground truth data
			std::string groundTruth_shapeFile, groundTruth_weightFile, groundTruth_motionFile;

			// Colormap
			ColormapRange *colormap;


			// Pointer to scene for computing graph cut initial correspondences
			Scene *ptr_scene;


			// Options for DeformGraphAll
			bool initGCMatchRef;
			real_type pruneJoint_minEdgeRatio;
			std::string loadFinalShape;								/**< Filename of the final reconstructed mesh */
			std::string loadRiggedShape;							/**< Filename of a rigged mesh */
			std::string loadRiggedWeights;
			std::string loadRiggedBVH;
			real_type finalRefSampleDist;
			bool computeTotalEfit;

			// 1. ASG construction
			int EKNNGraph_k;
			bool performRebuildASG;
			bool accumulateASGOnReference;							/**< Whether to accumulate ASG on reference pose -- for comparing with pure accumulation strategy */
			real_type DSGMaxEdgeStretchFactor;						/**< Maximum factor an edge can stretch on the DSG */
			real_type DSGMaxEdgeLen;								/**< Maximum length of an edge on the DSG */
			real_type finalDSGMaxEdgeStretchFactor;					/**< For final reconstructed DSG: Maximum factor an edge can stretch on the DSG */
			real_type finalDSGMaxEdgeLen;							/**< For final reconstructed DSG: Maximum length of an edge on the DSG */

			// 2. Matching occluded bones
			real_type minMatchRatio;								/**< Minimum ratio of remaining geometry to trigger a match of occluded bones */
			real_type occluded_maxCorrespDist, occluded_maxNormalAngleDifference;			/**< Correspondence outlier thresholds */
			real_type occluded_jointOptWeight;						/**< Weight for joint term in optimization*/

		private:
			// Initialization parameters
			real_type saved_boundaryCorrespMaxDist_Multiplier, saved_maxCorrespDistance_Multiplier, saved_maxNormalAngleDifference;
			FloatArray saved_TStepWeights;
		};
	}

	namespace DeformGraphAll_Utility {
		class FrameSolution;
		typedef std::map<int,DeformGraphAll_Utility::FrameSolution*> SolutionTable;
	}

	namespace DeformAll_Utility
	{
		// 1. Utility classes
		//////////////////////////////////////////////////////////////////////////
		// Class representing rigid transformation in axis-angle format
		//////////////////////////////////////////////////////////////////////////
		class SE3_6DOF
		{
		public:
			SE3_6DOF();
			SE3_6DOF( const real3_type& axisAngle, const real3_type& translation );
			SE3_6DOF( const Geometry::Matrix33f& rotMatrix, const real3_type& translation );
			SE3_6DOF( const real3_type& axisAngle, const Geometry::Matrix33f& axisAngleRotationMatrix, const real3_type& translation );

			// Creation from linear rotation and linear translation parameters
			void createFromLinear( const real3_type& linearRot, const real3_type& linearTr );

			// Access
			const Geometry::Matrix33f& rotMatrix() const;
			const real3_type& rot() const;
			const real3_type& tr() const;

			// Operators
			//const SE3_6DOF& operator+=( const SE3_6DOF& rhs );
			SE3_6DOF operator-() const;

			real3_type applyLinearInverse( const real3_type& rhs ) const;
			real3_type applyLinearInverse_Normal( const real3_type& rhs ) const;
			real3_type applyInverse( const real3_type& rhs ) const;
			real3_type applyInverse_Normal( const real3_type& rhs ) const;

		private:
			real3_type axisAngle, translation;

			// Cache the 3x3 rotation matrix
			Geometry::Matrix33f rotationMatrix;
		};

		// Inline functions
		inline const Geometry::Matrix33f& SE3_6DOF::rotMatrix() const {
			return rotationMatrix;
		}
		inline const real3_type& SE3_6DOF::rot() const {
			return axisAngle;
		}
		inline const real3_type& SE3_6DOF::tr() const {
			return translation;
		}

		// Operations with SE3_6DOF
		//const SE3_6DOF operator+( const SE3_6DOF& lhs, const SE3_6DOF& rhs);
		const inline SE3_6DOF operator*( const SE3_6DOF& lhs, const SE3_6DOF& rhs )
		{
			Geometry::Matrix33f rotMatrix = lhs.rotMatrix();
			real3_type translation = (rotMatrix * rhs.tr()) + lhs.tr();
			rotMatrix = rotMatrix * rhs.rotMatrix();

			return SE3_6DOF(lhs.rot() + rhs.rot(), rotMatrix, translation);

			////return SE3_6DOF(lhs.rot() + rhs.rot(), rhs.tr() + lhs.tr());
		};

		const inline real3_type operator*( const SE3_6DOF& lhs, const real3_type& rhs )
		{
			return (lhs.rotMatrix()*rhs + lhs.tr());
		};

		std::ostream& operator<<( std::ostream& out, const SE3_6DOF &tf );
		std::ostream& operator<<( std::ostream& out, const TransformList &tfList );
		//TransformList operator+( const TransformList& tfList_left, const TransformList& tfList_right );
		TransformList operator*( const TransformList& tfList_left, const TransformList& tfList_right );
		TransformList operator-( const TransformList& tfList );

		//////////////////////////////////////////////////////////////////////////
		// Class storing intermediary deformation of the shape
		//////////////////////////////////////////////////////////////////////////
		class ShapeDeformed
		{
		public:
			ShapeDeformed( const ShapeModel& shape );
			~ShapeDeformed();

			void draw( bool showInvalid ) const;
			void deform( const RefWeightList& weightList, const TransformList& tfList, bool inverse );

			// Access functions
			const PointList& getVertex() const;
			const PointList& getNormal() const;
			bool isValid( index_type idx ) const {
				return vertexValid[idx];
			}
			const ShapeModel& getOriginalShape() const {
				return originalShape;
			}
			const std::vector<bool>& getVertexValid() const {
				return vertexValid;
			}
			std::vector<bool>& getVertexValid() {
				return vertexValid;
			}
			const real3_type& getBoxMin() const {
				return box_min;
			}
			const real3_type& getBoxMax() const {
				return box_max;
			}
			void setValid( index_type idx, bool valid ) {
				vertexValid[idx] = valid;
			}

			// Output functions
			void writeGeometryOBJ( const char *file, bool flipNormals );

		private:
			// Data members
			const ShapeModel& originalShape;		/**< Pointer to the original shape */
			PointList *vertex, *normal;				/**< Vertex & Normal Data */
			real3_type box_min, box_max;			/**< Bounding Box */
			std::vector<bool> vertexValid;			/**< Valid bit for each vertex */
		};

		inline const PointList& ShapeDeformed::getVertex() const {
			return *vertex;
		}

		inline const PointList& ShapeDeformed::getNormal() const {
			return *normal;
		}

		//////////////////////////////////////////////////////////////////////////
		// Utility class for translating transformation variable indices
		class TfVarTranslate {
		public:
			TfVarTranslate( const DeformGraphAll_Utility::SolutionTable& frameSolution, int referenceFrameIndex, const std::vector<int>& excludedLabels );
			TfVarTranslate( int frameIndex, int numBones, const std::vector<int>& excludedLabels, const std::vector<int>& occludedLabels );

			bool validBone( int frameIndex, int boneIndex ) const;
			int linearIndex( int frameIndex, int boneIndex ) const;
			void frameBoneIndex( index_type varIndex, int& frameIndex, int& boneIndex ) const;
			int numTotalBones() const;
			void printMapping() const;
			void remove_linearIndices( const idxList& toRemove );			/**< Remove a list of linear indices */

			int getOriginalNumBonesPerFrame() const {
				return originalNumBonesPerFrame;
			}

		private:
			int originalNumBonesPerFrame;
			stdext::hash_map< int, std::vector<int> > frameBoneIndex_to_linearIndex;
			std::vector<int2_type> linearIndex_to_frameBoneIndex;
		};

		//////////////////////////////////////////////////////////////////////////
		// Storing weights found from the reference frame's vertex weight grid
		//////////////////////////////////////////////////////////////////////////
		class RefWeight {
		public:
			RefWeight();

			bool valid() const {
				return bones.size();
			}

			size_type size() const {
				return bones.size();
			}

			bool findLabel( int &label ) const;			/**< Find a label (= index with the maximum weight) */

			// Data members
			idxList bones;								/**< List of affecting bones */
			distList weights;							/**< List of affecting weights */

			// Related to grid cell
			int3_type cellIndex;						/**< Index of the enclosing cell in the reference grid */
			bool extendedCell;							/**< Is the enclosing cell not part of the reference grid, and we had to extend the grid? */
			idxList closestExistingCellLinearIndices;	/**< If it's an extended cell, what is(are) the closest existing cell(s)? */
			real3_type trilinearCoordinate;				/**< Store trilinear coordinate within the cell */
		};



		class ManagedFrameSequence {
		public:
			ManagedFrameSequence( const std::string& rootDir, const std::string& meshSequence );
			~ManagedFrameSequence();

			bool sequenceValid() const;							/**< Is the sequence valid? */
			bool frameValid( int frame ) const;					/**< Is a given frame valid? */
			Mesh3_real* getFrame( int frame, bool& cached );	/**< Return the frame */

		private:
			std::string rootDir, meshSequence;					/**< Root dir and sequence start file */
			FileSequence *frameFS;								/**< Managing the frame sequence. */
			idxList frameFS_indices;							/**< List of valid frame indices in the filename. */

			// Caching the loaded shapes in memory
			bool checkCache( int frame ) const;
			std::vector<Mesh3_real*> shapeCache;				/**< Local cache of shapes in memory. */
		};


		// Additional function declarations
		void drawShapeValid( const PointList& vertex, const PointList& normal, const ShapeModel& originalShape, const std::vector<bool>& vertexValid, bool showInvalid );
		void drawShapeWeights( const PointList& vertex, const PointList& normal, const ShapeModel& shape, const RefWeightList& weightList, const std::vector<bool> vertexValid, const std::vector<RealColor>& boneColors, real_type alpha, int showBone = -1 );
		void drawShapeWeightsValid( const PointList& vertex, const PointList& normal, const ShapeModel& mesh, const RefWeightList& shapeWeightList, const std::vector<bool>& validBit, const RealColor& shapeColor, const std::vector<RealColor>& compColors, const DrawOptions& drawOptions, bool showInvalid, bool enableLighting = false );
		void MatrixToAxisAngle( const Geometry::Matrix33f& R, real3_type &axis, real_type &angle );
		void AxisAngleToMatrix( const real3_type& axisAngle, Geometry::Matrix33f& m );

		taucsType* solveLinear( TermList& terms, int numVars, taucsType& totalResidual, taucsType& stepResidual, const std::string& logFile, taucsType diagAddValue/* = 0.0*/ );
		template<typename Real, int N_col> extern void solveLinear_denseSVD( const std::vector< Geometry::Vec<N_col,Real> >& lhsRows, const std::vector<Real>& rhsEntries, Geometry::Vec<N_col,Real>& soln );

		// F. Constructing T-Step optimization
		inline void helperJacSA( real_type boneWeight, const real3_type& sourcePos, const real3_type& targetNormal, data_type& R )
		{
			static real_type x_i1, x_i2, x_i3, n_j1, n_j2, n_j3;
			x_i1 = sourcePos[0];     x_i2 = sourcePos[1];     x_i3 = sourcePos[2];
			n_j1 = targetNormal[0];  n_j2 = targetNormal[1];  n_j3 = targetNormal[2];

			//data_type R(6);
			R.val[0] = (-n_j2*x_i3+n_j3*x_i2)*boneWeight;
			R.val[1] = (n_j1*x_i3-n_j3*x_i1)*boneWeight;
			R.val[2] = (-n_j1*x_i2+n_j2*x_i1)*boneWeight;
			R.val[3] = n_j1*boneWeight;
			R.val[4] = n_j2*boneWeight;
			R.val[5] = n_j3*boneWeight;

			//return R;
		};

		inline void helperJacPtSA( real_type w_i, real_type w_j, const real3_type& sourcePos, const real3_type& targetPos, data_type& R0, data_type& R1, data_type& R2 )
		{
			static real_type x_i1, x_i2, x_i3, x_j1, x_j2, x_j3;
			x_i1 = sourcePos[0];       x_i2 = sourcePos[1];       x_i3 = sourcePos[2];
			x_j1 = targetPos[0];       x_j2 = targetPos[1];       x_j3 = targetPos[2];

			//R.resize(3,data_type(12));
			//R[0].setZero();
			//R[1].setZero();
			//R[2].setZero();
			R0.val[1] = w_i*x_i3;
			R0.val[2] = -w_i*x_i2;
			R0.val[3] = w_i;
			R0.val[7] = -w_j*x_j3;
			R0.val[8] = w_j*x_j2;
			R0.val[9] = -w_j;
			R1.val[0] = -w_i*x_i3;
			R1.val[2] = w_i*x_i1;
			R1.val[4] = w_i;
			R1.val[6] = w_j*x_j3;
			R1.val[8] = -w_j*x_j1;
			R1.val[10] = -w_j;
			R2.val[0] = w_i*x_i2;
			R2.val[1] = -w_i*x_i1;
			R2.val[5] = w_i;
			R2.val[6] = -w_j*x_j2;
			R2.val[7] = w_j*x_j1;
			R2.val[11] = -w_j;
		};

		void helperJointJacPtSA( const real3_type& jointDefPos_i, const real3_type& jointDefPos_j, std::vector<data_type>& R );

		// Derivative of the ball joint constraint term, without pre-rotation, with the transformations in the correct order.
		// [ ri_13*(t_i2 - x_2) - ri_12*(t_i3 - x_3), ri_11*(t_i3 - x_3) - ri_13*(t_i1 - x_1), ri_12*(t_i1 - x_1) - ri_11*(t_i2 - x_2), c_i3*ri_12 - c_i2*ri_13 - ri_11, c_i1*ri_13 - ri_12 - c_i3*ri_11, c_i2*ri_11 - c_i1*ri_12 - ri_13, rj_12*(t_j3 - x_3) - rj_13*(t_j2 - x_2), rj_13*(t_j1 - x_1) - rj_11*(t_j3 - x_3), rj_11*(t_j2 - x_2) - rj_12*(t_j1 - x_1), rj_11 + c_j2*rj_13 - c_j3*rj_12, rj_12 - c_j1*rj_13 + c_j3*rj_11, rj_13 + c_j1*rj_12 - c_j2*rj_11]
		// [ ri_23*(t_i2 - x_2) - ri_22*(t_i3 - x_3), ri_21*(t_i3 - x_3) - ri_23*(t_i1 - x_1), ri_22*(t_i1 - x_1) - ri_21*(t_i2 - x_2), c_i3*ri_22 - c_i2*ri_23 - ri_21, c_i1*ri_23 - ri_22 - c_i3*ri_21, c_i2*ri_21 - c_i1*ri_22 - ri_23, rj_22*(t_j3 - x_3) - rj_23*(t_j2 - x_2), rj_23*(t_j1 - x_1) - rj_21*(t_j3 - x_3), rj_21*(t_j2 - x_2) - rj_22*(t_j1 - x_1), rj_21 + c_j2*rj_23 - c_j3*rj_22, rj_22 - c_j1*rj_23 + c_j3*rj_21, rj_23 + c_j1*rj_22 - c_j2*rj_21]
		// [ ri_33*(t_i2 - x_2) - ri_32*(t_i3 - x_3), ri_31*(t_i3 - x_3) - ri_33*(t_i1 - x_1), ri_32*(t_i1 - x_1) - ri_31*(t_i2 - x_2), c_i3*ri_32 - c_i2*ri_33 - ri_31, c_i1*ri_33 - ri_32 - c_i3*ri_31, c_i2*ri_31 - c_i1*ri_32 - ri_33, rj_32*(t_j3 - x_3) - rj_33*(t_j2 - x_2), rj_33*(t_j1 - x_1) - rj_31*(t_j3 - x_3), rj_31*(t_j2 - x_2) - rj_32*(t_j1 - x_1), rj_31 + c_j2*rj_33 - c_j3*rj_32, rj_32 - c_j1*rj_33 + c_j3*rj_31, rj_33 + c_j1*rj_32 - c_j2*rj_31]
		//
		// Evaluated at zero, it becomes this, which is used in the helperJointJacInvPtSA function that follows. 
		// [ ri_12*x_3 - ri_13*x_2, ri_13*x_1 - ri_11*x_3, ri_11*x_2 - ri_12*x_1, -ri_11, -ri_12, -ri_13, rj_13*x_2 - rj_12*x_3, rj_11*x_3 - rj_13*x_1, rj_12*x_1 - rj_11*x_2, rj_11, rj_12, rj_13]
		// [ ri_22*x_3 - ri_23*x_2, ri_23*x_1 - ri_21*x_3, ri_21*x_2 - ri_22*x_1, -ri_21, -ri_22, -ri_23, rj_23*x_2 - rj_22*x_3, rj_21*x_3 - rj_23*x_1, rj_22*x_1 - rj_21*x_2, rj_21, rj_22, rj_23]
		// [ ri_32*x_3 - ri_33*x_2, ri_33*x_1 - ri_31*x_3, ri_31*x_2 - ri_32*x_1, -ri_31, -ri_32, -ri_33, rj_33*x_2 - rj_32*x_3, rj_31*x_3 - rj_33*x_1, rj_32*x_1 - rj_31*x_2, rj_31, rj_32, rj_33]
		inline void helperJointJacInvPtSA( const real3_type& jointPos, const Geometry::Matrix33f& ri, const Geometry::Matrix33f& rj, data_type& R0, data_type& R1, data_type& R2 )
		{
			// This is expanded from the identity transformation [0,0,0,0,0,0].

			static real_type x_0, x_1, x_2;
			x_0 = jointPos[0];       x_1 = jointPos[1];       x_2 = jointPos[2];

			//R.clear();
			//R.resize(3, data_type(12, 0.0));

			R0.val[0] = ri(0,1)*x_2 - ri(0,2)*x_1;
			R0.val[1] = ri(0,2)*x_0 - ri(0,0)*x_2;
			R0.val[2] = ri(0,0)*x_1 - ri(0,1)*x_0;
			R0.val[3] = -ri(0,0);
			R0.val[4] = -ri(0,1);
			R0.val[5] = -ri(0,2);
			R0.val[6] = rj(0,2)*x_1 - rj(0,1)*x_2;
			R0.val[7] = rj(0,0)*x_2 - rj(0,2)*x_0;
			R0.val[8] = rj(0,1)*x_0 - rj(0,0)*x_1;
			R0.val[9] = rj(0,0);
			R0.val[10] = rj(0,1);
			R0.val[11] = rj(0,2);
			R1.val[0] = ri(1,1)*x_2 - ri(1,2)*x_1;
			R1.val[1] = ri(1,2)*x_0 - ri(1,0)*x_2;
			R1.val[2] = ri(1,0)*x_1 - ri(1,1)*x_0;
			R1.val[3] = -ri(1,0);
			R1.val[4] = -ri(1,1);
			R1.val[5] = -ri(1,2);
			R1.val[6] = rj(1,2)*x_1 - rj(1,1)*x_2;
			R1.val[7] = rj(1,0)*x_2 - rj(1,2)*x_0;
			R1.val[8] = rj(1,1)*x_0 - rj(1,0)*x_1;
			R1.val[9] = rj(1,0);
			R1.val[10] = rj(1,1);
			R1.val[11] = rj(1,2);
			R2.val[0] = ri(2,1)*x_2 - ri(2,2)*x_1;
			R2.val[1] = ri(2,2)*x_0 - ri(2,0)*x_2;
			R2.val[2] = ri(2,0)*x_1 - ri(2,1)*x_0;
			R2.val[3] = -ri(2,0);
			R2.val[4] = -ri(2,1);
			R2.val[5] = -ri(2,2);
			R2.val[6] = rj(2,2)*x_1 - rj(2,1)*x_2;
			R2.val[7] = rj(2,0)*x_2 - rj(2,2)*x_0;
			R2.val[8] = rj(2,1)*x_0 - rj(2,0)*x_1;
			R2.val[9] = rj(2,0);
			R2.val[10] = rj(2,1);
			R2.val[11] = rj(2,2);
		}

		// I. Solver State File I/O
		void saveTransformationList( const TransformList& tfList, std::ofstream& out );
		void loadTransformationList( TransformList& tfList, std::ifstream& in );
		template<typename T> extern void saveVectorList( const std::vector<T>& vec_list, std::ofstream& out );
		template<typename T> extern void loadVectorList( std::vector<T>& vec_list, std::ifstream& in );

		// L. Misc
		void checkEdgeLengths( const PointList& vertex, const ShapeModel& shape, real_type edgeThresholdSq, real_type cosAngleThreshold, std::vector<bool> &valid );
	}

	namespace DeformGraphAll_Utility
	{
		// Saving DeformGraphAll state: version #
		const float VER_NUM = 0.2f;
		const std::string version_msg("DeformGraphAll_VER");
		const std::string printHeader("--*> ");

		class FrameSolution
		{
		public:
			FrameSolution( const ShapeModel& shape, const int frameNumber );			/**< Barebones constructor (used for loading from file only) */
			FrameSolution( const ShapeModel& shape, const int numBones, const int frameNumber, const real_type sampleDist );
			virtual ~FrameSolution();

			// Member functions
			std::vector<bool> isBoneOccluded() const;														/**< Returns a boolean array indicating whether each label is occluded */
			std::vector<bool> isBoneOccludedOrExcluded( const std::vector<int>& excludedLabels) const;		/**< Returns a boolean array indicating whether each label is occluded or excluded*/
			const SpinImageStack& getSpinImages( const char* filename, const Scene& scene );
			const CurvInfoList& getCurvInfo( const char* filename, const Scene& scene );
			const distList* getCurvScales() const;
			void draw( const PointList& points, const DrawOptions& options, bool drawRef ) const;
			void setRefDeformed( DeformAll_Utility::ShapeDeformed *def );
			inline const DeformAll_Utility::ShapeDeformed* getRefDeformed() const { return toRef; }

			// Data members (saved in state file)
			const int frameNumber;							/**< Frame number */
			DeformAll_Utility::TransformList tfList;		/**< Rigid transformations for this frame */
			std::vector<int> occludedLabels;				/**< A list of occluded labels in this frame */
			idxList frameSamples;							/**< Sampling on this frame */

			// Data members (not saved in state file)
			const ShapeModel& originalShape;				/**< Original frame */
			bool solveThisFrame;							/**< A boolean indicating whether we should solve this frame */
			distList sampleError;							/**< For visualization: plot sample error per frame */

		private:
			// Feature matching (saved in separate files)
			SpinImageStack *sampleSI;
			CurvInfoList *sampleCurv;
			distList *curvScales;

			// New feature!  Transform geometry to reference frame
			DeformAll_Utility::ShapeDeformed *toRef;
			RealColor toRefColor;
		};

		// Define the type for a table of FrameSolutions
		typedef std::map<int,DeformGraphAll_Utility::FrameSolution*> SolutionTable;
		void getSortedFrameList( const DeformGraphAll_Utility::SolutionTable& frameSolution, idxList& result );

		// Finding closest points
		typedef std::map<int,int> ClosestPointMap;
		typedef std::map<int,ClosestPointMap> AllClosestPoints;
		real_type findClosestPoints( const DeformAll_Utility::AllSampleGraph& asg, const DeformAll_Utility::SparseTLArray& allTfList, const int targetFrameNumber, const ShapeModel& targetShape, const idxList& targetSampleIndices, const std::vector<bool>& isOccluded, ClosestPointMap& result, const DeformAll_Utility::Options& defOpt, distList *sampleError = NULL, bool computeError = false );

		// Initialization
		void setRandomLabels( DeformAll_Utility::AllSampleGraph& asg, const PointList& points, int numBones );
		void propogateInitialLabels_Ref( DeformGraphAll_Utility::FrameSolution *refSoln, const MatchCandidates& matchCand, DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult );
		void propogateInitialLabels_Ref( const PointList& asgVertex, const int numBones, const MatchCandidates& matchCand, DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult );
		void propogateInitialLabels( const MatchCandidates& matchCand, const DeformAll_Utility::AllSampleGraph& asg, const ClosestPointMap& closestPts, DeformAll_Utility::TransformList &tfResult );
		void propogateInitialLabels( const MatchCandidates& matchCand, const DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult );

		// Estimating Joints
		enum DetectedJointType {
			JOINT_NONE = 0,
			JOINT_BALL,
			JOINT_HINGE
		};
		struct Joint {
			DetectedJointType type;
			Geometry::Vec3d location;
			Geometry::Vec3d constraint;
		};
		typedef std::map<int,DeformGraphAll_Utility::Joint> JointMap;

		// Filling in occluded transformations
		void fillOccludedTfs( const DeformGraphAll_Utility::JointMap& jointMap, DeformGraphAll_Utility::FrameSolution& fillSoln, const DeformAll_Utility::Options &defOpt, const DeformGraphAll_Utility::SolutionTable& frameSolution );

		// Precomputing cell error
		typedef std::vector<distList> Error_LabelSample;
		typedef std::map<int,Error_LabelSample> FrameError_LabelSample;
		void precomputeSampleError( const ShapeModel& targetShape, const int targetFrameNumber, const DeformAll_Utility::SparseTLArray& allTfList, const DeformAll_Utility::AllSampleGraph& asg, const std::vector<bool>& isOccluded, const DeformGraphAll_Utility::ClosestPointMap& closestPts, const real_type normalWeight, bool accumulateASGOnRef, DeformGraphAll_Utility::Error_LabelSample& result );

		// Pruning graph edges
		void pruneEdges( const PointList& points, const DeformAll_Utility::SparseWeightList& weightList, BoostGraph& graph, const Deform::DeformAll_Utility::SparseTLArray& allTfList, real_type maxStretchFactor, real_type maxEdgeLength );
	}


	class DeformGraphAll : public DeformAlgorithm
	{
	public:
		DeformGraphAll( const DeformAll_Utility::Options& options );
		virtual ~DeformGraphAll();

		virtual void solveDeformation( const SolveOptions &options );					/**< Solve for the deformation */
		virtual void draw( const DrawOptions &options ) const;							/**< Draw the deformed mesh and related visualizations */
		void setOptions( const DeformAll_Utility::Options& options );
		bool init( const std::string& rootDir, const std::string& meshSequence );		/**< Initialize the solver */
		void setBoneColors( const std::vector<RealColor> &colors );				/**< Generate random colors for each bone */
		std::pair<real3_type,real3_type> getDisplayBBox() const;				/**< Get a bounding box of the scene displayed. */
		void interact( const real3_type& pos ) const;
		ShapeModel* getFinalShape() {
			return finalShape;
		}

		// Real-time Inverse Kinematics
		void setupInteractiveIK( ShapeModel*& resultShape, DeformGraphAll_Utility::JointMap*& resultJoints, DeformAll_Utility::RefWeightList*& resultWeights, std::vector<bool>*& resultValid, QMutex*& resultLock );

		// Get a string representing solver status
		QString solverStatusString() const;
		enum SolverStatus {
			SOLVER_IDLE = 0,
			SOLVER_INIT,
			SOLVER_T_STEP,
			SOLVER_W_STEP,
			SOLVER_OCCLUDED_ICP,
			SOLVER_UPDATE_SAMPLES
		};

	private:
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		static const int dim = 3;

		// Data Members
		const int referenceFrameIndex;
		int lastFrame;
		DeformAll_Utility::Options defOpt;
		DeformAll_Utility::ManagedFrameSequence* managedFS;					/**< Internally managed sequence of input range scan data */
		DeformAll_Utility::AllSampleGraph *allSampleGraph;					/**< ASG for solving weights */
		DeformGraphAll_Utility::SolutionTable frameSolution;
		std::vector<int> excludedLabels;									/**< List of labels that are not used */
		SolverStatus solveStatus;											/**< Data members for the current status of the solver */
		DeformGraphAll_Utility::AllClosestPoints sampleClosestPoints;
		DeformGraphAll_Utility::JointMap jointMap;
		DeformGraphAll_Utility::FrameError_LabelSample allFrameError;		/**< Precomputed error per frame/label/sample */
		DeformGraphAll_Utility::Error_LabelSample sumError;					/**< "frameError" summed over all frames */
		std::vector<int> translateToOrig;									/**< To prevent assigning excluded labels in graph cuts (used in DataFn) */
		bool recomputeClosestPtError_AllFrames;
		ShapeModel *finalShape;												/**< The final reconstructed shape */
		DeformAll_Utility::ShapeDeformed *finalShapeDef;					/**< The final shape, deformed */
		DeformAll_Utility::RefWeightList *finalShapeWeights;				/**< Weights on the final shape */


		// Drawing related
		mutable QMutex drawMutex;
		const PointList *tempPts1, *tempPts2;
		const BoostGraph *tempGraph1, *tempGraph2;
		mutable PointList drawVList, drawNList;
		mutable int lastDefFrame_ASG, lastDefFrame_finalShape;
		MatchCandidates *matchCandDisplay;
		int matchCandDisplayIndex;

		// IK-related variables: manipulate finalShape
		ShapeModel *IKMeshDef;
		std::vector<bool> *IKMeshValid;
		QMutex *IKMeshLock;

		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Frames and solutions
		bool isValidFrame( int frame ) const;								/**< Check if the specified frame is valid. */
		DeformGraphAll_Utility::FrameSolution* makeFrameSoln( int frameIndex, bool emptyFrame );
		DeformGraphAll_Utility::FrameSolution* getFrameSoln( int frameIndex ) const;

		// Initialization
		bool initNewFrame( int newFrame );
		void detectOcclusions( int frameIndex, std::vector<int>& occludedBones );
		void fillOccludedTfs_AllFrames( bool forceAll = false );
		void matchOccludedParts( DeformGraphAll_Utility::FrameSolution* newSoln );

		// Picking frames to solve using a sliding window (assumes frame #'s are in increasing order)
		void selectSolver_SlidingWindow( int windowSize );

		// Solving for transformations
		void findClosestPoints_AllFrames( bool forceAll = false );
		real_type findTransformations( int updateCorrespType );
		void countConstraints( const DeformAll_Utility::TfVarTranslate& varTranslate, idxList& distConstraints, idxList& jointConstraints ) const;
		ErrorTerm* makeDistanceTerm( const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s, const real3_type& t, const real_type ptWeight, const real_type plWeight ) const;
		ErrorTerm* makeSmoothnessTerm_BallJoint( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s, const real3_type& t, const DeformAll_Utility::TransformList& tfList ) const;
		void applyTfCorrection( taucsType stepSize, taucsType *correction, const real_type& s, const real3_type& t, const DeformAll_Utility::TfVarTranslate& varTranslate );

		// Debugging transformation solver
		void computeTfResidual( taucsType &totalRes, taucsType &linearRes, taucsType stepSize, taucsType *correction, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s, const real3_type& t, const real_type ptWeight, const real_type plWeight ) const;
		void computeJointResidual( taucsType &totalRes, taucsType &linearRes, taucsType stepSize, taucsType *correction, const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s, const real3_type& t, const DeformAll_Utility::TransformList& tfList ) const;

		void computeTfResidual( taucsType& totalRes, real3_type& avgPt, real_type& avgDist, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type ptWeight, const real_type plWeight ) const;
		real_type computeJointResidual( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const DeformAll_Utility::TransformList& tfList ) const;

		// Estimating joint locations and types
		void estimateJoints( bool printDebug = false );
		void estimateJoints_Motion( DeformGraphAll_Utility::JointMap& newJointMap, const DeformAll_Utility::SparseTLArray& allTfList, bool detectHinges );

		// Solving for weights
		real_type findSampleLabels();
		static real_type GraphCutDataFn( int p, int l, void *data );
		static real_type GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data );
		void precomputeSampleError_AllFrames( bool forceAll = false );				/**< Precomputing the value of E_fit for all samples/frames/labels */
		void replaceLabels( const std::vector<bool>& toReplace, const PointList& samplePoints, std::vector<int>* labelArray = NULL );
		bool discardSmallRegions( const PointList& samplePoints, std::vector<int>* labelArray = NULL );
		bool reuseExtraLabels( const PointList& samplePoints, std::vector<int>& labelArray );

		// Post-processing
		void rebuildASG( real_type maxStretchFactor, real_type maxEdgeLength, bool useAllPoints, bool keepJointEdges );
		void cleanASG();
		void cleanASGWeights( int minSizeThreshold );
		void cleanASGComponents( int minSizeThreshold );
		void filterASGVertices( const std::vector<bool> &filter );
		void postProcess();
		void makeWeightDistanceTerm( const int asgIdx, const ptList& boneRefSrcX, const std::vector<int>& boneCandidates, const real_type ptWeight, const real_type plWeight, std::vector<distList>& result, distList& rhs );
		void processFinalShape();
		void accumulateASG( real_type maxStretchFactor, real_type maxEdgeLength, bool keepJointEdges );		/**< To compare with pure accumulation strategy */

		// Saving/loading solver state to/from file
		void saveSolverState( const char *filename ) const;
		void saveASGState( const DeformAll_Utility::AllSampleGraph *asg, std::ofstream& out ) const;
		void saveFrameState( const DeformGraphAll_Utility::FrameSolution *frameSoln, std::ofstream& out ) const;
		bool loadSolverState( const char *filename, bool graphOnly );
		bool loadASGState( std::ifstream& in, const float version );
		bool loadFrameState( std::ifstream& in, const float version );
		void saveIKState( const char *filename ) const;

		// Drawing
		void drawASG( const PointList& points, const PointList& normals, bool showGraph, bool showTest, bool showRemovedEdges, bool showColors, int showComp, const std::vector<RealColor> &boneColors ) const;
		void drawJoints( const DeformAll_Utility::TransformList& tfList, const DrawOptions& options ) const;
		void drawCorresp( const PointList& vList, int frameNumber, const ShapeModel& frameShape, const DrawOptions& options ) const;
		void drawError( const PointList& vList, const DeformGraphAll_Utility::Error_LabelSample& frameError, const DrawOptions& options ) const;
		void drawSampleError( const PointList& vList, const distList& sampleError, const DrawOptions& options ) const;

		// Solver statistics
		void printSolverStats() const;
		void printJointTransformInfo( std::ostream &out ) const;

		// Real-time Inverse Kinematics
		void solveInteractiveIK();
		taucsType findTransformations_IK( const ShapeModel *shape, ShapeModel *deformed, std::vector<bool>& valid, Scene *scene, const DeformAll_Utility::RefWeightList &weights ) const;

		// Loading Initialization for Display
		bool loadInitialization( int index, const DeformGraphAll_Utility::FrameSolution *lastSoln, const DeformGraphAll_Utility::FrameSolution *newSoln );

		// Computing total value of E_fit
		void findTotal_Efit( real_type &totalDist, real_type &totalCount );

		// Transforming input frames to the reference frame and writing out obj files
		void writeTransformedFrames();

		// Transforming reconstructed DSG to each frame and writing out pts files
		void writeTransformedDSG();
	};
}

namespace BVHLoader
{
	class JointInfo
	{
	public:
		JointInfo( const std::string& name, const int idx ) : name(name), offset(0.f,0.f,0.f), stateIndex(idx)
		{
			XPosIndex = YPosIndex = ZPosIndex = -1;
			XRotIndex = YRotIndex = ZRotIndex = -1;
			XScaleIndex = YScaleIndex = ZScaleIndex = -1;
		}

		std::string name;
		real3_type offset;
		int XPosIndex, YPosIndex, ZPosIndex;
		int XRotIndex, YRotIndex, ZRotIndex;
		int XScaleIndex, YScaleIndex, ZScaleIndex;
		int stateIndex;
		std::string rotationOrder;
	};

	class Joint
	{
	public:
		Joint( std::tr1::shared_ptr<JointInfo> info ) : parent(), info(info)
		{

		}

		static void addChild( std::tr1::shared_ptr<Joint> &sharedThis, std::tr1::shared_ptr<Joint> child )
		{
			sharedThis->children.push_back(child);
			child->parent = sharedThis;
		}

		bool hasParent() const
		{
			std::tr1::shared_ptr<Joint> sharedParent = parent.lock();
			return sharedParent;
		}

		std::tr1::weak_ptr<Joint> getParent()
		{
			return parent;
		}

		int numChildren() const
		{
			return children.size();
		}

		std::tr1::shared_ptr<Joint> getChild(int i)
		{
			return children[i];
		}

		std::tr1::weak_ptr<JointInfo> getInfo()
		{
			return info;
		}

	private:
		std::tr1::weak_ptr<Joint> parent;
		std::tr1::shared_ptr<JointInfo> info;
		std::deque<std::tr1::shared_ptr<Joint> > children;
	};

	class MySkeleton 
	{
		static const long serialVersionUID = 1L;
		Joint root;
		int numJoints;

	public:
		MySkeleton(Joint root, int numJoints)
			: root(root), numJoints(numJoints)
		{
			//load();
		}

		Joint getRootFragile() {
			return root;
		}

		int getNumJoints() {
			return numJoints;
		}
	};

	typedef std::vector<std::tr1::shared_ptr<Deform::DeformAll_Utility::TransformList> > MySkeletonStateList;

	class MySkeletonStateSequence 
	{
	private:
		static const long serialVersionUID = 1L;
		MySkeletonStateList states;
		StringArray jointNames;

	public:
		MySkeletonStateSequence( const MySkeletonStateList& states, const StringArray& jointNames)
			: states(states), jointNames(jointNames)
		{
			// nothing for now!
		}

		int getNumStates() {
			return states.size();
		}

		std::tr1::shared_ptr<Deform::DeformAll_Utility::TransformList> getStateFragile(int i) {
			return states[i];
		}

		StringArray getJointNamesFragile() {
			return jointNames;
		}
	};

	Geometry::Matrix33f getRotMatrixFromRotations( float rx, float ry, float rz, const std::string& rotationOrder );
}

namespace Deform
{
	namespace DeformGraphAll_Utility
	{
		ErrorTerm* makeDistanceTerm_IK( const ShapeModel *shape, const CorrespondenceList &corresp, const DeformAll_Utility::RefWeightList &shapeWeights, const DeformAll_Utility::TransformList &tfList, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s, const real3_type& t );
		ErrorTerm* makeSmoothnessTerm_BallJoint_IK( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s, const real3_type& t, const DeformAll_Utility::TransformList& tfList, const DeformAll_Utility::Options& defOpt, const DeformGraphAll_Utility::JointMap& jointMap );

		std::tr1::shared_ptr<BVHLoader::Joint> loadJointsBVH( const std::string& bvhFilename, BVHLoader::MySkeletonStateList& statesList );
		void populateJointsFromBVH( std::tr1::weak_ptr<BVHLoader::Joint> root, int numTotalBones, const char* masterName, JointMap& joints, std::tr1::shared_ptr<DeformAll_Utility::TransformList> initialState );
		int2_type loadWeightFile( const std::string& weightFilename, DeformAll_Utility::RefWeightList& weights );
	}
}


#endif
