#ifndef __DEFORM_H__
#define __DEFORM_H__

#include <QThread>
#include "types.h"
#include "subject.h"

namespace Deform {

	// Enumerate deformation types.
	enum DeformTypes {
		DEFORM_NONE = 0,
		DEFORM_EMBEDDED,
		DEFORM_WEIGHTED,
		DEFORM_EM,
		DEFORM_ALL,
		DEFORM_GRAPH_ALL
	};

	class DrawOptions;
	class SolveOptions;

	//////////////////////////////////////////////////////////////////////////
	// Abstract class representing deformation algorithms.
	//////////////////////////////////////////////////////////////////////////
	class DeformAlgorithm : public Subject {
	public:
		DeformAlgorithm ( const DeformTypes type ) : deformType(type), iterationFinished(true) { };
		virtual ~DeformAlgorithm() { };

		virtual void solveDeformation( const SolveOptions &options ) = 0;			/**< Solve for the deformation */
		virtual void draw( const DrawOptions &options ) const = 0;					/**< Draw the deformed mesh and related visualizations */

		DeformTypes getType() const {
			return deformType;
		};

		bool finished() const {
			return iterationFinished;
		};

	protected:
		bool iterationFinished;						/**< This variable is set true when the solve iteration has finished */

	private:
		const DeformTypes deformType;				/**< Name of the algorithm (used for printing purposes) */
	};


	/**
	* Passing solver options to the various DeformAlgorithms.
	*/
	class SolveOptions
	{
	public:
		SolveOptions() {
			init();
		};

		int maxItr;				/**< Maximum number of iterations in the solver */
		bool solveIK;			/**< DeformAll SolveIK related parameters */

	private:
		inline void init() {
			maxItr = 0;
			solveIK = false;
		};
	};

	class SolveThread : public QThread
	{
	public:
		SolveThread() : deform(NULL), options() { }
		SolveThread( DeformAlgorithm * const deform , const SolveOptions &options ) : deform(deform), options(options) { }

		// Run the solver thread
		virtual void run() {
			// Set termination to be enabled
			this->setTerminationEnabled(true);

			if (deform) {
				deform->solveDeformation(options);
			}

			// Exit the solver thread
			this->exit();
		}

	private:
		DeformAlgorithm * const deform;		/**< Algorithm to execute (managed externally) */
		SolveOptions options;				/**< Solver options */
	};

	/**
	* A class useful for passing drawing options for the various DeformAlgorithm classes.
	*/
	class DrawOptions
	{
	public:
		bool showSourceModel;
		bool showTargetModel;
		bool showDeformedModel;
		bool showSourceCorresp;
		bool showTargetCorresp;
		bool showDeformedCorresp;
		bool showShapeWeights;
		bool showAllFrames;
		bool showGraph;
		bool showEffRad;
		bool renderTransparent;
		int showComponent;
		bool showGraphDeformed;
		bool showWireframe;
		bool showGraphFaces;
		int showExtendedGrid;
		float openGLPointSize;
		RealColor sourceModelColor;
		RealColor targetModelColor;
		RealColor deformedModelColor;
		RealColor pointColor;
		bool showDeformedModelPoints;
		int showErrorComponent;
		int displayFrame;
		int displayTargettingFrame;
		bool ignoreDisplayInterface;
		bool showSourceDeformed;
		bool showInvalidVertices;
		bool showSampledPoints;
		bool showExternalMeshDeformed;
		bool showJoints;
		float sampleRadius;
		int curvDisplayMode;
		int curvDisplayScale;
		std::vector<RealColor>* boneColors;
		bool showInitSrc, showInitResult;

		// Debugging point / cell affecting relationships
		bool showCellAffectedPoints;
		bool showPointAffectingCells;
		int cellPickIndex;
		int pointPickIndex;

		// Debugging cell error
		bool showCellError;
		bool showSampleError;		// Show per-sample error
	};

}

#endif
