#include "ShapeAnimation.h"
#include "FileSequence.h"
#include "util.h"
#include "scene.h"

using namespace std;
using namespace Geometry;

ShapeAnimation::ShapeAnimation()
{

}

ShapeAnimation::~ShapeAnimation()
{
	clear();
}


void ShapeAnimation::clear()
{
	// Deallocate the memory for each shape model
	for (index_type i = 0, i_end = data.size(); i < i_end; i++)
	{
		delete data[i];
	}

	data.clear();
}

size_type ShapeAnimation::numFrames() const
{
	return data.size();
}



void ShapeAnimation::load( const QString &baseDir, const FileSequence &fs, const Scene &scene )
{
	// Clear existing animation
	clear();

	// Load and populate the new animation
	cout << "Loading frames [" << fs.startFrame << "," << fs.endFrame << "] skip " << fs.skipFrame << "..." << endl;

	for (int i = fs.startFrame; i <= fs.endFrame; i += fs.skipFrame) {
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "Frame " << i << ":" << endl;
		add(baseDir, fs[i], scene, scene.autoLoadFrameData);
	}
}

void ShapeAnimation::add( const QString &baseDir, const QString &filename, const Scene &scene, int processFrames )
{
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Loading new shape..." << endl;

	// 1. Construct and load the appropriate shape
	ShapeModel *shape = loadShape(qPrintable(baseDir), qPrintable(filename), scene, processFrames);

	// 2. Add it to the list of shape frames
	data.push_back(shape);
}

void ShapeAnimation::add( ShapeModel *shape )
{
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Adding new shape..." << endl;

	data.push_back(shape);

	shape->printInfo();
}

void ShapeAnimation::replace( ShapeModel *shape, int frameIndex )
{
	if ((frameIndex < 0) || (frameIndex >= data.size())) {
		cerr << "Error in ShapeAnimation::frame(): frame # out of range.";
		return;
	}

	ShapeModel *old = data[frameIndex];
	data[frameIndex] = shape;

	// Delete the old frame
	if (old != shape) {
		delete old;
	}
}

/**
 * Draw a bounding box.
 */
void ShapeAnimation::drawBox( const ShapeModel &model ) const
{
	// Disable lighting but keep track of the previous state
	GLboolean lighting;
	glGetBooleanv(GL_LIGHTING, &lighting);
	glDisable(GL_LIGHTING);

	glLineWidth(1.0);
	glColor3f(0.7f,0.7f,0.7f);
	drawBoxOpenGL(model.box_min, model.box_max);

	// Draw center and min of the bounding box
	glPointSize(5.0);
	glBegin(GL_POINTS);
		//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
		glVertex3f(model.box_min[0], model.box_min[1], model.box_min[2]);
	glEnd();

	// Restore previous state
	if (lighting) glEnable(GL_LIGHTING);
}

/**
 * Draw the current frame.
 */
void ShapeAnimation::draw( index_type frame, bool showBox/* = false*/ ) const {
	if ((frame >= 0) && (frame < data.size())) {
		data[frame]->draw();

		// Draw bounding box
		if (showBox) {
			drawBox(*data[frame]);
		}
	}
}


ShapeModel* ShapeAnimation::operator[]( index_type frame )
{
	if ((frame < 0) || (frame >= data.size())) {
		cerr << "Error in ShapeAnimation::frame(): frame # out of range.";
		return NULL;
	}

	return data[frame];
}


const ShapeModel* ShapeAnimation::operator[]( index_type frame ) const
{
	if ((frame < 0) || (frame >= data.size())) {
		cerr << "Error in ShapeAnimation::frame(): frame # out of range.";
		return NULL;
	}

	return data[frame];
}

ShapeModel* ShapeAnimation::operator[]( const QString &filename )
{
	// Search the source names for the corresponding frame.
	for (index_type i = 0; i < numFrames(); i++) {
		if ((*data[i]).geometryFile().endsWith(filename)) {
			return data[i];
		}
	}

	return NULL;
}

int ShapeAnimation::frameNumber( const ShapeModel* model ) const
{
	if (model) {
		for (index_type i = 0; i < numFrames(); i++) {
			if (data[i] == model) {
				return i;
			}
		}
	}

	return -1;
}

bool ShapeAnimation::validFrame( int frameIndex ) const
{
	return (frameIndex >= 0) && (frameIndex < numFrames());
}

