#ifndef __SPIN_IMAGE_H__
#define __SPIN_IMAGE_H__

#include "types.h"

/**
 * Apply the spin map to a point.
 */
real2_type SpinMap( const real3_type &p, const real3_type &n, const real3_type &x );
real_type GeometricConsistency( const real3_type &p_s1, const real3_type &n_s1,
							    const real3_type &p_s2, const real3_type &n_s2,
								const real3_type &p_t1, const real3_type &n_t1,
								const real3_type &p_t2, const real3_type &n_t2 );

class ShapeModel;
class CorrespondenceList;
real_type GeometricConsistency_Diameter( ShapeModel &source, ShapeModel &target, CorrespondenceList &cl );


class SpinImage
{
public:

	bool accumulate( real2_type coord, real_type binSize );			/**< Accumulate the given value in the spin image. */
	int countBins() const;											/**< Count the number of non-empty bins. */

	void resize( int width );								/**< Allocate the space for the spin image. */
	void resize( int width, real_type val );				/**< Allocate the space for the spin image. */
	void reset();											/**< Reset all the values of the spin image. */
	bool invalid() const;									/**< Check if there's no data in this spin image. */

	int width() const {
		return data.size();
	}

	real_type& operator()( int i, int j ) {
		return data[i][j];
	}
	const real_type& operator()( int i, int j ) const {
		return data[i][j];
	}
	void swap( SpinImage& si ) {
		data.swap(si.data);
	}

private:
	std::vector< std::vector<real_type> > data;
};


real_type ComputeSimilarity( const SpinImage &P, const SpinImage &Q, real_type lambda );


class SpinImageStack
{
public:
	SpinImageStack() : binSize(0.0), imageWidth(0), supportAngle(0.0), lambda(0.0) { };

	size_type size() const {
		return data.size();
	}

	void resize( size_type n ) {
		data.resize(n);
	}

	void clear() {
		//data.clear();
		// Complete clear out the memory contents
		std::vector<SpinImage>().swap(data);
	}

	void add( const SpinImage &si ) {
		data.push_back(si);
	}

	SpinImage& operator[]( int i ) {
		return data[i];
	}

	const SpinImage& operator[]( int i ) const {
		return data[i];
	}

	void querySimilar( const SpinImage& query, idxList &result, const char *scoreOutput = NULL ) const;
	void querySimilar( const SpinImage& query, idxList &index, distList &similarityScores ) const;

	/**
	* Queries within a subset of the spin image stack.
	* @param[in] query The spin image to match.
	* @param[in] subset Indices <I>into the stack</I> selecting the subset of the spin image stack.
	* @param[out] result List of indices <I>into the subset</I> in descending order of similarity to the query spin image.
	* @param[out] similarityScores Similarity scores corresponding to the resulting indices.
	*/
	void querySubsetSimilar( const SpinImage& query, const idxList &subset, idxList &result, distList &similarityScores ) const;

	void copyParams( const SpinImageStack& sis );

	real_type binSize;
	int imageWidth;
	real_type supportAngle;
	real_type lambda;				/**< Half of median bin size */

	void saveStack( std::ofstream& out ) const;
	void loadStack( std::ifstream& in );

private:
	std::vector<SpinImage> data;
};

#endif
