#include "Correspondence.h"
#include "FindCorrespondence.h"
#include "util.h"

using namespace std;

real3_type CorrespondenceList::sourcePos( index_type i ) const
{
	return sourcePoints[i];
}

real3_type CorrespondenceList::targetPos( index_type i ) const
{
	return targetPoints[i];
}

bool CorrespondenceList::valid( index_type i ) const
{
	return (weight[i] > 0.0);
}

real_type CorrespondenceList::getWeight( index_type i ) const
{
	return weight[i];
}

size_type CorrespondenceList::size() const
{
	return sourcePoints.size();
}

void CorrespondenceList::clear()
{
	sourcePoints.clear();
	targetPoints.clear();
	sourceIndex.clear();
	targetIndex.clear();
	weight.clear();
}

void CorrespondenceList::add( real3_type sourcePt, real3_type targetPt, index_type sourceIndex /*= 0*/, index_type targetIndex /*= 0*/ )
{
	this->sourcePoints.add(sourcePt);
	this->targetPoints.add(targetPt);
	this->sourceIndex.push_back(sourceIndex);
	this->targetIndex.push_back(targetIndex);
	this->weight.push_back(1.0);
}


void CorrespondenceList::resize( size_type size )
{
	sourcePoints.resize(size);
	targetPoints.resize(size);
	sourceIndex.resize(size);
	targetIndex.resize(size);
	weight.resize(size,1.0);
}

void CorrespondenceList::reserve( size_type size )
{
	sourcePoints.reserve(size);
	targetPoints.reserve(size);
	sourceIndex.reserve(size);
	targetIndex.reserve(size);
	weight.reserve(size);
}

void CorrespondenceList::setSourcePos( index_type i, real3_type pt )
{
	sourcePoints[i] = pt;
}

void CorrespondenceList::setTargetPos( index_type i, real3_type pt )
{
	targetPoints[i] = pt;
}

void CorrespondenceList::remove( idxList list )
{
	sourcePoints.removePoints(list);
	targetPoints.removePoints(list);

	// Sort the list according to the index number
	std::sort(list.begin(), list.end());

	// Traversing the list in reverse, remove elements from the array
	for (int i = list.size()-1; i >= 0; i--) {
		// Remove each element from the list
		index_type c = list[i];

		// Find the iterator at c
		idxList::iterator itr = sourceIndex.begin();
		sourceIndex.erase(itr+c);
		itr = targetIndex.begin();
		targetIndex.erase(itr+c);
		std::vector<real_type>::iterator witr = weight.begin();
		weight.erase(witr + c);
	}
}

bool CorrespondenceList::load( std::string filename )
{
	QFile file(filename.c_str());
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		cout << "Warning in CorrespondenceList::load(): Cannot open file " << filename.c_str() << "!" << endl;
		return false;
	}

	// Clear the current list
	clear();
	cout << "*** Loading correspondences from: " << filename << "...";

	real3_type source, target;
	QTextStream in(&file);
	while (!in.atEnd()) {
		QString line = in.readLine();
		sscanf(qPrintable(line), "[%f,%f,%f] <-> [%f,%f,%f]",
			   &source[0], &source[1], &source[2],
			   &target[0], &target[1], &target[2]);

		add(source, target);
	}

	file.close();
	cout << "done!" << endl;
	return true;
}

bool CorrespondenceList::save( std::string filename ) const
{
	// Save the information to a file
	QFile file(filename.c_str());
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		cout << "Warning in CorrespondenceList::save(): Cannot open save file " << filename.c_str() << "!" << endl;
		return false;
	}

	cout << "Saving correspondences to: " << filename << "...";

	QTextStream out(&file);
	for (index_type i = 0; i < size(); i++) {
		out << sourcePos(i) << " <-> " << targetPos(i) << "\n";
	}

	file.close();
	cout << "done!" << endl;
	return true;
}

SE3 CorrespondenceList::registerICP( ShapeModel &target, const SE3 &tf )
{
	// Refine the correspondence using ICP
	if (size() < 3) {
		return tf;
	}

	//////////////////////////////////////////////////////////////////////////
	// The ICP algorithm
	SE3 tfp(tf);
	PointList closestPoints;
	int itr = 0;
	const int maxItr = 100;
	bool converged = false;

	// Initialization
	closestPoints.resize(size());
	for (index_type i = 0; i < size(); i++) {
		targetPoints[i] = tfp*sourcePoints[i];
	}

	//////////////////////////////////////////////////////////////////////////
	// ICP loop
	real_type lastRes = REAL_TYPE_MAX;
	while ((itr < maxItr) && !converged) {
		// Step 1. Compute the closest points
		for (index_type i = 0; i < size(); i++) {
			real_type temp;

			//target.pointSqDistance(targetPoints[i], closestPoints[i]);
			target.hybridClosestPoint(targetPoints[i], closestPoints[i], temp);
		}

		// Step 2. Compute the registration (point-to-point metric)
		tfp = registerQuat(sourcePoints, closestPoints);

		// Step 3. Apply the registration and find the residual
		real_type res = 0.0;
		for (index_type i = 0; i < size(); i++) {
			real3_type temp = tfp*sourcePoints[i];
			res += (targetPoints[i] - temp).square_norm();
			targetPoints[i] = temp;
		}

		// Step 4. Check termination criteria
		if (abs(lastRes - res) < REAL_TYPE_EPSILON) {
			converged = true;
		}

		// Update iteration info
		lastRes = res;
		itr++;
	}

	// Return the final transformation
	return tfp;
}

void CorrespondenceList::loadBinary( std::ifstream &in )
{
	// Read list of correspondences
	unsigned int numMatched;
	in.read((char*) &numMatched, sizeof(unsigned int));
	resize(numMatched);

	for (index_type j = 0; j < numMatched; j++) {
		// Store source index, source position, target index, target position
		unsigned int index;
		real3_type pos;

		in.read((char*) &index, sizeof(unsigned int));
		readBinary(in, pos);
		sourceIndex[j] = index;
		sourcePoints[j] = pos;

		in.read((char*) &index, sizeof(unsigned int));
		readBinary(in, pos);
		targetIndex[j] = index;
		targetPoints[j] = pos;
	}
}

void CorrespondenceList::saveBinary( std::ofstream &out ) const
{
	// Store list of correspondences
	unsigned int numMatched = size();
	out.write((char*) &numMatched, sizeof(unsigned int));

	for (index_type j = 0; j < numMatched; j++) {
		// Store source index, source position, target index, target position
		unsigned int index;
		real3_type pos;

		index = sourceIndex[j];
		pos = sourcePoints[j];
		out.write((char*) &index, sizeof(unsigned int));
		writeBinary(out, pos);

		index = targetIndex[j];
		pos = targetPoints[j];
		out.write((char*) &index, sizeof(unsigned int));
		writeBinary(out, pos);
	}
}
