#include "PointList.h"
#include <SylvainTools/simple_mesh.h>
#include "util.h"
#include <ctime>

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;
using namespace Geometry;

PointList::PointList() :
kdTree(NULL), dataPts(NULL), queryEpsilon(0), dirtyBit(false)
{

}

PointList::PointList( size_type initialSize ) :
data(initialSize), kdTree(NULL), dataPts(NULL), queryEpsilon(0), dirtyBit(false)
{

}

PointList::PointList( const PointList& rhs ) :
data(rhs.data), kdTree(NULL), dataPts(NULL), queryEpsilon(rhs.queryEpsilon), dirtyBit(false)
{
}

PointList::~PointList()
{
	clear();
}

const PointList& PointList::operator=( const PointList & rhs )
{
	if ( this != &rhs ) {
		clear();			/**< This also clears the ANN data structure, and thus the dirty bit */

		// Copy values
		this->queryEpsilon = rhs.queryEpsilon;
		this->data = rhs.data;
	}
	return *this;
}

void PointList::resize( size_type size )
{
	dirtyBit = true;
	data.resize(size);
}

void PointList::reserve( size_type size )
{
	data.reserve(size);
}

void PointList::add( real3_type pt )
{
	dirtyBit = true;
	data.push_back(pt);
}

void PointList::add( const ptList& pts )
{
	dirtyBit = true;

	// Resize the list of points, preserving the existing points
	size_type currentSize = size();
	size_type addSize = pts.size();
	data.resize(currentSize + addSize);

	// Populate the new points
	for (index_type i = 0; i < addSize; ++i) {
		data[currentSize + i] = pts[i];
	}
}

void PointList::add( const PointList& pts )
{
	add(pts.data);
}

void PointList::clear()
{
	clearANNData();
	data.clear();
}

void PointList::removePoints( idxList list )
{
	// Mark list as dirty first
	dirtyBit = true;

	// Sort the list according to the index number
	std::sort(list.begin(), list.end());

	// Traversing the list in reverse, remove elements from the array
	for (int i = list.size()-1; i >= 0; i--) {
		// Remove each element from the list
		index_type c = list[i];

		// Find the iterator at c
		ptList::iterator itr = data.begin();
		data.erase(itr+c);
	}
}

void PointList::queryKNN( real3_type q, size_type k, idxList& nearest, distList& dist, bool persistent /*= false*/ ) const
{
	ANNquery(q, k, -1.0, nearest, dist, persistent);
}

void PointList::queryRange( real3_type q, float radSq, idxList& inRange, distList& dist, bool persistent/* = false*/ ) const
{
	ANNquery(q, 0, radSq, inRange, dist, persistent);
}

void PointList::setANNEpsilon( double eps )
{
	if (queryEpsilon != eps) {
		dirtyBit = true;
		queryEpsilon = eps;
	}
}

void PointList::clearANNData() const
{
	if (dataPts) {
		annDeallocPts(dataPts);		/**< Also sets it to NULL */
	}
	if (kdTree) {
		delete kdTree;
		kdTree = NULL;
	}
	dirtyBit = false;
	//annClose();				// done with ANN
}

void PointList::ANNquery( real3_type q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent ) const
{
	unsigned int nPts = data.size();		// actual number of data points
	ANNpoint queryPt;						// query point
	ANNidxArray nnIdx = NULL;				// near neighbor indices
	ANNdistArray dists = NULL;				// near neighbor distances

	//*******************************************************//
	// TESTING: measure time spent on ANN queries
	MEASURE_PERF_NARG(TIMER_POINTLIST_ANN_QUERY,AddStart);
	//*******************************************************//

	// Reconstruct the k-d tree used for querying the points:
	// 1. If there is no existing persistent data structure available,
	// 2. Or the list of points has been modified in any way (keep track of the dirty bit).
	if (dirtyBit || !persistent || !dataPts || !kdTree) {

		//cout << " --> Creating query data structure..." << endl;
		//cout << " *C";

		clearANNData();				/**< This will reset the dirty bit as well */

		//////////////////////////////////////////////////////////////////////////
		// Populate data points
		dataPts = annAllocPts(nPts, dim);

		for (ptList::size_type i = 0; i < nPts; ++i) {
			for (int d = 0; d < dim; ++d) {
				dataPts[i][d] = data[i][d];

				// Check bad floating point values
				if (!_finite(data[i][d])) {
					cerr << "Error in PointList::ANNquery(): Non-finite floating point value at index " << i << "!" << endl;
					// TODO: Deal with this error in a better way
					dataPts[i][d] = 0;
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		kdTree = new ANNkd_tree(				// build search structure
								dataPts,		// the data points
								nPts,			// number of points
								dim);			// dimension of space
	}


	//////////////////////////////////////////////////////////////////////////
	// Convert they query point into an ANN-familiar format.
	queryPt = annAllocPt(dim);
	for (int d = 0; d < dim; ++d) {			// Copy the query point
		queryPt[d] = q[d];
	}


	//////////////////////////////////////////////////////////////////////////
	// CORE
	//////////////////////////////////////////////////////////////////////////
	if (radSq < 0) {
		//cout << " --> Performing k-nearest query..." << endl;

		// Perform k-nearest-neighbor search
		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		kdTree->annkSearch(
							queryPt,			// query point
							k,					// number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}
	else {
		//cout << " --> Performing range query..." << endl;

		// Perform range search; initially find the number of points within the range
		k = kdTree->annkFRSearch(
									queryPt,			// query point
									radSq,				// radius of range query
									0,					// initially 0
									NULL,				// return no nearest neighbors
									NULL,				// return no distances
									queryEpsilon);		// error bound

		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		// Repeat the search with the *correct* number of neighbors
		kdTree->annkFRSearch(
							queryPt,			// query point
							radSq,				// radius of range query
							k,					// *correct* number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}

	//////////////////////////////////////////////////////////////////////////
	// Copy result data
	//////////////////////////////////////////////////////////////////////////
	indices.resize(k);
	distances.resize(k);
	for (int i = 0; i < k; ++i) {
		indices[i] = nnIdx[i];
		distances[i] = sqrt(dists[i]);
	}


	//*******************************************************//
	// TESTING: measure time spent on ANN queries
	MEASURE_PERF_NARG(TIMER_POINTLIST_ANN_QUERY,Stop);
	//*******************************************************//

	//////////////////////////////////////////////////////////////////////////
	// Clean things up
	//////////////////////////////////////////////////////////////////////////
	delete [] nnIdx;
	delete [] dists;
	annDeallocPt(queryPt);
}

void PointList::swap( PointList &list )
{
	// Swap the array
	data.swap(list.data);

	// Swap ANN-related structures
	double temp = queryEpsilon;
	queryEpsilon = list.queryEpsilon;
	list.queryEpsilon = temp;

	ANNpointArray temp2 = dataPts;
	dataPts = list.dataPts;
	list.dataPts = temp2;

	ANNkd_tree *temp3 = kdTree;
	kdTree = list.kdTree;
	list.kdTree = temp3;

	bool temp4 = dirtyBit;
	dirtyBit = list.dirtyBit;
	list.dirtyBit = temp4;
}

const real_type* PointList::getArrayPtr() const
{
	return ((const real_type*)&(data[0]));
}

void PointList_Utility::generateRandom( PointList &list, size_type size )
{
	list.resize(size);

	// Use SFMT to generate random points in the unit interval.
	for (index_type i = 0; i < list.size(); i++) {
		for (int d = 0; d < DIMENSION; d++) {
			list[i][d] = genrand_real2();
		}
	}
}

void PointList_Utility::measureExtent( const PointList &list, real3_type &center, real_type &radius )
{
	const size_type n = list.size();

	// Measure the extent of the source points
	center[0] = center[1] = center[2] = 0;
	for (index_type i = 0; i < n; i++) {
		center += list[i];
	}
	center = center / ((real_type) n);

	radius = -REAL_TYPE_MAX;
	for (index_type i = 0; i < n; i++) {
		real_type distFromCenter = (list[i] - center).norm();

		if (distFromCenter > radius) {
			radius = distFromCenter;
		}
	}
}

void PointList_Utility::draw( const PointList &list, int type, real_type radius /*= 1 */ )
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < list.size(); i++) {
			drawSphere(list[i], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < list.size(); i++) {
			glVertex3f(list[i][0], list[i][1], list[i][2]);
		}
		glEnd();
	}
}

void PointList_Utility::draw( const PointList &list, int type, const idxList &subset, real_type radius /*= 1 */ )
{
	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			drawSphere(list[subset[i]], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			glVertex3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
		}
		glEnd();
	}
}

void PointList_Utility::drawWithNames( const PointList &list, int type, real_type radius /*= 1 */ )
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < list.size(); i++) {
			glPushName(i);
			drawSphere(list[i], radius);
			glPopName();
		}
	}
	else if (type == POINT_SQUARES) {
		for (index_type i = 0; i < list.size(); i++) {
			glPushName(i);
			glRasterPos3f(list[i][0], list[i][1], list[i][2]);
			glPopName();
		}
	}
}

void PointList_Utility::drawSubsetWithNames( const PointList &list, int type, const idxList &subset, real_type radius /*= 1 */ )
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			glPushName(subset[i]);
			drawSphere(list[subset[i]], radius);
			glPopName();
		}
	}
	else if (type == POINT_SQUARES) {
		for (index_type i = 0; i < subset.size(); i++) {
			glPushName(subset[i]);
			glRasterPos3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
			glPopName();
		}
	}
}

void PointList_Utility::drawWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, real_type radius /*= 1 */ )
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < list.size(); i++) {
			if (weights[i] > 0.0) {
				drawSphere(list[i], radius);
			}
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < list.size(); i++) {
			if (weights[i] > 0.0) {
				glVertex3f(list[i][0], list[i][1], list[i][2]);
			}
		}
		glEnd();
	}
}

void PointList_Utility::drawWithWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &weights, real_type radius /*= 1 */ )
{
	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			if (weights[subset[i]] > 0.0) {
				drawSphere(list[subset[i]], radius);
			}
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			if (weights[subset[i]] > 0.0) {
				glVertex3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
			}
		}
		glEnd();
	}
}

void PointList_Utility::drawWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, const RealColor &minColor, const RealColor &maxColor, real_type radius /*= 1 */ )
{
	RealColor color;

	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < list.size(); i++) {
			color = weights[i] * maxColor + (1.0 - weights[i]) * minColor;
			setOpenGLMaterialColor(color);
			drawSphere(list[i], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < list.size(); i++) {
			color = weights[i] * maxColor + (1.0 - weights[i]) * minColor;
			setOpenGLColor(color);
			glVertex3f(list[i][0], list[i][1], list[i][2]);
		}
		glEnd();
	}
}

void PointList_Utility::drawWithWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &weights, const RealColor &minColor, const RealColor &maxColor, real_type radius /*= 1 */ )
{
	RealColor color;

	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			color = weights[subset[i]] * maxColor + (1.0 - weights[subset[i]]) * minColor;
			setOpenGLMaterialColor(color);
			drawSphere(list[subset[i]], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			color = weights[subset[i]] * maxColor + (1.0 - weights[subset[i]]) * minColor;
			setOpenGLColor(color);
			glVertex3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
		}
		glEnd();
	}
}

void PointList_Utility::drawWithSubsetWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &subsetWeights, real_type radius /*= 1 */ )
{
	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			if (subsetWeights[i] > 0.0) {
				drawSphere(list[subset[i]], radius);
			}
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			if (subsetWeights[i] > 0.0) {
				glVertex3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
			}
		}
		glEnd();
	}
}

void PointList_Utility::drawWithSubsetWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &subsetWeights, const RealColor &minColor, const RealColor &maxColor, real_type radius /*= 1 */ )
{
	RealColor color;

	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			color = subsetWeights[i] * maxColor + (1.0 - subsetWeights[i]) * minColor;
			setOpenGLMaterialColor(color);
			drawSphere(list[subset[i]], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			color = subsetWeights[i] * maxColor + (1.0 - subsetWeights[i]) * minColor;
			setOpenGLColor(color);
			glVertex3f(list[subset[i]][0], list[subset[i]][1], list[subset[i]][2]);
		}
		glEnd();
	}
}

void PointList_Utility::drawWithNamesWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, real_type radius /*= 1 */ )
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < list.size(); i++) {
			if (weights[i] > 0.0) {
				glPushName(i);
				drawSphere(list[i], radius);
				glPopName();
			}
		}
	}
	else if (type == POINT_SQUARES) {
		for (index_type i = 0; i < list.size(); i++) {
			if (weights[i] > 0.0) {
				glPushName(i);
				glRasterPos3f(list[i][0], list[i][1], list[i][2]);
				glPopName();
			}
		}
	}
}

