#ifndef __POINT_LIST_H__
#define __POINT_LIST_H__

#include <ANN/ANN.h>
#include "types.h"

#define DIMENSION 3

enum POINT_DISPLAY {
	POINT_SPHERES = 1,
	POINT_SQUARES
};

/**
* A list of points in three dimensions, with k-nearest neighbor and range search capabilities.
* Uses the ANN library to find nearest neighbors.
*/
class PointList
{
public:
	PointList();											/**< Constructor */
	PointList( size_type initialSize );						/**< Constructor with initial size */
	~PointList();											/**< Destructor */
	PointList(const PointList& rhs);						/**< Copy Constructor */
	const PointList& operator=(const PointList & rhs);		/**< Assignment */


	//////////////////////////////////////////////////////////////////////////
	// Access
	//////////////////////////////////////////////////////////////////////////
	size_type size() const;									/**< Returns the number of points contained in this list. */
	real3_type operator[](index_type i) const;				/**< Return a constant reference to the i^th point in the list. */
	const real_type* getArrayPtr() const;

	// Perform a k-nearest-neighbor query or a range query.  Refer to the function ANNquery for details.
	void queryKNN(real3_type q, size_type k, idxList& nearest, distList& dist, bool persistent = false) const;
	void queryRange(real3_type q, float radSq, idxList& inRange, distList& dist, bool persistent = false) const;


	//////////////////////////////////////////////////////////////////////////
	// Modify
	//////////////////////////////////////////////////////////////////////////
	real3_type& operator[](index_type i);					/**< Return a modifiable reference to the i^th point in the list. */
	void resize( size_type n );								/**< Resize the array to contain n points. */
	void reserve( size_type n );							/**< Reserve the array space to contain n points. */
	void add( real3_type pt );								/**< Add a point to the list. */
	void add( const ptList& pts );							/**< Add a list of points. */
	void add( const PointList& pts );						/**< Add a list of points. */
	void removePoints( idxList list );						/**< Remove points from the list. */
	void clear();											/**< Clear the list of points. */
	void setANNEpsilon( double eps );						/**< Set the ANN approximation value for KNN and range queries. */
	void swap( PointList &list );							/**< Swap the contents with another list */


private:
	static const int dim = DIMENSION;			/**< Dimension */
	ptList data;								/**< List of points */

	//////////////////////////////////////////////////////////////////////////
	// ANN-related functionality
	//////////////////////////////////////////////////////////////////////////
	double queryEpsilon;						/**< Error bound for approximate nearest neighbor search. */
	mutable ANNpointArray dataPts;				/**< Persistent data array for multiple queries. */
	mutable ANNkd_tree* kdTree;					/**< Persistent search structure for multiple queries. */
	mutable bool dirtyBit;						/**< Valid bit to test whether the ANN data structures are still OK to use. */
	void clearANNData() const;					/**< Clear the persistent ANN data structures, if they exist. */

	/**
	* Perform either a k-nearest-neighbor query or a range query depending on the input parameters.
	* To perform a k-nearest neighbor query, set "radSq" to a negative number
    * Setting "radSq" to a positive number retrieves all points within distance sqrt(radSq) of the query point.
	* @param q The query point.
	* @param k The number of nearest neighbors to retrieve.
	* @param radSq The squared radius of the range query to perform.
	* @param indices The indices of the retrieved points.
	* @param distances The distance to the retrieved points.
	* @param persistent Tell the function whether to keep a persistent data structure for the query.
	*/
	void ANNquery( real3_type q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent ) const;
};


inline size_type PointList::size() const
{
	return data.size();
}

inline real3_type PointList::operator[]( index_type i ) const
{
	return data[i];
}

inline real3_type& PointList::operator[]( index_type i )
{
	dirtyBit = true;
	return data[i];
}



//////////////////////////////////////////////////////////////////////////
// Utility functions for the PointList class.
//////////////////////////////////////////////////////////////////////////
namespace PointList_Utility
{
	// Utility functions
	void generateRandom( PointList &list, size_type size );									/**< Generate random points in the unit interval. */
	void measureExtent( const PointList &list, real3_type &center, real_type &radius );		/**< Measure the maximum distance from the centroid of the point list. */

	/**
	* Draw the points as spheres in OpenGL.
	*/
	void draw( const PointList &list, int type, real_type radius = 1 );
	void draw( const PointList &list, int type, const idxList &subset, real_type radius = 1 );
	void drawWithNames( const PointList &list, int type, real_type radius = 1 );
	void drawWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, real_type radius = 1 );
	void drawWithWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &weights, real_type radius = 1 );
	void drawWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, const RealColor &minColor, const RealColor &maxColor, real_type radius = 1 );
	void drawWithWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &weights, const RealColor &minColor, const RealColor &maxColor, real_type radius = 1 );
	void drawWithSubsetWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &subsetWeights, real_type radius = 1 );
	void drawWithSubsetWeight( const PointList &list, int type, const idxList &subset, const std::vector<real_type> &subsetWeights, const RealColor &minColor, const RealColor &maxColor, real_type radius = 1 );
	void drawWithNamesWithWeight( const PointList &list, int type, const std::vector<real_type> &weights, real_type radius = 1 );
	void drawSubsetWithNames( const PointList &list, int type, const idxList &subset, real_type radius = 1 );
}

#endif
