#include "PointListN.h"
#include <ctime>
#include "util.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;
using namespace Geometry;

PointListN::PointListN( int dim ) :
dim(dim), kdTree(NULL), dataPts(NULL), queryEpsilon(0), dirtyBit(false)
{

}

PointListN::PointListN( const PointListN& rhs ) :
dim(rhs.dim), queryEpsilon(rhs.queryEpsilon), data(rhs.data), kdTree(NULL), dataPts(NULL), dirtyBit(false)
{

}

PointListN::~PointListN()
{
	clear();
}

const PointListN& PointListN::operator=( const PointListN & rhs )
{
	if ( this != &rhs ) {
		clear();

		// copy values
		this->queryEpsilon = rhs.queryEpsilon;
		this->data = rhs.data;
	}
	return *this;
}

void PointListN::resize( size_type n )
{
	dirtyBit = true;
	data.resize(n, data_type(dim));
}

void PointListN::reserve( size_type size )
{
	data.reserve(size);
}

void PointListN::add( const data_type& pt )
{
	dirtyBit = true;
	data.push_back(pt);
}

void PointListN::clear()
{
	clearANNData();
	data.clear();
}

const std::vector< data_type >* PointListN::getData() const
{
	return &data;
}

void PointListN::setPts( const std::vector<data_type> *pts )
{
	clear();
	reserve(pts->size());
	for (size_type i = 0; i < pts->size(); i++) {
		data.push_back((*pts)[i]);
	}
}

void PointListN::removePoints( idxList list )
{
	// Mark list as dirty first
	dirtyBit = true;

	// Sort the list according to the index number
	std::sort(list.begin(), list.end());

	// Traversing the list in reverse, remove elements from the array
	for (int i = list.size()-1; i >= 0; i--) {
		// Remove each element from the list
		index_type c = list[i];

		// Find the iterator at c
		std::vector<data_type>::iterator itr = data.begin();
		data.erase(itr+c);
	}
}

void PointListN::queryKNN( const data_type& q, size_type k, idxList& nearest, distList& dist, bool persistent /*= false*/ ) const
{
	ANNquery(q, k, -1.0, nearest, dist, persistent);
}

void PointListN::queryRange( const data_type& q, float radSq, idxList& inRange, distList& dist, bool persistent/* = false*/ ) const
{
	ANNquery(q, 0, radSq, inRange, dist, persistent);
}

void PointListN::setANNEpsilon( double eps )
{
	if (queryEpsilon != eps) {
		dirtyBit = true;
		queryEpsilon = eps;
	}
}

void PointListN::clearANNData() const
{
	if (dataPts) {
		annDeallocPts(dataPts);		/**< Also sets it to NULL */
	}
	if (kdTree) {
		delete kdTree;
		kdTree = NULL;
	}
	dirtyBit = false;
	//annClose();				// done with ANN
}

void PointListN::ANNquery( const data_type& q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent ) const
{
	unsigned int nPts = data.size();		// actual number of data points
	static ANNpoint queryPt = NULL;			// query point
	ANNidxArray nnIdx = NULL;				// near neighbor indices
	ANNdistArray dists = NULL;				// near neighbor distances

	//*******************************************************//
	// TESTING: measure time spent on ANN queries
	MEASURE_PERF_NARG(TIMER_POINTLIST_N_ANN_QUERY,AddStart);
	//*******************************************************//

	// Reconstruct the k-d tree used for querying the points:
	// 1. If there is no existing persistent data structure available,
	// 2. Or the list of points has been modified in any way (keep track of the dirty bit).
	if (dirtyBit || !persistent || !dataPts || !kdTree) {

		//cout << " --> Creating query data structure..." << endl;
		//cout << " *CN";

		clearANNData();				/**< This will reset the dirty bit as well */

		//////////////////////////////////////////////////////////////////////////
		// Populate data points
		dataPts = annAllocPts(nPts, dim);

		for (int i = 0; i < nPts; i++) {
			for (int d = 0; d < dim; d++) {
				dataPts[i][d] = data[i].val[d];

				// Check bad floating point values
				if (!_finite(data[i].val[d])) {
					cerr << "Error in PointListN::ANNquery(): Non-finite floating point value at index " << i << "!" << endl;
					// TODO: Deal with this error in a better way
					dataPts[i][d] = 0;
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		kdTree = new ANNkd_tree(				// build search structure
								dataPts,		// the data points
								nPts,			// number of points
								dim);			// dimension of space
	}


	//////////////////////////////////////////////////////////////////////////
	// Convert the query point into an ANN-familiar format.
	if (!queryPt) {
		queryPt = annAllocPt(dim);
	}

	for (int d = 0; d < dim; d++) {			// Copy the query point
		queryPt[d] = q.val[d];
	}


	//////////////////////////////////////////////////////////////////////////
	// CORE
	//////////////////////////////////////////////////////////////////////////
	if (radSq < 0) {
		//cout << " --> Performing k-nearest query..." << endl;

		// Perform k-nearest-neighbor search
		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		kdTree->annkSearch(
							queryPt,			// query point
							k,					// number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}
	else {
		//cout << " --> Performing range query..." << endl;

		// Perform range search; initially find the number of points within the range
		k = kdTree->annkFRSearch(
									queryPt,			// query point
									radSq,				// radius of range query
									0,					// initially 0
									NULL,				// return no nearest neighbors
									NULL,				// return no distances
									queryEpsilon);		// error bound

		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		// Repeat the search with the *correct* number of neighbors
		kdTree->annkFRSearch(
							queryPt,			// query point
							radSq,				// radius of range query
							k,					// *correct* number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound

		// Resize result arrays.. since we don't know k in advance here
		indices.resize(k);
		distances.resize(k);
	}

	//////////////////////////////////////////////////////////////////////////
	// Copy result data
	//////////////////////////////////////////////////////////////////////////
	// Assume indices and distances are already resized appropriately
	for (int i = 0; i < k; i++) {
		indices[i] = nnIdx[i];
		distances[i] = sqrt(dists[i]);
	}

	//*******************************************************//
	// TESTING: measure time spent on ANN queries
	MEASURE_PERF_NARG(TIMER_POINTLIST_N_ANN_QUERY,Stop);
	//*******************************************************//

	//////////////////////////////////////////////////////////////////////////
	// Clean things up
	//////////////////////////////////////////////////////////////////////////
	delete [] nnIdx;
	delete [] dists;
	//annDeallocPt(queryPt);
}
