#ifndef __POINT_LIST_N_H__
#define __POINT_LIST_N_H__

#include <ANN/ANN.h>
#include "types.h"

/**
* A list of points in N-dimensions, with k-nearest neighbor and range search capabilities.
* Uses the ANN library to find nearest neighbors.
*/
class PointListN
{
public:
	PointListN( int dim );									/**< Constructor */
	~PointListN();											/**< Destructor */
	PointListN(const PointListN& rhs);						/**< Copy Constructor */
	const PointListN& operator=(const PointListN & rhs);	/**< Assignment */


	//////////////////////////////////////////////////////////////////////////
	// Access
	//////////////////////////////////////////////////////////////////////////
	int getDim() const;										/**< Returns the dimensionality of the points. */
	size_type size() const;									/**< Returns the number of points contained in this list. */
	data_type operator[](index_type i) const;				/**< Return a constant reference to the i^th point in the list. */

	// Perform a k-nearest-neighbor query or a range query.  Refer to the function ANNquery for details.
	void queryKNN( const data_type& q, size_type k, idxList& nearest, distList& dist, bool persistent = false) const;
	void queryRange( const data_type& q, float radSq, idxList& inRange, distList& dist, bool persistent = false) const;


	//////////////////////////////////////////////////////////////////////////
	// Modify
	//////////////////////////////////////////////////////////////////////////
	data_type& operator[](index_type i);					/**< Return a modifiable reference to the i^th point in the list. */
	void resize( size_type n );								/**< Resize the array to contain n points. */
	void reserve( size_type n );							/**< Reserve the array space to contain n points. */
	void add( const data_type& pt );						/**< Add a point to the list. */
	void removePoints( idxList list );						/**< Remove points from the list. */
	void clear();											/**< Clear the list of points. */
	void setANNEpsilon( double eps );						/**< Set the ANN approximation value for KNN and range queries. */


	//////////////////////////////////////////////////////////////////////////
	// Manipulating entire list
	//////////////////////////////////////////////////////////////////////////
	const std::vector<data_type>* getData() const;			/**< Returns a pointer to the vector containing the list of all points. */
	void setPts(const std::vector<data_type> *pts);			/**< Copy and replace an existing list with the current list of points. */


private:
	const int dim;											/**< Dimension */
	std::vector<data_type> data;							/**< List of points */

	//////////////////////////////////////////////////////////////////////////
	// ANN-related functionality
	//////////////////////////////////////////////////////////////////////////
	double queryEpsilon;						/**< Error bound for approximate nearest neighbor search. */
	mutable ANNpointArray dataPts;				/**< Persistent data array for multiple queries. */
	mutable ANNkd_tree* kdTree;					/**< Persistent search structure for multiple queries. */
	mutable bool dirtyBit;						/**< Valid bit to test whether the ANN data structures are still OK to use. */
	void clearANNData() const;					/**< Clear the persistent ANN data structures, if they exist. */

	/**
	* Perform either a k-nearest-neighbor query or a range query depending on the input parameters.
	* To perform a k-nearest neighbor query, set "radSq" to a negative number
	* Setting "radSq" to a positive number retrieves all points within distance sqrt(radSq) of the query point.
	* @param q The query point.
	* @param k The number of nearest neighbors to retrieve.
	* @param radSq The squared radius of the range query to perform.
	* @param indices The indices of the retrieved points.
	* @param distances The distance to the retrieved points.
	* @param persistent Tell the function whether to keep a persistent data structure for the query.
	*/
	void ANNquery( const data_type& q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent ) const;
};


inline int PointListN::getDim() const
{
	return dim;
}

inline size_type PointListN::size() const
{
	return data.size();
}

inline data_type PointListN::operator[]( index_type i ) const
{
	return data[i];
}

inline data_type& PointListN::operator[]( index_type i )
{
	dirtyBit = true;
	return data[i];
}

#endif
