#include "geom_util.h"
#include <iostream>
#include <limits>
#include <QString>
#include <ShapeModel.h>

using namespace std;

void printEdgeList(const EdgeList& edgeList, std::ostream& out/* = std::cout*/, int numDigits/* = 0*/)
{
	QString format("e%1: (v%2, v%3) (t%4, t%5)");
	QString str;

	for (index_type ei = 0, eEnd = edgeList.size(); ei < eEnd; ei++) {
		const Edge& current = edgeList[ei];
		str = format.arg(ei, numDigits)
			.arg(current.vi.x, numDigits)
			.arg(current.vi.y, numDigits)
			.arg(current.ti.x, numDigits)
			.arg(current.ti.y, numDigits);

		out << qPrintable(str) << endl;
	}
	out << endl;
}

void printAdjList(const AdjList& adjList, ostream& out/* = std::cout*/, int numDigits/* = 0*/)
{
	QString basicFormat("v%1:%2");
	QString pairFormat(" (v%1, e%2)");
	QString str;

	for (index_type vi = 0, vEnd = adjList.size(); vi < vEnd; vi++) {
		QString listStr("");

		// print the list of adjacent vertices
		for (index_type ai = 0, aEnd = adjList[vi].size(); ai < aEnd; ai++) {
			listStr = listStr + pairFormat.arg(adjList[vi][ai].x,numDigits).arg(adjList[vi][ai].y,numDigits);
		}

		str = basicFormat.arg(vi, numDigits).arg(listStr);

		out << qPrintable(str) << endl;
	}
	out << endl;
}

void createEdgeList(const Mesh3_real& inMesh, EdgeList& outEdgeList)
{
	EdgeList edgeList;		/**< Edge list*/
	AdjList adjList;		/**< Vertex adjacency list */
	edgeList.clear();
	adjList.clear();

	index_type numVertices = inMesh.numVertices();
	index_type numTriangles = inMesh.triangle_vertex_index.size();

	adjList.resize(numVertices);

	// Construct a list of edges
	for (index_type ti = 0, tNum = numTriangles;
		ti < tNum; ti++)
	{
		// Grab the vertices
		index3_type tri = inMesh.triangle_vertex_index[ti];

		// For each edge
		for (int i = 1; i <= 3; i++) {
			int i0 = i - min(i,2); int i1 = i - i0;		// iterates in (0, 1), (0, 2), (1, 2)
			int v0, v1;									// grabs the vertices of the edges
			bool found;
			index_type j;

			// Check to see if the edge is already in the adjList
			// Do both orders of the edge (v0->v1 and v1->v0)
			for (int temp = 0; temp < 2; temp++) {
				// The endpoints of the edge
				if (temp == 0) {
					v0 = tri[i0]; v1 = tri[i1];
				}
				else {
					v0 = tri[i1]; v1 = tri[i0];
				}

				found = false;
				for (j = 0; j < adjList[v0].size(); j++) {
					if (adjList[v0][j].x == v1) {
						found = true;
						break;
					}
				}

				if (found) {
					// Update the edge with the current triangle index
					edgeList[adjList[v0][j].y].ti.y = ti;
				}
				else {
					index_type ei;

					if (temp == 0) {
						// Create a new edge
						Edge ne;
						ne.vi.x = v0; ne.vi.y = v1;
						ne.ti.x = ti;

						// push back onto the edgeList
						edgeList.push_back(ne);

						ei = edgeList.size()-1;
					}
					else {
						// The edge must be stored already in the adjList
						found = false;
						for (j = 0; j < adjList[v1].size(); j++) {
							if (adjList[v1][j].x == v0) {
								found = true;
								break;
							}
						}
						if (found) {
							ei = adjList[v1][j].y;
						}
						else {
							// This case should not happen
							throw "createEdgeList(): Reverse edge not found in adjList!";
						}
					}

					// Add to the adjacency list
					index2_type temp;
					temp.x = v1;
					temp.y = ei;
					adjList[v0].push_back(temp);
				}
			}
		}
	}

	// Assign output
	outEdgeList.swap(edgeList);
}

void computeStats( const EdgeList& edgeList, const AdjList& adjList )
{
	// Compute the min and max degree of a vertex
	index_type minDegree = adjList.size();
	index_type maxDegree = 0;
	for (index_type vi = 0, vEnd = adjList.size(); vi < vEnd; vi++) {
		unsigned int deg = adjList[vi].size();
		if (deg < minDegree) {
			minDegree = deg;
		}
		if (deg > maxDegree) {
			maxDegree = deg;
		}
	}

	// Find all boundary edges
	std::vector<index_type> bdyEdges;	/**< Mark boundary edges */
	bdyEdges.clear();
	for (index_type ei = 0, eEnd = edgeList.size(); ei < eEnd; ei++) {
		if (edgeList[ei].isBoundary()) {
			bdyEdges.push_back(ei);
		}
	}

	cout << "Minimum Degree: " << minDegree << endl;
	cout << "Maximum Degree: " << maxDegree << endl;
	cout << "Number of Boundary Edges: " << bdyEdges.size() << endl;
}


// Adapted from:
// Wild Magic Source Code
// David Eberly
// http://www.geometrictools.com
// Copyright (c) 1998-2007
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or (at
// your option) any later version.  The license is available for reading at
// either of the locations:
//     http://www.gnu.org/copyleft/lgpl.html
//     http://www.geometrictools.com/License/WildMagicLicense.pdf
//
// Version: 4.0.0 (2006/06/28) (box class)
// Version: 4.0.1 (2007/05/06) (point triangle intersection)
// Version: 4.0.2 (2008/09/15) (box-box intersection)
int pointTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2, real3_type &closestPoint, real_type &fSqrDistance )
{
	real3_type kDiff = t0 - p;
	real3_type kEdge0 = t1 - t0;
	real3_type kEdge1 = t2 - t0;
	real_type fA00 = kEdge0 * kEdge0;
	real_type fA01 = kEdge0 * kEdge1;
	real_type fA11 = kEdge1 * kEdge1;
	real_type fB0 = kDiff * kEdge0;
	real_type fB1 = kDiff * kEdge1;
	real_type fC = kDiff * kDiff;
	real_type fDet = abs(fA00*fA11-fA01*fA01);
	real_type fS = fA01*fB1-fA11*fB0;
	real_type fT = fA01*fB0-fA00*fB1;
	int regionType = -1;

	if (fS + fT <= fDet)
	{
		if (fS < (real_type)0.0)
		{
			if (fT < (real_type)0.0)  // region 4
			{
				regionType = 4;
				if (fB0 < (real_type)0.0)
				{
					fT = (real_type)0.0;
					if (-fB0 >= fA00)
					{
						fS = (real_type)1.0;
						fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
					}
					else
					{
						fS = -fB0/fA00;
						fSqrDistance = fB0*fS+fC;
					}
				}
				else
				{
					fS = (real_type)0.0;
					if (fB1 >= (real_type)0.0)
					{
						fT = (real_type)0.0;
						fSqrDistance = fC;
					}
					else if (-fB1 >= fA11)
					{
						fT = (real_type)1.0;
						fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
					}
					else
					{
						fT = -fB1/fA11;
						fSqrDistance = fB1*fT+fC;
					}
				}
			}
			else  // region 3
			{
				regionType = 3;
				fS = (real_type)0.0;
				if (fB1 >= (real_type)0.0)
				{
					fT = (real_type)0.0;
					fSqrDistance = fC;
				}
				else if (-fB1 >= fA11)
				{
					fT = (real_type)1.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else
				{
					fT = -fB1/fA11;
					fSqrDistance = fB1*fT+fC;
				}
			}
		}
		else if (fT < (real_type)0.0)  // region 5
		{
			regionType = 5;
			fT = (real_type)0.0;
			if (fB0 >= (real_type)0.0)
			{
				fS = (real_type)0.0;
				fSqrDistance = fC;
			}
			else if (-fB0 >= fA00)
			{
				fS = (real_type)1.0;
				fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
			}
			else
			{
				fS = -fB0/fA00;
				fSqrDistance = fB0*fS+fC;
			}
		}
		else  // region 0
		{
			regionType = 0;
			// minimum at interior point
			real_type fInvDet = ((real_type)1.0)/fDet;
			fS *= fInvDet;
			fT *= fInvDet;
			fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
				fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
		}
	}
	else
	{
		real_type fTmp0, fTmp1, fNumer, fDenom;

		if (fS < (real_type)0.0)  // region 2
		{
			regionType = 2;
			fTmp0 = fA01 + fB0;
			fTmp1 = fA11 + fB1;
			if (fTmp1 > fTmp0)
			{
				fNumer = fTmp1 - fTmp0;
				fDenom = fA00-2.0f*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fS = (real_type)1.0;
					fT = (real_type)0.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else
				{
					fS = fNumer/fDenom;
					fT = (real_type)1.0 - fS;
					fSqrDistance = fS*(fA00*fS+fA01*fT+2.0f*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
			else
			{
				fS = (real_type)0.0;
				if (fTmp1 <= (real_type)0.0)
				{
					fT = (real_type)1.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else if (fB1 >= (real_type)0.0)
				{
					fT = (real_type)0.0;
					fSqrDistance = fC;
				}
				else
				{
					fT = -fB1/fA11;
					fSqrDistance = fB1*fT+fC;
				}
			}
		}
		else if (fT < (real_type)0.0)  // region 6
		{
			regionType = 6;
			fTmp0 = fA01 + fB1;
			fTmp1 = fA00 + fB0;
			if (fTmp1 > fTmp0)
			{
				fNumer = fTmp1 - fTmp0;
				fDenom = fA00-((real_type)2.0)*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fT = (real_type)1.0;
					fS = (real_type)0.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else
				{
					fT = fNumer/fDenom;
					fS = (real_type)1.0 - fT;
					fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
			else
			{
				fT = (real_type)0.0;
				if (fTmp1 <= (real_type)0.0)
				{
					fS = (real_type)1.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else if (fB0 >= (real_type)0.0)
				{
					fS = (real_type)0.0;
					fSqrDistance = fC;
				}
				else
				{
					fS = -fB0/fA00;
					fSqrDistance = fB0*fS+fC;
				}
			}
		}
		else  // region 1
		{
			regionType = 1;
			fNumer = fA11 + fB1 - fA01 - fB0;
			if (fNumer <= (real_type)0.0)
			{
				fS = (real_type)0.0;
				fT = (real_type)1.0;
				fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
			}
			else
			{
				fDenom = fA00-2.0f*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fS = (real_type)1.0;
					fT = (real_type)0.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else
				{
					fS = fNumer/fDenom;
					fT = (real_type)1.0 - fS;
					fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
		}
	}

	// account for numerical round-off error
	if (fSqrDistance < (real_type)0.0)
	{
		fSqrDistance = (real_type)0.0;
	}

	closestPoint = t0 + fS*kEdge0 + fT*kEdge1;
	//m_afTriangleBary[1] = fS;
	//m_afTriangleBary[2] = fT;
	//m_afTriangleBary[0] = (real_type)1.0 - fS - fT;
	return regionType;
}


real_type pointBoxSqDistance ( const real3_type &p, const real3_type &boxMin, const real3_type &boxMax, real3_type &closestPoint )
{
	// work in the box's coordinate system
	real3_type boxCenter = (boxMin + boxMax) / (real_type)2.0;
	real3_type halfLength = (boxMax - boxMin) / (real_type)2.0;
	real3_type kDiff = p - boxCenter;

	// compute squared distance and closest point on box
	real_type fSqrDistance = (real_type)0.0, fDelta;
	real3_type kClosest;

	for (int i = 0; i < 3; i++)
	{
		kClosest[i] = kDiff[i];
		if (kClosest[i] < -halfLength[i])
		{
			fDelta = kClosest[i] + halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = -halfLength[i];
		}
		else if (kClosest[i] > halfLength[i])
		{
			fDelta = kClosest[i] - halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = halfLength[i];
		}
	}

	closestPoint = boxCenter + kClosest;

	return fSqrDistance;
}


real_type lowerBoundTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2 )
{
	// Find bounding box of triangle
	static real3_type boxMin, boxMax;
	boxMin[0] = boxMin[1] = boxMin[2] = REAL_TYPE_MAX;
	boxMax[0] = boxMax[1] = boxMax[2] = -REAL_TYPE_MAX;

	for (int d = 0; d < 3; d++) {
		if (t0[d] < boxMin[d])      boxMin[d] = t0[d];
		else if (t0[d] > boxMax[d]) boxMax[d] = t0[d];
		if (t1[d] < boxMin[d])      boxMin[d] = t1[d];
		else if (t1[d] > boxMax[d]) boxMax[d] = t1[d];
		if (t2[d] < boxMin[d])      boxMin[d] = t2[d];
		else if (t2[d] > boxMax[d]) boxMax[d] = t2[d];
	}

	//////////////////////////////////////////////////////////////////////////
	// Method 1: Lower-bound via distance to bounding box of triangle
	static real3_type temp;
	return pointBoxSqDistance(p, boxMin, boxMax, temp);

	//////////////////////////////////////////////////////////////////////////
	// Method 2: Approximate bounding sphere as center of box with length half of diagonal
	//real3_type center = (boxMax + boxMin) / (real_type)2.0;
	//real_type r = (boxMax - boxMin).norm() / (real_type)2.0;
	//real3_type diff = p - center;
	//return max(0.0, diff*diff - r*r);
}


// inverts a 3x3 matrix
void invert(Geometry::Matrix33f &mat) {
	float a00, a01, a02;
	float a10, a11, a12;
	float a20, a21, a22;

	a00 = mat(0,0); a10 = mat(1,0); a20 = mat(2,0);
	a01 = mat(0,1); a11 = mat(1,1); a21 = mat(2,1);
	a02 = mat(0,2); a12 = mat(1,2); a22 = mat(2,2);

	float div = 1.0 / (a00*a11*a22-a00*a21*a12-a10*a01*a22+a10*a21*a02+a20*a01*a12-a20*a11*a02);

	mat(0,0) = (a11*a22-a21*a12)*div;
	mat(1,0) = -(a10*a22-a20*a12)*div;
	mat(2,0) = (a10*a21-a20*a11)*div;

	mat(0,1) = -(a01*a22-a21*a02)*div;
	mat(1,1) = (a00*a22-a20*a02)*div;
	mat(2,1) = -(a00*a21-a20*a01)*div;

	mat(0,2) = (a01*a12-a11*a02)*div;
	mat(1,2) = -(a00*a12-a10*a02)*div;
	mat(2,2) = (a00*a11-a10*a01)*div;
}



void bilinearInvert3D(const real3_type &p,
					  const real3_type &v0,
					  const real3_type &v1,
					  const real3_type &v2,
					  const real3_type &v3,
					  real_type &alpha,
					  real_type &beta)
{
	real3_type n = cross(v1-v2,v0-v3);
	real_type A = n * (cross(v0-v2,v1-v3));
	real_type B = n * (cross(v1-v3,p-v2) - cross(v0-v2,p-v3));
	real_type C = n * (cross(p-v2,p-v3));

	if (B > 0) {
		beta = 0.5 * (-B - sqrt(B*B-4.0*A*C)) / A;
	}
	else {
		beta = 2.0 * C / (-B + sqrt(B*B-4.0*A*C));
	}

	real3_type s1 = (1.0 - beta)*v2 + beta*v0;
	real3_type s2 = (1.0 - beta)*v3 + beta*v1;

	alpha = ((p-s1)*(s2-s1)) / ((s2-s1)*(s2-s1));
}

real3_type trilinearNewtonUpdate( const real3_type &p, std::vector<real3_type> verts, const real3_type &weights, real3_type &f )
{
	Geometry::Matrix33f J;
	real_type t1, t2, t3;
	const real_type &x = p[0];
	const real_type &y = p[1];
	const real_type &z = p[2];
	const real_type &a = weights[0];
	const real_type &b = weights[1];
	const real_type &c = weights[2];
	const real_type &x_0 = (verts[0])[0]; const real_type &y_0 = (verts[0])[1]; const real_type &z_0 = (verts[0])[2];
	const real_type &x_1 = (verts[1])[0]; const real_type &y_1 = (verts[1])[1]; const real_type &z_1 = (verts[1])[2];
	const real_type &x_2 = (verts[2])[0]; const real_type &y_2 = (verts[2])[1]; const real_type &z_2 = (verts[2])[2];
	const real_type &x_3 = (verts[3])[0]; const real_type &y_3 = (verts[3])[1]; const real_type &z_3 = (verts[3])[2];
	const real_type &x_4 = (verts[4])[0]; const real_type &y_4 = (verts[4])[1]; const real_type &z_4 = (verts[4])[2];
	const real_type &x_5 = (verts[5])[0]; const real_type &y_5 = (verts[5])[1]; const real_type &z_5 = (verts[5])[2];
	const real_type &x_6 = (verts[6])[0]; const real_type &y_6 = (verts[6])[1]; const real_type &z_6 = (verts[6])[2];
	const real_type &x_7 = (verts[7])[0]; const real_type &y_7 = (verts[7])[1]; const real_type &z_7 = (verts[7])[2];

	t1 = 1.0-a;
	t2 = 1.0-b;
	t3 = 1.0-c;
	f[0] = -x+((t1*x_3+a*x_7)*c+(t1*x_2+a*x_6)*t3)*b+((t1*x_1+a*x_5)*c+(t1*x_0+a*x_4)*t3)*t2;
	f[1] = -y+((t1*y_3+a*y_7)*c+(t1*y_2+a*y_6)*t3)*b+((t1*y_1+a*y_5)*c+(t1*y_0+a*y_4)*t3)*t2;
	f[2] = -z+((t1*z_3+a*z_7)*c+(t1*z_2+a*z_6)*t3)*b+((t1*z_1+a*z_5)*c+(t1*z_0+a*z_4)*t3)*t2;

	t1 = -1.0+b;
	t2 = 1.0-c;
	t3 = 1.0-a;
	J(0,0) = ((-x_3+x_7)*b+(x_1-x_5)*t1)*c+((-x_2+x_6)*b+(x_0-x_4)*t1)*t2;
	J(0,1) = ((-x_5+x_7)*c+(-x_4+x_6)*t2)*a+((-x_1+x_3)*c+(-x_0+x_2)*t2)*t3;
	J(0,2) = ((-x_6+x_7)*b+(x_4-x_5)*t1)*a+((-x_2+x_3)*b+(x_0-x_1)*t1)*t3;
	J(1,0) = ((-y_3+y_7)*b+(y_1-y_5)*t1)*c+((-y_2+y_6)*b+(y_0-y_4)*t1)*t2;
	J(1,1) = ((-y_5+y_7)*c+(-y_4+y_6)*t2)*a+((-y_1+y_3)*c+(-y_0+y_2)*t2)*t3;
	J(1,2) = ((-y_6+y_7)*b+(y_4-y_5)*t1)*a+((-y_2+y_3)*b+(y_0-y_1)*t1)*t3;
	J(2,0) = ((-z_3+z_7)*b+(z_1-z_5)*t1)*c+((-z_2+z_6)*b+(z_0-z_4)*t1)*t2;
	J(2,1) = ((-z_5+z_7)*c+(-z_4+z_6)*t2)*a+((-z_1+z_3)*c+(-z_0+z_2)*t2)*t3;
	J(2,2) = ((-z_6+z_7)*b+(z_4-z_5)*t1)*a+((-z_2+z_3)*b+(z_0-z_1)*t1)*t3; 

	// Invert and solve the system
	invert(J);
	return -J*f;
}

real3_type invertTrilinear(const real3_type &p, std::vector<real3_type> verts, int maxItr/* = 10*/)
{
	int i;
	real3_type weights, update, f;
	weights[0] = weights[1] = weights[2] = 0.0;

	for (i = 0; i < maxItr; i++) {
		update = trilinearNewtonUpdate(p, verts, weights, f);
		weights = weights + update;

		if (f.norm() < 1e-5) {
			break;
		}
	}

	//cerr << "Terminated in iteration " << i << " with error " << f.norm() << endl;

	if (f.norm() > 1e-5) {
		// Set weights to infinity
		cerr << "Warning: invertTrilinear() failed; larger error after iteration " << i << " with error " << f.norm() << endl;
		weights[0] = weights[1] = weights[2] = REAL_TYPE_INF;
	}
	return weights;
}

bool hexahedronTestInside( const real3_type &p, std::vector<real3_type> verts )
{
	// Find the projected position under each face
	static const int hexFaces[6][4] = {{0,1,2,3},{4,6,5,7},{0,4,1,5},{2,3,6,7},{0,2,4,6},{1,5,3,7}};

	for (int i = 0; i < 6; i++) {
		real_type alpha = 0.0, beta = 0.0;
		real3_type s1, s2, projected, normal;

		bilinearInvert3D(p,
			verts[hexFaces[i][0]],
			verts[hexFaces[i][1]],
			verts[hexFaces[i][2]],
			verts[hexFaces[i][3]],
			alpha, beta);

		// Compute the clamped projected location
		alpha = std::min(1.0f, std::max(0.0f, alpha));
		beta = std::min(1.0f, std::max(0.0f, beta));
		s1 = (1.0 - beta)*(verts[hexFaces[i][2]]) + beta*(verts[hexFaces[i][0]]);
		s2 = (1.0 - beta)*(verts[hexFaces[i][3]]) + beta*(verts[hexFaces[i][1]]);
		projected = (1.0 - alpha)*s1 + alpha*s2;

		// Compute normal using alpha, beta values
		s1 = (1.0 - beta) * (verts[hexFaces[i][3]] - verts[hexFaces[i][2]]) +
				beta      * (verts[hexFaces[i][1]] - verts[hexFaces[i][0]]);
		s2 = (1.0 - alpha) * (verts[hexFaces[i][0]] - verts[hexFaces[i][2]]) +
				alpha      * (verts[hexFaces[i][1]] - verts[hexFaces[i][3]]);
		normal = cross(s1,s2);
		normal.normalize();

		// Test if it's on the correct side of the surface
		if ((normal * (p - projected)) < 0) {
			return false;
		}
	}

	return true;
}



template<typename Real>
bool IntrBox3Box3Test( const Box3<Real>* m_pkBox0, const Box3<Real>* m_pkBox1 )
{
	// Cutoff for cosine of angles between box axes.  This is used to catch
	// the cases when at least one pair of axes are parallel.  If this
	// happens, there is no need to test for separation along the
	// Cross(A[i],B[j]) directions.
	const Real fCutoff = (Real)1.0 - numeric_limits<Real>::epsilon();
	bool bExistsParallelPair = false;
	int i;

	// convenience variables
	const Geometry::Vec3<Real>* akA = m_pkBox0->Axis;
	const Geometry::Vec3<Real>* akB = m_pkBox1->Axis;
	const Real* afEA = m_pkBox0->Extent;
	const Real* afEB = m_pkBox1->Extent;

	// compute difference of box centers, D = C1-C0
	Geometry::Vec3<Real> kD = m_pkBox1->Center - m_pkBox0->Center;

	Real aafC[3][3];     // matrix C = A^T B, c_{ij} = Dot(A_i,B_j)
	Real aafAbsC[3][3];  // |c_{ij}|
	Real afAD[3];        // Dot(A_i,D)
	Real fR0, fR1, fR;   // interval radii and distance between centers
	Real fR01;           // = R0 + R1

	// axis C0+t*A0
	for (i = 0; i < 3; i++)
	{
		aafC[0][i] = akA[0] * akB[i];
		aafAbsC[0][i] = fabs(aafC[0][i]);
		if (aafAbsC[0][i] > fCutoff)
		{
			bExistsParallelPair = true;
		}
	}
	afAD[0] = akA[0] * kD;
	fR = fabs(afAD[0]);
	fR1 = afEB[0]*aafAbsC[0][0]+afEB[1]*aafAbsC[0][1]+afEB[2]*aafAbsC[0][2];
	fR01 = afEA[0] + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A1
	for (i = 0; i < 3; i++)
	{
		aafC[1][i] = akA[1] * akB[i];
		aafAbsC[1][i] = fabs(aafC[1][i]);
		if (aafAbsC[1][i] > fCutoff)
		{
			bExistsParallelPair = true;
		}
	}
	afAD[1] = akA[1] * kD;
	fR = fabs(afAD[1]);
	fR1 = afEB[0]*aafAbsC[1][0]+afEB[1]*aafAbsC[1][1]+afEB[2]*aafAbsC[1][2];
	fR01 = afEA[1] + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A2
	for (i = 0; i < 3; i++)
	{
		aafC[2][i] = akA[2] * akB[i];
		aafAbsC[2][i] = fabs(aafC[2][i]);
		if (aafAbsC[2][i] > fCutoff)
		{
			bExistsParallelPair = true;
		}
	}
	afAD[2] = akA[2] * kD;
	fR = fabs(afAD[2]);
	fR1 = afEB[0]*aafAbsC[2][0]+afEB[1]*aafAbsC[2][1]+afEB[2]*aafAbsC[2][2];
	fR01 = afEA[2] + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*B0
	fR = fabs(akB[0] * kD);
	fR0 = afEA[0]*aafAbsC[0][0]+afEA[1]*aafAbsC[1][0]+afEA[2]*aafAbsC[2][0];
	fR01 = fR0 + afEB[0];
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*B1
	fR = fabs(akB[1] * kD);
	fR0 = afEA[0]*aafAbsC[0][1]+afEA[1]*aafAbsC[1][1]+afEA[2]*aafAbsC[2][1];
	fR01 = fR0 + afEB[1];
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*B2
	fR = fabs(akB[2] * kD);
	fR0 = afEA[0]*aafAbsC[0][2]+afEA[1]*aafAbsC[1][2]+afEA[2]*aafAbsC[2][2];
	fR01 = fR0 + afEB[2];
	if (fR > fR01)
	{
		return false;
	}

	// At least one pair of box axes was parallel, so the separation is
	// effectively in 2D where checking the "edge" normals is sufficient for
	// the separation of the boxes.
	if (bExistsParallelPair)
	{
		return true;
	}

	// axis C0+t*A0xB0
	fR = fabs(afAD[2]*aafC[1][0]-afAD[1]*aafC[2][0]);
	fR0 = afEA[1]*aafAbsC[2][0] + afEA[2]*aafAbsC[1][0];
	fR1 = afEB[1]*aafAbsC[0][2] + afEB[2]*aafAbsC[0][1];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A0xB1
	fR = fabs(afAD[2]*aafC[1][1]-afAD[1]*aafC[2][1]);
	fR0 = afEA[1]*aafAbsC[2][1] + afEA[2]*aafAbsC[1][1];
	fR1 = afEB[0]*aafAbsC[0][2] + afEB[2]*aafAbsC[0][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A0xB2
	fR = fabs(afAD[2]*aafC[1][2]-afAD[1]*aafC[2][2]);
	fR0 = afEA[1]*aafAbsC[2][2] + afEA[2]*aafAbsC[1][2];
	fR1 = afEB[0]*aafAbsC[0][1] + afEB[1]*aafAbsC[0][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A1xB0
	fR = fabs(afAD[0]*aafC[2][0]-afAD[2]*aafC[0][0]);
	fR0 = afEA[0]*aafAbsC[2][0] + afEA[2]*aafAbsC[0][0];
	fR1 = afEB[1]*aafAbsC[1][2] + afEB[2]*aafAbsC[1][1];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A1xB1
	fR = fabs(afAD[0]*aafC[2][1]-afAD[2]*aafC[0][1]);
	fR0 = afEA[0]*aafAbsC[2][1] + afEA[2]*aafAbsC[0][1];
	fR1 = afEB[0]*aafAbsC[1][2] + afEB[2]*aafAbsC[1][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A1xB2
	fR = fabs(afAD[0]*aafC[2][2]-afAD[2]*aafC[0][2]);
	fR0 = afEA[0]*aafAbsC[2][2] + afEA[2]*aafAbsC[0][2];
	fR1 = afEB[0]*aafAbsC[1][1] + afEB[1]*aafAbsC[1][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A2xB0
	fR = fabs(afAD[1]*aafC[0][0]-afAD[0]*aafC[1][0]);
	fR0 = afEA[0]*aafAbsC[1][0] + afEA[1]*aafAbsC[0][0];
	fR1 = afEB[1]*aafAbsC[2][2] + afEB[2]*aafAbsC[2][1];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A2xB1
	fR = fabs(afAD[1]*aafC[0][1]-afAD[0]*aafC[1][1]);
	fR0 = afEA[0]*aafAbsC[1][1] + afEA[1]*aafAbsC[0][1];
	fR1 = afEB[0]*aafAbsC[2][2] + afEB[2]*aafAbsC[2][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	// axis C0+t*A2xB2
	fR = fabs(afAD[1]*aafC[0][2]-afAD[0]*aafC[1][2]);
	fR0 = afEA[0]*aafAbsC[1][2] + afEA[1]*aafAbsC[0][2];
	fR1 = afEB[0]*aafAbsC[2][1] + afEB[1]*aafAbsC[2][0];
	fR01 = fR0 + fR1;
	if (fR > fR01)
	{
		return false;
	}

	return true;
}

// Generate code for float
template bool IntrBox3Box3Test<float>( const Box3<float>* m_pkBox0, const Box3<float>* m_pkBox1 );
