#ifndef __GEOM_UTIL_H__
#define __GEOM_UTIL_H__

#include "types.h"

class Mesh3_real;

/**
* Edge type.
* Contains the endpoints of the edge (two vertices) and the two adjacent triangles.
* In particular we know that the edge is on the boundary if there is only one
* adjacent triangle.
*/
class Edge
{
public:
	Edge() {
		vi.x = 0; vi.y = 0;
		ti.x = 0; ti.y = 0;
	}

	/**
	 * Determine if the edge is on the boundary.
	 *
	 * This method assumes that the edge list has been constructed in full.
	 * After construction of the edge list, triangle index 0 can only be on the first element,
	 * since triangle 0 is the first triangle to be processed and its three edges are added in the first iteration.
	 * Therefore, if the second triangle index is the default value 0, then it is a boundary edge.
	 */
	bool isBoundary() const
	{
		return (ti.y == 0);
	}

	index2_type vi;		// adjacent vertices
	index2_type ti;		// adjacent triangles
};

typedef std::vector<Edge> EdgeList;
typedef std::vector<std::vector<index2_type> > AdjList;

/**
 * Create a list of edges (along with their adjacent triangles) for the input mesh.
 *
 * We actually maintain two lists: the adjacency list for vertices, and the list
 * of edges.  The adjacency list stores, for each vertex v0, a list of adjacent
 * vertices v_adj along with the edge index # for the edge (v0 -> v_adj).
 * We use this adjacency list to check if the edge is already present in the
 * edge list, avoiding duplicates.
 *
 * To generate the list of edges, for each triangle, we add the three edges into
 * the edge list while checking for duplicates in the adjacency list.
 *
 */
void createEdgeList(const Mesh3_real& inMesh, EdgeList& outEdgeList);

/**
 * Debugging: print the edge list
 */
void printEdgeList(const EdgeList& edgeList, std::ostream& out = std::cout, int numDigits = 0);

/**
 * Debugging: print the vertex adjacency list
 */
void printAdjList(const AdjList& adjList, std::ostream& out = std::cout, int numDigits = 0);

/**
 * Find the min and max degree, and compute a list of boundary vertices.
 */
void computeStats( const EdgeList& edgeList, const AdjList& adjList );



// Squared distance between a point and a triangle
// Returns the region where the closest point was found
// Region 0: Closest point was inside the triangle
// Region 1: Closest point was on the edge between t1 and t2
// Region 2: Closest point was vertex t2
// Region 3: Closest point was on the edge between t2 and t0
// Region 4: Closest point was vertex t0
// Region 5: Closest point was on the edge between t0 and t1
// Region 6: Closest point was vertex t1
int pointTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2, real3_type &closestPoint, real_type &fSqrDistance );
real_type pointBoxSqDistance ( const real3_type &p, const real3_type &boxMin, const real3_type &boxMax, real3_type &closestPoint );
real_type lowerBoundTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2 );


// Cross product.
inline real3_type cross(const real3_type &v1, const real3_type &v2)
{
	return real3_type(v1.y()*v2.z() - v1.z()*v2.y(), v1.z()*v2.x() - v1.x()*v2.z(), v1.x()*v2.y() - v1.y()*v2.x());
}


// Inverting trilinear interpolation (adapted from hexahedronTest project)
void invert(Geometry::Matrix33f &mat);
void bilinearInvert3D(const real3_type &p, const real3_type &v0, const real3_type &v1, const real3_type &v2, const real3_type &v3, real_type &alpha, real_type &beta);
real3_type trilinearNewtonUpdate( const real3_type &p, std::vector<real3_type> verts, const real3_type &weights, real3_type &f );
real3_type invertTrilinear(const real3_type &p, std::vector<real3_type> verts, int maxItr = 50);
bool hexahedronTestInside( const real3_type &p, std::vector<real3_type> verts );



//////////////////////////////////////////////////////////////////////////
// Intersection between oriented bounding boxes.
template <class Real>
class Box3
{
public:
	// A box has center C, axis directions U[0], U[1], and U[2] (all
	// unit-length vectors), and extents e[0], e[1], and e[2] (all nonnegative
	// numbers).  A point X = C+y[0]*U[0]+y[1]*U[1]+y[2]*U[2] is inside or
	// on the box whenever |y[i]| <= e[i] for all i.

	// construction
	Box3 ();  // uninitialized
	Box3 (const Geometry::Vec3<Real>& rkCenter, const Geometry::Vec3<Real>* akAxis,
		const Real* afExtent);
	Box3 (const Geometry::Vec3<Real>& rkCenter, const Geometry::Vec3<Real>& rkAxis0,
		const Geometry::Vec3<Real>& rkAxis1, const Geometry::Vec3<Real>& rkAxis2,
		Real fExtent0, Real fExtent1, Real fExtent2);

	void ComputeVertices (Geometry::Vec3<Real> akVertex[8]) const;

	Geometry::Vec3<Real> Center;
	Geometry::Vec3<Real> Axis[3];		// must be an orthonormal set of vectors
	Real Extent[3];						// must be nonnegative
};

//----------------------------------------------------------------------------
template <class Real>
Box3<Real>::Box3 ()
{
	// uninitialized
}
//----------------------------------------------------------------------------
template <class Real>
Box3<Real>::Box3 (const Geometry::Vec3<Real>& rkCenter, const Geometry::Vec3<Real>* akAxis,
				  const Real* afExtent)
				  :
Center(rkCenter)
{
	for (int i = 0; i < 3; i++)
	{
		Axis[i] = akAxis[i];
		Extent[i] = afExtent[i];
	}
}
//----------------------------------------------------------------------------
template <class Real>
Box3<Real>::Box3 (const Geometry::Vec3<Real>& rkCenter, const Geometry::Vec3<Real>& rkAxis0,
				  const Geometry::Vec3<Real>& rkAxis1, const Geometry::Vec3<Real>& rkAxis2, Real fExtent0,
				  Real fExtent1, Real fExtent2)
				  :
Center(rkCenter)
{
	Axis[0] = rkAxis0;
	Axis[1] = rkAxis1;
	Axis[2] = rkAxis2;
	Extent[0] = fExtent0;
	Extent[1] = fExtent1;
	Extent[2] = fExtent2;
}
//----------------------------------------------------------------------------
template <class Real>
void Box3<Real>::ComputeVertices (Geometry::Vec3<Real> akVertex[8]) const
{
	Geometry::Vec3<Real> akEAxis[3] =
	{
		Extent[0]*Axis[0],
		Extent[1]*Axis[1],
		Extent[2]*Axis[2]
	};

	akVertex[0] = Center - akEAxis[0] - akEAxis[1] - akEAxis[2];
	akVertex[1] = Center + akEAxis[0] - akEAxis[1] - akEAxis[2];
	akVertex[2] = Center + akEAxis[0] + akEAxis[1] - akEAxis[2];
	akVertex[3] = Center - akEAxis[0] + akEAxis[1] - akEAxis[2];
	akVertex[4] = Center - akEAxis[0] - akEAxis[1] + akEAxis[2];
	akVertex[5] = Center + akEAxis[0] - akEAxis[1] + akEAxis[2];
	akVertex[6] = Center + akEAxis[0] + akEAxis[1] + akEAxis[2];
	akVertex[7] = Center - akEAxis[0] + akEAxis[1] + akEAxis[2];
}
//----------------------------------------------------------------------------

typedef Box3<float> Box3f;
typedef Box3<double> Box3d;
//////////////////////////////////////////////////////////////////////////

template<typename Real> extern bool IntrBox3Box3Test( const Box3<Real>* m_pkBox0, const Box3<Real>* m_pkBox1 );


#endif
