#include "quat.h"
#include "FindCorrespondence.h"
#include "geom_util.h"
#include "DeformGraphAll.h"

// Convert a rotation matrix to a unit quaternion.
// From http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
void MatrixToUnitQuat( const Geometry::Matrix33f& m, real4_type& q )
{
	real_type trace = m(0,0) + m(1,1) + m(2,2) + 1.0;

	if( trace > REAL_TYPE_EPSILON ) {
		real_type s = 0.5 / sqrt(trace);
		q[0] = 0.25 / s;
		q[1] = ( m(2,1) - m(1,2) ) * s;
		q[2] = ( m(0,2) - m(2,0) ) * s;
		q[3] = ( m(1,0) - m(0,1) ) * s;
	}
	else {
		if ( m(0,0) > m(1,1) && m(0,0) > m(2,2) ) {
			real_type s = 2.0 * sqrt( 1.0 + m(0,0) - m(1,1) - m(2,2));
			q[0] = (m(2,1) - m(1,2) ) / s;
			q[1] = 0.25 * s;
			q[2] = (m(0,1) + m(1,0) ) / s;
			q[3] = (m(0,2) + m(2,0) ) / s;
		}
		else if (m(1,1) > m(2,2)) {
			real_type s = 2.0 * sqrt( 1.0 + m(1,1) - m(0,0) - m(2,2));
			q[0] = (m(0,2) - m(2,0) ) / s;
			q[1] = (m(0,1) + m(1,0) ) / s;
			q[2] = 0.25 * s;
			q[3] = (m(1,2) + m(2,1) ) / s;
		}
		else {
			real_type s = 2.0 * sqrt( 1.0 + m(2,2) - m(0,0) - m(1,1) );
			q[0] = (m(1,0) - m(0,1) ) / s;
			q[1] = (m(0,2) + m(2,0) ) / s;
			q[2] = (m(1,2) + m(2,1) ) / s;
			q[3] = 0.25 * s;
		}
	}
}


// input: unit quaternion 'q0', translation vector 't' 
// output: unit dual quaternion represented by 'qn' and 'qd'
// Modified from original code by Ladislav Kavan, kavanl@cs.tcd.ie
void QuatTrans2UDQ( const real4_type& q0, const real3_type& t, real4_type& qn, real4_type& qd )
{
	// non-dual part (just copy q0):
	for (int i=0; i<4; i++) qn[i] = q0[i];
	// dual part:
	qd[0] = -0.5*(t[0]*q0[1] + t[1]*q0[2] + t[2]*q0[3]);
	qd[1] = 0.5*( t[0]*q0[0] + t[1]*q0[3] - t[2]*q0[2]);
	qd[2] = 0.5*(-t[0]*q0[3] + t[1]*q0[0] + t[2]*q0[1]);
	qd[3] = 0.5*( t[0]*q0[2] - t[1]*q0[1] + t[2]*q0[0]);
}


// input: unit dual quaternion represented by 'qn' and 'qd'
// output: unit quaternion 'q0', translation vector 't'
// Modified from original code by Ladislav Kavan, kavanl@cs.tcd.ie
void UDQ2QuatTrans( const real4_type& qn, const real4_type& qd, real4_type& q0, real3_type& t )
{
	// regular quaternion (just copy the non-dual part):
	for (int i=0; i<4; i++) q0[i] = qn[i];
	// translation vector:
	t[0] = 2.0*(-qd[0]*qn[1] + qd[1]*qn[0] - qd[2]*qn[3] + qd[3]*qn[2]);
	t[1] = 2.0*(-qd[0]*qn[2] + qd[1]*qn[3] + qd[2]*qn[0] - qd[3]*qn[1]);
	t[2] = 2.0*(-qd[0]*qn[3] - qd[1]*qn[2] + qd[2]*qn[1] + qd[3]*qn[0]);
}


// input: dual quat. represented by 'qn' and 'qd' with non-zero non-dual part
// output: unit quaternion 'q0', translation vector 't'
// Modified from original code by Ladislav Kavan, kavanl@cs.tcd.ie
void DQ2QuatTrans( const real4_type& qn, const real4_type& qd, real4_type& q0, real3_type& t )
{
	float len = 0.0;
	for (int i=0; i<4; i++) len += qn[i] * qn[i];
	len = sqrt(len); 
	for (int i=0; i<4; i++) q0[i] = qn[i] / len;
	t[0] = 2.0*(-qd[0]*qn[1] + qd[1]*qn[0] - qd[2]*qn[3] + qd[3]*qn[2]) / len;
	t[1] = 2.0*(-qd[0]*qn[2] + qd[1]*qn[3] + qd[2]*qn[0] - qd[3]*qn[1]) / len;
	t[2] = 2.0*(-qd[0]*qn[3] - qd[1]*qn[2] + qd[2]*qn[1] + qd[3]*qn[0]) / len;
}


// Convert a dual quaternion to a rotation and translation.  Also performs normalization.
// Modified from original code by Ladislav Kavan, kavanl@cs.tcd.ie
void DQToMatrix( const real4_type& Qn, const real4_type& Qd, Geometry::Matrix33f& R, real3_type& t )
{	
	real_type len2 = Qn * Qn;
	real_type w = Qn[0], x = Qn[1], y = Qn[2], z = Qn[3];
	real_type t0 = Qd[0], t1 = Qd[1], t2 = Qd[2], t3 = Qd[3];

	R(0,0) = w*w + x*x - y*y - z*z; R(0,1) = 2*x*y - 2*w*z; R(0,2) = 2*x*z + 2*w*y;
	R(1,0) = 2*x*y + 2*w*z; R(1,1) = w*w + y*y - x*x - z*z; R(1,2) = 2*y*z - 2*w*x; 
	R(2,0) = 2*x*z - 2*w*y; R(2,1) = 2*y*z + 2*w*x; R(2,2) = w*w + z*z - x*x - y*y;

	t[0] = -2*t0*x + 2*w*t1 - 2*t2*z + 2*y*t3;
	t[1] = -2*t0*y + 2*t1*z - 2*x*t3 + 2*w*t2;
	t[2] = -2*t0*z + 2*x*t2 + 2*w*t3 - 2*t1*y;

	// Divide by the length
	R /= len2;
	t /= len2;
}

// Linear blend skinning: Transform point and normal
void lbs(const real3_type& x, real3_type& x_out,
		 const Deform::DeformAll_Utility::RefWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse )
{
	const int nBones = tfList.size();

	if (weights.size() == 1) {
		index_type j = weights.bones[0];

		if (inverse) {
			Geometry::Matrix33f rotTransposed = tfList[j].rotMatrix().transpose();
			x_out = rotTransposed * (x - tfList[j].tr());
		}
		else {
			x_out = tfList[j] * x;
		}
	}
	else if (weights.size() > 1) {
		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < weights.size(); ++i) {
			tf.rot += weights.weights[i] * tfList[weights.bones[i]].rotMatrix();
			tf.tr += weights.weights[i] * tfList[weights.bones[i]].tr();
		}

		// Apply transformation (apply inverse transpose to the normal)
		if (inverse) {
			invert(tf.rot);
			x_out = tf.rot * (x - tf.tr);
		}
		else {
			x_out = tf * x;
		}
	}
}

// Linear blend skinning: Transform point and normal
void lbs(const real3_type& x, const real3_type& n, real3_type& x_out, real3_type& n_out,
		 const Deform::DeformAll_Utility::RefWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse )
{
	const int nBones = tfList.size();

	if (weights.size() == 1) {
		index_type j = weights.bones[0];

		if (inverse) {
			Geometry::Matrix33f rotTransposed = tfList[j].rotMatrix().transpose();
			x_out = rotTransposed * (x - tfList[j].tr());
			n_out = rotTransposed * n;
		}
		else {
			x_out = tfList[j] * x;
			n_out = tfList[j].rotMatrix() * n;
		}
	}
	else if (weights.size() > 1) {
		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < weights.size(); ++i) {
			tf.rot += weights.weights[i] * tfList[weights.bones[i]].rotMatrix();
			tf.tr += weights.weights[i] * tfList[weights.bones[i]].tr();
		}

		// Apply transformation (apply inverse transpose to the normal)
		if (inverse) {
			n_out = tf.rot.transpose() * n;
			invert(tf.rot);
			x_out = tf.rot * (x - tf.tr);
		}
		else {
			x_out = tf * x;
			invert(tf.rot);
			n_out = tf.rot.transpose() * n;
		}
		n_out.normalize();
	}
}

// Linear blend skinning
real3_type lbs( const real3_type& x, const data_type& weights, const RigidDeform& tfList, bool inverse )
{
	const int nBones = tfList.size();

	// Find transformation
	SE3 tf;
	tf.setZero();
	for (int i = 0; i < nBones; i++) {
		tf.rot += weights.val[i] * tfList[i].rot;
		tf.tr += weights.val[i] * tfList[i].tr;
	}

	// Apply transformation (apply inverse transpose to the normal)
	real3_type out;
	if (inverse) {
		invert(tf.rot);
		out = tf.rot * (x - tf.tr);
	}
	else {
		out = tf * x;
	}

	return out;
}

// Linear blend skinning: Transform point and normal
void lbs(const real3_type& x, const real3_type& n, real3_type& x_out, real3_type& n_out,
		 const data_type& weights, const RigidDeform& tfList, bool inverse )
{
	const int nBones = tfList.size();
	
	// Find transformation
	SE3 tf;
	tf.setZero();
	for (int i = 0; i < nBones; i++) {
		tf.rot += weights.val[i] * tfList[i].rot;
		tf.tr += weights.val[i] * tfList[i].tr;
	}

	// Apply transformation (apply inverse transpose to the normal)
	if (inverse) {
		n_out = tf.rot.transpose() * n;
		invert(tf.rot);
		x_out = tf.rot * (x - tf.tr);
	}
	else {
		x_out = tf * x;
		invert(tf.rot);
		n_out = tf.rot.transpose() * n;
	}
	n_out.normalize();
}


// Basic dual quaternion skinning
real3_type dqs( const real3_type& x, const data_type& weights, const RigidDeform& tfList, bool inverse )
{
	const int nBones = tfList.size();

	// Find transformation
	real4_type qn, qd;
	for (int i = 0; i < nBones; i++) {
		real4_type boneQuat, boneQn, boneQd;
		MatrixToUnitQuat(tfList[i].rot, boneQuat);
		QuatTrans2UDQ(boneQuat, tfList[i].tr, boneQn, boneQd);
		qn += weights.val[i] * boneQn;
		qd += weights.val[i] * boneQd;
	}
	Geometry::Matrix33f R;
	real3_type t;
	DQToMatrix(qn, qd, R, t);

	// Apply transformation
	real3_type out;
	if (inverse) {
		out = R.transpose() * (x - t);
	}
	else {
		out = R * x + t;
	}

	return out;
}

// Basic dual quaternion skinning: Transform point and normal
void dqs(const real3_type& x, const real3_type& n, real3_type& x_out, real3_type& n_out,
		 const data_type& weights, const RigidDeform& tfList, bool inverse )
{
	const int nBones = tfList.size();

	// Find transformation
	real4_type qn, qd;
	for (int i = 0; i < nBones; i++) {
		real4_type boneQuat, boneQn, boneQd;
		MatrixToUnitQuat(tfList[i].rot, boneQuat);
		QuatTrans2UDQ(boneQuat, tfList[i].tr, boneQn, boneQd);
		qn += weights.val[i] * boneQn;
		qd += weights.val[i] * boneQd;
	}
	Geometry::Matrix33f R;
	real3_type t;
	DQToMatrix(qn, qd, R, t);

	// Apply transformation
	if (inverse) {
		x_out = R.transpose() * (x - t);
		n_out = R.transpose() * n;
	}
	else {
		x_out = R * x + t;
		n_out = R * n;
	}
}
