#include "util.h"
#include "taucsUtil.h"
#include <iostream>
#include <cmath>

#undef min
#undef max

using namespace std;
using namespace Deform;

// Declaration for the compiler
template void Deform::SparseMatrix<real_type>::fill(index_type row, index_type col, real_type val);
template taucs_ccs_matrix* Deform::SparseMatrix<real_type>::writeTaucsMatrix(size_type nRows, size_type nCols, bool transposed /* = false */ ) const;


template<typename T>
void SparseMatrix<T>::print()
{
	// Print in Matlab format; change to 1-based indices
	for (unsigned int i = 0; i < size(); i++) {
		cout << "M(" << index[i].x+1 << "," << index[i].y+1 << ") = " << data[i] << ";" << endl;
	}
}

template<typename T>
void SparseMatrix<T>::write( const char *filename ) const
{
	std::ofstream out;
	out.open(filename);

	if (out.fail()) {
		std::cerr << "Error in SparseMatrix<T>::write(): Could not open output file " << filename << " for writing!" << std::endl;
		return;
	}

	// Print in Matlab format; change to 1-based indices
	for (unsigned int i = 0; i < currentSize; i++) {
		out << index[i].x+1 << " " << index[i].y+1 << " " << (taucsType) data[i] << endl;
	}

	out.close();
}

template<typename T>
void SparseMatrix<T>::getTaucsColumns( std::vector< std::map<int,taucsType> > &result, size_type nCols ) const
{
	result.clear();
	result.resize(nCols);

	// gather the data and put them into columns
	for (size_type i = 0; i < currentSize; i++) {
		index2_type &idx = index[i];
		result[idx.y][idx.x] = (taucsType) data[i];
	}
}

template<typename T>
taucs_ccs_matrix * SparseMatrix<T>::writeTaucsMatrix(size_type nRows, size_type nCols, bool transposed/* = false*/ ) const
{
	//// Find the columns and use the CreateTaucsMatrixFromColumns() utility function.
	//std::vector< std::map<int,taucsType> > cols;
	//getTaucsColumns(cols, nCols);
	//return CreateTaucsMatrixFromColumns(cols, nRows, TAUCS_DOUBLE);		/**< Currently the taucsType is double; should follow the type defined in taucsaddon.h */

	//cout << "Total number of non-zeros: " << currentSize << endl;

#if 0
	// Alternative algorithm: avoid creating a vector of maps
	std::vector<int> linearIndex(currentSize);
	for (int i = 0; i < linearIndex.size(); ++i) {
		linearIndex[i] = i;
	}

	// Sort the entries
	if (transposed) {
		// Already sorted!! ^^
		//std::sort(linearIndex.begin(), linearIndex.end(), make_ForwardCompare(index, index2_rowAsc()));
	}
	else {
		std::sort(linearIndex.begin(), linearIndex.end(), make_ForwardCompare(index, index2_colAsc()));
	}

	if (false/*transposed*/)
	{
		int start = -1;

		for (int i = 1; i < currentSize; ++i) {
			// Do comparison
			const index_type *x = (const index_type*) &(index[i-1]);
			const index_type *y = (const index_type*) &(index[i]);
			bool result;

			//if (*(x+1) == *(y+1))
			//	result = (*x < *y);
			//else
			//	result = (*(x+1) < *(y+1));
			if (*x == *y)
				result = (*(x+1) < *(y+1));
			else
				result = (*x < *y);
			if (!result)
			{
				// Blah!
				printf("Sorting %d out of order: [%d,%d] [%d,%d]\n", i, index[i-1][0], index[i-1][1], index[i][0], index[i][1]);
				start = i;
				break;
			}
		}

		if (start > 0)
		{
			start = std::max(0, start-10);
			int end = std::min((int)currentSize-1, start + 20);

			cout << "********************************************************" << endl;
			cout << "Before: ";
			for (int i = start; i < end; ++i)
			{
				printf(" [%d,%d]", index[i][0], index[i][1]);
			}
			cout << endl;
			cout << "After: ";
			for (int i = start; i < end; ++i)
			{
				printf(" [%d,%d]", index[linearIndex[i]][0], index[linearIndex[i]][1]);
			}
			cout << endl;
			cout << "********************************************************" << endl;
		}
	}
#endif

	// count nnz:
	int nnz = currentSize;

	taucs_ccs_matrix *matC;
	if (transposed) {
		matC = taucs_ccs_create(nCols,nRows,nnz,TAUCS_DOUBLE);
	}
	else {
		matC = taucs_ccs_create(nRows,nCols,nnz,TAUCS_DOUBLE);
	}
	if (! matC)
		return NULL;

	// copy cols into matC
	int lastColumn = -1;
	int rowptrC = 0;
	for (int i = 0; i < currentSize; ++i) {
		index_type row, col;
#if 0
		if (transposed) {
			row = index[linearIndex[i]][1];
			col = index[linearIndex[i]][0];
		}
		else {
			row = index[linearIndex[i]][0];
			col = index[linearIndex[i]][1];
		}
#endif
		row = index[i].y;
		col = index[i].x;

		// Check if we're beginning a new column
		if (lastColumn != col) {
			for (int c = lastColumn+1; c <= col; ++c) {
				matC->colptr[c] = rowptrC;
			}
			lastColumn = col;
		}

		// Populate the entry
		matC->rowind[rowptrC] = row;
#if 0
		matC->taucs_values[rowptrC] = data[linearIndex[i]];
#endif
		matC->taucs_values[rowptrC] = data[i];
		++rowptrC;
	}
	if (transposed) {
		for (int c = lastColumn+1; c <= nRows; ++c) {
			matC->colptr[c] = rowptrC;
		}
	}
	else {
		for (int c = lastColumn+1; c <= nCols; ++c) {
			matC->colptr[c] = rowptrC;
		}
	}

	return matC;
}


ErrorTerm::ErrorTerm() : rows(0), rhs(NULL), weight(0)
{

}

ErrorTerm::ErrorTerm( real_type weight, size_type rows, size_type totalNonZero ) :
weight(weight), rows(rows), lhs(totalNonZero), rowWeight(rows, 1.0)   /**< By default, the rows are weight = 1.0 */
{
	rhs = new real_type[rows];
}

ErrorTerm::~ErrorTerm()
{
	clear();
}

void ErrorTerm::clear()
{
	rows = 0;
	weight = 0;
	lhs.clear();
	if (rhs) {
		delete[] rhs;
		rhs = NULL;
	}
	rowWeight.clear();
}

ErrorTerm* Deform::mergeTerms( TermList terms )
{
	// Merge all the terms together
	ErrorTerm *result = new ErrorTerm;

	// Determine the total number of terms
	size_type totalRows, totalNNZ;
	totalNNZ = totalRows = 0;

	for (size_type i = 0; i < terms.size(); i++) {
		totalRows += terms[i]->rows;
		totalNNZ += terms[i]->lhs.size();
	}

	// Allocate the space
	result->weight = 1;
	result->rows = totalRows;
	result->rhs = new real_type[result->rows];
	result->lhs.resize(totalNNZ);

	index2_type idx;
	real_type val;

	// Copy the terms with the weights multiplied
	size_type startRow = 0, trackRow = 0;

	taucs_printf("***** Merging terms:\n");

	for (size_type i = 0; i < terms.size(); i++)
	{
		real_type termWeight = terms[i]->weight;
		real_type rowWeight = 1.0;
		bool rowWeightExists = (terms[i]->rowWeight.size() == terms[i]->rows);
		std::vector<int> trackRowMapping(terms[i]->rows, -1);

		//// Debugging
		//if (rowWeightExists) {
		//	for (int row = 0; row < terms[i]->rows; ++row) {
		//		cout << "Weight of row " << row << ": " << terms[i]->rowWeight[row] << endl;
		//	}
		//}

		for (size_type row = 0; row < terms[i]->rows; row++) {
			if (rowWeightExists) {
				rowWeight = terms[i]->rowWeight[row];
			}
			if (rowWeight > 0.0) {
				result->rhs[trackRow] = termWeight * rowWeight * terms[i]->rhs[row];
				trackRowMapping[row] = trackRow;
				trackRow++;
			}
		}

		//// Debugging
		//for (int row = 0; row < terms[i]->rows; ++row) {
		//	cout << "Mapping " << row << " -> " << trackRowMapping[row] << endl;
		//}

		SparseMatrix<real_type> &thisLHS = terms[i]->lhs;
		for (size_type elt = 0; elt < thisLHS.size(); elt++) {
			thisLHS.get(elt, idx, val);

			if (rowWeightExists) {
				rowWeight = terms[i]->rowWeight[idx.x];
			}

			if (rowWeight > 0.0) {
				// To get the index in the merged system: modify just the row
				//idx[0] += startRow;
				idx.x = trackRowMapping[idx.x];

				// Populate the matrix
				result->lhs.fill(idx.x, idx.y, termWeight * rowWeight * val);
			}
		}

		taucs_printf("***** Term %d: range [%d - %d], filled %d elements.\n", i, startRow, trackRow-1, thisLHS.size());
		startRow = trackRow;
	}

	taucs_printf("***** Finished Merging Terms: %d rows total.\n", startRow);

	////////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < result->rows; ci++) {
	//	cout << "V[" << ci << "] = " << result->rhs[ci] << ";" << endl;
	//}
	//result->lhs.print();
	//cout << "Estimated size for the LHS (all terms): " << totalNNZ
	//	 << " vs. Actual size: " << result->lhs.size() << endl;

	return result;
}


void taucsPrintResultCodeMessage( int result )
{
	char *message;

	switch(result) {
		case TAUCS_SUCCESS:
			message = "*** TAUCS Solve Success!";
			break;
		case TAUCS_ERROR:
			message = "*** TAUCS Error: Generic error.";
			break;
		case TAUCS_ERROR_NOMEM:
			message = "*** TAUCS Error: NOMEM error.";
			break;
		case TAUCS_ERROR_BADARGS:
			message = "*** TAUCS Error: BADARGS error.";
			break;
		case TAUCS_ERROR_MAXDEPTH:
			message = "*** TAUCS Error: MAXDEPTH error.";
			break;
		case TAUCS_ERROR_INDEFINITE:
			message = "*** TAUCS Error: INDEFINITE error.";
			break;
		default:
			break;
	}

	cout << message << endl;
}

void taucsSubmatrixSymmetric( const taucs_ccs_matrix *M, taucs_ccs_matrix *result, std::vector<bool> keep )
{
	const int numCols = M->n;
	
	// Check flags and size
	if (!(M->flags & TAUCS_SYMMETRIC) || (keep.size() != numCols)) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Keep track of new indices
	vector<int> newIndices(keep.size());
	int count = 0;
	for (int i = 0; i < keep.size(); i++) {
		if (keep[i]) {
			newIndices[i] = count;
			count++;
		}
		else {
			newIndices[i] = -1;
		}
	}

	// Create result matrix (assume enough memory is allocated in M)
	int newDim = count;
	result->n = result->m = newDim;

	// Keep selected elements
	count = 0;
	for (int j = 0; j < numCols; j++) {
		// For each column to keep
		if (keep[j]) {

			result->colptr[newIndices[j]] = count;

			for (int li = M->colptr[j]; li < M->colptr[j+1]; li++) {

				int i = M->rowind[li];

				// For each row to keep
				if (keep[i]) {
					// Add the value and row index
					result->taucs_values[count] = M->taucs_values[li];
					result->rowind[count] = newIndices[i];
					count++;
				}
			}
		}
	}
	result->colptr[newDim] = count;

	//cerr << "*** Creating submatrix with dim=" << newDim << " and nnz=" << count << endl;
}

void taucsSubvector( const taucsType *x, taucsType *result, std::vector<bool> keep, int numActive )
{
	// Assume the dimensions match
	// Allocate matrix
	//taucsType *result = new taucsType[numActive];

	// Copy data
	int count = 0;
	for (int i = 0; i < keep.size(); i++) {
		if (keep[i]) {
			result[count] = x[i];
			count++;
		}
	}
}


taucsType computeL2Error( const taucs_ccs_matrix *A, const taucsType *x, const taucsType *b, taucsType *residual )
{
	taucsType L2 = 0.0;
	MulNonSymmMatrixVector(A, x, residual);
	for (int i = 0; i < A->m; i++) {
		residual[i] = residual[i] - b[i];
		L2 += residual[i]*residual[i];
	}

	return sqrt(L2);
}


// For re-weighting A^T * A using a squared diagonal weighting matrix W^2 to get result: A^T * W^2 * A
// Assumes A^T * A was computed already before and just replaces the values
// Assumes W^2 is given by the values in the diagonal
void reweightATA( const taucs_ccs_matrix *matA, const taucs_ccs_matrix *matB, const taucsType *weightSq, taucs_ccs_matrix *result )
{
	// Compatibility of dimensions        
	if ((matA->m != matB->n) || (matA->n != matB->m) || (result->n != result->m))
		return;
	if ((matA->flags & TAUCS_SYMMETRIC) || (matB->flags & TAUCS_LOWER) )
		return;
	if (!(result->flags & TAUCS_SYMMETRIC))
		return;

	// Reset memory to zeros
	memset(result->taucs_values, 0, sizeof(taucsType) * (result->colptr[result->n]));

	// (m x n)*(n x m) = (m x m)
	int m=matA->m;
	int n=matA->n;

	for (int i=0; i<m; ++i) {
		// Creating column i of the result
		int colBegin = result->colptr[i];
		int colEnd = result->colptr[i+1];

		// travel on bi
		for (int rowptrBi = matB->colptr[i];rowptrBi < matB->colptr[i+1];++rowptrBi) {
			int rowInd = matB->rowind[rowptrBi];
			taucsType biv = weightSq[rowInd] * matB->taucs_values[rowptrBi];			/**< Multiply with the weight as well */

			// make biv*a_{rowInd} and insert into column
			// Ignore anything above the diagonal!!
			for (int rowptrA=matA->colptr[rowInd];rowptrA<matA->colptr[rowInd+1];++rowptrA) {
				int rowA = matA->rowind[rowptrA];
				if (rowA >= i) {
					taucsType valA = matA->taucs_values[rowptrA];

					// insert valA*biv into the appropriate location (search for it)
					int rowptrC;
					for ( rowptrC = colBegin; rowptrC < colEnd; ++rowptrC ) {
						// Check if the row location is the same
						if (result->rowind[rowptrC] == rowA) {
							result->taucs_values[rowptrC] += valA*biv;
							break;
						}
					}
					if (rowptrC == colEnd) {
						cerr << "Warning in reweightATA(): couldn't find appropriate location for the value!" << endl;
					}
				}
			}
		}
		// now column i is created
	}
}

// For re-weighting A^T * b using a squared diagonal weighting matrix W^2 to get result: A^T * W^2 * b
// Assumes W^2 is given by the values in the diagonal
void reweightATb(const taucs_ccs_matrix *matA, const taucsType *b, const taucsType *weightSq, taucsType *result)
{
	// make result all zero
	memset(result, 0, matA->m * sizeof(taucsType));

	for (int col = 0; col < matA->n; ++col) {
		// going over column col of matA, multiplying it by b[col] and setting the appropriate values of vector result
		// Also weight vector b according to the weight matrix
		for (int p = matA->colptr[col]; p < matA->colptr[col+1]; ++p) {
			result[matA->rowind[p]] += weightSq[col] * b[col] * matA->taucs_values[p];
		}
	}
}


// Compute the squared re-weighting based on the residual, and also compute and return the L1 norm on the way
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps )
{
	taucsType L1 = 0.0;
	for (int i = 0; i < n; i++) {
		L1 += abs(residual[i]);

		// Compute weight = 1 ./ (1 + (res / eps).^2);
		residual[i] = residual[i] * invEps;
		residual[i] = 1.0 / (1.0 + residual[i]*residual[i]);
	}

	return L1;
}

// Compute the squared re-weighting based on the residual, and also compute and return the L1 norm on the way
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps, const std::vector<bool> &selectiveReweight )
{
	taucsType L1 = 0.0;
	for (int i = 0; i < n; i++) {
		L1 += abs(residual[i]);

		if (selectiveReweight[i]) {
			// Compute weight = 1 ./ (1 + (res / eps).^2);
			residual[i] = residual[i] * invEps;
			residual[i] = 1.0 / (1.0 + residual[i]*residual[i]);
		}
		else {
			residual[i] = 1.0;
		}
	}

	return L1;
}
