#ifndef __TAUCS_UTIL_H__
#define __TAUCS_UTIL_H__

#include <taucsaddon.h>
#include "types.h"

namespace Deform {

	/**
	* Sparse matrix class for generating matrices in CCS format.
	* This is a dynamically growing matrix, with 0-based indexing.
	*/
	template<typename T>
	class SparseMatrix {
	//private:
	//	/**< Total reserved space for storage */
	//	size_type dataSize()
	//	{
	//		if (data.size() != index.size()) {
	//			throw "SparseMatrix<T>::size(): Data and index sizes do not match!";
	//		}

	//		return data.size();
	//	};

	public:
		SparseMatrix() : maxRow(0), maxCol(0), currentSize(0)
		{
			data.clear();
			index.clear();
		};

		SparseMatrix( size_type nnz ): maxRow(0), maxCol(0), currentSize(0), data(nnz), index(nnz)
		{

		};

		/**< Number of non-zero elements */
		size_type size() {
			return currentSize;
		};

		/**< Reserve space for n elements */
		void resize(size_type n)
		{
			clear();
			data.resize(n);
			index.resize(n);
		};

		void clear()
		{
			data.clear();
			index.clear();
			maxCol = maxRow = currentSize = 0;
		};

		void fill(index_type row, index_type col, T val)
		{
			// Push back the value and the index
			if (currentSize < data.size()) {
				// Just fill in the value
				data[currentSize] = val;

				index2_type &store = index[currentSize];
				store.x = row;
				store.y = col;
			}
			else {
				index2_type temp;
				temp.x = row;
				temp.y = col;

				data.push_back(val);
				index.push_back(temp);
			}
			if (row > maxRow) {
				maxRow = row;
			}
			if (col > maxCol) {
				maxCol = col;
			}

			currentSize++;
		};

		void get(size_type seq, index2_type &idx, T &val) const
		{
			if (seq >= currentSize) {
				throw "SparseMatrix<T>::get(): non-zero element out of range.";
			}
			idx = index[seq];
			val = data[seq];
		};

		void print();
		void write( const char *filename ) const;
		void getTaucsColumns( std::vector< std::map<int,taucsType> > &result, size_type nCols ) const;
		taucs_ccs_matrix * writeTaucsMatrix(size_type nRows, size_type nCols, bool transposed = false ) const;

		inline size_type getMaxRow() const {
			return maxRow;
		}

		inline size_type getMaxCol() const {
			return maxCol;
		}

	private:
		size_type currentSize;			/**< Keep track of the number of non-zero elements */
		size_type maxRow;
		size_type maxCol;
		std::vector<T> data;
		std::vector<index2_type> index;
	};

	/**
	* Class for holding each error term.
	*/
	class ErrorTerm {
	public:
		ErrorTerm();
		ErrorTerm( real_type weight, size_type rows, size_type totalNonZero );
		~ErrorTerm();
		void clear();

		SparseMatrix<real_type> lhs;
		real_type *rhs;
		size_type rows;
		real_type weight;
		std::vector<real_type> rowWeight;
	};

	typedef std::vector<ErrorTerm*> TermList;

	// Merge multiple error terms
	ErrorTerm* mergeTerms(TermList terms);

}

// Utility functions
void taucsPrintResultCodeMessage( int result );

// Keep only selected rows and columns from the symmetric matrix M
void taucsSubmatrixSymmetric( const taucs_ccs_matrix *M, taucs_ccs_matrix *result, std::vector<bool> keep );
void taucsSubvector( const taucsType *x, taucsType *result, std::vector<bool> keep, int numActive );
taucsType computeL2Error( const taucs_ccs_matrix *A, const taucsType *x, const taucsType *b, taucsType *residual );
void reweightATA( const taucs_ccs_matrix *matA, const taucs_ccs_matrix *matB, const taucsType *weightSq, taucs_ccs_matrix *result );
void reweightATb(const taucs_ccs_matrix *matA, const taucsType *b, const taucsType *weightSq, taucsType * result);
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps );
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps, const std::vector<bool> &selectiveReweight );

#endif
