#ifndef __TYPES_H__
#define __TYPES_H__

#include <SylvainTools/geom.h>
#include <functional>
#include <cfloat>

#undef min
#undef max

// Type definitions
typedef float real_type;
typedef Geometry::Vec2<real_type> real2_type;
typedef Geometry::Vec3<real_type> real3_type;
typedef Geometry::Vec<4,real_type> real4_type;
typedef Geometry::Square_matrix<3,real_type> matrix3_type;

// Unsigned index type
typedef unsigned int index_type;
//typedef Geometry::Vec2<index_type> index2_type;
typedef struct _index2_type index2_type;
struct _index2_type
{
	index_type x, y;
};

typedef Geometry::Vec3<index_type> index3_type;

// Signed index type
typedef Geometry::Vec2<int> int2_type;
typedef Geometry::Vec3<int> int3_type;

// Size type
typedef unsigned int size_type;

typedef std::vector<index_type> idxList;
typedef std::vector<real_type> distList;
typedef std::vector<index3_type> idx3List;
typedef std::vector<int3_type> int3List;
typedef std::vector<real3_type> ptList;

#define REAL_TYPE_EPSILON FLT_EPSILON
#define REAL_TYPE_MAX FLT_MAX
#define REAL_TYPE_QNAN std::numeric_limits<float>::quiet_NaN()
#define REAL_TYPE_INF std::numeric_limits<float>::infinity()

typedef real3_type RealColor;
typedef Geometry::Vec3i IntColor;


// int3_type comparison operator
struct int3Comp : public std::binary_function<int3_type,int3_type,bool> {
	bool operator()(const int3_type &x, const int3_type &y) const {
		// Dictionary Ordering
		if (x[0] == y[0])
			if (x[1] == y[1])
				return (x[2] < y[2]);
			else
				return (x[1] < y[1]);
		else
			return (x[0] < y[0]);
	}
};


// Data type definition.
class data_type {
public:
	inline data_type(int d) : val(d, 0.0) { }

	inline data_type(int d, real_type initialValue) : val(d, initialValue) { }

	inline void setZero() {
		for (int i = 0; i < val.size(); ++i) {
			val[i] = 0.0;
		}
	}

	inline bool checkFinite() {
		for (int i = 0; i < val.size(); ++i) {
			if (!_finite(val[i])) {
				return false;
			}
		}
		return true;
	}

	inline data_type& operator+=(const data_type& v) {
		for (unsigned int i = 0; i < val.size(); ++i) {
			val[i] += v.val[i];
		}
		return *this;
	}

	inline data_type& operator*=(const real_type r) {
		for (unsigned int i = 0; i < val.size(); ++i) {
			val[i] *= r;
		}
		return *this;
	}

	inline int normL0() {
		int total = 0;
		for (unsigned int i = 0; i < val.size(); ++i) {
			if (abs(val[i]) > 0.0) {
				++total;
			}
		}

		return total;
	}

	inline void normalizeL1() {
		real_type total = normL1();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); ++i) {
				val[i] /= total;
			}
		}
	}

	inline real_type normL1() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); ++i) {
			total += abs(val[i]);
		}

		return total;
	}

	inline void normalizeL2() {
		real_type total = normL2();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); ++i) {
				val[i] /= total;
			}
		}
	}

	inline real_type normSqrL2() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); ++i) {
			total += val[i]*val[i];
		}

		return total;
	}

	inline real_type normL2() {
		return sqrt(normSqrL2());
	}

	inline real_type findMin( int &minIndex ) const
	{
		real_type minVal = REAL_TYPE_MAX;

		for (int i = 0; i < val.size(); ++i) {
			if (val[i] < minVal) {
				minVal = val[i];
				minIndex = i;
			}
		}

		return minVal;
	}

	inline real_type findMax( int &maxIndex ) const
	{
		real_type maxVal = -REAL_TYPE_MAX;

		for (int i = 0; i < val.size(); ++i) {
			if (val[i] > maxVal) {
				maxVal = val[i];
				maxIndex = i;
			}
		}

		return maxVal;
	}

	std::vector<real_type> val;
};

/**
 * Add two data vectors.
 */
inline data_type operator+(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); ++i) {
		res.val[i] = v1.val[i] + v2.val[i];
	}
	return res;
}

/**
* Subtract two data vectors.
*/
inline data_type operator-(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); ++i) {
		res.val[i] = v1.val[i] - v2.val[i];
	}
	return res;
}

/**
 * Component-wise multiply.
 */
inline data_type operator*(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); ++i) {
		res.val[i] = v1.val[i] * v2.val[i];
	}
	return res;
}

/**
 * Component-wise divide.
 */
inline data_type operator/(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); ++i) {
		res.val[i] = v1.val[i] / v2.val[i];
	}
	return res;
}


/**
 * Dot product.
 */
inline real_type dot(const data_type& v1, const data_type& v2) {
	// No size checking here!
	real_type res = 0;
	for (unsigned int i = 0; i < v1.val.size(); ++i) {
		res += v1.val[i] * v2.val[i];
	}
	return res;
}

inline data_type operator*(const real_type r, const data_type& v) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); ++i) {
		res.val[i] = r * v.val[i];
	}
	return res;
}

inline data_type operator*(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); ++i) {
		res.val[i] = v.val[i] * r;
	}
	return res;
}

inline data_type operator/(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); ++i) {
		res.val[i] = v.val[i] / r;
	}
	return res;
}


inline std::ostream& operator<<(std::ostream& out, const data_type &v) {
	out << "[";
	for (unsigned int i = 0; i < v.val.size(); ++i) {
		if (i > 0) {
			out << ",";
		}
		out << v.val[i];
	}
	out << "]";
	return out;
}

//inline QTextStream& operator<<(QTextStream& out, const data_type &v) {
//	out << "[";
//	for (int i = 0; i < v.val.size(); ++i) {
//		if (i > 0) {
//			out << ", ";
//		}
//		out << v.val[i];
//	}
//	out << "]";
//	return out;
//}
//


#endif
