#include <iomanip>
#include <fstream>
#include <sstream>
#include <QRegExp>
#include <QFile>
#include <QScrollBar>
#include <QLineEdit>
#include <glut.h>
#include "util.h"
#include "PointList.h"
#include <sys/stat.h>


extern "C" {
#include <SFMT/SFMT.h>
}


//stdext::hash_map<std::string, TimerPair> perfTimers;
std::vector<TimerPair> timerArray;

const char* timer_str_array[] =
{
	"TIMER_BLANK",
	"TIMER_MULTIFRAME_INITIALIZE",
	"TIMER_MULTIFRAME_GLOBAL_REG",
	"TIMER_MULTIFRAME_RESAMPLE_ASG",
	"TIMER_SIMULTANEOUS_T_STEP_1",
	"TIMER_SIMULTANEOUS_T_STEP_2",
	"TIMER_SIMULTANEOUS_T_STEP_3",
	"TIMER_SIMULTANEOUS_W_STEP_1",
	"TIMER_SIMULTANEOUS_W_STEP_2",
	"TIMER_SIMULTANEOUS_W_STEP_3",
	//
	//
	"TIMER_RANGE_QUERY",
	"TIMER_FIT_CIRCLE",
	"TIMER_POINTLIST_ANN_QUERY",
	"TIMER_POINTLIST_N_ANN_QUERY",
	//
	//
	"TIMER_SAMPLING_SPEED_TRIAL",
	"TIMER_SUB_SAMPLING_GEOMETRY",
	"TIMER_POPULATING_CORRESPONDENCE_CANDIDATES",
	"TIMER_MEAN_SHIFT_CLUSTERING",
	"TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP",
	"TIMER_PRECOMPUTING_CLOSEST_POINTS",
	"TIMER_GRAPH_CUT_OPTIMIZATION",
	//
	//
	"TIMER_RANSAC_TOTAL_OUTLIERS",
	"TIMER_RANSAC_TOTAL_CANDIDATES",
	//
	//
	"TIMER_NUM",
};


/**
* Write the contents of a vector to the stream.
*/
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v) {
	out << "[";
	for (int i = 0; i < 3; i++) {
		if (i != 0) out << ",";
		//out << std::setprecision(2) << setiosflags(std::ios_base::fixed)<< v[i] << ",";
		//out << std::setprecision(5) << setiosflags(std::ios_base::scientific)<< v[i] << ",";
		out << v[i];
	}
	out << "]";
	return out;
}

QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v) {
	out << "[" << v[0] << "," << v[1] << "," << v[2] << "]";
	return out;
}

template<typename T>
std::ostream& operator<<( std::ostream& out, const std::vector<T> &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

template<typename T>
QTextStream& operator<<( QTextStream& out, const std::vector<T> &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

// Forward declarations for the compiler to produce code
template std::ostream& operator<< <unsigned int>( std::ostream& out, const std::vector<unsigned int> &v );
template std::ostream& operator<< <int>( std::ostream& out, const std::vector<int> &v );
template std::ostream& operator<< <float>( std::ostream& out, const std::vector<float> &v );
template std::ostream& operator<< <double>( std::ostream& out, const std::vector<double> &v );
template QTextStream& operator<< <int>( QTextStream& out, const std::vector<int> &v );
template QTextStream& operator<< <float>( QTextStream& out, const std::vector<float> &v );

// Specialization for printing a boolean array
template<> std::ostream& operator<< <bool> ( std::ostream& out, const std::vector<bool> &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << ( v[i] ? "T" : "F" );
	}
	out << "]";
	return out;
}


// Grabbed from pbrt 1.03, core/mc.cpp
void ComputeStep1dCDF(float *f, int nSteps, float *c, float *cdf) {
	// Compute integral of step function at $x_i$
	int i;
	cdf[0] = 0.;
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] = cdf[i-1] + f[i-1] / nSteps;
	// Transform step function integral into cdf
	*c = cdf[nSteps];
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] /= *c;
}

// Grabbed from pbrt 1.03, core/mc.cpp
float SampleStep1d(float *f, float *cdf, float c, int nSteps, float u, float *pdf) {
	// Find surrounding cdf segments
	float *ptr = std::lower_bound(cdf, cdf+nSteps+1, u);
	int offset = (int) (ptr-cdf-1);
	// Return offset along current cdf segment
	u = (u - cdf[offset]) / (cdf[offset+1] - cdf[offset]);
	*pdf = f[offset] / c;
	return (offset + u) / nSteps;
}


void setOpenGLColor( const FloatArray &color, real_type alpha/* = 1.0*/ )
{
	if (color.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	glColor4f(color[0], color[1], color[2], alpha);
}

float interpSlider(QScrollBar *slider) {
	return ((float)slider->value()) / ((float)(slider->maximum() - slider->minimum()));
}

void setInterpSlider(QScrollBar *slider, float interp) {
	slider->setValue((slider->maximum() - slider->minimum())*interp + slider->minimum());
}

float QLineEditFloatValue(QLineEdit *edit) {
	bool ok;
	float value = edit->text().toFloat(&ok);

	if (!ok) {
		// Just set the value to zero
		value = 0.f;
		edit->setText(QString("%1").arg(value));
	}

	return value;
}

void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr) {
	// Print
	out << "R{" << ind+1 << "} = ";
	out << "[" << rot(0,0) << ", " << rot(0,1) << ", " << rot(0,2) << "\n        ";
	out <<        rot(1,0) << ", " << rot(1,1) << ", " << rot(1,2) << "\n        ";
	out <<        rot(2,0) << ", " << rot(2,1) << ", " << rot(2,2) << "];\n";

	out << "T{" << ind+1 << "} = ";
	out << "[" << tr[0] << ", " << tr[1] << ", " << tr[2] << "];\n";
	out << "\n";
}

void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos )
{
	real3_type centerPos = (minPos + maxPos) / 2;

	glBegin(GL_LINES);
	// diagonal
	//glVertex3f(minPos[0], minPos[1], minPos[2]);
	//glVertex3f(maxPos[0], maxPos[1], maxPos[2]);

	// 3 lines from min pos
	//glColor3f(1.f,0.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);

	//glColor3f(0.f,1.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], minPos[2]);

	//glColor3f(0.f,0.f,1.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with x set to max pos
	//glColor3f(1.f,1.f,1.f);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with y set to max pos
	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 2 lines from min pos with z set to max pos
	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 3 lines from max pos
	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);
	glEnd();	// End of: bounding box drawing

	//glBegin(GL_LINES);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0]-1, centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1]-1, centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]-1);
	//glEnd();
}

void drawSphere( const real3_type &pos, const float radius, GLint slices/* = 10*/, GLint stacks/* = 10*/ )
{
	glPushMatrix();
		glTranslatef(pos[0], pos[1], pos[2]);
		glutSolidSphere(radius, slices, stacks);
	glPopMatrix();
}

void drawBoxFilled( const real3_type& min_pos, const real3_type& max_pos )
{
	glBegin(GL_QUADS);
	// Front Face
	glNormal3f( 0.0f, 0.0f, 1.0f);// Normal Pointing Towards Viewer
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 1 (Front)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 2 (Front)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Front)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 4 (Front)
	// Back Face
	glNormal3f( 0.0f, 0.0f,-1.0f);// Normal Pointing Away From Viewer
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Back)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 2 (Back)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 3 (Back)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 4 (Back)
	// Top Face
	glNormal3f( 0.0f, 1.0f, 0.0f);// Normal Pointing Up
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 1 (Top)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 2 (Top)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Top)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 4 (Top)
	// Bottom Face
	glNormal3f( 0.0f,-1.0f, 0.0f);// Normal Pointing Down
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Bottom)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 2 (Bottom)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 3 (Bottom)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 4 (Bottom)
	// Right face
	glNormal3f( 1.0f, 0.0f, 0.0f);// Normal Pointing Right
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 1 (Right)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 2 (Right)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Right)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 4 (Right)
	// Left Face
	glNormal3f(-1.0f, 0.0f, 0.0f);// Normal Pointing Left
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Left)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 2 (Left)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 3 (Left)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 4 (Left)
	glEnd();// Done Drawing Quads
}


Geometry::Vec3f VecToVec3f( qglviewer::Vec v )
{
	return Geometry::Vec3f(v[0], v[1], v[2]);
}

qglviewer::Vec Vec3fToVec( Geometry::Vec3f v )
{
	return qglviewer::Vec(v[0], v[1], v[2]);
}

void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha /*= 1.f*/ )
{
	if (floatColor.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha /*= 1.f */ )
{
	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void TextProgress( double x, bool init/* = false*/ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;
			std::cout << "\b\b\b\b";
			if (currentBin >= 1000.1) {
				std::cout << "\b";
			}
			if (currentBin >= 10000.1) {
				std::cout << "\b";
			}
			std::cout << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}

void TextProgressWithLabel( double x, bool init /*= false*/, QString id /*= QString() */ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;
	int idLength = id.length();

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;

			// Backspace
			for (int i = 0; i < idLength+10; i++) {
				std::cout << "\b";
			}

			// Print out next progress
			std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}


Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle) {
	float sinAngle = sqrtf(1.f - cosAngle*cosAngle);

	// use rodrigues' formula
	Geometry::Matrix33f ssm;
	ssm(0,0) = 0.f;   ssm(0,1) = -axis[2]; ssm(0,2) = axis[1];
	ssm(1,0) = axis[2];  ssm(1,1) = 0.f;   ssm(1,2) = -axis[0];
	ssm(2,0) = -axis[1]; ssm(2,1) = axis[0];  ssm(2,2) = 0.f;

	Geometry::Matrix33f rot, id;
	id = Geometry::Matrix33f::identity();
	rot = id + ssm*sinAngle + ssm*ssm*(1.f - cosAngle);

	return rot;
}

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			temp = m(i,j);
			out.write((char*) &temp, sizeof(float));
		}
	}
}

void writeBinary( std::ofstream &out, const Geometry::Vec3f v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		temp = v[i];
		out.write((char*) &temp, sizeof(float));
	}
}

void readBinary( std::ifstream &in, Geometry::Matrix33f &m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			in.read((char*) &temp, sizeof(float));
			m(i,j) = temp;
		}
	}
}

void readBinary( std::ifstream &in, Geometry::Vec3f &v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		in.read((char*) &temp, sizeof(float));
		v[i] = temp;
	}
}

void bestCandidateSample_rebuildKNN( const PointList &pts, int n, idxList& indices )
{
	indices.resize(n);

	int pick, bestPick;
	idxList nearest;
	distList dist;
	real_type maxDist;
	const int pickNum = 10;
	PointList pickedPts;
	int remainingPoints = pts.size();

	// Initialize array marking which vertices are picked
	std::vector<bool> markedPicked;
	std::vector<int> pickCandidates;
	markedPicked.resize(pts.size(), false);
	pickCandidates.resize(pickNum);

	// Initialize by adding one random point
	pick = floor(((double)pts.size()) * genrand_real1());
	pickedPts.clear();
	pickedPts.reserve(n);
	pickedPts.add(pts[pick]);
	markedPicked[pick] = true;
	indices[0] = pick;

	// Remaining samples
	for (index_type i = 1; i < n; i++) {

		maxDist = -REAL_TYPE_MAX;
		for (int j = 0; j < pickNum; j++) {

			// 1. Pick candidate
			do {
				pick = floor(((double)pts.size()) * genrand_real1());
			} while (markedPicked[pick]);

			pickCandidates[j] = pick;
			markedPicked[pick] = true;

			// 2. Compute distance to current selected point set
			pickedPts.queryKNN(pts[pick], 1, nearest, dist, true);

			if (dist[0] > maxDist) {
				maxDist = dist[0];
				bestPick = pick;
			}
		}

		// Add the best candidate
		pickedPts.add(pts[bestPick]);
		indices[i] = bestPick;
		--remainingPoints;
		//pickedPts.clearANNData();

		// Un-mark the discarded candidates
		for (int j = 0; j < pickNum; j++) {
			pick = pickCandidates[j];
			if (pick != bestPick) {
				markedPicked[pick] = false;
			}
		}
	}
}

void bestCandidateSampleDist_rebuildKNN( const PointList &pts, const real_type sampleDist, idxList& indices, int nCache )
{
	indices.clear();

	// In this implementation we are going to rebuild the KNN
	// only after adding a certain amount of points, say 50-100 pts.
	// In the mean time, we have a cache of the 50-100 pts, and the distance
	// to these points are computed naively (comparing every single point).
	// This should perform much faster than rebuilding the KNN every
	// time after adding a point.
	if (nCache <= 0) {
		nCache = 1;
	}

	int pick, bestPick;
	const int pickNum = 10;
	const int avgNum = nCache;			// Number of trials to average the candidate distance
	const real_type avgDistCorrection = 1.75;		// TODO: Why does this constant make sense?!?

	// Initialize array marking which vertices are picked
	PointList pickedPts;
	const PointList& pickedPts_const = pickedPts;
	std::vector<bool> markedPicked;
	std::vector<int> pickCandidates;
	std::vector<real_type> lastCandidateDistances;
	real_type avgLastCandidateDistances = REAL_TYPE_MAX;
	markedPicked.resize(pts.size(), false);
	pickCandidates.resize(pickNum);
	lastCandidateDistances.resize(avgNum, REAL_TYPE_MAX);		// This will be a circular buffer

	// Initialize picked point cache
	PointList pickedPtCache(nCache);
	int numCached = 0;

	// Initialize by adding one random point
	pick = floor(((double)pts.size()) * genrand_real1());
	pickedPtCache[numCached++] = pts[pick];
	markedPicked[pick] = true;
	indices.push_back(pick);

	// Remaining samples
	while (avgLastCandidateDistances > sampleDist) {
		// If the cache is full, dump the points into the picked point pool
		if (numCached >= nCache) {
			pickedPts.add(pickedPtCache);
			numCached = 0;
		}

		real_type maxDist = -REAL_TYPE_MAX;
		for (int j = 0; j < pickNum; j++) {
			// 1. Pick candidate
			do {
				pick = floor(((double)pts.size()) * genrand_real1());
			} while (markedPicked[pick]);

			pickCandidates[j] = pick;
			markedPicked[pick] = true;

			// 2. Compute distance to current selected point set
			real_type closestDist = REAL_TYPE_MAX;
			if (pickedPts_const.size() > 0) {
				idxList nearest;
				distList dists;
				pickedPts_const.queryKNN(pts[pick], 1, nearest, dists, true);
				closestDist = dists[0];
			}

			// 3. Compute distance to each point in cache
			for (int k = 0; k < numCached; ++k) {
				real_type temp = (pts[pick] - pickedPtCache[k]).norm();
				if (temp < closestDist) {
					closestDist = temp;
				}
			}

			// 4. Compare and store maximum
			if (closestDist > maxDist) {
				maxDist = closestDist;
				bestPick = pick;
			}
		}

		// Add the best candidate
		pickedPtCache[numCached++] = pts[bestPick];
		indices.push_back(bestPick);

		// Un-mark the discarded candidates
		for (int j = 0; j < pickNum; j++) {
			pick = pickCandidates[j];
			if (pick != bestPick) {
				markedPicked[pick] = false;
			}
		}

		// Update the avg of lastCandidateDistances
		{
			const int numPickedSoFar = numCached + pickedPts_const.size() - 1;
			const int bufferIdx = (numPickedSoFar - 1) % avgNum;
			const real_type savedLastDist = lastCandidateDistances[bufferIdx];
			lastCandidateDistances[bufferIdx] = maxDist;

			// Compute average
			if (numPickedSoFar == avgNum)
			{
				avgLastCandidateDistances = 0.0;
				for (int j = 0; j < avgNum; ++j) {
					avgLastCandidateDistances += lastCandidateDistances[j];
				}
				avgLastCandidateDistances /= (real_type) avgNum;
				avgLastCandidateDistances *= avgDistCorrection;
			}
			else if (numPickedSoFar > avgNum)
			{
				avgLastCandidateDistances += avgDistCorrection * (maxDist - savedLastDist) / ((real_type) avgNum);
			}

			////////////////////////////////////////////////////////////////////////////////////////////////
			// Debugging: compare to the actual answer
			if (false && numPickedSoFar >= avgNum)
			{
				real_type maxActualDist = -REAL_TYPE_MAX;
				for (int j = 0, j_end = pts.size(); j < j_end; j++) {
					real_type closestDist = REAL_TYPE_MAX;

					// A. Compute distance to current selected point set
					if (pickedPts_const.size() > 0) {
						idxList nearest;
						distList dists;
						pickedPts_const.queryKNN(pts[j], 1, nearest, dists, true);
						closestDist = dists[0];
					}

					// B. Compute distance to each point in cache
					for (int k = 0; k < numCached; ++k) {
						real_type temp = (pts[j] - pickedPtCache[k]).norm();
						if (temp < closestDist) {
							closestDist = temp;
						}
					}

					// C. Compare and store maximum
					if (closestDist > maxActualDist) {
						maxActualDist = closestDist;
					}
				}

				std::cout << "Compare (Approx/Actual): " << avgLastCandidateDistances << ", " << maxActualDist << std::endl;
			}
			////////////////////////////////////////////////////////////////////////////////////////////////
		}
	}
}


idxList bestCandidateSampleDist( const PointList &pts, const real_type sampleDist )
{
	idxList indices;

	bestCandidateSampleDist_rebuildKNN(pts, sampleDist, indices, 100);

	// Sort indices
	sort(indices.begin(), indices.end());

	return indices;
}

void bestCandidateSample_rebuildKNN( const PointList &pts, int n, idxList& indices, int nCache )
{
	indices.resize(n);

	// In this implementation we are going to rebuild the KNN
	// only after adding a certain amount of points, say 50-100 pts.
	// In the mean time, we have a cache of the 50-100 pts, and the distance
	// to these points are computed naively (comparing every single point).
	// This should perform much faster than rebuilding the KNN every
	// time after adding a point.
	if (nCache <= 0) {
		nCache = 1;
	}

	int pick, bestPick;
	const int pickNum = 10;
	int remainingPoints = pts.size();

	// Initialize array marking which vertices are picked
	PointList pickedPts;
	std::vector<bool> markedPicked;
	std::vector<int> pickCandidates;
	pickedPts.reserve(n);
	markedPicked.resize(pts.size(), false);
	pickCandidates.resize(pickNum);

	// Initialize picked point cache
	PointList pickedPtCache(nCache);
	int numCached = 0;

	// Initialize by adding one random point
	pick = floor(((double)pts.size()) * genrand_real1());
	pickedPtCache[numCached++] = pts[pick];
	markedPicked[pick] = true;
	indices[0] = pick;

	// Remaining samples
	const PointList& pickedPts_const = pickedPts;
	for (index_type i = 1; i < n; i++) {
		// If the cache is full, dump the points into the picked point pool
		if (numCached >= nCache) {
			pickedPts.add(pickedPtCache);
			numCached = 0;
		}

		real_type maxDist = -REAL_TYPE_MAX;
		for (int j = 0; j < pickNum; j++) {
			// 1. Pick candidate
			do {
				pick = floor(((double)pts.size()) * genrand_real1());
			} while (markedPicked[pick]);

			pickCandidates[j] = pick;
			markedPicked[pick] = true;

			// 2. Compute distance to current selected point set
			real_type closestDist = REAL_TYPE_MAX;
			if (pickedPts_const.size() > 0) {
				idxList nearest;
				distList dists;
				pickedPts_const.queryKNN(pts[pick], 1, nearest, dists, true);
				closestDist = dists[0];
			}

			// 3. Compute distance to each point in cache
			for (int k = 0; k < numCached; ++k) {
				real_type temp = (pts[pick] - pickedPtCache[k]).norm();
				if (temp < closestDist) {
					closestDist = temp;
				}
			}

			// 4. Compare and store maximum
			if (closestDist > maxDist) {
				maxDist = closestDist;
				bestPick = pick;
			}
		}

		// Add the best candidate
		pickedPtCache[numCached++] = pts[bestPick];
		indices[i] = bestPick;
		--remainingPoints;

		// Un-mark the discarded candidates
		for (int j = 0; j < pickNum; j++) {
			pick = pickCandidates[j];
			if (pick != bestPick) {
				markedPicked[pick] = false;
			}
		}
	}
}

idxList bestCandidateSample( const PointList &pts, int n )
{
	idxList indices;

	if ((n <= 0) || (n >= pts.size()) || (pts.size() <= 10)) {
		// Just use all the vertices in the mesh
		indices.resize(pts.size());
		for (index_type i = 0; i < pts.size(); i++) {
			indices[i] = i;
		}
	}
	else {
		// Use a cache of 100 points.  This seems to work well in some speed tests
		// that sampled 20% of the points in 2 robot and 2 car range scans.
		bestCandidateSample_rebuildKNN(pts, n, indices, 100);

		////////////////////////////////////////////////////////////////////////////////////////////////
		// Debugging code below
		if (false) {
		////////////////////////////////////////////////////////////////////////////////////////////////
			int testSeed = 32098475;
			int nTrials = 100;
			int trialIncrement = 5;
			distList trialSpeed(nTrials);
			idxList cacheSizes(nTrials);
			real_type refTrialSpeed;

			std::cout << "Trial #1 of sampling " << n << " points..." << std::endl;
			init_gen_rand(testSeed);
			MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,Start);
			bestCandidateSample_rebuildKNN(pts, n, indices);
			MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,Stop);
			refTrialSpeed = MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,GetDurationInSecs);

			for (int i = 0; i < nTrials; ++i) {
				idxList temp;
				int cacheSize = i * trialIncrement;

				std::cout << "Trial #2 of sampling " << n << " points... with cache size " << cacheSize << std::endl;
				init_gen_rand(testSeed);
				MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,Start);
				bestCandidateSample_rebuildKNN(pts, n, temp, cacheSize);
				MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,Stop);
				trialSpeed[i] = MEASURE_PERF_NARG(TIMER_SAMPLING_SPEED_TRIAL,GetDurationInSecs);
				cacheSizes[i] = cacheSize;

				std::cout << "Comparing... ";
				bool same = true;
				for (int j = 0; j < n; ++j) {
					if (indices[j] != temp[j]) {
						same = false;
						break;
					}
				}
				if (same) {
					std::cout << "same!" << std::endl;
				}
				else {
					std::cout << "different!" << std::endl;
					pauseUntilEnter("Temp");
				}
			}

			// Print stats summary
			std::cout << "Ref trial speed: " << refTrialSpeed << " secs" << std::endl;
			for (int i = 0; i < nTrials; ++i) {
				std::cout << "Trial " << i << " (cache " << cacheSizes[i] << "): " << trialSpeed[i] << " secs" << std::endl;
			}

			pauseUntilEnter("Temp");
		}
	}

	// Sort indices
	sort(indices.begin(), indices.end());

	return indices;
}

void generateLightColors( int n, std::vector<RealColor> &colors )
{
	// Generate a set of random light colors using best candidate sampling
	const real_type maxThreshold = 0.8;
	const int numTrials = 10;

	colors.resize(n);
	for (int i = 0; i < n; i++) {
		real3_type temp;

		real_type maxDist = -REAL_TYPE_MAX;
		real3_type maxColor;

		for (int j = 0; j < numTrials; ++j) {
			// Use rejection sampling:  Just make sure that the maximum rgb value is at least ~0.7
			real_type maxValue;
			do {
				temp[0] = genrand_real1();
				temp[1] = genrand_real1();
				temp[2] = genrand_real1();
				maxValue = std::max(temp[0], std::max(temp[1],temp[2]));
			} while ( maxValue < maxThreshold );

			if (i == 0) {
				maxColor = temp;
				break;
			}
			// Compute the minimum distance to all picked colors
			real_type distToColors = REAL_TYPE_MAX;
			for (int k = 0; k < i; ++k) {
				real_type dist = (temp - colors[k]).square_norm();
				if (dist < distToColors) {
					distToColors = dist;
				}
			}

			// Keep track of the maximum of the minimum distance
			if (distToColors > maxDist) {
				maxDist = distToColors;
				maxColor = temp;
			}
		}

		colors[i] = maxColor;
	}
}


void writeVector( double* x, int size, const char *filename )
{
	std::ofstream out;
	out.open(filename);

	if (out.fail()) {
		std::cerr << "Error in writeVector(): Could not open output file " << filename << " for writing!" << std::endl;
		return;
	}

	// Write data
	for (index_type i = 0; i < size; i++) {
		out << x[i] << std::endl;
	}
	out.close();
}

void pauseUntilEnter( const char *message )
{
#ifdef ENABLE_PAUSE_UNTIL_ENTER
	std::cin.clear();
	std::cout << message << " (Press enter/return to continue)...";
	std::cin.ignore(1,0);
#endif
}

void replaceField( std::string& str, char field_delim, int num )
{
	int idx = str.find(field_delim);
	if (idx != -1) {
		// Replace consecutive field delimiter symbols with the frame number
		int fieldLen = 1;
		while (((idx+fieldLen) < str.length()) && (str[idx+fieldLen] == field_delim)) {
			++fieldLen;
		}

		// Output a formatted string
		std::ostringstream out;
		out << str.substr(0, idx);
		out.width(fieldLen);
		out.fill('0');
		out << num;
		out.width(0);
		out << str.substr(idx+fieldLen);

		// Replace the string
		str = out.str();
	}

	//// Testing for replaceField()
	//if (!deformOptions.autoSaveSolverState.empty()) {
	//	for (int i = 0; i <= 100; i += 5) {
	//		string s = deformOptions.autoSaveSolverState;
	//		replaceField(s, '#', i);
	//		cout << "Filename: " << s << endl;
	//	}
	//}
	//return;
}

bool getLastDigit( QString str, int& number, int& digits )
{
	// Setup replacing the last digit in the filename
	QRegExp rx2("(\\d+)");
	QString match;
	int newPos, pos = 0;

	while ((newPos = rx2.indexIn(str, pos+match.length())) != -1) {
		match = rx2.cap(1);
		pos = newPos;
	}

	if (!match.isEmpty()) {
		number = match.toInt();
		digits = match.length();
		//format = dir + file.left(pos) + QString("%1") + file.right(file.length() - pos - numDigits);

		//cout << "Match at " << numPos << ": " << qPrintable(match) << endl;
		//cout << "Format: " << qPrintable(format) << endl;
		return true;
	}
	else {
		// there was no match
		//startFrame = 0;
		//format = filename;
		return false;
	}
}

RealColor OpenGLWeightColor( real_type value )
{
	RealColor color;

	if ((value >= 0.0) && (value <= 1.0)) {
		// We have nice weights between 0 and 1 -- color it gray
		// Darker shade: weights closer to 1
		// Brighter shade: weights closer to 0
		value = 1.0 - value;
		color[0] = color[1] = color[2] = value;
	}
	else if (value < 0.0) {
		// If we have negative weights -- color it red
		// Darker red: less negative
		// Brighter red: more negative
		color[0] = -value;
		color[1] = 0.f;
		color[2] = 0.f;
	}
	else {
		// If our weights are greater than 1 -- color it blue
		// Darker blue: greater than 1 a little
		// Brighter blue: greater than 1 a lot
		color[0] = 0.f;
		color[1] = 0.f;
		color[2] = value - 1.f;
	}

	return color;
}

void setOpenGLWeightColor( real_type value, real_type alpha /*= 1.0f */ )
{
	RealColor color = OpenGLWeightColor(value);
	glColor4f(color[0], color[1], color[2], alpha);
}

template<typename T>
void sortedListUnion( const std::vector<T> &first, const std::vector<T> &second, std::vector<T> &result )
{
	// Make sure we have enough space in the destination
	result.resize(first.size() + second.size());

	// Perform the union between the two sorted lists
	std::vector<T>::iterator it = set_union(first.begin(), first.end(), second.begin(), second.end(), result.begin());

	// Size the result down to the right size
	result.resize(std::vector<T>::size_type(it - result.begin()));
	std::vector<T>(result).swap(result);		/**< Optional step: Deallocate memory (http://www.gotw.ca/gotw/054.htm) */
}

template<typename T>
void sortedListUnionMerge( const std::vector<T> &mergeList, std::vector<T> &to )
{
	std::vector<T> temp;
	sortedListUnion(mergeList, to, temp);
	to.swap(temp);
}

template void sortedListUnion<int>( const std::vector<int> &first, const std::vector<int> &second, std::vector<int> &result );
template void sortedListUnionMerge<int>( const std::vector<int> &mergeList, std::vector<int> &to );




void readPtsFormat( const char *filename, real3_type& box_min, real3_type& box_max, PointList& vertex, PointList& normal )
{
	std::ifstream in;
	in.open(filename, std::ios_base::binary);

	if (in.fail()) {
		std::cout << "Error in readPtsFormat(): could not open file " << filename << " for reading!" << std::endl;
		return;
	}

	std::cout << "*** Reading pts from " << filename << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write data in pts format (by Josiah Mason, http://students.cs.tamu.edu/jmanson/programs_wavelet_reconstruct.html)
	// Format description
	// 1. Header
	//    a. Bounding box min corner [FLOAT X, FLOAT Y, FLOAT Z]
	//    b. Bounding box max corner [FLOAT X, FLOAT Y, FLOAT Z]
	//    c. Number of points [INT]
	// 2. List of points (each point has the following format)
	//    a. Normal vector [FLOAT X, FLOAT Y, FLOAT Z]
	//    b. Position vector [FLOAT X, FLOAT Y, FLOAT Z]

	for (int d = 0; d < DIMENSION; ++d) {
		in.read((char*) &box_min[d], sizeof(real_type));
	}
	for (int d = 0; d < DIMENSION; ++d) {
		in.read((char*) &box_max[d], sizeof(real_type));
	}

	int n;
	in.read((char*) &n, sizeof(int));

	// Resize arrays
	vertex.resize(n);
	normal.resize(n);

	for (int i = 0; i < n; ++i) {
		for (int d = 0; d < DIMENSION; ++d) {
			in.read((char*) &normal[i][d], sizeof(real_type));
		}
		for (int d = 0; d < DIMENSION; ++d) {
			in.read((char*) &vertex[i][d], sizeof(real_type));
		}
	}

	in.close();

	//////////////////////////////////////////////////////////////////////////
	std::cout << " finished reading points!" << std::endl;
}


void writePtsFormat( const char *filename, const real3_type& box_min, const real3_type& box_max, const PointList& vertex, const PointList& normal )
{
	if (vertex.size() <= 0 || vertex.size() != normal.size()) {
		std::cout << "Error in writePtsFormat(): no points, or vertex/normal sizes don't match!" << std::endl;
		return;
	}

	std::ofstream out;
	out.open(filename, std::ios_base::binary);

	if (out.fail()) {
		std::cout << "Error in writePtsFormat(): could not open file " << filename << " for writing!" << std::endl;
		return;
	}

	int n = vertex.size();
	std::cout << "*** Writing " << n << " points to " << filename << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write data in pts format (by Josiah Mason, http://students.cs.tamu.edu/jmanson/programs_wavelet_reconstruct.html)
	// Format description
	// 1. Header
	//    a. Bounding box min corner [FLOAT X, FLOAT Y, FLOAT Z]
	//    b. Bounding box max corner [FLOAT X, FLOAT Y, FLOAT Z]
	//    c. Number of points [INT]
	// 2. List of points (each point has the following format)
	//    a. Normal vector [FLOAT X, FLOAT Y, FLOAT Z]
	//    b. Position vector [FLOAT X, FLOAT Y, FLOAT Z]

	for (int d = 0; d < DIMENSION; ++d) {
		out.write((char*) &box_min[d], sizeof(real_type));
	}
	for (int d = 0; d < DIMENSION; ++d) {
		out.write((char*) &box_max[d], sizeof(real_type));
	}
	out.write((char*) &n, sizeof(int));
	for (int i = 0; i < n; ++i) {
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &normal[i][d], sizeof(real_type));
		}
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &vertex[i][d], sizeof(real_type));
		}
	}

	out.close();

	//////////////////////////////////////////////////////////////////////////
	std::cout << " finished writing points!" << std::endl;
}

void write_bnpts_Format( const char *filename, const PointList& vertex, const PointList& normal )
{
	if (vertex.size() <= 0 || vertex.size() != normal.size()) {
		std::cout << "Error in write_bnpts_Format(): no points, or vertex/normal sizes don't match!" << std::endl;
		return;
	}

	std::ofstream out;
	out.open(filename, std::ios_base::binary);

	if (out.fail()) {
		std::cout << "Error in write_bnpts_Format(): could not open file " << filename << " for writing!" << std::endl;
		return;
	}

	const int n = vertex.size();
	std::cout << "*** Writing " << n << " points to " << filename << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write data in bnpts  format (by Michael Kazhdan, http://www.cs.jhu.edu/~misha/Code/PoissonRecon/)
	// Format description
	// 1. List of points (each point has the following format)
	//    a. Position vector [FLOAT X, FLOAT Y, FLOAT Z]
	//    b. Normal vector [FLOAT X, FLOAT Y, FLOAT Z]

	for (int i = 0; i < n; ++i) {
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &vertex[i][d], sizeof(real_type));
		}
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &normal[i][d], sizeof(real_type));
		}
	}

	out.close();

	//////////////////////////////////////////////////////////////////////////
	std::cout << " finished writing points!" << std::endl;
}

// From: http://www.techbytes.ca/techbyte103.html
bool FileExists( std::string strFilename )
{
	struct stat stFileInfo;
	bool blnReturn;
	int intStat;

	// Attempt to get the file attributes
	intStat = stat(strFilename.c_str(),&stFileInfo);

	if (intStat == 0)
	{
		// We were able to get the file attributes
		// so the file obviously exists.
		blnReturn = true; 
	}
	else
	{
		// We were not able to get the file attributes.
		// This may mean that we don't have permission to
		// access the folder which contains this file. If you
		// need to do that level of checking, lookup the
		// return values of stat which will give you
		// more details on why stat failed.
		blnReturn = false;
	}

	return(blnReturn);
}
