#ifndef __UTIL_H__
#define __UTIL_H__

#include "../setting.h"
#include <algorithm>
#include <iostream>
#include <hash_map>
#include <fstream>
#include <QString>
#include <QtOpenGL>
#include <QTextStream>
#include <Windows.h>
#include <QColor>
#include "types.h"
#include <libXMLParam/xmlParam.h>
#include <QGLViewer/qglviewer.h>

#undef min
#undef max

// Useful functions for printing arrays
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v);
QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v);
template<typename T> extern std::ostream& operator<<( std::ostream& out, const std::vector<T> &v );
template<typename T> extern QTextStream& operator<<( QTextStream& out, const std::vector<T> &v );

// Generating samples on a custom probability distribution
void ComputeStep1dCDF(float *f, int nSteps, float *c, float *cdf);
float SampleStep1d(float *f, float *cdf, float c, int nSteps, float u, float *pdf);

#ifdef MEASURE_TIMING

// High-performance timer collection. adapted from
// http://blog.kowalczyk.info/kb/high-resolution-timer-for-timing-code-fragments.html
// very simple, high-precision (at least in theory) timer for timing API calls
struct TimerPair {
	TimerPair() : totalDuration(0.0), iterations(0.0) {
		QueryPerformanceFrequency(&freq);
	}

	void Start() {
		totalDuration = 0.0;
		iterations = 0.0;
		QueryPerformanceCounter(&mTimeStart);
	}

	void AddStart() {
		iterations += 1.0;
		QueryPerformanceCounter(&mTimeStart);
	}

	void Stop() {
		QueryPerformanceCounter(&mTimeStop);
		totalDuration += (double) (mTimeStop.QuadPart - mTimeStart.QuadPart);
	}

	void Reset() {
		totalDuration = 0.0;
		iterations = 0.0;
	}

	double GetDurationInSecs() {
		return (totalDuration / ((double)freq.QuadPart));
	}

	void SetDurationInSecs( double secs ) {
		totalDuration = secs * ((double)freq.QuadPart);
	}

	LARGE_INTEGER freq;
	LARGE_INTEGER mTimeStart;
	LARGE_INTEGER mTimeStop;
	double totalDuration;
	double iterations;
};

// A collection of performance timers
//extern stdext::hash_map<std::string, TimerPair> perfTimers;
extern std::vector<TimerPair> timerArray;

// Convenient macros for measuring timing
#define MEASURE_PERF_MEMBER(x,y) timerArray[x].y
#define MEASURE_PERF_NARG(x,y) timerArray[x].y()
#define MEASURE_PERF_ARG(x,y,z) timerArray[x].y(z)

enum TIMER_TYPE
{
	TIMER_BLANK = 0,
	TIMER_MULTIFRAME_INITIALIZE,
	TIMER_MULTIFRAME_GLOBAL_REG,
	TIMER_MULTIFRAME_RESAMPLE_ASG,
	TIMER_SIMULTANEOUS_T_STEP_1,
	TIMER_SIMULTANEOUS_T_STEP_2,
	TIMER_SIMULTANEOUS_T_STEP_3,
	TIMER_SIMULTANEOUS_W_STEP_1,
	TIMER_SIMULTANEOUS_W_STEP_2,
	TIMER_SIMULTANEOUS_W_STEP_3,
	//
	//
	TIMER_RANGE_QUERY,
	TIMER_FIT_CIRCLE,
	TIMER_POINTLIST_ANN_QUERY,
	TIMER_POINTLIST_N_ANN_QUERY,
	//
	//
	TIMER_SAMPLING_SPEED_TRIAL,
	TIMER_SUB_SAMPLING_GEOMETRY,
	TIMER_POPULATING_CORRESPONDENCE_CANDIDATES,
	TIMER_MEAN_SHIFT_CLUSTERING,
	TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP,
	TIMER_PRECOMPUTING_CLOSEST_POINTS,
	TIMER_GRAPH_CUT_OPTIMIZATION,
	//
	//
	TIMER_RANSAC_TOTAL_OUTLIERS,
	TIMER_RANSAC_TOTAL_CANDIDATES,
	//
	//
	TIMER_NUM,
};

extern const char* timer_str_array[];

// Print timer
inline void printTime(const TIMER_TYPE timer) {
	std::cout << "***** " << timer_str_array[timer] << " took " << MEASURE_PERF_NARG(timer,GetDurationInSecs) << " seconds in " << MEASURE_PERF_MEMBER(timer,iterations) << " iterations." << std::endl;
}

// Print timer
inline std::ostream& printTime( std::ostream& out, const TIMER_TYPE timer ) {
	out << "***** " << timer_str_array[timer] << " took " << MEASURE_PERF_NARG(timer,GetDurationInSecs) << " seconds in " << MEASURE_PERF_MEMBER(timer,iterations) << " iterations." << std::endl;
	return out;
}

#else

// Convenient macros for measuring timing
#define MEASURE_PERF_MEMBER(x,y) NULL
#define MEASURE_PERF_NARG(x,y) NULL
#define MEASURE_PERF_ARG(x,y,z) NULL

#endif

inline void setOpenGLColor( QRgb color )
{
	unsigned char r = (unsigned char) qRed(color);
	unsigned char g = (unsigned char) qGreen(color);
	unsigned char b = (unsigned char) qBlue(color);
	unsigned char a = (unsigned char) qAlpha(color);
	glColor4ub(r,g,b,a);
}

inline void setOpenGLColor( const RealColor &color, real_type alpha = 1.0 )
{
	glColor4f(color[0], color[1], color[2], alpha);
}

void setOpenGLColor( const FloatArray &color, real_type alpha = 1.0 );
void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha = 1.f );
void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha = 1.f );

// UI controller utility functions
class QScrollBar;
class QLineEdit;
float interpSlider(QScrollBar *slider);
void setInterpSlider(QScrollBar *slider, float interp);
float QLineEditFloatValue(QLineEdit *edit);
void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr);

// Box drawing
void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos );
void drawBoxFilled( const real3_type& min_pos, const real3_type& max_pos );

// Sphere drawing
void drawSphere( const real3_type &pos, const float radius, GLint slices = 10, GLint stacks = 10 );
qglviewer::Vec Vec3fToVec( Geometry::Vec3f v );
Geometry::Vec3f VecToVec3f( qglviewer::Vec v );
QPoint clickedCoord( QMouseEvent *e );

// Enumerate correspondence display modes.
enum DISPLAYMODE {
	DISPLAY_SOURCE = 1,
	DISPLAY_TARGET,
	DISPLAY_SOURCE_AND_TARGET
};

/**
 * Display a text progress bar.
 * @param x The current index (when update), or the maximum index (when init).
 * @param init As true, this initializes the progress counter.  As false, it updates the counter.
 */
void TextProgress( double x, bool init = false );
void TextProgressWithLabel( double x, bool init = false, QString id = QString() );


// Rotation matrix about an axis.
Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle);

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m );
void writeBinary( std::ofstream &out, const Geometry::Vec3f v );
void readBinary( std::ifstream &in, Geometry::Matrix33f &m );
void readBinary( std::ifstream &in, Geometry::Vec3f &v );

// Best candidate sampling within a given point set.
class PointList;
idxList bestCandidateSample( const PointList &pts, int n );
idxList bestCandidateSampleDist( const PointList &pts, const real_type sampleDist );

// Generate light colors
void generateLightColors( int n, std::vector<RealColor> &colors );

// Write out the contents of a vector
void writeVector( double* x, int size, const char *filename );

// Pause until enter is pressed
void pauseUntilEnter( const char *message );

// Replace a field in a string with a number
void replaceField( std::string& str, char field, int num );

// Help from http://objectmix.com/c/358470-how-parallel-sort.html to perform sort on indices of the array
template < typename Sequence, class Predicate >
class ForwardCompare {
	const Sequence& the_sequence;
	const Predicate& the_predicate;

public:
	ForwardCompare ( Sequence const & ref, Predicate & pred ): the_sequence ( ref ), the_predicate ( pred ) { };
	bool operator() ( typename Sequence::size_type lhs, typename Sequence::size_type rhs ) const {
			return the_predicate(the_sequence[lhs], the_sequence[rhs]);
	}
};

template < typename Sequence, class Predicate >
ForwardCompare< Sequence, Predicate > make_ForwardCompare ( Sequence const & seq, Predicate & pred ) {
	return ( ForwardCompare< Sequence, Predicate >( seq, pred ) );
}

// Get the last digit in a string
bool getLastDigit( QString str, int& number, int& digits );

// Color a weight in OpenGL
RealColor OpenGLWeightColor( real_type value );
void setOpenGLWeightColor( real_type value, real_type alpha = 1.0f );

////// Sort ascending: first column, then row
////struct index2_colAsc : public std::binary_function<index2_type,index2_type,bool> {
////	inline bool operator()(const index2_type &x, const index2_type &y) const {
////		//if (x[1] == y[1])
////		//	return (x[0] < y[0]);
////		//else
////		//	return (x[1] < y[1]);
////		if (*(((index_type*)&x)+1) == *(((index_type*)&y)+1))
////			return (*((index_type*)&x) < *((index_type*)&y));
////		else
////			return (*(((index_type*)&x)+1) < *(((index_type*)&y)+1));
////	}
////};
////
////// Sort ascending: first row, then column
////struct index2_rowAsc : public std::binary_function<index2_type,index2_type,bool> {
////	inline bool operator()(const index2_type &x, const index2_type &y) const {
////		//if (x[0] == y[0])
////		//	return (x[1] < y[1]);
////		//else
////		//	return (x[0] < y[0]);
////		if (*((index_type*)&x) == *((index_type*)&y))
////			return (*(((index_type*)&x)+1) < *(((index_type*)&y)+1));
////		else
////			return (*((index_type*)&x) < *((index_type*)&y));
////	}
////};

// Sort ascending
template<typename T>
struct comp_asc: public std::binary_function<T,T,bool> {
	bool operator()(const T &x, const T &y) const {
		return (x < y);
	}
};

// Sort descending
template<typename T>
struct comp_desc: public std::binary_function<T,T,bool> {
	bool operator()(const T &x, const T &y) const {
		return (x > y);
	}
};

// Sort a key/data pair (key ascending)
template<typename Key, typename Data>
struct comp_key_data_asc: public std::binary_function<std::pair<Key,Data>,std::pair<Key,Data>,bool> {
	bool operator()(const std::pair<Key,Data> &x, const std::pair<Key,Data> &y) const {
		return (x.first < y.first);
	}
};

// Sort a key/data pair (key descending)
template<typename Key, typename Data>
struct comp_key_data_desc: public std::binary_function<std::pair<Key,Data>,std::pair<Key,Data>,bool> {
	bool operator()(const std::pair<Key,Data> &x, const std::pair<Key,Data> &y) const {
		return (x.first > y.first);
	}
};


// Enumerating unique pairs
template<typename integralType>
class EnumerateUniquePairs {
public:
	static integralType numTotalPairs( integralType n ) {
		return (n-1)*n / 2;
	}
	static std::pair<integralType,integralType> linearIndexToPairIndex( integralType n, integralType linearIndex ) {
		double B = (2*n-1);
		double largest_i = 0.5*(B - sqrt(B*B - 8.0*((double)linearIndex)));
		integralType solved_i = floor(largest_i + 1e-5);
		integralType solved_j = linearIndex - (2*n-solved_i-1)*solved_i/2 + (solved_i+1);
		return std::make_pair(solved_i, solved_j);
	}
	static integralType pairIndexToLinearIndex( integralType n, integralType i, integralType j) {
		return (2*n-i-1)*i/2 + (j-i-1);
	}

private:
	// Prevent someone from actually constructing an object of this type
	EnumerateUniquePairs();
	~EnumerateUniquePairs();
};



// Utility function for union-ing two lists
template<typename T> extern void sortedListUnion( const std::vector<T> &first, const std::vector<T> &second, std::vector<T> &result );
template<typename T> extern void sortedListUnionMerge( const std::vector<T> &mergeList, std::vector<T> &to );

// Reading / writing a list of points and normals
void readPtsFormat( const char *filename, real3_type& box_min, real3_type& box_max, PointList& vertex, PointList& normal );
void writePtsFormat( const char *filename, const real3_type& box_min, const real3_type& box_max, const PointList& vertex, const PointList& normal );
void write_bnpts_Format( const char *filename, const PointList& vertex, const PointList& normal );

// Checking if a file exists
bool FileExists( std::string strFilename );

#endif
