#ifndef __CORRESP_VIEWER_H__
#define __CORRESP_VIEWER_H__

#include <QGLViewer/qglviewer.h>
#include <QMutex>
#include <QSplitter>
#include "observer.h"
#include "Correspondence.h"
#include <QMenu>

class Viewer;
class Scene;
class Subject;
class CorrespViewer;

enum selectionMode {
	SELECT_VERTICES = 1,
	SELECT_CORRESP
};

enum showCorrespModes {
	SHOW_SOURCE_TO_TARGET = 1,
	SHOW_TARGET_TO_SOURCE
};

enum TransformSets {
	TRANSFORM_SET_PAIRS = 1,
	TRANSFORM_SET_CLUSTERS,
	TRANSFORM_SET_REGIONS
};

class CorrespViewerElement : public QGLViewer
{
	Q_OBJECT

public:
	CorrespViewerElement(CorrespViewer *v, Scene* s, int displayMode);		/**< Constructor: initialize members here. */
	virtual ~CorrespViewerElement();										/**< Destructor: destroy members here. */

	void saveScreenshot2( const QString &filename );
	void resetView();

public slots:
	void updateDisplay();

	// Correspondences actions
	void saveCorrespondences();
	void loadCorrespondences();
	void clearCorrespondences();

	// Switch correspondence viewing type
	void switchCorrespViewType();

	// Show correspondence result
	void DisplayCorrespondenceResult( bool show );

	// Load/Save correspondence result
	void loadCorrespResult();
	void saveCorrespResult();

	// Grab screenshot
	void saveScreenshot( bool promptFile = true );
	void customSaveSnapshot();								/**< Take a snapshot and save it in a specific file. */

	// Load/Save Camera Parameters
	void loadCamera(const QString& file = QString());
	void saveCamera();

	// Regenerate colors
	void regenerateColors();

protected :
	virtual void draw();
	virtual void init();
	virtual void animate();
	virtual void drawWithNames();
	void drawSelectionRectangle() const;
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);
	virtual void keyPressEvent(QKeyEvent *e);
	void endSelection(const QPoint&);

private:
	// Pointer to the entire scene
	Scene *scene;
	CorrespViewer *parent;

	// Correspondence Data
	CorrespondenceList *cList;

	// Mode
	const int displayMode;

	// Subset display
	int displaySubsetMode;
	int displaySubsetNum, displaySubsetIndex, displaySubsetTotal;

	//////////////////////////////////////////////////////////////////////////
	// Selection variables
	// Current rectangular selection
	QRect rectangle_;

	// Different selection modes
	enum SelectionMode { NONE, TOGGLE, MOVE, ADD };
	SelectionMode selectionMode_;

	// Point select threshold
	const int pointSelect;

	int storedFrameIndex;

	// String to save snapshot
	QString customSaveSnapshotFile;

	//////////////////////////////////////////////////////////////////////////
	// Show clustering result
	void showClusterMembers();
	void showVerifiedRegions();
	void showMotionSegmentation();
	void drawExtraEdges();
	void drawSubset();
	void drawTransformedSubset( int transformSet );
	void drawTextMessage( int i, QColor color, QString text );

	//////////////////////////////////////////////////////////////////////////
	// Menu variables and functions
	void createContextMenu();

	QMenu* contextMenu;
	QAction *dummyTitleAct;
	QAction *saveCorrespondencesAct;
	QAction *loadCorrespondencesAct;
	QAction *clearCorrespondencesAct;
	QAction *showMotionSegmentationAct;
	QAction *showModelsAct;
	QAction *switchCorrespViewTypeAct;
	QAction *loadCorrespResultAct;
	QAction *saveCorrespResultAct;
	QAction *showClusterMembersAct;
	QAction *showVerifiedRegionsAct;
	QAction *showExtraEdgesAct;
	QAction *saveScreenshotAct;
	QAction *showMotionSegmentationAsPointsAct;
	QAction *switchSelectModeAct;
	QAction *displayDeformEMResultAct;
	QAction *displayDeformAllResultAct;
	QAction *saveCameraAct;
	QAction *loadCameraAct;
	QAction *regenerateColorsAct;
	QAction *showOtherModelAct;

	friend class CorrespViewer;
};


class CorrespViewer : public QSplitter, public Observer {

	Q_OBJECT

public slots:
	void updateDisplay();
	void SwitchSelectMode();					/**< Switch selection mode */

public:
	CorrespViewer( Scene *s, Viewer *v );
	virtual ~CorrespViewer();

	virtual void Update(Subject* theChangedSubject);		/**< Update the display according to the subject */

	void toggleIdSelection(index_type id);
	void toggleTargetIdSelection(index_type id);

	// Selected points
	int selectMode;
	idxList selected, targetSelected;

	// Adding new correspondences
	bool setNewSourcePt, setNewTargetPt;
	Geometry::Vec3f newSourcePt, newTargetPt;

	// Showing forward or backward correspondences
	int showCorrespType;

	// Show correspondence result
	void DisplayCorrespondenceResult( bool show );

	const Viewer* getMasterViewer() const {
		return masterViewer;
	}

	// Capture registration results
	void CaptureResults( const std::string& name, const std::string& camFile );

protected:
	virtual void init();

private:
	// Pointer to the entire scene
	Scene *scene;
	Viewer *masterViewer;
	qglviewer::Camera *tempCamera;

	// Source and Target views
	CorrespViewerElement *sourceView, *targetView;

	// Scene change notify variable
	QMutex sceneMutex;
	bool sceneChanged;

signals:
	void updateViews();
};

#endif
