#include "ShapePoser.h"
#include "scene.h"
#include "viewer.h"
#include <algorithm>
#include "DeformGraphAll.h"
#include <QMutex>
#include <boost/tuple/tuple.hpp>
#include "../Util/util.h"

using namespace std;
using namespace Deform;
using namespace Deform::DeformAll_Utility;

ShapePoser::ShapePoser( Scene *sc, ShapeModel* s ) :
scene(sc), shape(s), pointSelect(3), selectionMode_(NONE),
contextMenu(NULL), loadPoseResultAct(NULL), savePoseResultAct(NULL), selectedGroup(-1),
showSource(true), def(NULL), showWeights(false), defVertexValid(NULL), showSourceWeights(false),
drawSelectionBoxes(true), shapeWeights(NULL), jointMap(NULL), jointDisplaySize(3.f),
setSourceMFMode(false), currentMotion(0)
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	if (!shape) {
		// Immediately kill
		close();
	}

	shapeColor[0] = 1.0f; shapeColor[1] = 0.7f; shapeColor[2] = 0.7f;

	generateLightColors(scene->graph_n, boneColors);
	sourceVertexValid.resize(shape->numVertices(), true);
}

ShapePoser::~ShapePoser()
{
	if (contextMenu) delete contextMenu;
	if (savePoseResultAct) delete savePoseResultAct;
	if (loadPoseResultAct) delete loadPoseResultAct;
}

void ShapePoser::init()
{
	createContextMenu();

	//////////////////////////////////////////////////////////////////////////
	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);
	setWindowTitle("Shape Poser");

	// Center the view on the source and target
	setSceneBoundingBox(qglviewer::Vec(shape->box_min[0], shape->box_min[1], shape->box_min[2]),
		qglviewer::Vec(shape->box_max[0], shape->box_max[1], shape->box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	showEntireScene();

	//////////////////////////////////////////////////////////////////////////
	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the max select buffer size
	setSelectBufferSize(100000);

	// Set antialiasing
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);

	//////////////////////////////////////////////////////////////////////////
	// Make camera the default manipulated frame.
	setManipulatedFrame( camera()->frame() );

	// Preserve CAMERA bindings, see setHandlerKeyboardModifiers documentation.
	// The frames can be move with the shof
	// The camera can always be moved with the Control key.
	setHandlerKeyboardModifiers(QGLViewer::FRAME, Qt::ShiftModifier);

	//////////////////////////////////////////////////////////////////////////
	// Move and resize the window
	this->move(530, 230);
	this->resize(800, 600);

	//////////////////////////////////////////////////////////////////////////
	scene->registerShapePoserInstance(this);

	this->setAnimationPeriod(0.f);
	startAnimation();

	//////////////////////////////////////////////////////////////////////////
	cout << "...done!" << endl;
}

Deform::DrawOptions ShapePoser::getDeformDrawOptions()
{
	Deform::DrawOptions options;

	// Populate option values from the viewer and the scene
	options.showSourceModel = false;
	options.showTargetModel = false;
	options.showDeformedModel = true;
	options.showSourceCorresp = false;
	options.showTargetCorresp = false;
	options.showDeformedCorresp = false;
	options.showShapeWeights = false;
	options.showAllFrames = false;
	options.displayFrame = 0;
	options.showGraph = false;
	options.showEffRad = false;
	options.renderTransparent = false;
	options.showComponent = -1;
	options.showGraphDeformed = false;
	options.showGraphFaces = false;
	options.showExtendedGrid = -1;
	options.showWireframe = false;
	options.openGLPointSize = 2.0f;
	options.sourceModelColor = RealColor(1.0f, 0.7f, 0.7f);
	options.deformedModelColor = RealColor(0.7f, 1.0f, 0.7f);
	options.targetModelColor = RealColor(0.7f, 0.7f, 1.0f);
	options.pointColor = RealColor(scene->pointColor[0],scene->pointColor[1],scene->pointColor[2]);
	//options.showErrorComponent = showErrorComponent;
	options.ignoreDisplayInterface = false;
	options.showSourceDeformed = false;
	options.showInvalidVertices = false;
	options.showSampledPoints = false;
	options.showExternalMeshDeformed = false;
	options.showJoints = false;

	return options;
}

namespace Deform
{
	namespace DeformGraphAll_Utility
	{
		void drawJoints( const JointMap& jointMap, const int numBones, const real_type jointSize, const std::vector<RealColor>& boneColors, const DrawOptions& options )
		{
			JointMap::const_iterator it;
			for (it = jointMap.begin(); it != jointMap.end(); ++it) {
				int bi, bj, linIdx = it->first;
				boost::tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);
				const Joint& joint = it->second;
				const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
				const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

				int numSamples = 0;
				if (joint.type == JOINT_BALL) {
					numSamples = 1;
				}
				else if (joint.type == JOINT_HINGE) {
					numSamples = 20;
				}

				for (int j = 0; j < numSamples; ++j) {
					real3_type jointLocation;

					if (joint.type == JOINT_BALL) {
						jointLocation = jointPos;
					}
					else if (joint.type == JOINT_HINGE) {
						real_type t = jointSize * 3.333 * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
						jointLocation = jointPos +  t*jointCon;
					}

					//////////////////////////////////////////////////////////////////////////
					glEnable(GL_LIGHTING);
					setOpenGLMaterialColor(boneColors[bi]);
					drawSphere(jointLocation, jointSize);
					setOpenGLMaterialColor(boneColors[bj]);
					drawSphere(jointLocation + real3_type(1.0, 1.0, 1.0), jointSize);
				}
			}
		}
	}
}

void ShapePoser::testLoadJoints( DeformAll_Utility::TransformList& tfList, int motionIndex )
{
	static bool firstDraw = true;
	static BVHLoader::MySkeletonStateList states;
	static std::tr1::shared_ptr<BVHLoader::Joint> root(static_cast<BVHLoader::Joint*>(NULL));
	static int numBones = 17;

	if (firstDraw)
	{
		firstDraw = false;
		const char *tmp = "K:\\HandTracking\\HandTrackingData\\DetailedHand2\\lefthand.bvh";
		root = Deform::DeformGraphAll_Utility::loadJointsBVH(tmp, states);
		//Deform::DeformGraphAll_Utility::populateJointsFromBVH(root, numBones, NULL, jointMap, states[0]);
		//cout << DeformGraphAll_Utility::printHeader << "Loading complete!  Number of loaded joints: " << jointMap.size() << endl;
	}

	const char *masterName = NULL;

	tfList.resize(numBones);

	// Walk the skeleton, mark joint locations
	typedef struct _Recit
	{
		std::tr1::weak_ptr<BVHLoader::Joint> joint;
		real3_type parentOffset;
		int index;
		int parentIndex;
		bool hasMasterParent;
		DeformAll_Utility::SE3_6DOF parentLocalTf;
	} Recit;
	std::deque<Recit> recQueue;

	Recit parent;
	parent.joint = root;
	parent.index = (masterName ? -1 : 0);
	parent.parentIndex = (masterName ? -1 : 0);
	parent.hasMasterParent = false;
	parent.parentOffset = real3_type(0.f, 0.f, 0.f);
	parent.parentLocalTf = DeformAll_Utility::SE3_6DOF();
	recQueue.push_back(parent);

	glLineWidth(5.f);

	do 
	{
		bool childOfMaster = false;
		real3_type currentOffset = real3_type(0.f, 0.f, 0.f);

		// Pull offset off queue
		parent = recQueue.front();
		recQueue.pop_front();

		// Visit child
		std::tr1::shared_ptr<BVHLoader::Joint> parentJoint(parent.joint);
		std::tr1::shared_ptr<BVHLoader::JointInfo> parentInfo(parentJoint->getInfo());

		int parentIndex = parentInfo->stateIndex;

		DeformAll_Utility::SE3_6DOF parentFrame;
		if (motionIndex >= 0 && motionIndex < states.size())
		{
			parentFrame = parent.parentLocalTf
				* DeformAll_Utility::SE3_6DOF(Geometry::Matrix33f::identity(), parentInfo->offset)
				* ((*states[motionIndex])[parentIndex]);
		}
		else
		{
			parentFrame = parent.parentLocalTf
				* DeformAll_Utility::SE3_6DOF(Geometry::Matrix33f::identity(), parentInfo->offset);
		}

		tfList[parentIndex] = parentFrame;

		if (parentInfo)
		{
			// Apply rotation and translation when adding the offset
			currentOffset = parentFrame.tr();

			if (parent.hasMasterParent || masterName == NULL || parentInfo->name == masterName)
			{
				childOfMaster = true;
			}
		}

		// This is the offset for this joint, record it
		// Joint is between this and the parent
		if (parent.hasMasterParent)
		{
			if (parent.parentIndex < 0 || parent.index < 0)
			{
				cerr << "Error!!! parent index " << parent.parentIndex << ", index " << parent.index << std::endl;
			}
			else
			{
				// Add currentOffset as joint
				glDisable(GL_LIGHTING);
				setOpenGLColor(boneColors[parentIndex]);
				glBegin(GL_LINES);
				glVertex3f(parent.parentOffset.x(), parent.parentOffset.y(), parent.parentOffset.z());
				glVertex3f(currentOffset.x(), currentOffset.y(), currentOffset.z());
				glEnd();

				glEnable(GL_LIGHTING);
				setOpenGLMaterialColor(boneColors[parent.parentIndex]);
				drawSphere(parent.parentOffset, jointDisplaySize);
				setOpenGLMaterialColor(boneColors[parentIndex]);
				drawSphere(currentOffset, jointDisplaySize);
				//cout << "Adding ball joint between " << current.parentIndex << " and " << current.index << std::endl;
			}
		}

		// Update queue
		for (int i = 0; i < parentJoint->numChildren(); ++i)
		{
			Recit anew;
			anew.joint = parentJoint->getChild(i);
			anew.index = parentIndex;
			anew.parentIndex = parent.index;
			anew.hasMasterParent = childOfMaster;
			anew.parentOffset = currentOffset;
			anew.parentLocalTf = parentFrame;
			//cout << "Concat: parent " << current.index << " with self " << curIndex << std::endl;
			//cout << " - Parent: " << current.parentLocalTf << std::endl;
			//cout <<"  - Self: " << anew.parentLocalTf << std::endl;
			recQueue.push_back(anew);

			std::tr1::shared_ptr<BVHLoader::Joint> childJoint(anew.joint);
			std::tr1::shared_ptr<BVHLoader::JointInfo> childInfo(childJoint->getInfo());
			//cout << "Found bone " << curIndex << " with name: " << childInfo->name << std::endl;
		}
	} while (!recQueue.empty());
}

void ShapePoser::draw()
{
	//cout << "source mf: " << sourceMF.translation().x << " " << sourceMF.translation().y << " " << sourceMF.translation().z << " " << std::endl;
	glPushMatrix();
	glMultMatrixd(sourceMF.matrix());

	// Draw the deformation result
	Deform::DrawOptions drawOptions = getDeformDrawOptions();
	scene->drawDeformation(drawOptions);

	if (showSource) {
		// Draw the shape
		glEnable(GL_LIGHTING);
		setOpenGLMaterialColor(shapeColor);
		
		if (shapeWeights)
		{
			drawOptions.showShapeWeights = showSourceWeights;
			drawShapeWeightsValid(shape->vertex, shape->normal, *shape, *shapeWeights, sourceVertexValid, shapeColor, boneColors, drawOptions, drawOptions.showInvalidVertices, true);
		}
		else
		{
			drawShapeValid(shape->vertex, shape->normal, *shape, sourceVertexValid, drawOptions.showInvalidVertices);
		}
	}

	if (def && defVertexValid && drawDefLock) {
		// Draw the shape
		glEnable(GL_LIGHTING);
		setOpenGLMaterialColor(shapeColor);

		drawDefLock->lock();
		drawOptions.showShapeWeights = showWeights;
		if (shapeWeights)
		{
			drawShapeWeightsValid(def->vertex, def->normal, *def, *shapeWeights, *defVertexValid, shapeColor, boneColors, drawOptions, drawOptions.showInvalidVertices, true);
		}
		else
		{
			drawShapeValid(def->vertex, def->normal, *def, *defVertexValid, drawOptions.showInvalidVertices);
		}
		drawDefLock->unlock();

		//for (int i = 0; i < boneColors.size(); ++i) {
		//	cout << "Bone color " << i << ": " << boneColors[i] << endl;
		//}
		//for (int i = 0; i < shapeWeights->size(); ++i) {
		//	cout << "Shape weights " << i << ": " << (*shapeWeights)[i].bones << " " << (*shapeWeights)[i].weights << endl;
		//}
	}

	glPopMatrix();


	//// Draw the selected vertices
	//glEnable(GL_LIGHTING);
	//setOpenGLMaterialColor(scene->pointColor);
	//PointList_Utility::draw(shape->vertex, POINT_SPHERES, selected, 0.5 * scene->pointSize());

	//if (jointMap)
	//{
	//	Deform::DeformGraphAll_Utility::drawJoints(*jointMap, 17, jointDisplaySize, boneColors, drawOptions);
	//}

	// Draws rectangular selection area. Could be done in postDraw() instead.
	if (selectionMode_ != NONE) {
		drawSelectionRectangle();
	}


	//////////////////////////////////////////////////////////////////////////
	// Draw the selection boxes
	if (drawSelectionBoxes) {
		glEnable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glDepthMask(GL_FALSE);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

		for (int i = 0; i < selectionGroups.size(); ++i) {
			if ((selectedGroup >= 0) && (selectedGroup < selectionGroups.size()) && (i == selectedGroup)) {
				setOpenGLMaterialColor(scene->sourceColor, 0.6f);
			}
			else {
				setOpenGLMaterialColor(scene->pointColor, 0.6f);
			}
			selectionGroups[i]->draw(showSource);
		}

		glDisable(GL_CULL_FACE);
		glDepthMask(GL_TRUE);
		glDisable(GL_BLEND);
	}
}

void ShapePoser::drawWithNames()
{
	if (selectionMode_ == NEW_GROUP) {
		glPointSize(5.0f);
		if (vertexSubsample.size()) {
			PointList_Utility::drawSubsetWithNames(shape->vertex, POINT_SQUARES, vertexSubsample);
		}
		else {
			PointList_Utility::drawWithNames(shape->vertex, POINT_SQUARES);
		}
	}
	else if (selectionMode_ == SELECT_GROUP) {
		for (int i = 0; i < selectionGroups.size(); ++i) {
			glPushName(i);
			selectionGroups[i]->draw(showSource);
			glPopName();
		}
	}
}

void ShapePoser::drawSelectionRectangle() const
{
	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	glColor4f(0.0f, 0.0f, 0.3f, 0.3f);
	glBegin(GL_QUADS);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glLineWidth(2.0f);
	glColor4f(0.4f, 0.4f, 0.5f, 0.5f);
	glBegin(GL_LINE_LOOP);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();
}

void ShapePoser::mousePressEvent( QMouseEvent *e )
{
	// Start selection.
	rectangle_ = QRect(e->pos(), e->pos());

	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier)) {
		selectionMode_ = NEW_GROUP;
	}
	//else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
	//	selectionMode_ = SELECT_GROUP;
	//}
	else {
		selectionMode_ = SELECT_GROUP;
		// Define selection window dimensions
		setSelectRegionWidth(20);
		setSelectRegionHeight(20);
		// Compute rectangle center and perform selection
		select(e->pos());
		// reset selection mode
		selectionMode_ = NONE;
		// Update display to show new selected objects
		updateGL();

		QGLViewer::mousePressEvent(e);
	}
}

void ShapePoser::mouseMoveEvent(QMouseEvent* e)
{
	if (selectionMode_ != NONE) {
		// Updates rectangle_ coordinates and redraws rectangle
		rectangle_.setBottomRight(e->pos());
		updateGL();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void ShapePoser::mouseReleaseEvent(QMouseEvent* e)
{
	if (selectionMode_ != NONE) {
		// Actual selection on the rectangular area.
		// Possibly swap left/right and top/bottom to make rectangle_ valid.
		rectangle_ = rectangle_.normalized();

		int rw, rh;
		rw = rectangle_.width();
		rh = rectangle_.height();

		if (rw <= 0) rw = 1;
		if (rh <= 0) rh = 1;

		// Define selection window dimensions
		setSelectRegionWidth(rw);
		setSelectRegionHeight(rh);

		// Compute rectangle center and perform selection
		select(rectangle_.center());

		// reset selection mode
		selectionMode_ = NONE;

		// Update display to show new selected objects
		updateGL();
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void ShapePoser::endSelection( const QPoint& )
{
	// Flush GL buffers
	glFlush();

	// Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
	GLint nbHits = glRenderMode(GL_RENDER);

	//cout << "Number of objects selected: " << nbHits << endl;

	// Interpret results : each object created 4 values in the selectBuffer().
	// (selectBuffer())[4*i+3] is the id pushed on the stack.
	if (nbHits > 0) {
		// If the rectangle is small, then select a single object of minimum depth.
		if ((rectangle_.width() < pointSelect) || (rectangle_.height() < pointSelect)) {
			GLuint zMin = (selectBuffer())[1];
			int minId = (selectBuffer())[3];

			for (int i=1; i<nbHits; ++i) {
				if ((selectBuffer())[4*i+1] < zMin) {
					zMin = (selectBuffer())[4*i+1];
					minId = (selectBuffer())[4*i+3];
				}
			}

			if (selectionMode_ == SELECT_GROUP) {
				selectedGroup = minId;
			}
		}
		else {
			if (selectionMode_ == NEW_GROUP) {
				SelectionGroup *newGroup = new SelectionGroup(shape->vertex);

				for (int i=0; i<nbHits; ++i) {
					newGroup->addIdSelection((selectBuffer())[4*i+3]);
				}
				newGroup->computeBox();

				selectionGroups.push_back(newGroup);
			}
		}
	}
	else {
		if (selectionMode_ == SELECT_GROUP) {
			selectedGroup = -1;
		}
	}
}

void ShapePoser::postSelection( const QPoint& )
{
	if ((selectedGroup < 0) || (selectedGroup >= selectionGroups.size())) {
		if (setSourceMFMode)
		{
			setManipulatedFrame( &sourceMF );
		}
		else
		{
			// Camera will be the default frame is no object is selected.
			setManipulatedFrame( camera()->frame() );
		}
	}
	else {
		setManipulatedFrame( selectionGroups[selectedGroup]->frame() );
	}

	displayMessage(tr("in postSelection"));
}

void ShapePoser::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
		updateGL();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::NoButton)) {
		// Clear selection
		if ((selectedGroup >= 0) && (selectedGroup < selectionGroups.size())) {
			int erasedIndex = selectedGroup;
			SelectionGroup* erasedGroup = selectionGroups[selectedGroup];
			selectedGroup = -1;
			// Camera will be the default frame is no object is selected.
			setManipulatedFrame( camera()->frame() );

			// Remove from the list of groups
			selectionGroups.erase(selectionGroups.begin() + erasedIndex);
			delete erasedGroup;

			//displayMessage(tr("Erased group %1, %2 groups left.").arg(erasedIndex).arg(selectionGroups.size()));
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		showSource = !showSource;
		displayMessage(tr("Displaying source: %1").arg(showSource?"Yes":"No"));
	}
	else if ((e->key()==Qt::Key_H) && (modifiers==Qt::NoButton)) {
		showWeights = !showWeights;
		displayMessage(tr("Displaying weights: %1").arg(showWeights?"Yes":"No"));
	}
	else if ((e->key()==Qt::Key_N) && (modifiers==Qt::NoButton)) {
		showSourceWeights = !showSourceWeights;
		displayMessage(tr("Displaying source weights: %1").arg(showSourceWeights?"Yes":"No"));
	}
	else if ((e->key()==Qt::Key_L) && (modifiers==Qt::NoButton)) {
		std::vector<RealColor> temp;
		generateLightColors(scene->graph_n, temp);
		boneColors.swap(temp);
		displayMessage(tr("Changing bone colors"));
	}
	else if ((e->key()==Qt::Key_V) && (modifiers==Qt::NoButton)) {
		drawSelectionBoxes = !drawSelectionBoxes;
		displayMessage(tr("Drawing selection boxes: %1").arg(drawSelectionBoxes?"Yes":"No"));
	}
	else if ((e->key()==Qt::Key_BracketLeft) && (modifiers==Qt::NoButton)) {
		jointDisplaySize *= 0.9f;
		displayMessage(tr("Decreasing joint display size: %1").arg(jointDisplaySize));
	}
	else if ((e->key()==Qt::Key_BracketRight) && (modifiers==Qt::NoButton)) {
		jointDisplaySize *= 1.1f;
		displayMessage(tr("Increasing joint display size: %1").arg(jointDisplaySize));
	}
	else if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		std::string which;

		if (this->manipulatedFrame() != this->camera()->frame())
		{
			setManipulatedFrame(this->camera()->frame());
			which = "camera";
			setSourceMFMode = false;
		}
		else
		{
			setManipulatedFrame(&sourceMF);
			which = "source mesh";
			setSourceMFMode = true;
		}
		displayMessage(tr("Setting manipulated frame to: %1").arg(which.c_str()));
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::NoButton)) {
		++currentMotion;
		displayMessage(tr("Current motion: %1").arg(currentMotion));
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::NoButton)) {
		--currentMotion;
		displayMessage(tr("Current motion: %1").arg(currentMotion));
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}

void ShapePoser::createContextMenu()
{
	contextMenu = new QMenu(tr("&Menu"), this);

	savePoseResultAct = new QAction(tr("&Save Pose Result..."), this);
	contextMenu->addAction(savePoseResultAct);
	connect(savePoseResultAct, SIGNAL(triggered()), this, SLOT(savePoseResult()));

	loadPoseResultAct = new QAction(tr("&Load Pose Result..."), this);
	contextMenu->addAction(loadPoseResultAct);
	connect(loadPoseResultAct, SIGNAL(triggered()), this, SLOT(loadPoseResult()));
}


void ShapePoser::savePoseResult()
{
	cerr << "Not implemented yet!" << endl;

	//// Add a new frame to the animation.
	//QString file = QFileDialog::getSaveFileName(this, tr("Save Mesh"), tr(""), tr("Meshes (*.obj)"));

	//if (!file.size()) {
	//	return;
	//}

	//if (shape->shapeType == SHAPE_MESH) {
	//	Mesh3_real* newShape = ((Mesh3_real*)shape)->pickSubset(selected);
	//	newShape->setFile(tr(scene->rootDir.c_str()), file);
	//	newShape->writeGeometryOBJ();
	//}
	//else {
	//	QMessageBox::information(this, tr("Error!"), tr("Saving out a point set is not implemented yet!"), QMessageBox::Ok);
	//}
}

void ShapePoser::loadPoseResult()
{
	cerr << "Not implemented yet!" << endl;

	//// Add a new frame to the animation.
	//QString file = QFileDialog::getOpenFileName(this, tr("Open Mesh"), tr(""), tr("Meshes (*.obj)"));

	//if (file.size() > 0) {
	//	shape = loadShape(tr(scene->rootDir.c_str()), file, *scene, false);
	//	selectAll();
	//}
}

void ShapePoser::updateInteractiveCorresp( CorrespondenceList& updateList ) const
{
	int totalSize = 0;
	for (int i = 0; i < selectionGroups.size(); ++i) {
		totalSize += selectionGroups[i]->size();
	}

	// Update the correspondence list
	if (totalSize > 0)
	{
		updateList.resize(totalSize);
		int startIdx = 0;
		for (int i = 0; i < selectionGroups.size(); ++i) {
			selectionGroups[i]->updateCorrespList(updateList, startIdx);
			startIdx += selectionGroups[i]->size();
		}
	}
	else
	{
		updateList.resize(0);
	}
}

void ShapePoser::subsampleConstraints( int maxConstraints )
{
	// Perform best candidate sampling on the vertices
	vertexSubsample = bestCandidateSample(shape->vertex, maxConstraints);
}

SE3 ShapePoser::getSE3( const qglviewer::ManipulatedFrame& mf )
{
	SE3 result;
	float m[3][3];
	qglviewer::Vec t = mf.translation();
	mf.rotation().getRotationMatrix(m);

	for (int i = 0; i < 3; ++i) {
		result.tr[i] = t[i];
		for (int j = 0; j < 3; ++j) {
			result.rot(i,j) = m[i][j];
		}
	}

	return result;
}

ShapePoser::SelectionGroup::SelectionGroup( const PointList& vertices ) :
vertices(vertices), box_min(0.0f,0.0f,0.0f), box_max(0.0f,0.0f,0.0f)
{

}

void ShapePoser::SelectionGroup::addIdSelection( index_type id )
{
	idxList::iterator found = std::find(selected.begin(), selected.end(), id);
	if (found == selected.end()) {
		selected.push_back(id);
	}
}

void ShapePoser::SelectionGroup::eraseIdSelection( index_type id )
{
	idxList::iterator found = std::find(selected.begin(), selected.end(), id);
	if (found != selected.end()) {
		selected.erase(found);
	}
}

void ShapePoser::SelectionGroup::computeBox()
{
	box_min[0] = box_min[1] = box_min[2] = REAL_TYPE_MAX;
	box_max[0] = box_max[1] = box_max[2] = -REAL_TYPE_MAX;

	for (int i = 0; i < selected.size(); ++i) {
		const real3_type& pos = vertices[selected[i]];

		for (int j = 0; j < 3; j++) {
			if (pos[j] < box_min[j]) {
				box_min[j] = pos[j];
			}
			if (pos[j] > box_max[j]) {
				box_max[j] = pos[j];
			}
		}
	}

	// Set the frame position
	box_center = box_min + (box_max - box_min) / 2.0f;
	mf.setPosition(box_center[0],box_center[1],box_center[2]);
}

void ShapePoser::SelectionGroup::draw( bool showSource ) const
{
	// Draw target box
	glPushMatrix();
	glMultMatrixd(mf.matrix());
	drawBoxFilled(box_min - box_center, box_max - box_center);
	glPopMatrix();

	if (showSource) {
		// Draw source box
		drawBoxFilled(box_min, box_max);

		// Draw line between boxes
		SE3 tf = getSE3(this->mf);
		real3_type center_tf = tf.tr;
		real3_type temp;

		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
		glColor3f(0.0f,0.0f,0.0f);
		glVertex3f(box_center[0],box_center[1],box_center[2]);
		glVertex3f(center_tf[0],center_tf[1],center_tf[2]);
		glVertex3f(box_min[0],box_min[1],box_min[2]);
		temp = tf * (box_min - box_center);
		glVertex3f(temp[0],temp[1],temp[2]);
		glVertex3f(box_max[0],box_max[1],box_max[2]);
		temp = tf * (box_max - box_center);
		glVertex3f(temp[0],temp[1],temp[2]);
		glEnd();

		glBegin(GL_POINTS);
		glVertex3f(box_min[0],box_min[1],box_min[2]);
		glVertex3f(box_max[0],box_max[1],box_max[2]);
		temp = tf * (box_min - box_center);
		glVertex3f(temp[0],temp[1],temp[2]);
		temp = tf * (box_max - box_center);
		glVertex3f(temp[0],temp[1],temp[2]);
		glEnd();
		glEnable(GL_LIGHTING);
	}
}

qglviewer::ManipulatedFrame* ShapePoser::SelectionGroup::frame()
{
	return &mf;
}

size_type ShapePoser::SelectionGroup::size() const
{
	return selected.size();
}

void ShapePoser::SelectionGroup::updateCorrespList( CorrespondenceList& updateList, int idxOffset ) const
{
	SE3 tf = getSE3(this->mf);

	for (int i = 0; i < selected.size(); ++i) {
		int idx = idxOffset + i;
		updateList.sourcePoints[idx] = vertices[selected[i]];
		updateList.sourceIndex[idx] = selected[i];

		// Transform the target points
		updateList.targetPoints[idx] = tf * (vertices[selected[i]] - box_center);
		updateList.targetIndex[idx] = 0;

		// Update the weights
		updateList.weight[idx] = 1.0f;
	}
}

