#ifndef __SHAPE_POSER_H__
#define __SHAPE_POSER_H__

#include <QGLViewer/qglviewer.h>
#include <QMenu>
#include <deque>
#include "types.h"

class Scene;
class ShapeModel;
class PointList;
class CorrespondenceList;
class SE3;
class QMutex;

namespace Deform {
	class DrawOptions;

	namespace DeformAll_Utility {
		class RefWeight;
		typedef std::deque<RefWeight> RefWeightList;
		class SE3_6DOF;
		typedef std::vector<SE3_6DOF> TransformList;
	}

	namespace DeformGraphAll_Utility {
		struct Joint;
		typedef std::map<int,Joint> JointMap;

	}
}

class ShapePoser : public QGLViewer
{
	Q_OBJECT

public:
	ShapePoser( Scene *sc, ShapeModel* s );				/**< Constructor: initialize members here. */
	virtual ~ShapePoser();								/**< Destructor: destroy members here. */

	void updateInteractiveCorresp( CorrespondenceList& updateList ) const;
	const ShapeModel* getShape() const {
		return shape;
	}
	void setDef( ShapeModel *def, std::vector<bool> *defVertexValid, QMutex *defLock ) {
		this->def = def;
		this->defVertexValid = defVertexValid;
		this->drawDefLock = defLock;
	}
	void setDef( ShapeModel *def, Deform::DeformGraphAll_Utility::JointMap* defJoints, Deform::DeformAll_Utility::RefWeightList *defWeights, std::vector<bool> *defVertexValid, QMutex *defLock ) {
		this->def = def;
		this->jointMap = defJoints;
		this->shapeWeights = defWeights;
		this->defVertexValid = defVertexValid;
		this->drawDefLock = defLock;
	}
	Deform::DeformAll_Utility::RefWeightList *getWeights() {
		return shapeWeights;
	}
	void subsampleConstraints( int maxConstraints );

public slots:
	// Load/Save Posing result
	void loadPoseResult();
	void savePoseResult();

protected :
	virtual void draw();
	virtual void init();
	virtual void drawWithNames();
	void drawSelectionRectangle() const;
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);
	virtual void keyPressEvent(QKeyEvent *e);
	virtual void endSelection(const QPoint&);
	virtual void postSelection( const QPoint& );

private:
	void testLoadJoints( Deform::DeformAll_Utility::TransformList& tfList, int motionIndex );

	// Pointer to the entire scene
	Scene *scene;

	// Pointer to the shape
	ShapeModel *shape, *def;
	std::vector<bool> *defVertexValid;
	RealColor shapeColor;
	Deform::DeformAll_Utility::RefWeightList *shapeWeights;
	Deform::DeformGraphAll_Utility::JointMap *jointMap;
	bool showWeights;
	std::vector<RealColor> boneColors;
	idxList vertexSubsample;
	QMutex *drawDefLock;
	bool showSourceWeights;
	std::vector<bool> sourceVertexValid;
	int currentMotion;

	//////////////////////////////////////////////////////////////////////////
	// Selection variables
	// Current rectangular selection
	QRect rectangle_;

	// Different selection modes
	enum SelectionMode { NONE, NEW_GROUP, SELECT_GROUP };
	SelectionMode selectionMode_;

	// Point select threshold
	const int pointSelect;

	//////////////////////////////////////////////////////////////////////////
	// Menu variables and functions
	void createContextMenu();
	QMenu* contextMenu;
	QAction *loadPoseResultAct;
	QAction *savePoseResultAct;

	//////////////////////////////////////////////////////////////////////////
	// Selection Groups

	static SE3 getSE3(const qglviewer::ManipulatedFrame& mf);
	
	class SelectionGroup {
	public:
		SelectionGroup( const PointList& vertices );

		void addIdSelection( index_type id );
		void eraseIdSelection( index_type id );
		void computeBox();
		void updateCorrespList( CorrespondenceList& updateList, int idxOffset ) const;
		size_type size() const;

		void draw( bool showSource ) const;
		qglviewer::ManipulatedFrame* frame();

	private:
		idxList selected;
		const PointList& vertices;
		real3_type box_min, box_max, box_center;
		qglviewer::ManipulatedFrame mf;
	};

	std::deque<SelectionGroup*> selectionGroups;
	int selectedGroup;

	//////////////////////////////////////////////////////////////////////////
	// Display options
	Deform::DrawOptions getDeformDrawOptions();
	bool showSource;
	bool drawSelectionBoxes;
	real_type jointDisplaySize;
	bool setSourceMFMode;
	qglviewer::ManipulatedFrame sourceMF;
};

#endif
