#include "TransformViewer.h"
#include "geom_util.h"
#include "scene.h"
#include "../Util/util.h"

#ifdef COMPUTE_TRIANGLE_DIST
#include <omp.h>
#endif

using namespace qglviewer;
using namespace std;

void loadRegistrationErrorBinary( const char *loadfileName, ShapeModel *targetShape, distList &error, ErrorStats &stats )
{
	// Load the data
	ifstream in;
	in.open(loadfileName, ios::binary);

	if (in.fail()) {
		cout << "Error in loading registration error stats: cannot open file " << loadfileName << " for reading!" << endl;
	}
	else {
		//cout << "Loading binary registration error stats in file " << loadfileName << "...";

		index_type numVertices;
		in.read((char*) &numVertices, sizeof(index_type));

		if (targetShape && numVertices == targetShape->numVertices()) {
			error.resize(numVertices);

			in.read((char*) &stats, sizeof(ErrorStats));
			in.read((char*) &(error[0]), sizeof(real_type) * targetShape->numVertices());

			//cout << " done!" << endl;
		}
		else
		{
			cout << endl << "--> Error in reading: number of vertices does not match with loaded reference shape!" << endl;
		}

		in.close();
	}
}

void ErrorStats::print( real_type depthScan_gridSpacing, size_t numVertices )
{
	const std::string units("x grid spacing");
	cout << "*** Alignment statistics:" << endl;
	cout << "*** Grid Spacing: " << depthScan_gridSpacing << endl;
	cout << "*** Number of valid comparisons: " << numValid << " out of total " << numVertices << endl;
	cout << "*** Maximum distance to closest point: " << maxError << units << endl;
	cout << "*** Minimum distance to closest point: " << minError << units << endl;
	cout << "*** Average distance to closest point: " << avgError << units << endl;
	cout << "*** Variance of the distances to closest point: " << varError << endl;
}

void ErrorStats::printCSV( size_t numVertices )
{
	cout << numValid << "," << numVertices << "," << ((real_type)numValid) / ((real_type)numVertices) << ",";
	cout << maxError << "," << minError << "," << avgError << "," << varError << endl;
}

TransformViewer::TransformViewer( Scene *sc ) :
scene(sc), showSource(true), showTarget(true),
useSameColor(false), showRegistrationError(false),
showWireframe(false), contextMenu(NULL), loadShapeAct(NULL),
useFrameSource(-1), useFrameTarget(-1), useFrameError(-1),
useFramePts(-1), INVALID_ERROR_MARK(-1.f)
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	// Set the window title.
	setWindowTitle("Shape Compare");

	setAcceptDrops(true);
}

TransformViewer::~TransformViewer()
{
	if (contextMenu) delete contextMenu;
	if (loadShapeAct) delete loadShapeAct;

	for (int i = 0; i < sourceFrames.size(); ++i)
	{
		delete sourceFrames[i];
	}

	for (int i = 0; i < targetFrames.size(); ++i)
	{
		delete targetFrames[i];
	}

	for (int i = 0; i < ptsFrames.size(); ++i)
	{
		delete ptsFrames[i];
	}

	for (int i = 0; i < normalsFrames.size(); ++i)
	{
		delete normalsFrames[i];
	}
}

void TransformViewer::init()
{
	createContextMenu();
	restoreStateFromFile();

	//////////////////////////////////////////////////////////////////////////
	// Disable using current color as color of material
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);

	//////////////////////////////////////////////////////////////////////////
	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the color of invalid vertices when computing the registration error
	invalidRegistrationColor.setRgbF(0.9f, 0.9f, 0.9f);

	depthScan_gridSpacing = scene->pointSize();

	// Set as orthographic camera
	camera()->setType(qglviewer::Camera::ORTHOGRAPHIC);

	scene->colormap.load("jet.txt");
}

void TransformViewer::draw()
{
	glEnable(GL_LIGHTING);

	if (showRegistrationError) {
		drawRegistrationError();
		//drawColormap();
	}
	else {
		if (showWireframe) {
			glDisable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		}
		else {
			glEnable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}

		if (showSource) {
			const ShapeModel *drawShape = getCompareShape_Source();

			if (drawShape)
			{
				glEnable(GL_LIGHTING);
				if (useSameColor) setOpenGLMaterialColor(scene->targetColor);
				else setOpenGLMaterialColor(scene->sourceColor);

				drawShape->draw();
			}
			else
			{
				const PointList *drawPts = getPts();

				if (drawPts)
				{
					glDisable(GL_LIGHTING);
					if (useSameColor) setOpenGLColor(scene->targetColor);
					else setOpenGLColor(scene->sourceColor);
					PointList_Utility::draw(*drawPts, POINT_SQUARES);
				}
			}
		}

		if (showTarget) {
			const ShapeModel *drawShape = getCompareShape_Target();

			if (drawShape)
			{
				glEnable(GL_LIGHTING);
				setOpenGLMaterialColor(scene->targetColor);
				drawShape->draw();
			}
		}
	}
}



void TransformViewer::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		if (showSource && showTarget) {
			showSource = true;
			showTarget = false;
		}
		else if (showSource && !showTarget) {
			showSource = false;
			showTarget = true;
		}
		else if (!showSource && showTarget) {
			showSource = true;
			showTarget = true;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		useSameColor = !useSameColor;
		updateGL();
	}
	else if ((e->key()==Qt::Key_E) && (modifiers==Qt::CTRL)) {
		computeRegistrationError();
		showRegistrationError = true;
		updateGL();
	}
	else if ((e->key()==Qt::Key_B) && (modifiers==Qt::CTRL)) {
		batchComputeError();
	}
	else if ((e->key()==Qt::Key_R) && (modifiers==Qt::CTRL)) {
		showRegistrationError = !showRegistrationError;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier)) {
		scene->colormapMax += depthScan_gridSpacing * 0.1;
		cout << "Colormap Range: (" << scene->colormapMin << "," << scene->colormapMax << ")" << endl;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier)) {
		scene->colormapMax -= depthScan_gridSpacing * 0.1;
		if (scene->colormapMax < 0)
		{
			scene->colormapMax = 0;
		}
		cout << "Colormap Range: (" << scene->colormapMin << "," << scene->colormapMax << ")" << endl;
		updateGL();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::CTRL)) {
		// Pick a new color for an invalid vertex using the color picker
		QColor color = QColorDialog::getColor(invalidRegistrationColor, this);
		if (color.isValid()) {
			invalidRegistrationColor = color;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::ALT)) {
		QString savefileName;
		savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the registration error stats...", tr(""), QString("*.data"));
		if (savefileName.length() > 0) {
			saveRegistrationErrorText(savefileName);
		}
	}
	else if ((e->key()==Qt::Key_W) && (modifiers==Qt::CTRL)) {
		showWireframe = !showWireframe;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::CTRL)) {
		if (showRegistrationError && useFrameError >= 0)
		{
			// Increment the frame number
			if (++useFrameError >= errorFrames.size()) {
				useFrameError = 0;
			}
			cerr << "Using error frame: " << useFrameError << endl;
		}
		else if (useFrameSource >= 0)
		{
			// Increment the frame number
			if (++useFrameSource >= sourceFrames.size()) {
				useFrameSource = 0;
			}
			cerr << "Using frame: " << useFrameSource << endl;
		}
		else if (useFramePts >= 0)
		{
			// Increment the frame number
			if (++useFramePts >= ptsFrames.size()) {
				useFramePts = 0;
			}
			cerr << "Using pts frame: " << useFramePts << endl;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::CTRL)) {
		if (showRegistrationError && useFrameError >= 0)
		{
			// Decrement the frame number
			if (--useFrameError < 0) {
				useFrameError = errorFrames.size() - 1;
			}
			cerr << "Using error frame: " << useFrameError << endl;
		}
		else if (useFrameSource >= 0)
		{
			// Decrement the frame number
			if (--useFrameSource < 0) {
				useFrameSource = sourceFrames.size() - 1;
			}
			cerr << "Using frame: " << useFrameSource << endl;
		}
		else if (useFramePts >= 0)
		{
			// Decrement the frame number
			if (--useFramePts < 0) {
				useFramePts = ptsFrames.size() - 1;
			}
			cerr << "Using pts frame: " << useFramePts << endl;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::ALT)) {
		if (showRegistrationError && useFrameError >= 0)
		{
			// Increment the frame number
			if (++useFrameError >= errorFrames.size()) {
				useFrameError = 0;
			}
			cerr << "Using frame: " << useFrameError << endl;
		}
		else if (useFrameTarget >= 0)
		{
			// Increment the frame number
			if (++useFrameTarget >= targetFrames.size()) {
				useFrameTarget = 0;
			}
			cerr << "Using frame: " << useFrameTarget << endl;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::ALT)) {
		if (showRegistrationError && useFrameError >= 0)
		{
			// Decrement the frame number
			if (--useFrameError < 0) {
				useFrameError = errorFrames.size() - 1;
			}
			cerr << "Using frame: " << useFrameError << endl;
		}
		else if (useFrameTarget >= 0)
		{
			// Decrement the frame number
			if (--useFrameTarget < 0) {
				useFrameTarget = targetFrames.size() - 1;
			}
			cerr << "Using frame: " << useFrameTarget << endl;
		}
		updateGL();
	}
	else
	{
		QGLViewer::keyPressEvent(e);
	}
}

void TransformViewer::computeRegistrationError()
{
	ShapeModel *targetShape = getCompareShape_Target();

#ifdef COMPUTE_TRIANGLE_DIST
	Mesh3_real *compShape = (Mesh3_real*) getCompareShape_Source();
	if (!targetShape || !compShape || compShape->shapeType != SHAPE_MESH) {
		cout << "Error in computeRegistrationError(): The shape to compare is not a mesh!" << endl;
		return;
	}
#else
	PointList *compPoints = getPts();
	if (!targetShape || !compPoints) {
		cout << "Error in computeRegistrationError(): No shape to compare!" << endl;
		return;
	}
#endif

#ifdef PRE_PRUNE_DIST
	const real_type maxDist_Multiplier = 10.0;
	const real_type maxDistSq = (maxDist_Multiplier * maxDist_Multiplier * depthScan_gridSpacing * depthScan_gridSpacing);
#endif

	cout << "Computing the registration error of all vertices in the shape..." << endl;
	cout << "--> Reference shape: " << qPrintable(targetShape->uid()) << endl;
#ifdef COMPUTE_TRIANGLE_DIST
	cout << "--> Comparison shape: " << qPrintable(compShape->uid()) << endl;
#else
	cout << "--> Comparison point list: " << compPoints->size() << " vertices" << endl;
#endif
#ifdef PRE_PRUNE_DIST
	cout << "--> Distance threshold to first vertex of triangle: " << maxDist_Multiplier << "x grid scan spacing" << endl;
#endif

	ErrorStats stats;
	registrationError.resize(targetShape->numVertices());

#ifdef COMPUTE_TRIANGLE_DIST
	// Set OpenMP properties
	const int numMaxThreads = 10;
	omp_set_dynamic(0);
	omp_set_num_threads(numMaxThreads);
#endif

	cout << "Progress:";
	TextProgress(targetShape->numVertices(), true);
	for (index_type i = 0; i < targetShape->numVertices(); i++)
	{
		// Skip the computation if vertex is not valid
		if (targetShape->markVertex[i] != VERTEX_VALID)
		{
			registrationError[i] = INVALID_ERROR_MARK;
			//cout << "Found boundary vertex in target: " << i << endl;
			continue;
		}

		real_type error = REAL_TYPE_MAX;
		const real3_type &ref = targetShape->vertex[i];

#ifdef COMPUTE_TRIANGLE_DIST
		const int t_end = compShape->triangle_vertex_index.size();
		std::vector<real_type> minDist(numMaxThreads, REAL_TYPE_MAX);
		std::vector<int> minRegionType(numMaxThreads);
		std::vector<index_type> minTriInd(numMaxThreads);

		// Compute the distance to all the triangles in the source mesh; find the minimum
		#pragma omp parallel
		{
			#pragma omp for nowait

			for(int t = 0; t < t_end; t++)
			{
				real_type dist;
				const index3_type& tvi = compShape->triangle_vertex_index[t];
				real3_type v1, v2, v3, closest;

				v1 = compShape->vertex[tvi[0]];
				v2 = compShape->vertex[tvi[1]];
				v3 = compShape->vertex[tvi[2]];

#ifdef PRE_PRUNE_DIST
				// Pre-prune strategy
				real3_type diff = ref - v1;
				real_type distApprox = diff.square_norm();
				if (distApprox < maxDistSq)
#endif
				{
					int type = pointTriangleSqDistance(ref, v1, v2, v3, closest, dist);

					if (dist < minDist[omp_get_thread_num()]) {
						minDist[omp_get_thread_num()] = dist;
						minRegionType[omp_get_thread_num()] = type;
						minTriInd[omp_get_thread_num()] = t;
					}
				}
			}
		}

		////////////////////////////////////////////////////////////////////////
		int mi = 0;
		for (int j = 0; j < numMaxThreads; ++j)
		{
			if (minDist[j] < error)
			{
				error = minDist[j];
				mi = j;
			}
		}

		// Check to see that the closest point is not on the boundary of compShape
		if (error == REAL_TYPE_MAX || compShape->triangleRegionOnBoundary(minTriInd[mi], minRegionType[mi]))
		{
			registrationError[i] = INVALID_ERROR_MARK;
			//cout << "Closest point is on the boundary for vertex: " << i << endl;
		}
		else
#else
		idxList nearest;
		distList dists;
		compPoints->queryKNN(ref, 1, nearest, dists, true);
		error = dists[0];
#endif
		{
			//// Divide the minimum error by the scene size
			//error = sqrt(error) / scene->totalScene_boundingBoxDiagonal;
			// Divide the minimum error by the grid spacing length
			error = sqrt(error) / depthScan_gridSpacing;

			if (error > stats.maxError) stats.maxError = error;
			if (error < stats.minError) stats.minError = error;
			stats.avgError += error;
			stats.varError += error*error;
			stats.numValid++;

			// Store result
			registrationError[i] = error;
		}

		//cout << "Error in vertex " << i << ": " << registrationError[i] << endl;
		TextProgress(i);
	}

	cout << " done!" << endl;

	// Compute average and variance
	stats.avgError = stats.avgError / ((real_type)stats.numValid);
	stats.varError = stats.varError / ((real_type)stats.numValid) - stats.avgError*stats.avgError;
	
	// Assign statistics
	registrationStats = stats;

	stats.print(depthScan_gridSpacing, targetShape->numVertices());
}

void TransformViewer::drawRegistrationError()
{
	ShapeModel *targetShape = getCompareShape_Target();
	distList *getError = getRegistrationError();
	if (!getError)
	{
		getError = &registrationError;
	}
	distList &error = *getError;

	// Draw the error on the target mesh
	if (!targetShape || error.size() != targetShape->numVertices()) {
		return;
	}

	if (targetShape->shapeType == SHAPE_MESH) {
		const Mesh3_real &drawShape = *((Mesh3_real*)targetShape);

		glDisable(GL_LIGHTING);
		glBegin(GL_TRIANGLES);
		for(size_type t = 0,t_end = drawShape.triangle_vertex_index.size();t < t_end;t++){
			const index3_type& tvi = drawShape.triangle_vertex_index[t];

			// Check to see if any of the vertices is invalid
			bool markTriangle = false;

			for (int i = 0; i < 3; i++) {
				if (error[tvi[i]] < 0) {
					markTriangle = true;
					//cout << " " << tvi[i];
					break;
				}
			}

			// Draw the triangle
			for(size_type i = 0;i < 3;i++) {
				const real3_type& v = drawShape.vertex[tvi[i]];

				// Set invalid comparison triangles to a special color
				if (markTriangle) {
					setOpenGLColor(invalidRegistrationColor.rgba());
				}
				else {
					setOpenGLColor(scene->getColormapColor(error[tvi[i]]));
				}
				glVertex3f(v[0],v[1],v[2]);
			}
		}
		glEnd();
	}
}

void TransformViewer::drawColormap()
{
	const int subdivisions = 100;
	float w = width();
	float h = height();
	float mapWidth = 30;
	float mapOffset = 20;
	float top = 0.25*h;
	float bottom = 0.75*h;
	float left = w - mapWidth - mapOffset;
	float right = w - mapOffset;
	float current = top;
	float offset = abs(bottom - top) / ((float)subdivisions);

	startScreenCoordinatesSystem();
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);

	glBegin(GL_QUAD_STRIP);
	for (int i = 0; i < subdivisions; i++) {
		RealColor currentColor = scene->colormap.getColor(((float)subdivisions - i - 1) / ((float)subdivisions - 1));
		setOpenGLColor(currentColor);
		glVertex2f(left,  current);
		glVertex2f(right, current);
		current += offset;
	}
	glEnd();

	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();

	// Draw the min and max labels
	drawText(left - mapOffset, top, tr("%1").arg(scene->colormapMax));
	drawText(left - mapOffset, bottom, tr("%1").arg(scene->colormapMin));
}

void TransformViewer::createContextMenu()
{
	contextMenu = new QMenu(tr("&Menu"), this);

	loadShapeAct = new QAction(tr("&Load Additional Shape..."), this);
	contextMenu->addAction(loadShapeAct);
	//connect(loadShapeAct, SIGNAL(triggered()), this, SLOT(loadNewFrame()));
}

void TransformViewer::processDropFiles( QStringList &fileList, const Qt::KeyboardModifiers modifiers )
{
	enum FILE_TYPES {
		FILE_TYPE_NONE = 0,
		FILE_TYPE_OBJ,
		FILE_TYPE_XML,
		FILE_TYPE_ERROR_DATA,
		FILE_TYPE_PTS,
		FILE_TYPE_NUM,
	};

	const char* const fileTypeStr[] = {
		"none",
		".obj",
		".xml",
		".err",
		".pts",
	};

	int allTypes = -1;
	vector<int> classify(fileList.size(), 0);
	for (int i = 0; i < fileList.size(); ++i)
	{
		QString &str = fileList[i];

		for (int j = 1; j < FILE_TYPE_NUM; ++j)
		{
			// Compare the tail of the two strings
			if (str.endsWith(tr(fileTypeStr[j]), Qt::CaseInsensitive))
			{
				classify[i] = j;
				break;
			}
		}

		// Check to see if all types are the same
		if (i == 0 && classify[i] > 0)
		{
			allTypes = classify[i];
		}
		else if (allTypes > 0 && allTypes != classify[i])
		{
			allTypes = -1;
		}
	}

	if (allTypes == FILE_TYPE_XML)
	{

	}
	else if (allTypes == FILE_TYPE_OBJ)
	{
		processObjSequence(fileList, (modifiers == Qt::CTRL));
	}
	else if (allTypes == FILE_TYPE_ERROR_DATA)
	{
		if (fileList.size())
		{
			ShapeModel *targetShape = getCompareShape_Target();
			errorFrames.resize(fileList.size());

			// Load the registration error file
			for (int i = 0; i < fileList.size(); ++i)
			{
				fileList[i].remove(0, 1);
				loadRegistrationErrorBinary(qPrintable(fileList[i]), targetShape, errorFrames[i], registrationStats);
				//registrationStats.print(depthScan_gridSpacing, targetShape->numVertices());

				cout << i << ",";
				registrationStats.printCSV(targetShape->numVertices());
			}
			useFrameError = 0;
		}
	}
	else if (allTypes == FILE_TYPE_PTS)
	{
		if (fileList.size())
		{
			real3_type box_min, box_max;
			ptsFrames.resize(fileList.size());
			normalsFrames.resize(fileList.size());

			// Load the registration error file
			for (int i = 0; i < fileList.size(); ++i)
			{
				ptsFrames[i] = new PointList();
				normalsFrames[i] = new PointList();
				fileList[i].remove(0, 1);
				readPtsFormat(qPrintable(fileList[i]), box_min, box_max, *(ptsFrames[i]), *(normalsFrames[i]));
			}
			useFramePts = 0;
		}
	}
}

void TransformViewer::processObjSequence( QStringList &fileList, bool loadTargetFrames )
{
	if (fileList.size() == 0)
	{
		//QString loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the reference shape...", scene->lastDir, tr("*.obj"));
		//if (loadfileName.length() == 0) {
		//	cout << "Viewer::CreateTransformViewer(): Loading canceled." << endl;
		//	return;
		//}

		return;
	}

	if (loadTargetFrames)
	{
		cout << "Opening " << fileList.size() << " reference shapes." << endl;
		int newUseFrameTarget = targetFrames.size();

		// Prompt for a scale factor
		bool ok;
		double scale = QInputDialog::getDouble(this, tr("Scale Factor"),
			tr("Factor Scale:"), 1.0, 0.0, 1e10, 4, &ok);
		if (!ok) {
			scale = 1.0;
		}

		// Prompt for the grid sample spacing
		if (newUseFrameTarget == 0)
		{
			double spacing = QInputDialog::getDouble(this, tr("Grid Sample Spacing"),
				tr("Grid Sample Spacing:"), 1.0, 0.0, 1e5, 7, &ok);
			if (!ok) {
				spacing = 1.0;
			}

			// Set the spacing
			depthScan_gridSpacing = spacing;
		}

		for (int i = 0, i_end = fileList.size(); i < i_end; ++i)
		{
			fileList[i].remove(0, 1);
			ShapeModel *shape = loadShape(scene->rootDir.c_str(), fileList[i], *scene, false);

			// Apply a uniform scale to the vertices and the bounding box
			if (scale != 1.0)
			{
				for (int i = 0; i < shape->numVertices(); ++i) {
					shape->vertex[i] = scale * shape->vertex[i];
				}
				for (int d = 0; d < 3; ++d)
				{
					shape->box_min[d] *= scale;
					shape->box_max[d] *= scale;
				}
			}

			targetFrames.push_back(shape);
		}


		// Set the shape to display
		ShapeModel *targetShape = targetFrames[newUseFrameTarget];
		setSceneBoundingBox(qglviewer::Vec(targetShape->box_min[0], targetShape->box_min[1], targetShape->box_min[2]),
			qglviewer::Vec(targetShape->box_max[0], targetShape->box_max[1], targetShape->box_max[2]));
		showEntireScene();
		useFrameTarget = newUseFrameTarget;
	}
	else
	{
		cout << "Opening " << fileList.size() << " comparison shapes." << endl;
		int newUseFrameSource = sourceFrames.size();

		for (int i = 0, i_end = fileList.size(); i < i_end; ++i)
		{
			fileList[i].remove(0, 1);
			ShapeModel *shape = loadShape(scene->rootDir.c_str(), fileList[i], *scene, false);
			sourceFrames.push_back(shape);
		}

		useFrameSource = newUseFrameSource;
	}
}

void TransformViewer::dragEnterEvent( QDragEnterEvent *event )
{
	//setText(tr("<drop content>"));
	//setBackgroundRole(QPalette::Highlight);

	event->acceptProposedAction();
	//emit changed(event->mimeData());
}

void TransformViewer::dropEvent( QDropEvent *event )
{
	const QMimeData *mimeData = event->mimeData();
	const Qt::KeyboardModifiers modifiers = event->keyboardModifiers();

	if (mimeData->hasImage())
	{
		//setPixmap(qvariant_cast<QPixmap>(mimeData->imageData()));
		cout << "Input is image" << endl;
	}
	else if (mimeData->hasHtml())
	{
		//setText(mimeData->html());
		//setTextFormat(Qt::RichText);
		cout << "Input is HTML" << endl << qPrintable(mimeData->html()) << endl;
	}
	else if (mimeData->hasText())
	{
		//setText(mimeData->text());
		//setTextFormat(Qt::PlainText);
		cout << "Input is text" << endl << qPrintable(mimeData->text()) << endl;
	}
	else if (mimeData->hasUrls())
	{
		QList<QUrl> urlList = mimeData->urls();
		QStringList strList;
		for (int i = 0; i < urlList.size(); ++i) {
			QString url = urlList.at(i).path();
			strList.append(url);
		}
		processDropFiles(strList, modifiers);
		//setText(text);
		//cout << qPrintable(text) << endl;
	}
	else
	{
		//setText(tr("Cannot display data"));
		cout << "Cannot display data" << endl;
	}

	//setBackgroundRole(QPalette::Dark);
	event->acceptProposedAction();
}

void TransformViewer::saveRegistrationErrorText( QString &savefileName )
{
	ShapeModel *targetShape = getCompareShape_Target();

	// If the registration error has been computed
	if (targetShape && registrationError.size() == targetShape->numVertices()) {
		// Save the data
		ofstream out;
		out.open(qPrintable(savefileName));
		ShapeModel *compShape = getCompareShape_Source();

		if (out.fail()) {
			cout << "Error in saving registration error stats: cannot open file " << qPrintable(savefileName) << " for writing!" << endl;
		}
		else {
			cout << "Saving registration error stats in file " << qPrintable(savefileName) << "...";

			out << "Registration error from target vertices to transformed source" << endl;

			// Output the source and target file names
			out << "Source shape:\t" << qPrintable(compShape->uid()) << endl;
			out << "Target shape:\t" << qPrintable(targetShape->uid()) << endl;
			//out << "Bounding box diagonal length\t" << scene->pointSize(1.0f) << endl;
			out << "Range scan grid spacing length\t" << depthScan_gridSpacing << endl;
			//out << "Vertex Index\tRegistration Error (% of bounding box diagonal)" << endl;
			out << "Vertex Index\tRegistration Error (multiple of grid spacing)" << endl;
			for (int i = 0; i < targetShape->numVertices(); i++) {
				out << i << "\t" << registrationError[i] << endl;
			}

			out.close();
		}
	}
}

void TransformViewer::saveRegistrationErrorBinary( const char *savefileName )
{
	ShapeModel *targetShape = getCompareShape_Target();

	// If the registration error has been computed
	if (targetShape && registrationError.size() == targetShape->numVertices()) {
		// Save the data
		ofstream out;
		out.open(savefileName, ios::binary);
		ShapeModel *compShape = getCompareShape_Source();

		if (out.fail()) {
			cout << "Error in saving registration error stats: cannot open file " << savefileName << " for writing!" << endl;
		}
		else {
			cout << "Saving binary registration error stats in file " << savefileName << "...";

			index_type numVertices = targetShape->numVertices();
			out.write((char*) &numVertices, sizeof(index_type));
			out.write((char*) &registrationStats, sizeof(ErrorStats));
			out.write((char*) &(registrationError[0]), sizeof(real_type) * targetShape->numVertices());

			out.close();
		}
	}
}

void TransformViewer::batchComputeError()
{
	if (!targetFrames.size() || (!sourceFrames.size() && !ptsFrames.size()) ||
		((targetFrames.size() != ptsFrames.size()) && (targetFrames.size() != sourceFrames.size()) && (targetFrames.size() != 1)))
	{
		cout << "No frames to compare or wrong sizes (" << targetFrames.size() << "," << sourceFrames.size() << ")!" << endl;
		return;
	}

	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the registration errors...", "", QString("*.err"));
	if (savefileName.length() == 0) {
		savefileName = tr("computeError_###.err");
	}

	if (targetFrames.size() == 1)
	{
		// Compare the single target frame to all source frames
		useFrameTarget = 0;

		for (int i = 0; i < sourceFrames.size(); ++i)
		{
			useFrameSource = i;
			computeRegistrationError();

			// Decide on a save file
			string saveFile(qPrintable(savefileName));
			replaceField(saveFile, '#', i);
			saveRegistrationErrorBinary(saveFile.c_str());
		}
	}
	else if (targetFrames.size() == sourceFrames.size())
	{
		for (int i = 0; i < sourceFrames.size(); ++i)
		{
			useFrameTarget = useFrameSource = i;
			computeRegistrationError();

			// Decide on a save file
			string saveFile(qPrintable(savefileName));
			replaceField(saveFile, '#', i);
			saveRegistrationErrorBinary(saveFile.c_str());
		}
	}
	else if (targetFrames.size() == ptsFrames.size())
	{
		for (int i = 0; i < ptsFrames.size(); ++i)
		{
			useFrameTarget = useFramePts = i;
			computeRegistrationError();

			// Decide on a save file
			string saveFile(qPrintable(savefileName));
			replaceField(saveFile, '#', i);
			saveRegistrationErrorBinary(saveFile.c_str());
		}
	}
}
