#include "scene.h"
#include <ctime>
#include <utility>
#include <iostream>
#include "FileSequence.h"
#include "DeformGraphAll.h"
#include "DeformEdit.h"
#include "ShapePoser.h"
#include "../Util/util.h"
#include "viewerCommon.h"

using namespace std;
using namespace Deform;

extern "C" {
	#include <SFMT/SFMT.h>
}

const char *globalForegroundColor = "black";
const char *globalBackgroundColor = "white";

Scene::Scene() : animation(NULL), deform(NULL), target(NULL), source(NULL), t(NULL),
cl(NULL), matchCandidates(NULL),
graph_n(10), graph_k(8), solveInitialRigid(0), solveFinalNNLS(false),
markCellsWithPoints(false), useAccumulatedGeometry(false), maxCorrespondenceDistance(0.0), maxNormalAngleDifference(90.0), boundaryCorrespondenceDistance(0.0),
solverMaxIterations(50), estepMaxIterations(20), projMaxIterations(50), IRLSMaxIterations(50),
gridDivisions(10), solveDiscreteWStep(true), solveSimultaneousReg(true), refFrameIndex(0),
numCorrespondences(20), bonesPerGridPoint(4),
maxAutoBoneNum(50), minBoneVertexSupport(20), numGridCellsExpand(1),
defaultLoadFrames(0), defaultSkipFrames(1),
depthScan_gridSpacing(0.0f),
sourceFrame(0), targetFrame(0), autoLoadFrameData(1),
autoProcessMode(AUTO_PROCESS_NONE),
currentBone(-1), colormapMin(0), colormapMax(0),
correspondenceType(CORRESPONDENCE_NONE),
robust_PC(false), ANNEpsilon_MC(0),
featureSamplesSource(1000), featureSamplesTarget(1000), featureType(FEATURE_PRINCIPAL_CURVATURES),
ANNEpsilon_FM(0), radius_FM(0), maxCandidates_FM(100), threshold_FM(0),
maxDist_FM(100), pruneMaxRatio_FM(1.0),
ANNEpsilon_MS(0), radius_MS(0), maxIter_MS(100), rate_MS(1),
ANNEpsilon_VR(0), minVerifyClusterSize(0), minMatchSize_VR(0), maxNumVerifyClusters_VR(-1), maxNumLabels_GC(-1),
growRegionDistThreshold_VR(1), skipVerification_VR(false), graphCutSolutionType_GC(2), useDummyLabel_GC(false),
mismatchError_GC(10000), dummyLabelSmoothError_GC(100), dummyLabelDataError_GC(100), bdySmoothnessKNN_GC(0), closestPtDistThreshold_GC(1),
smoothWeight_GC(1.0), dataWeight_GC(1.0), smoothTruncateValue_GC(100.0), dataTruncateValue_GC(100.0),
dataTermType_GC(GRAPH_CUT_DATA_POINT_TO_POINT), graphCutInstance_GC(GRAPH_CUT_INSTANCE_SYMMETRIC),
RANSAC_max_itr(0), RANSAC_sourceModel_ANNEpsilon(0), RANSAC_targetModel_ANNEpsilon(0),
RANSAC_inlier_queryRadius(0), RANSAC_inlier_epsilonBand(0), RANSAC_minimumInliersPercentage(0),
showClusterMode(0), poser(NULL), screenType(0),
refSampleDistance_Multiplier(2.0), splitRegionMinError(0.05),
maxJointDistance_Multiplier(20.0), hingeJointNumFrames(5), slidingWindowSize(5),
loadGridOnly(false), solveIK_DeformAll(false), occludedLabel_ValidCorrespRatioThreshold(0.1), finalSmoothWeights(0),
triggerGCMatchFinished(false), triggerFinalShot(false), initGCMatchRef(false), frameChangeTrigger(-1), minMatchRatio(0.1),
excludeLabelThreshold_percentage(0.01), finalRefSampleDistance_Multiplier(0.0), accumulateASGOnReference(false), computeTotalEfit(false)
{
	//////////////////////////////////////////////////////////////////////////
	// Pick random numbers
	//////////////////////////////////////////////////////////////////////////
	// Initialize random number generator
	init_gen_rand((unsigned)time( NULL ));
	for (index_type i = 0; i < 10000; i++) {
		// generate a random number
		cpi.push_back(genrand_real1());
	}

	//////////////////////////////////////////////////////////////////////////
	// Initialize color defaults
	sourceColor.resize(3);
	targetColor.resize(3);
	pointColor.resize(3);
	selectedColor.resize(3);
	sourceColor[0] = 0.7f; sourceColor[1] = 0.7f; sourceColor[2] = 1.0f;
	targetColor[0] = 1.0f; targetColor[1] = 0.7f; targetColor[2] = 0.7f;
	pointColor[0] = 1.0f; pointColor[1] = 1.0f; pointColor[2] = 0.0f;
	selectedColor[0] = 0.0f; selectedColor[1] = 1.0f; selectedColor[2] = 1.0f;

	// Initialize default option values
	setDefaultOptions();
}

Scene::~Scene()
{
	clear();
}

void Scene::setupDeform( int deformType )
{
	if ((deformType == DEFORM_GRAPH_ALL) && meshSequence.length() && rootDir.length()) {
		// Even if we don't have an animation, check if we have the "meshSequence" and "rootDir" information for a list of files.
		// Delete the existing deformation object and create a new one
		if (deform) {
			deform->Detach(this);
			delete deform;
			deform = NULL;
		}
		deform = new DeformGraphAll(getDeformAllOptions());
		((DeformGraphAll*)deform)->init(rootDir, meshSequence);
		deform->Attach(this);
	}
}

void Scene::setDefaultOptions()
{
	maxJointDistance_Multiplier = 20.f;
	hingeJointNumFrames = 3;
	slidingWindowSize = 5;
	splitRegionMinError = 0.4f;
	excludeLabelThreshold_percentage = 0.01f;
	minMatchRatio = 0.05;
	occludedLabel_ValidCorrespRatioThreshold = 0.1f;
	occluded_jointOptWeight = 1000.0f;
	accumulateASGOnReference = false;
	disableJointConstraint = false;
	disableHingeJoints = false;
}

Deform::DeformAll_Utility::Options Scene::getDeformAllOptions()
{	
	// Populate options
	Deform::DeformAll_Utility::Options options;
	real_type sceneGridSpacing = pointSize();

	////////////////////////////////////////////////////////////////////
	// Data dependent parameters
	options.sceneGridSpacing = sceneGridSpacing;
	options.refSampleDist = refSampleDistance_Multiplier * sceneGridSpacing;
	options.numBones = graph_n;

	////////////////////////////////////////////////////////////////////
	// Matching closest points
	options.maxCorrepDist = 10.f * sceneGridSpacing;			// maxCorrespondenceDistance * sceneGridSpacing;
	options.minNormalCosAngle = cos(45.0 / 180.0 * M_PI);		// cos(maxNormalAngleDifference / 180.0 * M_PI);
	options.bdyMaxCorrepDist = 1.f * sceneGridSpacing;			// boundaryCorrespondenceDistance * sceneGridSpacing;
	options.closestPt_normalWeight = 2.0f * sceneGridSpacing;		// used in ShapeMode::findClosestPoint, recreates the 

	// Detecting and constraining joints
	options.pruneJoint_minEdgeRatio = 0.15;
	options.maxJointDist = 20.f * sceneGridSpacing;				// maxJointDistance_Multiplier * sceneGridSpacing
	options.hingeJointNumFrames = 3;							// hingeJointNumFrames
	options.jointIsHingeThreshold = 0.1f;						// jointIsHingeThreshold, but fixed to 0.1
	options.hingeJointSamples = 20;
	options.hingeExtent = 10.0f * sceneGridSpacing;

	// Optimization
	options.TWeights = em_multiframeEWeights;					// options.WWeights.swap(FloatArray(2, 1.f));
	options.WWeights = em_DiscreteMWeights;						// options.TWeightsIK.swap(FloatArray(2, 1.f));
	options.WWeights[1] *= sceneGridSpacing;
	options.slidingWindowSize = 5;								// slidingWindowSize

	// Optimizing weights
	options.excludeLabelThreshold_num = 5;
	options.excludeLabelThreshold_ratio = 0.01f;				// excludeLabelThreshold_percentage;
	options.splitRegionMinError = 0.4f;							// splitRegionMinError;		/**< Prevent any minor reusing of labels for now, change back to 0.1 or 0.05 */

	// Solver control
	options.maxSolverIterations = 7;							// solverMaxIterations;
	options.maxTStepIterations = 30;							// estepMaxIterations;

	// DSG parameters
	options.finalRefSampleDist = refSampleDistance_Multiplier * 0.5 * sceneGridSpacing;
	options.EKNNGraph_k = 15;									// graph_k;
	options.DSGMaxEdgeStretchFactor = 2.0f;
	options.DSGMaxEdgeLen = 4.0f * options.refSampleDist;
	options.finalDSGMaxEdgeStretchFactor = 10.0f;
	options.finalDSGMaxEdgeLen = 10.0f * options.refSampleDist;

	// Occlusion handling parameters
	options.occludedLabel_ValidCorrespNumThreshold = 5;
	options.occludedLabel_ValidCorrespRatioThreshold = 0.1f;	// occludedLabel_ValidCorrespRatioThreshold;
	options.minMatchRatio = 0.05f;								// minMatchRatio;
	options.occluded_jointOptWeight = 1000.0f;
	options.occluded_maxCorrespDist = 50.f * sceneGridSpacing;
	options.occluded_maxNormalAngleDifference = cos(45.0 / 180.0 * M_PI);

#define USE_FILE_PARAMS
#ifdef USE_FILE_PARAMS
	options.maxJointDist = maxJointDistance_Multiplier * sceneGridSpacing;
	options.hingeJointNumFrames = hingeJointNumFrames;
	options.slidingWindowSize = slidingWindowSize;
	options.excludeLabelThreshold_ratio = excludeLabelThreshold_percentage;
	options.splitRegionMinError = splitRegionMinError;				/**< Prevent any minor reusing of labels for now, change back to 0.1 or 0.05 */
	options.minMatchRatio = minMatchRatio;
	options.occludedLabel_ValidCorrespRatioThreshold = occludedLabel_ValidCorrespRatioThreshold;
	options.occluded_jointOptWeight = occluded_jointOptWeight;
#endif


	////////////////////////////////////////////////////////////////////
	// State control
	options.refFrameIndex = refFrameIndex;
	options.accumulateASGOnReference = false;					// accumulateASGOnReference
	options.performRebuildASG = true;

#ifdef USE_FILE_PARAMS
	options.accumulateASGOnReference = accumulateASGOnReference;
	if (options.accumulateASGOnReference)
	{
		options.slidingWindowSize = 1;
	}

	// Whether to disable joint constraints
	if (disableJointConstraint)
	{
		options.maxJointDist = -100.0f;
	}
	if (disableHingeJoints)
	{
		options.jointIsHingeThreshold = -100.f;
	}
#endif

	options.autoSaveSolverState = autoSaveSolverState;
	options.autoLoadSolverState = autoLoadSolverState;
	options.groundTruth_shapeFile = groundTruth_shapeFile;
	options.groundTruth_weightFile = groundTruth_weightFile;
	options.groundTruth_motionFile = groundTruth_motionFile;
	options.autoSaveSampleSpinImages = autoSaveSampleSpinImages;
	options.autoLoadSampleSpinImages = autoLoadSampleSpinImages;
	options.writeFrameToRef = writeFrameToRef;
	options.writeRefToFrame = writeRefToFrame;
	options.saveTaucsErrorLog = saveTaucsErrorLog;
	options.loadGraphOnly = loadGridOnly;
	options.initGCMatchRef = initGCMatchRef;
	options.loadFinalShape = loadFinalShape;
	options.computeTotalEfit = computeTotalEfit;
	options.ptr_scene = this;		// For computing/loading graph cut correspondences
	options.TWeightsIK = em_EWeights;							// options.TWeights.swap(FloatArray(2, 1.f));

	// For solving interactive IK on a rigged mesh
	options.loadRiggedShape = loadRiggedShape;
	options.loadRiggedWeights = loadRiggedWeights;
	options.loadRiggedBVH = loadRiggedBVH;


	////////////////////////////////////////////////////////////////////
	// Rendering parameters
	delete options.colormap;
	options.colormap = new ColormapRange(colormap, colormapMin, colormapMax);


	////////////////////////////////////////////////////////////////////
	// IK Rendering Parameters
	options.edgeLenStretchThreshold_Sq = boundaryCorrespondenceDistance*boundaryCorrespondenceDistance*sceneGridSpacing*sceneGridSpacing;		/**< Just use bdy correspondence distance for now */
	options.deformCosAngleThreshold = 0.99;


	////////////////////////////////////////////////////////////////////
	return options;
}


void Scene::setDeformOptions()
{
	//if (deform && (deform->getType() == DEFORM_EM)) {
	//	((DeformEM*)deform)->setOptions(getDeformEMOptions());
	//}
	//else if (deform && (deform->getType() == DEFORM_ALL)) {
	//	((DeformAll*)deform)->setOptions(getDeformAllOptions());
	//}
	//else
	if (deform && (deform->getType() == DEFORM_GRAPH_ALL)) {
		((DeformGraphAll*)deform)->setOptions(getDeformAllOptions());
	}
}

void Scene::findCorrespondences()
{
	switch(correspondenceType) {
		case CORRESPONDENCE_LOAD_FROM_FILE:
			// 1. Try loading correspondences from file
			clearCorrespondenceList();
			cl = new CorrespondenceList();
			cl->load(correspondenceFile);
			break;
		case CORRESPONDENCE_GROUND_TRUTH:
			setGroundTruthCorrespondences();
			break;
		case CORRESPONDENCE_CLOSEST_POINTS:
		case CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS:
			setClosestPointCorrespondences();
			break;
		case CORRESPONDENCE_GRAPH_CUT:
			// 1. Use transformation clustering to find correspondences
			initMatchCandidates(*source, *target);
			solveGraphCutCorrespondences(matchCandidates, NULL, NULL, NULL, NULL);

			// 2. Extract the correspondences from the graph cut result
			packageCorrespondences();
			break;
		case CORRESPONDENCE_SPIN_IMAGES:
			// Just set closest points for now, DeformEM will take care of actually assigning spin image correspondences
			setClosestPointCorrespondences();
			break;
		default:
			break;
	}
}

void Scene::solveGraphCutCorrespondences( MatchCandidates *matchCand, const PointList* sampledSourcePoints, const PointList* sampledTargetPoints, const BoostGraph* sampledSourceGraph, const BoostGraph* sampledTargetGraph )
{
	if (matchCand && (matchCand->doneClustering() || matchCand->doneVerify())) {
		return;
	}

	MEASURE_PERF_NARG(TIMER_SUB_SAMPLING_GEOMETRY,Start);
	matchCand->subSampleGeometry(featureSamplesSource, featureSamplesTarget, pruneMaxRatio_FM);
	MEASURE_PERF_NARG(TIMER_SUB_SAMPLING_GEOMETRY,Stop);

	MEASURE_PERF_NARG(TIMER_POPULATING_CORRESPONDENCE_CANDIDATES,Start);
	if (featureType == FEATURE_SPIN_IMAGES) {
		matchCand->spinImageCandidates(maxCandidates_FM);
	}
	else {
		matchCand->findCandidateCorrespondence(radius_FM*radius_FM, maxCandidates_FM, ANNEpsilon_FM);
	}
	MEASURE_PERF_NARG(TIMER_POPULATING_CORRESPONDENCE_CANDIDATES,Stop);

	MEASURE_PERF_NARG(TIMER_MEAN_SHIFT_CLUSTERING,Start);
	matchCand->meanShiftBasic(radius_MS*radius_MS,rate_MS,maxIter_MS,ANNEpsilon_MS);
	MEASURE_PERF_NARG(TIMER_MEAN_SHIFT_CLUSTERING,Stop);

	MEASURE_PERF_NARG(TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP,Start);
	if (skipVerification_VR) {
		matchCand->skipVerify(minVerifyClusterSize, maxNumVerifyClusters_VR);
	}
	else {
		matchCand->verifyRegion(minVerifyClusterSize, maxNumVerifyClusters_VR, makeSceneRadius(growRegionDistThreshold_VR), ANNEpsilon_VR);
	}
	matchCand->sortVerifyResult(minMatchSize_VR, maxNumLabels_GC);
	MEASURE_PERF_NARG(TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP,Stop);

	// Precompute closest points
	MEASURE_PERF_NARG(TIMER_PRECOMPUTING_CLOSEST_POINTS,Start);
	if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_FORWARD || graphCutInstance_GC == GRAPH_CUT_INSTANCE_SYMMETRIC) {
		matchCand->precomputeClosestPts();
	}
	else if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_PATCH_FORWARD || graphCutInstance_GC == GRAPH_CUT_INSTANCE_PATCH_SYMMETRIC) {
		if (sampledSourcePoints && sampledTargetPoints) {
			matchCand->patchLocations_source = sampledSourcePoints;
			matchCand->patchLocations_target = sampledTargetPoints;
		}
		matchCand->precomputeClosestPatches();
	}
	MEASURE_PERF_NARG(TIMER_PRECOMPUTING_CLOSEST_POINTS,Stop);


	MEASURE_PERF_NARG(TIMER_GRAPH_CUT_OPTIMIZATION,Start);
	if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_SYMMETRIC) {
		matchCand->GraphCutRelabel(
										graphCutSolutionType_GC,
										dataTermType_GC,
										useDummyLabel_GC,
										bdySmoothnessKNN_GC,
										makeSceneRadius(closestPtDistThreshold_GC),
										makeSceneRadius(mismatchError_GC),
										makeSceneRadius(dummyLabelSmoothError_GC),
										makeSceneRadius(dummyLabelDataError_GC),
										smoothWeight_GC,
										dataWeight_GC,
										makeSceneRadius(smoothTruncateValue_GC),
										makeSceneRadius(dataTruncateValue_GC));
	}
	else if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_FORWARD) {
		matchCand->GraphCutRelabel_Forward(
												graphCutSolutionType_GC,
												dataTermType_GC,
												smoothWeight_GC,
												dataWeight_GC,
												makeSceneRadius(smoothTruncateValue_GC),
												makeSceneRadius(dataTruncateValue_GC));
	}
	else if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_PATCH_FORWARD) {
		matchCand->GraphCutRelabel_PatchForward(
			*sampledSourceGraph,
			graphCutSolutionType_GC,
			dataTermType_GC,
			smoothWeight_GC,
			dataWeight_GC,
			makeSceneRadius(smoothTruncateValue_GC),
			makeSceneRadius(dataTruncateValue_GC));
	}
	else if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_PATCH_SYMMETRIC) {
		matchCand->GraphCutRelabel_Patch(
			*sampledSourceGraph,
			*sampledTargetGraph,
			graphCutSolutionType_GC,
			dataTermType_GC,
			useDummyLabel_GC,
			bdySmoothnessKNN_GC,
			makeSceneRadius(closestPtDistThreshold_GC),
			makeSceneRadius(mismatchError_GC),
			makeSceneRadius(dummyLabelSmoothError_GC),
			makeSceneRadius(dummyLabelDataError_GC),
			smoothWeight_GC,
			dataWeight_GC,
			makeSceneRadius(smoothTruncateValue_GC),
			makeSceneRadius(dataTruncateValue_GC));
	}
	else {
		cerr << "Unknown graph cut instance type " << graphCutInstance_GC << "!" << endl;
	}
	MEASURE_PERF_NARG(TIMER_GRAPH_CUT_OPTIMIZATION,Stop);

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Running time summary" << endl;
	printTime(TIMER_SUB_SAMPLING_GEOMETRY);
	printTime(TIMER_POPULATING_CORRESPONDENCE_CANDIDATES);
	printTime(TIMER_MEAN_SHIFT_CLUSTERING);
	printTime(TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP);
	printTime(TIMER_PRECOMPUTING_CLOSEST_POINTS);
	printTime(TIMER_GRAPH_CUT_OPTIMIZATION);
	cout << "-------------------------------------------------------------------------------" << endl;
}

void Scene::packageCorrespondences()
{
	clearCorrespondenceList();
	cl = new CorrespondenceList(matchCandidates->package(matchCandidates->sortedMatchListFinal_source,
		matchCandidates->sortedTFListFinal_source, sourceSample, initialRadii, initialTransforms));
	graph_n = sourceSample.size();
}

void Scene::setGroundTruthCorrespondences()
{
	// Assumes point-to-point correspondences given by vertex indices
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating ground truth point correspondences...";

	index_type numVertices = source->numVertices();
	index_type numPick = min(numCorrespondences, cpi.size());

	if (numPick > numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			cl->add(source->vertex[i], target->vertex[i], i, i);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(source->vertex, numPick);
		for (index_type i = 0; i < numPick; i++) {
			cl->add(source->vertex[pickList[i]], target->vertex[pickList[i]], pickList[i], pickList[i]);
		}
	}

	cout << "done!" << endl;
}

void Scene::setClosestPointCorrespondences()
{
	// Assumes point-to-point correspondences given by closest points
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating closest point correspondences...";

	const PointList &sourceVertex = source->vertex;
	const PointList &targetVertex = target->vertex;
	index_type numVertices = source->numVertices();
	index_type numPick = numCorrespondences;
	idxList nearest;
	distList dists;

	if (numPick >= numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			real3_type p = sourceVertex[i];

			// Find closest point in the target surface
			targetVertex.queryKNN(p, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(p, targetVertex[nearest[0]], i, nearest[0]);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(sourceVertex, numPick);

		for (index_type i = 0; i < numPick; i++) {
			real3_type p = sourceVertex[pickList[i]];

			// Find closest point in the target surface
			targetVertex.queryKNN(p, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(p, targetVertex[nearest[0]], pickList[i], nearest[0]);
		}
	}

	cout << "done!" << endl;
}

void Scene::clear()
{
	clearAnimation();
	clearDeform();
}

void Scene::clearAnimation()
{
	if (animation) {
		delete animation;
		animation = NULL;
		target = source = NULL;
	}
}

void Scene::createNewAnimation() {
	clearAnimation();
	animation = new ShapeAnimation();
}

void Scene::loadNewAnimation( const FileSequence &fs )
{
	createNewAnimation();
	animation->load(QString(rootDir.c_str()), fs, *this);
}

void Scene::addAnimationFrame( QString filename, int processFrames )
{
	animation->add(QString(rootDir.c_str()), filename, *this, processFrames);
}

bool Scene::is_deformation_set_up() const
{
	if (deform) {
		return true;
	}
	return false;
}

bool Scene::is_deformation_set_up( DeformTypes deformType ) const
{
	if (deform && (deform->getType() == deformType)) {
		return true;
	}
	return false;
}

void Scene::clearDeform()
{
	// if there's a dangling solver thread
	if (t) {
		cout << "Warning in Scene::clearDeform(): dangling solver thread!  Please finish its execution.";
		t->wait();
		delete t;
		t = NULL;
	}

	if (deform) {
		// Detach from being notified by this deformation
		deform->Detach(this);

		// Delete the object
		delete deform;
		deform = NULL;
	}

	clearCorrespondenceList();
}

void Scene::solveDeformation()
{
	if (!deform) {
		cerr << "Error in solveDeformation(): Nothing to solve!" << endl;
	}
	else {
		// Check if there's already a thread
		if (t && !deform->finished()) {
			return;
		}

		// Create temporary variable to return solver options
		SolveOptions solveOptions;
		solveOptions.maxItr = solverMaxIterations;
		solveOptions.solveIK = solveIK_DeformAll;

		// Create new solver thread object
		t = new Deform::SolveThread(deform, solveOptions);

		cerr << "*** Solver thread " << t << " created!" << endl;

		t->start();
	}
}

void Scene::controlSolveThread( int signal, int type )
{
	//if (type == DEFORM_EM) {
	//	if (poser) {
	//		setSolveStepType(DEFORM_EM_INTERACTIVE);
	//	}
	//	else {
	//		setSolveStepType(DEFORM_EM_AUTO_STEP);
	//	}
	//	DeformEM_ControlSolveThread(signal);
	//}
	//else if ((type == DEFORM_ALL) || 
	if (type == DEFORM_GRAPH_ALL) {
		if (signal == SETUP_DEFORM) {
			setupDeform(type);
		}
		else if (signal == SOLVE_DEFORM_AUTO) {
			solveDeformation();
		}
	}
}

void Scene::checkExit()
{
	if (deform->finished()) {
		// Delete the solver thread if it's done
		if (t) {
			// Wait till the solver thread exits
			t->wait();

			cerr << "-------------------------------------------------------------------------------" << endl;
			cerr << "*** Solver thread " << t << " finished!" << endl;

			// Delete the thread
			delete t;
			t = NULL;
		}

		//// Check the final error between the deformed and target meshes
		//real_type finalError = meshNorm(*target, *(deform->deformationResult()));
		//cout << "Final Error between target and deformed meshes: " << finalError << endl;

		//// If we have embedded deformation, save the previous solution
		//if (deformType == DEFORM_EMBEDDED) {
		//	lastSolution = ((Deform::EmbeddedDeform*)deform)->getDeformation();
		//	cout << "Saved last solution of " << lastSolution.size() << "elements." << endl;
		//	cout << "First rotation: " << endl << lastSolution[0].rot << endl;
		//}

		//printTime(TIMER_"DeformEM Setup");
		//printTime(TIMER_"DeformEM RANSAC Correspondences");
		//printTime(TIMER_"DeformEM E-step");
		//printTime(TIMER_"DeformEM M-step");
		//printTime(TIMER_"DeformEM Registration");
		//if (solveFinalNNLS) {
		//	printTime(TIMER_"DeformEM Final NNLS Update");
		//}
		//if (solveInitialRigid) {
		//	cout << "****** RANSAC Total Candidates: " << MEASURE_PERF_MEMBER(TIMER_RANSAC_TOTAL_CANDIDATES,iterations) << endl;
		//	cout << "****** RANSAC Total Outliers: " << MEASURE_PERF_MEMBER(TIMER_RANSAC_TOTAL_OUTLIERS,iterations) << endl;
		//	double percentage = MEASURE_PERF_MEMBER(TIMER_RANSAC_TOTAL_OUTLIERS,iterations) / MEASURE_PERF_MEMBER(TIMER_RANSAC_TOTAL_CANDIDATES,iterations) * 100.0;
		//	cout << "****** RANSAC Outlier percentage: " << percentage << "%" << endl;
		//}

		//if ((autoProcessMode == AUTO_PROCESS_DEFORMATION) && !autoSaveDeformEMResult.empty()) {
		//	// Save the result
		//	saveDeformation(QString(autoSaveDeformEMResult.c_str()));

		//	//////////////////////////////////////////////////////////////////////////
		//	// Gather timings and statistics
		//	if (!autoSaveDeformEMTimings.empty()) {
		//		
		//		// Save out statistics in text format
		//		DeformEM &dem = *((DeformEM*)deform);
		//		ofstream out;
		//		out.open(autoSaveDeformEMTimings.c_str(), ios_base::out);

		//		if (out.fail()) {
		//			cerr << "Error in registration timings: could not open output file " << autoSaveDeformEMTimings << " for writing!" << endl;
		//		}
		//		else {
		//			out << "# Vertices (source)" << "\t" << source->numVertices() << "\n";
		//			out << "# Vertices (target)" << "\t" << target->numVertices() << "\n";
		//			out << "# Correspondences" << "\t" << cl->size() << "\n";
		//			if (solveInitialRigid) {
		//				out << "RANSAC Total Candidates" << "\t" << MEASURE_PERF_MEMBER(TIMER_"RANSAC Total Candidates",iterations) << "\n";
		//				out << "RANSAC Total Outliers" << "\t" << MEASURE_PERF_MEMBER(TIMER_"RANSAC Total Outliers",iterations) << "\n";
		//				double percentage = MEASURE_PERF_MEMBER(TIMER_"RANSAC Total Outliers",iterations) / MEASURE_PERF_MEMBER(TIMER_"RANSAC Total Candidates",iterations) * 100.0;
		//				out << "RANSAC Outlier percentage" << "\t" << percentage << "%\n";
		//			}
		//			dem.writeStats(out);

		//			out << "Setup Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM Setup",GetDurationInSecs) << "\n";
		//			out << "RANSAC Correspondence Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM RANSAC Correspondences",GetDurationInSecs) << "\n";
		//			out << "E-Step Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM E-step",GetDurationInSecs) << "\n";
		//			out << "E-Step Iterations" << "\t" << MEASURE_PERF_MEMBER(TIMER_"DeformEM E-step",iterations) << "\n";
		//			out << "M-Step Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM M-step",GetDurationInSecs) << "\n";
		//			out << "M-Step Iterations" << "\t" << MEASURE_PERF_MEMBER(TIMER_"DeformEM M-step",iterations) << "\n";
		//			out << "Total Registration Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM Registration",GetDurationInSecs) << "\n";
		//			if (solveFinalNNLS) {
		//				out << "Final NNLS Weight Time (secs)" << "\t" << MEASURE_PERF_NARG(TIMER_"DeformEM Final NNLS Update",GetDurationInSecs) << "\n";
		//			}

		//			out.close();
		//		}
		//	}

		//	// Exit
		//	exit(0);
		//}

		//// This temp trigger might cause the program to exit.
		//if ((autoProcessMode == AUTO_SIMULTANEOUS_ALIGN) && true) {
		//	tempTrigger = true;
		//}
		//if (autoProcessMode == AUTO_SIMULTANEOUS_ALIGN_GRAPH) {
		//	// Take a screenshot
		//	triggerFinalShot = true;
		//}
	}
}


ShapeAnimation* Scene::getAnimation()
{
	return animation;
}

void Scene::drawDeformation( const Deform::DrawOptions& options ) const
{
	//if (!deform) {
	//	// Draw the sampled points
	//	glPointSize(5.f);
	//	sourceSample.draw(POINT_SQUARES);
	//	return;
	//}

	// Draw the deformed result and related visualizations
	if (deform) {
		deform->draw(options);
	}

	//switch(deformType) {
	//case DEFORM_EMBEDDED:
	//	((EmbeddedDeform*)deform)->draw(showSourceModel, showGraph, renderTransparent);
	//	break;
	//case DEFORM_WEIGHTED:
	//	((WeightedDeform*)deform)->draw(showSourceModel, showGraph, showEffRad, renderTransparent);
	//	break;
	//case DEFORM_EM:
	//	((DeformEM*)deform)->draw(showSourceModel, showGraph, showEffRad, renderTransparent, currentBone);
	//	break;
	//default:
	//	break;
	//}
}

//void Scene::drawAccumulatedGeometry( Deform::DrawOptions *options ) const
//{
//	if (deform && (deform->getType() == DEFORM_EM)) {
//		((DeformEM*)deform)->drawAccumulatedGeometry(*options);
//	}
//	delete options;
//}

void Scene::Update( Subject* theChangedSubject )
{
	// Notify the observers of the scene
	Notify();
}


void Scene::resampleSource()
{
	if (source) {
		source->sampleVertices(graph_n, sourceSample);
	}
}

void Scene::setXMLParamList( xmlParam &p )
{
	// Old deformation parameters
	p.addParam("Graph_Nodes", "int", (void*)&graph_n);
	p.addParam("Graph_Neighbors", "int", (void*)&graph_k);

	// DeformEM parameters
	p.addParam("solverMaxIterations", "int", (void*)&solverMaxIterations);
	p.addParam("estepMaxIterations", "int", (void*)&estepMaxIterations);
	p.addParam("projMaxIterations", "int", (void*)&projMaxIterations);
	p.addParam("IRLSMaxIterations", "int", (void*)&IRLSMaxIterations);
	p.addParam("DeformEM_EWeights", "floatArray", (void*)&em_EWeights);
	p.addParam("DeformEM_MultiframeEWeights", "floatArray", (void*)&em_multiframeEWeights);
	p.addParam("DeformEM_MWeights", "floatArray", (void*)&em_MWeights);
	p.addParam("DeformEM_DiscreteMWeights", "floatArray", (void*)&em_DiscreteMWeights);
	p.addParam("GridDivisions", "int", (void*)&gridDivisions);
	p.addParam("numCorrespondences", "int", (void*)&numCorrespondences);
	p.addParam("bonesPerGridPoint", "int", (void*)&bonesPerGridPoint);
	p.addParam("maxAutoBoneNum", "int", (void*)&maxAutoBoneNum);
	p.addParam("minBoneVertexSupport", "int", (void*)&minBoneVertexSupport);
	p.addParam("numGridCellsExpand", "int", (void*)&numGridCellsExpand);
	p.addParam("solveInitialRigid", "int", (void*)&solveInitialRigid);
	p.addParam("solveFinalNNLS", "bool", (void*)&solveFinalNNLS);
	p.addParam("markCellsWithPoints", "bool", (void*)&markCellsWithPoints);
	p.addParam("useAccumulatedGeometry", "bool", (void*)&useAccumulatedGeometry);
	p.addParam("maxCorrespondenceDistance", "float", (void*)&maxCorrespondenceDistance);
	p.addParam("maxNormalAngleDifference", "float", (void*)&maxNormalAngleDifference);
	p.addParam("boundaryCorrespondenceDistance", "float", (void*)&boundaryCorrespondenceDistance);
	p.addParam("autoSaveDeformEMResult", "string", (void*)&autoSaveDeformEMResult);
	p.addParam("autoSaveDeformEMTimings", "string", (void*)&autoSaveDeformEMTimings);
	p.addParam("saveSolverIKState", "string", (void*)&saveSolverIKState);
	p.addParam("loadSolverIKState", "string", (void*)&loadSolverIKState);
	p.addParam("occludedLabel_ValidCorrespRatioThreshold", "float", (void*)&occludedLabel_ValidCorrespRatioThreshold);

	// DeformAll parameters
	p.addParam("refSampleDistance_Multiplier", "float", (void*)&refSampleDistance_Multiplier);
	//p.addParam("finalRefSampleDistance_Multiplier", "float", (void*)&finalRefSampleDistance_Multiplier);
	p.addParam("splitRegionMinError", "float", (void*)&splitRegionMinError);
	p.addParam("maxJointDistance_Multiplier", "float", (void*)&maxJointDistance_Multiplier);
	p.addParam("hingeJointNumFrames", "int", (void*)&hingeJointNumFrames);
	p.addParam("slidingWindowSize", "int", (void*)&slidingWindowSize);
	p.addParam("loadGridOnly", "int", (void*)&loadGridOnly);
	p.addParam("finalSmoothWeights", "int", (void*)&finalSmoothWeights);
	p.addParam("minMatchRatio", "float", (void*)&minMatchRatio);
	p.addParam("excludeLabelThreshold_percentage", "float", (void*)&excludeLabelThreshold_percentage);
	p.addParam("loadFinalShape", "string", (void*)&loadFinalShape);
	p.addParam("writeFrameToRef", "string", (void*)&writeFrameToRef);
	p.addParam("writeRefToFrame", "string", (void*)&writeRefToFrame);
	p.addParam("saveTaucsErrorLog", "string", (void*)&saveTaucsErrorLog);
	p.addParam("occluded_jointOptWeight", "float", (void*)&occluded_jointOptWeight);
	p.addParam("loadRiggedShape", "string", (void*)&loadRiggedShape);
	p.addParam("loadRiggedWeights", "string", (void*)&loadRiggedWeights);
	p.addParam("loadRiggedBVH", "string", (void*)&loadRiggedBVH);

	// Display parameters
	p.addParam("sourceColor", "floatArray", (void*)&sourceColor);
	p.addParam("targetColor", "floatArray", (void*)&targetColor);
	p.addParam("pointColor", "floatArray", (void*)&pointColor);
	p.addParam("selectedColor", "floatArray", (void*)&selectedColor);

	// Data files
	p.addParam("rootDir", "string", (void*)&rootDir);
	p.addParam("defaultFile", "string", (void*)&defaultFile);
	p.addParam("defaultLoadFrames", "int", (void*)&defaultLoadFrames);
	p.addParam("defaultSkipFrames", "int", (void*)&defaultSkipFrames);
	p.addParam("sourceFrame", "int", (void*)&sourceFrame);
	p.addParam("targetFrame", "int", (void*)&targetFrame);
	p.addParam("autoLoadFrameData", "int", (void*)&autoLoadFrameData);
	p.addParam("meshSequence", "string", (void*)&meshSequence);
	p.addParam("refFrameIndex", "int", (void*)&refFrameIndex);
	p.addParam("solveDiscreteWStep", "bool", (void*)&solveDiscreteWStep);
	p.addParam("solveSimultaneousReg", "bool", (void*)&solveSimultaneousReg);

	p.addParam("correspondenceFile", "string", (void*)&correspondenceFile);
	p.addParam("defaultColormapFile", "string", (void*)&defaultColormapFile);
	p.addParam("defaultSourceFile", "string", (void*)&defaultSourceFile);
	p.addParam("defaultTargetFile", "string", (void*)&defaultTargetFile);
	p.addParam("autoProcessMode", "int", (void*)&autoProcessMode);
	p.addParam("autoSaveCorrespondenceResult", "string", (void*)&autoSaveCorrespondenceResult);
	p.addParam("autoSaveSolverState", "string", (void*)&autoSaveSolverState);
	p.addParam("autoLoadSolverState", "string", (void*)&autoLoadSolverState);
	p.addParam("autoSaveSampleSpinImages", "string", (void*)&autoSaveSampleSpinImages);
	p.addParam("autoLoadSampleSpinImages", "string", (void*)&autoLoadSampleSpinImages);
	p.addParam("autoLoadGCMatchResult", "string", (void*)&autoLoadGCMatchResult);
	p.addParam("initGCMatchRef", "bool", (void*)&initGCMatchRef);
	p.addParam("autoLoadCameraParams", "string", (void*)&autoLoadCameraParams);
	p.addParam("autoSaveShot", "string", (void*)&autoSaveShot);
	p.addParam("accumulateASGOnReference", "bool", (void*)&accumulateASGOnReference);
	p.addParam("disableJointConstraint", "bool", (void*)&disableJointConstraint);
	p.addParam("disableHingeJoints", "bool", (void*)&disableHingeJoints);
	p.addParam("computeTotalEfit", "bool", (void*)&computeTotalEfit);

	// Correspondence parameters
	p.addParam("correspondenceType", "int", (void*)&correspondenceType);
	p.addParam("scaleFactors_MC", "floatArray", (void*)&scaleFactors_MC);
	p.addParam("ANNEpsilon_MC", "float", (void*)&ANNEpsilon_MC);
	p.addParam("colormapMin", "float", (void*)&colormapMin);
	p.addParam("colormapMax", "float", (void*)&colormapMax);
	p.addParam("featureSamplesSource", "int", (void*)&featureSamplesSource);
	p.addParam("featureSamplesTarget", "int", (void*)&featureSamplesTarget);
	p.addParam("featureType", "int", (void*)&featureType);
	p.addParam("ANNEpsilon_FM", "float", (void*)&ANNEpsilon_FM);
	p.addParam("radius_FM", "float", (void*)&radius_FM);
	p.addParam("maxCandidates_FM", "int", (void*)&maxCandidates_FM);
	p.addParam("threshold_FM", "float", (void*)&threshold_FM);
	p.addParam("ANNEpsilon_MS", "float", (void*)&ANNEpsilon_MS);
	p.addParam("radius_MS", "float", (void*)&radius_MS);
	p.addParam("maxIter_MS", "float", (void*)&maxIter_MS);
	p.addParam("rate_MS", "float", (void*)&rate_MS);
	p.addParam("scaleFactors_PC", "floatArray", (void*)&scaleFactors_PC);
	p.addParam("robust_PC", "bool", (void*)&robust_PC);
	p.addParam("minVerifyClusterSize", "int", (void*)&minVerifyClusterSize);
	p.addParam("maxDist_FM", "float", (void*)&maxDist_FM);
	p.addParam("ANNEpsilon_VR", "float", (void*)&ANNEpsilon_VR);
	
	p.addParam("depthScan_gridSpacing", "float", (void*)&depthScan_gridSpacing);

	p.addParam("RANSAC_max_itr", "int", (void*)&RANSAC_max_itr);
	p.addParam("RANSAC_sourceModel_ANNEpsilon", "float", (void*)&RANSAC_sourceModel_ANNEpsilon);
	p.addParam("RANSAC_targetModel_ANNEpsilon", "float", (void*)&RANSAC_targetModel_ANNEpsilon);
	p.addParam("RANSAC_inlier_queryRadius", "float", (void*)&RANSAC_inlier_queryRadius);
	p.addParam("RANSAC_inlier_epsilonBand", "float", (void*)&RANSAC_inlier_epsilonBand);
	p.addParam("RANSAC_minimumInliersPercentage", "float", (void*)&RANSAC_minimumInliersPercentage);

	p.addParam("pruneMaxRatio_FM", "float", (void*)&pruneMaxRatio_FM);
	p.addParam("minMatchSize_VR", "int", (void*)&minMatchSize_VR);
	p.addParam("maxNumVerifyClusters_VR", "int", (void*)&maxNumVerifyClusters_VR);
	p.addParam("skipVerification_VR", "bool", (void*)&skipVerification_VR);
	p.addParam("maxNumLabels_GC", "int", (void*)&maxNumLabels_GC);
	p.addParam("growRegionDistThreshold_VR", "float", (void*)&growRegionDistThreshold_VR);
	p.addParam("graphCutSolutionType_GC", "int", (void*)&graphCutSolutionType_GC);
	p.addParam("useDummyLabel_GC", "bool", (void*)&useDummyLabel_GC);
	p.addParam("bdySmoothnessKNN_GC", "int", (void*)&bdySmoothnessKNN_GC);
	p.addParam("closestPtDistThreshold_GC", "float", (void*)&closestPtDistThreshold_GC);
	p.addParam("mismatchError_GC", "float", (void*)&mismatchError_GC);
	p.addParam("dummyLabelSmoothError_GC", "float", (void*)&dummyLabelSmoothError_GC);
	p.addParam("dummyLabelDataError_GC", "float", (void*)&dummyLabelDataError_GC);
	p.addParam("smoothWeight_GC", "float", (void*)&smoothWeight_GC);
	p.addParam("dataWeight_GC", "float", (void*)&dataWeight_GC);
	p.addParam("smoothTruncateValue_GC", "float", (void*)&smoothTruncateValue_GC);
	p.addParam("dataTruncateValue_GC", "float", (void*)&dataTruncateValue_GC);
	p.addParam("dataTermType_GC", "int", (void*)&dataTermType_GC);
	p.addParam("graphCutInstance_GC", "int", (void*)&graphCutInstance_GC);

	// Spin Image Parameters
	p.addParam("binSizeMultiple_SI", "float", (void*)&binSizeMultiple_SI);
	p.addParam("imageWidth_SI", "int", (void*)&imageWidth_SI);
	p.addParam("maxSupportAngle_SI", "float", (void*)&maxSupportAngle_SI);

	// Compute alignment distance
	p.addParam("alignDistanceFile", "string", (void*)&alignDistanceFile);

	// Ground truth shape, weights, and motion parameters
	p.addParam("groundTruth_shapeFile", "string", (void*)&groundTruth_shapeFile);
	p.addParam("groundTruth_weightFile", "string", (void*)&groundTruth_weightFile);
	p.addParam("groundTruth_motionFile", "string", (void*)&groundTruth_motionFile);

	// Automatic window positioning
	p.addParam("screenType", "int", (void*)&screenType);
}

RealColor Scene::getColormapColor( real_type val ) const
{
	return colormap.getColor((val - colormapMin) / (colormapMax - colormapMin));
}

void Scene::regenerateColors()
{
	for (index_type l = 0; l < clusterColors.size(); l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

void Scene::generateClusterColors( size_type n )
{
	// Pick colors from the colormap for each label
	clusterColors.resize(n);

	for (index_type l = 0; l < n; l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

RealColor Scene::getClusterColor( index_type i )
{
	return clusterColors[i];
}

size_type Scene::getClusterSize( index_type i )
{
	return matchCandidates->clusterSizes[i];
}

void Scene::initMatchCandidates( ShapeModel &sm, ShapeModel &tm, bool forceReset/* = false*/ )
{
	// Re-init only if necessary
	if (forceReset || !matchCandidates ||
		(&sm != &matchCandidates->sourceModel) ||
		(&tm != &matchCandidates->targetModel)) {

		// Initialize structure
		if (matchCandidates) {
			delete matchCandidates;
		}
		matchCandidates = new MatchCandidates(sm, tm);
		cout << "Creating new MatchCandidates object..." << endl;
	}
}

bool Scene::clusterResultExists() const
{
	if (!matchCandidates || (!matchCandidates->doneClustering() && !matchCandidates->doneVerify())) {
		return false;
	}

	return true;
}

real_type Scene::makeSceneRadius( real_type percent ) const
{
	return pointSize() * percent;
}

real_type Scene::makeSceneRadiusSq( real_type percent ) const
{
	real_type radius = makeSceneRadius(percent);
	return (radius * radius);
}

distList Scene::makeSceneSqRadii( const FloatArray &percentList ) const
{
	// Generate scales as a percentage of the bounding box diagonal
	distList result;
	result.resize(percentList.size());

	for (int i = 0; i < percentList.size(); i++) {
		result[i] = makeSceneRadiusSq(percentList[i]);
	}

	return result;
}

void Scene::clearCorrespondenceList()
{
	if (cl) {
		delete cl;
		cl = NULL;
	}
}

real_type Scene::pointSize() const
{
	return depthScan_gridSpacing;
}

void Scene::init()
{
	lastDir = QString(rootDir.c_str());

	if (defaultColormapFile.length() > 0) {
		cout << "Opening default colormap: " << defaultColormapFile << endl;
		colormap.load(defaultColormapFile.c_str());
	}

	generateLightColors(graph_n, boneColors);
}

void Scene::setClosestPointCorrespondences_RejectBoundary()
{
	// Assumes point-to-point correspondences given by closest points
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating closest point correspondences (Rejecting Boundary)...";

	index_type numVertices = source->numVertices();
	index_type numPick = min(numCorrespondences, cpi.size());
	idxList nearest;
	distList dists;

	if (numPick > numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			real3_type sourceVertex = source->vertex[i];

			// Find closest point in the target surface
			target->vertex.queryKNN(sourceVertex, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(sourceVertex, target->vertex[nearest[0]], i, nearest[0]);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(source->vertex, numPick);
		for (index_type i = 0; i < numPick; i++) {
			real3_type sourceVertex = source->vertex[pickList[i]];

			// Find closest point in the target surface
			target->vertex.queryKNN(sourceVertex, 1, nearest, dists, true);

			// Check that target vertex is not on boundary
			if (target->markVertex[nearest[0]] == VERTEX_VALID) {
				// Add to the list of correspondences
				cl->add(sourceVertex, target->vertex[nearest[0]], pickList[i], nearest[0]);
			}
		}
	}

	cout << "done!" << endl;
}

void Scene::updateInteractiveCorresp( CorrespondenceList& updateList )
{
	if (!poser) {
		return;
	}
	poser->updateInteractiveCorresp(updateList);
}

void Scene::registerShapePoserInstance( ShapePoser *shapePoser )
{
	poser = shapePoser;

	// Pass the model to DeformAll and start solving in the current thread
	//if (deform && (deform->getType() == DEFORM_ALL)) {
	//	ShapeModel *def = NULL;
	//	std::vector<bool> *defValid = NULL;
	//	QMutex *defLock = NULL;
	//	((DeformAll*)deform)->setupInteractiveIK((shapePoser->getShape()), this, shapePoser->getWeights(), def, defValid, defLock);

	//	if (def && defValid && defLock) {
	//		shapePoser->setDef(def, defValid, defLock);
	//		solveIK_DeformAll = true;
	//		solveDeformation();
	//	}
	//}
	//else
	if (deform && (deform->getType() == DEFORM_GRAPH_ALL)) {
		// Do something similar to DEFORM_ALL case
		ShapeModel *def = NULL;
		DeformAll_Utility::RefWeightList *defWeights = NULL;
		DeformGraphAll_Utility::JointMap *defJoints = NULL;
		std::vector<bool> *defValid = NULL;
		QMutex *defLock = NULL;
		((DeformGraphAll*)deform)->setupInteractiveIK(def, defJoints, defWeights, defValid, defLock);

		if (def && defWeights && defValid && defLock) {
			shapePoser->setDef(def, defJoints, defWeights, defValid, defLock);
			solveIK_DeformAll = true;
			solveDeformation();
		}
	}
}

void Scene::clearShapePoserInstance()
{
	poser = NULL;
}

QString Scene::getSolverStatus() const
{
	if (!deform) {
		return QString();
	}
	//if (deform->getType() == DEFORM_EM) {
	//	return QString(((DeformEM*)deform)->solverStatus.c_str());
	//}
	//else if (deform->getType() == DEFORM_ALL) {
	//	return ((DeformAll*)deform)->solverStatusString();
	//}
	//else
	if (deform->getType() == DEFORM_GRAPH_ALL) {
		return ((DeformGraphAll*)deform)->solverStatusString();
	}

	return QString();
}

std::pair<real3_type,real3_type> Scene::getSolverDisplayBBox() const
{
	//if (!deform) {
	//	return make_pair(real3_type(),real3_type());
	//}
	//else if (deform->getType() == DEFORM_ALL) {
	//	return ((DeformAll*)deform)->getDisplayBBox();
	//}
	//else
	if (deform->getType() == DEFORM_GRAPH_ALL) {
		return ((DeformGraphAll*)deform)->getDisplayBBox();
	}

	return make_pair(real3_type(),real3_type());
}


void Scene::deformInteract( const real3_type& pos )
{
	if (deform && deform->getType() == DEFORM_GRAPH_ALL) {
		return ((DeformGraphAll*)deform)->interact(pos);
	}
}

ShapeModel* Scene::getFinalShape()
{
	if (deform && deform->getType() == DEFORM_GRAPH_ALL) {
		return ((DeformGraphAll*)deform)->getFinalShape();
	}
	else return NULL;
}

//ShapeModel* Scene::inverseDeform( ShapeModel &shape )
//{
//	return deform->inverseDeformation(shape);
//}

//void Scene::accumulateGeometry()
//{
//	if (deform && (deform->getType() != DEFORM_EM)) {
//		((DeformEM*)deform)->accumulateFrame();
//	}
//}

//void Scene::revertPreviousDeformation()
//{
//	////if (!deform || (deformType != DEFORM_EM) || (!deform->finished())) {
//	////	return;
//	////}
//
//	////// Get the maximum number of deformations to browse
//	////const std::vector<RigidDeform> &storedDeformations = ((DeformEM*)deform)->getSolvedDeformations();
//	////int browseMax = storedDeformations.size() - 1;
//
//	////if (browseMax < 0) {
//	////	cerr << "*** No previous solutions to revert!" << endl;
//	////	return;
//	////}
//	////else {
//	////	// Apply the last deformation
//	////	cerr << "*** Reverting to solution " << browseMax << "." << endl;
//	////	((DeformEM*)deform)->setVertexWeightsFromRawData(((DeformEM*)deform)->getSolvedWeights()[browseMax]);
//	////	((DeformEM*)deform)->applyDeformation(storedDeformations[browseMax]);
//	////}
//}

//void Scene::requestNextFrame()
//{
//	if (deform && (deform->getType() != DEFORM_EM)) {
//		((DeformEM*)deform)->requestNextFrame();
//	}
//}