#ifndef __SCENE_H__
#define __SCENE_H__

#include "subject.h"
#include "ShapeAnimation.h"
#include "deform.h"
#include <vector>
#include <QString>
#include <libXMLParam/xmlparam.h>
#include "Colormap.h"
#include "FindCorrespondence.h"
#include "BoostGraphDef.h"

// Global foreground and background colors for OpenGL viewer windows
extern const char *globalForegroundColor;
extern const char *globalBackgroundColor;

enum AutoProcessModes {
	AUTO_PROCESS_NONE = 0,
	AUTO_PROCESS_CORRESPONDENCE,
	AUTO_READ_CORRESPONDENCE_RESULT,
	AUTO_COMPUTE_ALIGNMENT_DISTANCE,
	AUTO_PROCESS_DEFORMATION,
	AUTO_READ_DEFORMATION_RESULT,
	AUTO_SIMULTANEOUS_ALIGN,
	AUTO_SIMULTANEOUS_ALIGN_GRAPH
};

class TransformViewer;
class QAbstractItemModel;
class DeformEM_GUI_Communicator;
class DeformEMOptions;
class ShapePoser;
namespace Deform {
	namespace DeformAll_Utility {
		class Options;
		class FrameSolution;
	}
}


class Scene : public Subject, public Observer
{
public:
	Scene();
	virtual ~Scene();

	// Initializing the scene
	void init();
	void setDefaultOptions();

	// Clearing the data
	void clear();
	void clearAnimation();
	void clearDeform();

	// Loading and Accessing the Mesh Animation
	void createNewAnimation();
	void loadNewAnimation(const FileSequence &fs);
	void addAnimationFrame( QString filename, int processFrames );
	ShapeAnimation* getAnimation();


	// Accessing, initializing and solving for the deformation
	void setupDeform(int deformType);
	void solveDeformation();
	void drawDeformation( const Deform::DrawOptions& options ) const;
	void controlSolveThread( int signal, int type );
	void checkExit();														/**< Checking if the deformation thread exits */
	QString getSolverStatus() const;
	std::pair<real3_type,real3_type> getSolverDisplayBBox() const;
	bool is_deformation_set_up() const;
	bool is_deformation_set_up( Deform::DeformTypes deformType ) const;
	void setDeformOptions();												/**< Set options for Deformation solver */
	void solutionViewModelChanged();


	// Interactive deformation solver
	void updateInteractiveCorresp( CorrespondenceList& updateList );
	void registerShapePoserInstance( ShapePoser *shapePoser );
	void clearShapePoserInstance();
	const ShapePoser *poser;



	// Simultaneous Deformation (DeformAll) Related Methods
	Deform::DeformAll_Utility::Options getDeformAllOptions();
	//void setupExternalMesh_DeformAll(ShapeModel *shape);



	// Updating the display
	virtual void Update(Subject* theChangedSubject);

	// Sampling the source mesh
	void resampleSource();

	// Deformation parameters
	ShapeModel *source, *target;		/**< Source and target meshes for the deformation */


	// Useful global variables
	QString lastDir;

	// Get a colormap color
	RealColor getColormapColor(real_type val) const;

	// Get the scene point size, based on the input scan's grid spacing
	real_type pointSize() const;

	// Reading deformation parameters from the file
	void setXMLParamList(xmlParam &p);

	// Weights for Error Terms
	FloatArray em_EWeights, em_multiframeEWeights, em_MWeights, em_DiscreteMWeights;

	// Defining Maximum # of Iterations for Solver
	int solverMaxIterations;						/**< Maximum total iterations for the solver */
	int estepMaxIterations;							/**< Maximum number of iterations for solving the transformations */
	int projMaxIterations;							/**< Maximum number of iterations for projecting affine transform to rigid transform */
	int IRLSMaxIterations;							/**< Maximum number of iterations for IRLS solver */

	// Setting up the deformation solver
	int graph_n;									/**< Total number of bones */
	int gridDivisions;								/**< Spacing between grid points */
	size_type numCorrespondences;					/**< Number of correspondences to generate */
	int bonesPerGridPoint;							/**< Maximum number of bones affecting each grid point (bone-influence set size) */
	int maxAutoBoneNum;								/**< Maximum number of bones to use */
	int minBoneVertexSupport;						/**< Minimum number of vertices each bone needs to influence */
	int numGridCellsExpand;							/**< Number of grid cells to expand the influence of each bone */
	int graph_k;									/**< Initializing bone influences: set bone radius to be distance to k^th nearest neighbor */
	int solveInitialRigid;							/**< Initially solve for a single rigid transformation */
	bool solveFinalNNLS;							/**< At the end, solve for final continuous weights */
	bool markCellsWithPoints;						/**< Initially when setting up the grid, mark cells with the mesh vertices, not triangles */
	bool useAccumulatedGeometry;					/**< Gradually accumulate geometry in the reference pose after solving each frame */
	float maxCorrespondenceDistance;				/**< Maximum distance between correspondences, given by multiple of grid resolution */
	float maxNormalAngleDifference;					/**< Maximum angle between normals of correspondences, given by angle in degrees */
	float boundaryCorrespondenceDistance;			/**< Maximum distance between correspondences, given by multiple of grid resolution, for points on boundary */

	// Deformation solver batch processing
	std::string autoSaveDeformEMResult;				/**< Automatically save the result from DeformEM */
	std::string autoSaveDeformEMTimings;			/**< Automatically save the stats and timings from DeformEM */
	std::string autoSaveSolverState;				/**< Automatically save the solver state (DeformAll) */
	std::string autoLoadSolverState;				/**< Automatically load the solver state (DeformAll) */
	std::string autoSaveSampleSpinImages;			/**< Automatically save sample spin images (DeformAll) */
	std::string autoLoadSampleSpinImages;			/**< Automatically load sample spin images (DeformAll) */

	FloatArray sourceColor, targetColor, pointColor, selectedColor;		/**< Color settings */

	std::string rootDir;							/**< Root data directory */
	std::string defaultFile;						/**< File to open at startup */
	int defaultLoadFrames;							/**< How many frames (at most) to load at startup */
	int defaultSkipFrames;							/**< Skip over frames when loading */
	int sourceFrame, targetFrame;					/**< Source and target mesh frame (animation frame #) */
	int autoLoadFrameData;							/**< Automatically load extra frame data */
	std::string meshSequence;						/**< (For DeformAll) Starting file for mesh sequence */
	int refFrameIndex;								/**< (For DeformAll) Starting index for mesh sequence */
	bool solveDiscreteWStep;						/**< (For DeformAll) Whether to solve for weights */
	bool solveSimultaneousReg;						/**< (For DeformAll) Whether to solve for simultaneous registration */

	std::string defaultSourceFile;					/**< Default source mesh to open at startup */
	std::string defaultTargetFile;					/**< Default target mesh to open at startup */
	int autoProcessMode;							/**< Automatic processing mode */
	std::string autoSaveCorrespondenceResult;		/**< Automatic save location for correspondence result */

	std::string correspondenceFile;					/**< File to open for correspondences */
	std::string defaultColormapFile;				/**< Colormap to use */

	// Display settings
	int currentBone;						/**< Current bone we are displaying */
	Colormap colormap;						/**< Colormap */
	float colormapMin, colormapMax;			/**< Min and max values for assigning colormap values */

	float totalScene_boundingBoxDiagonal;	/**< A measure of scene size */
	float depthScan_gridSpacing;			/**< Grid spacing size (in terms of world space coordinates) for depth scans */

	// Correspondence parameters
	int correspondenceType;					/**< What type of correspondences to generate */
	bool robust_PC;							/**< Whether to generate robust principal curvatures? */
	FloatArray scaleFactors_PC;				/**< Scales for generating the principal curvature frames (percentage of bbox diagonal) */
	FloatArray scaleFactors_MC;				/**< Sizes for the curvature descriptor (percentage of bbox diagonal) */
	float ANNEpsilon_MC;					/**< Range query epsilon for mean curvature estimation */
	int featureSamplesSource;				/**< Number of features to compute in the source mesh */
	int featureSamplesTarget;				/**< Number of features to compute in the target mesh */
	int featureType;						/**< Type of features to compute in each mesh */
	float ANNEpsilon_FM;					/**< Range query epsilon for feature matching */
	float radius_FM;						/**< Range query radius for feature matching */
	int maxCandidates_FM;					/**< Maximum number of candidates for feature matching */
	float threshold_FM;						/**< Minimum norm of a feature used for matching */
	float maxDist_FM;						/**< Maximum Euclidean distance for correspondence pairing (% of bbox diagonal) */
	float pruneMaxRatio_FM;					/**< Maximum ratio of principal curvatures for point pruning */

	float ANNEpsilon_MS;					/**< Range query epsilon for range query in feature space within mean-shift clustering */
	float radius_MS;						/**< Radius of kernel for mean-shift clustering */
	float maxIter_MS;						/**< Maximum # of iterations for mean-shift clustering */
	float rate_MS;							/**< Rate of convergence for mean-shift clustering */

	float ANNEpsilon_VR;					/**< Range query epsilon for region verification */
	int minVerifyClusterSize;					/**< Minimum size of a cluster for it to be matched & verified */
	int minMatchSize_VR;						/**< Minimum size of verified region */
	int maxNumVerifyClusters_VR;				/**< Maximum number of clusters for verification */
	int maxNumLabels_GC;						/**< Maximum number of labels for the graph cut */
	real_type growRegionDistThreshold_VR;		/**< Distance threshold for growing the verified region */
	int graphCutSolutionType_GC;				/**< What solution method to use in graph cut optimization */
	bool useDummyLabel_GC;						/**< Whether to assign a dummy label in the graph cut optimization? */
	int bdySmoothnessKNN_GC;					/**< Connecting k-nearest boundary vertices for smoothness constraints */
	real_type mismatchError_GC;					/**< Penalty for inconsistent labeling */
	real_type dummyLabelSmoothError_GC;			/**< Smoothness penalty for assigning the dummy label */
	real_type dummyLabelDataError_GC;			/**< Data penalty for assigning the dummy label */
	real_type closestPtDistThreshold_GC;		/**< Closest point distance threshold for graph cut */
	real_type smoothWeight_GC;					/**< Weight of smoothness term for graph cut error function */
	real_type dataWeight_GC;					/**< Weight of data term for graph cut error function */
	real_type smoothTruncateValue_GC;			/**< Maximum value of the smooth term for graph cut error function */
	real_type dataTruncateValue_GC;				/**< Maximum value of the data term for graph cut error function */
	int dataTermType_GC;						/**< Type of data term in error function */
	int graphCutInstance_GC;					/**< Graph cut instance setup: symmetric, forward, or backward */

	int RANSAC_max_itr;							/**< RANSAC correspondence: Number of iterations */
	float RANSAC_sourceModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_targetModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_inlier_queryRadius;			/**< RANSAC correspondence: Radius for gathering inlier candidates */
	float RANSAC_inlier_epsilonBand;			/**< RANSAC correspondence: Epsilon-band distance threshold for inliers */
	float RANSAC_minimumInliersPercentage;		/**< RANSAC correspondence: Minimum percentage of inliers for good fit */

	// Spin image parameters
	real_type binSizeMultiple_SI;				/**< Spin Images: Bin Size parameter as a multiple of the mesh resolution */
	int imageWidth_SI;							/**< Spin Images: Image Width */
	real_type maxSupportAngle_SI;				/**< Spin Images: Maximum Support Angle (in degrees) */

	// Skip verification step?
	bool skipVerification_VR;					/**< Skip verification step? */

	// Compute and save alignment distance
	std::string alignDistanceFile;				/**< Text filename to store the resulting distance */

	// Ground truth shape, weights, and motion parameters
	std::string groundTruth_shapeFile;			/**< Filename for ground truth shape */
	std::string groundTruth_weightFile;			/**< Filename for ground truth vertex weights */
	std::string groundTruth_motionFile;			/**< Filename for ground truth motion data (world-space transformations) */

	// Create radius values from the total scene size
	real_type makeSceneRadius( real_type percent ) const;
	real_type makeSceneRadiusSq( real_type percent ) const;
	distList makeSceneSqRadii( const FloatArray &percentList ) const;

	// Store colors for each cluster
	void regenerateColors();
	void generateClusterColors( size_type n );
	RealColor getClusterColor( index_type i );
	size_type getClusterSize( index_type i );
	std::vector<RealColor> clusterColors;

	// Store clustering data
	void initMatchCandidates( ShapeModel &sm, ShapeModel &tm, bool forceReset = false );
	bool clusterResultExists() const;
	MatchCandidates *matchCandidates;

	// Debugging: Visualizing the cluster modes.
	int showClusterMode;

	// Store initial bone locations and influence radii.
	distList initialRadii;
	RigidDeform initialTransforms;

	// Display parameters (automatic window positioning)
	int screenType;							/**< 0 = 1280x1024, 1 = 1900x1200, 3 = 800x600 */

	// Parameters for DeformAll
	float refSampleDistance_Multiplier;				/**< Minimum distance between samples in the reference frame */
	float splitRegionMinError;		/**< Minimum average corresp error needed to split a region to reuse labels*/
	float maxJointDistance_Multiplier;				/**< Maximum distance between an estimated joint location from the grid and from the transformations */
	int hingeJointNumFrames;						/**< Number of frames required to estimate hinge joints */
	int slidingWindowSize;							/**< Number of frames to solve simultaneously in T-step */
	bool loadGridOnly;								/**< Option to load the grid only */
	bool solveIK_DeformAll;							/**< Solve IK? */
	std::string saveSolverIKState;					/**< Save IK state? */
	std::string loadSolverIKState;					/**< Load IK state? */
	float occludedLabel_ValidCorrespRatioThreshold;	/**< Threshold for ratio of correspondences for detecting occlusion */
	float customError;								/**< Custom additional error for dealing with occluded labels in the W-step */
	int finalSmoothWeights;							/**< Perform uniform smoothing on final weights on grid */

	// Solve and package correspondence result
	void solveGraphCutCorrespondences( MatchCandidates *matchCand, const PointList* sampledSourcePoints, const PointList* sampledTargetPoints, const BoostGraph* sampledSourceGraph, const BoostGraph* sampledTargetGraph );
	void packageCorrespondences();

	// Save snapshots of Graph cut match result
	bool triggerGCMatchFinished;
	bool triggerFinalShot;
	std::string autoLoadCameraParams;				/**< Automatic loading for camera parameters */
	std::string autoSaveShot;						/**< Automatic saving of screenshots */

	// Parameters for DeformGraphAll
	ShapeModel* getFinalShape();					/**< Get the final reconstructed shape from DeformGraphAll */
	void deformInteract( const real3_type& pos );	/**< Interact with DeformGraphAll */
	std::string autoLoadGCMatchResult;				/**< Load pre-computed initial registration results */
	bool initGCMatchRef;							/**< For initialization, use graph cut from ref to new frame or not */
	int frameChangeTrigger;							/**< Changing frames in the solver */
	float minMatchRatio;							/**< Minimum ratio of remaining geometry to trigger a match of occluded bones */
	float excludeLabelThreshold_percentage;			/**< Minimum percentage of samples to be a valid region */
	std::string loadFinalShape;						/**< Filename of the final reconstructed mesh */
	std::vector<RealColor> boneColors;				/**< Colors for each bone */
	float finalRefSampleDistance_Multiplier;		/**< Smaller distance to generate final shape */
	bool accumulateASGOnReference;					/**< Whether to use accumulation or simultaneous registration strategy */
	bool disableJointConstraint;					/**< Disable using joint constraints */
	bool disableHingeJoints;						/**< Disable using hinge joints (i.e. use ball joints only) */
	bool computeTotalEfit;							/**< When loading a solution, whether to compute the total Efit value */
	std::string writeFrameToRef;					/**< Transform each frame to the reference and write to this file (*.obj format) */
	std::string writeRefToFrame;					/**< Transform the reference DSG to each frame and write to this file (*.pts format) */
	std::string saveTaucsErrorLog;					/**< In case we get an error solving the linear system, write the error log to this file (*.txt file extension should be omitted) */
	float occluded_jointOptWeight;					/**< Weight for joint constraint when matching reappearing parts. */
	std::string loadRiggedShape;					/**< Filename of a rigged mesh */
	std::string loadRiggedWeights;					/**< Filename of a rigged mesh's vertex weights */
	std::string loadRiggedBVH;						/**< Filename of a rigged mesh's skeleton */


private:
	ShapeAnimation *animation;				/**< A mesh animation sequence */

	Deform::DeformAlgorithm *deform;		/**< Deformation strategy */
	Deform::SolveThread *t;					/**< A separate thread used to solve the deformation */

	// Sampling correspondences
	std::vector<double> cpi;				/**< Random numbers to pick correspondence indices */
	CorrespondenceList *cl;					/**< New correspondence class (with only positions, not any mesh vertex indices) */

	// Sampling the mesh
	PointList sourceSample;
	void setGroundTruthCorrespondences();
	void setClosestPointCorrespondences();
	void setClosestPointCorrespondences_RejectBoundary();
	void findCorrespondences();
	void clearCorrespondenceList();

	// Debugging: save the previous solution
	RigidDeform lastSolution;
};

#endif
