#include <gl/glew.h>
#include "viewer.h"
#include <fstream>
#include "scene.h"
#include "FindCorrespondence.h"
#include "FileSequence.h"
#include "viewerCommon.h"
#include "../Util/util.h"
#include "../Util/openglUtil.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

/*! Returns "pseudo-distance" from (x,y) to ball of radius size.
\arg for a point inside the ball, it is proportional to the euclidean distance to the ball
\arg for a point outside the ball, it is proportional to the inverse of this distance (tends to
zero) on the ball, the function is continuous. */
float projectOnBall(float x, float y)
{
	// If you change the size value, change angle computation in deformedBallQuaternion().
	const float size       = 1.0f;
	const float size2      = size*size;
	const float size_limit = size2*0.5;

	const float d = x*x + y*y;
	return d < size_limit ? sqrt(size2 - d) : size_limit/sqrt(d);
}

QString genShowHideString( bool status, const char* content )
{
	return QString("%1 %2").arg(status?"Showing":"Hiding").arg(content);
}

void setViewerPosition( QWidget& widget, int x, int y, int width, int height )
{
	widget.move(x, y);
	if (width > 0 && height > 0)
	{
		widget.resize(width, height);
	}
}

Viewer::Viewer( Scene* s, const QGLFormat& format, QWidget* parent/*=0*/, const QGLWidget* shareWidget/*=0*/, Qt::WFlags flags/*=0 */ ) :
QGLViewer(format, parent, shareWidget, flags), scene(s), displayFrame(0), sceneChanged(false),
deformEdit(NULL), maxGraphNodes(1000), maxGraphNeighbors(50), maxOverlapFactor(0.1f),
correspViewer(NULL), tempList(NULL), interactionMode_(NONE), sourceFeatures(NULL),
targetFeatures(NULL), displayScale(0), autoLaunchDone(false), maxFeatureSpaceRange(0.03f),
curvDisplayMode(1), sourceQueryIndex(-1), showErrorComponent(0),
pointPickIndex(0), cellPickIndex(0), displayTargettingFrame(0), externalMeshLoaded(false),
maxDisplayFrame(INT_MAX), showInitSrc(false), showInitResult(false),
shapePoser(NULL), transformViewer(NULL), showOpenGLInfoAct(static_cast<QAction*>(NULL))
{
	// Attach to subject
	scene->Attach(this);
	setAcceptDrops(true);
}

Viewer::~Viewer()
{
	// Detach from subject
	scene->Detach(this);
}


void Viewer::drawFrame( const ShapeModel* frame )
{
	// Show samples, if they exist
	if (showSamplesAct->isChecked()) {
		if (scene->matchCandidates) {
			if (scene->matchCandidates->sourceSample && (&scene->matchCandidates->sourceModel == frame)) {
				frame = scene->matchCandidates->sourceSample;
			}
			if (scene->matchCandidates->targetSample && (&scene->matchCandidates->targetModel == frame)) {
				frame = scene->matchCandidates->targetSample;
			}
		}
	}

	// Display the shape
	if (showFeaturesAct->isChecked()) {
		frame->drawFeatureValues(*scene, displayScale);
	}
	else if (showCurvaturesAct->isChecked()) {
		frame->drawCurvatureValues(*scene, displayScale, curvDisplayMode);
	}
	else if (showFramesAct->isChecked()) {
		frame->drawFrames(frame->frames, frame->frameScales, frame->vertex, scene->pointSize(), displayScale, curvDisplayMode);
	}
	else if (showStatusAct->isChecked()) {
		frame->drawWithStatus();
	}
	else if (showCellIndexAct->isChecked()) {
		if (frame->cellIndex.size()) {
			glDisable(GL_LIGHTING);
			frame->drawCellIndex();
			glEnable(GL_LIGHTING);
		}
	}
	else if (showGeometryAct->isChecked()) {
		frame->draw();
	}

	// Draw vertices
	if (showVerticesAct->isChecked()) {
		setOpenGLMaterialColor(scene->pointColor);
		for (index_type i = 0; i < frame->numVertices(); i++) {
			drawSphere(frame->vertex[i], scene->pointSize() / 2.0f);
		}
	}

	// Draw frames
	if (showVertexFramesAct->isChecked()) {
		int compMax = frame->frameScales.size();
		int comp = min(compMax-1, displayScale);

		for (index_type i = 0; i < frame->numVertices(); i++) {
			const CurvInfo &coordFrame = frame->frames[i][comp];
			qglviewer::Vec from = Vec3fToVec(frame->vertex[i]);
			qglviewer::Vec vmin = Vec3fToVec(coordFrame.vmin);
			qglviewer::Vec vmax = Vec3fToVec(coordFrame.vmax);
			qglviewer::Vec vn = Vec3fToVec(coordFrame.n);

			setOpenGLMaterialColor(RealColor(0.6f, 0.6f, 1.0f));
			drawArrow(from, from + scene->pointSize()*vmin, scene->pointSize() / 25.0f);

			setOpenGLMaterialColor(RealColor(1.0f, 0.6f, 0.6f));
			drawArrow(from, from + scene->pointSize()*vmax, scene->pointSize() / 25.0f);

			setOpenGLMaterialColor(scene->pointColor);
			drawArrow(from, from + scene->pointSize()*vn, scene->pointSize() / 25.0f);
		}
	}


	// Display text status
	if (showCurvaturesAct->isChecked() || showFramesAct->isChecked()) {
		setOpenGLColor(foregroundColor().rgba());

		switch(curvDisplayMode) {
			case CURVINFO_NORMALS:
				drawText(10, height() - 24, tr("Showing frame normals"));
				break;
			case CURVINFO_MIN_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame minimum principal curvatures"));
				break;
			case CURVINFO_MAX_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame maximum principal curvatures"));
				break;
			default:
				break;
		}
	}
}


/**
 * Draw the scene.
 */
void Viewer::draw()
{
	// Enable multisampling
	//glEnable(GL_MULTISAMPLE);

	if (!autoLaunchDone) {
		autoLaunch();
	}

	//////////////////////////////////////////////////////////////////////////
	// Store current state
	//////////////////////////////////////////////////////////////////////////
	GLboolean colorMaterial, lighting;
	GLint polygonMode;
	glGetBooleanv(GL_LIGHTING, &lighting);
	glGetBooleanv(GL_COLOR_MATERIAL, &colorMaterial);
	glGetIntegerv(GL_POLYGON_MODE, &polygonMode);

	//////////////////////////////////////////////////////////////////////////
	// Set state
	//////////////////////////////////////////////////////////////////////////
	// Toggle wireframe display
	if (showWireframe()) {
		glDisable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else {
		//glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		//glCullFace(GL_BACK);
	}

	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);
	
	//////////////////////////////////////////////////////////////////////////////
	////// Draw Geometry
	////GLfloat color[4];
	////// Set alpha
	////color[3] = 1.0f;

	////if (scene->getAnimation()) {
	////	if (renderTransparent()) {
	////		glEnable(GL_BLEND);
	////		glDepthMask (GL_FALSE);
	////		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	////		color[3] = 0.7f;
	////	}

	////	if (showAllFramesAct->isChecked()) {
	////		color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;
	////		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

	////		for (int i = 0; i < scene->getAnimation()->numFrames(); i += 10) {
	////			drawFrame((*scene->getAnimation())[i]);
	////		}
	////	}
	////	else {
	////		//if (showTargetModel()) {
	////		//	color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
	////		//	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	////		//	drawFrame(scene->target);
	////		//}
	////		if (showSourceModel()) {
	////			color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
	////			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

	////			drawFrame(scene->source);
	////		}
	////	}

	////	////// Draw deformation result
	////	////if (showDeformedModel() && scene->deformResult()) {
	////	////	if (showDeformedSameColorAct->isChecked()) {
	////	////		color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
	////	////	}
	////	////	else {
	////	////		color[0] = 0.7f;  color[1] = 1.0f;  color[2] = 0.7f;
	////	////	}
	////	////	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	////	////	drawFrame(scene->deformResult());
	////	////}
	////	// Draw deformation result
	////	//scene->drawDeformation(getDeformDrawOptions());


	////	// Draw the bounding box, if necessary
	////	if (showBbox()) {
	////		scene->getAnimation()->drawBox(*scene->target);
	////	}

	////	if (renderTransparent()) {
	////		glDepthMask (GL_TRUE);
	////		glDisable(GL_BLEND);
	////	}
	////}


	// Draw the deformation result
	scene->drawDeformation(getDeformDrawOptions());

	//// Draw solver status on bottom of this window
	//{
	//	QString statusMsg = scene->getSolverStatus();
	//	QFont useFont;
	//	useFont.setBold(true);
	//	useFont.setPixelSize(20);
	//	glDisable(GL_LIGHTING);
	//	glColor3f(0.0f,0.0f,0.0f);
	//	drawText(25, height() - 25, statusMsg, useFont);
	//	glEnable(GL_LIGHTING);
	//}


	//////////////////////////////////////////////////////////////////////////
	// Feature proximity testing
	drawSphereTest();

	// Draw the selected model point
	if (scene->matchCandidates && sourceQueryIndex >= 0) {
		glDisable(GL_LIGHTING);
		setOpenGLColor(foregroundColor().rgba());
		glPointSize(5.f);
		glBegin(GL_POINTS);
		real3_type v = scene->matchCandidates->sourceModel.vertex[sourceQueryIndex];
		glVertex3f(v[0], v[1], v[2]);
		glEnd();
		glEnable(GL_LIGHTING);
	}

	
	// Draw correspondence candidates
	setOpenGLMaterialColor(scene->pointColor);
	PointList_Utility::draw(sourceQuery, POINT_SPHERES, scene->pointSize() * 2.0f);
	setOpenGLMaterialColor(scene->selectedColor);
	PointList_Utility::draw(targetMatch, POINT_SPHERES, scene->pointSize() * 2.0f);

	//////////////////////////////////////////////////////////////////////////
	// Visualize pose clustering result
	if (visualizeClusterAct->isChecked()) {
		visualizeClusterMode();
	}

	//////////////////////////////////////////////////////////////////////////
	// Draw axis at scene center
	if (axisIsDrawn()) {
		qglviewer::Vec center = sceneCenter();
		glPushMatrix();
		glTranslatef(center[0], center[1], center[2]);
		QGLViewer::drawAxis(sceneRadius()); // Or any scale
		glPopMatrix();
	}

	//////////////////////////////////////////////////////////////////////////
	// Restore OpenGL state
	//////////////////////////////////////////////////////////////////////////
	glPolygonMode(GL_FRONT_AND_BACK, polygonMode);

	//if (colorMaterial)glEnable(GL_COLOR_MATERIAL);
	//else glDisable(GL_COLOR_MATERIAL);

	//if (lighting) glEnable(GL_LIGHTING);
	//else glDisable(GL_LIGHTING);
	glDisable(GL_LIGHTING);
}

void Viewer::pointListTest() {
	GLfloat color[4];

	//////////////////////////////////////////////////////////////////////////
	// Unit test for testing capabilities of PointList
	if (!tempList) {
		// Create a random point list
		tempList = new PointList();
		PointList_Utility::generateRandom(*tempList, 1000);
	}
	color[0] = 1.f;  color[1] = 1.f;  color[2] = 0.0f;  color[3] = 1.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	PointList_Utility::draw(*tempList, POINT_SPHERES, 0.005f);

	idxList list;
	distList dist;
	float rad = 1;
	real3_type q(0, 0, 0);
	//tempList->queryRange(q, rad*rad, list, dist, true);
	tempList->queryKNN(q, 100, list, dist, true);
	color[0] = 0.f;  color[1] = 0.f;  color[2] = 1.f;  color[3] = 1.f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	PointList_Utility::draw(*tempList, POINT_SPHERES, list, 0.006f);

	glEnable(GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	color[0] = 0.f;  color[1] = 1.f;  color[2] = 1.f;  color[3] = 0.5f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	drawSphere(q, rad, 15, 15);
	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
}


void Viewer::colormapTest() {
	// Colormap rendering test using spiral
	const float nbSteps = 200.0f;
	glBegin(GL_QUAD_STRIP);
	for (int i=0; i<nbSteps; ++i)
	{
		const RealColor color = scene->colormap.getColor(i/(nbSteps-1));
		const float ratio = i/nbSteps;
		const float angle = 21.0f*ratio;
		const float c = cos(angle);
		const float s = sin(angle);
		const float r1 = 1.0f - 0.8f*ratio;
		const float r2 = 0.8f - 0.8f*ratio;
		const float alt = ratio - 0.5f;
		const float nor = 0.5f;
		const float up = sqrt(1.0f-nor*nor);

		setOpenGLColor(color);
		setOpenGLMaterialColor(color);
		//cout << "Color: " << i/(nbSteps-1) << ", " << color << endl;
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r1*c, alt, r1*s);
		setOpenGLMaterialColor(color);
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r2*c, alt+0.05f, r2*s);
	}
	glEnd();
}

void Viewer::drawSphereTest()
{
	if (scene->source) {
		glDisable(GL_LIGHTING);
		glColor3f(1,1,1);

		// Plot the selected points
		glPointSize(5.f);
		glBegin(GL_POINTS);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p = scene->source->vertex[inRange[i]];
			glVertex3f(p[0],p[1],p[2]);
		}
		glEnd();

		// Plot the selected normals
		glBegin(GL_LINES);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p1 = scene->source->vertex[inRange[i]];
			real3_type p2 = p1 + 5*scene->pointSize()*scene->source->normal[inRange[i]];
			glVertex3f(p1[0],p1[1],p1[2]);
			glVertex3f(p2[0],p2[1],p2[2]);
		}
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// Draw fitted sphere
		setOpenGLMaterialColor(RealColor(0.8f,0.8f,1.0f), 0.5f);
		drawSphere(fitCenter, fitRadius, 30, 30);

		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}

	// Test Sphere Fitting
	if (interactionMode_ == DEFINE_SPHERE) {
		glDisable(GL_LIGHTING);
		setOpenGLColor(foregroundColor().rgba());
		glBegin(GL_LINES);
		glVertex3f(clickStartPos[0],clickStartPos[1],clickStartPos[2]);
		glVertex3f(clickEndPos[0],clickEndPos[1],clickEndPos[2]);
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		setOpenGLMaterialColor(RealColor(0.8f,0.8f,1.0f), 0.5f);
		drawSphere(clickStartPos, (clickEndPos-clickStartPos).norm(), 30, 30);
		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}	
}


void Viewer::animate()
{
	if (sceneChanged) {
		//cerr << "Scene contents updated!" << endl;
		updateGL();

		//// Automatically save a snapshot
		//saveSnapshot();
		//cerr << "***** Saved snapshot!" << endl;

		sceneMutex.lock();
		sceneChanged = false;
		scene->checkExit();

		if (scene->triggerGCMatchFinished) {
			setGraphShowAct->setChecked(!setGraphShowAct->isChecked());
			displayFrame = scene->refFrameIndex;

			//// Capture snapshot of current frame
			//customSaveSnapshotFile = QString("%1_eknn.png").arg(scene->autoSaveCorrespondenceResult.c_str());
			//connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal
			//updateGL();
			//disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal

			// Capture snapshots of registration result
			captureCorrespViewerResult(800, 610);
			exit(0);
			scene->triggerGCMatchFinished = false;
		}

		if (scene->triggerFinalShot) {
			// Capture snapshot of current frame before exiting
			if (scene->autoSaveShot.length())
			{
				string saveFile = scene->autoSaveShot;
				replaceField(saveFile, '#', displayFrame);
				customSaveSnapshotFile = QString(saveFile.c_str());
				connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal
				updateGL();
				disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
			}

			// Exit
			exit(0);

			// In case we comment out exit..
			scene->triggerFinalShot = false;
		}

		sceneMutex.unlock();
	}

	if (scene->frameChangeTrigger >= 0) {
		// Capture snapshot of current frame before changing to the next
		if (scene->autoSaveShot.length())
		{
			string saveFile = scene->autoSaveShot;
			replaceField(saveFile, '#', displayFrame);
			customSaveSnapshotFile = QString(saveFile.c_str());
			connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal
			updateGL();
			disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
		}

		// Set the frame / maximum frame
		displayFrame = scene->frameChangeTrigger;
		maxDisplayFrame = scene->frameChangeTrigger;
		scene->frameChangeTrigger = -1;
	}
}


/**
 * Initialize the viewer.
 */
void Viewer::init()
{
	// Create context menu.
	createContextMenu();

	// Restore previous viewer state.
	restoreStateFromFile();

	/////////////////////////////////////////////////////
	//       Keyboard shortcut customization           //
	//      Changes standard action key bindings       //
	/////////////////////////////////////////////////////

	// Define 'Control+Q' as the new exit shortcut (default was 'Escape')
	setShortcut(EXIT_VIEWER, Qt::CTRL+Qt::Key_Q);

	// Set 'Control+F' as the FPS toggle state key.
	setShortcut(DISPLAY_FPS, Qt::CTRL+Qt::Key_F);

	// Set 'Control+Y' as the camera fly mode toggle key.
	setShortcut(CAMERA_MODE, Qt::CTRL+Qt::Key_Y);

	// Define 'Control+H' as the new help shortcut (default was 'H')
	setShortcut(HELP, Qt::CTRL+Qt::Key_H);

	// Disable some shortcuts
	setShortcut(MOVE_CAMERA_LEFT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_RIGHT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_UP, Qt::NoButton);
	setShortcut(MOVE_CAMERA_DOWN, Qt::NoButton);

	// Disable draw grid toggle shortcut (default was 'G')
	//setShortcut(DRAW_GRID, 0);

	// Add custom key description (see keyPressEvent).
	setKeyDescription(Qt::Key_Space, "Opens Viewer Menu");
	setKeyDescription(Qt::CTRL+Qt::Key_Left, "Shows previous frame in the animation");
	setKeyDescription(Qt::CTRL+Qt::Key_Right, "Shows next frame in the animation");
	setKeyDescription(Qt::KeypadModifier+Qt::Key_Enter, "Generate deformation graph");
	setKeyDescription(Qt::CTRL+Qt::Key_Return, "Generates and solves for the deformation");
	setKeyDescription(Qt::SHIFT+Qt::Key_Return, "Step one iteration in the deformation solver");
	setKeyDescription(Qt::Key_F, "Flip between displaying source and target model");

	//////////////////////////////////////////////////////////////////////////
	// Initialize rendering
	//////////////////////////////////////////////////////////////////////////
	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));
	QDir::setCurrent(tr(scene->rootDir.c_str()));

	//// Opens help window
	//help();

	// Start the animation
	startAnimation();

	// Initialize scene variables
	scene->init();
}

void Viewer::autoLaunch()
{
	autoLaunchDone = true;
	autoProcess();
}

void Viewer::autoProcess()
{
	if (scene->autoProcessMode == AUTO_PROCESS_NONE) {
		return;
	}
	else if (scene->autoProcessMode == AUTO_SIMULTANEOUS_ALIGN_GRAPH) {
		// Use DeformGraphAll
		useDeformGraphAllAct->setChecked(true);

		if (scene->autoLoadCameraParams.length()) {
			cout << "Loading display settings from " << scene->autoLoadCameraParams << endl;
			loadDisplaySettings(QString(scene->autoLoadCameraParams.c_str()));
		}

		// Display correspondences window & correspondences
		CreateCorrespViewer();

		// Display deformation editor window
		//CreateDeformEditor();

		// Automatically position windows
		if (scene->screenType == 1) {
			if (deformEdit) {
				deformEditPosition(15, -5);
			}
			if (correspViewer) {
				correspViewerPosition(530, 230, 800, 600);
			}
			setViewerPosition(static_cast<QGLViewer&>(*this), 125, 10, 1220, 1105);

			// Useful for capturing movies
			//setViewerPosition(static_cast<QGLViewer&>(*this), 530, 230, 800, 600);
		}
		else if (scene->screenType == 0) {
			if (deformEdit) {
				deformEditPosition(-30, -5);
			}
			setViewerPosition(static_cast<QGLViewer&>(*this), 125, 0, 680, 960);
		}
		else if (scene->screenType == 2) {
			setViewerPosition(static_cast<QGLViewer&>(*this), 125, 0, 320, 240);
		}
		else if (scene->screenType == 3) {
			setViewerPosition(static_cast<QGLViewer&>(*this), 125, 230, 800, 600);
		}

		// Automatically setup the deformation
		DeformEditorChanged(SETUP_DEFORM, 1.0f);

		// Start solver
		DeformEditorChanged(SOLVE_DEFORM_AUTO, 1.0f);

		if (scene->autoLoadCameraParams.length()) {
			cout << "Loading display settings from " << scene->autoLoadCameraParams << endl;
			loadDisplaySettings(QString(scene->autoLoadCameraParams.c_str()));
		}
	}
}

QString Viewer::helpString() const
{
	QString text("<h2>S i m p l e V i e w e r</h2>");
	text += "Use the mouse to move the camera around the object. ";
	text += "You can respectively revolve around, zoom and translate with the three mouse buttons. ";
	text += "Left and middle buttons pressed together rotate around the camera view direction axis<br><br>";
	text += "Pressing <b>Alt</b> and one of the function keys (<b>F1</b>..<b>F12</b>) defines a camera keyFrame. ";
	text += "Simply press the function key again to restore it. Several keyFrames define a ";
	text += "camera path. Paths are saved when you quit the application and restored at next start.<br><br>";
	text += "Press <b>F</b> to display the frame rate, <b>A</b> for the world axis, ";
	text += "<b>Alt+Return</b> for full screen mode and <b>Control+S</b> to save a snapshot. ";
	text += "See the <b>Keyboard</b> tab in this window for a complete shortcut list.<br><br>";
	text += "Double clicks automates single click actions: A left button double click aligns the closer axis with the camera (if close enough). ";
	text += "A middle button double click fits the zoom of the camera and the right button re-centers the scene.<br><br>";
	text += "A left button double click while holding right button pressed defines the camera <i>Revolve Around Point</i>. ";
	text += "See the <b>Mouse</b> tab and the documentation web pages for details.<br><br>";
	text += "Press <b>Escape</b> to exit the viewer.";
	return text;
}

void Viewer::add()
{
	// Add a new frame to the animation.
	QStringList files = QFileDialog::getOpenFileNames(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));

	if (files.size() > 0) {
		// Input from the user: process the local frame information?
		QMessageBox::StandardButton reply = QMessageBox::question(this, tr("Process local frames?"),
												tr("Load / compute local curvature frames when adding?"),
												QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
		bool processFrames = true;

		if (reply == QMessageBox::Cancel) {
			return;
		}

		if (reply == QMessageBox::No) {
			processFrames = false;
		}

		QStringList list = files;
		QStringList::Iterator it = list.begin();
		while(it != list.end()) {
			// Add the frame to the animation
			scene->addAnimationFrame(*it, processFrames);
			++it;
		}
	}
}


void Viewer::open(const char *file/* = NULL*/, int maxFrames/* = 0*/, int skipFrames/* = 1*/) {
	QString fileName;
	if (file) {
		fileName = QString(file);
	}
	else {
		fileName = QFileDialog::getOpenFileName(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));
	}
	if (fileName.isEmpty()) {
		return;
	}

	FileSequence fs(fileName);

	if (!fs.valid()) {
		return;
	}

	scene->lastDir = grabPath(fileName);
	int numGoodFrames = 0;
	fs.skipFrame = skipFrames;

	if (fs.hasNumber()) {
		QString newFile = fs[fs.startFrame];
		int frame = fs.startFrame;

		while(QFile::exists(newFile)) {
			//cout << "Good frame " << numGoodFrames << " in file! " << qPrintable(newFile) << endl;
			numGoodFrames++;
			frame += fs.skipFrame;
			newFile = fs[frame];
		}

		fs.endFrame = frame - fs.skipFrame;
	}

	if ((maxFrames > 0) && (numGoodFrames > maxFrames)) {
		// Clamp the number of frames we are loading, if necessary
		fs.endFrame = fs.startFrame + (maxFrames - 1)*fs.skipFrame;
	}

	//////////////////////////////////////////////////////////////////////////
	// Load all the frames

	// Clear all the current data
	scene->clear();

	// Load the new data
	scene->loadNewAnimation(fs);
	displayFrame = 0;

	// Center the view on the new data
	ShapeModel &temp = *(*scene->getAnimation())[displayFrame];
	fitViewToShape(&temp);

	// Set the target mesh
	scene->source = scene->target = &temp;
	setTargetModelShowAct->setChecked(true);
}


void Viewer::clearResult()
{
	scene->clearDeform();
}

void Viewer::setGraphNodes()
{
	// Dialog box to set the number of graph nodes
	bool ok;
	int nodes = QInputDialog::getInteger(this, tr("Set Number of Graph Nodes"),
		tr("Nodes:"), scene->graph_n, 0, 1e5, 1, &ok);

	if (ok) {
		scene->graph_n = nodes;
	}
}

void Viewer::resampleGraph()
{
	scene->resampleSource();
}

bool Viewer::showWireframe() const {
	return setWireframeModeAct->isChecked();
}


bool Viewer::showBbox() {
	return setBboxModeAct->isChecked();
}

bool Viewer::showSourceModel()
{
	return setSourceModelShowAct->isChecked();
}

bool Viewer::showTargetModel()
{
	return setTargetModelShowAct->isChecked();
}

bool Viewer::showDeformedModel()
{
	return setDeformedModelShowAct->isChecked();
}

int Viewer::whichDeform() {
	if (useEDAct->isChecked()) {
		return Deform::DEFORM_EMBEDDED;
	}
	else if (useWDAct->isChecked()) {
		return Deform::DEFORM_WEIGHTED;
	}
	else if (useDeformEMAct->isChecked()) {
		return Deform::DEFORM_EM;
	}
	else if (useDeformAllAct->isChecked()) {
		return Deform::DEFORM_ALL;
	}
	else if (useDeformGraphAllAct->isChecked()) {
		return Deform::DEFORM_GRAPH_ALL;
	}

	return 0;
}


bool Viewer::renderTransparent() const {
	return setTransparencyAct->isChecked();
}

void Viewer::createContextMenu()
{
	// Create menu bar.
	contextMenu = new QMenu(tr("&Context"), this);
	fileMenu = new QMenu(tr("&File"), this);
	displayMenu = new QMenu(tr("&Display"), this);
	toolsMenu = new QMenu(tr("&Tools"), this);
	optionsMenu = new QMenu(tr("&Options"), this);

	// Top-level menus
	contextMenu->addMenu(fileMenu);
	contextMenu->addMenu(displayMenu);
	contextMenu->addMenu(toolsMenu);
	contextMenu->addMenu(optionsMenu);

	//////////////////////////////////////////////////////////////////////////
	// File Menu
	openAct = new QAction(tr("&Open Sequence..."), this);
	fileMenu->addAction(openAct);
	connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

	addAct = new QAction(tr("&Add Additional Frame..."), this);
	fileMenu->addAction(addAct);
	connect(addAct, SIGNAL(triggered()), this, SLOT(add()));

	saveFramesAct = new QAction(tr("&Save Frames..."), this);
	fileMenu->addAction(saveFramesAct);
	connect(saveFramesAct, SIGNAL(triggered()), this, SLOT(saveFrames()));


	//////////////////////////////////////////////////////////////////////////
	// Tools Menu
	simplifyCurrentShapeAct = new QAction(tr("&Simplify Current Frame..."), this);
	toolsMenu->addAction(simplifyCurrentShapeAct);
	connect(simplifyCurrentShapeAct, SIGNAL(triggered()), this, SLOT(simplifyCurrentShape()));

	batchSimplifyAct = new QAction(tr("&Batch Simplify..."), this);
	toolsMenu->addAction(batchSimplifyAct);
	connect(batchSimplifyAct, SIGNAL(triggered()), this, SLOT(batchSimplify()));

	shapeSliceAct = new QAction(tr("&Slice Shape..."), this);
	toolsMenu->addAction(shapeSliceAct);
	connect(shapeSliceAct, SIGNAL(triggered()), this, SLOT(CreateShapeSlicer()));

	assignTransformsAct = new QAction(tr("&Assign Transforms..."), this);
	toolsMenu->addAction(assignTransformsAct);
	connect(assignTransformsAct, SIGNAL(triggered()), this, SLOT(CreateTransformViewer()));

	assignTransformsReverseAct = new QAction(tr("&Assign Transforms (Reverse)..."), this);
	toolsMenu->addAction(assignTransformsReverseAct);
	connect(assignTransformsReverseAct, SIGNAL(triggered()), this, SLOT(CreateTransformViewerReverse()));

	captureVideoAct = new QAction(tr("&Capture Animation Sequence"), this);
	toolsMenu->addAction(captureVideoAct);
	connect(captureVideoAct, SIGNAL(triggered()), this, SLOT(CaptureResultSnapshots_DeformGraphAll()));

	captureVideoAct2 = new QAction(tr("&Capture Animation Sequence 2"), this);
	toolsMenu->addAction(captureVideoAct2);
	connect(captureVideoAct2, SIGNAL(triggered()), this, SLOT(CaptureResultSnapshots_DeformGraphAll_Model()));

	captureVideoAct3 = new QAction(tr("&Capture Animation Sequence 3"), this);
	toolsMenu->addAction(captureVideoAct3);
	connect(captureVideoAct3, SIGNAL(triggered()), this, SLOT(CaptureResultSnapshots_DeformGraphAll_Init()));

	shapePoseAct = new QAction(tr("&Pose Shape..."), this);
	toolsMenu->addAction(shapePoseAct);
	connect(shapePoseAct, SIGNAL(triggered()), this, SLOT(CreateShapePoser()));

	convertBOBJAct = new QAction(tr("&Convert BOBJ to OBJ..."), this);
	toolsMenu->addAction(convertBOBJAct);
	connect(convertBOBJAct, SIGNAL(triggered()), this, SLOT(Convert_BOBJ_To_OBJ()));

	//////////////////////////////////////////////////////////////////////////
	// Display Menu

	saveDisplaySettingsAct = new QAction(tr("&Save Display Settings..."), this);
	displayMenu->addAction(saveDisplaySettingsAct);
	connect(saveDisplaySettingsAct, SIGNAL(triggered()), this, SLOT(saveDisplaySettings()));

	loadDisplaySettingsAct = new QAction(tr("&Load Display Settings..."), this);
	displayMenu->addAction(loadDisplaySettingsAct);
	connect(loadDisplaySettingsAct, SIGNAL(triggered()), this, SLOT(loadDisplaySettings()));

	displayMenu->addSeparator();

	setWireframeModeAct = new QAction(tr("&Wireframe Mode"), this);
	setWireframeModeAct->setCheckable(true);
	displayMenu->addAction(setWireframeModeAct);

	setTransparencyAct = new QAction(tr("&Transparency Mode"), this);
	setTransparencyAct->setCheckable(true);
	displayMenu->addAction(setTransparencyAct);

	fitViewToShapeAct = new QAction(tr("&Fit View to Current Frame..."), this);
	displayMenu->addAction(fitViewToShapeAct);
	connect(fitViewToShapeAct, SIGNAL(triggered()), this, SLOT(fitViewToShape()));

	displayMenu->addSeparator();

	setSourceModelShowAct = new QAction(tr("&Show Source Model"), this);
	setSourceModelShowAct->setCheckable(true);
	displayMenu->addAction(setSourceModelShowAct);
	setSourceModelShowAct->setChecked(true);

	setTargetModelShowAct = new QAction(tr("&Show Target Model"), this);
	setTargetModelShowAct->setCheckable(true);
	displayMenu->addAction(setTargetModelShowAct);

	setDeformedModelShowAct = new QAction(tr("&Show Deformed Model"), this);
	setDeformedModelShowAct->setCheckable(true);
	displayMenu->addAction(setDeformedModelShowAct);

	sourceCorrespShowAct = new QAction(tr("&Show Source Correspondence Points"), this);
	sourceCorrespShowAct->setCheckable(true);
	displayMenu->addAction(sourceCorrespShowAct);

	targetCorrespShowAct = new QAction(tr("&Show Target Correspondence Points"), this);
	targetCorrespShowAct->setCheckable(true);
	displayMenu->addAction(targetCorrespShowAct);

	deformedCorrespShowAct = new QAction(tr("&Show Deformed Correspondence Points"), this);
	deformedCorrespShowAct->setCheckable(true);
	displayMenu->addAction(deformedCorrespShowAct);

	showShapeWeightsAct = new QAction(tr("&Show Weights on Mesh"), this);
	showShapeWeightsAct->setCheckable(true);
	displayMenu->addAction(showShapeWeightsAct);
	showShapeWeightsAct->setChecked(true);
	
	showAllFramesAct = new QAction(tr("&Show All Frames"), this);
	showAllFramesAct->setCheckable(true);
	displayMenu->addAction(showAllFramesAct);

	showDeformedSameColorAct = new QAction(tr("&Deformed Model: Use target color"), this);
	showDeformedSameColorAct->setCheckable(true);
	displayMenu->addAction(showDeformedSameColorAct);

	displayMenu->addSeparator();

	showSamplesAct = new QAction(tr("&Show Samples"), this);
	showSamplesAct->setCheckable(true);
	displayMenu->addAction(showSamplesAct);

	visualizeClusterAct = new QAction(tr("&Visualize Cluster Result"), this);
	visualizeClusterAct->setCheckable(true);
	displayMenu->addAction(visualizeClusterAct);

	showVerticesAct = new QAction(tr("&Show Vertices"), this);
	showVerticesAct->setCheckable(true);
	displayMenu->addAction(showVerticesAct);

	showVertexFramesAct = new QAction(tr("&Show Vertex Frames"), this);
	showVertexFramesAct->setCheckable(true);
	displayMenu->addAction(showVertexFramesAct);

	displayMenu->addSeparator();
	modelDisplayGroup = new QActionGroup(this);

	showGeometryAct = new QAction(tr("&Show Geometry"), this);
	showGeometryAct->setCheckable(true);
	displayMenu->addAction(showGeometryAct);
	modelDisplayGroup->addAction(showGeometryAct);
	showGeometryAct->setChecked(true);

	showStatusAct = new QAction(tr("&Show Vertex Status"), this);
	showStatusAct->setCheckable(true);
	displayMenu->addAction(showStatusAct);
	modelDisplayGroup->addAction(showStatusAct);

	showFeaturesAct = new QAction(tr("&Show Features"), this);
	showFeaturesAct->setCheckable(true);
	displayMenu->addAction(showFeaturesAct);
	modelDisplayGroup->addAction(showFeaturesAct);

	showCurvaturesAct = new QAction(tr("&Show Curvatures"), this);
	showCurvaturesAct->setCheckable(true);
	displayMenu->addAction(showCurvaturesAct);
	modelDisplayGroup->addAction(showCurvaturesAct);

	showFramesAct = new QAction(tr("&Show Frames"), this);
	showFramesAct->setCheckable(true);
	displayMenu->addAction(showFramesAct);
	modelDisplayGroup->addAction(showFramesAct);

	showCellIndexAct = new QAction(tr("&Show Cell Index"), this);
	showCellIndexAct->setCheckable(true);
	displayMenu->addAction(showCellIndexAct);
	modelDisplayGroup->addAction(showCellIndexAct);

	displayMenu->addSeparator();

	setBboxModeAct = new QAction(tr("&Show Bounding Box"), this);
	setBboxModeAct->setCheckable(true);
	displayMenu->addAction(setBboxModeAct);

	setGraphShowAct = new QAction(tr("&Show Graph"), this);
	setGraphShowAct->setCheckable(true);
	displayMenu->addAction(setGraphShowAct);
	setGraphShowAct->setChecked(true);

	showDeformedGraphAct = new QAction(tr("&Show Graph Deformed"), this);
	showDeformedGraphAct->setCheckable(true);
	displayMenu->addAction(showDeformedGraphAct);
	showDeformedGraphAct->setChecked(true);
	
	showGraphFacesAct = new QAction(tr("&Show Graph Faces"), this);
	showGraphFacesAct->setCheckable(true);
	displayMenu->addAction(showGraphFacesAct);
	showGraphFacesAct->setChecked(false);

	setEffRadShowAct = new QAction(tr("&Show Influence Spheres"), this);
	setEffRadShowAct->setCheckable(true);
	displayMenu->addAction(setEffRadShowAct);

	showSampleErrorAct = new QAction(tr("&Show Per-Sample Error"), this);
	showSampleErrorAct->setCheckable(true);
	displayMenu->addAction(showSampleErrorAct);

	showCellErrorAct = new QAction(tr("&Show Grid Cell Error"), this);
	showCellErrorAct->setCheckable(true);
	displayMenu->addAction(showCellErrorAct);

	showSourceDeformedAct = new QAction(tr("&Show Source Shape Deformed"), this);
	showSourceDeformedAct->setCheckable(true);
	displayMenu->addAction(showSourceDeformedAct);

	showInvalidVerticesAct = new QAction(tr("&Show Invalid Vertices"), this);
	showInvalidVerticesAct->setCheckable(true);
	displayMenu->addAction(showInvalidVerticesAct);

	showSampledPointsAct = new QAction(tr("&Show Sampled Points"), this);
	showSampledPointsAct->setCheckable(true);
	displayMenu->addAction(showSampledPointsAct);

	showExternalMeshDeformedAct = new QAction(tr("&Show external mesh deformed..."), this);
	showExternalMeshDeformedAct->setCheckable(true);
	displayMenu->addAction(showExternalMeshDeformedAct);

	showJointsAct = new QAction(tr("&Show Joints"), this);
	showJointsAct->setCheckable(true);
	displayMenu->addAction(showJointsAct);
	showJointsAct->setChecked(true);

	displayMenu->addSeparator();

	changeBackgroundColorAct = new QAction(tr("&Change Background Color..."), this);
	displayMenu->addAction(changeBackgroundColorAct);
	connect(changeBackgroundColorAct, SIGNAL(triggered()), this, SLOT(changeBackgroundColor()));

	changeForegroundColorAct = new QAction(tr("&Change Foreground Color..."), this);
	displayMenu->addAction(changeForegroundColorAct);
	connect(changeForegroundColorAct, SIGNAL(triggered()), this, SLOT(changeForegroundColor()));

	saveColormapAct = new QAction(tr("&Save Colormap..."), this);
	displayMenu->addAction(saveColormapAct);
	connect(saveColormapAct, SIGNAL(triggered()), this, SLOT(saveColormap()));

	loadColormapAct = new QAction(tr("&Load Colormap..."), this);
	displayMenu->addAction(loadColormapAct);
	connect(loadColormapAct, SIGNAL(triggered()), this, SLOT(loadColormap()));

	setColormapLimitsAct = new QAction(tr("&Set Colormap Range..."), this);
	displayMenu->addAction(setColormapLimitsAct);
	connect(setColormapLimitsAct, SIGNAL(triggered()), this, SLOT(setColormapLimits()));


	//////////////////////////////////////////////////////////////////////////
	// Options menu
	clearResultAct = new QAction(tr("&Clear Result"), this);
	optionsMenu->addAction(clearResultAct);
	connect(clearResultAct, SIGNAL(triggered()), this, SLOT(clearResult()));

	setGraphNodesAct = new QAction(tr("&Graph Nodes..."), this);
	optionsMenu->addAction(setGraphNodesAct);
	connect(setGraphNodesAct, SIGNAL(triggered()), this, SLOT(setGraphNodes()));

	resampleGraphAct = new QAction(tr("&Resample Graph"), this);
	optionsMenu->addAction(resampleGraphAct);
	connect(resampleGraphAct, SIGNAL(triggered()), this, SLOT(resampleGraph()));

	showDeformEditorAct = new QAction(tr("&Edit Deformation Parameters..."), this);
	optionsMenu->addAction(showDeformEditorAct);
	connect(showDeformEditorAct, SIGNAL(triggered()), this, SLOT(CreateDeformEditor()));

	showCorrespViewerAct = new QAction(tr("&Show Correspondences..."), this);
	optionsMenu->addAction(showCorrespViewerAct);
	connect(showCorrespViewerAct, SIGNAL(triggered()), this, SLOT(CreateCorrespViewer()));

	showCorrespEditorAct = new QAction(tr("&Change Correspondence Parameters..."), this);
	optionsMenu->addAction(showCorrespEditorAct);
	connect(showCorrespEditorAct, SIGNAL(triggered()), this, SLOT(CreateCorrespEditor()));

	computeAlignDistanceAct = new QAction(tr("&Compute Alignment Distance..."), this);
	optionsMenu->addAction(computeAlignDistanceAct);
	connect(computeAlignDistanceAct, SIGNAL(triggered()), this, SLOT(ComputeAlignDistance()));

	showAccumulatedGeometryViewerAct = new QAction(tr("&Show Accumulated Geometry..."), this);
	optionsMenu->addAction(showAccumulatedGeometryViewerAct);
	connect(showAccumulatedGeometryViewerAct, SIGNAL(triggered()), this, SLOT(CreateAccumulatedGeometryViewer()));

	showDisplayOptionsAct = new QAction(tr("&Show Display Options..."), this);
	optionsMenu->addAction(showDisplayOptionsAct);
	connect(showDisplayOptionsAct, SIGNAL(triggered()), this, SLOT(CreateDisplayOptions()));

	setFeatureThresholdAct = new QAction(tr("&Set Feature L2 norm Threshold..."), this);
	optionsMenu->addAction(setFeatureThresholdAct);
	connect(setFeatureThresholdAct, SIGNAL(triggered()), this, SLOT(setFeatureThreshold()));

	optionsMenu->addSeparator()->setText(tr("&Deformation Type"));

	deformGroup = new QActionGroup(this);

	useEDAct = new QAction(tr("&Embedded Deformation"), this);
	useEDAct->setCheckable(true);
	optionsMenu->addAction(useEDAct);
	deformGroup->addAction(useEDAct);

	useWDAct = new QAction(tr("&Weighted Deformation"), this);
	useWDAct->setCheckable(true);
	optionsMenu->addAction(useWDAct);
	deformGroup->addAction(useWDAct);

	useDeformEMAct = new QAction(tr("&EM Deformation"), this);
	useDeformEMAct->setCheckable(true);
	optionsMenu->addAction(useDeformEMAct);
	deformGroup->addAction(useDeformEMAct);

	useDeformAllAct = new QAction(tr("&Simultaneous Deform"), this);
	useDeformAllAct->setCheckable(true);
	optionsMenu->addAction(useDeformAllAct);
	deformGroup->addAction(useDeformAllAct);

	useDeformGraphAllAct = new QAction(tr("&Simultaneous Deform w/ASG"), this);
	useDeformGraphAllAct->setCheckable(true);
	optionsMenu->addAction(useDeformGraphAllAct);
	deformGroup->addAction(useDeformGraphAllAct);

	showOpenGLInfoAct = std::tr1::shared_ptr<QAction>(new QAction(tr("&Show OpenGL Information"), this));
	optionsMenu->addAction(showOpenGLInfoAct.get());
	connect(showOpenGLInfoAct.get(), SIGNAL(triggered()), this, SLOT(showOpenGLInfo()));

	// Default value
	useDeformGraphAllAct->setChecked(true);
}

/**
 * Rotate the camera frame about an axis with angle degrees, about the revolve-around point.
 */
void Viewer::rotateAbout(qglviewer::Vec up, float angle) {
	// rotate camera about the current up vector
	qglviewer::Vec o, p, v, cup;
	qglviewer::Quaternion delta;

	up.normalize();
	o = camera()->revolveAroundPoint();
	p = camera()->position();
	v = p + camera()->viewDirection();
	cup = p + camera()->upVector();

	delta.setAxisAngle(up, angle/180.0*M_PI);

	p = delta.rotate(p-o) + o;
	v = delta.rotate(v-o) + o - p;
	cup = delta.rotate(cup-o) + o - p;
	
	camera()->setPosition(p);
	camera()->setViewDirection(v);
	camera()->setUpVector(cup);
}

void Viewer::rotateCamera_DeformedBall( float x, float y, int steps)
{
	qglviewer::Camera* camera = this->camera();
	
	// Points on the deformed ball
	float dx = x;
	float dy = -y;

	const qglviewer::Vec p1(0.0, 0.0, projectOnBall(0.0, 0.0));
	const qglviewer::Vec p2(dx, dy, projectOnBall(dx, dy));

	// Approximation of rotation angle
	// Should be divided by the projectOnBall size, but it is 1.0
	const qglviewer::Vec axis = cross(p2,p1);
	//const float angle = 2.0 * asin(sqrt(axis.squaredNorm() / p1.squaredNorm() / p2.squaredNorm()));
	const float angle = 2.0 * M_PI / ((double)steps);
	qglviewer::Quaternion rot = qglviewer::Quaternion(axis, angle);

	//cout << "Rotating: " << rot[0] << "," << rot[1] << "," << rot[2] << "," << rot[3] << endl;

	camera->frame()->rotateAroundPoint(rot, camera->revolveAroundPoint());
}

void Viewer::keyPressEvent(QKeyEvent *e)
{
	// Get event modifiers key
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	bool handled = false;
	if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = this->geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);
		//cout << "(x,y,w,h) = (" << win.x() << "," << win.y() << "," << win.width() << "," << win.height() << ")" << endl;

		contextMenu->exec(center);
		handled = true;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::CTRL)) {
		// increment the frame number
		//if (++displayFrame >= scene->getAnimation()->numFrames()) {
		//	displayFrame = 0;
		//}
		//displayMessage(updateFrame(displayFrame));
		if (displayFrame < maxDisplayFrame) {
			++displayFrame;
		}
		displayMessage(tr("Showing frame %1").arg(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::CTRL)) {
		// decrement the frame number
		//if (--displayFrame < 0) {
		//	displayFrame = scene->getAnimation()->numFrames() - 1;
		//}
		//displayMessage(updateFrame(displayFrame));
		if (displayFrame > 0) {
			--displayFrame;
		}
		displayMessage(tr("Showing frame %1").arg(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Up) && (modifiers==Qt::CTRL)) {
		// increment the targetting frame number
		++displayTargettingFrame;
		displayMessage(tr("Switching target frame %1").arg(displayTargettingFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Down) && (modifiers==Qt::CTRL)) {
		// decrement the targetting frame number
		--displayTargettingFrame;
		displayMessage(tr("Switching target frame %1").arg(displayTargettingFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Return) && (modifiers==Qt::NoButton)) {
		// This used to be the trigger for solving graph cut correspondences.

		//// Solve for the correspondences and display the result
		//scene->initMatchCandidates(*(scene->source), *(scene->target));
		//scene->solveGraphCutCorrespondences();
		//CreateCorrespViewer();
		//correspViewer->DisplayCorrespondenceResult(true);

		////scene->setupDeform(whichDeform());
		////// Display the deformed mesh
		////setDeformedModelShowAct->setChecked(true);
		////displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier)) {
		//scene->graph_n += 10;
		//scene->setupDeform(whichDeform());
		//// Display the deformed mesh
		//setDeformedModelShowAct->setChecked(true);
		//displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));

		scene->currentBone += 1;
		if (scene->currentBone >= scene->graph_n) {
			scene->currentBone = scene->graph_n-1;
		}
		displayMessage(tr("Showing bone %1...").arg(scene->currentBone));


		handled = true;
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier)) {
		//scene->graph_n -= 10;
		//if (scene->graph_n < 10) {
		//	scene->graph_n = 10;
		//}
		//scene->setupDeform(whichDeform());
		//// Display the deformed mesh
		//setDeformedModelShowAct->setChecked(true);
		//displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));

		scene->currentBone -= 1;
		if (scene->currentBone < -1) {
			scene->currentBone = -1;
		}

		if (scene->currentBone < 0) {
			displayMessage(tr("Showing all bones..."));
		}
		else {
			displayMessage(tr("Showing bone %1...").arg(scene->currentBone));
		}

		handled = true;
	}
	else if ((e->key()==Qt::Key_Return) && (modifiers==Qt::SHIFT)) {
		scene->solveDeformation();
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::NoButton)) {
		//rotateAbout(qglviewer::Vec(0, 1, 0), 10.0f);
		rotateCamera_DeformedBall(1.0, 0.0, 20);
		displayMessage(tr("Rotating left"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::NoButton)) {
		//rotateAbout(qglviewer::Vec(0, 1, 0), -10.0f);
		rotateCamera_DeformedBall(-1.0, 0.0, 20);
		displayMessage(tr("Rotating right"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Up) && (modifiers==Qt::NoButton)) {
		//rotateAbout(qglviewer::Vec(1, 0, 0), 10.0f);
		rotateCamera_DeformedBall(0.0, 1.0, 20);
		displayMessage(tr("Rotating up"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Down) && (modifiers==Qt::NoButton)) {
		//rotateAbout(qglviewer::Vec(1, 0, 0), -10.0f);
		rotateCamera_DeformedBall(0.0, -1.0, 20);
		displayMessage(tr("Rotating down"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		// Toggle showing the source model
		setSourceModelShowAct->setChecked(!setSourceModelShowAct->isChecked());
		displayMessage(genShowHideString(setSourceModelShowAct->isChecked(), "source model"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_D) && (modifiers==Qt::NoButton)) {
		// Toggle showing the deformed model
		setDeformedModelShowAct->setChecked(!setDeformedModelShowAct->isChecked());
		displayMessage(genShowHideString(setDeformedModelShowAct->isChecked(), "deformed model"));

		//QString message;
		//// Flip between displaying the target and the deformed meshes
		//if ((showDeformedModel() && showTargetModel()) || (!showDeformedModel() && !showTargetModel()) || (!showDeformedModel() && showTargetModel())) {
		//	setDeformedModelShowAct->setChecked(true);
		//	setTargetModelShowAct->setChecked(false);
		//	message = tr("Showing deformed model");
		//}
		//else {
		//	setDeformedModelShowAct->setChecked(false);
		//	setTargetModelShowAct->setChecked(true);

		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->target);
		//	message = tr("Showing target model (frame %1)").arg(displayFrame);
		//}
		//displayMessage(message);

		handled = true;
	}
	else if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		// Toggle showing the target model
		setTargetModelShowAct->setChecked(!setTargetModelShowAct->isChecked());
		displayMessage(genShowHideString(setTargetModelShowAct->isChecked(), "target model"));

		//QString message;
		//// Flip between displaying the source and the target meshes
		//if ((showSourceModel() && showTargetModel()) || (!showSourceModel() && !showTargetModel()) || (!showSourceModel() && showTargetModel())) {
		//	setSourceModelShowAct->setChecked(true);
		//	setTargetModelShowAct->setChecked(false);

		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->source);
		//	message = tr("Showing source model (frame %1)").arg(displayFrame);
		//}
		//else {
		//	setSourceModelShowAct->setChecked(false);
		//	setTargetModelShowAct->setChecked(true);
		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->target);
		//	message = tr("Showing target model (frame %1)").arg(displayFrame);
		//}
		//displayMessage(message);

		handled = true;
	}
	else if ((e->key()==Qt::Key_W) && (modifiers==Qt::NoButton)) {
		// Toggle showing the source correspondence points
		sourceCorrespShowAct->setChecked(!sourceCorrespShowAct->isChecked());
		displayMessage(genShowHideString(sourceCorrespShowAct->isChecked(), "source correspondences"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_E) && (modifiers==Qt::NoButton)) {
		// Toggle showing the deformed correspondence points
		deformedCorrespShowAct->setChecked(!deformedCorrespShowAct->isChecked());
		displayMessage(genShowHideString(deformedCorrespShowAct->isChecked(), "deformed correspondences"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_R) && (modifiers==Qt::NoButton)) {
		// Toggle showing the target correspondence points
		targetCorrespShowAct->setChecked(!targetCorrespShowAct->isChecked());
		displayMessage(genShowHideString(targetCorrespShowAct->isChecked(), "target correspondences"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_H) && (modifiers==Qt::NoButton)) {
		// Toggle showing the weights painted on the shape
		showShapeWeightsAct->setChecked(!showShapeWeightsAct->isChecked());
		displayMessage(genShowHideString(showShapeWeightsAct->isChecked(), "weights on mesh"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_A) && (modifiers==Qt::NoButton)) {
		// Toggle showing all frames
		showAllFramesAct->setChecked(!showAllFramesAct->isChecked());
		displayMessage(genShowHideString(showAllFramesAct->isChecked(), "all frames"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::NoButton)) {
		displayScale++;
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::NoButton)) {
		displayScale = max(0, displayScale-1);
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::CTRL)) {
		curvDisplayMode++;
		if (curvDisplayMode > 4) {
			curvDisplayMode = 1;
		}
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::CTRL)) {
		curvDisplayMode--;
		if (curvDisplayMode < 1) {
			curvDisplayMode = 4;
		}
	}
	else if ((e->key()==Qt::Key_T) && (modifiers==Qt::CTRL)) {
		// Generate a random point
		real3_type meshMin = scene->target->box_min;
		real3_type meshMax = scene->target->box_max;

		for (int d = 0; d < 3; d++) {
			real_type a = genrand_real1();
			testPoint[d] = (1.0 - a)*meshMin[d] + a*meshMax[d];
		}

		real3_type diffMesh;
		testDMesh = scene->target->pointSqDistance(testPoint, testCMesh);
		diffMesh = testPoint - testCMesh;
		cout << "Squared distance to mesh: " << testDMesh << " v.s. " << diffMesh*diffMesh << endl;

		//// Set the box and triangle
		//testBoxMin[0] = 0.3f; testBoxMax[0] = 0.5f;
		//testBoxMin[1] = 0.4f; testBoxMax[1] = 0.6f;
		//testBoxMin[2] = 0.5f; testBoxMax[2] = 0.7f;
		//testT0[0] = 0.1f; testT1[0] = 0.4f; testT2[0] = 0.0f;
		//testT0[1] = 0.1f; testT1[1] = 0.4f; testT2[1] = 0.7f;
		//testT0[2] = 0.1f; testT1[2] = 0.4f; testT2[2] = 0.8f;

		//real3_type diffBox, diffTri;
		//real_type lbDTri = lowerBoundTriangleSqDistance(testPoint, testT0, testT1, testT2);

		//// Compute distances
		//testDBox = pointBoxSqDistance(testPoint, testBoxMin, testBoxMax, testCBox);
		//testDTri = pointTriangleSqDistance(testPoint, testT0, testT1, testT2, testCTri);

		//diffBox = testPoint - testCBox;
		//diffTri = testPoint - testCTri;

		//cout << "Squared distance to box: " << testDBox << " v.s. " << diffBox*diffBox << endl;
		//cout << "Squared distance to triangle: " << testDTri << " v.s. " << diffTri*diffTri << endl;
		//cout << "Lower-bound distance to triangle: " << lbDTri << endl;

		handled = true;
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::CTRL+Qt::ALT)) {
		//if (scene->matchCandidates) {
		//	MEASURE_PERF_NARG(TIMER_SPIN_IMAGE_STACK,Start);
		//	scene->matchCandidates->sourceModel.computeSpinImages(scene->binSizeMultiple_SI, scene->imageWidth_SI, scene->maxSupportAngle_SI);
		//	MEASURE_PERF_NARG(TIMER_SPIN_IMAGE_STACK,Stop);
		//	printTime(TIMER_SPIN_IMAGE_STACK);
		//}
		// Compute the spin image for the current selected vertex
		if (sourceQueryIndex >= 0) {
			cout << "Computing spin image for vertex " << sourceQueryIndex << endl;

			// Save the spin image to disk for viewing
			//scene->matchCandidates->sourceModel.computeSpinImage(sourceQueryIndex);
			scene->matchCandidates->sourceModel.saveSingleSpinImage("siTemp.data", sourceQueryIndex);
		}
	}
	// Browse solved deformations
	else if ((e->key()==Qt::Key_BracketLeft) && (modifiers==Qt::CTRL)) {
		//scene->browseDeformation(-1);
		showErrorComponent = min(2, showErrorComponent+1);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_BracketRight) && (modifiers==Qt::CTRL)) {
		//scene->browseDeformation(+1);
		showErrorComponent = max(0, showErrorComponent-1);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_G) && (modifiers==Qt::NoButton)) {
		// Toggle showing the graph
		setGraphShowAct->setChecked(!setGraphShowAct->isChecked());
		displayMessage(genShowHideString(setGraphShowAct->isChecked(), "graph structure"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_V) && (modifiers==Qt::NoButton)) {
		// Toggle deforming the graph
		showDeformedGraphAct->setChecked(!showDeformedGraphAct->isChecked());
		displayMessage(genShowHideString(showDeformedGraphAct->isChecked(), "deformed graph"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_B) && (modifiers==Qt::NoButton)) {
		// Toggle showing the graph faces
		showGraphFacesAct->setChecked(!showGraphFacesAct->isChecked());
		displayMessage(genShowHideString(showGraphFacesAct->isChecked(), "graph faces"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::NoButton)) {
		// Generate random bone colors
		generateLightColors(scene->graph_n, scene->boneColors);
		displayMessage(tr("Changing bone colors"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_W) && (modifiers==Qt::CTRL)) {
		// Toggle showing per-sample error
		showSampleErrorAct->setChecked(!showSampleErrorAct->isChecked());
		displayMessage(genShowHideString(showSampleErrorAct->isChecked(), "sample error"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_E) && (modifiers==Qt::CTRL)) {
		// Toggle showing grid cell error
		showCellErrorAct->setChecked(!showCellErrorAct->isChecked());
		displayMessage(genShowHideString(showCellErrorAct->isChecked(), "grid cell error"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier+Qt::CTRL)) {
		++pointPickIndex;
		displayMessage(tr("Showing point %1...").arg(pointPickIndex));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier+Qt::CTRL)) {
		--pointPickIndex;
		displayMessage(tr("Showing point %1...").arg(pointPickIndex));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier+Qt::ALT)) {
		++cellPickIndex;
		displayMessage(tr("Showing cell %1...").arg(cellPickIndex));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier+Qt::ALT)) {
		--cellPickIndex;
		displayMessage(tr("Showing cell %1...").arg(cellPickIndex));
		handled = true;
	}
	else if ((e->key()==Qt::Key_X) && (modifiers==Qt::NoButton)) {
		// Toggle showing deformed source
		showSourceDeformedAct->setChecked(!showSourceDeformedAct->isChecked());
		displayMessage(genShowHideString(showSourceDeformedAct->isChecked(), "deformed source"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_I) && (modifiers==Qt::NoButton)) {
		// Toggle showing deformed source
		showInvalidVerticesAct->setChecked(!showInvalidVerticesAct->isChecked());
		displayMessage(genShowHideString(showInvalidVerticesAct->isChecked(), "invalid vertices"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Y) && (modifiers==Qt::NoButton)) {
		// Toggle showing deformed source
		showSampledPointsAct->setChecked(!showSampledPointsAct->isChecked());
		displayMessage(genShowHideString(showSampledPointsAct->isChecked(), "sampled points"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Z) && (modifiers==Qt::NoButton)) {
		// Load the external mesh if it doesn't exist
		if (!externalMeshLoaded) {
			setupExternalMeshDeformAll();
		}

		// Toggle showing deformed external mesh
		if (externalMeshLoaded) {
			showExternalMeshDeformedAct->setChecked(!showExternalMeshDeformedAct->isChecked());
		}
		displayMessage(genShowHideString(showExternalMeshDeformedAct->isChecked(), "deformed external mesh"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_J) && (modifiers==Qt::NoButton)) {
		// Toggle showing joints
		showJointsAct->setChecked(!showJointsAct->isChecked());
		displayMessage(genShowHideString(showJointsAct->isChecked(), "joints"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_O) && (modifiers==Qt::CTRL)) {
		// Toggle showing initialization source
		showInitSrc = !showInitSrc;
		displayMessage(genShowHideString(showInitSrc, "Initialization Source"));
		handled = true;
	}
	else if ((e->key()==Qt::Key_P) && (modifiers==Qt::CTRL)) {
		// Toggle showing initialization result
		showInitResult = !showInitResult;
		displayMessage(genShowHideString(showInitResult, "Initialization Result"));
		handled = true;
	}


	if (!handled) {
		QGLViewer::keyPressEvent(e);
	}

	emit updateChildDisplay();
}

real3_type Viewer::closestPointOnLine( QPoint pos, real3_type x )
{
	// Find the line in 3D corresponding to the cursor
	qglviewer::Vec o, d;
	real3_type orig, dir;
	camera()->convertClickToLine(pos, o, d);
	orig = VecToVec3f(o);
	dir = VecToVec3f(d);

	// Find closest point on line
	real_type t = - ((orig-x) * dir) / (dir*dir);
	return orig + t*dir;
}

void Viewer::mousePressEvent( QMouseEvent *e )
{
	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		updateGL();

		// Move the selected vertex along the shape
		bool found;
		clickStartPos = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));
		clickEndPos = clickStartPos;

		if (found) {
			//cout << "Setting start position " << clickStartPos << endl;
			interactionMode_ = DEFINE_SPHERE;
		}
		else {
			interactionMode_ = NONE;
		}
	}
	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::CTRL)) {
		updateGL();

		// TODO: Track the mouse stroke
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void Viewer::mouseMoveEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

		//cout << "Move: Setting end position to " << clickEndPos << endl;
		updateGL();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void Viewer::mouseReleaseEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
			clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

			//cout << "Release: Setting end position to " << clickEndPos << endl;

			// Perform sphere fit
			//sphereFit();

			// Find feature space points close in proximity
			featureProximity();

			updateGL();
		}

		interactionMode_ = NONE;
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void Viewer::Update( Subject* theChangedSubject )
{
	// Implement updating of opengl here by setting a variable and looping in an animation loop
	sceneMutex.lock();
	sceneChanged = true;
	sceneMutex.unlock();

	//cerr << "Updated by scene at: " << theChangedSubject << endl;
	//updateGL();
}

QString Viewer::updateFrame( int frame )
{
	QString message;

	// determine the mode
	if (showSourceModel() && !showTargetModel()) {
		scene->source = (*scene->getAnimation())[displayFrame];
		message = tr("Showing source frame %1").arg(displayFrame);
	}
	else if (!showSourceModel() && showTargetModel()) {
		scene->target = (*scene->getAnimation())[displayFrame];
		message = tr("Showing target frame %1").arg(displayFrame);
	}
	else {
		message = tr("View only the source/target frame to update.");
	}

	return message;
}

void Viewer::customSaveSnapshot()
{
	cout << "Saving snapshot to file: " << qPrintable(customSaveSnapshotFile) << endl << flush;
	saveSnapshot(customSaveSnapshotFile);
}


void Viewer::captureVideo()
{
	// Iterate through each frame and take a screenshot
	// Show only the source model
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(true);
	setDeformedModelShowAct->setChecked(false);

	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the animation", scene->lastDir, tr("*.%1").arg(snapshotFormat().toLower()));
	if (savefileName.length() == 0) {
		cout << "Video capture cancelled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	const int numFrames = scene->getAnimation()->numFrames();

	FileSequence seq(savefileName);
	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	for (unsigned int i = 0; i < numFrames; i++) {
		scene->target = (*scene->getAnimation())[i];

		// Update the display and take a screenshot
		customSaveSnapshotFile = seq[i];
		updateGL();
	}
	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal

}

void Viewer::saveColormap()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the colormap...", scene->lastDir, QString("*.txt"));
	if (savefileName.length() == 0) {
		cout << "Viewer::loadColormap(): Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	scene->colormap.save(qPrintable(savefileName));
}

void Viewer::loadColormap()
{
	QString loadfileName;
	loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the colormap...", scene->lastDir, tr("*.txt"));
	if (loadfileName.length() == 0) {
		cout << "Viewer::loadColormap(): Loading canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(loadfileName);
	}

	scene->colormap.load(qPrintable(loadfileName));
}

void Viewer::setColormapLimits()
{
	// Set the range of the colormap
	real_type newMin, newMax;

	// Set the resolution of the screen here.
	bool ok;

	newMin = QInputDialog::getDouble(this, tr("Colormap minimum range"),
				tr("Input Minimum:"), scene->colormapMin, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	newMax = QInputDialog::getDouble(this, tr("Colormap maximum range"),
				tr("Input Maximum:"), scene->colormapMax, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->colormapMin = newMin;
	scene->colormapMax = newMax;
	updateGL();
}

void Viewer::setFeatureThreshold()
{
	// Set the range of the colormap
	real_type newVal;

	// Set the resolution of the screen here.
	bool ok;

	newVal = QInputDialog::getDouble(this, tr("Feature threshold display"),
		tr("Input Threshold:"), scene->threshold_FM, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->threshold_FM = newVal;
	updateGL();
}

void Viewer::saveTransformSpacePoints()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the feature points...", scene->lastDir, QString("*.data"));
	if (savefileName.length() == 0) {
		cout << "Viewer::saveFeatureSpacePoints(): Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	scene->matchCandidates->writeData(qPrintable(savefileName));
}


void Viewer::sphereFit()
{
	if (!scene->source) {
		return;
	}

	PointList rangePts, rangeNormals;
	real3_type sample = clickStartPos;
	real_type radSq = (clickEndPos - clickStartPos).square_norm();

	cout << "Fitting sphere at: " << sample << " with radius^2 " << radSq << endl;

	scene->source->vertex.queryRange(sample, radSq, inRange, distRange, true);

	// Gather the point locations and the normals
	rangePts.resize(inRange.size());
	rangeNormals.resize(inRange.size());
	for (index_type j = 0; j < inRange.size(); j++) {
		// Throw out the point if the normal is zero
		if (scene->source->normal[inRange[j]].norm() >= REAL_TYPE_EPSILON) {
			rangePts[j] = scene->source->vertex[inRange[j]];
			rangeNormals[j] = scene->source->normal[inRange[j]];
		}
	}

	// Fit a circle at this position
	if (FitSphere(rangePts, rangeNormals, sample, radSq)) {
		// Store the mean curvature (inverse of the radius) as the feature
		//features.val[s] = 1.0 / sqrt(radSq);
		cout << "Fit sphere: center " << sample << " radius^2 " << radSq << endl;
		fitCenter = sample;
		fitRadius = sqrt(radSq);
	}
	else {
		// If the fit is degenerate, then the curvature is 0.
		//features.val[s] = 0;
		cout << "Fit plane: dist from origin " << radSq << endl;
	}
}



void Viewer::featureProximity()
{
	//// (Testing)
	//scene->deformInteract(clickStartPos);
	//return;


	idxList inRange;

	cout << "Computing feature proximity... ";

	if (!scene->matchCandidates) {
		cout << "no matchCandidates." << endl;
		return;
	}

	// Find closest vertex in source model
	const PointList& sourceVertex = scene->matchCandidates->sourceModel.vertex;
	const SpinImageStack& sourceSI = scene->matchCandidates->sourceModel.spinImageStack;
	idxList indices;
	distList dists;
	sourceVertex.queryKNN(clickStartPos, 1, indices, dists, true);
	sourceQueryIndex = indices[0];

	cout << "selecting vertex " << sourceQueryIndex << "... ";

	//// Write the query spin image to disk
	//sourceShape.saveSingleSpinImage("J:\\singleSpinImage.data", sourceQueryIndex);

	// Find closest point in target model
	//sourceQuery.resize(1);
	//sourceQuery[0] = sourceShape.vertex[sourceQueryIndex];

	//real_type temp;
	//real3_type closest;
	//scene->matchCandidates->targetModel.hybridClosestPoint(sourceQuery[0], closest, temp);
	//targetMatch.clear();
	//targetMatch.add(closest);

	//return;

	// Match spin image features
	if (scene->featureType == FEATURE_SPIN_IMAGES) {
		const ShapeModel &targetShape = scene->matchCandidates->targetModel;
		sourceQuery.resize(1);
		sourceQuery[0] = sourceVertex[sourceQueryIndex];

		const SpinImage &query = sourceSI[sourceQueryIndex];
		//targetShape.spinImageStack.querySimilar(query, inRange);
		distList scores;
		targetShape.spinImageStack.querySimilar(query, inRange, scores);

		targetMatch.clear();
		cout << "Matched vertices:";
		for (index_type i = 0, i_end = std::min((unsigned int)scene->maxCandidates_FM, inRange.size()); i < i_end; i++) {
			targetMatch.add(targetShape.vertex[inRange[i]]);
			cout << " " << inRange[i];
		}
		cout << endl;

		return;
	}

	//// Match spin image features only within sampled points in source and target
	//if (scene->featureType == FEATURE_SPIN_IMAGES) {
	//	if (!scene->matchCandidates->sourceSample || !scene->matchCandidates->targetSample) {
	//		return;
	//	}

	//	ShapeModel &sourceSample = *scene->matchCandidates->sourceSample;
	//	ShapeModel &targetSample = *scene->matchCandidates->targetSample;

	//	distList distances;
	//	sourceSample.queryKNN(clickStartPos, 1, inRange, distances, true);
	//	sourceQuery.resize(1);
	//	sourceQuery[0] = sourceSample.vertex[inRange[0]];

	//	const SpinImage &query = sourceSample.spinImageStack[inRange[0]];
	//	targetSample.spinImageStack.querySimilar(query, inRange);
	//	//targetSample.spinImageStack.querySimilar(query, inRange, "J:\\scores.txt");

	//	targetMatch.clear();
	//	for (index_type i = 0, i_end = min(scene->maxCandidates_FM, inRange.size()); i < i_end; i++) {
	//		targetMatch.add(targetSample.vertex[inRange[i]]);
	//	}

	//	return;
	//}

	//////////////////////////////////////////////////////////////////////////////
	// Match other features
	distList distances;
	const real_type radSq = scene->radius_FM * scene->radius_FM;	/**< Range query radius in feature space */
	const int maxCandidates = scene->maxCandidates_FM;				/**< Maximum number of candidates extracted */

	if (!scene->matchCandidates || !scene->matchCandidates->sourceSample || !scene->matchCandidates->targetSample) {
		return;
	}
	ShapeModel *src = scene->matchCandidates->sourceSample;
	ShapeModel *tgt = scene->matchCandidates->targetSample;

	// Set ANN Epsilon for feature space query
	tgt->features->setANNEpsilon(scene->ANNEpsilon_FM);

	// Find closest feature point in source sample
	src->queryKNN(clickStartPos, 1, inRange, distances, true);
	sourceQuery.resize(1);
	sourceQuery[0] = src->vertex[inRange[0]];
	data_type queryFeature = (*src->features)[inRange[0]];

	// Query feature space
	tgt->features->queryRange(queryFeature, radSq, inRange, distances, true);
	//tgt->features->queryKNN(queryFeature, 7, inRange, distances, true);

	// Populate result
	size_type n = inRange.size();
	if (maxCandidates > 0) {
		n = std::min(n, (size_type)maxCandidates);
	}
	targetMatch.clear();
	targetMatch.reserve(n);

	const real_type maxPairDist = scene->makeSceneRadius(scene->maxDist_FM);

	for (index_type i = 0; i < n; i++) {
		// Threshold the Euclidean distance
		real3_type targetPoint = tgt->vertex[inRange[i]];
		if ((sourceQuery[0] - targetPoint).norm() < maxPairDist) {
			targetMatch.add(tgt->vertex[inRange[i]]);
		}
	}
}


void Viewer::testFeatureSelect()
{
	// Just sample the source and target meshes
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->subSampleGeometry(scene->featureSamplesSource, scene->featureSamplesTarget, scene->pruneMaxRatio_FM);

	// Test the sampling of smoothness edges on the boundary
	scene->matchCandidates->shapeBoundaryKNNEdges(scene->matchCandidates->sourceModel, scene->bdySmoothnessKNN_GC, scene->matchCandidates->sourceExtraEdges);
	scene->matchCandidates->shapeBoundaryKNNEdges(scene->matchCandidates->targetModel, scene->bdySmoothnessKNN_GC, scene->matchCandidates->targetExtraEdges);

	//// Recompute principal curvatures and update display for the currently displayed mesh
	//ShapeModel *shape = selectCurrentModel();

	//shape->computeFrames(scene->makeSceneSqRadii(scene->scaleFactors_PC), true);
}

void Viewer::testFeatureMatch()
{
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->subSampleGeometry(scene->featureSamplesSource, scene->featureSamplesTarget, scene->pruneMaxRatio_FM);

	if (scene->featureType == FEATURE_SPIN_IMAGES) {
		scene->matchCandidates->spinImageCandidates(scene->maxCandidates_FM);
	}
	else {
		scene->matchCandidates->findCandidateCorrespondence(scene->radius_FM*scene->radius_FM,
															scene->maxCandidates_FM,
															scene->ANNEpsilon_FM);
	}


	//scene->matchCandidates->RANSAC_Candidates(scene->RANSAC_max_itr,
	//										  scene->RANSAC_sourceModel_ANNEpsilon,
	//										  scene->RANSAC_targetModel_ANNEpsilon,
	//										  scene->makeSceneRadiusSq(scene->RANSAC_inlier_queryRadius),
	//										  scene->makeSceneRadiusSq(scene->RANSAC_inlier_epsilonBand),
	//										  scene->RANSAC_minimumInliersPercentage / 100.0f);
}

/**
 * Assumes that point correspondences are already defined between the source and target meshes.
 */
void Viewer::verifyTransformationSpace()
{
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->groundTruthCorrespondence();
}

void Viewer::testFeatureCluster()
{
	// Test the mean-shift clustering
	scene->matchCandidates->meanShiftBasic(
		scene->radius_MS*scene->radius_MS,
		scene->rate_MS,
		scene->maxIter_MS,
		scene->ANNEpsilon_MS
		);

	//scene->matchCandidates->pruneClusterGeometricConsistency(0.25f);
}

void Viewer::testRefineCluster()
{
	//const int n = 54;
	//real_type leftx[] = {1.1083, 1.29, 0.9052, 0.6289, 0.9923, 1.2578, 1.1075, 0.5549, 1.0859, 0.2728, 0.5033, 0.9631, 0.6627, 0.2577, 0.0999, -0.4068, -0.6887, -0.8531, -1.425, -1.2356, -1.3816, -1.0528, -0.7782, -0.6249, -1.0701, 0.1044, -0.3676, 0.9159, 1.2793, 1.0869, 0.5343, 0.716, 1.2469, 0.9465, 0.6702, 0.9357, 0.1059, 0.2211, 0.6901, 0.5399, 0.0983, -0.2399, -0.2339, -0.7978, -1.1265, -1.4125, -1.0712, -1.3632, -1.1988, -0.6496, -0.343, -0.7955, -0.2082, -0.2143};
	//real_type lefty[] = {1.3408, 1.0535, 1.2876, 1.3628, 0.7881, 0.3829, 0.5093, 0.6597, -0.1507, 0.651, 0.0934, -0.3888, -0.136, -0.3828, 0.9476, 0.6691, 0.6928, 0.905, 1.1257, 0.5715, 0.453, 0.0285, -0.0977, 0.1233, -0.3348, -0.6039, -0.3662, 1.4579, 0.8832, 1.0002, 1.1507, 0.8633, 0.0529, 0.3057, 0.3809, -0.0243, 0.3635, 0.0847, -0.5005, -0.3741, -0.1534, 0.9566, 0.3725, 0.4199, 0.8444, 0.9547, 0.3593, 0.1222, -0.09, -0.3425, 0.0997, -0.461, -0.5956, -0.1451};
	//real_type leftz[] = {-1.3836, -1.3241, -0.9513, -0.7652, -0.6462, -0.8966, -0.6831, -0.3109, -0.8118, -0.2333, -0.307, -0.8971, -0.47, -0.4777, -0.2677, -0.2207, -0.2879, -0.4311, -1.2739, -0.6398, -0.8478, -0.5615, -0.4434, -0.2862, -0.8595, -0.6349, -0.4661, -1.1972, -1.0782, -0.8918, -0.5196, -0.4601, -0.961, -0.5339, -0.3478, -0.5983, -0.1925, -0.2294, -0.7689, -0.5553, -0.3147, -0.2615, -0.1862, -0.3207, -0.607, -1.0284, -0.4966, -0.9127, -0.7695, -0.5333, -0.2189, -0.7414, -0.6291, -0.3089};
	//real_type rightx[] = {1.2175, 1.2861, 1.0737, 0.886, 1.0232, 1.1259, 1.0497, 0.6743, 0.8796, 0.447, 0.5, 0.7139, 0.5615, 0.1684, 0.3804, -0.0878, -0.319, -0.4024, -0.8895, -0.8149, -0.9846, -0.8178, -0.6234, -0.4169, -0.9628, -0.038, -0.3675, 1.1113, 1.2485, 1.1423, 0.7669, 0.8355, 1.0409, 0.8885, 0.7008, 0.8034, 0.2462, 0.2727, 0.4719, 0.3957, 0.1069, 0.113, -0.0212, -0.4835, -0.6503, -0.8939, -0.7315, -1.071, -0.9876, -0.5986, -0.1858, -0.7684, -0.3059, -0.1611};
	//real_type righty[] = {0.3913, 0.1783, 0.6062, 0.8012, 0.3753, -0.0696, 0.1516, 0.5416, -0.3483, 0.6347, 0.1872, -0.5291, -0.0867, -0.1744, 0.8477, 0.7994, 0.8486, 0.9523, 0.8502, 0.7284, 0.5827, 0.3754, 0.2823, 0.4719, -0.009, -0.364, -0.0057, 0.6052, 0.1793, 0.3932, 0.7833, 0.5703, -0.3196, 0.1228, 0.3178, -0.1271, 0.5041, 0.2804, -0.4887, -0.2675, 0.0996, 0.93, 0.5865, 0.6848, 0.8922, 0.8411, 0.6247, 0.3334, 0.2297, 0.0435, 0.4227, -0.1021, -0.2796, 0.1839};
	//real_type rightz[] = {-1.8853, -1.7139, -1.57, -1.489, -1.1461, -1.1272, -1.0377, -0.8757, -0.8379, -0.8364, -0.6196, -0.7893, -0.6103, -0.5223, -1.0073, -0.8828, -0.9634, -1.1748, -1.9117, -1.2056, -1.3168, -0.894, -0.7139, -0.7022, -0.9364, -0.534, -0.5644, -1.8134, -1.4705, -1.3985, -1.2365, -1.0651, -1.0273, -0.8483, -0.7673, -0.7484, -0.6886, -0.5803, -0.6511, -0.5616, -0.5316, -1.0305, -0.7118, -0.8731, -1.2959, -1.6643, -0.9942, -1.2166, -1.0053, -0.645, -0.6216, -0.7562, -0.555, -0.5526};
	//PointList left, right;
	//left.resize(n);
	//right.resize(n);
	//for (index_type i = 0; i < n; i++) {
	//	left[i][0] = leftx[i];
	//	left[i][1] = lefty[i];
	//	left[i][2] = leftz[i];
	//	right[i][0] = rightx[i];
	//	right[i][1] = righty[i];
	//	right[i][2] = rightz[i];
	//}
	//registerQuat(left, right);

	//// Refine the currently selected cluster using ICP
	//if (scene->clusterResultExists()) {
	//	scene->matchCandidates->refineCluster(scene->showClusterMode);
	//}

	if (scene->skipVerification_VR) {
		scene->matchCandidates->skipVerify(scene->minVerifyClusterSize, scene->maxNumVerifyClusters_VR);
	}
	else {
		scene->matchCandidates->verifyRegion(scene->minVerifyClusterSize, scene->maxNumVerifyClusters_VR, scene->makeSceneRadius(scene->growRegionDistThreshold_VR), scene->ANNEpsilon_VR);
	}
	scene->matchCandidates->sortVerifyResult(scene->minMatchSize_VR, scene->maxNumLabels_GC);
}

void Viewer::testGraphCutRelabel()
{
	if (scene->clusterResultExists()) {
		if (scene->graphCutInstance_GC == GRAPH_CUT_INSTANCE_SYMMETRIC) {
			scene->matchCandidates->GraphCutRelabel(
				scene->graphCutSolutionType_GC,
				scene->dataTermType_GC,
				scene->useDummyLabel_GC,
				scene->bdySmoothnessKNN_GC,
				scene->makeSceneRadius(scene->closestPtDistThreshold_GC),
				scene->makeSceneRadius(scene->mismatchError_GC),
				scene->makeSceneRadius(scene->dummyLabelSmoothError_GC),
				scene->makeSceneRadius(scene->dummyLabelDataError_GC),
				scene->smoothWeight_GC,
				scene->dataWeight_GC,
				scene->makeSceneRadius(scene->smoothTruncateValue_GC),
				scene->makeSceneRadius(scene->dataTruncateValue_GC));
		}
		else if (scene->graphCutInstance_GC == GRAPH_CUT_INSTANCE_FORWARD) {
			scene->matchCandidates->GraphCutRelabel_Forward(
				scene->graphCutSolutionType_GC,
				scene->dataTermType_GC,
				scene->smoothWeight_GC,
				scene->dataWeight_GC,
				scene->makeSceneRadius(scene->smoothTruncateValue_GC),
				scene->makeSceneRadius(scene->dataTruncateValue_GC));
		}
		else {
			cerr << "Unknown graph cut instance type " << scene->graphCutInstance_GC << "!" << endl;
		}
	}
}

void Viewer::testSampleCorrespondences()
{
	scene->initMatchCandidates(*(scene->source), *(scene->target));
	scene->solveGraphCutCorrespondences(scene->matchCandidates, NULL, NULL, NULL, NULL);

	// Extract the correspondences
	//scene->packageCorrespondences();
}

ShapeModel* Viewer::selectCurrentModel()
{
	if (showAllFramesAct->isChecked()) {
		return (*scene->getAnimation())[0];
	}
	else {
		if (showTargetModel()) {
			return scene->target;
		}
		if (showSourceModel()) {
			return scene->source;
		}
		if (!showSourceModel() && !showTargetModel()) {
			return scene->source;
			//cout << "Please display a mesh for computing curvatures." << endl;
		}
	}

	return NULL;
}

void Viewer::saveFrames()
{
	//// Save the per-vertex frames of the source and target meshes.
	//const CurvInfoList *pCurvDataS = &scene->source->frames;
	//const CurvInfoList *pCurvDataT = &scene->target->frames;
	//if (!pCurvDataS || !pCurvDataT) {
	//	return;
	//}

	////////////////////////////////////////////////////////////////////////////
	//QString sourceName = scene->source->geometryFile();
	//QString savefileName;
	//savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save frames...",
	//	QString("%1/frames-%2%3.data").arg(scene->lastDir).arg(sourceName.left(sourceName.size()-4)).arg(useRobustCurvaturesAct->isChecked()?"-robust":""),
	//	QString("*.data"));

	//if (savefileName.length() == 0) {
	//	cout << "Viewer::saveCurvatures(): Save canceled." << endl;
	//	return;
	//}
	//else {
	//	scene->lastDir = grabPath(savefileName);
	//}
	//saveMeshFrames(qPrintable(savefileName), scene->source, *pCurvDataS);

	////////////////////////////////////////////////////////////////////////////
	//sourceName = scene->target->geometryFile();
	//savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save frames...",
	//	QString("%1/frames-%2%3.data").arg(scene->lastDir).arg(sourceName.left(sourceName.size()-4)).arg(useRobustCurvaturesAct->isChecked()?"-robust":""),
	//	QString("*.data"));

	//if (savefileName.length() == 0) {
	//	cout << "Viewer::saveCurvatures(): Save canceled." << endl;
	//	return;
	//}
	//else {
	//	scene->lastDir = grabPath(savefileName);
	//}
	//saveMeshFrames(qPrintable(savefileName), scene->target, *pCurvDataT);
}

void Viewer::testDistanceCode()
{
	glDisable(GL_LIGHTING);

	// Draw point
	glPointSize(5.0f);
	glBegin(GL_POINTS);
	glColor3f(1.0f, 0.0f, 0.0f);
	glVertex3f(testPoint[0],testPoint[1],testPoint[2]);
	glColor3f(0.0f, 0.0f, 1.0f);
	glVertex3f(testCMesh[0],testCMesh[1],testCMesh[2]);
	//glColor3f(0.0f, 1.0f, 0.0f);
	//glVertex3f(testCBox[0],testCBox[1],testCBox[2]);
	//glColor3f(0.0f, 0.0f, 1.0f);
	//glVertex3f(testCTri[0],testCTri[1],testCTri[2]);
	glEnd();

	//// Draw box
	//glColor3f(1.0f, 1.0f, 1.0f);
	//drawBox(testBoxMin, testBoxMax);

	//// Draw triangle
	//glBegin(GL_TRIANGLES);
	//glVertex3f(testT0[0],testT0[1],testT0[2]);
	//glVertex3f(testT1[0],testT1[1],testT1[2]);
	//glVertex3f(testT2[0],testT2[1],testT2[2]);
	//glEnd();
}

void Viewer::visualizeClusterMode()
{
	if (!scene->clusterResultExists()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Visualize the cluster modes by applying the mode transformation to the source and comparing with the target.
	float color[4];
	size_type numClusters = scene->matchCandidates->clusterModes.size();
	scene->showClusterMode = std::max(0, std::min((int)numClusters-1, scene->showClusterMode));
	//SE3	ctf = scene->matchCandidates->getClusterTransform(scene->showClusterMode);
	SE3 ctf = scene->matchCandidates->tfList[scene->showClusterMode];

	// Draw source mesh w/ transformation
	glPushMatrix();
	ctf.applyGL();
	color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	scene->source->draw();
	glPopMatrix();

	// Draw target mesh
	color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	scene->target->draw();

	// Draw correspondence pairs for this cluster
	glDisable(GL_LIGHTING);
	glBegin(GL_POINTS);
	glPointSize(10.f);
	for (index_type i = 0, i_end = scene->matchCandidates->size(); i < i_end; i++ ) {
		// Draw all the pairs with this label
		if (scene->matchCandidates->clusterLabels[i] == scene->showClusterMode) {
			//real3_type sp = scene->matchCandidates->sourceSample->vertex[scene->matchCandidates->sourceIndex[i]];
			real3_type tp = scene->matchCandidates->targetSample->vertex[scene->matchCandidates->targetIndex[i]];

			//glColor3f(0.5f, 0.5f, 1.0f);
			//glVertex3f(sp[0],sp[1],sp[2]);
			glColor3f(1.0f, 0.2f, 0.2f);
			glVertex3f(tp[0],tp[1],tp[2]);
		}
	}
	glEnd();

	displayMessage(QString("Showing cluster %1 / %2 (%3 pairs)").arg(scene->showClusterMode).arg(numClusters).arg(scene->getClusterSize(scene->showClusterMode)));

	////////////////////////////////////////////////////////////////////////////
	//// Draw each verified region
	//size_type numRegions = scene->matchCandidates->matchList.size();
	//scene->showClusterMode = max(0, min(numRegions-1, scene->showClusterMode));
	//if (!scene->matchCandidates->validCluster(scene->showClusterMode)) {
	//	return;
	//}

	//CorrespondenceList &list = scene->matchCandidates->matchList[scene->showClusterMode];
	//glDisable(GL_LIGHTING);
	//glPointSize(3.0f);
	//glBegin(GL_POINTS);
	//for (index_type i = 0; i < list.size(); i++) {
	//	real3_type temp;

	//	// Draw source vertex
	//	glColor3f(0.7f,0.7f,1.0f);
	//	temp = list.sourcePos(i);
	//	glVertex3f(temp[0],temp[1],temp[2]);

	//	// Draw corresponding target vertex
	//	glColor3f(1.0f,0.7f,0.7f);
	//	temp = list.targetPos(i);
	//	glVertex3f(temp[0],temp[1],temp[2]);
	//}
	//glEnd();

	//displayMessage(QString("Showing region %1 with %2 vertices").arg(scene->showClusterMode).arg(list.size()));

	////////////////////////////////////////////////////////////////////////////
	//// Draw each verified region  -->  Moved to CorrespViewer
	//size_type numRegions = scene->matchCandidates->matchList.size();

	//if (scene->clusterColors.size() < numRegions) {
	//	// If not enough colors, generate more
	//	scene->generateClusterColors(numRegions);
	//}

	//glDisable(GL_LIGHTING);
	//glPointSize(3.0f);
	//glBegin(GL_POINTS);
	//for (index_type ci = 0; ci < numRegions; ci++) {
	//	if (!scene->matchCandidates->validCluster(ci)) {
	//		continue;
	//	}

	//	setOpenGLColor(scene->getClusterColor(ci));

	//	CorrespondenceList &list = scene->matchCandidates->matchList[ci];
	//	for (index_type i = 0; i < list.size(); i++) {
	//		real3_type temp;

	//		// Draw corresponding target vertex
	//		temp = list.targetPos(i);
	//		glVertex3f(temp[0],temp[1],temp[2]);
	//	}
	//}
	//glEnd();
}


void Viewer::fitViewToShape( const ShapeModel *shape /*= NULL */ )
{
	real3_type box_min, box_max;

	if (!shape) {
		std::pair<real3_type,real3_type> box = scene->getSolverDisplayBBox();
		box_min = box.first;
		box_max = box.second;

		// If the bounding box is empty
		if (box_min == box_max) {
			shape = selectCurrentModel();
		}
	}

	if (shape) {
		box_min = shape->box_min;
		box_max = shape->box_max;
	}

	// Fit the view to the current shape
	setSceneBoundingBox(Vec3fToVec(box_min), Vec3fToVec(box_max));

	camera()->setZClippingCoefficient(10.f);
	camera()->setZNearCoefficient(1e-2f);
	showEntireScene();

	//cout << "Camera near clipping plane: " << camera()->zNear() << endl;
	//cout << "Camera far clipping plane: " << camera()->zFar() << endl;
}


void Viewer::simplifyCurrentShape()
{
	//// Removed from this release in order to remove VTK library

	////// Prompt for the simplification factor
	////bool ok;
	////double fac = QInputDialog::getDouble(this, tr("Simplification Factor"),
	////	tr("Factor Reduction:"), 0.5f, 0.0f, 1.0f, 4, &ok);

	////if (ok) {
	////	ShapeModel *newShape = ((Mesh3_real*)selectCurrentModel())->simplifyGeometry(fac);
	////	scene->getAnimation()->add(newShape);
	////}
}

void Viewer::batchSimplify()
{
	//// Removed from this release in order to remove VTK library

	////// Get a list of files
	////QStringList files = QFileDialog::getOpenFileNames(this, tr("Select meshes to simplify..."), scene->lastDir, tr("Meshes (*.obj)"));
	////if (files.size() == 0) {
	////	return;
	////}

	////// Prompt for a simplification factor
	////bool ok;
	////double fac = QInputDialog::getDouble(this, tr("Simplification Factor"),
	////	tr("Factor Simplification:"), 0.0f, 0.0f, 1.0f, 4, &ok);
	////if (!ok) {
	////	return;
	////}

	////// Prompt for a scale factor
	////double scale = QInputDialog::getDouble(this, tr("Scale Factor"),
	////	tr("Factor Scale:"), 1.0f, 0.0f, 1e10, 4, &ok);
	////if (!ok) {
	////	return;
	////}

	////// Simplify each file
	////QStringList list = files;
	////QStringList::Iterator it = list.begin();
	////while(it != list.end()) {
	////	QString filename = *it;

	////	// Load and simplify
	////	Mesh3_real loaded;
	////	loaded.setFile(QString(scene->rootDir.c_str()), filename);
	////	Mesh3_real *simplified = loaded.simplifyGeometry(fac, scale);
	////	delete simplified;

	////	++it;
	////}
}

void Viewer::changeBackgroundColor()
{
	QColor color = QColorDialog::getColor(this->backgroundColor(), this);
	if (color.isValid()) {
		this->setBackgroundColor(color);
	}
}

void Viewer::changeForegroundColor()
{
	QColor color = QColorDialog::getColor(this->foregroundColor(), this);
	if (color.isValid()) {
		this->setForegroundColor(color);
	}
}

real_type Viewer::ComputeAlignDistance()
{
	real_type error = REAL_TYPE_MAX;

	if (scene->matchCandidates) {
		error = scene->matchCandidates->maxSymmetricHausdorff();

		cout << "Maximum Symmetric Hausdorff Error: " << error << " ("
			 << error / scene->totalScene_boundingBoxDiagonal * 100.0 << "% of bbox diagonal)" << endl;
	}

	return error;
}

void Viewer::CaptureResultSnapshots_DeformGraphAll()
{
	if (!scene->is_deformation_set_up()) {
		return;
	}

	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	// Reset the view
	if (!loadDisplaySettings()) {
		return;
	}

	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(false);
	setGraphShowAct->setChecked(false);
	showShapeWeightsAct->setChecked(true);	
	showJointsAct->setChecked(false);
	showExternalMeshDeformedAct->setChecked(false);
	setTransparencyAct->setChecked(false);
	real_type rotateIncrement = 0.0f;
	//real_type rotateIncrement = -360.0 / ((double)maxDisplayFrame);

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	for (int i = 0; i <= maxDisplayFrame; ++i) {
		// Change the frame
		displayFrame = i;

		// Rotate the model
		rotateAbout(qglviewer::Vec(0, 1, 0), rotateIncrement);

		// Shot 1: show the input frame
		setSourceModelShowAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2-shot01-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		updateGL();
		setSourceModelShowAct->setChecked(false);

		// Shot 2: show the ASG
		setGraphShowAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2-shot02-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		updateGL();
		setGraphShowAct->setChecked(false);

		if (externalMeshLoaded) {
			//// Shot 2: show the ASG
			//setGraphShowAct->setChecked(true);
			//customSaveSnapshotFile = tr("%1/%2-shot03-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
			//updateGL();
			//setGraphShowAct->setChecked(false);

			// Shot 3: show reconstructed mesh
			showExternalMeshDeformedAct->setChecked(true);
			customSaveSnapshotFile = tr("%1/%2-shot04-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
			updateGL();
			showExternalMeshDeformedAct->setChecked(false);

			// Shot 4: show reconstructed mesh with joints
			showJointsAct->setChecked(true);
			showExternalMeshDeformedAct->setChecked(true);
			setTransparencyAct->setChecked(true);
			customSaveSnapshotFile = tr("%1/%2-shot05-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
			updateGL();
			showJointsAct->setChecked(false);
			showExternalMeshDeformedAct->setChecked(false);
			setTransparencyAct->setChecked(false);
		}
	}

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}

void Viewer::CaptureResultSnapshots_DeformGraphAll_Model()
{
	if (!scene->is_deformation_set_up() || !externalMeshLoaded) {
		return;
	}

	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	int numSteps = QInputDialog::getInt(this, tr("Number of Steps"), tr("Please enter the number of steps:"), 60);

	// Reset the view
	if (!loadDisplaySettings()) {
		return;
	}

	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(false);
	setGraphShowAct->setChecked(false);
	showShapeWeightsAct->setChecked(true);	
	showJointsAct->setChecked(false);
	showExternalMeshDeformedAct->setChecked(false);
	setTransparencyAct->setChecked(false);

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	// Change the frame to the reference frame
	displayFrame = scene->refFrameIndex;

	for (int i = 0; i < numSteps; ++i) {
		//// Shot 2: show the ASG
		//setGraphShowAct->setChecked(true);
		//setSourceModelShowAct->setChecked(true);
		//customSaveSnapshotFile = tr("%1/%2-shot02-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		//updateGL();
		//setSourceModelShowAct->setChecked(false);
		//setGraphShowAct->setChecked(false);

		// Shots 6,7: show reconstructed mesh
		showExternalMeshDeformedAct->setChecked(true);
		//showShapeWeightsAct->setChecked(false);
		//customSaveSnapshotFile = tr("%1/%2-shot06-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		//updateGL();
		showShapeWeightsAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2-shot07-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		updateGL();
		showExternalMeshDeformedAct->setChecked(false);

		// Shot 8,9: show reconstructed mesh with joints
		showJointsAct->setChecked(true);
		showExternalMeshDeformedAct->setChecked(true);
		setTransparencyAct->setChecked(true);
		//showShapeWeightsAct->setChecked(false);
		//customSaveSnapshotFile = tr("%1/%2-shot08-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		//updateGL();
		showShapeWeightsAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2-shot09-frame%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		updateGL();
		showJointsAct->setChecked(false);
		showExternalMeshDeformedAct->setChecked(false);
		setTransparencyAct->setChecked(false);

		// Rotate the camera
		rotateCamera_DeformedBall(-1.0, 0.0, numSteps);
	}

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}

void Viewer::CaptureResultSnapshots_DeformGraphAll_Init()
{
	if (!scene->is_deformation_set_up()) {
		return;
	}

	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	// Reset the view
	if (!loadDisplaySettings()) {
		return;
	}

	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(false);
	setGraphShowAct->setChecked(false);
	showShapeWeightsAct->setChecked(true);	
	showJointsAct->setChecked(false);
	showExternalMeshDeformedAct->setChecked(false);
	setTransparencyAct->setChecked(false);

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	for (displayFrame = scene->refFrameIndex; displayFrame < maxDisplayFrame; ++displayFrame) {
		// Shot 10: show initialization source
		setSourceModelShowAct->setChecked(true);
		showInitSrc = true;
		showInitResult = false;
		customSaveSnapshotFile = tr("%1/%2-shot10-frame%3.png").arg(dir).arg(headerText).arg(displayFrame, 4, 10, QLatin1Char('0'));
		updateGL();

		// Shot 11: show initialization result
		showInitSrc = false;
		showInitResult = true;
		customSaveSnapshotFile = tr("%1/%2-shot11-frame%3.png").arg(dir).arg(headerText).arg(displayFrame, 4, 10, QLatin1Char('0'));
		updateGL();
	}

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}

Deform::DrawOptions Viewer::getDeformDrawOptions() const
{
	Deform::DrawOptions options;

	// Populate option values from the viewer and the scene
	options.showSourceModel = setSourceModelShowAct->isChecked();
	options.showTargetModel = setTargetModelShowAct->isChecked();
	options.showDeformedModel = setDeformedModelShowAct->isChecked();
	options.showSourceCorresp = sourceCorrespShowAct->isChecked();
	options.showTargetCorresp = targetCorrespShowAct->isChecked();
	options.showDeformedCorresp = deformedCorrespShowAct->isChecked();
	options.showShapeWeights = showShapeWeightsAct->isChecked();
	options.showAllFrames = showAllFramesAct->isChecked();
	options.displayFrame = displayFrame;
	options.displayTargettingFrame = displayTargettingFrame;
	options.showGraph = setGraphShowAct->isChecked();
	options.showEffRad = setEffRadShowAct->isChecked();
	options.renderTransparent = renderTransparent();
	options.showComponent = scene->currentBone;
	options.showGraphDeformed = showDeformedGraphAct->isChecked();
	options.showGraphFaces = showGraphFacesAct->isChecked();
	options.showExtendedGrid = -1;
	options.showWireframe = showWireframe();
	options.openGLPointSize = 2.0f;
	options.sourceModelColor = RealColor(1.0f, 0.7f, 0.7f);
	options.deformedModelColor = RealColor(0.7f, 1.0f, 0.7f);
	//options.deformedModelColor = RealColor(0.7f, 0.7f, 1.0f);
	options.targetModelColor = RealColor(0.7f, 0.7f, 1.0f);
	options.pointColor = RealColor(scene->pointColor[0],scene->pointColor[1],scene->pointColor[2]);
	options.showErrorComponent = showErrorComponent;
	options.ignoreDisplayInterface = false;
	options.showSourceDeformed = showSourceDeformedAct->isChecked();
	options.showInvalidVertices = showInvalidVerticesAct->isChecked();
	options.showSampledPoints = showSampledPointsAct->isChecked();
	options.showExternalMeshDeformed = externalMeshLoaded && showExternalMeshDeformedAct->isChecked();   // Show deformed external mesh
	options.showJoints = showJointsAct->isChecked();
	options.sampleRadius = scene->pointSize();
	options.curvDisplayMode = curvDisplayMode;
	options.curvDisplayScale = displayScale;
	options.boneColors = &(scene->boneColors);

	// Option to show initialization
	options.showInitSrc = showInitSrc;
	options.showInitResult = showInitResult;


	// Debugging
	options.showCellError = showCellErrorAct->isChecked();
	options.showCellAffectedPoints = true;
	options.showPointAffectingCells = false;
	options.cellPickIndex = cellPickIndex;
	options.pointPickIndex = pointPickIndex;
	options.showSampleError = showSampleErrorAct->isChecked();

	return options;
}

bool Viewer::Convert_BOBJ_To_OBJ()
{
	QString qfilename = QFileDialog::getOpenFileName(this, tr("Open Binary OBJ"), scene->lastDir, tr("Meshes (*.bobj)"));
	if (qfilename.isEmpty()) {
		return false;
	}
	QString qSaveFilename = QFileDialog::getSaveFileName(this, tr("Save OBJ"), scene->lastDir, tr("Meshes (*.obj)"));
	if (qSaveFilename.isEmpty()) {
		return false;
	}

	ifstream in;
	in.open(qPrintable(qfilename), ios_base::binary);

	if (in.fail()) {
		cout << "Error in Convert_BOBJ_To_OBJ(): could not open file " << qPrintable(qfilename) << " for reading!" << endl;
		return false;
	}

	ofstream out;
	out.open(qPrintable(qSaveFilename));

	if (out.fail()) {
		in.close();
		cout << "Error in Convert_BOBJ_To_OBJ(): could not open file " << qPrintable(qSaveFilename) << " for writing!" << endl;
		return false;
	}

	// General structure of this sort of thing: http://www.velocityreviews.com/forums/t283118-fstreamread.html
	//while (m_aa.read(&c, sizeof(c))
	//{
	//	m_oo.write(&c, sizeof(c));
	//	cout<<"\nout to c:"<<(int)c;
	//}
	//if (m_aa.eof())
	//{
	//	// ok we hit end of file
	//}
	//else
	//{
	//	// some other sort of error
	//}

	cout << "Converting BOBJ to OBJ: " << qPrintable(qfilename) << " --> " << qPrintable(qSaveFilename) << endl;

	char type;
	while (in.read(&type, sizeof(char))) {
		if (type == 0) {
			// Read/Write a vertex
			out << 'v';

			float temp;
			for (int i = 0; i < 3; ++i) {
				in.read((char *) &temp, sizeof(float));
				out << ' ' << temp;
			}

			out << '\n';
		}
		else if (type == 1) {
			// Read/Write a triangle
			out << 'f';

			int temp;
			for (int i = 0; i < 3; ++i) {
				in.read((char *) &temp, sizeof(int));
				out << ' ' << temp;
			}

			out << '\n';
		}
		else {
			// Ignore others
		}
	}
	if (!in.eof()) {
		cerr << "Some sort of file error in Convert_BOBJ_To_OBJ()." << endl;
	}

	in.close();
	out.close();

	cout << "Converting complete!" << endl;

	return true;
}

void Viewer::setupExternalMeshDeformAll()
{
	if (scene->is_deformation_set_up(Deform::DEFORM_GRAPH_ALL)) {
		// Just assume it's loaded
		externalMeshLoaded = true;
		return;
	}

	//if (!scene->is_deformation_set_up(Deform::DEFORM_ALL)) {
	//	return;
	//}

	//if (!externalMeshLoaded) {
	//	// Load the mesh
	//	QString qfilename = QFileDialog::getOpenFileName(this, tr("Open OBJ"), scene->lastDir, tr("Meshes (*.obj)"));
	//	if (qfilename.isEmpty()) {
	//		return;
	//	}
	//	ShapeModel *shape = loadShape(tr(scene->rootDir.c_str()), qfilename, *scene, 0);

	//	// Register the mesh with the DeformAll instance
	//	scene->setupExternalMesh_DeformAll(shape);

	//	// Set the state
	//	externalMeshLoaded = true;
	//}
}

void Viewer::saveDisplaySettings()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the camera parameters...", QString(), QString("*.xml"));
	if (savefileName.length() == 0) {
		return;
	}

	QDomDocument document("DeformGraphAll");
	QDomElement ds = document.createElement("DisplaySettings");
	document.appendChild(ds);
	ds.appendChild( camera()->domElement("Camera", document) );

	// Add bone colors
	QDomElement bc = document.createElement("BoneColors");
	for (int i = 0; i < scene->graph_n; ++i) {
		QDomElement temp = document.createElement(tr("c%1").arg(i, 4, 10, QLatin1Char('0')));
		temp.appendChild(document.createTextNode(tr("%1,%2,%3").arg(scene->boneColors[i][0]).arg(scene->boneColors[i][1]).arg(scene->boneColors[i][2])));
		bc.appendChild(temp);
	}
	ds.appendChild(bc);

	// Add resolution settings
	QDomElement res = document.createElement("Window");
	res.appendChild(document.createTextNode(tr("%1,%2,%3,%4").arg(this->x()).arg(this->y()).arg(this->width()).arg(this->height())));
	ds.appendChild(res);

	QFile f(savefileName);
	if (f.open(QIODevice::WriteOnly))
	{
		QTextStream out(&f);
		document.save(out, 2);
	}
}

bool Viewer::loadDisplaySettings(QString file/* = QString()*/)
{
	if (file.length() == 0) {
		// Get the save file name
		file = QFileDialog::getOpenFileName(this, "Choose a file name to load the display parameters...", QString(), QString("*.xml"));
		if (file.length() == 0) {
			return false;
		}
	}

	// Load DOM from file
	QDomDocument document;
	QFile f(file);
	if (f.open(QIODevice::ReadOnly)) {
		document.setContent(&f);
		f.close();
	}

	//cout << "Document: " << qPrintable(document.toString()) << endl;
	QDomElement child = document.documentElement().firstChild().toElement();

	while (!child.isNull()) {
		if (child.tagName() == "Camera") {
			cout << "Applying camera settings" << endl;
			camera()->initFromDOMElement(child);
		}

		if (child.tagName() == "BoneColors") {
			cout << "Applying bone colors" << endl;
			QDomElement temp = child.firstChild().toElement();
			while (!temp.isNull()) {
				QString tagName = temp.tagName().remove(0,1);
				int idx = tagName.toInt();
				QStringList list1 = temp.text().split(",");

				// Read the color from the text
				real3_type color;
				for (int i = 0; i < list1.size(); ++i) {
					color[i] = list1[i].toFloat();
				}
				cout << "Color for bone " << idx << ": " << color << endl;
				scene->boneColors[idx] = color;

				temp = temp.nextSibling().toElement();
			}
		}

		if (child.tagName() == "Window") {
			cout << "Applying Resolution" << endl;
			QStringList list1 = child.text().split(",");
			int2_type loc, res;
			loc[0] = list1[0].toInt();
			loc[1] = list1[1].toInt();
			res[0] = list1[2].toInt();
			res[1] = list1[3].toInt();

			this->move(loc[0], loc[1]);
			this->resize(res[0], res[1]);
		}

		//cout << "Type of next sibling: " << child.nextSibling().nodeType() << endl;
		child = child.nextSibling().toElement();
	}

	return true;
}

void Viewer::dragEnterEvent( QDragEnterEvent *event )
{
	//setText(tr("<drop content>"));
	//setBackgroundRole(QPalette::Highlight);

	event->acceptProposedAction();
	//emit changed(event->mimeData());
}

void Viewer::dropEvent( QDropEvent *event )
{
	const QMimeData *mimeData = event->mimeData();

	if (mimeData->hasImage()) {
		//setPixmap(qvariant_cast<QPixmap>(mimeData->imageData()));
		cout << "Input is image" << endl;
	} else if (mimeData->hasHtml()) {
		//setText(mimeData->html());
		//setTextFormat(Qt::RichText);
		cout << qPrintable(mimeData->html()) << endl;
	} else if (mimeData->hasText()) {
		//setText(mimeData->text());
		//setTextFormat(Qt::PlainText);
		cout << qPrintable(mimeData->text()) << endl;
	} else if (mimeData->hasUrls()) {
		QList<QUrl> urlList = mimeData->urls();
		QString text;
		for (int i = 0; i < urlList.size() && i < 32; ++i) {
			QString url = urlList.at(i).path();
			text += url + QString("\n");
		}
		//setText(text);
		cout << qPrintable(text) << endl;
	} else {
		//setText(tr("Cannot display data"));
		cout << "Cannot display data" << endl;
	}

	//setBackgroundRole(QPalette::Dark);
	event->acceptProposedAction();
}

void Viewer::showOpenGLInfo()
{
	printOpenGLInfo();
}
