#include <QGLViewer/qglviewer.h>
#include <QtGui>
#include <QMutex>
#include <memory>
#include "observer.h"
#include "subject.h"
#include "PointList.h"
#include "PointListN.h"
#include "FindCorrespondence.h"

class Scene;
class CorrespViewer;
class DeformEdit;
class CorrespEdit;
class MatchCandidates;
class FeatureViewer;
class ShapeSlicer;
class ShapePoser;
class TransformViewer;
class AccumulatedGeometryViewer;
class DisplayOptions;
namespace Deform { class DrawOptions; }

/**
 * Viewer class.
 *
 * Responsibilities:
 *  - Displaying meshes on screen
 *  - Implementing menu & menu functionality
 */
class Viewer : public QGLViewer, public Observer
{
	Q_OBJECT

public:
	Viewer( Scene* s, const QGLFormat& format, QWidget* parent=0, const QGLWidget* shareWidget=0, Qt::WFlags flags=0 );	/**< Constructor: initialize members here. */
	virtual ~Viewer();										/**< Destructor: destroy members here. */

	virtual void Update(Subject* theChangedSubject);		/**< Update the display according to the subject */
	Deform::DrawOptions getDeformDrawOptions() const;		/**< Gather the drawing options for drawing the deformed model */

private slots:
	void open(const char *file = NULL, int maxFrames = 0, int skipFrames = 1);	/**< Open at most "maxFrames" frames of a mesh animation sequence. */
	void add();

	void clearResult();										/**< Clear the deformation result. */
	void setGraphNodes();									/**< Set the number of graph nodes. */
	void resampleGraph();									/**< Resample the graph node locations. */
	void customSaveSnapshot();								/**< Take a snapshot and save it in a specific file. */
	void captureVideo();									/**< Take a video of the mesh animation sequence. */

	void CreateDeformEditor();								/**< Display the dialog for editing deformation parameters */
	void DeformEditorDestroyed();							/**< Cleanup after the editor has been closed */
	void DeformEditorChanged(int which, double value);		/**< Handle changes in parameter values */

	void CreateCorrespViewer();								/**< Display the correspondence viewer */
	void CorrespViewerDestroyed();							/**< Cleanup after the correspondence viewer has been closed */

	void saveColormap();									/**< Save the current colormap */
	void loadColormap();									/**< Load the current colormap */
	void setColormapLimits();								/**< Set colormap limits */

	void setFeatureThreshold();								/**< Set threshold on feature for display */

	void autoLaunch();										/**< Code to execute after the first draw */
	void autoProcess();										/**< Automatic processing for batch jobs */

	void saveFrames();

	void simplifyCurrentShape();							/**< Simplify the current displayed shape. */
	void batchSimplify();									/**< Simplify a batch of shapes. */
	void fitViewToShape( const ShapeModel *shape = NULL );	/**< Fit the view to the current shape. */

	void changeBackgroundColor();							/**< Change background color */
	void changeForegroundColor();							/**< Change foreground color */

	void CreateShapePoser();								/**< Display the shape poser */
	void ShapePoserDestroyed();								/**< Cleanup after the shape poser has been closed */

	void CreateTransformViewer();							/**< Display the transform viewer (source-to-target)  */
	void TransformViewerDestroyed();						/**< Cleanup after the transform viewer has been closed */

	real_type ComputeAlignDistance();						/**< Compute the alignment distance. */

	void CaptureResultSnapshots_DeformGraphAll();			/**< Capture a sequence of snapshots showing the result. */
	void CaptureResultSnapshots_DeformGraphAll_Model();		/**< Capture a sequence of snapshots showing the result. */
	void CaptureResultSnapshots_DeformGraphAll_Init();		/**< Capture a sequence of snapshots showing the result. */

	bool Convert_BOBJ_To_OBJ();								/**< Convert binary obj files to normal obj files */

	void saveDisplaySettings();
	bool loadDisplaySettings(QString file = QString());

	void showOpenGLInfo();

protected:
	virtual void draw();
	virtual void init();
	virtual void animate();
	virtual QString helpString() const;
	virtual	void keyPressEvent(QKeyEvent *e);
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);
	virtual void dragEnterEvent(QDragEnterEvent *event);
	virtual void dropEvent(QDropEvent *event);

private:
	//////////////////////////////////////////////////////////////////////////
	// Context Menu Related Methods and Members
	//////////////////////////////////////////////////////////////////////////
	QMenu *contextMenu;
	QMenu *fileMenu;
	QMenu *displayMenu;
	QMenu *toolsMenu;
	QMenu *optionsMenu;

	QAction *openAct;
	QAction *addAct;
	QAction *setWireframeModeAct;
	QAction *setBboxModeAct;
	QAction *setSourceModelShowAct;
	QAction *setTargetModelShowAct;
	QAction *setDeformedModelShowAct;
	QAction *sourceCorrespShowAct;
	QAction *targetCorrespShowAct;
	QAction *deformedCorrespShowAct;
	QAction *setGraphShowAct;
	QAction *setEffRadShowAct;
	QAction *setTransparencyAct;
	QAction *clearResultAct;
	QAction *setGraphNodesAct;
	QAction *resampleGraphAct;
	QAction *captureVideoAct;
	QAction *captureVideoAct2;
	QAction *captureVideoAct3;
	QAction *captureVideoAct4;
	QAction *showDeformEditorAct;
	QAction *showAllFramesAct;
	QAction *showCorrespViewerAct;
	QAction *showAccumulatedGeometryViewerAct;
	QAction *saveColormapAct;
	QAction *loadColormapAct;
	QAction *setColormapLimitsAct;
	QAction *setFeatureThresholdAct;
	QAction *showCorrespEditorAct;
	QAction *saveFramesAct;
	QAction *visualizeClusterAct;
	QAction *simplifyCurrentShapeAct;
	QAction *batchSimplifyAct;
	QAction *fitViewToShapeAct;
	QAction *changeBackgroundColorAct;
	QAction *changeForegroundColorAct;
	QAction *shapeSliceAct;
	QAction *shapePoseAct;
	QAction *assignTransformsAct;
	QAction *assignTransformsReverseAct;
	QAction *showVerticesAct;
	QAction *showVertexFramesAct;
	QAction *computeAlignDistanceAct;
	QAction *showDeformedSameColorAct;
	QAction *showDeformedGraphAct;
	QAction *showGraphFacesAct;
	QAction *showDisplayOptionsAct;
	QAction *showShapeWeightsAct;
	QAction *showCellErrorAct;
	QAction *showSampleErrorAct;
	QAction *convertBOBJAct;
	QAction *showSourceDeformedAct;
	QAction *showInvalidVerticesAct;
	QAction *showSampledPointsAct;
	QAction *showExternalMeshDeformedAct;
	QAction *showJointsAct;
	QAction *saveDisplaySettingsAct;
	QAction *loadDisplaySettingsAct;

	QActionGroup *modelDisplayGroup;
	QAction *showFeaturesAct;
	QAction *showCurvaturesAct;
	QAction *showFramesAct;
	QAction *showGeometryAct;
	QAction	*showSamplesAct;
	QAction *showStatusAct;
	QAction *showCellIndexAct;

	QActionGroup *deformGroup;
	QAction *useEDAct;
	QAction *useWDAct;
	QAction *useDeformEMAct;
	QAction *useDeformAllAct;
	QAction *useDeformGraphAllAct;

	std::tr1::shared_ptr<QAction> showOpenGLInfoAct;

	void createContextMenu();


	//////////////////////////////////////////////////////////////////////////
	// Scene data
	//////////////////////////////////////////////////////////////////////////
	Scene *scene;

	//////////////////////////////////////////////////////////////////////////
	// State methods and parameters
	//////////////////////////////////////////////////////////////////////////
	bool showWireframe() const;
	bool showBbox();
	bool showSourceModel();
	bool showTargetModel();
	bool showDeformedModel();
	int whichDeform();
	bool renderTransparent() const;
	QString updateFrame(int frame);
	void drawSphereTest();
	void drawCurvatureScales();
	void colormapTest();
	void pointListTest();
	void saveTransformSpacePoints();
	ShapeModel *selectCurrentModel();
	void drawFrame( const ShapeModel* frame );
	int showErrorComponent;

	void rotateAbout(qglviewer::Vec up, float angle);
	void rotateCamera_DeformedBall( float x, float y, int steps );
	int displayFrame, displayTargettingFrame, maxDisplayFrame;

	// Scene change notify variable
	QMutex sceneMutex;
	bool sceneChanged;

	// Snapshot file
	QString customSaveSnapshotFile;

	//////////////////////////////////////////////////////////////////////////
	// Set deform parameters
	DeformEdit* deformEdit;
	float maxGraphNodes, maxGraphNeighbors, maxOverlapFactor;

	CorrespViewer *correspViewer;
	ShapePoser *shapePoser;
	TransformViewer *transformViewer;

	//////////////////////////////////////////////////////////////////////////
	// Test capabilities of PointList
	PointList *tempList;

	//////////////////////////////////////////////////////////////////////////
	// Test sphere-fitting
	enum InteractionMode { NONE = 0, DEFINE_SPHERE, DRAW_STROKE };
	InteractionMode interactionMode_;
	real3_type clickStartPos, clickEndPos;
	real3_type fitCenter;
	real_type fitRadius;

	real3_type closestPointOnLine( QPoint pos, real3_type x );
	void sphereFit();
	void testFeatureSelect();
	void featureProximity();
	void testFeatureMatch();
	void testFeatureCluster();
	void testRefineCluster();
	void verifyTransformationSpace();
	void visualizeClusterMode();
	void testGraphCutRelabel();
	void testSampleCorrespondences();

	// Range query for mesh
	idxList inRange;
	distList distRange;

	// Range query in feature space
	int sourceQueryIndex;
	PointList sourceQuery, targetMatch;

	// Curvature values
	int displayScale;
	PointListN *sourceFeatures, *targetFeatures;

	// Maximum values for various correspondence parameters
	float maxFeatureSpaceRange;

	// Principal curvatures testing variables
	int curvDisplayMode;

	bool autoLaunchDone;


	// Debugging: test point-box and point-triangle distance code.
	real3_type testBoxMin, testBoxMax;
	real3_type testT0, testT1, testT2;
	real3_type testPoint, testCBox, testCTri, testCMesh;
	real_type testDBox, testDTri, testDMesh;
	void testDistanceCode();

	// Debugging: show grid cells that affect each point
	int pointPickIndex, cellPickIndex;


	// Load an additional mesh (the reconstructed object) and
	// ...show it deformed according to weights and transformations
	void setupExternalMeshDeformAll();
	bool externalMeshLoaded;

	// Show initalization result
	bool showInitSrc, showInitResult;

	bool captureCorrespViewerResult(int winH, int winV);
	void correspViewerPosition( int x, int y, int width = 0, int height = 0 );
	void deformEditPosition( int x, int y, int width = 0, int height = 0 );

signals:
	void updateChildDisplay();
};

void setViewerPosition( QWidget& widget, int x, int y, int width, int height );
