#include "DeformGraphAll.h"
#include <GraphCutLib/GCoptimization.h>
#include <LapackLib/dgetrf.h>
#include <LapackLib/nnls.h>
#include "scene.h"
#include "ShapeModel.h"
#include "FindCorrespondence.h"
#include "FileSequence.h"
#include "quat.h"
#include "util.h"
#include "geom_util.h"
#include "taucsUtil.h"
#include <psapi.h>
#include <boost/algorithm/string.hpp>

#define OPT_TF_PREMULT

#undef min
#undef max

int TEST_COMP_TYPE = 2;

using namespace Deform;
using namespace Deform::DeformAll_Utility;

namespace Deform {
	namespace DeformAll_Utility {

		const BoostEdgeWeightType BOOST_EDGE_WEIGHT_TYPE_MAX = REAL_TYPE_MAX;

		struct IndexedWeight {
			IndexedWeight() : index(0), weight(0.0) { };
			IndexedWeight(int i, real_type w) : index(i), weight(w) { };

			int index;				/**< Index of the bone */
			real_type weight;		/**< Weight of the bone */
		};
		typedef std::vector<IndexedWeight> SparseWeight;
		typedef std::deque<SparseWeight> SparseWeightList;

		bool operator==( const SparseWeight& w1, const SparseWeight& w2 )
		{
			if (w1.size() != w2.size()) return false;
			if (w1.size() == 0) return true;
			if (w1.size() > 1) return false;
			if (w1[0].index != w2[0].index) return false;
			return true;
		}

		typedef std::map<int,const TransformList*> SparseTLArray;



		class AllSampleGraph
		{
		public:
			AllSampleGraph( int k, real_type t );

			// Just add samples of a new frame to the graph, with no weights
			void initASG( const ShapeModel &shape, const idxList& newSampleIndices, const int frameIndex );
			// Add samples of a new frame to the graph, with prior knowledge of weights
			void addSamples( const ShapeModel &shape, const idxList& newSampleIndices, const RefWeightList& newSampleWeights, const int newFrameIndex, const SparseTLArray& allTfList );
			// Add samples of a new frame to the graph, with previous sampling of the weight function
			void resampleASG( const ShapeModel &shape, const idxList& newSampleIndices, const PointList& refSamples, const RefWeightList& refSampleWeights, const std::vector<bool>& isOccluded, const int newFrameIndex, const SparseTLArray& allTfList );

			// Rebuild graph
			void rebuild( const ShapeModel &shape, const idxList& newSampleIndices, const int newFrameIndex, const std::vector<bool>& isOccluded, const std::vector<PointList>& refSamples_Bone, const std::vector<PointList>& refNormals_Bone, const SparseTLArray& allTfList, real_type minDist, bool accumulateOnRef, int refFrameIndex );
			void pickSamples_NoOverlap( const PointList& existingPts, const PointList& existingNormals, const PointList& newPoints, const PointList& newNormals, const idxList& subset, idxList& pickedResult ) const;

			// Remove edges that are stretched too far
			void pruneEdges( const Deform::DeformAll_Utility::SparseTLArray& allTfList, real_type maxStretchFactor, real_type maxEdgeLength, const DeformGraphAll_Utility::JointMap *jointMap );

			// Transform sample positions
			void getVertexNormal_Ref( const SparseTLArray& allTfList, PointList& vList, PointList& nList ) const;
			void getVertexNormal_Frame( const int frameIndex, const SparseTLArray& allTfList, PointList& vList, PointList& nList ) const;


			inline int numSamples() const {
				return vertex.size();
			}

			// Parameters
			int EKNNgraph_k;
			real_type sampleDistThreshold;

			// Data members
			BoostGraph graph;						/**< Euclidean k-nearest neighbor graph for selected source vertices */

			// Q: Should we maintain these explicitly, or should we always use the frame/shape indices to lookup from the original data source?
			// -> If we always use frame/shape indices, then we need to keep all the frames in memory...
			PointList vertex, normal;				/**< Vertex/normal in the original frames */

			// Q: Why do we need to maintain these?
			SparseWeightList weightList;
			idxList fIndex, sIndex;					/**< fIndex = Frame number of each sample,  sIndex = sampling index # from */


			// Testing and displaying via OpenGL
			PointList vertex_disp, normal_disp;
			idxList testIndex;
			std::list< std::pair<int,int> > testList;
			std::vector<distList> testScores;
		};
	}
}

// Function declarations
float MemoryToMB(int bytes);
DWORD CalculateWSPrivate(DWORD processID, const SYSTEM_INFO &sysInfo);
void lbs(const real3_type& x, real3_type& x_out, const Deform::DeformAll_Utility::SparseWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse );
void lbs(const real3_type& x, const real3_type& n, real3_type& x_out, real3_type& n_out, const Deform::DeformAll_Utility::SparseWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse );
void makeAllTFList( const Deform::DeformGraphAll_Utility::SolutionTable& frameSolution, Deform::DeformAll_Utility::SparseTLArray& allTfList );
int findLabel( const Deform::DeformAll_Utility::SparseWeight& weight );
void printAllTFList( const Deform::DeformAll_Utility::SparseTLArray& allTfList );
void extrapolateWeights( const std::vector<PointList>& refSamples_Bone, const std::vector<PointList>& refNormals_Bone, const std::vector<bool>& isOccluded, const PointList& newPts, const idxList& sampleIdx, const Deform::DeformAll_Utility::TransformList& tfList, real_type minDist, std::vector<int>& label, std::vector<distList> *scores = NULL );
void printMatrixInfo( taucs_ccs_matrix *m );
void addToDiagonal_Symmetric( taucs_ccs_matrix *m, taucsType value );

using namespace std;
using namespace boost;
using namespace Deform;


void Deform::DeformAll_Utility::AxisAngleToMatrix( const real3_type& axisAngle, Geometry::Matrix33f& m )
{
	// Implementing Rodrigues's formula
	static real_type c1, c2, c3;
	static real_type t1,t2,t3,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20;
	c1 = axisAngle[0]; c2 = axisAngle[1], c3 = axisAngle[2];

	t10 = c3*c3;
	t11 = c2*c2;
	t12 = c1*c1;
	t18 = -t11-t12;
	t9 = t10-t18;

	if (t9 < 1e-7) {
		m(0,0) = 1.0;
		m(0,1) = -c3;
		m(0,2) = c2;
		m(1,0) = c3;
		m(1,1) = 1.0;
		m(1,2) = -c1;
		m(2,0) = -c2;
		m(2,1) = c1;
		m(2,2) = 1.0;
	}
	else {
		t13 = sqrt(t9);
		t20 = sin(t13)/t13;
		t19 = (1.0-cos(t13))/t9;
		t17 = c2*t20;
		t16 = c1*t20;
		t15 = c3*t19;
		t14 = c3*t20;
		t3 = c2*c1*t19;
		t2 = c1*t15;
		t1 = c2*t15;
		m(0,0) = 1.0+(-t10-t11)*t19;
		m(0,1) = -t14+t3;
		m(0,2) = t17+t2;
		m(1,0) = t14+t3;
		m(1,1) = 1.0+(-t10-t12)*t19;
		m(1,2) = -t16+t1;
		m(2,0) = -t17+t2;
		m(2,1) = t16+t1;
		m(2,2) = 1.0+t18*t19;
	}
}


/**
* Axis-angle conversion from matrix.  Assumes that the rotation is determinant +1.
* Grabbed from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToAngle/index.htm.
*/
void Deform::DeformAll_Utility::MatrixToAxisAngle( const Geometry::Matrix33f& R, real3_type &axis, real_type &angle )
{
	const real_type epsilon = 0.1f;
	const real_type eepsilon = 0.01f;
	const real_type ax = R(2,1)-R(1,2);
	const real_type ay = R(0,2)-R(2,0);
	const real_type az = R(1,0)-R(0,1);
	const real_type trace = R(0,0) + R(1,1) + R(2,2);

	// If the matrix is symmetric
	if ((abs(ax) < eepsilon) && (abs(ay) < eepsilon) && (abs(az) < eepsilon))
	{
		// If the matrix is identity
		if ((abs(R(2,1)+R(1,2)) < epsilon) &&
			(abs(R(0,2)+R(2,0)) < epsilon) &&
			(abs(R(1,0)+R(0,1)) < epsilon) &&
			(abs(trace - 3.0) < epsilon)) {
				axis[0] = 1.0;
				axis[1] = 0.0;
				axis[2] = 0.0;
				angle = 0.0;
		}
		else {
			// Otherwise, this matrix is a 180 degree rotation
			angle = M_PI;

			if (R(0,0) >= R(1,1)) {
				if (R(0,0) >= R(2,2)) {
					//00 is max diag
					axis[0] = sqrt(0.5 * (R(0,0)+1));
					axis[1] = R(0,1) * 0.5 / axis[0];
					axis[2] = R(0,2) * 0.5 / axis[0];
				} else {
					//22 is max
					axis[2] = sqrt(0.5 * (R(2,2)+1));
					axis[0] = R(0,2) * 0.5 / axis[2];
					axis[1] = R(1,2) * 0.5 / axis[2];
				}
			} else {
				if (R(1,1) >= R(2,2)) {
					//11 is max diag
					axis[1] = sqrt(0.5 * (R(1,1)+1));
					axis[0] = R(0,1) * 0.5 / axis[1];
					axis[2] = R(1,2) * 0.5 / axis[1];
				} else {
					//22 is max
					axis[2] = sqrt(0.5 * (R(2,2)+1));
					axis[0] = R(0,2) * 0.5 / axis[2];
					axis[1] = R(1,2) * 0.5 / axis[2];
				}
			}
		}
	}
	else {
		real_type s = sqrt(ax*ax + ay*ay + az*az);

		if (abs(s) < eepsilon)
			s = 1.0;

		s = 1.0 / s;
		axis[0] = ax * s;
		axis[1] = ay * s;
		axis[2] = az * s;
		angle = acos((trace - 1.0) / 2.0);
	}
}


void Deform::DeformAll_Utility::drawShapeValid( const PointList& vertex, const PointList& normal, const ShapeModel& originalShape, const std::vector<bool>& vertexValid, bool showInvalid )
{
	if (originalShape.shapeType == SHAPE_MESH) {
		const Mesh3_real& originalMesh = ((const Mesh3_real&)originalShape);

		glBegin(GL_TRIANGLES);
		for(size_type t = 0,t_end = originalMesh.triangle_vertex_index.size();t < t_end;t++) {
			const index3_type& tvi = originalMesh.triangle_vertex_index[t];
			//const index3_type& tni = originalMesh.triangle_normal_index[t];

			// Draw if all three vertices of the triangle are valid
			if (vertexValid[tvi[0]] && vertexValid[tvi[1]] && vertexValid[tvi[2]]) {
				for(size_type i = 0;i < 3;i++) {
					real3_type v = vertex[tvi[i]];
					//const real3_type& n = normal[tni[i]];
					real3_type n = normal[tvi[i]];

					glNormal3f(n[0],n[1],n[2]);
					glVertex3f(v[0],v[1],v[2]);
				}
			}
		}
		glEnd();

		if (showInvalid) {
			// Show the invalid triangles with a wireframe and dark shading
			setOpenGLMaterialColor(RealColor(0.0,0.0,0.0));
			glDisable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glBegin(GL_TRIANGLES);
			for(size_type t = 0,t_end = originalMesh.triangle_vertex_index.size();t < t_end;t++) {
				const index3_type& tvi = originalMesh.triangle_vertex_index[t];
				//const index3_type& tni = originalMesh.triangle_normal_index[t];

				// Draw if some vertex of the triangle is invalid
				if (!(vertexValid[tvi[0]] && vertexValid[tvi[1]] && vertexValid[tvi[2]])) {
					for(size_type i = 0;i < 3;i++) {
						real3_type v = vertex[tvi[i]];
						//const real3_type& n = normal[tni[i]];
						real3_type n = normal[tvi[i]];

						glNormal3f(n[0],n[1],n[2]);
						glVertex3f(v[0],v[1],v[2]);
					}
				}
			}
			glEnd();
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}
	}
}


void Deform::DeformAll_Utility::drawShapeWeights( const PointList& vertex, const PointList& normal, const ShapeModel& shape, const RefWeightList& weightList, const std::vector<bool> vertexValid, const std::vector<RealColor>& boneColors, real_type alpha, int showBone /*= -1 */ )
{
	if (shape.shapeType != SHAPE_MESH) {
		return;
	}
	const idx3List& triangle_vertex_index = ((Mesh3_real&)shape).triangle_vertex_index;

	// Draw triangle mesh, with bone weights painted on
	glBegin(GL_TRIANGLES);
	for(size_type t = 0,t_end = triangle_vertex_index.size(); t < t_end; ++t) {
		const index3_type& tvi = triangle_vertex_index[t];

		// Draw if all three vertices of the triangle are valid
		if (weightList[tvi[0]].size() && weightList[tvi[1]].size() && weightList[tvi[2]].size() &&
			vertexValid[tvi[0]] && vertexValid[tvi[1]] && vertexValid[tvi[2]]) {
				for (int i = 0; i < 3; ++i) {
					const real3_type& v = vertex[tvi[i]];
					const real3_type& n = normal[tvi[i]];

					// Find the color of the vertex
					idxList bones = weightList[tvi[i]].bones;
					distList weights = weightList[tvi[i]].weights;

					if (showBone < 0) {
						RealColor color(0.0,0.0,0.0);
						for (int j = 0; j < bones.size(); ++j) {
							color += weights[j] * boneColors[bones[j]];
						}
						glColor4f(color[0],color[1],color[2], alpha);
					}
					else {
						// Set default color for zero weight
						setOpenGLWeightColor(0.0, alpha);

						// Search for the bone if non-zero weight
						for (int j = 0; j < bones.size(); ++j) {
							if (showBone == bones[j]) {
								setOpenGLWeightColor(weights[j], alpha);
								break;
							}
						}
					}

					// Vertex position
					glNormal3f(n[0],n[1],n[2]);
					glVertex3f(v[0],v[1],v[2]);
				}
		}
	}
	glEnd();
}

void Deform::DeformAll_Utility::drawShapeWeightsValid( const PointList& vertex, const PointList& normal, const ShapeModel& mesh, const RefWeightList& shapeWeightList, const std::vector<bool>& validBit, const RealColor& shapeColor, const std::vector<RealColor>& compColors, const DrawOptions& drawOptions, bool showInvalid, bool enableLighting /*= false */ )
{
	real_type alpha = 1.0f;
	if (drawOptions.renderTransparent) {
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		alpha = 0.4f;
	}

	if (drawOptions.showShapeWeights && (shapeWeightList.size() == vertex.size())) {
		GLboolean lighting, colorMaterial;
		glGetBooleanv(GL_LIGHTING, &lighting);
		glGetBooleanv(GL_COLOR_MATERIAL, &colorMaterial);

		// Change lighting
		if (enableLighting) {
			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
		}
		else {
			glDisable(GL_LIGHTING);
			glDisable(GL_COLOR_MATERIAL);
		}

		drawShapeWeights(vertex, normal, mesh, shapeWeightList, validBit, compColors, alpha, drawOptions.showComponent);

		// Restore lighting state
		if (lighting) glEnable(GL_LIGHTING);
		else glDisable(GL_LIGHTING);
		if (colorMaterial) glEnable(GL_COLOR_MATERIAL);
		else glDisable(GL_COLOR_MATERIAL);
	}
	else {
		setOpenGLMaterialColor(shapeColor);
		if (vertex.size() == validBit.size()) {
			drawShapeValid(vertex, normal, mesh, validBit, showInvalid);
		}
		else {
			mesh.draw();
		}
	}
	if (drawOptions.renderTransparent) {
		glDisable(GL_CULL_FACE);
		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
	}
}

void Deform::DeformAll_Utility::Options::setFromGUI( const Options& options )
{
	// Keep initialization parameters
	real_type temp1 = this->saved_boundaryCorrespMaxDist_Multiplier;
	real_type temp2 = this->saved_maxCorrespDistance_Multiplier;
	real_type temp3 = this->saved_maxNormalAngleDifference;
	FloatArray temp4 = this->saved_TStepWeights;

	// Copy parameters
	(*this) = options;

	// Copy initialization parameters
	this->saved_boundaryCorrespMaxDist_Multiplier = temp1;
	this->saved_maxCorrespDistance_Multiplier = temp2;
	this->saved_maxNormalAngleDifference = temp3;
	this->saved_TStepWeights = temp4;
}

Deform::DeformAll_Utility::SE3_6DOF::SE3_6DOF() :
axisAngle(0.0,0.0,0.0), translation(0.0,0.0,0.0), rotationMatrix(Geometry::Matrix33f::identity())
{

}

Deform::DeformAll_Utility::SE3_6DOF::SE3_6DOF( const real3_type& axisAngle, const real3_type& translation ) :
axisAngle(axisAngle), translation(translation), rotationMatrix(Geometry::Matrix33f::identity())
{
	// Initialize the rotation matrix
	AxisAngleToMatrix(axisAngle, rotationMatrix);
}

Deform::DeformAll_Utility::SE3_6DOF::SE3_6DOF( const Geometry::Matrix33f& rotMatrix, const real3_type& translation ) :
axisAngle(0.0,0.0,0.0), translation(translation), rotationMatrix(rotMatrix)
{
	// Initialize the axis-angle
	real_type angle;
	MatrixToAxisAngle(rotationMatrix, axisAngle, angle);
	axisAngle = angle * axisAngle;
}

Deform::DeformAll_Utility::SE3_6DOF::SE3_6DOF( const real3_type& axisAngle, const Geometry::Matrix33f& axisAngleRotationMatrix, const real3_type& translation ) :
axisAngle(axisAngle), translation(translation), rotationMatrix(axisAngleRotationMatrix)
{

}

//const SE3_6DOF& Deform::DeformAll_Utility::SE3_6DOF::operator+=( const SE3_6DOF& rhs )
//{
//	// Update the axis angle and the translation
//	axisAngle += rhs.axisAngle;
//	translation += rhs.translation;
//
//	// Also update the rotation matrix
//	AxisAngleToMatrix(axisAngle, rotationMatrix);
//
//	return *this;
//}

Deform::DeformAll_Utility::SE3_6DOF Deform::DeformAll_Utility::SE3_6DOF::operator-() const
{
	Geometry::Matrix33f rot_transpose = rotationMatrix.transpose();
	real3_type invTranslation = -(rot_transpose * translation);
	return SE3_6DOF(-axisAngle, rot_transpose, invTranslation);
}

real3_type Deform::DeformAll_Utility::SE3_6DOF::applyLinearInverse( const real3_type& rhs ) const
{
	real3_type result = rhs - translation;
	result = result - (axisAngle ^ result);
	return result;
}

real3_type Deform::DeformAll_Utility::SE3_6DOF::applyLinearInverse_Normal( const real3_type& rhs ) const
{
	return rhs - (axisAngle ^ rhs);
}

real3_type Deform::DeformAll_Utility::SE3_6DOF::applyInverse( const real3_type& rhs ) const
{
	Geometry::Matrix33f rotTr = rotationMatrix.transpose();
	return rotTr * (rhs - translation);
}

real3_type Deform::DeformAll_Utility::SE3_6DOF::applyInverse_Normal( const real3_type& rhs ) const
{
	Geometry::Matrix33f rotTr = rotationMatrix.transpose();
	return rotTr * rhs;
}

// Code from Niloy Mitra to convert a linear rotation/translation to Plucker coordinates, then to a rigid transformation.
void Deform::DeformAll_Utility::SE3_6DOF::createFromLinear( const real3_type& linearRot, const real3_type& linearTr )
{
	real_type linearRot_length = linearRot.norm();

	if (linearRot_length < 1e-8){
		axisAngle = real3_type(0.0, 0.0, 0.0);
		rotationMatrix = Geometry::Matrix33f::identity();
		translation = linearTr;
	}
	else {
		real3_type axes_vec, moment_vec, axes_pt, move_vec;
		double screw_par, delta, tau, a0, a1, a2, a3;

		axes_vec      = linearRot / linearRot_length;
		screw_par     = (linearRot * linearTr) / linearRot.square_norm();

		moment_vec    = (linearTr - linearRot*screw_par) / linearRot_length;
		axes_pt       = axes_vec ^ moment_vec;

		delta         = atan(linearRot_length);
		tau           = delta / 2.0;
		a0            = cos(tau);
		a1            = axes_vec[0]*sin(tau);
		a2            = axes_vec[1]*sin(tau);
		a3            = axes_vec[2]*sin(tau);

		// Assign the rotation
		rotationMatrix(0,0) = a0*a0 + a1*a1 - a2*a2 - a3*a3;
		rotationMatrix(0,1) = 2*(a1*a2-a0*a3);
		rotationMatrix(0,2) = 2*(a1*a3+a0*a2);

		rotationMatrix(1,0) = 2*(a1*a2+a0*a3);
		rotationMatrix(1,1) = a0*a0 - a1*a1 + a2*a2 - a3*a3;
		rotationMatrix(1,2) = 2*(a2*a3-a0*a1);

		rotationMatrix(2,0) = 2*(a1*a3-a0*a2);
		rotationMatrix(2,1) = 2*(a2*a3+a0*a1);
		rotationMatrix(2,2) = a0*a0 - a1*a1 - a2*a2 + a3*a3;

		// Assign the translation
		move_vec = axes_vec * screw_par * delta;
		translation = move_vec + axes_pt - rotationMatrix * axes_pt;


		// Initialize the axis-angle
		real_type angle;
		MatrixToAxisAngle(rotationMatrix, axisAngle, angle);
		axisAngle = angle * axisAngle;
	}
}

//const SE3_6DOF Deform::DeformAll_Utility::operator+( const SE3_6DOF& lhs, const SE3_6DOF& rhs )
//{
//	SE3_6DOF result(lhs);
//	result += rhs;
//	return result;
//}

//Deform::DeformAll_Utility::TransformList Deform::DeformAll_Utility::operator+( const TransformList& tfList_left, const TransformList& tfList_right )
//{
//	if (tfList_left.size() != tfList_right.size()) {
//		cerr << "Error in TransformList operator+(): size of tfLists do not match!" << endl;
//		return TransformList();
//	}
//
//	TransformList result(tfList_left);
//	for (int i = 0; i < tfList_left.size(); ++i) {
//		result[i] += tfList_right[i];
//	}
//	return result;
//}

Deform::DeformAll_Utility::TransformList Deform::DeformAll_Utility::operator*( const TransformList& tfList_left, const TransformList& tfList_right )
{
	if (tfList_left.size() != tfList_right.size()) {
		cerr << "Error in TransformList operator+(): size of tfLists do not match!" << endl;
		return TransformList();
	}

	TransformList result(tfList_left.size());
	for (int i = 0; i < tfList_left.size(); ++i) {
		result[i] = tfList_left[i] * tfList_right[i];
	}
	return result;
}

Deform::DeformAll_Utility::TransformList Deform::DeformAll_Utility::operator-( const TransformList& tfList )
{
	TransformList result(tfList.size());
	for (int i = 0; i < tfList.size(); ++i) {
		result[i] = -tfList[i];
	}
	return result;
}

std::ostream& Deform::DeformAll_Utility::operator<<( std::ostream& out, const SE3_6DOF &tf ) {
	const real3_type& rot = tf.rot();
	const real3_type& tr = tf.tr();

	out << "[" << rot[0] << "," << rot[1] << "," << rot[2];
	out << " | " << tr[0] << "," << tr[1] << "," << tr[2] << "]";

	return out;
}

std::ostream& Deform::DeformAll_Utility::operator<<( std::ostream& out, const TransformList &tfList ) {
	for (int i = 0; i < tfList.size(); ++i) {
		if (i > 0) out << endl;
		out << "tf" << i << ": " << tfList[i];
	}
	return out;
}

Deform::DeformAll_Utility::ShapeDeformed::ShapeDeformed( const ShapeModel& shape ) :
originalShape(shape), vertex(NULL), normal(NULL)
{
	// Copy the vertex and normal information
	vertex = new PointList(originalShape.vertex);
	normal = new PointList(originalShape.normal);

	// Copy the bounding box information
	box_min = originalShape.box_min;
	box_max = originalShape.box_max;

	// Reset the valid bits
	vertexValid.resize(vertex->size(), true);
}

Deform::DeformAll_Utility::ShapeDeformed::~ShapeDeformed()
{
	// Clean up
	delete vertex;
	delete normal;
}

void Deform::DeformAll_Utility::ShapeDeformed::draw( bool showInvalid ) const
{
	drawShapeValid(*vertex, *normal, originalShape, vertexValid, showInvalid);
}

void Deform::DeformAll_Utility::ShapeDeformed::deform( const RefWeightList& weightList, const TransformList& tfList, bool inverse )
{
	size_type n_vertices = vertex->size();

	if (weightList.size() != n_vertices) {
		cerr << "Error in ShapeDeformed::deform(): weight list size doesn't match!" << endl;
		return;
	}

	// Initialize the bounding box for the mesh
	for (int d = 0; d < 3; d++) {
		box_min[d] = REAL_TYPE_MAX;
		box_max[d] = -REAL_TYPE_MAX;
	}

	// Deform using the precomputed weight list
	for (size_type i = 0; i < n_vertices; i++)
	{
		if (!weightList[i].valid()) {
			// Reset to original shape's vertex and normal
			(*vertex)[i] = originalShape.vertex[i];
			(*normal)[i] = originalShape.normal[i];

			// Mark this vertex as invalid, and move on to the next vertex
			vertexValid[i] = false;
			continue;
		}

		vertexValid[i] = true;

		// Deform vertex and normal
		lbs(originalShape.vertex[i], originalShape.normal[i], (*vertex)[i], (*normal)[i], weightList[i], tfList, inverse);

		//transformVertexNormal(weightList[i], tfList, originalShape.vertex[i], originalShape.normal[i], (*vertex)[i], (*normal)[i]);

		//if (i == 49) {
		//	// Display how it deforms
		//	cout << "Vertex " << i << ": " << originalShape.vertex[i] << endl;
		//	for(int j = 0; j < (int)weightList[i].size(); ++j) {
		//		cout << "*** Bone " << weightList[i].bones[j] << ": " << weightList[i].weights[j] << endl;
		//	}
		//	cout << "Final Position: " << (*vertex)[i] << endl;
		//}

		// Update bounding box
		for (int j = 0; j < 3; j++) {
			if ((*vertex)[i][j] < box_min[j]) {
				box_min[j] = (*vertex)[i][j];
			}
			if ((*vertex)[i][j] > box_max[j]) {
				box_max[j] = (*vertex)[i][j];
			}
		}
	}
}

void Deform::DeformAll_Utility::ShapeDeformed::writeGeometryOBJ( const char *file, bool flipNormals )
{
	// Check shape type
	if (originalShape.shapeType != SHAPE_MESH) {
		cerr << "Error in ShapeDeformed::writeGeometryOBJ(): Shape is not a mesh!" << endl;
		return;
	}

	// Check if the file already exists
	if (QFile::exists(file)) {
		cerr << "Error in ShapeDeformed::writeGeometryOBJ(): File " << file << " already exists." << endl;
		return;
	}

	ofstream out;
	out.open(file);

	if (out.fail()) {
		cerr << "Error in Mesh3_real::writeGeometryOBJ(): Could not open output file " << file << " for writing!" << endl;
		return;
	}

	cout << "Saving vertex and triangle data in " << file << " ...";

	// Make a list of valid vertex indices
	std::vector<int> translateIndex(vertex->size());
	int count = 0;
	for (int i = 0; i < vertex->size(); ++i) {
		if (vertexValid[i]) {
			translateIndex[i] = count;
			++count;
		}
		else {
			translateIndex[i] = -1;
		}
	}

	// Based on http://www.nada.kth.se/~asa/Ray/vertface2obj.m
	// Write vertex data
	for (index_type i = 0; i < vertex->size(); i++) {
		if (vertexValid[i]) {
			out << "v";
			if (flipNormals) {
				out << " " << (*vertex)[i][0];
				out << " " << (*vertex)[i][2];
				out << " " << (*vertex)[i][1];
			}
			else {
				out << " " << (*vertex)[i][0];
				out << " " << (*vertex)[i][1];
				out << " " << (*vertex)[i][2];
			}
			out << "\n";
		}
	}

	// Write triangle data
	for (index_type i = 0; i < ((Mesh3_real&)originalShape).triangle_vertex_index.size(); i++) {
		const index3_type& tri = ((Mesh3_real&)originalShape).triangle_vertex_index[i];
		if (vertexValid[tri[0]] && vertexValid[tri[1]] && vertexValid[tri[2]]) {
			out << "f";
			for (int c = 0; c < 3; c++) {
				out << " " << translateIndex[tri[c]]+1;
			}
			out << "\n";
		}
	}

	out.close();

	cout << "...done!" << endl;
}

Deform::DeformAll_Utility::RefWeight::RefWeight()
{
	// Do nothing for this constructor; use the default values.
}

bool Deform::DeformAll_Utility::RefWeight::findLabel( int &label ) const
{
	if (!valid()) {
		return false;
	}

	// Find the bone with max weight; if there's a tie, this will return the lowest label index
	real_type maxWeight = -REAL_TYPE_MAX;
	for (int i = 0; i < bones.size(); ++i) {
		if (weights[i] > maxWeight) {
			label = bones[i];
			maxWeight = weights[i];
		}
	}

	return true;
}



Deform::DeformAll_Utility::AllSampleGraph::AllSampleGraph( int k, real_type t ) : EKNNgraph_k(k), sampleDistThreshold(t)
{
	// Nothing for now!
}

void Deform::DeformAll_Utility::AllSampleGraph::initASG( const ShapeModel &shape, const idxList& sampleIndices, const int frameNumber )
{
	// Add these samples
	const int size = sampleIndices.size();
	vertex.resize(size);
	normal.resize(size);
	weightList.resize(size);
	fIndex.resize(size);
	sIndex.resize(size);
	for (int i = 0; i < size; ++i) {
		int addIdx = sampleIndices[i];

		vertex[i] = shape.vertex[addIdx];
		normal[i] = shape.normal[addIdx];
		weightList[i] = SparseWeight(1, IndexedWeight(0, 1.0));			/**< Assign just uniform weights for now */
		fIndex[i] = frameNumber;
		sIndex[i] = i;
	}
}

void Deform::DeformAll_Utility::AllSampleGraph::addSamples( const ShapeModel &shape, const idxList& newSampleIndices, const RefWeightList& newSampleWeights, const int newFrameIndex, const SparseTLArray& allTfList )
{
	// Check that newSampleIndices and newSampleWeights are the same size
	if (newSampleWeights.size() != newSampleIndices.size()) {
		cerr << "Error in addSamples(): # of weights must match # of indices!" << endl;
		return;
	}

	// Transform all existing samples to the new frame's pose
	PointList vertex_t, normal_t;
	getVertexNormal_Frame(newFrameIndex, allTfList, vertex_t, normal_t);

	// Pick samples that are far enough from the existing samples
	const PointList& vertex_t_const = vertex_t;
	idxList addedIndex;
	for (int i = 0; i < newSampleIndices.size(); ++i) {
		if (numSamples() <= EKNNgraph_k) {
			addedIndex.push_back(i);
		}
		else if (newSampleWeights[i].valid()) {
			idxList nearest;
			distList dists;
			vertex_t_const.queryKNN(shape.vertex[newSampleIndices[i]], 1, nearest, dists, true);		/**< Nearest existing sample? */

			// Compute distance using strategy from Pekelny & Gotsman (EG08):
			// project the distance ONTO the tangent plane at the nearest point.
			real3_type diff = shape.vertex[newSampleIndices[i]] - vertex_t_const[nearest[0]];
			real_type dist = normal_t[nearest[0]]*diff;
			dist = sqrt(diff*diff - dist*dist);

			// Add if the distance is far enough
			if (dist > sampleDistThreshold) {
				addedIndex.push_back(i);
			}
		}
	}


	// Add these samples
	const int currentSize = vertex_t.size();
	const int addedSize = addedIndex.size();
	vertex.resize(currentSize + addedSize);
	normal.resize(currentSize + addedSize);
	weightList.resize(currentSize + addedSize);
	fIndex.resize(currentSize + addedSize);
	sIndex.resize(currentSize + addedSize);
	//vertex_t.resize(currentSize + addedSize);
	//normal_t.resize(currentSize + addedSize);
	for (int i = 0; i < addedSize; ++i) {
		int addIdx = addedIndex[i];

		for (int j = 0; j < newSampleWeights[addIdx].size(); ++j) {
			IndexedWeight iw(newSampleWeights[addIdx].bones[j], newSampleWeights[addIdx].weights[j]);
			weightList[currentSize + i].push_back(iw);
		}
		vertex[currentSize + i] = shape.vertex[newSampleIndices[addIdx]];
		normal[currentSize + i] = shape.normal[newSampleIndices[addIdx]];
		fIndex[currentSize + i] = newFrameIndex;
		sIndex[currentSize + i] = addIdx;
		//vertex_t[currentSize + i] = newSamples[addIdx];
		//normal_t[currentSize + i] = shape.normal[newSampleIndices[addIdx]];
	}

	//// (Testing code below)
	//vertex_disp.swap(vertex_t);
}

void Deform::DeformAll_Utility::AllSampleGraph::resampleASG( const ShapeModel &shape, const idxList& newSampleIndices, const PointList& refSamples, const RefWeightList& refSampleWeights, const std::vector<bool>& isOccluded, const int newFrameIndex, const SparseTLArray& allTfList )
{
	// Check that newSampleIndices and newSampleWeights are the same size
	if (refSampleWeights.size() != refSamples.size()) {
		cerr << "Error in resampleASG(): # of ref weights must match # of ref samples!" << endl;
		return;
	}


	// Transform all existing samples to the new frame's pose
	PointList vertex_t, normal_t;
	getVertexNormal_Frame(newFrameIndex, allTfList, vertex_t, normal_t);

	// Transform the reference samples to the new frame
	SparseTLArray::const_iterator it = allTfList.find(newFrameIndex);
	if (it == allTfList.end()) {
		cerr << "Error in resampleASG(): Cannot find transformation for target frame " << newFrameIndex << endl;
		return;
	}
	const TransformList& tfList = *(it->second);
	PointList ref_t(refSamples.size());
	for (int i = 0; i < refSamples.size(); ++i) {
		lbs(refSamples[i], ref_t[i], refSampleWeights[i], tfList, true);
	}

	// Pick samples that are far enough from the existing samples
	idxList addedIndex;
	pickSamples_NoOverlap(vertex_t, normal_t, shape.vertex, shape.normal, newSampleIndices, addedIndex);

	// Add these samples
	const PointList& ref_t_const = ref_t;
	const int currentSize = vertex_t.size();
	const int addedSize = addedIndex.size();
	vertex.reserve(currentSize + addedSize);
	normal.reserve(currentSize + addedSize);
	fIndex.reserve(currentSize + addedSize);
	sIndex.reserve(currentSize + addedSize);
	vertex_t.reserve(currentSize + addedSize);
	normal_t.reserve(currentSize + addedSize);
	testIndex.clear();
	for (int i = 0; i < addedSize; ++i) {
		int addIdx = addedIndex[i];

		// Find the closest refSample
		idxList nearest;
		distList dists;
		ref_t_const.queryKNN(shape.vertex[newSampleIndices[addIdx]], 1, nearest, dists, true);
		const RefWeight& nearestWeightVal = refSampleWeights[nearest[0]];

		// Check for occlusion
		SparseWeight sp_nearestWeightVal(nearestWeightVal.size());
		bool occluded = false;
		for (int j = 0; j < nearestWeightVal.size(); ++j) {
			if (isOccluded[nearestWeightVal.bones[j]]) {
				occluded = true;
				break;
			}
			sp_nearestWeightVal[j].index = nearestWeightVal.bones[j];
			sp_nearestWeightVal[j].weight = nearestWeightVal.weights[j];
		}

		if (!occluded) {
			// Assign to the new sample
			weightList.push_back(sp_nearestWeightVal);
			vertex.add(shape.vertex[newSampleIndices[addIdx]]);
			normal.add(shape.normal[newSampleIndices[addIdx]]);
			fIndex.push_back(newFrameIndex);
			sIndex.push_back(addIdx);

			// (Testing code below)
			vertex_t.add(shape.vertex[newSampleIndices[addIdx]]);
			normal_t.add(shape.normal[newSampleIndices[addIdx]]);
			testIndex.push_back(vertex_t.size()-1);
		}
	}

	// (Testing code below)
	vertex_disp.swap(vertex_t);
}

void Deform::DeformAll_Utility::AllSampleGraph::rebuild( const ShapeModel &shape, const idxList& newSampleIndices, const int newFrameIndex, const std::vector<bool>& isOccluded, const std::vector<PointList>& refSamples_Bone, const std::vector<PointList>& refNormals_Bone, const SparseTLArray& allTfList, real_type minDist, bool accumulateOnRef, int refFrameIndex )
{
	// Transform all existing samples to the new frame's pose
	PointList vertex_t, normal_t;
	getVertexNormal_Frame(newFrameIndex, allTfList, vertex_t, normal_t);

	// Transform each bone's reference samples to the new frame
	SparseTLArray::const_iterator it = allTfList.find(newFrameIndex);
	if (it == allTfList.end()) {
		cerr << "Error in resampleASG(): Cannot find transformation for target frame " << newFrameIndex << endl;
		return;
	}
	const TransformList& tfList = *(it->second);
	const int numBones = tfList.size();


	// Pick samples that are far enough from the existing samples
	idxList farEnough;
	pickSamples_NoOverlap(vertex_t, normal_t, shape.vertex, shape.normal, newSampleIndices, farEnough);

	// Extrapolate weights using the existing samples
	std::vector<int> newLabels;
	extrapolateWeights(refSamples_Bone, refNormals_Bone, isOccluded, shape.vertex, newSampleIndices, tfList, minDist, newLabels);


	// Add new samples that are far enough and have extrapolated weights
	const int currentSize = vertex_t.size();
	const int addedSize = farEnough.size();
	vertex.reserve(currentSize + addedSize);
	normal.reserve(currentSize + addedSize);
	fIndex.reserve(currentSize + addedSize);
	sIndex.reserve(currentSize + addedSize);
	vertex_t.reserve(currentSize + addedSize);
	normal_t.reserve(currentSize + addedSize);
	testIndex.clear();
	for (int i = 0; i < addedSize; ++i) {
		int addIdx = farEnough[i];
		const real3_type& newPt = shape.vertex[newSampleIndices[addIdx]];
		const real3_type& newNormal = shape.normal[newSampleIndices[addIdx]];

		// Check extrapolated label
		if (newLabels[addIdx] >= 0) {
			DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(newLabels[addIdx], 1.0));

			// Assign to the new sample
			weightList.push_back(sw);

			if (accumulateOnRef) {
				// Transform samples to reference
				real3_type refPt, refNormal;
				lbs(newPt, newNormal, refPt, refNormal, sw, tfList, false);

				vertex.add(refPt);
				normal.add(refNormal);
				fIndex.push_back(refFrameIndex);
				sIndex.push_back(-1);
			}
			else {
				vertex.add(newPt);
				normal.add(newNormal);
				fIndex.push_back(newFrameIndex);
				sIndex.push_back(addIdx);
			}

			//// (Testing code below)
			//vertex_t.add(newPt);
			//normal_t.add(newNormal);
			//testIndex.push_back(vertex_t.size()-1);
		}
	}

	//// (Testing code below)
	//vertex_disp.swap(vertex_t);
}

void Deform::DeformAll_Utility::AllSampleGraph::pickSamples_NoOverlap( const PointList& existingPts, const PointList& existingNormals, const PointList& newPoints, const PointList& newNormals, const idxList& subset, idxList& pickedResult ) const
{
	const int numExisting = existingPts.size();

	// Pick samples that are far enough from the existing samples
	idxList addedIndex;
	for (int i = 0, i_end = subset.size(); i < i_end; ++i) {
		if (numExisting <= EKNNgraph_k) {
			addedIndex.push_back(i);
		}
		else {
			idxList nearest;
			distList dists;
			existingPts.queryKNN(newPoints[subset[i]], 1, nearest, dists, true);		/**< Nearest existing sample? */

			// Compute distance using strategy from Pekelny & Gotsman (EG08):
			// project the distance ONTO the tangent plane at the nearest point.
			real3_type diff = newPoints[subset[i]] - existingPts[nearest[0]];
			real_type distToPl = abs(existingNormals[nearest[0]]*diff);
			real_type distOnPl = sqrt(diff*diff - distToPl*distToPl);
			real_type distPt = sqrt(diff*diff);
			real_type cosAngle = newNormals[subset[i]] * existingNormals[nearest[0]];

			// Compare angles between the normals
			if (cosAngle < 0 && distPt > sampleDistThreshold) {
				addedIndex.push_back(i);
			}
			else if (cosAngle > 0 && distOnPl > sampleDistThreshold) {
				addedIndex.push_back(i);
			}
		}
	}

	pickedResult.swap(addedIndex);
}


real_type outlierT( const distList& scores, idxList& sortIndex )
{
	// Sort indices into scores in descending order
	const int numScores = scores.size();
	sortIndex.resize(numScores);
	for (int i = 0; i < numScores; ++i) {
		sortIndex[i] = i;
	}
	std::sort(sortIndex.begin(), sortIndex.end(), make_ForwardCompare(scores, comp_desc<real_type>()));

	// Find the outlier threshold
	real_type maxScore = scores[sortIndex[0]];
	real_type Ml = scores[sortIndex[floor(3.0*((double)numScores) / 4.0)]];
	real_type Mu = scores[sortIndex[floor(((double)numScores) / 4.0)]];
	//return (Mu + 1.5*(Mu - Ml));								/**< Moderate outliers */
	//return (Mu + 3.0*(Mu - Ml));								/**< Significant outliers */
	return 3.0*Mu;
}

void extrapolateWeights( const std::vector<PointList>& refSamples_Bone, const std::vector<PointList>& refNormals_Bone, const std::vector<bool>& isOccluded, const PointList& newPts, const idxList& sampleIdx, const Deform::DeformAll_Utility::TransformList& tfList, real_type minDist, std::vector<int>& label, std::vector<distList> *storeScores /*= NULL */ )
{
	const int numBones = tfList.size();
	std::vector<PointList> refV_t(numBones), refN_t(numBones);
	for (int i = 0; i < numBones; ++i) {
		DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(i, 1.0));
		const PointList& rbV = refSamples_Bone[i];
		const PointList& rbN = refNormals_Bone[i];
		PointList pts(rbV.size()), normals(rbN.size());

		for (int j = 0; j < rbV.size(); ++j) {
			lbs(rbV[j], rbN[j], pts[j], normals[j], sw, tfList, true);
		}

		// Add to the list
		refV_t[i].swap(pts);
		refN_t[i].swap(normals);
	}

	//for (int i = 0; i < numBones; ++i) {
	//	cout << "-- Group " << i << ": " << refV_t[i].size() << endl;
	//}

	// Extrapolate labels from the existing sample set
	label.resize(sampleIdx.size());
	if (storeScores) {
		storeScores->resize(sampleIdx.size());
	}
	for (int i = 0, i_end = sampleIdx.size(); i < i_end; ++i) {
		int addIdx = sampleIdx[i];
		const real3_type& newPt = newPts[addIdx];

		// Find the closest distance to each bone
		distList scores(numBones);
		real_type scoreSum = 0.0, minDistPt = REAL_TYPE_MAX;
		for (int j = 0, j_end = numBones; j < j_end; ++j) {
			const PointList& bonePts = refV_t[j];
			const PointList& boneNormals = refN_t[j];

			if (bonePts.size()) {
				idxList nearest;
				distList dists;
				bonePts.queryKNN(newPt, 1, nearest, dists, true);

				// Compute distance using strategy from Pekelny & Gotsman (EG08):
				// combine pt-pt and pt-pl distance.
				real3_type diff = newPt - bonePts[nearest[0]];
				real_type distPt = sqrt(diff*diff);

				if (distPt < REAL_TYPE_EPSILON) {
					scores[j] = REAL_TYPE_INF;
				}
				else {
					//real_type distPl = abs(boneNormals[nearest[0]]*diff);
					//scores[j] = 1.0 / (0.8*distPl + 0.2*distPt);
					scores[j] = 1.0 / distPt;
					scoreSum += scores[j];
				}

				// Keep track of the minimum distPt
				if (distPt < minDistPt) {
					minDistPt = distPt;
				}
			}
			else {
				scores[j] = -REAL_TYPE_INF;
			}
		}
		if (scoreSum < REAL_TYPE_EPSILON) {
			scoreSum = 1.0;
		}

		// Normalize scores
		for (int j = 0, j_end = numBones; j < j_end; ++j) {
			scores[j] = scores[j] / scoreSum;
		}
		// Determine if the max score is significant (an outlier)
		idxList sortIndex(numBones);
		real_type thres = outlierT(scores, sortIndex);
		int maxBone = sortIndex[0];

		// Check confidence threshold
		//if (isOccluded[maxBone] || (minDistPt > minDist && scores[maxBone] < thres)) {
		if (isOccluded[maxBone] || scores[maxBone] < thres) {
			label[i] = -1;
		}
		else {
			label[i] = maxBone;
		}

		// Optional: store scores
		if (storeScores) {
			(*storeScores)[i].swap(scores);
		}
	}
}

void Deform::DeformAll_Utility::AllSampleGraph::pruneEdges( const Deform::DeformAll_Utility::SparseTLArray& allTfList, real_type maxStretchFactor, real_type maxEdgeLength, const DeformGraphAll_Utility::JointMap *jointMap )
{
	std::list< std::pair<int,int> > removeEdgeList;

	PointList vRef, nRef;
	getVertexNormal_Ref(allTfList, vRef, nRef);

	const int numBones = allTfList.begin()->second->size();
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		int vi1 = index[source(*ei,graph)];
		int vi2 = index[target(*ei,graph)];
		const SparseWeight& w1 = weightList[vi1];
		const SparseWeight& w2 = weightList[vi2];
		const real3_type& pt1 = vRef[vi1];
		const real3_type& pt2 = vRef[vi2];
		bool noJoint = true;
		if (jointMap) {
			int i = findLabel(w1);
			int j = findLabel(w2);
			// Check if there is no valid joint between the labels
			int lidx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones, min(i,j), max(i,j));
			noJoint = (jointMap->find(lidx) == jointMap->end());
		}

		// Check if vertices have different labels
		if (w1 != w2 && noJoint) {
			real3_type x1, x2;

			// Min and max edge lengths
			real_type minLength = REAL_TYPE_MAX, maxLength = -REAL_TYPE_MAX;
			Deform::DeformAll_Utility::SparseTLArray::const_iterator it;
			for (it = allTfList.begin(); it != allTfList.end(); ++it) {
				lbs(pt1, x1, w1, *(it->second), true);
				lbs(pt2, x2, w2, *(it->second), true);

				real_type len = (x1 - x2).norm();
				if (len < minLength) {
					minLength = len;
				}
				if (len > maxLength) {
					maxLength = len;
				}
			}

			// Exceeds threshold?
			real_type stretch = maxLength / minLength;
			if (stretch > maxStretchFactor) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
			else if (maxLength > maxEdgeLength) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
		}
		else {
			real_type len = (pt1 - pt2).norm();
			if (len > maxEdgeLength) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
		}
	}

	// Remove the edges
	std::list< std::pair<int,int> >::const_iterator it;
	for (it = removeEdgeList.begin(); it != removeEdgeList.end(); ++it) {
		BoostVertex v1 = boost::vertex(it->first,graph);
		BoostVertex v2 = boost::vertex(it->second,graph);
		BoostEdge e;
		bool temp;
		tie(e,temp) = edge(v1,v2,graph);
		if (temp) {
			remove_edge(e,graph);
		}
		else {
			cerr << "Unable to find edge (" << it->first << "," << it->second << ") in graph!" << endl;
		}
	}

	// Testing
	testList.swap(removeEdgeList);
}

void Deform::DeformAll_Utility::AllSampleGraph::getVertexNormal_Ref( const SparseTLArray& allTfList, PointList& vList, PointList& nList ) const
{
	vList.resize(numSamples());
	nList.resize(numSamples());

	for (int i = 0, i_end = numSamples(); i < i_end; ++i) {
		if (!weightList[i].size()) {
			cerr << "Error in getVertexNormal_Ref(): No weights for sample " << i << endl;
			return;
		}
		SparseTLArray::const_iterator it = allTfList.find(fIndex[i]);
		if (it == allTfList.end()) {
			cerr << "Error in getVertexNormal_Ref(): Cannot find transformation for vertex " << i << ", frame " << fIndex[i] << endl;
			return;
		}

		// Transform the sample to the reference
		lbs(vertex[i], normal[i], vList[i], nList[i], weightList[i], *(it->second), false);
	}
}

void Deform::DeformAll_Utility::AllSampleGraph::getVertexNormal_Frame( const int frameIndex, const SparseTLArray& allTfList, PointList& vList, PointList& nList ) const
{
	vList.resize(numSamples());
	nList.resize(numSamples());

	// Find the transformations of the specific frame to transform to
	SparseTLArray::const_iterator it = allTfList.find(frameIndex);
	if (it == allTfList.end()) {
		cerr << "Error in getVertexNormal_Frame(): Cannot find transformation for target frame " << frameIndex << endl;
		return;
	}
	const TransformList& tfList = *(it->second);

#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: pre-multiplication of transform list
	SparseTLArray preMultTf;
	for (it = allTfList.begin(); it != allTfList.end(); ++it)
	{
		const int frameIndex = it->first;
		TransformList* newMultiple = new TransformList();
		*newMultiple = (-tfList) * *(it->second);
		preMultTf[frameIndex] = newMultiple;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif

	// Transform each sample
	for (int i = 0, i_end = numSamples(); i < i_end; ++i) {
		if (!weightList[i].size()) {
			cerr << "Error in getVertexNormal_Frame(): No weights for sample " << i << endl;
			return;
		}

		if (frameIndex == fIndex[i]) {
			vList[i] = vertex[i];
			nList[i] = normal[i];
		}
		else {
#ifdef OPT_TF_PREMULT
			// Using pre-multiplication
			it = preMultTf.find(fIndex[i]);
			if (it == preMultTf.end()) {
				cerr << "Error in findClosestPoints(): Cannot find transformation for sample " << i << ", frame " << fIndex[i] << endl;
				break;
			}
			if (TEST_COMP_TYPE == 1) {
				cerr << "Bad option!!" << endl;
				break;
			}
			else if (TEST_COMP_TYPE == 2) {
				lbs(vertex[i], normal[i], vList[i], nList[i], weightList[i], *(it->second), false);
			}
#else
			it = allTfList.find(fIndex[i]);
			if (it == allTfList.end()) {
				cerr << "Error in getVertexNormal_Frame(): Cannot find transformation for vertex " << i << ", frame " << fIndex[i] << endl;
				return;
			}
			// Transform the sample to the reference, then to the specified frame
			if (TEST_COMP_TYPE == 1) {
				real3_type tempV, tempN;
				lbs(vertex[i], normal[i], tempV, tempN, weightList[i], *(it->second), false);
				lbs(tempV, tempN, vList[i], nList[i], weightList[i], tfList, true);
			}
			else if (TEST_COMP_TYPE == 2) {
				DeformAll_Utility::TransformList temp = (-tfList) * *(it->second);
				lbs(vertex[i], normal[i], vList[i], nList[i], weightList[i], temp, false);
			}
#endif
		}
	}

#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: delete pre-multiplication of transform list
	for (it = preMultTf.begin(); it != preMultTf.end(); ++it)
	{
		delete it->second;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif
}

float MemoryToMB(int bytes)
{
	return ((float) bytes) / 1024.f / 1024.f;
}

DWORD CalculateWSPrivate(DWORD processID, const SYSTEM_INFO &sysInfo)
{
	// Adapted from: http://www.codeproject.com/KB/cpp/XPWSPrivate.aspx
	// and http://social.msdn.microsoft.com/Forums/en-US/windowssdk/thread/6f79156f-1087-415e-ab79-8a3a9d063527
	HANDLE hProcess = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, processID );
	if ( !hProcess )
	{
		cout << "Cannot open process " << processID << " for information!" << endl;
		return 0;
	}

	PSAPI_WORKING_SET_INFORMATION wsTemp;
	PSAPI_WORKING_SET_INFORMATION* pWSInfo = NULL;
	BYTE* pByte = NULL;

	memset (&wsTemp, 0, sizeof(PSAPI_WORKING_SET_INFORMATION));

	BOOL bRet = QueryWorkingSet( hProcess, (PVOID)&wsTemp, sizeof(PSAPI_WORKING_SET_INFORMATION));

	if ((bRet == FALSE) && (GetLastError() == ERROR_BAD_LENGTH))
	{
		// wychang: the "+ 1000" is a fudge factor..
		size_t stSize = sizeof(PSAPI_WORKING_SET_INFORMATION) + ((wsTemp.NumberOfEntries - 1 + 1000) * sizeof(PSAPI_WORKING_SET_BLOCK));
		pByte = new BYTE[stSize];
		memset (pByte, 0, stSize);
		pWSInfo = (PSAPI_WORKING_SET_INFORMATION*)pByte;

		bRet = QueryWorkingSet( hProcess, (PVOID)pWSInfo, stSize );
		//cout << "Querying working set information with " << stSize << " bytes, returned " << bRet << endl;
	}
	else
	{
		pWSInfo = &wsTemp;
	}

	CloseHandle( hProcess );

	DWORD dSharedPages = 0;
	DWORD dPrivatePages = 0; 
	DWORD dPageTablePages = 0;

	if (bRet != FALSE)
	{
		for ( DWORD i = 0; i < pWSInfo->NumberOfEntries; i++ )
		{
			if (pWSInfo->WorkingSetInfo[i].Shared)
			{
				++dSharedPages;
			}
			else // private one
			{
				++dPrivatePages;
			}
		}
	}
	else
	{
		cout << "Error while querying working set for " << processID << endl;
	}

	delete [] pByte;
	pByte = NULL;
	pWSInfo = NULL;

	return (sysInfo.dwPageSize * dPrivatePages);
}

// Linear blend skinning: Transform point
void lbs(const real3_type& x, real3_type& x_out,
		 const Deform::DeformAll_Utility::SparseWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse )
{
	const int nBones = tfList.size();

	if (weights.size() == 1) {
		int j = weights[0].index;

		if (inverse) {
			Geometry::Matrix33f rotTransposed = tfList[j].rotMatrix().transpose();
			x_out = rotTransposed * (x - tfList[j].tr());
		}
		else {
			x_out = tfList[j] * x;
		}
	}
	else if (weights.size() > 1) {
		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < weights.size(); ++i) {
			tf.rot += weights[i].weight * tfList[weights[i].index].rotMatrix();
			tf.tr += weights[i].weight * tfList[weights[i].index].tr();
		}

		// Apply transformation (apply inverse transpose to the normal)
		if (inverse) {
			invert(tf.rot);
			x_out = tf.rot * (x - tf.tr);
		}
		else {
			x_out = tf * x;
		}
	}
}

// Linear blend skinning: Transform point and normal
void lbs(const real3_type& x, const real3_type& n, real3_type& x_out, real3_type& n_out,
		 const Deform::DeformAll_Utility::SparseWeight& weights, const Deform::DeformAll_Utility::TransformList& tfList, bool inverse )
{
	const int nBones = tfList.size();

	if (weights.size() == 1) {
		int j = weights[0].index;

		if (inverse) {
			Geometry::Matrix33f rotTransposed = tfList[j].rotMatrix().transpose();
			x_out = rotTransposed * (x - tfList[j].tr());
			n_out = rotTransposed * n;
		}
		else {
			x_out = tfList[j] * x;
			n_out = tfList[j].rotMatrix() * n;
		}
	}
	else if (weights.size() > 1) {
		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < weights.size(); ++i) {
			tf.rot += weights[i].weight * tfList[weights[i].index].rotMatrix();
			tf.tr += weights[i].weight * tfList[weights[i].index].tr();
		}

		// Apply transformation (apply inverse transpose to the normal)
		if (inverse) {
			n_out = tf.rot.transpose() * n;
			invert(tf.rot);
			x_out = tf.rot * (x - tf.tr);
		}
		else {
			x_out = tf * x;
			invert(tf.rot);
			n_out = tf.rot.transpose() * n;
		}
		n_out.normalize();
	}
}

void weightedColor( const Deform::DeformAll_Utility::SparseWeight& weights, const std::vector<RealColor> &colors, RealColor& result, int showComp )
{
	result[0] = result[1] = result[2] = 0.0;

	if (showComp < 0) {
		for (int i = 0, i_end = weights.size(); i < i_end; ++i) {
			result += weights[i].weight * colors[weights[i].index];
		}
	}
	else {
		// White if it's not the selected component, black otherwise
		if (findLabel(weights) != showComp) {
			result[0] = result[1] = result[2] = 1.0;
		}
	}
}

void makeEKNNgraph( int k, const PointList& points, BoostGraph& EKNNgraph )
{
	const int n = points.size();

	cout << "*** Constructing Euclidean-KNN graph for " << n << " points...";

	// 1. Construct the Euclidean k-nearest neighbor graph
	std::deque< std::list<int> > tempAdjList(n);
	for (int i = 0; i < n; ++i) {
		// For each point, find the k-nearest neighbors
		const real3_type& q = points[i];
		idxList nearest;
		distList dists;
		points.queryKNN(q, k, nearest, dists, true);

		// Add edges in the graph
		for (int j = 0; j < nearest.size(); ++j) {
			int idx = nearest[j];

			// Add both directions of the edge, avoid duplicating this vertex
			if (i != idx) {
				tempAdjList[i].push_back(idx);
				tempAdjList[idx].push_back(i);
			}
		}
	}

	// 2. Graph post-processing: remove duplicate edges and add to graph structure
	EKNNgraph = BoostGraph(n);
	for (int i = 0; i < n; ++i) {
		std::list<int>& adjList = tempAdjList[i];

		// Sort & remove consecutive duplicate values
		adjList.sort();
		adjList.unique();

		// Add all unique edges to the graph
		std::list<int>::const_iterator it;
		for (it = adjList.begin(); it != adjList.end(); ++it) {
			if (*it > i) {
				add_edge(i, *it, (points[i] - points[*it]).norm(), EKNNgraph);
			}
		}
	}


	//// Testing: accessing adjacent vertices
	//BoostIndexMap EKNNindex = get(vertex_index, EKNNgraph);
	//BoostVertexIterator vi_start, vi;
	//tie(vi_start, vi) = vertices(EKNNgraph);

	//for (int i = 0; i < n; ++i) {
	//	std::list<int>& adjList = tempAdjList[i];

	//	// Print out all values
	//	//cerr << "*** Adjacency of vertex " << i << ": [";
	//	std::list<int>::const_iterator it;
	//	for (it = adjList.begin(); it != adjList.end(); ++it)
	//	{
	//		if (it != adjList.begin()) cerr << " ";
	//		cerr << *it;
	//	}
	//	cerr << "]" << endl;


	//	const BoostVertex& v = *(vi_start + i);
	//	//cerr << "*** Adjacency of vertex " << EKNNindex[v] << ": [";
	//	BoostAdjacencyIterator ai, ai_start, ai_end;
	//	tie(ai_start, ai_end) = adjacent_vertices(v, EKNNgraph);
	//	for (ai = ai_start; ai != ai_end; ++ai)
	//	{
	//		if (ai != ai_start) cerr << " ";
	//		cerr << EKNNindex[*ai];
	//	}
	//	cerr << "]" << endl;

	//	if (i != EKNNindex[v]) {
	//		cerr << "Index mismatch!" << endl;
	//	}
	//}

	cout << " done!" << endl;
}

void drawBoostGraph( const BoostGraph& graph, const PointList& points )
{
	// Check sizes
	if (num_vertices(graph) != points.size()) {
		cerr << "Graph sizes (" << num_vertices(graph) << "!=" << points.size() << ") does not match!" << endl;
		return;
	}

	// 1. Draw the points
	PointList_Utility::draw(points, POINT_SQUARES);

	// 2. Draw the edges (uniquely!)
	glBegin(GL_LINES);
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		const real3_type& pt1 = points[index[source(*ei,graph)]];
		const real3_type& pt2 = points[index[target(*ei,graph)]];

		glVertex3f(pt1[0], pt1[1], pt1[2]);
		glVertex3f(pt2[0], pt2[1], pt2[2]);
	}
	glEnd();
}

void drawBoostGraph( const BoostGraph& graph, const PointList& points, const PointList& normals )
{
	// Check sizes
	if (num_vertices(graph) != points.size()) {
		cerr << "Graph sizes (" << num_vertices(graph) << "!=" << points.size() << ") does not match!" << endl;
		return;
	}

	// 1. Draw the points
	glBegin(GL_POINTS);
	for (index_type i = 0; i < points.size(); i++) {
		glNormal3f(normals[i][0], normals[i][1], normals[i][2]);
		glVertex3f(points[i][0], points[i][1], points[i][2]);
	}
	glEnd();

	// 2. Draw the edges (uniquely!)
	glBegin(GL_LINES);
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		const real3_type& pt1 = points[index[source(*ei,graph)]];
		const real3_type& pt2 = points[index[target(*ei,graph)]];
		const real3_type& n1 = normals[index[source(*ei,graph)]];
		const real3_type& n2 = normals[index[target(*ei,graph)]];

		glVertex3f(pt1[0], pt1[1], pt1[2]);
		glVertex3f(n1[0], n1[1], n1[2]);
		glVertex3f(pt2[0], pt2[1], pt2[2]);
		glVertex3f(n2[0], n2[1], n2[2]);
	}
	glEnd();
}

void drawWeightedBoostGraph( const BoostGraph& graph, const PointList& points, const Deform::DeformAll_Utility::SparseWeightList& weightList, const std::vector<RealColor>& colors, int showComp )
{
	// Check sizes
	if (num_vertices(graph) != points.size()) {
		cerr << "Graph sizes (" << num_vertices(graph) << "!=" << points.size() << ") does not match!" << endl;
		return;
	}

	// 1. Draw the points
	RealColor temp1, temp2;
	glBegin(GL_POINTS);
	for (index_type i = 0; i < points.size(); i++) {
		weightedColor(weightList[i], colors, temp1, showComp);
		setOpenGLColor(temp1);
		glVertex3f(points[i][0], points[i][1], points[i][2]);
	}
	glEnd();

	// 2. Draw the edges (uniquely!)
	glBegin(GL_LINES);
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		const int i1 = index[source(*ei,graph)];
		const int i2 = index[target(*ei,graph)];
		const real3_type& pt1 = points[i1];
		const real3_type& pt2 = points[i2];
		weightedColor(weightList[i1], colors, temp1, showComp);
		weightedColor(weightList[i2], colors, temp2, showComp);

		setOpenGLColor(temp1);
		glVertex3f(pt1[0], pt1[1], pt1[2]);
		setOpenGLColor(temp2);
		glVertex3f(pt2[0], pt2[1], pt2[2]);
	}
	glEnd();
}

void drawWeightedBoostGraph( const BoostGraph& graph, const PointList& points, const PointList& normals, const Deform::DeformAll_Utility::SparseWeightList& weightList, const std::vector<RealColor>& colors, int showComp )
{
	// Check sizes
	if (num_vertices(graph) != points.size()) {
		cerr << "Graph sizes (" << num_vertices(graph) << "!=" << points.size() << ") does not match!" << endl;
		return;
	}

	// 1. Draw the points
	RealColor temp1, temp2;
	glBegin(GL_POINTS);
	for (index_type i = 0; i < points.size(); i++) {
		weightedColor(weightList[i], colors, temp1, showComp);
		setOpenGLColor(temp1);
		glNormal3f(normals[i][0], normals[i][1], normals[i][2]);
		glVertex3f(points[i][0], points[i][1], points[i][2]);
	}
	glEnd();

	// 2. Draw the edges (uniquely!)
	glBegin(GL_LINES);
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		const int i1 = index[source(*ei,graph)];
		const int i2 = index[target(*ei,graph)];
		const real3_type& pt1 = points[i1];
		const real3_type& pt2 = points[i2];
		const real3_type& n1 = normals[i1];
		const real3_type& n2 = normals[i2];
		weightedColor(weightList[i1], colors, temp1, showComp);
		weightedColor(weightList[i2], colors, temp2, showComp);

		setOpenGLColor(temp1);
		glNormal3f(n1[0], n1[1], n1[2]);
		glVertex3f(pt1[0], pt1[1], pt1[2]);
		setOpenGLColor(temp2);
		glNormal3f(n2[0], n2[1], n2[2]);
		glVertex3f(pt2[0], pt2[1], pt2[2]);
	}
	glEnd();
}

void makeAllTFList( const Deform::DeformGraphAll_Utility::SolutionTable& frameSolution, Deform::DeformAll_Utility::SparseTLArray& allTfList )
{
	allTfList.clear();

	Deform::DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		Deform::DeformGraphAll_Utility::FrameSolution* frameSoln = it->second;
		allTfList[it->first] = &(frameSoln->tfList);
	}
}

int findLabel( const Deform::DeformAll_Utility::SparseWeight& weight )
{
	if (weight.size() == 0) {
		return -1;
	}
	else {
		// Return index corresponding to maximum weight
		int maxIndex = weight[0].index;
		real_type maxWeight = weight[0].weight;

		for (int i = 1, i_end = weight.size(); i < i_end; ++i) {
			if (weight[i].weight > maxWeight) {
				maxIndex = weight[i].index;
				maxWeight = weight[i].weight;
			}
		}

		return maxIndex;
	}
}

void printAllTFList( const Deform::DeformAll_Utility::SparseTLArray& allTfList )
{
	DeformAll_Utility::SparseTLArray::const_iterator it;
	for (it = allTfList.begin(); it != allTfList.end(); ++it) {
		cout << DeformGraphAll_Utility::printHeader << "Frame " << it->first << endl;
		const DeformAll_Utility::TransformList& tfList = *(it->second);
		for (int i = 0; i < tfList.size(); ++i) {
			cout << DeformGraphAll_Utility::printHeader << "-- " << i << " rot " << tfList[i].rot() << endl;
			cout << DeformGraphAll_Utility::printHeader << "-- " << i << " tr  " << tfList[i].tr() << endl;
		}
	}
}

Deform::DeformAll_Utility::TfVarTranslate::TfVarTranslate( const DeformGraphAll_Utility::SolutionTable& frameSolution, int referenceFrameIndex, const std::vector<int>& excludedLabels ) :
originalNumBonesPerFrame(frameSolution.begin()->second->tfList.size())
{
	//////////////////////////////////////////////////////////////////////////
	// Initialize transformation linear indices for a table of solutions
	//////////////////////////////////////////////////////////////////////////

	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int frameIndex = it->first;
		const DeformGraphAll_Utility::FrameSolution& frameSoln = *(it->second);
		int frameNumBones = frameSoln.tfList.size();
		std::vector<int> linearIndices(frameNumBones);

		// Take into account excluded and occluded bones in this frame
		std::vector<bool> frameBoneExcluded = frameSoln.isBoneOccludedOrExcluded(excludedLabels);

		// Issue a warning if the number of bones don't match up
		if (frameNumBones != originalNumBonesPerFrame) {
			cerr << "Warning in TfVarTranslate(): number of bones in frame " << frameIndex << " doesn't match with first frame!" << endl;
		}

		for (int i = 0; i < frameNumBones; ++i) {
			// Check if this frame is not activated, or if the bone is excluded.
			// --->>> then don't optimize for the transformation
			if (frameIndex == referenceFrameIndex || !it->second->solveThisFrame || frameBoneExcluded[i]) {
				linearIndices[i] = -1;
			}
			else {
				linearIndices[i] = linearIndex_to_frameBoneIndex.size();

				// Add entry to linear list of bones
				linearIndex_to_frameBoneIndex.push_back(int2_type(frameIndex,i));
			}
		}

		// Add to the hash table
		frameBoneIndex_to_linearIndex[frameIndex] = linearIndices;
	}
}

Deform::DeformAll_Utility::TfVarTranslate::TfVarTranslate( int frameIndex, int numBones, const std::vector<int>& excludedLabels, const std::vector<int>& occludedLabels ) :
originalNumBonesPerFrame(numBones)
{
	//////////////////////////////////////////////////////////////////////////
	// Initialize transformation linear indices for a single solution
	//////////////////////////////////////////////////////////////////////////

	// Populate a boolean array indicating whether each bone is excluded
	std::vector<bool> isBoneExcluded(originalNumBonesPerFrame, false);
	for (int i = 0; i < excludedLabels.size(); ++i) {
		isBoneExcluded[excludedLabels[i]] = true;
	}
	// Take into account occluded bones
	for (int i = 0; i < occludedLabels.size(); ++i) {
		isBoneExcluded[occludedLabels[i]] = true;
	}

	std::vector<int> linearIndices(originalNumBonesPerFrame);

	for (int i = 0; i < originalNumBonesPerFrame; ++i) {
		// Check if the bone is excluded --->>> then don't optimize for the transformation.
		if (isBoneExcluded[i]) {
			linearIndices[i] = -1;
		}
		else {
			linearIndices[i] = linearIndex_to_frameBoneIndex.size();

			// Add entry to linear list of bones
			linearIndex_to_frameBoneIndex.push_back(int2_type(frameIndex,i));
		}
	}

	// Add to the hash table
	frameBoneIndex_to_linearIndex[frameIndex] = linearIndices;
}


bool Deform::DeformAll_Utility::TfVarTranslate::validBone( int frameIndex, int boneIndex ) const
{
	stdext::hash_map< int, std::vector<int> >::const_iterator it = frameBoneIndex_to_linearIndex.find(frameIndex);

	if (it == frameBoneIndex_to_linearIndex.end()) {
		return false;
	}

	return it->second[boneIndex] >= 0;
}

int Deform::DeformAll_Utility::TfVarTranslate::linearIndex( int frameIndex, int boneIndex ) const
{
	stdext::hash_map< int, std::vector<int> >::const_iterator it = frameBoneIndex_to_linearIndex.find(frameIndex);

	if (it == frameBoneIndex_to_linearIndex.end()) {
		return -1;
	}

	return it->second[boneIndex];
}

void Deform::DeformAll_Utility::TfVarTranslate::frameBoneIndex( index_type varIndex, int& frameIndex, int& boneIndex ) const
{
	frameIndex = linearIndex_to_frameBoneIndex[varIndex][0];
	boneIndex = linearIndex_to_frameBoneIndex[varIndex][1];
}

int Deform::DeformAll_Utility::TfVarTranslate::numTotalBones() const
{
	return linearIndex_to_frameBoneIndex.size();
}

void Deform::DeformAll_Utility::TfVarTranslate::printMapping() const
{
	//// Print the frame/bone index to linear index mapping
	//stdext::hash_map< int, std::vector<int> >::const_iterator it;
	//for (it = frameBoneIndex_to_linearIndex.begin(); it != frameBoneIndex_to_linearIndex.end(); ++it) {
	//	const std::vector<int>& linearIndices = it->second;
	//	for (int i = 0; i < linearIndices.size(); ++i) {
	//		cout << "Frame " << it->first << " bone " << i << ": " << linearIndices[i] << endl;
	//	}
	//}

	for (int i = 0; i < numTotalBones(); ++i) {
		const int2_type& temp = linearIndex_to_frameBoneIndex[i];
		cout << "Bone " << i << ": frame " << temp[0] << ", bone "<< temp[1] << endl;
	}
}

void Deform::DeformAll_Utility::TfVarTranslate::remove_linearIndices( const idxList& toRemove )
{
	const int numOldBones = numTotalBones();

	// 1. Make a boolean array
	std::vector<bool> isValid(numOldBones, true);
	for (int i = 0; i < toRemove.size(); ++i) {
		isValid[toRemove[i]] = false;
	}

	// 2. Translate from old linear indices to new linear indices
	std::vector<int2_type> new_LinearIndex_to_frameBoneIndex;
	std::vector<int> translateOld_to_New(numOldBones, -1);
	int numNewBones = 0;
	for (int i = 0; i < numOldBones; ++i) {
		if (isValid[i]) {
			new_LinearIndex_to_frameBoneIndex.push_back(linearIndex_to_frameBoneIndex[i]);
			translateOld_to_New[i] = numNewBones;
			++numNewBones;
		}
	}

	// 3. Apply the translation to frameBoneIndex_to_linearIndex
	stdext::hash_map< int, std::vector<int> >::iterator it;
	for (it = frameBoneIndex_to_linearIndex.begin(); it != frameBoneIndex_to_linearIndex.end(); ++it) {
		std::vector<int>& linearIndices = it->second;
		for (int i = 0; i < linearIndices.size(); ++i) {
			if (linearIndices[i] >= 0) {
				// Apply the translation
				linearIndices[i] = translateOld_to_New[linearIndices[i]];
			}
		}
	}

	// 4. Swap old and new linear indices
	linearIndex_to_frameBoneIndex.swap(new_LinearIndex_to_frameBoneIndex);
}

Deform::DeformAll_Utility::ManagedFrameSequence::ManagedFrameSequence( const std::string& rootDir, const std::string& meshSequence ) :
rootDir(rootDir), meshSequence(meshSequence), frameFS(NULL)
{
	frameFS = new FileSequence(meshSequence);
	if (frameFS->valid()) {
		frameFS_indices = frameFS->autoDetect();
		shapeCache.resize(frameFS_indices.size(), NULL);
	}
}

Deform::DeformAll_Utility::ManagedFrameSequence::~ManagedFrameSequence()
{
	// Delete cached shapes
	for (int i = 0; i < shapeCache.size(); ++i) {
		if (shapeCache[i]) {
			delete shapeCache[i];
			shapeCache[i] = NULL;
		}
	}

	delete frameFS;
}

bool Deform::DeformAll_Utility::ManagedFrameSequence::sequenceValid() const
{
	return (frameFS->valid() && frameFS_indices.size());
}

bool Deform::DeformAll_Utility::ManagedFrameSequence::frameValid( int frame ) const
{
	return ((frame >= 0) && (frame < frameFS_indices.size()));
}

Mesh3_real* Deform::DeformAll_Utility::ManagedFrameSequence::getFrame( int frame, bool& cached )
{
	cached = false;

	if (frameValid(frame)) {
		if (checkCache(frame)) {
			cached = true;
			return shapeCache[frame];
		}
		else {
			QString frameFile = (*frameFS)[frameFS_indices[frame]];

			// TODO: Expand this implementation to detect the shape type
			// *Assumption:* For now, assume this is a mesh
			Mesh3_real *shape = new Mesh3_real();

			shape->setFile(QString(rootDir.c_str()), frameFile);
			shape->loadGeometry();
			shape->printInfo();

			// Cache the shape and return
			shapeCache[frame] = shape;

			return shape;
		}
	}

	return NULL;
}

bool Deform::DeformAll_Utility::ManagedFrameSequence::checkCache( int frame ) const
{
	return (shapeCache[frame] != NULL);
}

real_type Deform::DeformGraphAll_Utility::findClosestPoints( const DeformAll_Utility::AllSampleGraph& asg, const DeformAll_Utility::SparseTLArray& allTfList, const int targetFrameNumber, const ShapeModel& targetShape, const idxList& targetSampleIndices, const std::vector<bool>& isOccluded, ClosestPointMap& result, const DeformAll_Utility::Options& defOpt, distList *sampleError/* = NULL*/, bool computeError/* = false*/ )
{
	ClosestPointMap closestPts;

	// Find the transformations of the specific frame to transform to
	DeformAll_Utility::SparseTLArray::const_iterator it = allTfList.find(targetFrameNumber);
	if (it == allTfList.end()) {
		cerr << "Error in findClosestPoints(): Cannot find transformation for target frame " << targetFrameNumber << endl;
		return 0.0;
	}
	const DeformAll_Utility::TransformList& tfList = *(it->second);

#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: pre-multiplication of transform list
	SparseTLArray preMultTf;
	for (it = allTfList.begin(); it != allTfList.end(); ++it)
	{
		const int frameIndex = it->first;
		TransformList* newMultiple = new TransformList();
		*newMultiple = (-tfList) * *(it->second);
		preMultTf[frameIndex] = newMultiple;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif

	bool storeError = (sampleError && ((asg.numSamples()+1) == sampleError->size()));
	real_type totalDist = 0.0;
	int sampleErrorCount = 0;

	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		if (asg.fIndex[i] < targetFrameNumber && asg.weightList[i].size()) {
			// Check occlusion
			bool occluded = false;
			for (int j = 0, j_end = asg.weightList[i].size(); j < j_end; ++j) {
				if (isOccluded[asg.weightList[i][j].index]) {
					occluded = true;
					break;
				}
			}

			if (!occluded) {
				real3_type queryPos, queryNormal;

#ifdef OPT_TF_PREMULT
				// Using premultiplication
				it = preMultTf.find(asg.fIndex[i]);
				if (it == preMultTf.end()) {
					cerr << "Error in findClosestPoints(): Cannot find transformation for sample " << i << ", frame " << asg.fIndex[i] << endl;
					break;
				}
				if (TEST_COMP_TYPE == 1) {
					cerr << "Bad option!!" << endl;
					break;
				}
				else if (TEST_COMP_TYPE == 2) {
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, asg.weightList[i], *(it->second), false);
				}
#else
				// Transform the sample to the reference, then to the specified frame
				it = allTfList.find(asg.fIndex[i]);
				if (it == allTfList.end()) {
					cerr << "Error in findClosestPoints(): Cannot find transformation for sample " << i << ", frame " << asg.fIndex[i] << endl;
					break;
				}
				// Transform the sample to the reference, then to the specified frame
				if (TEST_COMP_TYPE == 1) {
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, asg.weightList[i], *(it->second), false);
					lbs(queryPos, queryNormal, queryPos, queryNormal, asg.weightList[i], tfList, true);
				}
				else if (TEST_COMP_TYPE == 2) {
					DeformAll_Utility::TransformList temp = (-tfList) * *(it->second);
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, asg.weightList[i], temp, false);
				}
#endif

				// Find the closest point
				ClosestPointInfo cpi = targetShape.findClosestPoint(queryPos, queryNormal, defOpt.closestPt_normalWeight);
				index_type closestIdx = cpi.index;


				// Perform outlier detection:
				// 1. Distance threshold (when target on boundary)
				// 2. Distance threshold
				// 3. Normal angle threshold (assumes that normals are already normalized)
				real3_type diff = queryPos - targetShape.vertex[closestIdx];
				real_type cdist = diff.norm();
				real_type cosAngle = queryNormal * targetShape.normal[closestIdx];

				if ((targetShape.markVertex[closestIdx] == VERTEX_VALID || cdist <= defOpt.bdyMaxCorrepDist)
					&& cdist <= defOpt.maxCorrepDist
					&& cosAngle >= defOpt.minNormalCosAngle)
				{
					// Store the closest point
					closestPts[i] = closestIdx;

					// Compute and add error
					if (computeError) {
						real_type plDiff = diff * targetShape.normal[closestIdx];
						plDiff = (DeformAll_Utility::ERROR_POINT_WEIGHT*diff.square_norm() + DeformAll_Utility::ERROR_PLANE_WEIGHT*plDiff*plDiff);
						totalDist += plDiff;
						if (storeError) {
							(*sampleError)[i] = plDiff;
						}
						++sampleErrorCount;
					}
				}
			}
		}
		else if (asg.fIndex[i] == targetFrameNumber) {
			// Store sample's index
			closestPts[i] = targetSampleIndices[asg.sIndex[i]];
		}
		// Otherwise, if the sample's frame index is greater than the target index, then sample doesn't exist in the target,
		// ... since it's sampled from a future frame (thus sample is unobserved in target frame). => Don't do anything for this case.
	}

#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: delete pre-multiplication of transform list
	for (it = preMultTf.begin(); it != preMultTf.end(); ++it)
	{
		delete it->second;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif

	// Assign result
	result.swap(closestPts);

	// Store count of sample error
	if (storeError) {
		(*sampleError)[asg.numSamples()] = sampleErrorCount;
	}

	return totalDist;
}

void Deform::DeformGraphAll_Utility::propogateInitialLabels_Ref( DeformGraphAll_Utility::FrameSolution *refSoln, const MatchCandidates& matchCand, DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult )
{
	const int refFrameIndex = refSoln->frameNumber;
	const ShapeModel refShape = refSoln->originalShape;
	const int numBones = refSoln->tfList.size();
	const idxList refSampleIndices = refSoln->frameSamples;

	// Count the number of labels used (on lastSoln)
	int maxLabels = matchCand.sortedTFList.size();
	idxList numVerticesPerLabel(maxLabels, 0), sortIndex(maxLabels);
	std::vector<int> markLabels(maxLabels, -1);
	for (int i = 0, i_end = matchCand.graphCutLabels_source.size(); i < i_end; ++i) {
		int label = matchCand.graphCutLabels_source[i];
		if (label >= 0 && label < maxLabels) {
			++numVerticesPerLabel[label];
		}
	}

	// Sort indices into numVerticesPerLabel in descending order
	for (int i = 0; i < maxLabels; ++i) {
		sortIndex[i] = i;
	}
	std::sort(sortIndex.begin(), sortIndex.end(), make_ForwardCompare(numVerticesPerLabel, comp_desc<index_type>()));

	// Count the number of labels that are used
	int usedLabels = 0;
	for (int i = 0; i < maxLabels; ++i) {
		if (numVerticesPerLabel[sortIndex[i]] > 0) {
			++usedLabels;
		}
	}

	// Mark the labels to propagate
	int top = min(usedLabels, numBones);
	int topVertices = 0;
	for (int i = 0; i < top; ++i) {
		markLabels[sortIndex[i]] = i;
		topVertices += numVerticesPerLabel[sortIndex[i]];
	}

	// Make a list of vertices whose label is in the top B
	PointList markedLabelVertexList;
	idxList markedLabelVIndexList;
	if (usedLabels > top) {
		cout << DeformGraphAll_Utility::printHeader << "Reducing labels down to " << numBones << endl;
		markedLabelVertexList.reserve(topVertices);
		markedLabelVIndexList.reserve(topVertices);
		for (int i = 0, i_end = matchCand.graphCutLabels_source.size(); i < i_end; ++i) {
			int label = matchCand.graphCutLabels_source[i];
			if (label >= 0 && label < maxLabels && markLabels[label] >= 0) {
				markedLabelVertexList.add(refShape.vertex[i]);
				markedLabelVIndexList.push_back(i);
			}
		}
	}

	// Copy the labels over to the ASG samples
	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		if (asg.fIndex[i] == refFrameIndex) {
			int label = matchCand.graphCutLabels_source[refSampleIndices[asg.sIndex[i]]];

			if (label < 0 || label >= maxLabels || markLabels[label] < 0) {
				// If label is not valid, or it's not in the top,
				// ... instead find the closest point in markedLabelVertexList and assign its label to the sample
				idxList nearest;
				distList dists;
				markedLabelVertexList.queryKNN(refShape.vertex[refSampleIndices[asg.sIndex[i]]], 1, nearest, dists, true);
				label = matchCand.graphCutLabels_source[markedLabelVIndexList[nearest[0]]];
			}

			// Assign the label
			DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(markLabels[label], 1.0));
			asg.weightList[i].swap(sw);
		}
	}

	// Copy the transformations
	DeformAll_Utility::TransformList tfList(numBones);
	for (int i = 0; i < top; ++i) {
		const SE3& tf_orig = matchCand.sortedInverseTFList[sortIndex[i]];
		tfList[i] = DeformAll_Utility::SE3_6DOF(tf_orig.rot, tf_orig.tr);
	}
	// Assign to result
	tfResult.swap(tfList);

	//////////////////////////////////////////////////////////////////////////
	cout << DeformGraphAll_Utility::printHeader << "Propagating initial labels to reference ASG..." << endl;
	cout << DeformGraphAll_Utility::printHeader << "Top " << usedLabels << " labels by region size:" << endl;
	for (int i = 0; i < usedLabels; ++i) {
		cout << DeformGraphAll_Utility::printHeader << "-- Label " << sortIndex[i] << ": " << numVerticesPerLabel[sortIndex[i]] << " vertices" << endl;
	}
}

void Deform::DeformGraphAll_Utility::propogateInitialLabels_Ref( const PointList& asgVertex, const int numBones, const MatchCandidates& matchCand, DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult )
{
	// Count the number of labels used (on lastSoln)
	int maxLabels = matchCand.sortedTFList.size();
	idxList numVerticesPerLabel(maxLabels, 0), sortIndex(maxLabels);
	std::vector<int> markLabels(maxLabels, -1);
	for (int i = 0, i_end = matchCand.graphCutLabels_source.size(); i < i_end; ++i) {
		int label = matchCand.graphCutLabels_source[i];
		if (label >= 0 && label < maxLabels) {
			++numVerticesPerLabel[label];
		}
	}

	// Sort indices into numVerticesPerLabel in descending order
	for (int i = 0; i < maxLabels; ++i) {
		sortIndex[i] = i;
	}
	std::sort(sortIndex.begin(), sortIndex.end(), make_ForwardCompare(numVerticesPerLabel, comp_desc<index_type>()));

	// Count the number of labels that are used
	int usedLabels = 0;
	for (int i = 0; i < maxLabels; ++i) {
		if (numVerticesPerLabel[sortIndex[i]] > 0) {
			++usedLabels;
		}
	}

	// Mark the labels to propagate
	int top = min(usedLabels, numBones);
	int topVertices = 0;
	for (int i = 0; i < top; ++i) {
		markLabels[sortIndex[i]] = i;
		topVertices += numVerticesPerLabel[sortIndex[i]];
	}

	// Make a list of vertices whose label is in the top B
	PointList markedLabelVertexList;
	idxList markedLabelVIndexList;
	if (usedLabels > top) {
		cout << DeformGraphAll_Utility::printHeader << "Reducing labels down to " << numBones << endl;
		markedLabelVertexList.reserve(topVertices);
		markedLabelVIndexList.reserve(topVertices);
		for (int i = 0, i_end = matchCand.graphCutLabels_source.size(); i < i_end; ++i) {
			int label = matchCand.graphCutLabels_source[i];
			if (label >= 0 && label < maxLabels && markLabels[label] >= 0) {
				markedLabelVertexList.add(asgVertex[i]);
				markedLabelVIndexList.push_back(i);
			}
		}
	}

	// Copy the labels over to the ASG samples
	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		int label = matchCand.graphCutLabels_source[i];

		if (label < 0 || label >= maxLabels || markLabels[label] < 0) {
			// If label is not valid, or it's not in the top,
			// ... instead find the closest point in markedLabelVertexList and assign its label to the sample
			idxList nearest;
			distList dists;
			markedLabelVertexList.queryKNN(asgVertex[i], 1, nearest, dists, true);
			label = matchCand.graphCutLabels_source[markedLabelVIndexList[nearest[0]]];
		}

		// Assign the label
		DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(markLabels[label], 1.0));
		asg.weightList[i].swap(sw);
	}

	// Copy the transformations
	DeformAll_Utility::TransformList tfList(numBones);
	for (int i = 0; i < top; ++i) {
		const SE3& tf_orig = matchCand.sortedInverseTFList[sortIndex[i]];
		tfList[i] = DeformAll_Utility::SE3_6DOF(tf_orig.rot, tf_orig.tr);
	}
	// Assign to result
	tfResult.swap(tfList);

	//////////////////////////////////////////////////////////////////////////
	cout << DeformGraphAll_Utility::printHeader << "Propagating initial labels to reference ASG..." << endl;
	cout << DeformGraphAll_Utility::printHeader << "Top " << usedLabels << " labels by region size:" << endl;
	for (int i = 0; i < usedLabels; ++i) {
		cout << DeformGraphAll_Utility::printHeader << "-- Label " << sortIndex[i] << ": " << numVerticesPerLabel[sortIndex[i]] << " vertices" << endl;
	}
}

void Deform::DeformGraphAll_Utility::propogateInitialLabels( const MatchCandidates& matchCand, const DeformAll_Utility::AllSampleGraph& asg, const ClosestPointMap& closestPts, DeformAll_Utility::TransformList &tfResult )
{
	const int numBones = tfResult.size();
	idxList total(numBones,0), overlap(numBones,0);
	std::vector< std::map<int,int> > matchedLabels(numBones);

	// Compute overlap of ASG with matchCand result
	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		int label = findLabel(asg.weightList[i]);
		++total[label];

		ClosestPointMap::const_iterator it = closestPts.find(i);
		if (it != closestPts.end()) {
			++overlap[label];

			// Add matchCand label to list
			int mcLabel = matchCand.graphCutLabels_source[it->second];
			if (matchedLabels[label].find(mcLabel) == matchedLabels[label].end()) {
				matchedLabels[label][mcLabel] = 0;
			}
			else {
				++matchedLabels[label][mcLabel];
			}
		}
	}

	// Compute weighted average of overlapping transformations
	DeformAll_Utility::TransformList tfList(numBones);
	for (int i = 0; i < numBones; ++i) {
		if (overlap[i] > 0) {
			// Compute DLB weighted average sw of combineTf
			real4_type qn, qd;
			for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
				const SE3& tf = matchCand.sortedInverseTFList[it->first];
				real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);

				real4_type boneQuat, boneQn, boneQd;
				MatrixToUnitQuat(tf.rot, boneQuat);
				QuatTrans2UDQ(boneQuat, tf.tr, boneQn, boneQd);
				qn += weight * boneQn;
				qd += weight * boneQd;
			}
			Geometry::Matrix33f R;
			real3_type t;
			DQToMatrix(qn, qd, R, t);

			// Assign to list
			tfList[i] = DeformAll_Utility::SE3_6DOF(R,t);
		}
	}
	// Assign to result
	tfResult.swap(tfList);

	//// Apply transformation of maximum overlap
	//DeformAll_Utility::TransformList tfList(numBones);
	//for (int i = 0; i < numBones; ++i) {
	//	if (overlap[i] > 0) {
	//		int maxIndex = -1;
	//		real_type maxWeight = -REAL_TYPE_MAX;
	//		for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
	//			real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);

	//			if (weight > maxWeight) {
	//				maxIndex = it->first;
	//				maxWeight = weight;
	//			}
	//		}

	//		// Assign to list
	//		const SE3& tf = matchCand.sortedInverseTFList[maxIndex];
	//		tfList[i] = DeformAll_Utility::SE3_6DOF(tf.rot, tf.tr);
	//	}
	//}
	//// Assign to result
	//tfResult.swap(tfList);


	//////////////////////////////////////////////////////////////////////////
	cout << DeformGraphAll_Utility::printHeader << "Propagating initial labels to frame ASG..." << endl;
	for (int i = 0; i < numBones; ++i) {
		cout << DeformGraphAll_Utility::printHeader << "-- Label " << i << " (" << overlap[i] << "/" << total[i] << "): ";
		if (overlap[i] > 0) {
			int maxIndex = -1;
			real_type maxWeight = -REAL_TYPE_MAX;
			for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
				if (it != matchedLabels[i].begin()) cout << ",";
				real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);
				cout << it->first << "(" << weight*100.0 << "%)";
				if (weight > maxWeight) {
					maxIndex = it->first;
					maxWeight = weight;
				}
			}
			cout << " (max: " << maxIndex << ")";
		}
		cout << endl;
	}
}

void Deform::DeformGraphAll_Utility::propogateInitialLabels( const MatchCandidates& matchCand, const DeformAll_Utility::AllSampleGraph& asg, DeformAll_Utility::TransformList &tfResult )
{
	const int numBones = tfResult.size();
	idxList total(numBones,0), overlap(numBones,0);
	std::vector< std::map<int,int> > matchedLabels(numBones);

	// Compute overlap of ASG with matchCand result
	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		int label = findLabel(asg.weightList[i]);
		++total[label];
		++overlap[label];

		// Add matchCand label to list
		//int mcLabel = matchCand.graphCutLabels_source[i];
		int mcLabel = matchCand.graphCutLabels_target[i];
		if (matchedLabels[label].find(mcLabel) == matchedLabels[label].end()) {
			matchedLabels[label][mcLabel] = 0;
		}
		else {
			++matchedLabels[label][mcLabel];
		}
	}

	// Compute weighted average of overlapping transformations
	DeformAll_Utility::TransformList tfList(numBones);
	for (int i = 0; i < numBones; ++i) {
		if (overlap[i] > 0) {
			// Compute DLB weighted average sw of combineTf
			real4_type qn, qd;
			for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
				//const SE3& tf = matchCand.sortedInverseTFList[it->first];
				const SE3& tf = matchCand.sortedTFList[it->first];
				real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);

				real4_type boneQuat, boneQn, boneQd;
				MatrixToUnitQuat(tf.rot, boneQuat);
				QuatTrans2UDQ(boneQuat, tf.tr, boneQn, boneQd);
				qn += weight * boneQn;
				qd += weight * boneQd;
			}
			Geometry::Matrix33f R;
			real3_type t;
			DQToMatrix(qn, qd, R, t);

			// Assign to list
			tfList[i] = DeformAll_Utility::SE3_6DOF(R,t);
		}
	}
	// Assign to result
	tfResult.swap(tfList);

	//// Apply transformation of maximum overlap
	//DeformAll_Utility::TransformList tfList(numBones);
	//for (int i = 0; i < numBones; ++i) {
	//	if (overlap[i] > 0) {
	//		int maxIndex = -1;
	//		real_type maxWeight = -REAL_TYPE_MAX;
	//		for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
	//			real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);

	//			if (weight > maxWeight) {
	//				maxIndex = it->first;
	//				maxWeight = weight;
	//			}
	//		}

	//		// Assign to list
	//		const SE3& tf = matchCand.sortedInverseTFList[maxIndex];
	//		tfList[i] = DeformAll_Utility::SE3_6DOF(tf.rot, tf.tr);
	//	}
	//}
	//// Assign to result
	//tfResult.swap(tfList);


	//////////////////////////////////////////////////////////////////////////
	cout << DeformGraphAll_Utility::printHeader << "Propagating initial labels to frame ASG..." << endl;
	for (int i = 0; i < numBones; ++i) {
		cout << DeformGraphAll_Utility::printHeader << "-- Label " << i << " (" << overlap[i] << "/" << total[i] << "): ";
		if (overlap[i] > 0) {
			int maxIndex = -1;
			real_type maxWeight = -REAL_TYPE_MAX;
			for (std::map<int,int>::const_iterator it = matchedLabels[i].begin(); it != matchedLabels[i].end(); ++it) {
				if (it != matchedLabels[i].begin()) cout << ",";
				real_type weight = ((real_type)it->second) / ((real_type)overlap[i]);
				cout << it->first << "(" << weight*100.0 << "%)";
				if (weight > maxWeight) {
					maxIndex = it->first;
					maxWeight = weight;
				}
			}
			cout << " (max: " << maxIndex << ")";
		}
		cout << endl;
	}
}

void Deform::DeformGraphAll_Utility::setRandomLabels( DeformAll_Utility::AllSampleGraph& asg, const PointList& points, int numBones )
{
	cout << DeformGraphAll_Utility::printHeader << "Setting random labeling for " << numBones << " labels...";

	// Select random locations sampled from ASG points
	idxList indices(1);
	distList dists(1);
	indices = bestCandidateSample(points, numBones);
	PointList temp(numBones);
	for (int i = 0; i < numBones; ++i) {
		temp[i] = points[indices[i]];
	}

	// Find the closest bone position & assign the corresponding label (like k-means clustering)
	for (int i = 0, i_end = asg.numSamples(); i < i_end; ++i) {
		((const PointList&)temp).queryKNN(points[i], 1, indices, dists, true);
		asg.weightList[i].swap(DeformAll_Utility::SparseWeight(1, DeformAll_Utility::IndexedWeight(indices[0], 1.0)));
	}
	cout << " done!" << endl;
}

void Deform::DeformGraphAll_Utility::fillOccludedTfs( const DeformGraphAll_Utility::JointMap& jointMap, DeformGraphAll_Utility::FrameSolution& fillSoln, const DeformAll_Utility::Options &defOpt, const DeformGraphAll_Utility::SolutionTable& frameSolution )
{
	const std::vector<int>& occludedLabels = fillSoln.occludedLabels;
	DeformAll_Utility::TransformList& tfList = fillSoln.tfList;
	const int hingeJointSamples = defOpt.hingeJointSamples;
	const real_type hingeExtent = defOpt.hingeExtent;

	bool printDebug = true;
	if (printDebug) pauseUntilEnter("Filling occluded transformations");

	// Note: We shouldn't have to worry about excludedLabels because those will not be connected to any other bone (won't show up in jointValid)

	// Find bones to fill
	const int numBones = tfList.size();
	std::deque<int> toFill;
	std::vector<bool> tfFilled(numBones, true);

	if (printDebug) cout << "---***>>> Filling occluded transformations (frame " << fillSoln.frameNumber << "):";
	for (int i = 0; i < occludedLabels.size(); ++i) {
		tfFilled[occludedLabels[i]] = false;
		toFill.push_back(occludedLabels[i]);
		if (printDebug) cout << " " << occludedLabels[i];
	}
	if (printDebug) cout << endl;

	// Fill the rest of the labels
	int itr = 0;
	const int maxItr = toFill.size() * toFill.size();
	while (!toFill.empty() && (itr < maxItr)) {
		int tfToFill = toFill.front();
		toFill.pop_front();
		++itr;

		// Gather valid (filled) neighboring transformations
		if (printDebug) cout << "---***>>> Gathering neighbors for " << tfToFill << "..." << endl;
		std::vector<int> filledNeighbors;
		int totalNeighbors = 0;
		for (int ni = 0; ni < numBones; ++ni) {
			if (tfToFill != ni) {
				int bi, bj;
				bi = min(tfToFill, ni);
				bj = max(tfToFill, ni);

				// See if the joint is valid
				int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones,bi,bj);				
				if (jointMap.find(linIdx) != jointMap.end()) {
					if (printDebug) cout << "---***>>> -- Joint between " << bi << " and " << bj << " exists: ";
					// See if the tf is filled in
					if (tfFilled[ni]) {
						if (printDebug) cout << "Storing.";
						// Store the tf
						filledNeighbors.push_back(ni);
					}
					++totalNeighbors;
					if (printDebug) cout << endl;
				}
			}
		}

		if (printDebug) {
			cout << "---***>>> -- Filling " << tfToFill << ", neighbors: [";
			for (int i = 0; i < filledNeighbors.size(); ++i)
			{
				if (i != 0) cout << ",";
				cout << filledNeighbors[i];
			}
			cout << "]" << endl;
		}

		const int numNeighbors = filledNeighbors.size();
		if (numNeighbors == 0) {
			if (printDebug) cout << "---***>>> --> Push back" << endl;
			// Push it to the back of the queue
			toFill.push_back(tfToFill);
		}
		else if (numNeighbors == 1) {
			if (printDebug) cout << "---***>>> --> Seek relative tf from frame:";

			// Seek the last valid relative transformation to that part
			SE3_6DOF relativeTf;
			const int ni = filledNeighbors[0];
			const int fillFrameNumber = fillSoln.frameNumber;

			DeformGraphAll_Utility::SolutionTable::const_iterator it = frameSolution.find(fillFrameNumber);
			if (it != frameSolution.end())
			{
				DeformGraphAll_Utility::SolutionTable::const_reverse_iterator rev_it(it);
				bool seekDone = false;

				// Seek the frame solution table in reverse order starting from the frame just before
				while (!seekDone && rev_it != frameSolution.rend())
				{
					int frameIndex = rev_it->first;
					DeformGraphAll_Utility::FrameSolution *frameSoln = rev_it->second;
					std::vector<bool>& frameOccluded = frameSoln->isBoneOccluded();
			
					if (printDebug) cout << " " << frameIndex;

					// Check if both transformations are not occluded in this frame
					if (!frameOccluded[ni] && !frameOccluded[tfToFill])
					{
						// Compute the relative transformation
						relativeTf = frameSoln->tfList[tfToFill] * (-frameSoln->tfList[ni]);
						seekDone = true;
						if (printDebug) cout << " found!!" << endl;
					}
					++rev_it;
				}

				if (!seekDone)
				{
					if (printDebug) cout << " failed!!" << endl;
				}
			}

			// Apply the relative transformation to the transformation of the connected neighbor
			tfList[tfToFill] = relativeTf * tfList[ni];
			tfFilled[tfToFill] = true;
		}
		else
		{
			if (printDebug) cout << "---***>>> --> Best fit" << endl;

			// If there are two or more neighbors, compute the best-fitting transformation based on the joints
			Geometry::Matrix33f sumRot, conM;
			real3_type centerX, centerY;
			ptList jointX, jointY;
			distList weights;

			// Set quantities to zero
			conM = sumRot = 0.0*sumRot;
			centerX = centerY = 0.0*centerX;

			// Iterate through each neighbor
			for (int i = 0; i < numNeighbors; ++i) {
				int ni, bi, bj;
				ni = filledNeighbors[i];
				bi = min(tfToFill, ni);
				bj = max(tfToFill, ni);

				// Store reference and transformed joint locations
				int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones,bi,bj);				
				const DeformGraphAll_Utility::Joint& joint = jointMap.find(linIdx)->second;
				const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
				const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

				int numSamples = 0;
				real_type weight = 1.0;
				if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
					numSamples = 1;
				}
				else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
					numSamples = hingeJointSamples;
					weight = 1.0 / ((real_type)hingeJointSamples);
				}

				for (int j = 0; j < numSamples; ++j) {
					real3_type jointLocation;

					if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
						jointLocation = jointPos;
					}
					else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
						real_type t = hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
						jointLocation = jointPos +  t*jointCon;
					}

					// Use just the reference position
					jointX.push_back(jointLocation);
					jointY.push_back(-tfList[ni] * jointLocation);
					weights.push_back(weight);

					// Compute sums
					centerX += jointX[jointX.size()-1];
					centerY += jointY[jointY.size()-1];
				}

				// Invert
				sumRot += tfList[ni].rotMatrix().transpose();
			}

			// Compute centroid
			const int numConstraints = jointX.size();
			centerX /= ((real_type)numConstraints);
			centerY /= ((real_type)numConstraints);

			// Center point set / populate constraint matrix
			for (int i = 0; i < numConstraints; ++i) {
				jointX[i] -= centerX;
				jointY[i] -= centerY;

				for (int row = 0; row < DIMENSION; ++row) {
					for (int col = 0; col < DIMENSION; ++col) {
						conM(row,col) += weights[i] * weights[i] * jointX[i][row] * jointY[i][col];
					}
				}
			}

			// Compute SVD
			Geometry::Matrix33f	S, U, VT, tempR, signFlipM;
			conM.computeSVD(S,U,VT);
			tempR = VT.transpose() * U.transpose();
			signFlipM = Geometry::Matrix33f::identity();
			if (det(tempR) < 0) {
				signFlipM(2,2) = -1.0;
			}

			// Find optimal rotation
			Geometry::Matrix33f optR;
			if (numConstraints == 2) {
				// Resolve rotation ambiguity by minimizing relative rotation to neighboring tfs
				Geometry::Matrix33f R_A_Matrix, R_B_Matrix, R_C_Matrix, RotX;
				R_A_Matrix = R_B_Matrix = R_C_Matrix = 0.0 * R_A_Matrix;
				R_A_Matrix(0,0) = 1.0;
				R_B_Matrix(1,1) = 1.0; R_B_Matrix(2,2) = 1.0;
				R_C_Matrix(1,2) = 1.0; R_C_Matrix(2,1) = -1.0;

				real_type R_A, R_B, R_C;
				Geometry::Matrix33f tempM;
				tempM = VT.transpose() * R_A_Matrix * signFlipM * U.transpose() * sumRot.transpose();
				R_A = tempM.trace();
				tempM = VT.transpose() * R_B_Matrix * signFlipM * U.transpose() * sumRot.transpose();
				R_B = tempM.trace();
				tempM = VT.transpose() * R_C_Matrix * signFlipM * U.transpose() * sumRot.transpose();
				R_C = tempM.trace();

				real_type optAngle = atan(R_C/R_B);
				real_type err = R_A + R_B*cos(optAngle) + R_C*sin(optAngle);
				real_type err2 = R_A - R_B*cos(optAngle) - R_C*sin(optAngle);
				if (err < err2) {
					optAngle += M_PI;
				}

				RotX = R_A_Matrix + cos(optAngle)*R_B_Matrix + sin(optAngle)*R_C_Matrix;
				optR = VT.transpose() * RotX * signFlipM * U.transpose();
			}
			else {
				// Return the det = +1 rotation
				optR = VT.transpose() * signFlipM * U.transpose();
			}
			// Find optimal translation
			real3_type optT = centerY - optR*centerX;

			// Assign the result
			tfList[tfToFill] = DeformAll_Utility::SE3_6DOF(optR, optT);
			tfList[tfToFill] = -tfList[tfToFill];
			tfFilled[tfToFill] = true;

			//// Debugging: print information
			//cout << "Filling occluded bone " << tfToFill << endl;
			//cout << "Rotation: " << endl << optR;
			//cout << "Translation: " << optT << endl;
		}
	}

	if (toFill.size()) {
		// If there are still remaining transformations to fill
		if (printDebug) cout << "---***>>> WARNING! Filling remaining tfs with identity:";

		//// Option 1 --> Use the largest region's transformation
		//const int maxLabel = getMaxRegionLabel();
		//for (int i = 0; i < toFill.size(); ++i) {
		//	tfList[toFill[i]] = tfList[maxLabel];
		//}

		// Option 2 --> Use identity
		for (int i = 0; i < toFill.size(); ++i) {
			tfList[toFill[i]] = DeformAll_Utility::SE3_6DOF();
			if (printDebug) cout << " " << toFill[i];
		}
		if (printDebug) cout << endl;
	}
}

void Deform::DeformGraphAll_Utility::precomputeSampleError(
	const ShapeModel& targetShape, const int targetFrameNumber, const DeformAll_Utility::SparseTLArray& allTfList,
	const DeformAll_Utility::AllSampleGraph& asg, const std::vector<bool>& isOccluded,
	const DeformGraphAll_Utility::ClosestPointMap& closestPts,
	const real_type normalWeight, bool accumulateASGOnRef,
	DeformGraphAll_Utility::Error_LabelSample& result )
{
	// Find the transformations of the specific frame to transform to
	DeformAll_Utility::SparseTLArray::const_iterator itTf = allTfList.find(targetFrameNumber);
	if (itTf == allTfList.end()) {
		cerr << "Error in precomputeSampleError(): Cannot find transformation for target frame " << targetFrameNumber << endl;
		return;
	}
	const DeformAll_Utility::TransformList& tgtTfList = *(itTf->second);


#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: pre-multiplication of transform list
	SparseTLArray preMultTf;
	for (itTf = allTfList.begin(); itTf != allTfList.end(); ++itTf)
	{
		const int frameIndex = itTf->first;
		TransformList* newMultiple = new TransformList();
		*newMultiple = (-tgtTfList) * *(itTf->second);
		preMultTf[frameIndex] = newMultiple;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif

	const int numBones = tgtTfList.size();
	const int numSamples = asg.numSamples();
	DeformGraphAll_Utility::Error_LabelSample errorLS(numBones,distList(numSamples,0.0));

	for (int l = 0; l < numBones; ++l) {
		DeformAll_Utility::SparseWeight sw(1,DeformAll_Utility::IndexedWeight(l,1.0));

		// If bone l is occluded, defer the assignment until later
		if (!isOccluded[l]) {
			// For each sample that has a closest point on this frame
			// i.e. distance not computed for samples that don't have a valid corresponding point computed using the currently assigned weight
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it;
			for (it = closestPts.begin(); it != closestPts.end(); ++it) {
				const int i = it->first;
				real3_type queryPos, queryNormal;


#ifdef OPT_TF_PREMULT
				// Using pre-multiplication
				itTf = preMultTf.find(asg.fIndex[i]);
				if (itTf == preMultTf.end()) {
					cerr << "Error in findClosestPoints(): Cannot find transformation for sample " << i << ", frame " << asg.fIndex[i] << endl;
					break;
				}
				if (TEST_COMP_TYPE == 1) {
					cerr << "Bad option!!" << endl;
					break;
				}
				else if (TEST_COMP_TYPE == 2) {
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, sw, *(itTf->second), false);
				}
#else
				// Find transformation to reference for this sample
				itTf = allTfList.find(asg.fIndex[i]);
				if (itTf == allTfList.end()) {
					cerr << "Error in precomputeSampleError(): Cannot find transformation for sample " << i << ", frame " << asg.fIndex[i] << endl;
					break;
				}
				if (TEST_COMP_TYPE == 1) {
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, sw, *(itTf->second), false);
					lbs(queryPos, queryNormal, queryPos, queryNormal, sw, tgtTfList, true);
				}
				else if (TEST_COMP_TYPE == 2) {
					DeformAll_Utility::TransformList temp = (-tgtTfList) * *(itTf->second);
					lbs(asg.vertex[i], asg.normal[i], queryPos, queryNormal, sw, temp, false);
				}
#endif

				// Find the closest point, compute distance
				ClosestPointInfo cpi = targetShape.findClosestPoint(queryPos, queryNormal, normalWeight);
				real3_type targetP, targetN;
				targetP = targetShape.vertex[cpi.index];
				targetN = targetShape.normal[cpi.index];

				// Store error
				real3_type diff = (queryPos - targetP);
				real_type plDiff = diff * targetN;

				// To make it units of distance (instead of distance squared), square root the total quantity
				errorLS[l][i] = sqrtf(DeformAll_Utility::ERROR_POINT_WEIGHT*diff.square_norm() + DeformAll_Utility::ERROR_PLANE_WEIGHT*plDiff*plDiff);
			}
		}
	}

#ifdef OPT_TF_PREMULT
	////////////////////////////////////////////////////////////////////////////////
	// Optimization: delete pre-multiplication of transform list
	for (itTf = preMultTf.begin(); itTf != preMultTf.end(); ++itTf)
	{
		delete itTf->second;
	}
	////////////////////////////////////////////////////////////////////////////////
#endif

	// For all occluded bones: assign, at each sample, the error of the currently assigned label
	// ... with a small epsilon added/subtracted from the error.
	for (int l = 0; l < numBones; ++l) {
		if (isOccluded[l]) {
			// For each sample that has a closest point on this frame
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it;
			for (it = closestPts.begin(); it != closestPts.end(); ++it) {
				const int i = it->first;

				// Find the current label of the sample
				const int currentLabel = findLabel(asg.weightList[i]);

				if (accumulateASGOnRef) {
					// If we're performing the accumulated strategy, then we don't want to assign the error of the current label.
					// Instead we want to prevent assigning this label on other places, and restrict the label to the same
					// location. --> Set the error high if the current label on this sample is not l, otherwise 0.
					if (currentLabel != l) {
						errorLS[l][i] = REAL_TYPE_MAX;
					}
				}
				else {
					// The current label may be occluded as well!!
					if (isOccluded[currentLabel]) {
						// Among all the labels that's not occluded, find the smallest error
						real_type minError = REAL_TYPE_MAX;
						for (int j = 0; j < numBones; ++j) {
							if (!isOccluded[j] && errorLS[j][i] < minError) {
								minError = errorLS[j][i];
							}
						}
						// Assign it as the error of the occluded bone, minus a smoothness penalty
						errorLS[l][i] = minError;
					}
					else {
						errorLS[l][i] = errorLS[currentLabel][i];
					}
				}
			}
		}
	}


	// Assign result
	result.swap(errorLS);
}

void Deform::DeformGraphAll_Utility::getSortedFrameList( const DeformGraphAll_Utility::SolutionTable& frameSolution, idxList& result )
{
	idxList frameList;
	frameList.reserve(frameSolution.size());
	for (DeformGraphAll_Utility::SolutionTable::const_iterator it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		frameList.push_back(it->first);
	}
	sort(frameList.begin(), frameList.end());

	result.swap(frameList);
}

void Deform::DeformGraphAll_Utility::pruneEdges( const PointList& points, const DeformAll_Utility::SparseWeightList& weightList, BoostGraph& graph, const Deform::DeformAll_Utility::SparseTLArray& allTfList, real_type maxStretchFactor, real_type maxEdgeLength )
{
	std::list< std::pair<int,int> > removeEdgeList;

	const int numBones = allTfList.begin()->second->size();
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		int vi1 = index[source(*ei,graph)];
		int vi2 = index[target(*ei,graph)];
		const DeformAll_Utility::SparseWeight& w1 = weightList[vi1];
		const DeformAll_Utility::SparseWeight& w2 = weightList[vi2];
		const real3_type& pt1 = points[vi1];
		const real3_type& pt2 = points[vi2];

		// Check if vertices have different labels
		if (w1 != w2) {
			real3_type x1, x2;

			// Min and max edge lengths
			real_type minLength = REAL_TYPE_MAX, maxLength = -REAL_TYPE_MAX;
			Deform::DeformAll_Utility::SparseTLArray::const_iterator it;
			for (it = allTfList.begin(); it != allTfList.end(); ++it) {
				lbs(pt1, x1, w1, *(it->second), true);
				lbs(pt2, x2, w2, *(it->second), true);

				real_type len = (x1 - x2).norm();
				if (len < minLength) {
					minLength = len;
				}
				if (len > maxLength) {
					maxLength = len;
				}
			}

			// Exceeds threshold?
			real_type stretch = maxLength / minLength;
			if (stretch > maxStretchFactor) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
			else if (maxLength > maxEdgeLength) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
		}
		else {
			real_type len = (pt1 - pt2).norm();
			if (len > maxEdgeLength) {
				removeEdgeList.push_back(make_pair(vi1, vi2));
			}
		}
	}

	// Remove the edges
	std::list< std::pair<int,int> >::const_iterator it;
	for (it = removeEdgeList.begin(); it != removeEdgeList.end(); ++it) {
		BoostVertex v1 = boost::vertex(it->first,graph);
		BoostVertex v2 = boost::vertex(it->second,graph);
		BoostEdge e;
		bool temp;
		tie(e,temp) = edge(v1,v2,graph);
		if (temp) {
			remove_edge(e,graph);
		}
		else {
			cerr << "Unable to find edge (" << it->first << "," << it->second << ") in graph!" << endl;
		}
	}

	//// Testing
	//testList.swap(removeEdgeList);
}

Deform::DeformGraphAll_Utility::FrameSolution::FrameSolution( const ShapeModel& shape, const int frameNumber ) :
frameNumber(frameNumber), originalShape(shape), solveThisFrame(true), sampleSI(NULL), sampleCurv(NULL), toRef(NULL)
{
}

Deform::DeformGraphAll_Utility::FrameSolution::FrameSolution( const ShapeModel& shape, const int numBones, const int frameNumber, const real_type sampleDist ) :
frameNumber(frameNumber), originalShape(shape), tfList(numBones), solveThisFrame(true), sampleSI(NULL), sampleCurv(NULL), toRef(NULL)
{
	if (sampleDist < REAL_TYPE_EPSILON || (sampleDist < 1.1 * shape.SampleResolution()))
	{
		frameSamples = bestCandidateSample(shape.vertex, shape.vertex.size());
	}
	else
	{
		// Sample the source vertices using best candidate sampling
		frameSamples = bestCandidateSampleDist(shape.vertex, sampleDist);
	}

	cout << DeformGraphAll_Utility::printHeader << "Sampled " << frameSamples.size() << " points (out of " << shape.vertex.size() << ", fraction " << ((real_type)frameSamples.size()) / ((real_type)shape.vertex.size()) << ") from frame " << frameNumber << endl;
}

Deform::DeformGraphAll_Utility::FrameSolution::~FrameSolution()
{
	delete sampleSI;
	delete sampleCurv;
}

std::vector<bool> Deform::DeformGraphAll_Utility::FrameSolution::isBoneOccluded() const
{
	// Populate a boolean array indicating whether each label is occluded
	std::vector<bool> result(tfList.size(), false);
	for (int i = 0; i < occludedLabels.size(); ++i) {
		result[occludedLabels[i]] = true;
	}

	return result;
}

std::vector<bool> Deform::DeformGraphAll_Utility::FrameSolution::isBoneOccludedOrExcluded( const std::vector<int>& excludedLabels ) const
{
	// Populate a boolean array indicating whether each label is occluded or excluded
	std::vector<bool> result(tfList.size(), false);
	for (int i = 0; i < excludedLabels.size(); ++i) {
		result[excludedLabels[i]] = true;
	}
	for (int i = 0; i < occludedLabels.size(); ++i) {
		result[occludedLabels[i]] = true;
	}

	return result;
}

const SpinImageStack& Deform::DeformGraphAll_Utility::FrameSolution::getSpinImages( const char* filename, const Scene& scene )
{
	if (!sampleSI) {
		// Temporarily load/compute the spin images for the shape
		ShapeModel& shape = const_cast<ShapeModel&>(originalShape);
		sampleSI = new SpinImageStack;

		// 1. Try to load sampled curvature directly
		string saveFile(filename);
		replaceField(saveFile, '#', frameNumber);
		bool success = shape.loadSpinImages(saveFile.c_str(), *sampleSI, false);

		// 2. Compute the frames for the entire shape, and sample it
		if (!success) {
			shape.computeSpinImages(scene.binSizeMultiple_SI, scene.imageWidth_SI, scene.maxSupportAngle_SI);

			// Now find the spin images that correspond to the surface samples
			sampleSI->copyParams(shape.spinImageStack);
			sampleSI->resize(frameSamples.size());
			for (int i = 0; i < frameSamples.size(); ++i) {
				(*sampleSI)[i].swap(shape.spinImageStack[frameSamples[i]]);
			}

			// Now clear all memory associated with the original surface stack
			shape.spinImageStack.clear();

			// Save the sampled frames
			shape.saveSpinImages(saveFile.c_str(), *sampleSI);
		}
	}

	return *sampleSI;
}

const CurvInfoList& Deform::DeformGraphAll_Utility::FrameSolution::getCurvInfo( const char* filename, const Scene& scene )
{
	if (!sampleCurv) {
		ShapeModel& shape = const_cast<ShapeModel&>(originalShape);
		sampleCurv = new CurvInfoList;
		curvScales = new distList;

		// 1. Try to load sampled curvature directly
		string saveFile(filename);
		replaceField(saveFile, '#', frameNumber);
		bool success = shape.loadFrames(saveFile.c_str(), *sampleCurv, *curvScales, false);

		// 2. Compute the frames for the entire shape, and sample it
		if (!success) {
			*curvScales = scene.makeSceneSqRadii(scene.scaleFactors_PC);
			shape.computeFrames(*curvScales);

			// Now find the curvature info that correspond to the surface samples
			sampleCurv->resize(frameSamples.size());
			for (int i = 0; i < frameSamples.size(); ++i) {
				(*sampleCurv)[i].swap(shape.frames[frameSamples[i]]);
			}

			// Now clear all memory associated with the original curv info
			shape.frames.clear();
			shape.frameScales.clear();

			// Save the sampled frames
			shape.saveFrames(saveFile.c_str(), QString(""), *sampleCurv, *curvScales);
		}
	}

	return *sampleCurv;
}

const distList* Deform::DeformGraphAll_Utility::FrameSolution::getCurvScales() const
{
	return curvScales;
}

void Deform::DeformGraphAll_Utility::FrameSolution::draw( const PointList& points, const DrawOptions& options, bool drawRef ) const
{
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);

	// Draw reference frame scanned geometry
	if (options.showSourceModel) {
		if (drawRef) {
			setOpenGLMaterialColor(options.targetModelColor);
		}
		else {
			setOpenGLMaterialColor(options.sourceModelColor);
		}
		originalShape.draw();
		//refSoln->originalShape.drawFrames(defOpt.sceneGridSpacing, 0, CURVINFO_NORMALS);
	}

	// Draw reference samples
	if (options.showSourceCorresp) {
		setOpenGLMaterialColor(options.pointColor);
		PointList_Utility::draw(originalShape.vertex, POINT_SPHERES, frameSamples, options.sampleRadius * 1.0);
	}
	if (options.showDeformedCorresp && sampleCurv) {
		glLineWidth(3.0);
		originalShape.drawFrames(*sampleCurv, *curvScales, points, options.sampleRadius * 3.0, options.curvDisplayScale, options.curvDisplayMode);
	}
	if (options.showSourceDeformed) {
		setOpenGLMaterialColor(toRefColor);
		toRef->draw(false);
	}
}

void Deform::DeformGraphAll_Utility::FrameSolution::setRefDeformed( DeformAll_Utility::ShapeDeformed *def )
{
	toRef = def;
	
	// Generate a random light color for it
	std::vector<RealColor> colors;
	generateLightColors(1, colors);
	toRefColor = colors[0];
}


Deform::DeformGraphAll::DeformGraphAll( const DeformAll_Utility::Options& options ) :
DeformAlgorithm(DEFORM_GRAPH_ALL), managedFS(NULL), allSampleGraph(NULL), defOpt(options),
referenceFrameIndex(options.refFrameIndex), lastFrame(options.refFrameIndex), solveStatus(SOLVER_IDLE),
recomputeClosestPtError_AllFrames(true), finalShape(NULL), finalShapeDef(NULL), finalShapeWeights(NULL),
lastDefFrame_ASG(-1), lastDefFrame_finalShape(-1), IKMeshDef(NULL), IKMeshValid(NULL), IKMeshLock(NULL),
matchCandDisplay(NULL), matchCandDisplayIndex(-1), tempPts1(NULL), tempPts2(NULL), tempGraph1(NULL), tempGraph2(NULL)
{

}

Deform::DeformGraphAll::~DeformGraphAll()
{
	// Delete frame data
	for (DeformGraphAll_Utility::SolutionTable::const_iterator it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;
		if (frameSoln) {
			delete frameSoln;
		}
	}

	// Delete data members
	if (managedFS) delete managedFS;
	if (allSampleGraph) delete allSampleGraph;
}

bool Deform::DeformGraphAll::init( const std::string& rootDir, const std::string& meshSequence )
{
	if (managedFS) {
		// Seems to have been called twice?
		cerr << "Error in DeformGraphAll::init(): managedFS exists, init() seems to have been called twice?" << endl;
		return false;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << DeformGraphAll_Utility::printHeader << "Initializing DeformGraphAll..." << endl;

	managedFS = new DeformAll_Utility::ManagedFrameSequence(rootDir, meshSequence);

	if (!managedFS->sequenceValid()) {
		cerr << "Error in DeformGraphAll::init(): error loading mesh sequence " << meshSequence << ". Aborting." << endl;

		delete managedFS;
		managedFS = NULL;
		return false;
	}
	if (!isValidFrame(referenceFrameIndex)) {
		cerr << "Error in DeformGraphAll::init(): cannot load reference frame!" << endl;
		return false;
	}

	cout << DeformGraphAll_Utility::printHeader << "Loading reference frame..." << endl;

	drawMutex.lock();
	// Use the first frame as the reference pose
	const DeformGraphAll_Utility::FrameSolution *refSoln = makeFrameSoln(referenceFrameIndex, false);
	if (!refSoln) {
		drawMutex.unlock();
		return false;
	}

	// Create initial ASG
	allSampleGraph = new DeformAll_Utility::AllSampleGraph(defOpt.EKNNGraph_k, defOpt.refSampleDist);
	allSampleGraph->initASG(refSoln->originalShape, refSoln->frameSamples, refSoln->frameNumber);

	// Rebuild EKNN graph on the reference frame
	DeformAll_Utility::SparseTLArray allTfList;
	PointList vRef, nRef;
	makeAllTFList(frameSolution, allTfList);
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);
	makeEKNNgraph(allSampleGraph->EKNNgraph_k, vRef, allSampleGraph->graph);
	drawMutex.unlock();

	cout << DeformGraphAll_Utility::printHeader << "Finished initialization!" << endl;
	return true;
}

std::pair<real3_type,real3_type> Deform::DeformGraphAll::getDisplayBBox() const
{
	// Return the bounding box information for the reference frame.
	const DeformGraphAll_Utility::FrameSolution* refSoln = getFrameSoln(referenceFrameIndex);
	if (refSoln) {
		return make_pair(refSoln->originalShape.box_min, refSoln->originalShape.box_max);
	}
	return make_pair(real3_type(),real3_type());
}

DeformGraphAll_Utility::FrameSolution* Deform::DeformGraphAll::makeFrameSoln( int frameIndex, bool emptyFrame )
{
	// Check if we already made the solution
	DeformGraphAll_Utility::SolutionTable::iterator it = frameSolution.find(frameIndex);
	if (it != frameSolution.end()) {
		cerr << "Error in makeFrameSoln(): already made for frame " << frameIndex << "!" << endl;
		return it->second;
	}

	bool cached;
	Mesh3_real* shape = managedFS->getFrame(frameIndex, cached);

	if (shape && !cached) {
		// Compute / load additional data
		shape->setSampleResolution(defOpt.sceneGridSpacing);
		shape->computeNormals(true);
		shape->markBoundary_ComputeResolution();
	}

	// Create the solution
	DeformGraphAll_Utility::FrameSolution *soln;
	if (emptyFrame) {
		soln = new DeformGraphAll_Utility::FrameSolution(*shape, frameIndex);
	}
	else {
		soln = new DeformGraphAll_Utility::FrameSolution(*shape, defOpt.numBones, frameIndex, defOpt.refSampleDist );
	}

	// Add the solution
	frameSolution[frameIndex] = soln;

	return soln;
}

DeformGraphAll_Utility::FrameSolution* Deform::DeformGraphAll::getFrameSoln( int frameIndex ) const
{
	DeformGraphAll_Utility::SolutionTable::const_iterator it = frameSolution.find(frameIndex);

	if (it != frameSolution.end()) {
		return it->second;
	}

	cerr << "Error in getFrameSoln(): Couldn't find a solution for the frame " << frameIndex << "!" << endl;
	return NULL;
}

bool Deform::DeformGraphAll::isValidFrame( int frame ) const
{
	if (managedFS) {
		return managedFS->frameValid(frame);
	}
	return false;
}

void Deform::DeformGraphAll::setOptions( const DeformAll_Utility::Options& options )
{
	defOpt.setFromGUI(options);

	//// Testing: change deformation composition type
	//if (TEST_COMP_TYPE == 1) TEST_COMP_TYPE = 2;
	//else if (TEST_COMP_TYPE == 2) TEST_COMP_TYPE = 1;

	//// (Testing) Update closest points
	//DeformGraphAll_Utility::FrameSolution *newSoln = frameSolution.find(lastFrame+1)->second;
	//int newFrame = newSoln->frameNumber;
	//DeformAll_Utility::SparseTLArray allTfList;
	//makeAllTFList(frameSolution, allTfList);
	//DeformGraphAll_Utility::ClosestPointMap closestPts;
	//std::vector<bool> isOccluded = newSoln->isBoneOccludedOrExcluded(excludedLabels);
	//DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, newFrame, newSoln->originalShape, isOccluded, closestPts, defOpt);
	//sampleClosestPoints[newFrame].swap(closestPts);			/**< Assign to global list, delete old object if it's there */
}

void Deform::DeformGraphAll::solveDeformation( const SolveOptions &options )
{
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << DeformGraphAll_Utility::printHeader << "Registration using DeformGraphAll..." << endl;

	if (options.solveIK) {
		solveInteractiveIK();
		return;
	}

	if (!defOpt.loadRiggedShape.empty() && defOpt.ptr_scene) {
		// Load the final reconstructed shape
		finalShape = loadShape(QString(defOpt.ptr_scene->rootDir.c_str()), QString(defOpt.loadRiggedShape.c_str()), *defOpt.ptr_scene, 0);
		const int nVerts = finalShape->numVertices();

		//////////////////////////////////////////////////////////////////////////
		// Load BVH file for joints
		BVHLoader::MySkeletonStateList states;
		std::tr1::shared_ptr<BVHLoader::Joint> root =  DeformGraphAll_Utility::loadJointsBVH(defOpt.loadRiggedBVH, states);
		Deform::DeformGraphAll_Utility::populateJointsFromBVH(root, defOpt.numBones, NULL, jointMap, states[0]);
		cout << DeformGraphAll_Utility::printHeader << "Loading complete!  Number of loaded joints: " << jointMap.size() << endl;

		//////////////////////////////////////////////////////////////////////////
		// Load weight file for vertex weights
		finalShapeDef = new DeformAll_Utility::ShapeDeformed(*finalShape);
		finalShapeWeights = new DeformAll_Utility::RefWeightList;

		int2_type nVertsBones = DeformGraphAll_Utility::loadWeightFile(defOpt.loadRiggedWeights, *finalShapeWeights);

		if (nVerts != nVertsBones[0])
		{
			cerr << "Error! Number of vertices " << nVerts << " in mesh doesn't match number " << nVertsBones[0] << " in weight file!" << std::endl;
		}

		if (defOpt.ptr_scene->graph_n != nVertsBones[1])
		{
			cerr << "Error! Number of bones in scene " << defOpt.ptr_scene->graph_n << " in mesh doesn't match number " << nVertsBones[1] << " in weight file!  Replacing value." << std::endl;
			defOpt.ptr_scene->graph_n = nVertsBones[1];
		}

		// Change the displayed frame
		if (defOpt.ptr_scene) {
			defOpt.ptr_scene->frameChangeTrigger = lastFrame;
		}		

		goto SolverFinished;
	}

	if (!defOpt.loadFinalShape.empty() && defOpt.ptr_scene) {
		// Load the final reconstructed shape
		finalShape = loadShape(QString(defOpt.ptr_scene->rootDir.c_str()), QString(defOpt.loadFinalShape.c_str()), *defOpt.ptr_scene, 0);
		processFinalShape();
		goto SolverFinished;
	}

	// Load solver state
	if (!defOpt.autoLoadSolverState.empty()) {
		drawMutex.lock();
		bool success = loadSolverState(defOpt.autoLoadSolverState.c_str(), defOpt.loadGraphOnly);
		drawMutex.unlock();

		if (success) {
			// Post-processing
			DeformAll_Utility::SparseTLArray allTfList;
			makeAllTFList(frameSolution, allTfList);
			PointList vRef, nRef;
			allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

			discardSmallRegions(vRef);				/**< Discard small regions. This initializes the "excludedLabels" array. */
			estimateJoints();
			selectSolver_SlidingWindow(defOpt.slidingWindowSize);

			//// TESTING
			//cleanASGComponents();
			//pauseUntilEnter("Cleaning Test Done");
			//// END TESTING

			// Change the displayed frame
			if (defOpt.ptr_scene) {
				defOpt.ptr_scene->frameChangeTrigger = lastFrame;
			}

			//// (Testing! For taking screenshots) Build a cleaned version of the ASG
			//rebuildASG(10.0, 4.0*defOpt.refSampleDist, false, false);
			//cleanASG();

			recomputeClosestPtError_AllFrames = !defOpt.accumulateASGOnReference;

			printSolverStats();

			////////////////////////////////////////////////////////////////////////////
			// Testing! For computing total E_fit value
			if (defOpt.computeTotalEfit)
			{
				real_type totalEfit, totalEfit_count;
				findTotal_Efit(totalEfit, totalEfit_count);

				// Really pause after computing E_fit
				std::cin.clear();
				std::cout << "Total E_fit (value/count/avg): " << totalEfit << ", " << totalEfit_count << ", " << totalEfit / totalEfit_count << " (Press enter/return to continue)...";
				std::cin.ignore(1,0);
			}

			////////////////////////////////////////////////////////////////////////////
			// Save joint and transformation information to a text file
			while (false)
			{
				real_type newMinEdgeRatio, newMaxJointDist;
				cout << "Enter new value for min incident edge ratio (currently " << defOpt.pruneJoint_minEdgeRatio << "): ";
				cin >> newMinEdgeRatio;
				cout << "Enter new value for max joint dist (currently " << defOpt.maxJointDist << "): ";
				cin >> newMaxJointDist;

				// Re-estimate joints
				defOpt.pruneJoint_minEdgeRatio = newMinEdgeRatio;
				defOpt.maxJointDist = newMaxJointDist;
				estimateJoints();

				// Re-write data
				ofstream outFile("D:\\recon_jointTfData.txt");
				if (outFile) {
					printJointTransformInfo(outFile);
				}
			}

			//writeTransformedFrames();
			//writeTransformedDSG();

			//// Really pause
			//std::cin.clear();
			//std::cout << "Loaded Solver State: " << " (Press enter/return to continue)...";
			//std::cin.ignore(1,0);

			////////////////////////////////////////////////////////////////////////////
			pauseUntilEnter("Loaded Solver State");
			goto ResumeSolver;			/**< Jump to where we last saved the solver state */
		}
		else {
			goto SolverFinished;
		}
	}
	else {
		pauseUntilEnter("Starting Registration");
	}

	while (isValidFrame(lastFrame+1)) {
		int nextFrame = lastFrame+1;

		// Initial registration
		solveStatus = SOLVER_INIT;
		MEASURE_PERF_NARG(TIMER_MULTIFRAME_INITIALIZE,AddStart);
		if (!initNewFrame(nextFrame)) {
			MEASURE_PERF_NARG(TIMER_MULTIFRAME_INITIALIZE,Stop);
			break;
		}
		MEASURE_PERF_NARG(TIMER_MULTIFRAME_INITIALIZE,Stop);

		// Change the displayed frame
		if (defOpt.ptr_scene) {
			defOpt.ptr_scene->frameChangeTrigger = nextFrame;
		}

		//pauseUntilEnter("Init done");

		cout << "-------------------------------------------------------------------------------" << endl;
		cout << DeformGraphAll_Utility::printHeader << "Optimizing global registration..." << endl;
		MEASURE_PERF_NARG(TIMER_MULTIFRAME_GLOBAL_REG,AddStart);

		// Select subset of frames to solve (sliding window)
		selectSolver_SlidingWindow(defOpt.slidingWindowSize);

		taucsType TStepRes = 0.0, lastTStepRes = 0.0;
		int numItr = 0;
		while (true) {
			solveStatus = SOLVER_T_STEP;

			if (recomputeClosestPtError_AllFrames) {
				// W-step had a global affect; update closest points in all frames.
				findClosestPoints_AllFrames(true);
			}

			// T-step
			TStepRes = findTransformations(CORRESPONDENCE_CLOSEST_POINTS);

			if (!defOpt.accumulateASGOnReference) {
				// (Testing) Re-match occluded parts
				DeformGraphAll_Utility::FrameSolution* newSoln = getFrameSoln(nextFrame);
				fillOccludedTfs(jointMap, *newSoln, defOpt, frameSolution);
				matchOccludedParts(newSoln);
			}

			// Determine stop conditions
			cout << "(Overall) " << numItr << " / " << defOpt.maxSolverIterations;
			cout << " (Residual=" << TStepRes << ", Diff=" << abs(TStepRes - lastTStepRes) << ")";
			cout << endl;
			if ((abs(TStepRes - lastTStepRes) < defOpt.stopEpsilon_sqRt*(1+TStepRes))) {
				cout << "(Overall) done! (objective)" << endl;
				break;
			}
			if ((defOpt.maxSolverIterations > 0) && (numItr+1 >= defOpt.maxSolverIterations)) {
				cout << "(Overall) done! (max iteration)" << endl;
				break;
			}
			++numItr;
			lastTStepRes = TStepRes;

			// W-step
			solveStatus = SOLVER_W_STEP;
			findSampleLabels();

			//pauseUntilEnter("temp");
			//detectOcclusions(nextFrame, newSoln->occludedLabels);
			//pauseUntilEnter("temp");
		}
		// Fill in occluded transformations in all frames
		fillOccludedTfs_AllFrames();
		MEASURE_PERF_NARG(TIMER_MULTIFRAME_GLOBAL_REG,Stop);

		//pauseUntilEnter("Global Reg Done");

		// Rebuild the ASG
		solveStatus = SOLVER_UPDATE_SAMPLES;
		if (defOpt.performRebuildASG) {
			MEASURE_PERF_NARG(TIMER_MULTIFRAME_RESAMPLE_ASG,AddStart);

			if (defOpt.accumulateASGOnReference) {
				accumulateASG(defOpt.DSGMaxEdgeStretchFactor, defOpt.DSGMaxEdgeLen, true);
			}
			else {
				rebuildASG(defOpt.DSGMaxEdgeStretchFactor, defOpt.DSGMaxEdgeLen, false, true);

				// Clean any small stray vertex components as a post-process
				cleanASGComponents(5);

				// During rebuilding, the samples themselves have changed, so we cleared all the closest point / error data.
				// Therefore, we need to recompute these again on all frames.
				recomputeClosestPtError_AllFrames = true;
			}

			MEASURE_PERF_NARG(TIMER_MULTIFRAME_RESAMPLE_ASG,Stop);
		}

		// Move to next frame
		lastFrame = nextFrame;

		// Save solver state
		if (!defOpt.autoSaveSolverState.empty()) {
			// Save to a different file for every frame, if the string contains '#' symbols
			string saveFile = defOpt.autoSaveSolverState;
			replaceField(saveFile, '#', lastFrame);

			// Save state
			saveSolverState(saveFile.c_str());
			printSolverStats();
		}

		//pauseUntilEnter("Moving to next frame");

		// Resume solver here (right where we saved the solver state)
ResumeSolver: NULL;

		//// TESTING: REMOVE THIS LINE AFTER USE
		//break;
	}

	solveStatus = SOLVER_IDLE;

	// Perform post-processing
	postProcess();

	//////////////////////////////////////////////////////////////////////////
SolverFinished:
	iterationFinished = true;
	Notify();						/**< Notify the viewer to update the display. */
}

void Deform::DeformGraphAll::postProcess()
{
	// 1. Build an ASG with very small sample spacing, and prune ALL edges that stretch
	defOpt.refSampleDist = defOpt.finalRefSampleDist;
	rebuildASG(defOpt.finalDSGMaxEdgeStretchFactor, defOpt.finalDSGMaxEdgeLen, true, false);

	//// 2. Clean ASG
	//cleanASG();

	// Save dense ASG to file
	if (!defOpt.autoSaveSolverState.empty()) {
		// Save to a different file for every frame, if the string contains '#' symbols
		string saveFile = defOpt.autoSaveSolverState;
		replaceField(saveFile, '#', lastFrame+1);

		// Save state
		saveSolverState(saveFile.c_str());
		printSolverStats();


		// Output the list of points in ASCII format
		PointList vRef, nRef;
		DeformAll_Utility::SparseTLArray allTfList;
		makeAllTFList(frameSolution, allTfList);
		allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

		// Compute the bounding box of the points
		real3_type box_min, box_max;
		for (int d = 0; d < 3; d++) {
			box_min[d] = REAL_TYPE_MAX;
			box_max[d] = -REAL_TYPE_MAX;
		}
		for (int i = 0; i < vRef.size(); ++i) {
			for(int d = 0; d < dim; d++){
				// Calculate the bounding box of the mesh
				if (vRef[i][d] < box_min[d]) {
					box_min[d] = vRef[i][d];
				}
				if (vRef[i][d] > box_max[d]) {
					box_max[d] = vRef[i][d];
				}
			}
		}

		// Write formats
		std::string ptsFile = defOpt.autoSaveSolverState;
		ptsFile.insert(ptsFile.length() - 5, "_pts");
		replaceField(ptsFile, '#', lastFrame);
		writePtsFormat(ptsFile.c_str(), box_min, box_max, vRef, nRef);

		ptsFile = defOpt.autoSaveSolverState;
		ptsFile.insert(ptsFile.length() - 5, "_bnpts");
		replaceField(ptsFile, '#', lastFrame);
		write_bnpts_Format(ptsFile.c_str(), vRef, nRef);
	}

	//ofstream out;
	//out.open(ptsFile.c_str());
	//if (out.fail()) {
	//	cout << "Error in DeformGraphAll::postProcess(): could not open file " << ptsFile << " for writing!" << endl;
	//}
	//else {
	//	cout << DeformGraphAll_Utility::printHeader << "Saving points to " << ptsFile;
	//	for (int i = 0; i < vRef.size(); ++i) {
	//		const real3_type& pt = vRef[i];
	//		out << pt[0] << " " << pt[1] << " " << pt[2] << endl;
	//	}
	//	out.close();
	//	cout << "...done!" << endl;
	//}


	//// Testing
	//// Update the closest point correspondences
	//findClosestPoints_AllFrames(true);
	//pauseUntilEnter("Closest points updated!");
	//findSmoothWeights();
	//pauseUntilEnter("Smooth weights found!");
}

void Deform::DeformGraphAll::processFinalShape()
{
	bool success = false;
	if (!defOpt.autoLoadSolverState.empty()) {
		drawMutex.lock();
		success = loadSolverState(defOpt.autoLoadSolverState.c_str(), false);
		drawMutex.unlock();
	}

	// If we loaded the ASG properly
	if (finalShape && success) {
		// Estimate joints based on solved frame motion
		estimateJoints();

		// Clean up graph
		rebuildASG(defOpt.DSGMaxEdgeStretchFactor, defOpt.DSGMaxEdgeLen, false, false);
		cleanASGWeights(20);
		cleanASGWeights(20);

		// Initialize deformed final shape and weights
		finalShapeDef = new DeformAll_Utility::ShapeDeformed(*finalShape);
		finalShapeWeights = new DeformAll_Utility::RefWeightList;

		// 1. Extrapolate the labels onto the shape
		DeformAll_Utility::SparseTLArray allTfList;
		makeAllTFList(frameSolution, allTfList);
		PointList vRef, nRef;
		allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);
		discardSmallRegions(vRef);				/**< Discard small regions. This initializes the "excludedLabels" array. */

		const int nVerts = finalShape->numVertices();
		std::vector<int> finalLabels(nVerts);

		// Strategy 1: Just use nearest neighbor
		for (int i = 0; i < nVerts; ++i) {
			// Assign the label of the closest ASG sample
			idxList nearest;
			distList dists;
			vRef.queryKNN(finalShape->vertex[i], 1, nearest, dists, true);
			finalLabels[i] = findLabel(allSampleGraph->weightList[nearest[0]]);
		}

		//// Strategy 2: Use weight extrapolation
		//const int numBones = defOpt.numBones;
		//std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
		//for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
		//	int label = findLabel(allSampleGraph->weightList[i]);
		//	vRef_bone[label].add(vRef[i]);
		//	nRef_bone[label].add(nRef[i]);
		//}
		//DeformAll_Utility::TransformList tfList(defOpt.numBones);
		//std::vector<bool> isOccluded(defOpt.numBones, false);
		//idxList newSampleIndices(nVerts);
		//for (int i = 0; i < nVerts; ++i) {
		//	newSampleIndices[i] = i;
		//}
		//extrapolateWeights(vRef_bone, nRef_bone, isOccluded, finalShape->vertex, newSampleIndices, tfList, defOpt.refSampleDist, finalLabels);

		// 2. Translate the labels to finalShapeWeights
		finalShapeWeights->resize(nVerts);
		for (int i = 0; i < nVerts; ++i) {
			if (finalLabels[i] >= 0) {
				(*finalShapeWeights)[i].bones.push_back(finalLabels[i]);
				(*finalShapeWeights)[i].weights.push_back(1.0f);
			}
		}

		// Change the displayed frame
		if (defOpt.ptr_scene) {
			defOpt.ptr_scene->frameChangeTrigger = lastFrame;
		}
	}
}

bool Deform::DeformGraphAll::loadInitialization( int index, const DeformGraphAll_Utility::FrameSolution *lastSoln, const DeformGraphAll_Utility::FrameSolution *newSoln )
{
	if (matchCandDisplay && (index == matchCandDisplayIndex)) {
		return true;
	}

	if (defOpt.ptr_scene->autoLoadGCMatchResult.empty()) {
		return false;
	}

	cout << DeformGraphAll_Utility::printHeader << "Loading GC initial correspondences for frame " << index << "..." << endl;

	// If the string contains '#' symbols, substitute the actual number
	string saveFile = defOpt.ptr_scene->autoLoadGCMatchResult.c_str();
	replaceField(saveFile, '#', index);

	// Check if the file exists
	if (QFile::exists(QString(saveFile.c_str()))) {
		if (matchCandDisplay && (index != matchCandDisplayIndex)) {
			delete matchCandDisplay;
		}

		// Graph-cut initial correspondences between lastFrame and newFrame
		matchCandDisplay = new MatchCandidates(const_cast<ShapeModel&>(lastSoln->originalShape), const_cast<ShapeModel&>(newSoln->originalShape));

		// Load the correspondence result
		matchCandDisplay->loadGraphCutResult(saveFile.c_str());

		// Error checking
		if (lastSoln->originalShape.numVertices() != matchCandDisplay->graphCutLabels_source.size() || 
			newSoln->originalShape.numVertices() != matchCandDisplay->graphCutLabels_target.size()) {
				cerr << "Error in DeformGraphAll::initNewFrame(): Graph cut result doesn't match with loaded shapes!" << endl;
				delete matchCandDisplay;
				matchCandDisplay = NULL;
				return false;
		}

		matchCandDisplayIndex = index;
		return true;
	}
	else {
		return false;
	}
}

void Deform::DeformGraphAll::draw( const DrawOptions &options ) const
{
	drawMutex.lock();

	if (options.displayFrame > referenceFrameIndex && (options.showInitSrc || options.showInitResult)) {
		int drawFrameIndex = options.displayFrame;
		const DeformGraphAll_Utility::FrameSolution *drawSolution1 = getFrameSoln(drawFrameIndex - 1);
		const DeformGraphAll_Utility::FrameSolution *drawSolution2 = getFrameSoln(drawFrameIndex);
		PointList noPts;

		// Draw target frame
		drawSolution2->draw(noPts, options, true);

		if (options.showInitSrc) {
			// Draw the current frame and the previous frame

			// Draw source frame
			drawSolution1->draw(noPts, options, false);
		}
		else if (options.showInitResult && const_cast<Deform::DeformGraphAll*>(this)->loadInitialization(drawFrameIndex-1, drawSolution1, drawSolution2)) {
			// Visualize the graph cut initialization result
			const std::vector<index_type> *p_labels;
			const ShapeModel *p_shape;

			if (!defOpt.ptr_scene) {
				cerr << "Pointer to scene needed!" << endl << flush;
			}

			// SHOW_SOURCE_TO_TARGET
			p_labels = &(matchCandDisplay->graphCutLabels_source);
			p_shape = &(matchCandDisplay->sourceModel);
			//// SHOW_TARGET_TO_SOURCE
			//p_labels = &(matchCandDisplay->graphCutLabels_target);
			//p_shape = &(matchCandDisplay->targetModel);

			size_type numRegions = matchCandDisplay->sortedTFList.size();
			if (defOpt.ptr_scene->clusterColors.size() < numRegions) {
				// If not enough colors, generate more
				defOpt.ptr_scene->generateClusterColors(numRegions);
			}

			// If it's a mesh, color the entire mesh
			if (p_shape->shapeType == SHAPE_MESH) {
				glEnable(GL_LIGHTING);
				glEnable(GL_BLEND);
				glEnable(GL_CULL_FACE);
				glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
				glCullFace(GL_BACK);
				//glDepthMask (GL_FALSE);
				glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

				Mesh3_real &mesh = *((Mesh3_real*)p_shape);

				glBegin(GL_TRIANGLES);
				for(size_type t = 0, t_end = mesh.triangle_vertex_index.size(); t < t_end; t++) {
					const index3_type& tvi = mesh.triangle_vertex_index[t];

					// Check that all three labels not dummy
					bool noDummy = true;
					for(size_type i = 0;i < 3;i++) {
						index_type label = (*p_labels)[tvi[i]];

						if (label >= numRegions) {
							noDummy = false;
						}
					}

					// Draw the triangle
					if (noDummy) {
						for(size_type i = 0;i < 3;i++) {
							index_type label = (*p_labels)[tvi[i]];
							real3_type v = mesh.vertex[tvi[i]];
							real3_type n = mesh.normal[tvi[i]];

							//// Apply inverse transformation = SHOW_TARGET_TO_SOURCE
							//v = matchCandDisplay->sortedInverseTFList[label] * v;
							//n = matchCandDisplay->sortedInverseTFList[label].rot * n;

							// Apply forward transformation	= SHOW_SOURCE_TO_TARGET
							v = matchCandDisplay->sortedTFList[label] * v;
							n = matchCandDisplay->sortedTFList[label].rot * n;

							setOpenGLMaterialColor(defOpt.ptr_scene->getClusterColor(label), 0.7f);
							glNormal3f(n[0],n[1],n[2]);
							glVertex3f(v[0],v[1],v[2]);
						}
					}
				}
				glEnd();

				//glDepthMask (GL_TRUE);
				glDisable(GL_BLEND);
				glDisable(GL_CULL_FACE);
			}
		}

		drawMutex.unlock();
		return;
	}

	// Create samples on frame to display
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int drawFrameIndex = it->first;

		// Except for displaying all frames, display only the selected frame.
		if (options.showAllFrames || (options.displayFrame == drawFrameIndex)) {
			const DeformGraphAll_Utility::FrameSolution *drawSolution = it->second;

			// Draw input frames
			PointList noPts;
			drawSolution->draw(noPts, options, (drawFrameIndex == referenceFrameIndex));

			// Draw Joints
			if (options.showJoints) {
				drawJoints(-drawSolution->tfList, options);
			}

			// Testing
			if (options.showDeformedModel && tempPts1 && tempGraph1) {
				glDisable(GL_LIGHTING);
				glColor3f(0.0,0.0,0.0);
				drawBoostGraph(*tempGraph1, *tempPts1);
			}
			if (options.showTargetModel && tempPts2 && tempGraph2) {
				glDisable(GL_LIGHTING);
				glColor3f(0.0,0.0,0.0);
				drawBoostGraph(*tempGraph2, *tempPts2);
			}

			// Draw deformed external mesh
			if (!options.showAllFrames && options.showExternalMeshDeformed && finalShapeDef && finalShapeWeights) {
				if (lastDefFrame_finalShape != drawFrameIndex) {
					finalShapeDef->deform(*finalShapeWeights, drawSolution->tfList, true);

					std::vector<bool> &valid = finalShapeDef->getVertexValid();

					// Reset the valid bit
					for (int i = 0; i < finalShapeDef->getVertex().size(); ++i) {
						valid[i] = true;
					}

					cout << "Pruning edges with length greater than: " << defOpt.edgeLenStretchThreshold_Sq << endl;

					// Since there are artifacts when you deform, remove all vertices that are part of an edge exceeding a length threshold
					DeformAll_Utility::checkEdgeLengths(finalShapeDef->getVertex(), finalShapeDef->getOriginalShape(), defOpt.edgeLenStretchThreshold_Sq, defOpt.deformCosAngleThreshold, valid);
				}

				DeformAll_Utility::drawShapeWeightsValid(
					finalShapeDef->getVertex(), finalShapeDef->getNormal(), finalShapeDef->getOriginalShape(),
					*finalShapeWeights, finalShapeDef->getVertexValid(),
					options.targetModelColor, (*options.boneColors), options, options.showInvalidVertices, true);

				// Update last displayed frame
				lastDefFrame_finalShape = drawFrameIndex;
			}
			else {
				if (true || lastDefFrame_ASG != drawFrameIndex || num_vertices(allSampleGraph->graph) != drawVList.size()) {
					// Regenerate ASG samples on this frame
					allSampleGraph->getVertexNormal_Frame(drawFrameIndex, allTfList, drawVList, drawNList);
				}

				// Draw this frame
				if (drawFrameIndex == referenceFrameIndex) {
					if (options.showCellError) {
						// Draw total error for all frames
						if (sumError.size() == defOpt.numBones) {
							drawError(drawVList, sumError, options);
						}
					}
				}
				else {
					if (options.showCellError) {
						// Draw error for each sample, each label
						DeformGraphAll_Utility::FrameError_LabelSample::const_iterator it2 = allFrameError.find(drawFrameIndex);
						if (it2 != allFrameError.end()) {
							const DeformGraphAll_Utility::Error_LabelSample& frameError = it2->second;
							drawError(drawVList, frameError, options);
						}
					}

					if (options.showSampleError && (drawSolution->sampleError.size() == allSampleGraph->numSamples())) {
						// Draw error for each sample
						drawSampleError(drawVList, drawSolution->sampleError, options);
					}
				}

				if (options.showTargetCorresp) {
					drawCorresp(drawVList, drawFrameIndex, drawSolution->originalShape, options);
				}

				drawASG(drawVList, drawNList, options.showGraph, options.showSampledPoints, options.showInvalidVertices, options.showShapeWeights, options.showComponent, *(options.boneColors));

				// Update last displayed frame
				lastDefFrame_ASG = drawFrameIndex;
			}
		}
	}

	drawMutex.unlock();
}

void Deform::DeformGraphAll::drawASG( const PointList& points, const PointList& normals, bool showGraph, bool showTest, bool showRemovedEdges, bool showColors, int showComp, const std::vector<RealColor> &boneColors ) const
{
	if (allSampleGraph) {
		glDisable(GL_LIGHTING);

		if (showTest) {
			// (Testing)
			glPointSize(8.0);
			glColor3f(1.0,1.0,0.0);
			PointList_Utility::draw(allSampleGraph->vertex_disp, POINT_SQUARES, allSampleGraph->testIndex);

			// (Testing)
			glPointSize(5.0);
			glColor3f(0.0,0.0,1.0);
			PointList_Utility::draw(allSampleGraph->vertex_disp, POINT_SQUARES);

			// (Testing)
			glPointSize(3.0);
			glColor3f(1.0,0.0,0.0);
			PointList_Utility::draw(allSampleGraph->normal_disp, POINT_SQUARES);
		}

		if (showRemovedEdges) {
			glLineWidth(4.0);
			glColor3f(1.0,0.0,0.0);
			glBegin(GL_LINES);
			std::list< std::pair<int,int> >::const_iterator it;
			for (it = allSampleGraph->testList.begin(); it != allSampleGraph->testList.end(); ++it) {
				const real3_type& pt1 = points[it->first];
				const real3_type& pt2 = points[it->second];

				glVertex3f(pt1[0], pt1[1], pt1[2]);
				glVertex3f(pt2[0], pt2[1], pt2[2]);
			}
			glEnd();
		}

		if (showGraph) {
			glLineWidth(1.0);
			glPointSize(5.0);
			if (showColors) {
				glEnable(GL_LIGHTING);
				glEnable(GL_COLOR_MATERIAL);
				drawWeightedBoostGraph(allSampleGraph->graph, points, normals, allSampleGraph->weightList, boneColors, showComp);
				glDisable(GL_COLOR_MATERIAL);
			}
			else {
				glColor3f(0.0,0.0,1.0);
				drawBoostGraph(allSampleGraph->graph, points);
			}
		}

		if (false)
		{
			if (allSampleGraph->vertex_disp.size() != allSampleGraph->numSamples())
			{
				DeformAll_Utility::SparseTLArray allTfList;
				makeAllTFList(frameSolution, allTfList);
				allSampleGraph->getVertexNormal_Ref(allTfList, allSampleGraph->vertex_disp, allSampleGraph->normal_disp);
			}

			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			glColor3f(1.0, 1.0, 0.0);
			PointList_Utility::draw(allSampleGraph->vertex_disp, POINT_SPHERES, 0.5);
		}
	}
}

void Deform::DeformGraphAll::drawJoints( const DeformAll_Utility::TransformList& tfList, const DrawOptions& options ) const
{
	const int numBones = defOpt.numBones;
	const real_type jointSize = defOpt.sceneGridSpacing * 3.0;

	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);
		const DeformGraphAll_Utility::Joint& joint = it->second;
		const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
		const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

		int numSamples = 0;
		if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
			numSamples = 1;
		}
		else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
			numSamples = defOpt.hingeJointSamples;
		}

		for (int j = 0; j < numSamples; ++j) {
			real3_type jointLocation;

			if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
				jointLocation = jointPos;
			}
			else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				real_type t = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
				jointLocation = jointPos +  t*jointCon;
			}

			//////////////////////////////////////////////////////////////////////////
			glEnable(GL_LIGHTING);
			if (options.showGraphDeformed) {
				setOpenGLMaterialColor((*options.boneColors)[bi]);
				drawSphere(tfList[bi] * jointLocation, jointSize);
				setOpenGLMaterialColor((*options.boneColors)[bj]);
				drawSphere(tfList[bj] * jointLocation, jointSize);
			}
			else {
				// Draw two smaller spheres
				jointLocation[0] -= 0.25 * jointSize;
				setOpenGLMaterialColor((*options.boneColors)[bi]);
				drawSphere(jointLocation, 0.5 * jointSize);

				jointLocation[0] += 0.5 * jointSize;
				setOpenGLMaterialColor((*options.boneColors)[bj]);
				drawSphere(jointLocation, 0.5 * jointSize);
			}
		}
	}
}

void Deform::DeformGraphAll::drawCorresp( const PointList& vList, int frameNumber, const ShapeModel& frameShape, const DrawOptions& options ) const
{
	// Show correspondences on this frame (draw source ASG points deformed to this frame)
	DeformGraphAll_Utility::AllClosestPoints::const_iterator it2 = sampleClosestPoints.find(frameNumber);
	if (it2 != sampleClosestPoints.end()) {
		// Draw the corresponding positions
		glDisable(GL_LIGHTING);
		glPointSize(8.0);
		glBegin(GL_POINTS);
		const DeformGraphAll_Utility::ClosestPointMap& closestPts = it2->second;
		DeformGraphAll_Utility::ClosestPointMap::const_iterator it3;
		for (it3 = closestPts.begin(); it3 != closestPts.end(); ++it3) {
			const int srcASGIdx = it3->first;
			const int tgtShapeIdx = it3->second;
			const real3_type& targetPoint = frameShape.vertex[tgtShapeIdx];
			setOpenGLColor(options.pointColor);
			glVertex3f(vList[srcASGIdx][0],vList[srcASGIdx][1],vList[srcASGIdx][2]);
			glColor3f(1.0,0.0,0.0);
			glVertex3f(targetPoint[0],targetPoint[1],targetPoint[2]);
		}
		glEnd();

		// Draw all of the transformed source points again, to indicate points with no correspondence
		glColor3f(0.0,0.0,0.0);
		PointList_Utility::draw(vList, POINT_SQUARES);
	}
}

void Deform::DeformGraphAll::drawSampleError( const PointList& vList, const distList& sampleError, const DrawOptions& options ) const
{
	const int numSamples = allSampleGraph->numSamples();

	glDisable(GL_LIGHTING);
	glPointSize(8.0);
	glBegin(GL_POINTS);
	for (int i = 0; i < numSamples; ++i) {
		real_type errorValue = sampleError[i];

		if (errorValue > 0.0) {
			setOpenGLColor(defOpt.colormap->getColor(errorValue));
			glVertex3f(vList[i][0],vList[i][1],vList[i][2]);
		}
	}
	glEnd();
}

void Deform::DeformGraphAll::drawError( const PointList& vList, const DeformGraphAll_Utility::Error_LabelSample& frameError, const DrawOptions& options ) const
{
	const int drawLabel = options.showComponent;
	const int numSamples = allSampleGraph->numSamples();

	glDisable(GL_LIGHTING);
	glPointSize(8.0);
	glBegin(GL_POINTS);
	for (int i = 0; i < numSamples; ++i) {
		real_type errorValue;
		if (drawLabel < 0 || drawLabel >= defOpt.numBones) {
			// Pick the currently assigned label
			errorValue = frameError[findLabel(allSampleGraph->weightList[i])][i];
		}
		else {
			errorValue = frameError[drawLabel][i];
		}

		if (errorValue > 0.0) {
			setOpenGLColor(defOpt.colormap->getColor(errorValue));
			glVertex3f(vList[i][0],vList[i][1],vList[i][2]);
		}
	}
	glEnd();
}

bool Deform::DeformGraphAll::initNewFrame( int newFrame )
{
	DeformGraphAll_Utility::FrameSolution *newSoln = makeFrameSoln(newFrame, false), *lastSoln = getFrameSoln(lastFrame);
	if (!newSoln) {
		return false;
	}

	MatchCandidates *matchCand = NULL;
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	if (defOpt.initGCMatchRef && defOpt.ptr_scene && defOpt.ptr_scene->autoSaveCorrespondenceResult.length() == 0)
	{
		cerr << DeformGraphAll_Utility::printHeader << "No save file given for computing initialization (use autoSaveCorrespondenceResult)!" << endl;
		return false;
	}
	else if (defOpt.initGCMatchRef && defOpt.ptr_scene) {
		// Graph-cut initial correspondences between lastFrame and newFrame
		matchCand = new MatchCandidates(const_cast<ShapeModel&>(lastSoln->originalShape), const_cast<ShapeModel&>(newSoln->originalShape));

		// Load frames and spin images for the last frame and new frame
		Scene& scene = *defOpt.ptr_scene;
		distList frameScales = scene.makeSceneSqRadii(scene.scaleFactors_PC);
		const_cast<ShapeModel&>(lastSoln->originalShape).computeFrames(frameScales);
		const_cast<ShapeModel&>(lastSoln->originalShape).computeSpinImages(scene.binSizeMultiple_SI, scene.imageWidth_SI, scene.maxSupportAngle_SI);
		const_cast<ShapeModel&>(newSoln->originalShape).computeFrames(frameScales);
		const_cast<ShapeModel&>(newSoln->originalShape).computeSpinImages(scene.binSizeMultiple_SI, scene.imageWidth_SI, scene.maxSupportAngle_SI);

		// Subsample locations on the accumulated reference frame and on the new frame
		const int nSample = scene.numCorrespondences;
		PointList lastSample(nSample), newSample(nSample);
		idxList lastSampleIdx = bestCandidateSample(lastSoln->originalShape.vertex, nSample);
		idxList newSampleIdx = bestCandidateSample(newSoln->originalShape.vertex, nSample);
		for (int i = 0; i < nSample; ++i) {
			lastSample[i] = lastSoln->originalShape.vertex[lastSampleIdx[i]];
			newSample[i] = newSoln->originalShape.vertex[newSampleIdx[i]];
		}

		// Create graphs
		BoostGraph lastGraph, newGraph;
		makeEKNNgraph(defOpt.EKNNGraph_k, lastSample, lastGraph);
		makeEKNNgraph(defOpt.EKNNGraph_k, newSample, newGraph);

		// Compute the result
		scene.solveGraphCutCorrespondences(matchCand, &lastSample, &newSample, &lastGraph, &newGraph);
		matchCand->saveGraphCutResult(scene.autoSaveCorrespondenceResult.c_str());

		// Display the result
		scene.matchCandidates = matchCand;

		// Compute total time
		double total = 0.0;
		total += MEASURE_PERF_NARG(TIMER_SUB_SAMPLING_GEOMETRY,GetDurationInSecs);
		total += MEASURE_PERF_NARG(TIMER_POPULATING_CORRESPONDENCE_CANDIDATES,GetDurationInSecs);
		total += MEASURE_PERF_NARG(TIMER_MEAN_SHIFT_CLUSTERING,GetDurationInSecs);
		total += MEASURE_PERF_NARG(TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP,GetDurationInSecs);
		total += MEASURE_PERF_NARG(TIMER_PRECOMPUTING_CLOSEST_POINTS,GetDurationInSecs);
		total += MEASURE_PERF_NARG(TIMER_GRAPH_CUT_OPTIMIZATION,GetDurationInSecs);
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "Total time: " << total << " secs" << endl;
		cout << "-------------------------------------------------------------------------------" << endl;

		// Save the timings into a file
		std::string timingFile = scene.autoSaveCorrespondenceResult;
		timingFile = timingFile.substr(0,timingFile.length()-5) + std::string("_timing.txt");
		cout << "Saving timing to: " << timingFile << endl;
		ofstream out;
		out.open(timingFile.c_str());
		if (out.fail()) {
			cout << "Error: could not open file " << timingFile << " for writing!" << endl;
		}
		else {
			out << "-------------------------------------------------------------------------------" << endl;
			out << "Running time summary" << endl;
			printTime(out, TIMER_SUB_SAMPLING_GEOMETRY);
			printTime(out, TIMER_POPULATING_CORRESPONDENCE_CANDIDATES);
			printTime(out, TIMER_MEAN_SHIFT_CLUSTERING);
			printTime(out, TIMER_EXTRACTING_DENSE_CORRESPONDENCES_ICP);
			printTime(out, TIMER_PRECOMPUTING_CLOSEST_POINTS);
			printTime(out, TIMER_GRAPH_CUT_OPTIMIZATION);
			out << "-------------------------------------------------------------------------------" << endl;
			out << "Total time: " << total << " secs" << endl;
			out << "-------------------------------------------------------------------------------" << endl;
			out.close();
		}

		//pauseUntilEnter("Finished Graph Cut Correspondences");

		// This will trigger a screenshot and exit the program entirely
		defOpt.ptr_scene->triggerGCMatchFinished = true;

		// Exit the initialization
		return false;

		////// Gather the geometry on the reference frame
		////Scene& scene = *defOpt.ptr_scene;
		////ShapeModel refShape(SHAPE_POINTS);
		////ShapeModel& newShape = const_cast<ShapeModel&>(newSoln->originalShape);
		////PointList lastPts, lastNormals;
		////const int numSamples = allSampleGraph->numSamples();
		////allSampleGraph->getVertexNormal_Ref(allTfList, refShape.vertex, refShape.normal);
		////allSampleGraph->getVertexNormal_Frame(lastFrame, allTfList, lastPts, lastNormals);
		////refShape.setSampleResolution(defOpt.sceneGridSpacing);
		////refShape.markVertex.resize(numSamples, VERTEX_VALID);

		////// Load file format
		////std::string siFile = defOpt.autoSaveSolverState;
		////std::string curvFile = defOpt.autoSaveSolverState;
		////std::string GCFile = defOpt.autoSaveSolverState;
		////siFile.insert(siFile.length() - 5, "_SI");
		////curvFile.insert(curvFile.length() - 5, "_CV");
		////GCFile.insert(GCFile.length() - 5, "_GC");
		//////cout << DeformGraphAll_Utility::printHeader << "Spin image save file is: " << siFile << endl;
		//////cout << DeformGraphAll_Utility::printHeader << "Curv info save file is: " << curvFile << endl;


		////// Load the frames and spin images
		////refShape.frames.resize(numSamples);
		////refShape.spinImageStack.resize(numSamples);
		////for (int i = 0; i < numSamples; ++i) {
		////	DeformGraphAll_Utility::FrameSolution* frameSoln = getFrameSoln(allSampleGraph->fIndex[i]);
		////	const SpinImageStack& sampleSI = frameSoln->getSpinImages(siFile.c_str(), scene);
		////	const CurvInfoList& sampleCurv = frameSoln->getCurvInfo(curvFile.c_str(), scene);
		////	const distList *curvScales = frameSoln->getCurvScales();

		////	refShape.frames[i] = sampleCurv[allSampleGraph->sIndex[i]];
		////	refShape.spinImageStack[i] = sampleSI[allSampleGraph->sIndex[i]];

		////	if (!refShape.frameScales.size() && curvScales) {
		////		refShape.frameScales = *curvScales;

		////		cout << DeformGraphAll_Utility::printHeader << "Computed curvature scales: [";
		////		for (int j = 0; j < curvScales->size(); ++j)
		////		{
		////			if (j != 0) cout << ",";
		////			cout << (*curvScales)[j];
		////		}
		////		cout << "]" << endl;
		////	}
		////	if (!refShape.spinImageStack.binSize) {
		////		refShape.spinImageStack.copyParams(sampleSI);

		////		cout << "Loaded spin image features..." << endl;
		////		cout << "*** Bin Size: " << refShape.spinImageStack.binSize << endl;
		////		cout << "*** Image Width: " << refShape.spinImageStack.imageWidth << endl;
		////		cout << "*** Support Angle: " << refShape.spinImageStack.supportAngle << " degrees" << endl;
		////		cout << "*** Expected overlap: " << refShape.spinImageStack.lambda << " bins" << endl;
		////	}
		////}

		////// Load frames and spin images for the new shape
		////distList frameScales = scene.makeSceneSqRadii(scene.scaleFactors_PC);
		////newShape.computeFrames(frameScales);
		////newShape.computeSpinImages(scene.binSizeMultiple_SI, scene.imageWidth_SI, scene.maxSupportAngle_SI);

		////// Subsample locations on the accumulated reference frame and on the new frame
		////const int nRef = defOpt.ptr_scene->numCorrespondences;
		////const int nNew = defOpt.ptr_scene->numCorrespondences;
		////PointList refSample(nRef), lastSample(nRef), newSample(nNew);
		////DeformAll_Utility::SparseWeightList refWeights(nRef);
		////idxList refSampleIdx = bestCandidateSample(refShape.vertex, nRef);
		////idxList newSampleIdx = bestCandidateSample(newSoln->originalShape.vertex, nNew);
		////for (int i = 0; i < nRef; ++i) {
		////	refSample[i] = refShape.vertex[refSampleIdx[i]];
		////	lastSample[i] = lastPts[refSampleIdx[i]];
		////	refWeights[i] = allSampleGraph->weightList[refSampleIdx[i]];
		////}
		////for (int i = 0; i < nNew; ++i) {
		////	newSample[i] = newSoln->originalShape.vertex[newSampleIdx[i]];
		////}

		////// Create graphs
		////BoostGraph refGraph, newGraph;
		////makeEKNNgraph(defOpt.EKNNGraph_k, refSample, refGraph);
		////makeEKNNgraph(defOpt.EKNNGraph_k, newSample, newGraph);

		////// Prune edges of graph
		////DeformGraphAll_Utility::pruneEdges(refSample, refWeights, refGraph, allTfList, 2.0, REAL_TYPE_MAX);

		////drawMutex.lock();
		////tempPts1 = &refSample;
		////tempPts2 = &newSample;
		////tempGraph1 = &refGraph;
		////tempGraph2 = &newGraph;
		////drawMutex.unlock();


		////// Graph-cut initial correspondences between lastFrame and newFrame
		////cout << DeformGraphAll_Utility::printHeader << "Computing GC initial correspondences..." << endl;

		////// Graph-cut initial correspondences between lastFrame and newFrame
		////matchCand = new MatchCandidates(refShape, newShape);
		////defOpt.ptr_scene->matchCandidates = matchCand;		/**< (Testing) Display the result */
		////pauseUntilEnter("Solving GCMatch");
		////defOpt.ptr_scene->solveGraphCutCorrespondences(matchCand, &refSample, &newSample, &refGraph, &newGraph);
		//////matchCand = new MatchCandidates(newShape, refShape);
		//////defOpt.ptr_scene->matchCandidates = matchCand;		/**< (Testing) Display the result */
		//////defOpt.ptr_scene->solveGraphCutCorrespondences(matchCand, &newSample, &refSample, &newGraph, &refGraph);

		////// Save result
		////replaceField(GCFile, '#', lastFrame);
		////matchCand->saveGraphCutResult(GCFile.c_str());


		////// We have a labeling directly on the ASG
		////// --> Propagate the information to initialize the new frame
		////if (lastFrame == referenceFrameIndex) {
		////	DeformGraphAll_Utility::propogateInitialLabels_Ref(refShape.vertex, defOpt.numBones, *matchCand, *allSampleGraph, newSoln->tfList);

		////	PointList vNew, nNew;
		////	allSampleGraph->getVertexNormal_Frame(newFrame, allTfList, vNew, nNew);
		////	discardSmallRegions(vNew);
		////}
		////else {
		////	DeformGraphAll_Utility::propogateInitialLabels(*matchCand, *allSampleGraph, newSoln->tfList);
		////}

		////pauseUntilEnter("Applied initial transformations");

		////drawMutex.lock();
		////tempPts1 = NULL;
		////tempPts2 = NULL;
		////tempGraph1 = NULL;
		////tempGraph2 = NULL;
		////drawMutex.unlock();
	}
	else {
		cout << DeformGraphAll_Utility::printHeader << "Loading GC initial correspondences..." << endl;

		// If the string contains '#' symbols, substitute the actual number
		string saveFile = defOpt.ptr_scene->autoLoadGCMatchResult.c_str();
		replaceField(saveFile, '#', lastFrame);

		// Check if the file exists
		if (QFile::exists(QString(saveFile.c_str()))) {
			// Graph-cut initial correspondences between lastFrame and newFrame
			matchCand = new MatchCandidates(const_cast<ShapeModel&>(lastSoln->originalShape), const_cast<ShapeModel&>(newSoln->originalShape));

			// Load the correspondence result
			matchCand->loadGraphCutResult(saveFile.c_str());

			// Error checking
			if (lastSoln->originalShape.numVertices() != matchCand->graphCutLabels_source.size() || 
				newSoln->originalShape.numVertices() != matchCand->graphCutLabels_target.size()) {
					cerr << "Error in DeformGraphAll::initNewFrame(): Graph cut result doesn't match with loaded shapes!" << endl;
					delete matchCand;
					return false;
			}


			// Propagate initial labels/transformations
			if (lastFrame == referenceFrameIndex) {
				DeformGraphAll_Utility::propogateInitialLabels_Ref(lastSoln, *matchCand, *allSampleGraph, newSoln->tfList);

				// Discard labels with small regions
				PointList vNew, nNew;
				allSampleGraph->getVertexNormal_Frame(newFrame, allTfList, vNew, nNew);
				discardSmallRegions(vNew);
			}
			else {
				DeformGraphAll_Utility::ClosestPointMap closestPts;
				//std::vector<bool> isOccluded = lastSoln->isBoneOccludedOrExcluded(excludedLabels);
				// Don't consider any bone occluded in this case
				std::vector<bool> isOccluded(defOpt.numBones, false);
				DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, lastFrame, lastSoln->originalShape, lastSoln->frameSamples, isOccluded, closestPts, defOpt);
				DeformGraphAll_Utility::propogateInitialLabels(*matchCand, *allSampleGraph, closestPts, newSoln->tfList);

				// Concatenate the transformations from lastSoln
				newSoln->tfList = lastSoln->tfList * newSoln->tfList;

				//// Just use the last solution's pose as the initialization
				//newSoln->tfList = lastSoln->tfList;
			}
		}
		else {
			if (lastFrame == referenceFrameIndex) {
				// TODO: Randomly subdivide the first frame

				// Discard labels with small regions
				PointList vNew, nNew;
				allSampleGraph->getVertexNormal_Frame(newFrame, allTfList, vNew, nNew);
				discardSmallRegions(vNew);
			}
			else {
				// Just use the last solution's pose as the initialization
				newSoln->tfList = lastSoln->tfList;
			}
		}
	}

	// Detect occlusions, fill in occluded tfs
	detectOcclusions(newFrame, newSoln->occludedLabels);
	fillOccludedTfs(jointMap, *newSoln, defOpt, frameSolution);
	detectOcclusions(newFrame, newSoln->occludedLabels);					/**< Try testing occlusions again */

	//// Re-match occluded parts... Do this after at least refining transformations once!
	//matchOccludedParts(newSoln);


	//if (lastFrame == referenceFrameIndex) {
	//	// If this is the first frame after the reference, exclude occluded bones
	//	for (int i = 0, i_end = newSoln->occludedLabels.size(); i < i_end; ++i) {
	//		excludedLabels.push_back(newSoln->occludedLabels[i]);
	//	}
	//	std::sort(excludedLabels.begin(), excludedLabels.end());
	//	newSoln->occludedLabels.clear();

	//	// Replace labels
	//	std::vector<bool> isOccluded = newSoln->isBoneOccludedOrExcluded(excludedLabels);
	//	PointList vNew, nNew;
	//	allSampleGraph->getVertexNormal_Frame(newFrame, allTfList, vNew, nNew);
	//	replaceLabels(isOccluded, vNew);
	//}

	// Clean up
	delete matchCand;
	return true;


	//////////////////////////////////////////////////////////////////////////////
	////// TESTING CODE (July 8 2009: tested initial label/tf propagation code on reference frame)
	//////////////////////////////////////////////////////////////////////////////
	////// Try random labelling on reference
	////DeformGraphAll_Utility::setRandomLabels(*allSampleGraph, allSampleGraph->vertex, defOpt.numBones);
	////printAllTFList(allTfList);
	////pauseUntilEnter("Test");

	////// Propagate initial labels/transformations
	////if (lastFrame == referenceFrameIndex) {
	////	//DeformGraphAll_Utility::propogateInitialLabels_Ref(lastSoln, *matchCand, *allSampleGraph, newSoln->tfList);

	////	DeformGraphAll_Utility::ClosestPointMap closestPts;
	////	std::vector<bool> isOccluded = lastSoln->isBoneOccludedOrExcluded(excludedLabels);
	////	DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, lastFrame, lastSoln->originalShape, isOccluded, closestPts, defOpt);
	////	DeformGraphAll_Utility::propogateInitialLabels(*matchCand, *allSampleGraph, closestPts, newSoln->tfList);
	////}
	////printAllTFList(allTfList);
	////pauseUntilEnter("Test");

	////// Discard small regions on the new frame
	////PointList vNew, nNew;
	////allSampleGraph->getVertexNormal_Frame(newFrame, allTfList, vNew, nNew);
	////discardSmallRegions(vNew);
	////pauseUntilEnter("Test");

	////// Detect occlusions
	////detectOcclusions(newFrame, newSoln->occludedLabels);

	////// (Testing) Display the result
	////defOpt.ptr_scene->matchCandidates = matchCand;
	////pauseUntilEnter("Test");
}

void Deform::DeformGraphAll::replaceLabels( const std::vector<bool>& toReplace, const PointList& samplePoints, std::vector<int>* labelArray /*= NULL */ )
{
	int numSamples = allSampleGraph->numSamples();
	if (labelArray) numSamples = labelArray->size();

	// Replace excluded labels with the closest non-excluded label
	// a. Build the search structure
	PointList nonExcludedCells;
	std::vector<int> nonExcludedCellLabels;
	for (int i = 0; i < numSamples; ++i) {
		int label = -1;
		if (labelArray) label = (*labelArray)[i];
		else label = findLabel(allSampleGraph->weightList[i]);

		if (!toReplace[label]) {
			nonExcludedCells.add(samplePoints[i]);
			nonExcludedCellLabels.push_back(label);
		}
	}

	// b. Remove excluded labels
	for (int i = 0; i < numSamples; ++i) {
		int label = -1;
		if (labelArray) label = (*labelArray)[i];
		else label = findLabel(allSampleGraph->weightList[i]);

		if (toReplace[label]) {
			idxList nearest;
			distList dists;
			const real3_type& queryPos = samplePoints[i];

			nonExcludedCells.queryKNN(queryPos, 1, nearest, dists, true);
			const int newLabel = nonExcludedCellLabels[nearest[0]];

			// Assign the label
			if (labelArray) {
				(*labelArray)[i] = newLabel;
			}
			else {
				DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(newLabel, 1.0));
				allSampleGraph->weightList[i].swap(sw);
			}
		}
	}
}

bool Deform::DeformGraphAll::discardSmallRegions( const PointList& samplePoints, std::vector<int>* labelArray /*= NULL */ )
{
	const int numLabels = defOpt.numBones;
	int numSamples = allSampleGraph->numSamples();
	if (labelArray) numSamples = labelArray->size();
	idxList samplesPerLabel(numLabels,0);

	// Count the number of samples per label
	for (int i = 0; i < numSamples; ++i) {
		int label = -1;
		if (labelArray) label = (*labelArray)[i];
		else label = findLabel(allSampleGraph->weightList[i]);

		++samplesPerLabel[label];
	}

	// Remove regions that cover less than some threshold percent of the entire graph
	const int excludeThreshold = std::max((double)defOpt.excludeLabelThreshold_num, defOpt.excludeLabelThreshold_ratio * ((double)numSamples));
	std::vector<bool> toExclude(numLabels, false);
	std::vector<int> newExcludedLabels;

	int replaceCount = 0;					/**< Count how many samples will we replace the label */
	for (int i = 0; i < numLabels; ++i) {
		if (samplesPerLabel[i] < excludeThreshold) {
			newExcludedLabels.push_back(i);
			toExclude[i] = true;
			replaceCount += samplesPerLabel[i];

			DeformGraphAll_Utility::SolutionTable::iterator it;
			for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
				// Reset the transformations associated with the excluded labels
				it->second->tfList[i] = DeformAll_Utility::SE3_6DOF();

				// If the label was previously marked as occluded, then remove it
				std::vector<int> newOccludedLabels;
				std::vector<int>& oldOccludedLabels = it->second->occludedLabels;
				for (int j = 0; j < oldOccludedLabels.size(); ++j) {
					if (oldOccludedLabels[j] != i) {
						newOccludedLabels.push_back(oldOccludedLabels[j]);
					}
				}
				oldOccludedLabels.swap(newOccludedLabels);
			}
		}
	}

	// Assign to the global excludedLabels variable
	excludedLabels.swap(newExcludedLabels);

	if (replaceCount > 0) {
		// Replace labels that are to be excluded with the closest non-excluded label
		replaceLabels(toExclude, samplePoints, labelArray);
	}

	// Determine if we had newly discarded some labels
	bool excludedLabels_sameAsBefore = (excludedLabels.size() == newExcludedLabels.size());
	if (excludedLabels_sameAsBefore) {
		// Determine if the contents are the same
		for (int i = 0; i < excludedLabels.size(); ++i) {
			if (excludedLabels[i] != newExcludedLabels[i]) {
				excludedLabels_sameAsBefore = false;
			}
		}
	}
	bool labelDiscarded = (replaceCount > 0) || (!excludedLabels_sameAsBefore);


	//////////////////////////////////////////////////////////////////////////
	cout << DeformGraphAll_Utility::printHeader << "Discarding regions with less than " << excludeThreshold << " samples..." << endl;
	cout << DeformGraphAll_Utility::printHeader << "-- #samples/exclusion per label: [";
	for (int i = 0; i < numLabels; i++) {
		if (i > 0) cout << ",";
		cout << samplesPerLabel[i];
		if (toExclude[i]) cout << "*";
	}
	cout << "]" << endl;
	cout << DeformGraphAll_Utility::printHeader << "-- Label replaced for " << replaceCount << " samples" << endl;
	if (labelDiscarded) cout << DeformGraphAll_Utility::printHeader << "Some labels were newly discarded!" << endl;

	return labelDiscarded;
}

void Deform::DeformGraphAll::detectOcclusions( int frameIndex, std::vector<int>& occludedBones )
{
	// Find closest point correspondences in the frame
	DeformGraphAll_Utility::FrameSolution *frameSoln = frameSolution.find(frameIndex)->second;
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	DeformGraphAll_Utility::ClosestPointMap closestPts;
	//std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
	// Don't consider any bone occluded in this case
	std::vector<bool> isOccluded(defOpt.numBones, false);
	DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, frameIndex, frameSoln->originalShape, frameSoln->frameSamples, isOccluded, closestPts, defOpt);

	// Count total and overlap
	idxList total(defOpt.numBones,0), overlap(defOpt.numBones,0);
	for (int i = 0, i_end = allSampleGraph->numSamples(); i < i_end; ++i) {
		// Find the region that the sample belongs to
		int label = findLabel(allSampleGraph->weightList[i]);
		++total[label];

		DeformGraphAll_Utility::ClosestPointMap::const_iterator it2 = closestPts.find(i);
		if (it2 != closestPts.end()) {
			// Has a correspondence
			++overlap[label];
		}
	}

	// Check occlusion threshold
	cout << DeformGraphAll_Utility::printHeader << "Detecting occlusion status for frame " << frameIndex
		<< "(threshold " << defOpt.occludedLabel_ValidCorrespRatioThreshold << "):" << endl;

	occludedBones.clear();
	for (int i = 0, i_end = defOpt.numBones; i < i_end; ++i) {
		if (total[i] > 0) {
			real_type ratio = ((real_type)overlap[i]) / ((real_type)total[i]);

			cout << DeformGraphAll_Utility::printHeader << "-- Bone " << i << " ratio is: " << ratio << "(" << overlap[i] << "/" << total[i] << ")";
			if (ratio < defOpt.occludedLabel_ValidCorrespRatioThreshold || 
				overlap[i] < defOpt.occludedLabel_ValidCorrespNumThreshold)
			{
				cout << " --> occluded!" << endl;
				occludedBones.push_back(i);
			}
			else {
				cout << endl;
			}
		}
	}
}

void Deform::DeformGraphAll::fillOccludedTfs_AllFrames( bool forceAll /*= false */ )
{
	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int frameIndex = it->first;
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;

		if (frameIndex != referenceFrameIndex) {
			if (forceAll || frameSoln->solveThisFrame) {
				cout << "On frame " << frameIndex << endl;
				fillOccludedTfs(jointMap, *frameSoln, defOpt, frameSolution);
			}
		}
	}
}

void Deform::DeformGraphAll::selectSolver_SlidingWindow( int windowSize )
{
	idxList frameList;
	getSortedFrameList(frameSolution, frameList);

	if (frameList.size()) {
		// Activate the last "windowSize" frames and deactivate the rest
		const int numFrames = frameList.size();
		for (int i = 0; i < numFrames; ++i) {
			int frameNumber = frameList[numFrames - i - 1];

			if (i < windowSize && frameNumber != referenceFrameIndex) {
				// Activate
				getFrameSoln(frameNumber)->solveThisFrame = true;
			}
			else {
				// Deactivate
				getFrameSoln(frameNumber)->solveThisFrame = false;
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Debugging: print the status
	cout << DeformGraphAll_Utility::printHeader << "Solving on frames";
	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		cout << " " << it->first << (it->second->solveThisFrame ? "*" : "");
	}
	cout << endl;
}

void Deform::DeformGraphAll::findTotal_Efit( real_type &totalDist, real_type &totalCount )
{
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	totalDist = 0.0;
	totalCount = 0.0;

	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int frameIndex = it->first;
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;
		frameSoln->sampleError.clear();
		frameSoln->sampleError.resize(allSampleGraph->numSamples() + 1, 0.0);

		if (frameIndex != referenceFrameIndex) {
			DeformGraphAll_Utility::ClosestPointMap closestPts;
			std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
			totalDist += DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, frameIndex, frameSoln->originalShape, frameSoln->frameSamples, isOccluded, closestPts, defOpt, &(frameSoln->sampleError), true);
			totalCount += frameSoln->sampleError[allSampleGraph->numSamples()];
		}
	}
}

void Deform::DeformGraphAll::findClosestPoints_AllFrames( bool forceAll /*= false */ )
{
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int frameIndex = it->first;
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;

		if (frameIndex != referenceFrameIndex) {
			if (forceAll || frameSoln->solveThisFrame) {
				DeformGraphAll_Utility::ClosestPointMap closestPts;
				std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
				DeformGraphAll_Utility::findClosestPoints(*allSampleGraph, allTfList, frameIndex, frameSoln->originalShape, frameSoln->frameSamples, isOccluded, closestPts, defOpt);

				drawMutex.lock();
				sampleClosestPoints[frameIndex].swap(closestPts);			/**< Assign to global list, delete old object if it's there */
				drawMutex.unlock();
			}
		}
	}
}

real_type Deform::DeformGraphAll::findTransformations( int updateCorrespType )
{
	bool done = false;
	int TStepNumItr = 1;
	taucsType Tres = 0.0, lastTres = 0.0, stepRes = 0.0;
	taucsType trustRadius = 1.0;

	const real_type ptWeight = DeformAll_Utility::ERROR_POINT_WEIGHT;
	const real_type plWeight = DeformAll_Utility::ERROR_PLANE_WEIGHT;
	const real_type ptWeightSqrt = sqrt(DeformAll_Utility::ERROR_POINT_WEIGHT);
	const real_type plWeightSqrt = sqrt(DeformAll_Utility::ERROR_PLANE_WEIGHT);

	//// Print out occlusion information per bone
	//cout << "--------------------------------------------------------------------" << endl;
	//cout << "Currently Excluded Labels: [";
	//for (int i = 0; i < excludedLabels.size(); ++i)
	//{
	//	if (i != 0) cout << ",";
	//	cout << excludedLabels[i];
	//}
	//cout << "]" << endl;
	//DeformGraphAll_Utility::SolutionTable::const_iterator it_frame;
	//for (it_frame = frameSolution.begin(); it_frame != frameSolution.end(); ++it_frame) {
	//	cout << "Currently Occluded Labels for frame " << it_frame->first << ": [";
	//	const std::vector<int>& frOcc = it_frame->second->occludedLabels;
	//	for (int i = 0; i < frOcc.size(); ++i)
	//	{
	//		if (i != 0) cout << ",";
	//		cout << frOcc[i];
	//	}
	//	cout << "]" << endl;
	//}
	//cout << "--------------------------------------------------------------------" << endl;


	DeformAll_Utility::TfVarTranslate varTranslate(frameSolution, referenceFrameIndex, excludedLabels);
	//varTranslate.printMapping();

	// Find joints and bones
	estimateJoints();

	// Reset timers
	//MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_1,Reset);
	//MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_2,Reset);
	//MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_3,Reset);
	
	//cout << DeformGraphAll_Utility::printHeader << "T-Step:";

	DeformGraphAll_Utility::SolutionTable::const_iterator it;

	// T-step loop
	while (!done) {
		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_1,AddStart);
		// (1) Update the closest point in all frames
		if (updateCorrespType == CORRESPONDENCE_CLOSEST_POINTS) {
			findClosestPoints_AllFrames();
		}

		idxList numDist, numJoint, toRemove;
		countConstraints(varTranslate, numDist, numJoint);
		//cout << "Distance constraints per bone: [";
		//for (int i = 0; i < numDist.size(); ++i)
		//{
		//	if (i != 0) cout << ",";
		//	cout << numDist[i];
		//}
		//cout << "]" << endl;
		//cout << "Joint constraints per bone: [";
		//for (int i = 0; i < numJoint.size(); ++i)
		//{
		//	if (i != 0) cout << ",";
		//	cout << numJoint[i];
		//}
		//cout << "]" << endl;
		// Check if there are bones with no distance constraints
		for (int i = 0; i < numDist.size(); ++i) {
			int numConstraints = numDist[i] + numJoint[i];
			if (numConstraints < defOpt.occludedLabel_ValidCorrespNumThreshold) {
				int f, b;
				varTranslate.frameBoneIndex(i, f, b);
				cout << DeformGraphAll_Utility::printHeader << "WARNING!!! Frame " << f << ", bone " << b << " has " << numConstraints << " distance/joint constraints!" << endl;
				toRemove.push_back(i);
			}
		}
		if (toRemove.size()) {
			// Remove bones with no distance constraints
			cout << DeformGraphAll_Utility::printHeader << "Removing bones with less than " << defOpt.occludedLabel_ValidCorrespNumThreshold << " constraints." << endl;
			cout << DeformGraphAll_Utility::printHeader << "Mapping before: ";
			varTranslate.printMapping();

			varTranslate.remove_linearIndices(toRemove);

			cout << DeformGraphAll_Utility::printHeader << "Mapping after: ";
			varTranslate.printMapping();
		}
		if (varTranslate.numTotalBones() == 0)
		{
			// Abort if there are no bones remaining
			done = true;
			break;
		}


		////// Compute translation/scaling
		////real3_type avgPt;
		////real_type avgDist;
		////computeTfResidual(temp1, avgPt, avgDist, varTranslate, ptWeight, plWeight);
		////real_type s = sqrt(2.f) / avgDist;
		////real3_type t = avgPt;
		real_type s = 1.0;
		real3_type t(0.0,0.0,0.0);


		////// Compute initial error residual
		////taucsType totalRes = 0.0, fitRes = 0.0, jointRes = 0.0, dummyF = 0.0, dummyJ = 0.0;
		////computeTfResidual(fitRes, dummyF, 0.0, testSoln, varTranslate, s, t, ptWeight, plWeight);
		////for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		////	computeJointResidual(temp1, temp2, 0.0, testSoln, varTranslate, it->first, s, t, it->second->tfList);
		////	jointRes += temp1, dummyJ += temp2;
		////}
		////totalRes = defOpt.TWeights[0]*fitRes + defOpt.TWeights[1]*jointRes;
		////cout << TStepNumItr << "> Initial residual,  " << totalRes << "," << fitRes << "," << jointRes << endl;

		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_1,Stop);


		//////////////////////////////////////////////////////////////////////////
		// Solving for backward transformations in distance term; with ball joint constraint
		//////////////////////////////////////////////////////////////////////////
		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_2,AddStart);
		// (2) Construct system
		TermList terms;
		if (defOpt.TWeights[0] > 0.0) {
			ErrorTerm *distanceTerm = makeDistanceTerm(varTranslate, s, t, ptWeightSqrt, plWeightSqrt);
			distanceTerm->weight = sqrt(defOpt.TWeights[0]);
			terms.push_back(distanceTerm);
		}
		if (defOpt.TWeights[1] > 0.0) {
			// Use ball joint constraints
			DeformGraphAll_Utility::SolutionTable::const_iterator it;
			for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
				ErrorTerm *smoothnessTerm = makeSmoothnessTerm_BallJoint(varTranslate, it->first, s, t, it->second->tfList);

				if (smoothnessTerm)
				{
					smoothnessTerm->weight = sqrt(defOpt.TWeights[1]);
					terms.push_back(smoothnessTerm);
				}
			}
		}
		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_2,Stop);


		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_3,AddStart);
		// (3) Solve system and apply solution
		//taucsType *soln = DeformAll_Utility::solveDogLeg(terms, numVars, trustRadius);
		taucsType *soln = DeformAll_Utility::solveLinear(terms, 6 * varTranslate.numTotalBones(), Tres, stepRes, defOpt.saveTaucsErrorLog, 0.0);


		////taucsType totalResU = 0.0, totalResL = 0.0, fitResU = 0.0, jointResU = 0.0, fitResL = 0.0, jointResL = 0.0;
		////computeTfResidual(fitResU, fitResL, 1.0, soln, varTranslate, s, t, ptWeight, plWeight);
		////for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		////	computeJointResidual(temp1, temp2, 1.0, soln, varTranslate, it->first, s, t, it->second->tfList);
		////	jointResU += temp1, jointResL += temp2;
		////}
		////totalResU = defOpt.TWeights[0]*fitResU + defOpt.TWeights[1]*jointResU;
		////totalResL = defOpt.TWeights[0]*fitResL + defOpt.TWeights[1]*jointResL;
		////cout << TStepNumItr << "> Updated residualU, " << totalResU << "," << fitResU << "," << jointResU << endl;
		////cout << TStepNumItr << "> Updated residualL, " << totalResL << "," << fitResL << "," << jointResL << endl;
		////cout << TStepNumItr << "> Return by Solver,  " << Tres << "," << stepRes << endl;

		////taucsType gainRatio = (totalRes - totalResU) / (totalRes - totalResL);
		////cout << TStepNumItr << "> Gain ratio: " << gainRatio << endl;


		applyTfCorrection(1.0, soln, s, t, varTranslate);
		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_3,Stop);

		//cout << "*** (T-Step) " << TStepNumItr << " / " << defOpt.maxTStepIterations;
		//cout << " (Residual=" << Tres << ", Diff=" << abs(Tres - lastTres) << ")" << endl;
		cout << " " << TStepNumItr;

		//pauseUntilEnter("Test T-Step");

		// (4) Determine stop conditions and update
		if ((TStepNumItr > 1) && (abs(Tres - lastTres) < defOpt.stopEpsilon_sqRt*(1+Tres))) {
			//cout << "*** (T-Step) done! (objective)" << endl;
			done = true;
		}
		if ((defOpt.maxTStepIterations > 0) && (TStepNumItr >= defOpt.maxTStepIterations)) {
			//cout << "*** (T-Step) done! (max iteration)" << endl;
			done = true;
		}
		++TStepNumItr;
		lastTres = Tres;

		// (5) Cleanup: Delete error terms & solution
		for (int i = 0; i < terms.size(); ++i) {
			if (terms[i]) {
				delete terms[i];
				terms[i] = NULL;
			}
		}
		terms.clear();
		delete[] soln;
	}

	//cout << "*** (T-step) Total number of iterations: " << TStepNumItr << endl;
	//cout << endl;

	//printTime(TIMER_SIMULTANEOUS_T_STEP_1);
	//printTime(TIMER_SIMULTANEOUS_T_STEP_2);
	//printTime(TIMER_SIMULTANEOUS_T_STEP_3);

	return lastTres;
}

void Deform::DeformGraphAll::countConstraints( const DeformAll_Utility::TfVarTranslate& varTranslate, idxList& distConstraints, idxList& jointConstraints ) const
{
	const int numBones = defOpt.numBones;
	idxList numDistConstraints(varTranslate.numTotalBones(), 0);
	idxList numJointConstraints(varTranslate.numTotalBones(), 0);

	DeformGraphAll_Utility::AllClosestPoints::const_iterator it;
	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);

		// Check that we are solving for the transformations of this frame
		if (targetSoln->solveThisFrame) {
			const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it2;
			for (it2 = closestPts.begin(); it2 != closestPts.end(); ++it2) {
				// Add this correspondence to the optimization
				const int srcASGIdx = it2->first;
				const DeformAll_Utility::SparseWeight& sw = allSampleGraph->weightList[srcASGIdx];
				const int sourceFrameIdx = allSampleGraph->fIndex[srcASGIdx];

				if (sourceFrameIdx == targetFrameIdx)
				{
					continue;
				}

				// (1) LHS: For each bone affecting this correspondence
				for (int bi = 0, bi_end = sw.size(); bi < bi_end; ++bi) {
					int boneIndex = sw[bi].index;

					// Populate entries for source frame
					if (varTranslate.validBone(sourceFrameIdx, boneIndex)) {
						++numDistConstraints[varTranslate.linearIndex(sourceFrameIdx, boneIndex)];
					}

					// Populate entries for target frame
					if (varTranslate.validBone(targetFrameIdx, boneIndex)) {
						++numDistConstraints[varTranslate.linearIndex(targetFrameIdx, boneIndex)];
					}
				}
			}
		}
	}


	DeformGraphAll_Utility::JointMap::const_iterator it_joint;
	DeformGraphAll_Utility::SolutionTable::const_iterator it_frame;
	for (it_frame = frameSolution.begin(); it_frame != frameSolution.end(); ++it_frame) {
		const int targetFrameIndex = it_frame->first;
		for (it_joint = jointMap.begin(); it_joint != jointMap.end(); ++it_joint) {
			int bi, bj, linIdx = it_joint->first;
			tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

			if ((solveStatus == SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) || varTranslate.validBone(targetFrameIndex, bj))) ||
				(solveStatus != SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) && varTranslate.validBone(targetFrameIndex, bj))))
			{
				int numSamples = 0;
				const DeformGraphAll_Utility::Joint& joint = it_joint->second;

				if (joint.type == DeformGraphAll_Utility::JOINT_BALL)
				{
					numSamples = 1;
				}
				else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE)
				{
					numSamples = defOpt.hingeJointSamples;
				}

				if (varTranslate.validBone(targetFrameIndex, bi)) {
					numJointConstraints[varTranslate.linearIndex(targetFrameIndex, bi)] += numSamples;
				}
				if (varTranslate.validBone(targetFrameIndex, bj)) {
					numJointConstraints[varTranslate.linearIndex(targetFrameIndex, bj)] += numSamples;
				}
			}
		}
	}

	distConstraints.swap(numDistConstraints);
	jointConstraints.swap(numJointConstraints);
}

ErrorTerm* Deform::DeformGraphAll::makeDistanceTerm( const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s,  const real3_type& t, const real_type ptWeight, const real_type plWeight ) const
{
	const int termsPerBonePerCorresp = 6;

	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;

	DeformGraphAll_Utility::AllClosestPoints::const_iterator it;
	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);

		// Check that we are solving for the transformations of this frame
		if (targetSoln->solveThisFrame) {
			const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it2;
			for (it2 = closestPts.begin(); it2 != closestPts.end(); ++it2) {
				// Add this correspondence to the optimization
				const int srcASGIdx = it2->first;
				const int sourceFrameIdx = allSampleGraph->fIndex[srcASGIdx];
				if (sourceFrameIdx == targetFrameIdx)
				{
					continue;
				}

				const DeformAll_Utility::SparseWeight& sw = allSampleGraph->weightList[srcASGIdx];
				bool populated = false;

				// (1) LHS: For each bone affecting this correspondence
				for (int bi = 0, bi_end = sw.size(); bi < bi_end; ++bi) {
					int boneIndex = sw[bi].index;

					// Populate entries for source frame
					if (varTranslate.validBone(sourceFrameIdx, boneIndex)) {
						totalLHSnnz += 9;								/**< Point-to-point metric */
						totalLHSnnz += termsPerBonePerCorresp;			/**< Point-to-plane metric */
						populated = true;
					}

					// Populate entries for target frame
					if (varTranslate.validBone(targetFrameIdx, boneIndex)) {
						totalLHSnnz += 9;								/**< Point-to-point metric */
						totalLHSnnz += termsPerBonePerCorresp;			/**< Point-to-plane metric */
						populated = true;
					}
				}

				// Check to see if any entries were populated at all
				if (populated) {
					totalRows += 3;			/**< Point-to-point metric */
					++totalRows;			/**< Point-to-plane metric */
				}
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Create the term
	ErrorTerm *distanceTerm = new ErrorTerm(1.0, totalRows, totalLHSnnz);

	// Keep track of the row as we populate the matrices
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	// Prepare
	data_type Rpt0(12), Rpt1(12), Rpt2(12), Rpl(12), Rsrc(6), Rtgt(6);
	Rpt0.setZero();
	Rpt1.setZero();
	Rpt2.setZero();
	Rpl.setZero();
	Rsrc.setZero();
	Rtgt.setZero();

	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);

		// Check that we are solving for the transformations of this frame
		if (targetSoln->solveThisFrame) {
			const ShapeModel& targetShape = targetSoln->originalShape;
			const DeformAll_Utility::TransformList& targetTf = targetSoln->tfList;

			const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it2;
			for (it2 = closestPts.begin(); it2 != closestPts.end(); ++it2) {

				const int srcASGIdx = it2->first;
				const int sourceFrameIdx = allSampleGraph->fIndex[srcASGIdx];
				if (sourceFrameIdx == targetFrameIdx)
				{
					continue;
				}

				// Add this correspondence to the optimization
				const int tgtShapeIdx = it2->second;
				const DeformAll_Utility::SparseWeight& sw = allSampleGraph->weightList[srcASGIdx];
				const DeformGraphAll_Utility::FrameSolution *sourceSoln = getFrameSoln(sourceFrameIdx);
				const DeformAll_Utility::TransformList& sourceTf = sourceSoln->tfList;
				bool populated = false;
				real3_type sourceX, sourceN, targetX, targetN;
				sourceX = allSampleGraph->vertex[srcASGIdx];
				sourceN = allSampleGraph->normal[srcASGIdx];
				targetX = targetShape.vertex[tgtShapeIdx];
				targetN = targetShape.normal[tgtShapeIdx];

				// (1) LHS: For each bone affecting this correspondence
				for (int bi = 0, bi_end = sw.size(); bi < bi_end; ++bi) {
					int boneIndex = sw[bi].index;
					real_type boneWeight = sw[bi].weight;
					real3_type bone_defSourceX, bone_defSourceN, bone_defTargetX, bone_defTargetN;
					DeformAll_Utility::SE3_6DOF id;

					// Deform positions using the current bone transformation
					bone_defSourceX = sourceTf[boneIndex] * sourceX;
					bone_defSourceN = sourceTf[boneIndex].rotMatrix() * sourceN;
					bone_defTargetX = targetTf[boneIndex] * targetX;
					bone_defTargetN = targetTf[boneIndex].rotMatrix() * targetN;

					//// Apply translation and scaling to the positions
					//bone_defSourceX = s*(bone_defSourceX - t);
					//bone_defTargetX = s*(bone_defTargetX - t);

					// Compute point-to-point metric
					DeformAll_Utility::helperJacPtSA(boneWeight, boneWeight, bone_defSourceX, bone_defTargetX, Rpt0, Rpt1, Rpt2);

					// Compute point-to-plane metric
					// Type (1) SA rotation w/ no normal rotation
					DeformAll_Utility::helperJacSA(boneWeight, bone_defSourceX, bone_defTargetN, Rsrc);
					DeformAll_Utility::helperJacSA(boneWeight, bone_defTargetX, bone_defTargetN, Rtgt);
					for (int col = 0; col < 6; ++col) {
						Rpl.val[col] = Rsrc.val[col];
						Rpl.val[col+6] = -1.0 * Rtgt.val[col];
					}

					bool sourceValid = varTranslate.validBone(sourceFrameIdx, boneIndex);
					bool targetValid = varTranslate.validBone(targetFrameIdx, boneIndex);
					index_type srcVarCol = 6 * varTranslate.linearIndex(sourceFrameIdx, boneIndex);
					index_type tgtVarCol = 6 * varTranslate.linearIndex(targetFrameIdx, boneIndex);
					populated = populated || sourceValid || targetValid;

					//if (sourceValid && targetValid && srcVarCol >= tgtVarCol)
					//{
					//	printf("********************************************************\n");
					//	printf("** Error in sorting.. frames (%d, %d) ********************\n", sourceFrameIdx, targetFrameIdx);
					//	printf("********************************************************\n");
					//}

					// Populate entries for point-to-point metric
					for (int col = 0; sourceValid && col < 6; ++col) {
						if (Rpt0.val[col] != 0.0)
							distanceTerm->lhs.fill(TrackRow+0, srcVarCol+col, Rpt0.val[col]);
					}

					for (int col = 0; targetValid && col < 6; ++col) {
						if (Rpt0.val[col+6] != 0.0)
							distanceTerm->lhs.fill(TrackRow+0, tgtVarCol+col, Rpt0.val[col+6]);
					}

					for (int col = 0; sourceValid && col < 6; ++col) {
						if (Rpt1.val[col] != 0.0)
							distanceTerm->lhs.fill(TrackRow+1, srcVarCol+col, Rpt1.val[col]);
					}

					for (int col = 0; targetValid && col < 6; ++col) {
						if (Rpt1.val[col+6] != 0.0)
							distanceTerm->lhs.fill(TrackRow+1, tgtVarCol+col, Rpt1.val[col+6]);
					}

					for (int col = 0; sourceValid && col < 6; ++col) {
						if (Rpt2.val[col] != 0.0)
							distanceTerm->lhs.fill(TrackRow+2, srcVarCol+col, Rpt2.val[col]);
					}

					for (int col = 0; targetValid && col < 6; ++col) {
						if (Rpt2.val[col+6] != 0.0)
							distanceTerm->lhs.fill(TrackRow+2, tgtVarCol+col, Rpt2.val[col+6]);
					}

					// Populate entries for point-to-plane metric
					for (int col = 0; sourceValid && col < termsPerBonePerCorresp; ++col) {
						if (Rpl.val[col] != 0.0) {
							distanceTerm->lhs.fill(TrackRow+3, srcVarCol+col, Rpl.val[col]);
						}
					}

					for (int col = 0; targetValid && col < termsPerBonePerCorresp; ++col) {
						if (Rpl.val[col+6] != 0.0) {
							distanceTerm->lhs.fill(TrackRow+3, tgtVarCol+col, Rpl.val[col+6]);
						}
					}
				}

				// Check to see if any entries were populated at all
				if (populated) {
					real3_type defSourceX, defTargetX, defTargetN;
					defSourceX = vRef[srcASGIdx];
					lbs(targetX, targetN, defTargetX, defTargetN, sw, targetTf, false);

					//// Apply translation and scaling to the positions
					//defSourceX = s*(defSourceX - t);
					//defTargetX = s*(defTargetX - t);

					// (2) RHS: Compute residual for point-to-point metric
					real3_type residualPt = defSourceX - defTargetX;
					for (int k = 0; k < 3; ++k) {
						distanceTerm->rhs[TrackRow+k] = -residualPt[k];
						distanceTerm->rowWeight[TrackRow+k] = ptWeight;
					}

					// Increment row
					TrackRow += 3;

					// (2) RHS: Compute residual for point-to-plane metric
					real_type residualPl = (defSourceX - defTargetX)*defTargetN;
					distanceTerm->rhs[TrackRow] = -residualPl;
					distanceTerm->rowWeight[TrackRow] = plWeight;

					// Increment row
					++TrackRow;
				}
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < distanceTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << distanceTerm->rhs[ci] << ";" << endl;
	//}
	//distanceTerm->lhs.print();
	if ((totalLHSnnz != distanceTerm->lhs.size()) || (TrackRow != distanceTerm->rows)) {
		cout << "*** Distance Term:" << endl;
		cout << "Estimated size for Jacobian: " << totalLHSnnz
			<< " vs. Actual size: " << distanceTerm->lhs.size() << endl;
		cout << "*** Estimated # rows: " << distanceTerm->rows
			<< " vs. Actual # rows: " << TrackRow << endl;
	}

	return distanceTerm;
}

ErrorTerm* Deform::DeformGraphAll::makeSmoothnessTerm_BallJoint( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s,  const real3_type& t, const DeformAll_Utility::TransformList& tfList ) const
{
	// Initialize variables
	const int numBones = defOpt.numBones;

	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		if ((solveStatus == SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) || varTranslate.validBone(targetFrameIndex, bj))) ||
			(solveStatus != SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) && varTranslate.validBone(targetFrameIndex, bj))))
		{
			const DeformGraphAll_Utility::Joint& joint = it->second;
			int numSamples = 1;
			if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				numSamples = defOpt.hingeJointSamples;
			}
			//int numSamples = defOpt.hingeJointSamples;

			if (varTranslate.validBone(targetFrameIndex, bi)) {
				totalLHSnnz += 18 * numSamples;
			}
			if (varTranslate.validBone(targetFrameIndex, bj)) {
				totalLHSnnz += 18 * numSamples;
			}
			totalRows += 3 * numSamples;
		}
	}

	if (totalLHSnnz == 0 || totalRows == 0)
	{
		return NULL;
	}

	//////////////////////////////////////////////////////////////////////////
	// Create the term
	ErrorTerm *smoothTerm = new ErrorTerm(1.0, totalRows, totalLHSnnz);

	// Keep track of the row as we populate the matrices
	index_type TrackRow = 0;

	data_type RJoint0(12), RJoint1(12), RJoint2(12);
	RJoint0.setZero();
	RJoint1.setZero();
	RJoint2.setZero();

	//////////////////////////////////////////////////////////////////////////
	// Iterate through each unique pair
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);
		index_type varCol_bi = 6 * varTranslate.linearIndex(targetFrameIndex, bi);		/**< Column where the variable belongs */
		index_type varCol_bj = 6 * varTranslate.linearIndex(targetFrameIndex, bj);		/**< Column where the variable belongs */

		if (varCol_bi > varCol_bj)
		{
			// Swap bi, bj
			int temp = bi;
			bi = bj;
			bj = temp;

			index_type tempcol = varCol_bi;
			varCol_bi = varCol_bj;
			varCol_bj = tempcol;
		}

		// In the case of solving for the transformations of occluded bones,
		// ...activate the joint constraint if either bone is activated.
		// Otherwise, in normal operation, active the joint if both bones are activated.
		if ((solveStatus == SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) || varTranslate.validBone(targetFrameIndex, bj))) ||
			(solveStatus != SOLVER_OCCLUDED_ICP && (varTranslate.validBone(targetFrameIndex, bi) && varTranslate.validBone(targetFrameIndex, bj))))
		{
			const DeformGraphAll_Utility::Joint& joint = it->second;
			const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
			const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

			int numSamples = 0;
			if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
				numSamples = 1;
			}
			else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				numSamples = defOpt.hingeJointSamples;
			}
			//int numSamples = defOpt.hingeJointSamples;

			DeformAll_Utility::SE3_6DOF tInv_bi = -tfList[bi];
			DeformAll_Utility::SE3_6DOF tInv_bj = -tfList[bj];
			bool biValid = varTranslate.validBone(targetFrameIndex, bi);
			bool bjValid = varTranslate.validBone(targetFrameIndex, bj);

			for (int j = 0; j < numSamples; ++j) {
				real3_type jointLocation;

				if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
					jointLocation = jointPos;
				}
				else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
					real_type t = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
					jointLocation = jointPos +  t*jointCon;
				}

				//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				// Populate optimization matrices
				//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

				// Transform the joint position with each bone transformation
				// Idea 1: constrain inverse

				// Populate LHS matrix entries using inverse transformation
				DeformAll_Utility::helperJointJacInvPtSA(jointLocation, tInv_bi.rotMatrix(), tInv_bj.rotMatrix(), RJoint0, RJoint1, RJoint2);

				for (int col = 0; biValid && col < 6; ++col) {
					if (RJoint0.val[col] != 0.0)
						smoothTerm->lhs.fill(TrackRow+0, varCol_bi+col, RJoint0.val[col]);
				}

				for (int col = 0; bjValid && col < 6; ++col) {
					if (RJoint0.val[col+6] != 0.0)
						smoothTerm->lhs.fill(TrackRow+0, varCol_bj+col, RJoint0.val[col+6]);
				}
				
				for (int col = 0; biValid && col < 6; ++col) {
					if (RJoint1.val[col] != 0.0)
						smoothTerm->lhs.fill(TrackRow+1, varCol_bi+col, RJoint1.val[col]);
				}

				for (int col = 0; bjValid && col < 6; ++col) {
					if (RJoint1.val[col+6] != 0.0)
						smoothTerm->lhs.fill(TrackRow+1, varCol_bj+col, RJoint1.val[col+6]);
				}

				for (int col = 0; biValid && col < 6; ++col) {
					if (RJoint2.val[col] != 0.0)
						smoothTerm->lhs.fill(TrackRow+2, varCol_bi+col, RJoint2.val[col]);
				}

				for (int col = 0; bjValid && col < 6; ++col) {
					if (RJoint2.val[col+6] != 0.0)
						smoothTerm->lhs.fill(TrackRow+2, varCol_bj+col, RJoint2.val[col+6]);
				}

				// Populate RHS residual entries
				real3_type jointDef_bi, jointDef_bj;
				jointDef_bi = tInv_bi * jointLocation;
				jointDef_bj = tInv_bj * jointLocation;
				real3_type residualJoint = jointDef_bi - jointDef_bj;
				for (int k = 0; k < 3; ++k) {
					smoothTerm->rhs[TrackRow+k] = -residualJoint[k];

					// TODO: assign a weight for each joint if necessary
					//smoothTerm->rowWeight[TrackRow+k] = ;
				}

				// Increment row
				TrackRow += 3;
			}
		}
	}

	////////////////////////////////////////////////////////////////////////
	// Debugging: print out results
	//for (int ci = 0; ci < smoothTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << smoothTerm->rhs[ci] << ";" << endl;
	//}
	//distanceTerm->lhs.print();
	if ((totalLHSnnz != smoothTerm->lhs.size()) || (TrackRow != smoothTerm->rows)) {
		cout << "*** Smoothness (Ball Joint) Term:" << endl;
		cout << "Estimated size for Jacobian: " << totalLHSnnz
			<< " vs. Actual size: " << smoothTerm->lhs.size() << endl;
		cout << "*** Estimated # rows: " << smoothTerm->rows
			<< " vs. Actual # rows: " << TrackRow << endl;
	}

	return smoothTerm;
}

void Deform::DeformGraphAll::computeTfResidual( taucsType &totalRes, taucsType &linearRes, taucsType stepSize, taucsType *correction, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s, const real3_type& t, const real_type ptWeight, const real_type plWeight ) const
{
	totalRes = 0.0, linearRes = 0.0;

	DeformGraphAll_Utility::AllClosestPoints::const_iterator it;
	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);

		// Check that we are solving for the transformations of this frame
		if (targetSoln->solveThisFrame) {
			const ShapeModel& targetShape = targetSoln->originalShape;
			const DeformAll_Utility::TransformList& targetTf = targetSoln->tfList;

			const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it2;
			for (it2 = closestPts.begin(); it2 != closestPts.end(); ++it2) {
				// Add this correspondence to the optimization
				const int srcASGIdx = it2->first;
				const int tgtShapeIdx = it2->second;
				const DeformAll_Utility::SparseWeight& sw = allSampleGraph->weightList[srcASGIdx];
				const int sourceFrameIdx = allSampleGraph->fIndex[srcASGIdx];
				const DeformGraphAll_Utility::FrameSolution *sourceSoln = getFrameSoln(sourceFrameIdx);
				const DeformAll_Utility::TransformList& sourceTf = sourceSoln->tfList;
				bool populated = false;
				real3_type sourceX, sourceN, targetX, targetN;
				sourceX = allSampleGraph->vertex[srcASGIdx];
				sourceN = allSampleGraph->normal[srcASGIdx];
				targetX = targetShape.vertex[tgtShapeIdx];
				targetN = targetShape.normal[tgtShapeIdx];

				int boneIndex = findLabel(sw);
				if (varTranslate.validBone(sourceFrameIdx, boneIndex) || varTranslate.validBone(targetFrameIdx, boneIndex)) {
					real3_type defSourceX, defSourceN, defTargetX, defTargetN, residualPt;
					real3_type linearRotSrc, linearTrSrc, linearRotTgt, linearTrTgt;
					real_type residualPl;
					int idxSrc, idxTgt;

					// Linearized Residual
					if (varTranslate.validBone(sourceFrameIdx, boneIndex)) {
						idxSrc = 6*varTranslate.linearIndex(sourceFrameIdx, boneIndex);
						linearRotSrc = real3_type(stepSize*correction[idxSrc],stepSize*correction[idxSrc+1],stepSize*correction[idxSrc+2]);
						linearTrSrc = real3_type(stepSize*correction[idxSrc+3],stepSize*correction[idxSrc+4],stepSize*correction[idxSrc+5]);
					}
					if (varTranslate.validBone(targetFrameIdx, boneIndex)) {
						idxTgt = 6*varTranslate.linearIndex(targetFrameIdx, boneIndex);
						linearRotTgt = real3_type(stepSize*correction[idxTgt],stepSize*correction[idxTgt+1],stepSize*correction[idxTgt+2]);
						linearTrTgt = real3_type(stepSize*correction[idxTgt+3],stepSize*correction[idxTgt+4],stepSize*correction[idxTgt+5]);
					}

					// Apply transformations
					defSourceX = sourceTf[boneIndex] * sourceX;
					defSourceN = sourceTf[boneIndex].rotMatrix() * sourceN;
					defTargetX = targetTf[boneIndex] * targetX;
					defTargetN = targetTf[boneIndex].rotMatrix() * targetN;

					// Apply translation and scaling to the positions
					defSourceX = s*(defSourceX - t);
					defTargetX = s*(defTargetX - t);

					residualPt = (defSourceX - defTargetX) + (cross(linearRotSrc,defSourceX) + linearTrSrc) - (cross(linearRotTgt,defTargetX) + linearTrTgt);
					residualPl = (defSourceX - defTargetX)*defTargetN + (cross(defSourceX,defTargetN)*linearRotSrc + defTargetN*linearTrSrc) - (cross(defTargetX,defTargetN)*linearRotTgt + defTargetN*linearTrTgt);
					linearRes += ptWeight * residualPt.square_norm() + plWeight * residualPl * residualPl;


					// Actual Residual
					DeformAll_Utility::SE3_6DOF correctionSrc, correctionTgt;
					correctionSrc.createFromLinear(linearRotSrc, linearTrSrc);
					correctionTgt.createFromLinear(linearRotTgt, linearTrTgt);

					//cout << "Original:   " << sourceTf[boneIndex].rot() << " " << sourceTf[boneIndex].tr() << endl;
					//cout << "Correction: " << correctionSrc.rot() << " " << correctionSrc.tr() << endl;

					// Apply transformations
					defSourceX = correctionSrc * defSourceX;
					defSourceN = correctionSrc.rotMatrix() * defSourceN;
					defTargetX = correctionTgt * defTargetX;
					defTargetN = correctionTgt.rotMatrix() * defTargetN;

					residualPt = defSourceX - defTargetX;
					residualPl = (defSourceX - defTargetX)*defTargetN;
					totalRes += ptWeight * residualPt.square_norm() + plWeight * residualPl * residualPl;
				}
			}
		}
	}
}

void Deform::DeformGraphAll::computeTfResidual( taucsType& totalRes, real3_type& avgPt, real_type& avgDist, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type ptWeight, const real_type plWeight ) const
{
	PointList allDef;
	totalRes = 0.0;

	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	DeformGraphAll_Utility::AllClosestPoints::const_iterator it;
	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);

		// Check that we are solving for the transformations of this frame
		if (targetSoln->solveThisFrame) {
			const ShapeModel& targetShape = targetSoln->originalShape;
			const DeformAll_Utility::TransformList& targetTf = targetSoln->tfList;

			const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it2;
			for (it2 = closestPts.begin(); it2 != closestPts.end(); ++it2) {
				// Add this correspondence to the optimization
				const int srcASGIdx = it2->first;
				const int tgtShapeIdx = it2->second;
				const DeformAll_Utility::SparseWeight& sw = allSampleGraph->weightList[srcASGIdx];
				const int sourceFrameIdx = allSampleGraph->fIndex[srcASGIdx];
				const DeformGraphAll_Utility::FrameSolution *sourceSoln = getFrameSoln(sourceFrameIdx);
				const DeformAll_Utility::TransformList& sourceTf = sourceSoln->tfList;
				bool populated = false;
				real3_type sourceX, sourceN, targetX, targetN;
				sourceX = allSampleGraph->vertex[srcASGIdx];
				sourceN = allSampleGraph->normal[srcASGIdx];
				targetX = targetShape.vertex[tgtShapeIdx];
				targetN = targetShape.normal[tgtShapeIdx];

				// (1) LHS: For each bone affecting this correspondence
				for (int bi = 0, bi_end = sw.size(); bi < bi_end; ++bi) {
					int boneIndex = sw[bi].index;

					// Populate entries for source frame
					if (varTranslate.validBone(sourceFrameIdx, boneIndex)) {
						populated = true;
					}

					// Populate entries for target frame
					if (varTranslate.validBone(targetFrameIdx, boneIndex)) {
						populated = true;
					}
				}

				// Check to see if any entries were populated at all
				if (populated) {
					real3_type defSourceX, defTargetX, defTargetN;
					defSourceX = vRef[srcASGIdx];
					lbs(targetX, targetN, defTargetX, defTargetN, sw, targetTf, false);

					// (2) RHS: Compute residual for point-to-point metric
					real3_type residualPt = defSourceX - defTargetX;

					// (2) RHS: Compute residual for point-to-plane metric
					real_type residualPl = (defSourceX - defTargetX)*defTargetN;

					totalRes += ptWeight * residualPt.square_norm() + plWeight * residualPl * residualPl;

					// Add to list
					allDef.add(defSourceX);
					allDef.add(defTargetX);
				}
			}
		}
	}

	// Statistics
	avgDist = avgPt[0] = avgPt[1] = avgPt[2] = 0.0;
	const int numPts = allDef.size();
	for (int i = 0; i < numPts; ++i) {
		avgPt += allDef[i];
	}
	avgPt = avgPt / ((real_type)numPts);
	for (int i = 0; i < numPts; ++i) {
		avgDist += (allDef[i] - avgPt).norm();
	}
	avgDist = avgDist / ((real_type)numPts);
	//cout << "Average Point: " << avgPt << endl;
	//cout << "Average Distance: " << avgDist << endl;
	//cout << "Averaged over " << numPts << " constraints." << endl;
}

void Deform::DeformGraphAll::computeJointResidual( taucsType &totalRes, taucsType &linearRes, taucsType stepSize, taucsType *correction, const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s, const real3_type& t, const DeformAll_Utility::TransformList& tfList ) const
{
	totalRes = 0.0, linearRes = 0.0;
	const int numBones = defOpt.numBones;

	// Iterate through each unique pair
	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		if (!varTranslate.validBone(targetFrameIndex, bi) || !varTranslate.validBone(targetFrameIndex, bj)) {
			continue;
		}

		const DeformGraphAll_Utility::Joint& joint = it->second;
		const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
		const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

		int numSamples = 0;
		if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
			numSamples = 1;
		}
		else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
			numSamples = defOpt.hingeJointSamples;
		}

		for (int j = 0; j < numSamples; ++j) {
			real3_type jointLocation;

			if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
				jointLocation = jointPos;
			}
			else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				real_type t = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
				jointLocation = jointPos +  t*jointCon;
			}

			// Transform the joint position with each bone transformation
			real3_type jointDef_bi, jointDef_bj, residualPt;
			std::vector<data_type> RJoint;
			real3_type linearRotSrc, linearTrSrc, linearRotTgt, linearTrTgt;
			int idxSrc, idxTgt;

			// Linearized Residual
			if (varTranslate.validBone(targetFrameIndex, bi)) {
				idxSrc = 6*varTranslate.linearIndex(targetFrameIndex, bi);
				linearRotSrc = real3_type(stepSize*correction[idxSrc],stepSize*correction[idxSrc+1],stepSize*correction[idxSrc+2]);
				linearTrSrc = real3_type(stepSize*correction[idxSrc+3],stepSize*correction[idxSrc+4],stepSize*correction[idxSrc+5]);
			}
			if (varTranslate.validBone(targetFrameIndex, bj)) {
				idxTgt = 6*varTranslate.linearIndex(targetFrameIndex, bj);
				linearRotTgt = real3_type(stepSize*correction[idxTgt],stepSize*correction[idxTgt+1],stepSize*correction[idxTgt+2]);
				linearTrTgt = real3_type(stepSize*correction[idxTgt+3],stepSize*correction[idxTgt+4],stepSize*correction[idxTgt+5]);
			}

			// Apply transformations
			jointDef_bi = (-tfList[bi]) * jointLocation;
			jointDef_bj = (-tfList[bj]) * jointLocation;
			// Apply translation and scaling to the positions
			jointDef_bi = s*(jointDef_bi - t);
			jointDef_bj = s*(jointDef_bj - t);
			residualPt = (jointDef_bi - jointDef_bj) - (cross(linearRotSrc,jointDef_bi) + linearTrSrc) + (cross(linearRotTgt,jointDef_bj) + linearTrTgt);
			linearRes += residualPt.square_norm();

			// Actual Residual
			DeformAll_Utility::SE3_6DOF correctionSrc, correctionTgt;
			correctionSrc.createFromLinear(linearRotSrc, linearTrSrc);
			correctionTgt.createFromLinear(linearRotTgt, linearTrTgt);

			// Apply transformations
			jointDef_bi = (-correctionSrc) * jointDef_bi;
			jointDef_bj = (-correctionTgt) * jointDef_bj;
			residualPt = (jointDef_bi - jointDef_bj);
			totalRes += residualPt.square_norm();
		}
	}
}

real_type Deform::DeformGraphAll::computeJointResidual( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const DeformAll_Utility::TransformList& tfList ) const
{
	real_type totalRes = 0.0;
	const int numBones = defOpt.numBones;

	// Iterate through each unique pair
	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		if (!varTranslate.validBone(targetFrameIndex, bi) || !varTranslate.validBone(targetFrameIndex, bj)) {
			continue;
		}

		const DeformGraphAll_Utility::Joint& joint = it->second;
		const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
		const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

		int numSamples = 0;
		if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
			numSamples = 1;
		}
		else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
			numSamples = defOpt.hingeJointSamples;
		}

		for (int j = 0; j < numSamples; ++j) {
			real3_type jointLocation;

			if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
				jointLocation = jointPos;
			}
			else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				real_type t = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
				jointLocation = jointPos +  t*jointCon;
			}

			real3_type jointDef_bi, jointDef_bj;
			std::vector<data_type> RJoint;

			// Transform the joint position with each bone transformation
			// Idea 1: constrain inverse
			jointDef_bi = (-tfList[bi]) * jointLocation;
			jointDef_bj = (-tfList[bj]) * jointLocation;

			// Populate RHS residual entries
			real3_type residualJoint = jointDef_bi - jointDef_bj;

			totalRes += residualJoint.square_norm();
		}
	}

	return totalRes;
}

void Deform::DeformGraphAll::applyTfCorrection( taucsType stepSize, taucsType *correction, const real_type& s, const real3_type& t, const DeformAll_Utility::TfVarTranslate& varTranslate )
{
	int frameIndex, boneIndex;

	for (int i = 0; i < varTranslate.numTotalBones(); ++i) {
		int idx = 6*i;
		varTranslate.frameBoneIndex(i, frameIndex, boneIndex);
		DeformAll_Utility::SE3_6DOF& updateTf = getFrameSoln(frameIndex)->tfList[boneIndex];

		real3_type linearRot(stepSize*correction[idx],stepSize*correction[idx+1],stepSize*correction[idx+2]);
		real3_type linearTr(stepSize*correction[idx+3],stepSize*correction[idx+4],stepSize*correction[idx+5]);
		//DeformAll_Utility::SE3_6DOF correctionTf(linearRot, linearTr);

		// Compare projection using Plucker coordinates
		DeformAll_Utility::SE3_6DOF correctionTf2;
		correctionTf2.createFromLinear(linearRot, linearTr);

		//if ((correctionTf.tr() - correctionTf2.tr()).norm() > 0.01) {
		//	cout << "*** Comparison:" << endl;
		//	cout << "Ordinary Rotation Matrix: " << correctionTf.rotMatrix() << endl;
		//	cout << "Ordinary Axis-Angle: " << correctionTf.rot() << endl;
		//	cout << "Ordinary Translation: " << correctionTf.tr() << endl;
		//	cout << "Plucker Rotation Matrix: " << correctionTf2.rotMatrix() << endl;
		//	cout << "Plucker Axis-Angle: " << correctionTf2.rot() << endl;
		//	cout << "Plucker Translation: " << correctionTf2.tr() << endl;
		//}

		//// Transform the translation
		//real3_type tNew = correctionTf2.tr() / s - correctionTf2.rotMatrix() * t + t;
		//DeformAll_Utility::SE3_6DOF correct(correctionTf2.rotMatrix(), tNew);
		//updateTf = correct * updateTf;

		updateTf = correctionTf2 * updateTf;
	}
}

void Deform::DeformGraphAll::estimateJoints(bool printDebug/* = false*/)
{
	const int numBones = defOpt.numBones;
	const int numPairs = EnumerateUniquePairs<int>::numTotalPairs(numBones);
	DeformGraphAll_Utility::JointMap newJointMap;

	// Roughly estimate joint 3D locations
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	// 1. Count boundary edges
	idxList numEdges(numPairs,0);
	std::vector<Geometry::Vec3d> sumPts(numPairs,Geometry::Vec3d(0.0,0.0,0.0));
	BoostIndexMap index = get(vertex_index, allSampleGraph->graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(allSampleGraph->graph); ei != ei_end; ++ei) {
		const int idx1 = index[source(*ei,allSampleGraph->graph)];
		const int idx2 = index[target(*ei,allSampleGraph->graph)];
		const int l1 = findLabel(allSampleGraph->weightList[idx1]);
		const int l2 = findLabel(allSampleGraph->weightList[idx2]);

		if (l1 != l2) {
			const int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones, min(l1,l2), max(l1,l2));
			const real3_type& p1 = vRef[idx1];
			const real3_type& p2 = vRef[idx2];
			sumPts[linIdx] += Geometry::Vec3d(p1[0],p1[1],p1[2]);
			sumPts[linIdx] += Geometry::Vec3d(p2[0],p2[1],p2[2]);
			++numEdges[linIdx];
		}
	}

	// 2. Select most significant joints for each bone
	for (int i = 0; i < numBones; ++i) {
		if (printDebug) cout << DeformGraphAll_Utility::printHeader << "Joints related to bone " << i << ":" << endl;

		// Find the maximum total weight among all the joints related to this bone
		int maxEdges = 0;
		for (int j = 0; j < numBones; ++j) {
			if (i != j) {
				int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones, min(i,j), max(i,j));
				if (numEdges[linIdx] > maxEdges) {
					maxEdges = numEdges[linIdx];
				}
			}
		}

		if (maxEdges > 0) {
			// Find the ratio of each joint's weight w.r.t the maximum joint weight
			// To prune out small joints, we require that this ration is greater than a threshold (currently 0.5)
			// To disable pruning, set the ratio to be 0.0
			const double minRatio = defOpt.pruneJoint_minEdgeRatio;
			for (int j = i+1; j < numBones; ++j) {
				int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numBones, min(i,j), max(i,j));
				double ratio = ((double)numEdges[linIdx]) / ((double)maxEdges);

				if (printDebug && ratio > 0) {
					cout << DeformGraphAll_Utility::printHeader << "-- Pair (" << i << "," << j << ") interaction ratio to max: " << ratio << " (numEdges/maxEdges = ";
					cout << numEdges[linIdx] << "/" << maxEdges << ")";
				}

				if (ratio > minRatio) {
					// Add the joint
					DeformGraphAll_Utility::Joint newJoint;
					newJoint.type = DeformGraphAll_Utility::JOINT_BALL;
					newJoint.location = sumPts[linIdx] / (2.0*((double)numEdges[linIdx]));
					newJointMap[linIdx] = newJoint;
					if (printDebug && ratio > 0) cout << " --> Passed." << endl;
				}
				else {
					if (printDebug && ratio > 0) cout << " --> Failed." << endl;
				}
			}
		}
	}

	//// Testing
	//drawMutex.lock();
	//jointMap.swap(newJointMap);
	//drawMutex.unlock();
	//pauseUntilEnter("Test");
	//drawMutex.lock();
	//jointMap.swap(newJointMap);
	//drawMutex.unlock();

	estimateJoints_Motion(newJointMap, allTfList, frameSolution.size() > defOpt.hingeJointNumFrames);

	// Update the joint information
	drawMutex.lock();
	jointMap.swap(newJointMap);
	drawMutex.unlock();
}

void Deform::DeformGraphAll::estimateJoints_Motion( DeformGraphAll_Utility::JointMap& newJointMap, const DeformAll_Utility::SparseTLArray& allTfList, bool detectHinges )
{
	const bool printDebug = false;
	const int numBones = defOpt.numBones;
	std::vector<int> removeJoints;

	//////////////////////////////////////////////////////////////////////////
	// Estimate the true joint location based on the motion in each frame
	DeformGraphAll_Utility::JointMap::iterator it;
	for (it = newJointMap.begin(); it != newJointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);
		const DeformGraphAll_Utility::Joint& joint = it->second;

		// Construct linear least squares system for estimating the joint location
		std::vector< Geometry::Vec<DIMENSION,double> > lhsRows;
		std::vector<double> rhsEntries;

		// Add per-frame transformation constraints
		DeformAll_Utility::SparseTLArray::const_iterator it;
		for (it = allTfList.begin(); it != allTfList.end(); ++it) {
			const int frameIndex = it->first;

			// Skip the reference frame and any new frames
			if (frameIndex == referenceFrameIndex || frameIndex > lastFrame) {
				continue;
			}

			// Get the transformations for this frame
			const DeformAll_Utility::TransformList& tfList = *(it->second);

			// Avoid occluded bones
			std::vector<bool> isBoneOccluded = getFrameSoln(frameIndex)->isBoneOccludedOrExcluded(excludedLabels);

			if (!isBoneOccluded[bi] && !isBoneOccluded[bj]) {
				// Add the constraint
				DeformAll_Utility::SE3_6DOF tf_bi, tf_bj;
				tf_bi = -tfList[bi];
				tf_bj = -tfList[bj];

				Geometry::Matrix33f M = tf_bi.rotMatrix() - tf_bj.rotMatrix();
				real3_type t = tf_bi.tr() - tf_bj.tr();

				for (int d = 0; d < DIMENSION; ++d) {
					Geometry::Vec<DIMENSION,double> temp;
					temp[0] = M(d,0); temp[1] = M(d,1); temp[2] = M(d,2);
					lhsRows.push_back(temp);
					rhsEntries.push_back(-t[d]);
				}
			}
		}


		// Add location constraint
		const double lambda = 0.1;
		const Geometry::Vec3d& gridEstLoc = joint.location;
		for (int d = 0; d < DIMENSION; ++d) {
			Geometry::Vec<3,double> temp;
			temp[0] = 0.0; temp[1] = 0.0; temp[2] = 0.0;
			temp[d] = lambda;
			lhsRows.push_back(temp);
			rhsEntries.push_back(lambda * gridEstLoc[d]);
		}

		// Solve SVD
		Geometry::Vec3d solvedLoc;
		DeformAll_Utility::solveLinear_denseSVD(lhsRows, rhsEntries, solvedLoc);

		//////////////////////////////////////////////////////////////////////////
		// Hinge Detection
		//////////////////////////////////////////////////////////////////////////
		// Remove location constraint
		for (int d = 0; d < DIMENSION; ++d) {
			lhsRows.pop_back();
			rhsEntries.pop_back();
		}

		Geometry::Vec<DIMENSION,double> jointPoint, jointLineDir;
		bool hinge = false;
		double hingeThreshold = defOpt.jointIsHingeThreshold;

		if (!detectHinges) {
			jointPoint = solvedLoc;
		}
		else
		{
			// Compute SVD
			// Allocate the storage space for the input/output arrays
			const int N_col = DIMENSION;
			const int N_row = lhsRows.size();
			const int minMN = min(N_row,N_col);
			double **A = new double*[N_row];
			double *pS = new double[minMN];
			double **pU = new double*[N_row];
			double **pVT = new double*[minMN];
			for (int i = 0; i < N_row; ++i) {
				A[i] = new double[N_col];
				pU[i] = new double[minMN];
				for (int j = 0; j < N_col; j++) {
					A[i][j] = lhsRows[i][j];
				}
			}
			for (int i = 0; i < minMN; ++i) {
				pVT[i] = new double[N_col];
			}


			//cout << "A = [" << endl;
			//displayMatrix(N_row, N_col, A);
			//cout << "]" << endl;



			// Compute SVD using LAPACK
			LapackLib::dgesvd(A, N_row, N_col, pS, pU, pVT);



			//cout << "U = [" << endl;
			//displayMatrix(N_row, minMN, pU);
			//cout << "]" << endl;

			//cout << "VT = [" << endl;
			//displayMatrix(minMN, N_col, pVT);
			//cout << "]" << endl;

			//cout << "rhs: [" << endl;
			//for (int i = 0; i < rhsEntries.size(); ++i) {
			//	cout << rhsEntries[i] << endl;
			//}
			//cout << "]" << endl;



			// Look at the singular values and determine joint type
			double sing_sum = 0.0;
			for (int i = 0; i < minMN; ++i) {
				sing_sum += pS[i];
			}
			double ratio = pS[minMN-1] / sing_sum;

			if (ratio < hingeThreshold) {
				if (printDebug) {
					cout << "--> Hinge joint detected (ratio:" << ratio << ")" << endl;
				}

				hinge = true;
				// Solve for hinge line
				double **aug = new double*[minMN];
				double **L = new double*[minMN];
				double **U = new double*[minMN];
				for (int i = 0; i < minMN; ++i) {
					aug[i] = new double[N_col+1];
					L[i] = new double[N_col+1];
					U[i] = new double[N_col+1];
					for (int col = 0; col <= N_col; ++col) {
						aug[i][col] = 0.0;
						L[i][col] = 0.0;
						U[i][col] = 0.0;
					}
				}

				// left: compute S*VT
				for (int i = 0; i < minMN; ++i) {
					for (int j = 0; j < N_col; ++j) {
						aug[i][j] = pS[i] * pVT[i][j];
					}
				}

				// right: compute -U^T * rhs
				for (int i = 0; i < N_col; ++i) {
					aug[i][N_col] = 0.0;
					for (int j = 0; j < N_row; ++j) {
						aug[i][N_col] += pU[j][i] * rhsEntries[j];
					}
				}

				// Zero out the last row of aug
				for (int col = 0; col <= N_col; ++col) {
					aug[minMN-1][col] = 0.0;
				}

				//cout << "aug:" << endl;
				//displayMatrix(minMN, N_col+1, aug);

				// Compute LU factorization
				LapackLib::dgetrf(minMN, N_col+1, aug, L, U);

				//cout << "L: " << endl;
				//displayMatrix(minMN, N_col+1, L);
				//cout << "U: " << endl;
				//displayMatrix(minMN, N_col+1, U);

				// Perform rref
				double temp = 1.0 / U[0][0];
				for (int col = 0; col <= N_col; ++col) {
					U[0][col] *= temp;
				}
				temp = 1.0 / U[1][1];
				for (int col = 0; col <= N_col; ++col) {
					U[1][col] *= temp;
				}
				temp = U[0][1];
				for (int col = 0; col <= N_col; ++col) {
					U[0][col] -= U[1][col]*temp;
				}

				//cout << "rref U: " << endl;
				//displayMatrix(minMN, N_col+1, U);

				jointPoint[0] = U[0][N_col]; jointPoint[1] = U[1][N_col]; jointPoint[2] = 0.0;
				jointLineDir[0] = -U[0][N_col-1]; jointLineDir[1] = -U[1][N_col-1]; jointLineDir[2] = 1.0;

				//cout << "Line Point: " << jointPoint << endl;
				//cout << "Line Dir: " << jointLineDir << endl;

				temp = - (jointLineDir * (jointPoint - gridEstLoc)) / jointLineDir.square_norm();
				jointPoint += jointLineDir * temp;
				jointLineDir /= jointLineDir.norm();

				//cout << "Line Point: " << jointPoint << endl;
				//cout << "Line Dir: " << jointLineDir << endl;

				for (int i = 0; i < minMN; ++i) {
					delete aug[i];
					delete L[i];
					delete U[i];
				}
				delete aug;
				delete L;
				delete U;
			}
			else {
				if (printDebug) {
					cout << "--> Ball joint detected (ratio:" << ratio << ")" << endl;
				}

				jointPoint = solvedLoc;
			}

			// Delete the vector and matrices
			for (int i = 0; i < N_row; ++i) {
				delete[] A[i];
				delete[] pU[i];
			}
			for (int i = 0; i < minMN; ++i) {
				delete[] pVT[i];
			}
			delete[] A;
			delete[] pS;
			delete[] pU;
		}

		//////////////////////////////////////////////////////////////////////////
		// Prune based on distance from estimated joint location to solved joint location
		double dist = (jointPoint - gridEstLoc).norm();
		if (printDebug) {
			// Print information on screen
			cout << DeformGraphAll_Utility::printHeader << "Found joint location for (" << bi << "," << bj << ") using " << (lhsRows.size() / 3) - 1 << " tfs." << endl;
			cout << DeformGraphAll_Utility::printHeader << "Distance to location on grid: " << dist;
		}
		if (dist > defOpt.maxJointDist) {
			if (printDebug) cout << " --> Joint Pruned.";

			// Mark the joint for removal
			removeJoints.push_back(linIdx);
		}
		else {
			if (printDebug) cout << endl;

			// Construct joint
			DeformGraphAll_Utility::Joint newJoint;
			newJoint.location = jointPoint;
			if (hinge) {
				newJoint.type = DeformGraphAll_Utility::JOINT_HINGE;
				newJoint.constraint = jointLineDir;
			}
			else {
				newJoint.type = DeformGraphAll_Utility::JOINT_BALL;
			}
			newJointMap[linIdx] = newJoint;
		}
	}

	// Remove joints
	for (int i = 0; i < removeJoints.size(); ++i) {
		DeformGraphAll_Utility::JointMap::iterator it = newJointMap.find(removeJoints[i]);
		if (it != newJointMap.end()) {
			newJointMap.erase(it);
		}
		else {
			cerr << "Error in estimateJoints_Motion(): can't find joint " << removeJoints[i] << " that needs to be removed!!" << endl;
		}
	}
}

void Deform::DeformGraphAll::precomputeSampleError_AllFrames( bool forceAll /*= false */ )
{
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		int frameIndex = it->first;
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;

		if (frameIndex != referenceFrameIndex) {
			if (forceAll || frameSoln->solveThisFrame) {
				// Find the closest points (computed using the currently assigned weights)
				DeformGraphAll_Utility::AllClosestPoints::const_iterator it3 = sampleClosestPoints.find(frameIndex);
				if (it3 == sampleClosestPoints.end()) {
					cerr << "Error in precomputeSampleError_AllFrames(): cannot find closest points for frame " << frameIndex << endl;
					break;
				}
				const DeformGraphAll_Utility::ClosestPointMap& closestPts = it3->second;

				DeformGraphAll_Utility::Error_LabelSample errorLS;
				std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
				DeformGraphAll_Utility::precomputeSampleError(frameSoln->originalShape, frameIndex, allTfList, *allSampleGraph, isOccluded, closestPts, defOpt.closestPt_normalWeight, defOpt.accumulateASGOnReference, errorLS);

				// Assign the result
				drawMutex.lock();
				allFrameError[frameIndex].swap(errorLS);
				drawMutex.unlock();
			}
		}
	}


	// Add up the total error
	const int numBones = defOpt.numBones;
	const int numSamples = allSampleGraph->numSamples();
	DeformGraphAll_Utility::Error_LabelSample totalLS(numBones,distList(numSamples,0.0));
	DeformGraphAll_Utility::Error_LabelSample countLS(numBones,distList(numSamples,0.0));

	DeformGraphAll_Utility::FrameError_LabelSample::const_iterator it2;
	for (it2 = allFrameError.begin(); it2 != allFrameError.end(); ++it2)
	{
		const int frameIndex = it2->first;
		const DeformGraphAll_Utility::Error_LabelSample& frameLS = it2->second;

		// Find the closest points (computed using the currently assigned weights)
		DeformGraphAll_Utility::AllClosestPoints::const_iterator it3 = sampleClosestPoints.find(frameIndex);
		if (it3 == sampleClosestPoints.end()) { cerr << "Error in precomputeSampleError_AllFrames(): cannot find closest points for frame " << frameIndex << endl; break; }
		const DeformGraphAll_Utility::ClosestPointMap& closestPts = it3->second;

		for (int l = 0; l < numBones; ++l)
		{
			// For each sample that has a closest point on this frame
			// i.e. distance not computed for samples that don't have a valid corresponding point computed using the currently assigned weight
			DeformGraphAll_Utility::ClosestPointMap::const_iterator it4;
			for (it4 = closestPts.begin(); it4 != closestPts.end(); ++it4)
			{
				const int i = it4->first;
				totalLS[l][i] += frameLS[l][i];
				countLS[l][i] += 1.0;
			}
		}
	}

	// Compute the average error per label and per sample
	for (int l = 0; l < numBones; ++l)
	{
		for (int i = 0; i < numSamples; ++i)
		{
			if (countLS[l][i] > 0.0)
			{
				totalLS[l][i] /= countLS[l][i];
			}
			//else
			//{
			//	cout << "Error in precomputeSampleError_AllFrames(): the number of samples in totalLS[" << l << "][" << i << "] is " << countLS[l][i] << ", final error is: " << totalLS[l][i] << "!" << endl;
			//}
		}
	}

	// Assign result
	drawMutex.lock();
	sumError.swap(totalLS);
	drawMutex.unlock();
}

real_type Deform::DeformGraphAll::findSampleLabels()
{
	MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_1,AddStart);

	// We will exclude the excluded labels from being assigned in the graph cut optimization.
	std::vector<bool> isExcluded(defOpt.numBones, false);
	for (int i = 0; i < excludedLabels.size(); ++i) {
		isExcluded[excludedLabels[i]] = true;
	}
	std::vector<int> translateToGC;
	translateToOrig.clear();
	for (int i = 0; i < defOpt.numBones; ++i) {
		if (!isExcluded[i]) {
			translateToGC.push_back(translateToOrig.size());
			translateToOrig.push_back(i);
		}
		else {
			translateToGC.push_back(-1);
		}
	}

	// Solve for the labels on each sample
	real_type energy;
	const int numSites = allSampleGraph->numSamples();
	const int numLabels = translateToOrig.size();
	GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph(numSites, numLabels);

	findClosestPoints_AllFrames();		/**< The closest point for the currently assigned weight will change sumError. */
	precomputeSampleError_AllFrames(recomputeClosestPtError_AllFrames);

	// Reset the recomputation of closest points / errors on all frames
	recomputeClosestPtError_AllFrames = false;

	MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_1,Stop);

	try {
		cout << DeformGraphAll_Utility::printHeader << "Optimizing graph cuts with " << numLabels << " labels..." << endl;
		cout << DeformGraphAll_Utility::printHeader << "Label translation to graph cut: [";
		for (int i = 0; i < translateToGC.size(); i++)
		{
			if (i != 0) cout << ",";
			cout << translateToGC[i];
		}
		cout << "]" << endl;
		cout << DeformGraphAll_Utility::printHeader << "Label translation to original: [";
		for (int i = 0; i < translateToOrig.size(); i++)
		{
			if (i != 0) cout << ",";
			cout << translateToOrig[i];
		}
		cout << "]" << endl;
		cout << DeformGraphAll_Utility::printHeader << "Data/Smoothness term weights: " << defOpt.WWeights[0] << "/" << defOpt.WWeights[1] << endl;

		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_2,AddStart);

		// Use function pointers for the data and smoothness costs
		gc->setDataCost(&DeformGraphAll::GraphCutDataFn, this);
		gc->setSmoothCost(&DeformGraphAll::GraphCutSmoothFn, this);

		// Set up edges
		BoostEdgeIterator ei, ei_end;
		BoostIndexMap index = get(vertex_index, allSampleGraph->graph);
		for (tie(ei, ei_end) = edges(allSampleGraph->graph); ei != ei_end; ++ei) {
			int p = index[source(*ei,allSampleGraph->graph)];
			int q = index[target(*ei,allSampleGraph->graph)];
			gc->setNeighbors(p,q);
		}

		// Initialize the labeling with the current sample labels
		for (int i = 0; i < numSites; i++) {
			gc->setLabel(i, translateToGC[findLabel(allSampleGraph->weightList[i])]);
		}

		// Perform optimization
		energy = gc->compute_energy();
		cout << "*** Before optimization energy is: " << energy << endl;
		energy = gc->expansion();
		cout << "*** After expansion, energy is: " << energy << endl;

		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_2,Stop);


		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_3,AddStart);
		
		// Read out the labeling
		std::vector<int> assignedLabels(numSites);
		for (int i = 0; i < numSites; ++i) {
			assignedLabels[i] = translateToOrig[gc->whatLabel(i)];
		}

		// If performing accumulation strategy, then replace back the occluded labels
		if (defOpt.accumulateASGOnReference) {
			std::vector<bool> isOccluded = getFrameSoln(lastFrame+1)->isBoneOccluded();
			for (int i = 0; i < numSites; i++) {
				int origLabel = findLabel(allSampleGraph->weightList[i]);
				if (isOccluded[origLabel]) {
					assignedLabels[i] = origLabel;
				}
			}
		}

		//pauseUntilEnter("Solved Weight Optimization");

		// Remove small regions and reuse extra labels
		DeformAll_Utility::SparseTLArray allTfList;
		makeAllTFList(frameSolution, allTfList);
		PointList vRef, nRef;
		allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);
		bool temp1, temp2;
		temp1 = discardSmallRegions(vRef, &assignedLabels);
		temp2 = reuseExtraLabels(vRef, assignedLabels);

		// Recompute closest point error on all frames if we discarded or reused labels
		if (!defOpt.accumulateASGOnReference && (temp1 || temp2)) {
			recomputeClosestPtError_AllFrames = true;
		}

		//pauseUntilEnter("Done reusing the labels");

		// Assign the labels
		drawMutex.lock();
		for (int i = 0; i < numSites; ++i) {
			DeformAll_Utility::SparseWeight sw(1, DeformAll_Utility::IndexedWeight(assignedLabels[i], 1.0));
			allSampleGraph->weightList[i].swap(sw);
		}
		drawMutex.unlock();

		MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_3,Stop);
	}
	catch (GCException e) {
		e.Report();
	}

	// Free memory associated with graph cut data structure
	delete gc;

	// Return final energy
	return energy;
}

real_type Deform::DeformGraphAll::GraphCutDataFn( int p, int l, void *data )
{
	DeformGraphAll& deformGraphAll = *((DeformGraphAll*)data);
	real_type errorValue = deformGraphAll.sumError[deformGraphAll.translateToOrig[l]][p];
	return deformGraphAll.defOpt.WWeights[0]*errorValue;
}

real_type Deform::DeformGraphAll::GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data )
{
	DeformGraphAll& deformGraphAll = *((DeformGraphAll*)data);
	real_type errorValue = 0.f;

	// Constant penalty if labels are different
	if (l1 != l2) {
		errorValue = 1.f;
	}

	return deformGraphAll.defOpt.WWeights[1]*errorValue;
}

bool Deform::DeformGraphAll::reuseExtraLabels( const PointList& samplePoints, std::vector<int>& labelArray )
{
	bool labelReused = false;
	const int numLabels = defOpt.numBones;
	const int numSamples = allSampleGraph->numSamples();
	data_type totalLabelError(numLabels, 0.0);
	data_type avgLabelError(numLabels, 0.0);
	data_type countCorrespPerLabel(numLabels, 0.0);

	for (int i = 0; i < numSamples; ++i) {
		int label = labelArray[i];
		totalLabelError.val[label] += sumError[label][i];
		countCorrespPerLabel.val[label] += 1.0;
	}
	// Compute average error per label, in units of scene grid spacing
	for (int i = 0; i < numLabels; i++) {
		if (countCorrespPerLabel.val[i] > 0.0) {
			avgLabelError.val[i] = totalLabelError.val[i] / (countCorrespPerLabel.val[i] * defOpt.sceneGridSpacing);
		}
	}

	cout << DeformGraphAll_Utility::printHeader << "Reusing extra labels..." << endl;
	cout << DeformGraphAll_Utility::printHeader << "-- Total:   " << totalLabelError << endl;
	cout << DeformGraphAll_Utility::printHeader << "-- Count:   " << countCorrespPerLabel << endl;
	cout << DeformGraphAll_Utility::printHeader << "-- Average: " << avgLabelError << endl;
	cout << DeformGraphAll_Utility::printHeader << ">> Split when avg error is greater than " << defOpt.splitRegionMinError << endl;


	while (excludedLabels.size()) {
		// a. Find the label that needs improvement
		int maxErrorLabel;
		real_type maxError = avgLabelError.findMax(maxErrorLabel);
		cout << DeformGraphAll_Utility::printHeader << "------ Label with maximum avg error: " << maxErrorLabel << " with error " << maxError << endl;
		avgLabelError.val[maxErrorLabel] = 0.0;		// Don't address this label any further after this iteration


		if (maxError <= defOpt.splitRegionMinError) {
			break;
		}
		else {
			// Find the extra label
			int extraLabel = excludedLabels[0];

			// Randomly subdivide the label with the largest error
			// a. Find all cells belonging to the maximum label
			PointList maxLabelSamples;
			idxList maxLabelSamplesIndex;
			for (int i = 0; i < numSamples; ++i) {
				if (labelArray[i] == maxErrorLabel) {
					maxLabelSamples.add(samplePoints[i]);
					maxLabelSamplesIndex.push_back(i);
				}
			}

			// b. Find two points that are somewhat apart within the label's region
			idxList sampling = bestCandidateSample(maxLabelSamples, 2);
			real3_type center1, center2;
			center1 = maxLabelSamples[sampling[0]];
			center2 = maxLabelSamples[sampling[1]];

			// c. Perform the relabeling
			for (int i = 0; i < maxLabelSamples.size(); i++) {
				// Evaluate distance
				real_type val1 = (maxLabelSamples[i] - center1).norm();
				real_type val2 = (maxLabelSamples[i] - center2).norm();

				if (val1 < val2) {
					// Assign new label
					labelArray[maxLabelSamplesIndex[i]] = extraLabel;
				}
			}

			// d. Update
			// Copy the deformation in the new label, for all frames
			DeformGraphAll_Utility::SolutionTable::const_iterator frame_it;
			for (frame_it = frameSolution.begin(); frame_it != frameSolution.end(); ++frame_it) {
				DeformGraphAll_Utility::FrameSolution *frameSoln = frame_it->second;
				DeformAll_Utility::TransformList& tfList = frameSoln->tfList;
				tfList[extraLabel] = tfList[maxErrorLabel];

				// Remove any occlusion information for substituted labels
				std::vector<int> newArray;
				for (int i = 0; i < frameSoln->occludedLabels.size(); ++i) {
					if (frameSoln->occludedLabels[i] != extraLabel) {
						newArray.push_back(frameSoln->occludedLabels[i]);
					}
				}
				frameSoln->occludedLabels.swap(newArray);
			}

			// Reset the label use stats
			discardSmallRegions(samplePoints, &labelArray);

			// Mark flag that we did reuse a label
			labelReused = true;
		}
	}
	if (labelReused) cout << DeformGraphAll_Utility::printHeader << "Some labels were reused." << endl;

	return labelReused;
}

void Deform::DeformGraphAll::rebuildASG( real_type maxStretchFactor, real_type maxEdgeLength, bool useAllPoints, bool keepJointEdges )
{
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	// Divide the points into groups corresponding to each bone
	const int numBones = defOpt.numBones;
	std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
	for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
		int label = findLabel(allSampleGraph->weightList[i]);
		vRef_bone[label].add(vRef[i]);
		nRef_bone[label].add(nRef[i]);
	}

	//cout << "Divided " << vRef.size() << " points into " << numBones << " groups:" << endl;
	//for (int i = 0; i < numBones; ++i) {
	//	cout << "-- Group " << i << ": " << vRef_bone[i].size() << endl;
	//}

	// Completely resample the ASG from all frames
	idxList frameList;
	getSortedFrameList(frameSolution, frameList);

	DeformAll_Utility::AllSampleGraph *newASG, *oldASG;
	oldASG = allSampleGraph;
	newASG = new DeformAll_Utility::AllSampleGraph(defOpt.EKNNGraph_k, defOpt.refSampleDist);

	for (int i = 0; i < frameList.size(); ++i) {
		cout << DeformGraphAll_Utility::printHeader << "Adding samples from frame " << frameList[i];

		// Add this frame's samples to the mix, re-sampling previous ASG
		DeformGraphAll_Utility::FrameSolution* frameSoln = frameSolution.find(frameList[i])->second;
		std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
		const idxList* sampleIdx = &(frameSoln->frameSamples);

		// Check if we want to use all points in this frame
		if (useAllPoints) {
			idxList allPointIdx(frameSoln->originalShape.numVertices());
			for (int i = 0; i < allPointIdx.size(); ++i) {
				allPointIdx[i] = i;
			}

			// IMPORTANT SIDE EFFECT!!! Replace the frame's samples entirely.
			frameSoln->frameSamples.swap(allPointIdx);
			sampleIdx = &(frameSoln->frameSamples);
		}

		newASG->rebuild(frameSoln->originalShape, *sampleIdx, frameList[i], isOccluded, vRef_bone, nRef_bone, allTfList, defOpt.refSampleDist, false, referenceFrameIndex);

		cout << " --> " << newASG->numSamples() << endl;
	}


	// Rebuild EKNN graph on the reference frame
	newASG->getVertexNormal_Ref(allTfList, vRef, nRef);
	makeEKNNgraph(newASG->EKNNgraph_k, vRef, newASG->graph);

	// Prune edges: look at transformations for all frames
	cout << "Graph size V,E (before pruning): " << num_vertices(newASG->graph) << "," << num_edges(newASG->graph) << endl;
	if (keepJointEdges) {
		newASG->pruneEdges(allTfList, maxStretchFactor, maxEdgeLength, &jointMap);
	}
	else {
		newASG->pruneEdges(allTfList, maxStretchFactor, maxEdgeLength, NULL);
	}
	cout << "Graph size V,E (after pruning):  " << num_vertices(newASG->graph) << "," << num_edges(newASG->graph) << endl;


	drawMutex.lock();
	// Replace previous ASG
	allSampleGraph = newASG;

	// Clear closest points and errors
	sampleClosestPoints.clear();
	allFrameError.clear();
	sumError.clear();
	drawMutex.unlock();

	// Clean up
	delete oldASG;
}


void Deform::DeformGraphAll::filterASGVertices( const std::vector<bool> &filter )
{
	const int numVertices = allSampleGraph->numSamples();

	if (filter.size() != numVertices) {
		cerr << "Error in filterASGVertices(): filter size doesn't match number of samples!" << endl;
		return;
	}

	// 4. Remove the vertices which have degree less than some threshold
	std::vector<int> newVertexTrans(numVertices, -1);
	int idx = 0;
	for (int i = 0; i < numVertices; ++i) {
		if (filter[i]) {
			newVertexTrans[i] = idx;
			++idx;
		}
	}
	const int numCleanedVertices = idx;

	// Replace old ASG with cleaned-up ASG
	cout << "Number of vertices in cleaned graph: " << numCleanedVertices << endl;
	DeformAll_Utility::AllSampleGraph *newASG, *oldASG;
	oldASG = allSampleGraph;
	newASG = new DeformAll_Utility::AllSampleGraph(defOpt.EKNNGraph_k, defOpt.refSampleDist);

	newASG->weightList.resize(numCleanedVertices);
	newASG->vertex.resize(numCleanedVertices);
	newASG->normal.resize(numCleanedVertices);
	newASG->fIndex.resize(numCleanedVertices);
	newASG->sIndex.resize(numCleanedVertices);

	for (int i = 0; i < oldASG->vertex.size(); ++i) {
		if (newVertexTrans[i] >= 0) {
			// Add to the list
			int idx = newVertexTrans[i];
			newASG->weightList[idx] = oldASG->weightList[i];
			newASG->vertex[idx] = oldASG->vertex[i];
			newASG->normal[idx] = oldASG->normal[i];
			newASG->fIndex[idx] = oldASG->fIndex[i];
			newASG->sIndex[idx] = oldASG->sIndex[i];
		}
	}

	// Translate edges
	newASG->graph = BoostGraph(numCleanedVertices);
	BoostIndexMap index = get(vertex_index, oldASG->graph);
	BoostEdgeWeightMap_const edgeWeight = get(edge_weight, (const BoostGraph&) oldASG->graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(oldASG->graph); ei != ei_end; ++ei) {
		int e1 = newVertexTrans[index[source(*ei,oldASG->graph)]];
		int e2 = newVertexTrans[index[target(*ei,oldASG->graph)]];
		if (e1 >= 0 && e2 >= 0) {
			add_edge(e1, e2, edgeWeight[*ei], newASG->graph);
		}
	}

	cout << "filterASGVertices(): Graph size V,E (before cleaning): " << num_vertices(oldASG->graph) << "," << num_edges(oldASG->graph) << endl;
	cout << "filterASGVertices(): Graph size V,E (after cleaning):  " << num_vertices(newASG->graph) << "," << num_edges(newASG->graph) << endl;
	//pauseUntilEnter("Replacing with cleaned graph");

	drawMutex.lock();
	// Replace previous ASG
	allSampleGraph = newASG;
	// Clear closest points and errors
	sampleClosestPoints.clear();
	allFrameError.clear();
	sumError.clear();
	drawMutex.unlock();

	// Clean up
	delete oldASG;
}

void Deform::DeformGraphAll::cleanASG()
{
	// 2. Count the degrees of each vertex
	int maxDegree = 0;
	idxList vertexDegree(allSampleGraph->vertex.size(), 0);
	BoostGraph& graph = allSampleGraph->graph;
	BoostIndexMap index = get(vertex_index, graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(graph); ei != ei_end; ++ei) {
		int temp = ++vertexDegree[index[source(*ei,graph)]];
		if (temp > maxDegree) {
			maxDegree = temp;
		}
		temp = ++vertexDegree[index[target(*ei,graph)]];
		if (temp > maxDegree) {
			maxDegree = temp;
		}
	}
	cout << "Maximum degree: " << maxDegree << endl;

	// 3. Make a histogram of all the degrees
	idxList degreeHist(maxDegree+1,0);
	for (int i = 0; i < vertexDegree.size(); ++i) {
		++degreeHist[vertexDegree[i]];
	}

	for (int i = 0; i < degreeHist.size(); ++i) {
		cout << "# vertices having degree " << i << ": " << degreeHist[i] << endl;
	}

	// 4. Remove the vertices which have degree less than some threshold
	int minDegree = 5;
	std::vector<bool> vertexFilter(vertexDegree.size());
	for (int i = 0; i < vertexDegree.size(); ++i) {
		vertexFilter[i] = (vertexDegree[i] >= minDegree);
	}

	filterASGVertices(vertexFilter);
}

void Deform::DeformGraphAll::cleanASGWeights( int minSizeThreshold )
{
	// Rebuilding graph (don't prune edges)
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);
	drawMutex.lock();
	makeEKNNgraph(allSampleGraph->EKNNgraph_k, vRef, allSampleGraph->graph);
	drawMutex.unlock();
	cout << "Graph size V,E (no pruning): " << num_vertices(allSampleGraph->graph) << "," << num_edges(allSampleGraph->graph) << endl;

	pauseUntilEnter("Cleaning ASG Weights");

	const int numVertices = allSampleGraph->numSamples();
	std::vector<int> componentLabel(numVertices, 0);
	std::vector<int> componentSize(1,-1);
	int compNumber = 0;
	BoostIndexMap index = get(vertex_index, allSampleGraph->graph);

	while (true) {
		// 1. Find an unvisited vertex
		int unvisited = -1;
		for (int i = 0; i < numVertices; ++i) {
			if (componentLabel[i] == 0) {
				unvisited = i;
				break;
			}
		}
		if (unvisited == -1)  {
			break;
		}

		// Create a new component label
		++compNumber;
		componentSize.push_back(0);
		std::deque<int> visitQueue;
		visitQueue.push_back(unvisited);

		//cout << "Creating component: " << compNumber << endl;

		while (!visitQueue.empty()) {
			// 1. Pop from the front
			int i = visitQueue.front();
			visitQueue.pop_front();

			if (componentLabel[i] == 0) {
				// 2. Visit the vertex
				componentLabel[i] = compNumber;
				++componentSize[compNumber];
				int currentWeight = findLabel(allSampleGraph->weightList[i]);

				//cout << "Visiting: " << i << "(w" << currentWeight << ") -->";

				// 3. Visit all neighbors that have the same weight
				BoostVertex current = vertex(i, allSampleGraph->graph);
				BoostAdjacencyIterator ai, ai_end;
				for (tie(ai, ai_end) = adjacent_vertices(current, allSampleGraph->graph); ai != ai_end; ++ai) {
					BoostVertex neighbor = *ai;
					int neighbor_index = index[neighbor];

					// Check if it has the same weight and has not been visited yet
					int neighborWeight = findLabel(allSampleGraph->weightList[neighbor_index]);
					//cout << " " << neighbor_index  << "(w" << neighborWeight << ")";

					if ((currentWeight == neighborWeight) && (componentLabel[neighbor_index] == 0)) {
						// Add to queue
						visitQueue.push_back(neighbor_index);
						//cout << "*";
					}
				}

				//cout << endl;
			}
		}
	}

	cout << "Total number of components: " << compNumber;
	cout << "Component sizes: [";
	for (int i = 0; i < componentSize.size(); ++i)
	{
		if (i != 0) cout << " ";
		cout << componentSize[i];
	}
	cout << "]" << endl;

	// Apply threshold to component size
	// Look at neighboring vertices to find a replacement label
	for (int i = 0; i < numVertices; ++i) {
		if (componentSize[componentLabel[i]] <= minSizeThreshold) {
			int currentWeight = findLabel(allSampleGraph->weightList[i]);
			BoostVertex current = vertex(i, allSampleGraph->graph);
			BoostAdjacencyIterator ai, ai_end;
			for (tie(ai, ai_end) = adjacent_vertices(current, allSampleGraph->graph); ai != ai_end; ++ai) {
				BoostVertex neighbor = *ai;
				int neighbor_index = index[neighbor];
				int neighborWeight = findLabel(allSampleGraph->weightList[neighbor_index]);

				if ((currentWeight != neighborWeight) && (componentSize[componentLabel[neighbor_index]] > minSizeThreshold)) {
					allSampleGraph->weightList[i][0].index = neighborWeight;
				}
			}
		}
	}

	//// List of valid vertices
	//PointList vRef, nRef;
	//DeformAll_Utility::SparseTLArray allTfList;
	//makeAllTFList(frameSolution, allTfList);
	//allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);
	//PointList validVertexList;
	//for (int i = 0; i < numVertices; ++i) {
	//	if (componentSize[componentLabel[i]] >= threshold) {
	//		validVertexList.add(vRef[i]);
	//	}
	//}

	//// Replace label for invalid vertices
	//for (int i = 0; i < numVertices; ++i) {
	//	if (componentSize[componentLabel[i]] < threshold) {
	//		// Find nearest neighbor
	//		idxList nearest;
	//		distList dist;
	//		validVertexList.queryKNN(vRef[i], 1, nearest, dist, true);

	//		// Assign label of nearest neighbor
	//		allSampleGraph->weightList[i][0].index = findLabel(allSampleGraph->weightList[nearest[0]]);
	//	}
	//}
}

void Deform::DeformGraphAll::cleanASGComponents( int minSizeThreshold )
{
	//pauseUntilEnter("Pruning small ASG components");

	/////////////////////////////////////////////////////////////////////
	// NOTE: The following code is taken from cleanASGWeights() with minor modifications.
	/////////////////////////////////////////////////////////////////////
	const int numVertices = allSampleGraph->numSamples();
	std::vector<int> componentLabel(numVertices, 0);
	std::vector<int> componentSize(1,-1);
	int compNumber = 0;
	BoostIndexMap index = get(vertex_index, allSampleGraph->graph);

	while (true) {
		// 1. Find an unvisited vertex
		int unvisited = -1;
		for (int i = 0; i < numVertices; ++i) {
			if (componentLabel[i] == 0) {
				unvisited = i;
				break;
			}
		}
		if (unvisited == -1)  {
			break;
		}

		// Create a new component label
		++compNumber;
		componentSize.push_back(0);
		std::deque<int> visitQueue;
		visitQueue.push_back(unvisited);

		//cout << "Creating component: " << compNumber << endl;

		while (!visitQueue.empty()) {
			// 1. Pop from the front
			int i = visitQueue.front();
			visitQueue.pop_front();

			if (componentLabel[i] == 0) {
				// 2. Visit the vertex
				componentLabel[i] = compNumber;
				++componentSize[compNumber];

				//cout << "Visiting: " << i << " -->";

				// 3. Visit all neighbors of this vertex
				BoostVertex current = vertex(i, allSampleGraph->graph);
				BoostAdjacencyIterator ai, ai_end;
				for (tie(ai, ai_end) = adjacent_vertices(current, allSampleGraph->graph); ai != ai_end; ++ai) {
					BoostVertex neighbor = *ai;
					int neighbor_index = index[neighbor];

					// Check if has not been visited yet
					//cout << " " << neighbor_index;
					if (componentLabel[neighbor_index] == 0) {
						// Add to queue
						visitQueue.push_back(neighbor_index);
						//cout << "*";
					}
				}

				//cout << endl;
			}
		}
	}

	cout << "Total number of components: " << compNumber;
	cout << "Component sizes: [";
	for (int i = 0; i < componentSize.size(); ++i)
	{
		if (i != 0) cout << " ";
		cout << componentSize[i];
	}
	cout << "]" << endl;


	//////////////////////////////////////////////////////////////////////
	// Filter out small components: Apply threshold to component size
	//////////////////////////////////////////////////////////////////////
	std::vector<bool> vertexFilter(numVertices);
	for (int i = 0; i < numVertices; ++i) {
		vertexFilter[i] = (componentSize[componentLabel[i]] > minSizeThreshold);
	}

	filterASGVertices(vertexFilter);
}

void Deform::DeformGraphAll::saveSolverState( const char *filename ) const
{
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in DeformGraphAll::saveSolverState(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	cout << DeformGraphAll_Utility::printHeader << "Saving solver state to " << filename << endl;

	// Save version information
	out.write(DeformGraphAll_Utility::version_msg.c_str(), sizeof(char) * (DeformGraphAll_Utility::version_msg.length()+1));
	out.write((char*) &DeformGraphAll_Utility::VER_NUM, sizeof(float));

	// Save the state of the ASG
	saveASGState(allSampleGraph, out);

	// Save data for each frame solution
	size_type numFrameSolutions = frameSolution.size();
	out.write((char*) &numFrameSolutions, sizeof(size_type));
	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		saveFrameState(it->second, out);
	}

	// Save the last solved frame index
	out.write((char*) &lastFrame, sizeof(int));

	// Save timing statistics
	double timing1 = MEASURE_PERF_NARG(TIMER_MULTIFRAME_INITIALIZE,GetDurationInSecs);
	double timing2 = MEASURE_PERF_NARG(TIMER_MULTIFRAME_GLOBAL_REG,GetDurationInSecs);
	double timing3 = MEASURE_PERF_NARG(TIMER_MULTIFRAME_RESAMPLE_ASG,GetDurationInSecs);
	out.write((char*) &timing1, sizeof(double));
	out.write((char*) &timing2, sizeof(double));
	out.write((char*) &timing3, sizeof(double));

	// New in VER_NUM 0.1f: Save information for joints
	const int numJoints = jointMap.size();
	out.write((char*) &numJoints, sizeof(int));

	DeformGraphAll_Utility::JointMap::const_iterator it_joint;
	for (it_joint = jointMap.begin(); it_joint != jointMap.end(); ++it_joint) {
		int jointIndex = it_joint->first;
		const DeformGraphAll_Utility::Joint& jointInfo = it_joint->second;

		out.write((char *) &jointIndex, sizeof(int));
		out.write((char *) &(jointInfo.type), sizeof(int));

		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char *) &(jointInfo.location[d]), sizeof(double));
		}

		if (jointInfo.type == DeformGraphAll_Utility::JOINT_HINGE) {
			for (int d = 0; d < DIMENSION; ++d) {
				out.write((char *) &(jointInfo.constraint[d]), sizeof(double));
			}
		}
	}

	// New in VER_NUM 0.2f: Save more timing information
	double timing4 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_1,GetDurationInSecs);
	double timing5 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_2,GetDurationInSecs);
	double timing6 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_T_STEP_3,GetDurationInSecs);
	double timing7 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_1,GetDurationInSecs);
	double timing8 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_2,GetDurationInSecs);
	double timing9 = MEASURE_PERF_NARG(TIMER_SIMULTANEOUS_W_STEP_3,GetDurationInSecs);
	double timing10 = MEASURE_PERF_NARG(TIMER_POINTLIST_ANN_QUERY,GetDurationInSecs);
	double timing11 = MEASURE_PERF_NARG(TIMER_POINTLIST_N_ANN_QUERY,GetDurationInSecs);
	out.write((char*) &timing4, sizeof(double));
	out.write((char*) &timing5, sizeof(double));
	out.write((char*) &timing6, sizeof(double));
	out.write((char*) &timing7, sizeof(double));
	out.write((char*) &timing8, sizeof(double));
	out.write((char*) &timing9, sizeof(double));
	out.write((char*) &timing10, sizeof(double));
	out.write((char*) &timing11, sizeof(double));


	out.close();

	cout << DeformGraphAll_Utility::printHeader << "Finished saving solver state!" << endl;
}

void Deform::DeformGraphAll::saveASGState( const DeformAll_Utility::AllSampleGraph *asg, std::ofstream& out ) const
{
	// Parameters
	out.write((char*) &asg->EKNNgraph_k, sizeof(int));
	out.write((char*) &asg->sampleDistThreshold, sizeof(real_type));

	int numSamples = asg->numSamples();
	out.write((char*) &numSamples, sizeof(int));

	// Write the samples
	for (int i = 0; i < numSamples; ++i) {
		for (int j = 0; j < dim; ++j) {
			out.write((char*)&(asg->vertex[i][j]), sizeof(real_type));
		}
	}
	for (int i = 0; i < numSamples; ++i) {
		for (int j = 0; j < dim; ++j) {
			out.write((char*)&(asg->normal[i][j]), sizeof(real_type));
		}
	}
	for (int i = 0; i < numSamples; ++i) {
		out.write((char*)&(asg->fIndex[i]), sizeof(index_type));
	}
	for (int i = 0; i < numSamples; ++i) {
		out.write((char*)&(asg->sIndex[i]), sizeof(index_type));
	}

	// Write the weights
	for (int i = 0; i < numSamples; ++i) {
		int weightSize = asg->weightList[i].size();
		out.write((char*) &weightSize, sizeof(int));
		for (int j = 0; j < weightSize; ++j) {
			out.write((char*) &(asg->weightList[i][j].index), sizeof(int));
			out.write((char*) &(asg->weightList[i][j].weight), sizeof(real_type));
		}
	}

	// Write the graph edges
	int numEdges = num_edges(asg->graph);
	out.write((char*) &numEdges, sizeof(int));

	BoostIndexMap index = get(vertex_index, asg->graph);
	BoostEdgeWeightMap_const edgeWeights = get(edge_weight, asg->graph);
	BoostEdgeIterator ei, ei_end;
	for (tie(ei, ei_end) = edges(asg->graph); ei != ei_end; ++ei) {
		int v1 = index[source(*ei,asg->graph)];
		int v2 = index[target(*ei,asg->graph)];
		BoostEdgeWeightType w = edgeWeights[*ei];
		out.write((char*) &v1, sizeof(int));
		out.write((char*) &v2, sizeof(int));
		out.write((char*) &w, sizeof(BoostEdgeWeightType));
	}
}

void Deform::DeformGraphAll::saveFrameState( const DeformGraphAll_Utility::FrameSolution *frameSoln, std::ofstream& out ) const
{
	// Save basic information
	out.write((char*) &(frameSoln->frameNumber), sizeof(int));

	// Save data
	DeformAll_Utility::saveTransformationList(frameSoln->tfList, out);
	DeformAll_Utility::saveVectorList(frameSoln->occludedLabels, out);
	DeformAll_Utility::saveVectorList(frameSoln->frameSamples, out);
}

bool Deform::DeformGraphAll::loadSolverState( const char *filename, bool graphOnly )
{
	ifstream in;
	in.open(filename, ios_base::binary);

	if (in.fail()) {
		cerr << "Error in DeformGraphAll::loadSolverState(): could not open file " << filename << " for reading!" << endl;
		return false;
	}

	cout << DeformGraphAll_Utility::printHeader << "Loading solver state from " << filename << "..." << endl;

	//////////////////////////////////////////////////////////////////////////
	// Check to see if there is version information
	float verNum;
	char *tempBuffer = new char[DeformGraphAll_Utility::version_msg.length()+1];
	in.read(tempBuffer, sizeof(char) * (DeformGraphAll_Utility::version_msg.length()+1));
	in.read((char*) &verNum, sizeof(float));

	if (strcmp(tempBuffer, DeformGraphAll_Utility::version_msg.c_str()) != 0) {			/**< Compare the initial version string */
		// There is no version information, exit with error
		cerr << "Error in DeformGraphAll::loadSolverState(): no version information found!" << endl;
		return false;
	}
	cout << DeformGraphAll_Utility::printHeader << "Save file version: " << verNum << endl;
	//////////////////////////////////////////////////////////////////////////

	// Load the state of the graph
	if (!loadASGState(in, verNum)) {
		return false;
	}

	if (graphOnly) {
		in.close();
		return true;
	}

	// Load data for each frame solution
	size_type numFrameSolutions;
	in.read((char*) &numFrameSolutions, sizeof(size_type));

	for (index_type i = 0; i < numFrameSolutions; ++i) {
		if (!loadFrameState(in, verNum)) {
			return false;
		}
	}

	// Load the last solved frame index
	in.read((char*) &lastFrame, sizeof(int));

	// Load timing statistics
	double timing1, timing2, timing3;
	in.read((char*) &timing1, sizeof(double));
	in.read((char*) &timing2, sizeof(double));
	in.read((char*) &timing3, sizeof(double));


	MEASURE_PERF_ARG(TIMER_MULTIFRAME_INITIALIZE,SetDurationInSecs,timing1);
	MEASURE_PERF_ARG(TIMER_MULTIFRAME_GLOBAL_REG,SetDurationInSecs,timing2);
	MEASURE_PERF_ARG(TIMER_MULTIFRAME_RESAMPLE_ASG,SetDurationInSecs,timing3);

	// New in VER_NUM 0.1f: Load information for joints
	if (verNum >= 0.1f) {
		DeformGraphAll_Utility::JointMap newJointMap;

		int numJoints;
		in.read((char*) &numJoints, sizeof(int));

		for (int i = 0; i < numJoints; ++i) {
			int jointIndex;
			DeformGraphAll_Utility::Joint tempJoint;
			in.read((char *) &jointIndex, sizeof(int));
			in.read((char *) &(tempJoint.type), sizeof(int));

			for (int d = 0; d < DIMENSION; ++d) {
				in.read((char *) &(tempJoint.location[d]), sizeof(double));
			}

			if (tempJoint.type == DeformGraphAll_Utility::JOINT_HINGE) {
				for (int d = 0; d < DIMENSION; ++d) {
					in.read((char *) &(tempJoint.constraint[d]), sizeof(double));
				}
			}

			//cout << "Read joint " << jointIndex << endl;
			newJointMap[jointIndex] = tempJoint;
		}

		jointMap.swap(newJointMap);
	}

	// New in VER_NUM 0.1f: Load extra timing information
	if (verNum >= 0.2f) {
		double timing4, timing5, timing6, timing7, timing8, timing9, timing10, timing11;
		in.read((char*) &timing4, sizeof(double));
		in.read((char*) &timing5, sizeof(double));
		in.read((char*) &timing6, sizeof(double));
		in.read((char*) &timing7, sizeof(double));
		in.read((char*) &timing8, sizeof(double));
		in.read((char*) &timing9, sizeof(double));
		in.read((char*) &timing10, sizeof(double));
		in.read((char*) &timing11, sizeof(double));

		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_T_STEP_1,SetDurationInSecs,timing4);
		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_T_STEP_2,SetDurationInSecs,timing5);
		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_T_STEP_3,SetDurationInSecs,timing6);
		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_W_STEP_1,SetDurationInSecs,timing7);
		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_W_STEP_2,SetDurationInSecs,timing8);
		MEASURE_PERF_ARG(TIMER_SIMULTANEOUS_W_STEP_3,SetDurationInSecs,timing9);
		MEASURE_PERF_ARG(TIMER_POINTLIST_ANN_QUERY,SetDurationInSecs,timing10);
		MEASURE_PERF_ARG(TIMER_POINTLIST_N_ANN_QUERY,SetDurationInSecs,timing11);
	}

	in.close();

	cout << DeformGraphAll_Utility::printHeader << "Loading complete!  Last frame solved: " << lastFrame << endl;
	return true;
}

bool Deform::DeformGraphAll::loadASGState( std::ifstream& in, const float version )
{
	// Parameters
	int k;
	real_type t;
	in.read((char*) &k, sizeof(int));
	in.read((char*) &t, sizeof(real_type));

	// Replace parameters
	DeformAll_Utility::AllSampleGraph *asg = allSampleGraph;
	asg->EKNNgraph_k = k;
	asg->sampleDistThreshold = t;

	int numSamples;
	in.read((char*) &numSamples, sizeof(int));

	cout << DeformGraphAll_Utility::printHeader << "Loading ASG (" << k << "," << t << ") with " << numSamples << " samples." << endl;

	// Read the samples
	asg->vertex.resize(numSamples);
	asg->normal.resize(numSamples);
	asg->fIndex.resize(numSamples);
	asg->sIndex.resize(numSamples);
	for (int i = 0; i < numSamples; ++i) {
		for (int j = 0; j < dim; ++j) {
			in.read((char*)&(asg->vertex[i][j]), sizeof(real_type));
		}
	}
	for (int i = 0; i < numSamples; ++i) {
		for (int j = 0; j < dim; ++j) {
			in.read((char*)&(asg->normal[i][j]), sizeof(real_type));
		}
	}
	for (int i = 0; i < numSamples; ++i) {
		in.read((char*)&(asg->fIndex[i]), sizeof(index_type));
	}
	for (int i = 0; i < numSamples; ++i) {
		in.read((char*)&(asg->sIndex[i]), sizeof(index_type));
	}

	// Read the weights
	asg->weightList.resize(numSamples);
	for (int i = 0; i < numSamples; ++i) {
		int weightSize;
		in.read((char*) &weightSize, sizeof(int));
		asg->weightList[i].resize(weightSize);
		for (int j = 0; j < weightSize; ++j) {
			in.read((char*) &(asg->weightList[i][j].index), sizeof(int));
			in.read((char*) &(asg->weightList[i][j].weight), sizeof(real_type));
		}
	}

	// Read the graph edges
	int numEdges;
	in.read((char*) &numEdges, sizeof(int));
	asg->graph = BoostGraph(numSamples);
	for (int i = 0; i < numEdges; ++i) {
		int v1, v2;
		BoostEdgeWeightType w;
		in.read((char*) &v1, sizeof(int));
		in.read((char*) &v2, sizeof(int));
		in.read((char*) &w, sizeof(BoostEdgeWeightType));
		add_edge(v1, v2, w, asg->graph);
	}

	return true;
}

bool Deform::DeformGraphAll::loadFrameState( std::ifstream& in, const float version )
{
	// Load basic information
	int frameNumber;
	in.read((char*) &frameNumber, sizeof(int));

	cout << DeformGraphAll_Utility::printHeader << "Loading frame: " << frameNumber << endl;

	// Try retrieving the frame
	if (!isValidFrame(frameNumber)) {
		cerr << "Error in loadFrameState(): cannot load frame " << frameNumber << "." << endl;
		return false;
	}

	DeformGraphAll_Utility::FrameSolution *frameSoln = makeFrameSoln(frameNumber, true);

	// Load data
	DeformAll_Utility::loadTransformationList(frameSoln->tfList, in);
	DeformAll_Utility::loadVectorList(frameSoln->occludedLabels, in);
	DeformAll_Utility::loadVectorList(frameSoln->frameSamples, in);

	return true;
}

void Deform::DeformGraphAll::saveIKState( const char *filename ) const
{
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in DeformGraphAll::saveIKState(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	cout << "*** Saving solver state for IK to " << filename << "...";


	// 1. Save the state of the ASG
	saveASGState(allSampleGraph, out);

	// 2. Save a small amount of per-frame data
	int numFrames = frameSolution.size();
	out.write((char*) &numFrames, sizeof(int));
	DeformGraphAll_Utility::SolutionTable::const_iterator it;
	for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;

		// Save basic information
		out.write((char*) &(frameSoln->frameNumber), sizeof(int));

		// Save data
		DeformAll_Utility::saveTransformationList(frameSoln->tfList, out);
		DeformAll_Utility::saveVectorList(frameSoln->occludedLabels, out);
	}


	// 3. Save information for joints
	const int numJoints = jointMap.size();
	out.write((char*) &numJoints, sizeof(int));

	DeformGraphAll_Utility::JointMap::const_iterator it_joint;
	for (it_joint = jointMap.begin(); it_joint != jointMap.end(); ++it_joint) {
		int jointIndex = it_joint->first;
		const DeformGraphAll_Utility::Joint& jointInfo = it_joint->second;

		out.write((char *) &jointIndex, sizeof(int));
		out.write((char *) &(jointInfo.type), sizeof(int));

		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char *) &(jointInfo.location[d]), sizeof(double));
		}

		if (jointInfo.type == DeformGraphAll_Utility::JOINT_HINGE) {
			for (int d = 0; d < DIMENSION; ++d) {
				out.write((char *) &(jointInfo.constraint[d]), sizeof(double));
			}
		}
	}

	out.close();

	cout << "Finished saving solver state for IK!" << endl;
}

void Deform::DeformGraphAll::printSolverStats() const
{
	unsigned int totalPts = 0;
	for (DeformGraphAll_Utility::SolutionTable::const_iterator it = frameSolution.begin(); it != frameSolution.end(); ++it) {
		totalPts += it->second->originalShape.numVertices();
	}

	double totalTime = MEASURE_PERF_NARG(TIMER_MULTIFRAME_INITIALIZE,GetDurationInSecs) + MEASURE_PERF_NARG(TIMER_MULTIFRAME_GLOBAL_REG,GetDurationInSecs) + MEASURE_PERF_NARG(TIMER_MULTIFRAME_RESAMPLE_ASG,GetDurationInSecs);

	cout << DeformGraphAll_Utility::printHeader << "------------------------------------------------------------" << endl;
	cout << DeformGraphAll_Utility::printHeader << "Current timings:" << endl;
	cout << DeformGraphAll_Utility::printHeader << "Total time: " << totalTime << " seconds." << endl;
	printTime(TIMER_MULTIFRAME_INITIALIZE);
	printTime(TIMER_MULTIFRAME_GLOBAL_REG);
	printTime(TIMER_MULTIFRAME_RESAMPLE_ASG);
	printTime(TIMER_SIMULTANEOUS_T_STEP_1);
	printTime(TIMER_SIMULTANEOUS_T_STEP_2);
	printTime(TIMER_SIMULTANEOUS_T_STEP_3);
	printTime(TIMER_SIMULTANEOUS_W_STEP_1);
	printTime(TIMER_SIMULTANEOUS_W_STEP_2);
	printTime(TIMER_SIMULTANEOUS_W_STEP_3);
	printTime(TIMER_POINTLIST_ANN_QUERY);			// TESTING: measure time spent on ANN queries
	printTime(TIMER_POINTLIST_N_ANN_QUERY);			// TESTING: measure time spent on ANN queries
	cout << DeformGraphAll_Utility::printHeader << "Total number of bones: " << defOpt.numBones << endl;
	cout << DeformGraphAll_Utility::printHeader << "Total number of bones used: " << defOpt.numBones - excludedLabels.size() << endl;
	cout << DeformGraphAll_Utility::printHeader << "Total number of frames: " << frameSolution.size() << endl;
	cout << DeformGraphAll_Utility::printHeader << "Sliding window size: " << defOpt.slidingWindowSize << endl;
	cout << DeformGraphAll_Utility::printHeader << "Average number of points/frame: " << ((double)totalPts)/((double)frameSolution.size()) << endl;
	cout << DeformGraphAll_Utility::printHeader << "Total number of points: " << totalPts << endl;
	cout << DeformGraphAll_Utility::printHeader << "Number of samples on ASG: " << allSampleGraph->numSamples() << endl;
	cout << DeformGraphAll_Utility::printHeader << "Number of edges on ASG: " << num_edges(allSampleGraph->graph) << endl;
	cout << DeformGraphAll_Utility::printHeader << "------------------------------------------------------------" << endl;

	cout << DeformGraphAll_Utility::printHeader << "Memory Usage Information:" << endl;

	// Get memory usage information for all loaded shapes
	{
		int totalMemory = 0;
		int numFrames = frameSolution.size();
		DeformGraphAll_Utility::SolutionTable::const_iterator it;
		for (it = frameSolution.begin(); it != frameSolution.end(); ++it) {
			DeformGraphAll_Utility::FrameSolution *frameSoln = it->second;
			const Mesh3_real *frameMesh = (const Mesh3_real*) &(frameSoln->originalShape);

			totalMemory += frameMesh->vertex.size() * sizeof(real3_type);
			totalMemory += frameMesh->normal.size() * sizeof(real3_type);
			totalMemory += frameMesh->triangle_vertex_index.size() * sizeof(index3_type);
			totalMemory += frameMesh->triangle_normal_index.size() * sizeof(index3_type);
		}

		cout << DeformGraphAll_Utility::printHeader << "Total " << MemoryToMB(totalMemory) << " MB loaded for " << numFrames << " frames." << endl;
	}

	// from MSDN: ``Collecting Memory Usage Information For a Process''
	// http://msdn.microsoft.com/en-us/library/ms682050(v=VS.85).aspx
	{
		// http://msdn.microsoft.com/en-us/magazine/cc300448.aspx
		// Show GetNativeSystemInfo().  If you're running in a WOW session
		// (e.g., an X86 app running on IA64), it'll return the capabilities
		// of the actual OS, not the WOW info.  If not running in WOW, you'll
		// get the same info as GetSystemInfo
		SYSTEM_INFO sysInfo;
		GetNativeSystemInfo( &sysInfo );
		cout << DeformGraphAll_Utility::printHeader << "Native System Info - Processor type: " << sysInfo.dwProcessorType << " page size: " << sysInfo.dwPageSize << endl;

		DWORD processID = GetCurrentProcessId();
		HANDLE hProcess;
		PROCESS_MEMORY_COUNTERS pmc;

		// Print the process identifier.
		cout << DeformGraphAll_Utility::printHeader << "Process ID: " <<  processID << endl;

		// Print information about the memory usage of the process.
		hProcess = OpenProcess(  PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, processID );

		if ((NULL != hProcess) && GetProcessMemoryInfo( hProcess, &pmc, sizeof(pmc) ) )
		{
			cout << DeformGraphAll_Utility::printHeader << "PageFaultCount: " << pmc.PageFaultCount << " faults" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "PeakWorkingSetSize: " << MemoryToMB(pmc.PeakWorkingSetSize) << " MB" << endl;
			cout << DeformGraphAll_Utility::printHeader << "WorkingSetSize: " << MemoryToMB(pmc.WorkingSetSize) << " MB" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "QuotaPeakPagedPoolUsage: " << MemoryToMB(pmc.QuotaPeakPagedPoolUsage) << " MB" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "QuotaPagedPoolUsage: " << MemoryToMB(pmc.QuotaPagedPoolUsage) << " MB" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "QuotaPeakNonPagedPoolUsage: " << MemoryToMB(pmc.QuotaPeakNonPagedPoolUsage) << " MB" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "QuotaNonPagedPoolUsage: " << MemoryToMB(pmc.QuotaNonPagedPoolUsage) << " MB" << endl;
			//cout << DeformGraphAll_Utility::printHeader << "PagefileUsage: " << MemoryToMB(pmc.PagefileUsage) << " MB" << endl; 
			//cout << DeformGraphAll_Utility::printHeader << "PeakPagefileUsage: " << MemoryToMB(pmc.PeakPagefileUsage) << " MB" << endl;
		}

		CloseHandle( hProcess );

		cout << DeformGraphAll_Utility::printHeader << "Private Working Set Size: " << MemoryToMB(CalculateWSPrivate(processID, sysInfo)) << " MB" << endl;
	}

	cout << DeformGraphAll_Utility::printHeader << "------------------------------------------------------------" << endl;
}


void Deform::DeformGraphAll::printJointTransformInfo( std::ostream &out ) const
{
	const int numBones = defOpt.numBones;

	// Write to file: number of bones
	out << "Number of parts: " << numBones << endl;

	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);
		const DeformGraphAll_Utility::Joint& joint = it->second;
		const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
		const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

		// Write to file: bi, bj, joint type, joint position, joint constraint (direction of hinge)
		out << bi << ", " << bj << ", " << joint.type << ", ";
		out << jointPos.x() << ", " << jointPos.y() << ", " << jointPos.z() << ", ";
		out << jointCon.x() << ", " << jointCon.y() << ", " << jointCon.z() << endl;
	}

	DeformGraphAll_Utility::SolutionTable::const_iterator it2;
	for (it2 = frameSolution.begin(); it2 != frameSolution.end(); ++it2) {
		int drawFrameIndex = it2->first;
		const DeformGraphAll_Utility::FrameSolution *drawSolution = it2->second;

		if (drawFrameIndex != referenceFrameIndex)
		{
			// Save the inverse transformations (going from the reference to the frame)
			TransformList tfList = -drawSolution->tfList;

			for (int i = 0; i < numBones; ++i)
			{
				SE3_6DOF &tf = tfList[i];
				const Geometry::Matrix33f &R = tf.rotMatrix();
				const real3_type &tr = tf.tr();

				out << drawFrameIndex << ", ";
				out << R(0,0) << ", " << R(1,0) << ", " << R(2,0) << ", " << 0.0 << ", ";
				out << R(0,1) << ", " << R(1,1) << ", " << R(2,1) << ", " << 0.0 << ", ";
				out << R(0,2) << ", " << R(1,2) << ", " << R(2,2) << ", " << 0.0 << ", ";
				out << tr.x() << ", " << tr.y() << ", " << tr.z() << ", " << 1.0 << endl;
			}
		}
	}
}


void Deform::DeformGraphAll::matchOccludedParts( DeformGraphAll_Utility::FrameSolution* newSoln )
{
	if (!defOpt.ptr_scene || !newSoln->occludedLabels.size()) {
		return;
	}


	const int newFrameNumber = newSoln->frameNumber;
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	// Divide the points into groups corresponding to each bone
	const int numBones = defOpt.numBones;
	std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
	for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
		int label = findLabel(allSampleGraph->weightList[i]);
		vRef_bone[label].add(vRef[i]);
		nRef_bone[label].add(nRef[i]);
	}

	// Transform all existing samples to the new frame's pose
	PointList vertex_t, normal_t;
	allSampleGraph->getVertexNormal_Frame(newFrameNumber, allTfList, vertex_t, normal_t);

	// Transform each bone's reference samples to the new frame
	const DeformAll_Utility::TransformList& tfList = newSoln->tfList;
	std::vector<bool> isOccluded = newSoln->isBoneOccludedOrExcluded(excludedLabels);

	// Extrapolate weights using the existing samples
	std::vector<int> newLabels;
	extrapolateWeights(vRef_bone, nRef_bone, isOccluded, newSoln->originalShape.vertex, newSoln->frameSamples, tfList, defOpt.refSampleDist, newLabels, &(allSampleGraph->testScores));


	// Add these samples
	allSampleGraph->vertex_disp.resize(newSoln->frameSamples.size());
	allSampleGraph->testIndex.clear();
	for (int i = 0, i_end = newSoln->frameSamples.size(); i < i_end; ++i) {
		const real3_type& newPt = newSoln->originalShape.vertex[newSoln->frameSamples[i]];
		// Check extrapolated weights
		if (newLabels[i] < 0) {
			// The sample's unmatched
			allSampleGraph->testIndex.push_back(i);
		}
		allSampleGraph->vertex_disp[i] = newPt;
	}


	//////////////////////////////////////////////////////////////////////////
	// Gather the occluded parts on the reference frame and the unmatched parts on the new frame
	Scene& scene = *defOpt.ptr_scene;
	ShapeModel refShape(SHAPE_POINTS), newShape(SHAPE_POINTS);
	idxList refShape_ASGIndex, newShape_frameIndex;
	refShape.setSampleResolution(defOpt.sceneGridSpacing);
	newShape.setSampleResolution(defOpt.sceneGridSpacing);

	// Add the occluded bones to the refShape
	for (int i = 0, i_end = allSampleGraph->numSamples(); i < i_end; ++i) {
		int label = findLabel(allSampleGraph->weightList[i]);
		if (isOccluded[label]) {
			// Add this sample to the refShape
			refShape.vertex.add(vRef[i]);
			refShape.normal.add(nRef[i]);
			refShape_ASGIndex.push_back(i);
		}
	}
	// Add the unmatched samples to the newShape
	for (int i = 0, i_end = newSoln->frameSamples.size(); i < i_end; ++i) {
		const real3_type& newPt = newSoln->originalShape.vertex[newSoln->frameSamples[i]];
		const real3_type& newNormal = newSoln->originalShape.normal[newSoln->frameSamples[i]];

		// Check extrapolated weights
		if (newLabels[i] < 0) {
			// The sample's unmatched
			newShape.vertex.add(newPt);
			newShape.normal.add(newNormal);
			newShape_frameIndex.push_back(i);
		}
	}

	// Make sure to mark the vertices as valid!
	const int refShapeNum = refShape.numVertices();
	const int newShapeNum = newShape.numVertices();
	refShape.markVertex.resize(refShape.vertex.size(), VERTEX_VALID);
	newShape.markVertex.resize(newShape.vertex.size(), VERTEX_VALID);


	//////////////////////////////////////////////////////////////////////////
	// If there's a significant amount of unmatched data left
	real_type unmatchedRatio = ((real_type)newShapeNum) / ((real_type)newSoln->frameSamples.size());

	cout << DeformGraphAll_Utility::printHeader << "Unmatched data ratio: " << unmatchedRatio << endl;
	if (unmatchedRatio > defOpt.minMatchRatio) {
		// ICP initialization between occluded bones and remaining data
		cout << DeformGraphAll_Utility::printHeader << "-- Performing ICP on occluded parts with remaining data" << endl;

		// For pausing in the middle
		//pauseUntilEnter("Matching occluded parts");

		// Create temporary frame using the remaining points
		DeformGraphAll_Utility::FrameSolution *tempSoln = new DeformGraphAll_Utility::FrameSolution(newShape, defOpt.numBones, newFrameNumber, 0.0);
		tempSoln->tfList = tfList;
		cout << DeformGraphAll_Utility::printHeader << "Excluding labels:";
		for (int i = 0; i < isOccluded.size(); ++i) {
			if (!isOccluded[i]) {
				tempSoln->occludedLabels.push_back(i);
				cout << " " << i;
			}
		}
		cout << endl;

		// Swap out the closest points
		DeformGraphAll_Utility::ClosestPointMap closestTemp;
		sampleClosestPoints[newFrameNumber].swap(closestTemp);

		// Change closest point thresholds, joint constraint weight
		real_type maxDist_saved = defOpt.maxCorrepDist;
		real_type minCosAngle_saved = defOpt.minNormalCosAngle;
		real_type jointWeight_saved = defOpt.TWeights[1];
		defOpt.maxCorrepDist = defOpt.occluded_maxCorrespDist;
		defOpt.minNormalCosAngle = defOpt.occluded_maxNormalAngleDifference;
		defOpt.TWeights[1] = defOpt.occluded_jointOptWeight;

		solveStatus = SOLVER_OCCLUDED_ICP;
		frameSolution.find(newFrameNumber)->second = tempSoln;		/**< Substitute temporary solution */
		selectSolver_SlidingWindow(1);								/**< Disable all but the new frame */
		findTransformations(CORRESPONDENCE_CLOSEST_POINTS);			/**< Solve for the transformations */

		// Restore original solution, closest points, thresholds & weights
		sampleClosestPoints[newFrameNumber].swap(closestTemp);
		defOpt.maxCorrepDist = maxDist_saved;
		defOpt.minNormalCosAngle = minCosAngle_saved;
		defOpt.TWeights[1] = jointWeight_saved;
		solveStatus = SOLVER_IDLE;
		frameSolution.find(newFrameNumber)->second = newSoln;
		selectSolver_SlidingWindow(defOpt.slidingWindowSize);

		// Assign the transformations
		for (int i = 0; i < newSoln->occludedLabels.size(); ++i) {
			const int idx = newSoln->occludedLabels[i];
			newSoln->tfList[idx] = tempSoln->tfList[idx];
		}

		// Try testing occlusions again
		detectOcclusions(newFrameNumber, newSoln->occludedLabels);
		fillOccludedTfs(jointMap, *newSoln, defOpt, frameSolution);
	}

	return;



	//////////////////////////////////////////////////////////////////////////
	// Load the frames and spin images
	refShape.frames.resize(refShapeNum);
	refShape.spinImageStack.resize(refShapeNum);
	newShape.frames.resize(newShapeNum);
	newShape.spinImageStack.resize(newShapeNum);

	for (int i = 0; i < refShapeNum; ++i) {
		const int idx = refShape_ASGIndex[i];
		DeformGraphAll_Utility::FrameSolution* frameSoln = getFrameSoln(allSampleGraph->fIndex[idx]);
		const SpinImageStack& sampleSI = frameSoln->getSpinImages(defOpt.getSIFileName().c_str(), scene);
		const CurvInfoList& sampleCurv = frameSoln->getCurvInfo(defOpt.getCurvFileName().c_str(), scene);
		const distList *curvScales = frameSoln->getCurvScales();

		refShape.frames[i] = sampleCurv[allSampleGraph->sIndex[idx]];
		refShape.spinImageStack[i] = sampleSI[allSampleGraph->sIndex[idx]];

		if (!refShape.frameScales.size() && curvScales) {
			refShape.frameScales = *curvScales;

			cout << DeformGraphAll_Utility::printHeader << "Computed curvature scales: [";
			for (int j = 0; j < curvScales->size(); ++j)
			{
				if (j != 0) cout << ",";
				cout << (*curvScales)[j];
			}
			cout << "]" << endl;
		}
		if (!refShape.spinImageStack.binSize) {
			refShape.spinImageStack.copyParams(sampleSI);

			cout << "Loaded spin image features..." << endl;
			cout << "*** Bin Size: " << refShape.spinImageStack.binSize << endl;
			cout << "*** Image Width: " << refShape.spinImageStack.imageWidth << endl;
			cout << "*** Support Angle: " << refShape.spinImageStack.supportAngle << " degrees" << endl;
			cout << "*** Expected overlap: " << refShape.spinImageStack.lambda << " bins" << endl;
		}
	}
	// Load the frames and spin images
	{
		const SpinImageStack& sampleSI = newSoln->getSpinImages(defOpt.getSIFileName().c_str(), scene);
		const CurvInfoList& sampleCurv = newSoln->getCurvInfo(defOpt.getCurvFileName().c_str(), scene);
		const distList *curvScales = newSoln->getCurvScales();
		newShape.frameScales = *curvScales;
		newShape.spinImageStack.copyParams(sampleSI);

		for (int i = 0; i < newShapeNum; ++i) {
			const int idx = newShape_frameIndex[i];
			newShape.frames[i] = sampleCurv[idx];
			newShape.spinImageStack[i] = sampleSI[idx];
		}

		cout << DeformGraphAll_Utility::printHeader << "Computed curvature scales: [";
		for (int j = 0; j < curvScales->size(); ++j)
		{
			if (j != 0) cout << ",";
			cout << (*curvScales)[j];
		}
		cout << "]" << endl;

		cout << "Loaded spin image features..." << endl;
		cout << "*** Bin Size: " << refShape.spinImageStack.binSize << endl;
		cout << "*** Image Width: " << refShape.spinImageStack.imageWidth << endl;
		cout << "*** Support Angle: " << refShape.spinImageStack.supportAngle << " degrees" << endl;
		cout << "*** Expected overlap: " << refShape.spinImageStack.lambda << " bins" << endl;
	}

	// Sample
	PointList& refSample = refShape.vertex;
	PointList& newSample = newShape.vertex;


	// Create graphs
	BoostGraph refGraph, newGraph;
	makeEKNNgraph(defOpt.EKNNGraph_k, refSample, refGraph);
	makeEKNNgraph(defOpt.EKNNGraph_k, newSample, newGraph);

	// Prune edges of graph... TODO?
	//DeformGraphAll_Utility::pruneEdges(refSample, refWeights, refGraph, allTfList, 2.0, REAL_TYPE_MAX);

	drawMutex.lock();
	tempPts1 = &refSample;
	tempPts2 = &newSample;
	tempGraph1 = &refGraph;
	tempGraph2 = &newGraph;
	drawMutex.unlock();



	// Graph-cut initial correspondences between lastFrame and newFrame
	cout << DeformGraphAll_Utility::printHeader << "Computing GC on occluded subproblem..." << endl;

	// Graph-cut initial correspondences between lastFrame and newFrame
	MatchCandidates *matchCand = new MatchCandidates(refShape, newShape);
	defOpt.ptr_scene->matchCandidates = matchCand;		/**< (Testing) Display the result */
	pauseUntilEnter("Solving GCMatch");
	defOpt.ptr_scene->solveGraphCutCorrespondences(matchCand, &refSample, &newSample, &refGraph, &newGraph);





	pauseUntilEnter("Constructing occlusion matching");

	drawMutex.lock();
	tempPts1 = NULL;
	tempPts2 = NULL;
	tempGraph1 = NULL;
	tempGraph2 = NULL;
	drawMutex.unlock();
}

void Deform::DeformGraphAll::interact( const real3_type& pos ) const
{
	if (allSampleGraph && allSampleGraph->vertex_disp.size()) {
		idxList nearest;
		distList dists;
		allSampleGraph->vertex_disp.queryKNN(pos, 1, nearest, dists, true);
		int idx = nearest[0];

		// Display the scores of the closest sample point
		const int numBones = defOpt.numBones;
		const distList& scores = allSampleGraph->testScores[idx];

		idxList sortIndex(numBones);
		real_type thres = outlierT(scores, sortIndex);
		int maxBone = sortIndex[0];

		for (int j = 0, j_end = numBones; j < j_end; ++j) {
			cout << "Vertex " << idx << " score " << sortIndex[j] << ": " << scores[sortIndex[j]] << endl;
		}
		cout << "-- Winner is bone " << maxBone << " (threshold " << thres << ")" << endl;

		//const real_type& scoreSum = scores[numBones];
		//real_type maxScore = -REAL_TYPE_MAX;
		//int maxBone = -1;
		//for (int j = 0, j_end = numBones; j < j_end; ++j) {
		//	cout << "Vertex " << idx << " score " << j << ": " << scores[j] << "/" << scoreSum;
		//	real_type temp = scores[j] / scoreSum;
		//	if (temp > maxScore) {
		//		maxScore = temp;
		//		maxBone = j;
		//	}
		//	cout << " = " << temp << endl;
		//}
		//cout << "-- Winner is bone " << maxBone << endl;
	}
}

real_type evalError( const std::vector<distList>& lhsRows, const distList& rhsEntries, const distList& soln )
{
	real_type error = 0.0;
	const int N_row = lhsRows.size();
	const int N_col = lhsRows[0].size();
	for (int i = 0; i < N_row; ++i) {
		real_type mult = 0.0;
		for (int j = 0; j < N_col; j++) {
			mult += lhsRows[i][j] * soln[j];
		}
		mult = mult - rhsEntries[i];
		error += mult*mult;
	}

	return error;
}

void printMatrixInfo( taucs_ccs_matrix *m )
{
	cerr << "# cols = " << m->n << endl;
	cerr << "# rows = " << m->m << endl;
	cerr << "flags = " << m->flags << endl;
}

void addToDiagonal_Symmetric( taucs_ccs_matrix *m, taucsType value )
{
	// Print diagonal entries of the symmetric matrix
	int n = m->n;

	//int nnz = m->colptr[n];
	//for (int i = 0; i < nnz; ++i) {
	//	cout << m->taucs_values[i] << endl;
	//}

	for (int i = 0; i < n; ++i) {
		int location = m->colptr[i];

		// Perform some error checking
		int row = m->rowind[location];
		if (i != row) {
			cout << "Warning in addToDiagonal_Symmetric(): Supposed diagonal element (" << i << "," << row << ")=" << m->taucs_values[location] << endl;
		}
		else {
			m->taucs_values[location] += value;
		}
	}
}


void Deform::DeformGraphAll::makeWeightDistanceTerm( const int asgIdx, const ptList& boneRefSrcX, const std::vector<int>& boneCandidates, const real_type ptWeight, const real_type plWeight, std::vector<distList>& result, distList& rhs )
{
	const int nCand = boneCandidates.size();
	//distList sumError(nCand,0.0);

	DeformGraphAll_Utility::AllClosestPoints::const_iterator it;
	for (it = sampleClosestPoints.begin(); it != sampleClosestPoints.end(); ++it) {
		const int targetFrameIdx = it->first;
		const DeformGraphAll_Utility::FrameSolution *targetSoln = getFrameSoln(targetFrameIdx);
		const ShapeModel& targetShape = targetSoln->originalShape;
		const DeformAll_Utility::TransformList& targetTf = targetSoln->tfList;

		// Find target points of this sample on targetFrameIdx
		const DeformGraphAll_Utility::ClosestPointMap& closestPts = it->second;
		DeformGraphAll_Utility::ClosestPointMap::const_iterator it2 = closestPts.find(asgIdx);
		if (it2 != closestPts.end()) {
			const int tgtShapeIdx = it2->second;
			real3_type targetX = targetShape.vertex[tgtShapeIdx], targetN = targetShape.normal[tgtShapeIdx];

			//cout << DeformGraphAll_Utility::printHeader << "---- Adding target point " << tgtShapeIdx << " (frame " << targetFrameIdx << ")" << endl;

			// Temporary variables to store the constraints
			distList ptX(nCand), ptY(nCand), ptZ(nCand), pl(nCand);

			// LHS: For each bone affecting this correspondence
			for (int bi = 0; bi < nCand; ++bi) {
				int boneIndex = boneCandidates[bi];

				real3_type bone_defSourceX;
				bone_defSourceX = (-targetTf[boneIndex]) * boneRefSrcX[bi];
				ptX[bi] = ptWeight*bone_defSourceX[0];
				ptY[bi] = ptWeight*bone_defSourceX[1];
				ptZ[bi] = ptWeight*bone_defSourceX[2];
				pl[bi] = plWeight*bone_defSourceX*targetN;

				//sumError[bi] += (bone_defSourceX - targetX).square_norm();
				//if (bi == 0) {
				//	cout << DeformGraphAll_Utility::printHeader << "---- Default Error: " << (bone_defSourceX - targetX).norm();
				//	cout << " (" << bone_defSourceX << " - " << targetX << ")" << endl;
				//}
			}

			// Add constraints
			const int currentSize = result.size();
			result.resize(currentSize+4);
			rhs.resize(currentSize+4);
			result[currentSize+0].swap(ptX);
			result[currentSize+1].swap(ptY);
			result[currentSize+2].swap(ptZ);
			result[currentSize+3].swap(pl);
			rhs[currentSize+0] = ptWeight*targetX[0];
			rhs[currentSize+1] = ptWeight*targetX[1];
			rhs[currentSize+2] = ptWeight*targetX[2];
			rhs[currentSize+3] = plWeight*targetX*targetN;
		}
	}

	//cout << DeformGraphAll_Utility::printHeader << "-- Sum of error: [";
	//for (int bi = 0; bi < nCand; ++bi) {
	//	if (bi != 0) cout << ",";
	//	cout << sumError[bi];
	//}
	//cout << "]" << endl;
}

void Deform::DeformGraphAll::accumulateASG( real_type maxStretchFactor, real_type maxEdgeLength, bool keepJointEdges )
{
	// In the accumulation strategy, we always add samples on the reference pose.  So we can just use asg->vertex.
	const PointList& vRef = allSampleGraph->vertex;
	const PointList& nRef = allSampleGraph->normal;

	// Divide the points into groups corresponding to each bone
	const int numBones = defOpt.numBones;
	std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
	for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
		int label = findLabel(allSampleGraph->weightList[i]);
		vRef_bone[label].add(vRef[i]);
		nRef_bone[label].add(nRef[i]);
	}

	//cout << "Divided " << vRef.size() << " points into " << numBones << " groups:" << endl;
	//for (int i = 0; i < numBones; ++i) {
	//	cout << "-- Group " << i << ": " << vRef_bone[i].size() << endl;
	//}

	// Accumulate samples using the very last frame
	idxList frameList;
	getSortedFrameList(frameSolution, frameList);
	const int lastFrameIndex = frameList[frameList.size() - 1];

	DeformAll_Utility::AllSampleGraph *newASG, *oldASG;
	oldASG = allSampleGraph;
	newASG = new DeformAll_Utility::AllSampleGraph(defOpt.EKNNGraph_k, defOpt.refSampleDist);
	*newASG = *oldASG;

	DeformAll_Utility::SparseTLArray allTfList;
	allTfList.clear();
	allTfList[referenceFrameIndex] = &(getFrameSoln(referenceFrameIndex)->tfList);
	allTfList[lastFrameIndex] = &(getFrameSoln(lastFrameIndex)->tfList);

	{
		cout << DeformGraphAll_Utility::printHeader << "Adding samples from frame " << lastFrameIndex;

		// Add this frame's samples to the mix, re-sampling previous ASG
		DeformGraphAll_Utility::FrameSolution* frameSoln = getFrameSoln(lastFrameIndex);
		std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);
		const idxList* sampleIdx = &(frameSoln->frameSamples);

		//// Check if we want to use all points in this frame (removed for now)
		//idxList allPointIdx(frameSoln->originalShape.numVertices());
		//if(useAllPoints) {
		//	for (int i = 0; i < allPointIdx.size(); ++i) {
		//		allPointIdx[i] = i;
		//	}
		//	sampleIdx = &allPointIdx;
		//}

		newASG->rebuild(frameSoln->originalShape, *sampleIdx, lastFrameIndex, isOccluded, vRef_bone, nRef_bone, allTfList, defOpt.refSampleDist, true, referenceFrameIndex);

		cout << " --> " << newASG->numSamples() << endl;
	}

	// Rebuild EKNN graph on the reference frame
	makeEKNNgraph(newASG->EKNNgraph_k, newASG->vertex, newASG->graph);

	// Prune edges: look at transformations for all frames
	cout << "Graph size V,E (before pruning): " << num_vertices(newASG->graph) << "," << num_edges(newASG->graph) << endl;
	if (keepJointEdges) {
		newASG->pruneEdges(allTfList, maxStretchFactor, maxEdgeLength, &jointMap);
	}
	else {
		newASG->pruneEdges(allTfList, maxStretchFactor, maxEdgeLength, NULL);
	}
	cout << "Graph size V,E (after pruning):  " << num_vertices(newASG->graph) << "," << num_edges(newASG->graph) << endl;


	drawMutex.lock();
	// Replace previous ASG
	allSampleGraph = newASG;

	// Clear closest points and errors
	sampleClosestPoints.clear();
	allFrameError.clear();
	sumError.clear();
	drawMutex.unlock();

	// Clean up
	delete oldASG;
}

void Deform::DeformGraphAll::setupInteractiveIK( ShapeModel*& resultShape, DeformGraphAll_Utility::JointMap*& resultJoints, DeformAll_Utility::RefWeightList*& resultWeights, std::vector<bool>*& resultValid, QMutex*& resultLock )
{
	// Create data structures
	IKMeshDef = new Mesh3_real((const Mesh3_real&)*finalShape);
	IKMeshValid = new std::vector<bool>(finalShape->numVertices(), true);
	IKMeshLock = new QMutex();

	// Shared variables with ShapePoser
	resultShape = IKMeshDef;
	resultJoints = &jointMap;
	resultWeights = finalShapeWeights;
	resultValid = IKMeshValid;
	resultLock = IKMeshLock;
}

void Deform::DeformGraphAll::solveInteractiveIK()
{
	if (!finalShape || !defOpt.ptr_scene || !IKMeshDef || !IKMeshValid || !IKMeshLock) {
		cerr << "Error in DeformGraphAll::solveInteractiveIK(): Variables not set up yet!" << endl;
		return;
	}

	// Find the transformations continuously using the mesh
	findTransformations_IK(finalShape, IKMeshDef, *IKMeshValid, defOpt.ptr_scene, *finalShapeWeights);

	// After finished
	return;
}

taucsType Deform::DeformGraphAll::findTransformations_IK( const ShapeModel *shape, ShapeModel *deformed, std::vector<bool>& valid, Scene *scene, const DeformAll_Utility::RefWeightList &weights ) const
{
	if (defOpt.TWeightsIK.size() != 2) {
		cerr << "Error in findTransformations_IK()! Not enough weights for each term!" << endl;
		return 0.0;
	}

	bool done = false;
	int TStepNumItr = 1;
	taucsType Tres = 0.0, lastTres = DBL_MAX, stepRes = 0.0;
	const int numBones = defOpt.numBones;
	std::vector<int> empty(0);

	DeformAll_Utility::TfVarTranslate varTranslate(0, numBones, excludedLabels, empty);
	varTranslate.printMapping();

	// Initialize correspondence list
	CorrespondenceList corresp;
	DeformAll_Utility::TransformList tfList(numBones), idList(numBones);

	// Initialize temporary structures for computing deformed result (double-buffering for geometry)
	const int numV = shape->numVertices();
	PointList tempV(numV), tempN(numV);
	std::vector<bool> *tempValid = new std::vector<bool>(numV, true);


	// T-step loop
	while (!done) {
		for (int i = 0; i < 100; ++i) {
			cout << "\b";
		}
		cout << "itr=" << TStepNumItr++;

		// Update correspondences
		scene->updateInteractiveCorresp(corresp);
		cout << ", # corresp=" << corresp.size() << " weights: " << defOpt.TWeightsIK[0] << " " << defOpt.TWeightsIK[1];
		if (!corresp.size()) {
			continue;
		}

		// Compute translation/scaling
		//real3_type avgPt;
		//real_type avgDist;
		//ptList constraintPosList;
		//{
		//	// Average correspondence positions
		//	for (int ci = 0; ci < corresp.size(); ++ci) {
		//		constraintPosList.push_back(shape->vertex[corresp.sourceIndex[ci]]);
		//		constraintPosList.push_back(avgPt += corresp.targetPos(ci));
		//	}
		//	// Average joint positions
		//	DeformGraphAll_Utility::JointMap::const_iterator it;
		//	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		//		int bi, bj, linIdx = it->first;
		//		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		//		// Otherwise, in normal operation, active the joint if both bones are activated.
		//		if (varTranslate.validBone(0, bi) && varTranslate.validBone(0, bj))
		//		{
		//			const DeformGraphAll_Utility::Joint& joint = it->second;
		//			const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
		//			const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

		//			int numSamples = 0;
		//			if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
		//				numSamples = 1;
		//			}
		//			else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
		//				numSamples = defOpt.hingeJointSamples;
		//			}
		//			//int numSamples = defOpt.hingeJointSamples;

		//			for (int j = 0; j < numSamples; ++j) {
		//				real3_type jointLocation;

		//				if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
		//					jointLocation = jointPos;
		//				}
		//				else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
		//					real_type t = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
		//					jointLocation = jointPos +  t*jointCon;
		//				}

		//				constraintPosList.push_back(jointLocation);
		//			}
		//		}
		//	}

		//	// Compute stats
		//	avgDist = avgPt[0] = avgPt[1] = avgPt[2] = 0.0;
		//	const int numPts = constraintPosList.size();
		//	for (int i = 0; i < numPts; ++i) {
		//		avgPt += constraintPosList[i];
		//	}
		//	avgPt = avgPt / ((real_type)numPts);
		//	for (int i = 0; i < numPts; ++i) {
		//		avgDist += (constraintPosList[i] - avgPt).norm();
		//	}
		//	avgDist = avgDist / ((real_type)numPts);
		//	//cout << " Avg Pt/Dist/Num: " << avgPt << "/" << avgDist << "/" << numPts;
		//}
		//real_type s = sqrt(2.f) / avgDist;
		//real3_type t = avgPt;
		real_type s = 1.0;
		real3_type t(0.0,0.0,0.0);

		//////////////////////////////////////////////////////////////////////////
		// (1) Construct system
		//////////////////////////////////////////////////////////////////////////
		TermList terms;
		if (defOpt.TWeightsIK[0] > 0.0) {
			ErrorTerm *distanceTerm = DeformGraphAll_Utility::makeDistanceTerm_IK(shape, corresp, weights, tfList, varTranslate, s, t);
			distanceTerm->weight = sqrt(defOpt.TWeightsIK[0]);
			terms.push_back(distanceTerm);
		}
		if (defOpt.TWeightsIK[1] > 0.0) {
			ErrorTerm *smoothnessTerm = DeformGraphAll_Utility::makeSmoothnessTerm_BallJoint_IK(varTranslate, 0, s, t, tfList, defOpt, jointMap);
			smoothnessTerm->weight = sqrt(defOpt.TWeightsIK[1]);
			terms.push_back(smoothnessTerm);
		}


		//////////////////////////////////////////////////////////////////////////
		// (2) Solve system
		//////////////////////////////////////////////////////////////////////////
		taucsType *soln = DeformAll_Utility::solveLinear(terms, 6 * varTranslate.numTotalBones(), Tres, stepRes, defOpt.saveTaucsErrorLog, 0.001);

		// Apply transformations
		real_type stepSize = 1.0;
		for (int i = 0; i < varTranslate.numTotalBones(); ++i) {
			int idx = 6*i;
			int frameIndex, boneIndex;
			varTranslate.frameBoneIndex(i, frameIndex, boneIndex);
			DeformAll_Utility::SE3_6DOF& updateTf = tfList[boneIndex];

			real3_type linearRot(stepSize*soln[idx],stepSize*soln[idx+1],stepSize*soln[idx+2]);
			real3_type linearTr(stepSize*soln[idx+3],stepSize*soln[idx+4],stepSize*soln[idx+5]);

			// Projection using Plucker coordinates
			DeformAll_Utility::SE3_6DOF correctionTf2;
			correctionTf2.createFromLinear(linearRot, linearTr);
			//updateTf = correctionTf2 * updateTf;

			// Transform the translation
			real3_type tNew = correctionTf2.tr() / s - correctionTf2.rotMatrix() * t + t;
			DeformAll_Utility::SE3_6DOF correct(correctionTf2.rotMatrix(), tNew);
			updateTf = correct * updateTf;
		}

		// Deform using the precomputed weight list
		for (size_type i = 0; i < numV; ++i) {
			if (weights[i].valid()) {				
				// Deform vertex and normal
				lbs(shape->vertex[i], shape->normal[i], tempV[i], tempN[i], weights[i], tfList, false);
			}

			(*tempValid)[i] = weights[i].valid();
		}
		// Check edge lengths
		DeformAll_Utility::checkEdgeLengths(tempV, *shape, defOpt.edgeLenStretchThreshold_Sq, defOpt.deformCosAngleThreshold, *tempValid);

		// Swap result
		IKMeshLock->lock();
		deformed->vertex.swap(tempV);
		deformed->normal.swap(tempN);
		valid.swap(*tempValid);
		IKMeshLock->unlock();

		//////////////////////////////////////////////////////////////////////////
		// (6) Cleanup: Delete error terms & solution
		//////////////////////////////////////////////////////////////////////////
		for (int i = 0; i < terms.size(); ++i) {
			if (terms[i]) {
				delete terms[i];
				terms[i] = NULL;
			}
		}
		terms.clear();
		delete[] soln;
	}

	cout << "Total number of iterations: " << TStepNumItr << endl;
	return lastTres;
}


ErrorTerm* Deform::DeformGraphAll_Utility::makeDistanceTerm_IK( const ShapeModel *shape, const CorrespondenceList &corresp, const DeformAll_Utility::RefWeightList &shapeWeights, const DeformAll_Utility::TransformList &tfList, const DeformAll_Utility::TfVarTranslate& varTranslate, const real_type& s,  const real3_type& t )
{
	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	for (index_type ci = 0; ci < corresp.size(); ++ci)
	{
		index_type shapeIndex = corresp.sourceIndex[ci];
		const DeformAll_Utility::RefWeight& weights = shapeWeights[shapeIndex];

		// (1) LHS: For each bone affecting this correspondence
		for (int bi = 0; bi < weights.size(); ++bi) {
			totalLHSnnz += 9;								/**< Point-to-point metric */
		}

		// Increment row
		totalRows += 3;
	}


	//////////////////////////////////////////////////////////////////////////
	// Create the term
	ErrorTerm *distanceTerm = new ErrorTerm(1.0, totalRows, totalLHSnnz);

	// Keep track of the row as we populate the matrices
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	//real_type totalResidual = 0.0;

	data_type Rpt0(12), Rpt1(12), Rpt2(12);
	Rpt0.setZero();
	Rpt1.setZero();
	Rpt2.setZero();

	for (index_type ci = 0; ci < corresp.size(); ++ci)
	{
		//real_type targetWeight = corresp.getWeight(ci);
		index_type shapeIndex = corresp.sourceIndex[ci];
		real3_type sourceX = shape->vertex[shapeIndex];
		real3_type targetX = corresp.targetPos(ci);
		real3_type deformedX(0.0, 0.0, 0.0);
		const DeformAll_Utility::RefWeight& weights = shapeWeights[shapeIndex];
		bool populated = false;

		// (1) LHS: For each bone affecting this correspondence
		for (int bi = 0; bi < weights.size(); ++bi) {
			// Find bone's current transformation
			int boneIndex = weights.bones[bi];
			real_type boneWeight = weights.weights[bi];
			std::vector<data_type> Rpt;

			// Deform positions using the current bone transformation
			real3_type boneDefX = tfList[boneIndex]*sourceX;

			// Apply translation and scaling to the positions
			boneDefX = s*(boneDefX - t);
			targetX = s*(targetX - t);

			// Compute point-to-point metric
			DeformAll_Utility::helperJacPtSA(boneWeight, boneWeight, boneDefX, targetX, Rpt0, Rpt1, Rpt2);

			if (varTranslate.validBone(0, boneIndex)) {
				// Populate matrix entries
				index_type varCol = 6 * varTranslate.linearIndex(0, boneIndex);		/**< Column where the variable belongs */

				// Populate entries for point-to-point metric
				for (int col = 0; col < 6; ++col) {
					if (Rpt0.val[col] != 0.0) {
						distanceTerm->lhs.fill(TrackRow+0, varCol+col, Rpt0.val[col]);
					}
					if (Rpt1.val[col] != 0.0) {
						distanceTerm->lhs.fill(TrackRow+1, varCol+col, Rpt1.val[col]);
					}
					if (Rpt2.val[col] != 0.0) {
						distanceTerm->lhs.fill(TrackRow+2, varCol+col, Rpt2.val[col]);
					}
				}

				populated = true;
			}

			// For RHS in (2): Compute deformed point
			deformedX += boneWeight * boneDefX;
		}

		// (2) RHS: Compute residuals (Check to see if any entries were populated at all)
		if (populated) {
			// (2) RHS: Compute residual for point-to-point metric
			real3_type residualPt = deformedX - targetX;

			for (int k = 0; k < 3; ++k) {
				distanceTerm->rhs[TrackRow+k] = -residualPt[k];
				//distanceTerm->rowWeight[TrackRow+k] = targetWeight;
			}

			// Increment row
			TrackRow += 3;
		}
	}


	////////////////////////////////////////////////////////////////////////
	// Debugging: print out results
	//for (int ci = 0; ci < distanceTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << distanceTerm->rhs[ci] << ";" << endl;
	//}
	//distanceTerm->lhs.print();
	if ((totalLHSnnz != distanceTerm->lhs.size()) || (TrackRow != distanceTerm->rows)) {
		cout << "*** Distance Term:" << endl;
		cout << "Estimated size for Jacobian: " << totalLHSnnz
			<< " vs. Actual size: " << distanceTerm->lhs.size() << endl;
		cout << "*** Estimated # rows: " << distanceTerm->rows
			<< " vs. Actual # rows: " << TrackRow << endl;
	}

	return distanceTerm;
}

ErrorTerm* Deform::DeformGraphAll_Utility::makeSmoothnessTerm_BallJoint_IK( const DeformAll_Utility::TfVarTranslate& varTranslate, int targetFrameIndex, const real_type& s,  const real3_type& t, const DeformAll_Utility::TransformList& tfList, const DeformAll_Utility::Options& defOpt, const DeformGraphAll_Utility::JointMap& jointMap )
{
	// Initialize variables
	const int numBones = defOpt.numBones;

	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	DeformGraphAll_Utility::JointMap::const_iterator it;
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		if (varTranslate.validBone(targetFrameIndex, bi) && varTranslate.validBone(targetFrameIndex, bj))
		{
			const DeformGraphAll_Utility::Joint& joint = it->second;
			//int numSamples = 1;
			//if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
			//	numSamples = defOpt.hingeJointSamples;
			//}
			int numSamples = defOpt.hingeJointSamples;

			totalLHSnnz += 18 * numSamples;
			totalRows += 3 * numSamples;
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// Create the term
	ErrorTerm *smoothTerm = new ErrorTerm(1.0, totalRows, totalLHSnnz);

	// Keep track of the row as we populate the matrices
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	// Iterate through each unique pair
	for (it = jointMap.begin(); it != jointMap.end(); ++it) {
		int bi, bj, linIdx = it->first;
		tie(bi,bj) = EnumerateUniquePairs<int>::linearIndexToPairIndex(numBones, linIdx);

		// Otherwise, in normal operation, active the joint if both bones are activated.
		if (varTranslate.validBone(targetFrameIndex, bi) && varTranslate.validBone(targetFrameIndex, bj))
		{
			const DeformGraphAll_Utility::Joint& joint = it->second;
			const real3_type jointPos(joint.location[0],joint.location[1],joint.location[2]);
			const real3_type jointCon(joint.constraint[0],joint.constraint[1],joint.constraint[2]);

			//int numSamples = 0;
			//if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
			//	numSamples = 1;
			//}
			//else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
			//	numSamples = defOpt.hingeJointSamples;
			//}
			int numSamples = defOpt.hingeJointSamples;

			for (int j = 0; j < numSamples; ++j) {
				real3_type jointLocation;

				if (joint.type == DeformGraphAll_Utility::JOINT_BALL) {
					jointLocation = jointPos;
				}
				else if (joint.type == DeformGraphAll_Utility::JOINT_HINGE) {
					real_type hingeT = defOpt.hingeExtent * ( 2.0 * ((real_type)j) / ((real_type)numSamples-1) - 1.0);
					jointLocation = jointPos +  hingeT*jointCon;
				}

				//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				// Populate optimization matrices
				//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				std::vector<data_type> RJoint;

				// Transform the joint position with each bone transformation
				real3_type jointDef_bi, jointDef_bj;
				jointDef_bi = tfList[bi] * jointLocation;
				jointDef_bj = tfList[bj] * jointLocation;

				// Apply translation and scaling to the positions
				jointDef_bi = s*(jointDef_bi - t);
				jointDef_bj = s*(jointDef_bj - t);

				// Populate LHS matrix entries
				DeformAll_Utility::helperJointJacPtSA(jointDef_bi, jointDef_bj, RJoint);

				index_type varCol_bi = 6 * varTranslate.linearIndex(targetFrameIndex, bi);		/**< Column where the variable belongs */
				for (int k = 0; k < 3; ++k) {
					for (int col = 0; col < 6; ++col) {
						if (RJoint[k].val[col] != 0.0) {
							smoothTerm->lhs.fill(TrackRow+k, varCol_bi+col, RJoint[k].val[col]);
						}
					}
				}
				index_type varCol_bj = 6 * varTranslate.linearIndex(targetFrameIndex, bj);		/**< Column where the variable belongs */
				for (int k = 0; k < 3; ++k) {
					for (int col = 0; col < 6; ++col) {
						if (RJoint[k].val[col+6] != 0.0) {
							smoothTerm->lhs.fill(TrackRow+k, varCol_bj+col, RJoint[k].val[col+6]);
						}
					}
				}

				// Populate RHS residual entries
				real3_type residualJoint = jointDef_bi - jointDef_bj;
				for (int k = 0; k < 3; ++k) {
					smoothTerm->rhs[TrackRow+k] = -residualJoint[k];

					// TODO: assign a weight for each joint if necessary
					//smoothTerm->rowWeight[TrackRow+k] = ;
				}

				// Increment row
				TrackRow += 3;
			}
		}
	}

	////////////////////////////////////////////////////////////////////////
	// Debugging: print out results
	//for (int ci = 0; ci < smoothTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << smoothTerm->rhs[ci] << ";" << endl;
	//}
	//distanceTerm->lhs.print();
	if ((totalLHSnnz != smoothTerm->lhs.size()) || (TrackRow != smoothTerm->rows)) {
		cout << "*** Smoothness (Ball Joint) Term:" << endl;
		cout << "Estimated size for Jacobian: " << totalLHSnnz
			<< " vs. Actual size: " << smoothTerm->lhs.size() << endl;
		cout << "*** Estimated # rows: " << smoothTerm->rows
			<< " vs. Actual # rows: " << TrackRow << endl;
	}

	return smoothTerm;
}

Geometry::Matrix33f BVHLoader::getRotMatrixFromRotations( float rx, float ry, float rz, const std::string& rotationOrder )
{
	double cosA, sinA, angle;
	Geometry::Matrix33f rotationX, rotationY, rotationZ, total;
	rotationX = rotationY = rotationZ = total = Geometry::Matrix33f::identity();

	angle = ((double)rx) * M_PI / 180.0;
	cosA = cos(angle); sinA = sin(angle);
	rotationX(1,1) = cosA; rotationX(1,2) = -sinA;
	rotationX(2,1) = sinA; rotationX(2,2) = cosA;

	angle = ((double)ry) * M_PI / 180.0;
	cosA = cos(angle); sinA = sin(angle);
	rotationY(0,0) = cosA; rotationY(0,2) = sinA;
	rotationY(2,0) = -sinA; rotationY(2,2) = cosA;

	angle = ((double)rz) * M_PI / 180.0;
	cosA = cos(angle); sinA = sin(angle);
	rotationZ(0,0) = cosA; rotationZ(0,1) = -sinA;
	rotationZ(1,0) = sinA; rotationZ(1,1) = cosA;

	for (int i=0; i<3; i++) {
		if (rotationOrder[i] == 'X') {
			total = total * rotationX;
		}

		if (rotationOrder[i] == 'Y') {
			total = total * rotationY;
		}

		if (rotationOrder[i] == 'Z') {
			total = total * rotationZ;
		}
	}

	return total;
}

namespace Deform
{
	namespace DeformGraphAll_Utility
	{
		void populateJointsFromBVH( std::tr1::weak_ptr<BVHLoader::Joint> root, int numTotalBones, const char* masterName, JointMap& joints, std::tr1::shared_ptr<DeformAll_Utility::TransformList> initialState )
		{
			// Walk the skeleton, mark joint locations
			typedef struct _Recit
			{
				std::tr1::weak_ptr<BVHLoader::Joint> joint;
				real3_type parentOffset;
				int parentIndex;
				bool hasMasterParent;
				DeformAll_Utility::SE3_6DOF parentLocalTf;
			} Recit;
			std::deque<Recit> recQueue;

			Recit parent;
			parent.joint = root;
			parent.parentIndex = (masterName ? -1 : 0);
			parent.hasMasterParent = false;
			parent.parentOffset = real3_type(0.f, 0.f, 0.f);
			parent.parentLocalTf = DeformAll_Utility::SE3_6DOF();
			recQueue.push_back(parent);

			glLineWidth(5.f);

			do 
			{
				bool childOfMaster = false;
				real3_type currentOffset = real3_type(0.f, 0.f, 0.f);

				// Pull offset off queue
				parent = recQueue.front();
				recQueue.pop_front();

				// Visit child
				std::tr1::shared_ptr<BVHLoader::Joint> parentJoint(parent.joint);
				std::tr1::shared_ptr<BVHLoader::JointInfo> parentInfo(parentJoint->getInfo());

				int parentIndex = parentInfo->stateIndex;

				DeformAll_Utility::SE3_6DOF parentFrame = parent.parentLocalTf
						* DeformAll_Utility::SE3_6DOF(Geometry::Matrix33f::identity(), parentInfo->offset)
						* ((*initialState)[parentIndex]);

				if (parentInfo)
				{
					// Apply rotation and translation when adding the offset
					currentOffset = parentFrame.tr();

					if (parent.hasMasterParent || masterName == NULL || parentInfo->name == masterName)
					{
						childOfMaster = true;
					}
				}

				// This is the offset for this joint, record it
				// Joint is between this and the parent
				if (parent.hasMasterParent)
				{
					if (parent.parentIndex < 0)
					{
						cerr << "Error!!! parent index " << parent.parentIndex << " is < 0!" << std::endl;
					}
					else
					{
						//// Add currentOffset as joint
						//glDisable(GL_LIGHTING);
						//setOpenGLColor(boneColors[parentIndex]);
						//glBegin(GL_LINES);
						//glVertex3f(parent.parentOffset.x(), parent.parentOffset.y(), parent.parentOffset.z());
						//glVertex3f(currentOffset.x(), currentOffset.y(), currentOffset.z());
						//glEnd();

						//glEnable(GL_LIGHTING);
						//setOpenGLMaterialColor(boneColors[parent.parentIndex]);
						//drawSphere(parent.parentOffset, jointDisplaySize);
						//setOpenGLMaterialColor(boneColors[parentIndex]);
						//drawSphere(currentOffset, jointDisplaySize);

						Joint j;
						j.type = JOINT_BALL;
						j.location = Geometry::Vec3d(currentOffset.x(), currentOffset.y(), currentOffset.z());
						int linIdx = EnumerateUniquePairs<int>::pairIndexToLinearIndex(numTotalBones, parent.parentIndex, parentIndex);
						joints[linIdx] = j;
						cout << "Adding ball joint between " << parent.parentIndex << " and " << parentIndex << std::endl;
					}
				}

				// Update queue
				for (int i = 0; i < parentJoint->numChildren(); ++i)
				{
					Recit anew;
					anew.joint = parentJoint->getChild(i);
					anew.parentIndex = parentIndex;
					anew.hasMasterParent = childOfMaster;
					anew.parentOffset = currentOffset;
					anew.parentLocalTf = parentFrame;
					//cout << "Concat: parent " << current.index << " with self " << curIndex << std::endl;
					//cout << " - Parent: " << current.parentLocalTf << std::endl;
					//cout <<"  - Self: " << anew.parentLocalTf << std::endl;
					recQueue.push_back(anew);

					std::tr1::shared_ptr<BVHLoader::Joint> childJoint(anew.joint);
					std::tr1::shared_ptr<BVHLoader::JointInfo> childInfo(childJoint->getInfo());
					//cout << "Found bone " << curIndex << " with name: " << childInfo->name << std::endl;
				}
			} while (!recQueue.empty());
		}
	}
}

std::tr1::shared_ptr<BVHLoader::Joint> Deform::DeformGraphAll_Utility::loadJointsBVH( const std::string& bvhFilename, BVHLoader::MySkeletonStateList& statesList )
{
	ifstream in;
	in.open(bvhFilename.c_str());

	if (in.fail()) {
		cerr << "Error in DeformGraphAll_Utility::loadJointsBVH(): could not open file " << bvhFilename << " for reading!" << endl;
		return std::tr1::shared_ptr<BVHLoader::Joint>(static_cast<BVHLoader::Joint*>(NULL));
	}

	cout << DeformGraphAll_Utility::printHeader << "Loading BVH joints from " << bvhFilename << "..." << endl;

	std::tr1::shared_ptr<BVHLoader::Joint> root(static_cast<BVHLoader::Joint*>(NULL));
	std::tr1::shared_ptr<BVHLoader::Joint> currentJoint(static_cast<BVHLoader::Joint*>(NULL));
	std::tr1::shared_ptr<BVHLoader::JointInfo> currentJointInfo(static_cast<BVHLoader::JointInfo*>(NULL));
	std::vector<std::tr1::shared_ptr<BVHLoader::JointInfo> > jointInfoList;
	StringArray jointNamesList;
	int totalNumChannels = 0;
	bool startedMotion = false;
	std::string line;

	while(!std::getline(in, line).eof())
	{
		StringArray elementsWSpaces;
		trim(line);
		split(elementsWSpaces, line, is_any_of("[\t ]"), token_compress_on);

		// remove extra spaces
		int count = 0;
		for (int i = 0; i < elementsWSpaces.size(); i++)
		{
			if (elementsWSpaces[i] != "")
				count++;
		}

		StringArray elements(count);
		count = 0;
		for (int i = 0; i < elementsWSpaces.size(); i++)
		{
			if (elementsWSpaces[i] != "")
				elements[count++] = elementsWSpaces[i];
		}

		if (starts_with(line, "ROOT")) {
			// only parse first figure
			if (root != NULL) {
				do 
				{
					if (std::getline(in, line).eof())
					{
						goto end;
					}

					trim(line);
				} while (!starts_with(line, "MOTION"));
				startedMotion = true;
			}
			else {
				std::tr1::shared_ptr<BVHLoader::JointInfo> newJointInfo(new BVHLoader::JointInfo(elements[1], jointNamesList.size()));
				std::tr1::shared_ptr<BVHLoader::Joint> newJoint(new BVHLoader::Joint(newJointInfo));
				currentJoint = newJoint;
				root = newJoint;
				currentJointInfo = newJointInfo;
				jointNamesList.push_back(elements[1]);
				jointInfoList.push_back(currentJointInfo);
			}
		}

		else if (starts_with(line, "OFFSET") && currentJoint) {
			currentJointInfo->offset[0] = atof(elements[1].c_str());
			currentJointInfo->offset[1] = atof(elements[2].c_str());
			currentJointInfo->offset[2] = atof(elements[3].c_str());
		}

		else if (starts_with(line, "CHANNELS") && currentJoint) {
			int numChannels = atoi(elements[1].c_str());
			currentJointInfo->rotationOrder = "";
			for (int i=0; i<numChannels; i++) {
				if (elements[2 + i] == "Xrotation") {
					currentJointInfo->rotationOrder += "X";
					currentJointInfo->XRotIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Yrotation") {
					currentJointInfo->rotationOrder += "Y";
					currentJointInfo->YRotIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Zrotation") {
					currentJointInfo->rotationOrder += "Z";
					currentJointInfo->ZRotIndex = totalNumChannels+i;
				}

				if (elements[2 + i] == "Xposition") {
					currentJointInfo->XPosIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Yposition") {
					currentJointInfo->YPosIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Zposition") {
					currentJointInfo->ZPosIndex = totalNumChannels+i;
				}

				if (elements[2 + i] == "Xscale") {
					currentJointInfo->XScaleIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Yscale") {
					currentJointInfo->YScaleIndex = totalNumChannels+i;
				}
				if (elements[2 + i] == "Zscale") {
					currentJointInfo->ZScaleIndex = totalNumChannels+i;
				}
			}
			totalNumChannels += numChannels;
		}

		else if (starts_with(line, "End Site") && currentJoint) {
			do 
			{
				if (std::getline(in, line).eof())
				{
					goto end;
				}

				trim(line);
			} while (!starts_with(line, "}"));
		}

		// start a new joint
		else if (starts_with(line, "JOINT") && currentJoint) {
			cerr << "Creating joint: " << elements[1] << std::endl;
			currentJointInfo = std::tr1::shared_ptr<BVHLoader::JointInfo>(new BVHLoader::JointInfo(elements[1], jointNamesList.size()));
			std::tr1::shared_ptr<BVHLoader::Joint> newJoint = std::tr1::shared_ptr<BVHLoader::Joint>(new BVHLoader::Joint(currentJointInfo));
			BVHLoader::Joint::addChild(currentJoint, newJoint);
			currentJoint = newJoint;
			jointNamesList.push_back(elements[1]);
			jointInfoList.push_back(currentJointInfo);
		}

		// return to the parent joint
		else if (starts_with(line, "}") && currentJoint) {
			std::tr1::shared_ptr<BVHLoader::JointInfo> info(currentJoint->getInfo());
			cerr << "Done with joint: " << info->name << std::endl;
			currentJoint = currentJoint->getParent().lock();
		} 

		else if (starts_with(line, "MOTION")) {
			startedMotion = true;
		} 

		else if (startedMotion && !starts_with(line, "Frame") ) {

			if (elements.size() != totalNumChannels) {
				cerr << "Line: " << statesList.size() << std::endl;
				cerr << elements.size() << " != " << totalNumChannels << std::endl;
				goto end;
			}

			std::tr1::shared_ptr<DeformAll_Utility::TransformList> newState = 
				std::tr1::shared_ptr<DeformAll_Utility::TransformList>(new DeformAll_Utility::TransformList());

			float currentXscale = 1.f;
			float currentYscale = 1.f;
			float currentZscale = 1.f;
			for (int i=0; i<jointInfoList.size(); i++)
			{
				std::tr1::shared_ptr<BVHLoader::JointInfo> info(jointInfoList[i]);
				real3_type translation;
				Geometry::Matrix33f rotation = Geometry::Matrix33f::identity();
				
				if (info->XPosIndex != -1) {
					translation.x() += atof(elements[info->XPosIndex].c_str());
				}
				if (info->YPosIndex != -1) {
					translation.y() += atof(elements[info->YPosIndex].c_str());
				}
				if (info->ZPosIndex != -1) {
					translation.z() += atof(elements[info->ZPosIndex].c_str());
				}
				// XXX: Hack!  Need to add proper hierarchical handling of scale in the future
				translation.x() *= currentXscale;
				translation.y() *= currentYscale;
				translation.z() *= currentZscale;

				if (info->XScaleIndex != -1) {
					currentXscale *= atof(elements[info->XScaleIndex].c_str());
				}
				if (info->YScaleIndex != -1) {
					currentYscale *= atof(elements[info->YScaleIndex].c_str());
				}
				if (info->ZScaleIndex != -1) {
					currentZscale *= atof(elements[info->ZScaleIndex].c_str());
				}

				if ( !(info->XRotIndex == -1 || info->YRotIndex == -1 || info->ZRotIndex == -1) ) {
					rotation = BVHLoader::getRotMatrixFromRotations(
						atof(elements[info->XRotIndex].c_str()), 
						atof(elements[info->YRotIndex].c_str()), 
						atof(elements[info->ZRotIndex].c_str()), info->rotationOrder);
				}

				newState->push_back(DeformAll_Utility::SE3_6DOF(rotation, translation));
			}

			statesList.push_back(newState);
		}
	}

end:

	//BVHLoader::MySkeleton* mySkeleton = new BVHLoader::MySkeleton(root, jointNamesList.size());

	//BVHLoader::MySkeletonStateSequence* myStates = new BVHLoader::MySkeletonStateSequence(statesList, jointNamesList);

	// TODO
	//mySkeleton.initializeJointMap(myStates.getJointNamesFragile());

	//mySkeleton.setState(myStates.getStateFragile(0));

	in.close();

	cout << DeformGraphAll_Utility::printHeader << "Loading complete!  Number of loaded joints: " << jointNamesList.size() << " and motion frames: " << statesList.size() << endl;

	return root;
}

int2_type Deform::DeformGraphAll_Utility::loadWeightFile( const std::string& weightFilename, DeformAll_Utility::RefWeightList& weights )
{
	int2_type result;
	ifstream in;
	in.open(weightFilename.c_str());

	if (in.fail()) {
		cerr << "Error in DeformGraphAll_Utility::loadWeightFile(): could not open file " << weightFilename << " for reading!" << endl;
		return result;
	}

	cout << DeformGraphAll_Utility::printHeader << "Loading weights from " << weightFilename << "..." << endl;

	int nVerts, nBones;
	in >> nVerts >> nBones;

	weights.resize(nVerts);
	for (int i = 0; i < nVerts; ++i)
	{
		int maxBone = -1;
		real_type maxWeight = -REAL_TYPE_MAX;
		for (int j = 0; j < nBones; ++j)
		{
			real_type weight;
			in >> weight;
			if (weight > maxWeight)
			{
				maxBone = j;
				maxWeight = weight;
			}
		}
		weights[i].bones.push_back(maxBone);
		weights[i].weights.push_back(1.f);
	}

	in.close();

	cout << DeformGraphAll_Utility::printHeader << "Loading complete!  Number of loaded vertices/bones: " << nVerts << ", " << nBones << endl;

	result.x() = nVerts;
	result.y() = nBones;

	return result;
}

QString Deform::DeformGraphAll::solverStatusString() const
{
	QString msg;

	switch(solveStatus) {
		case SOLVER_IDLE:
			msg = QString("Idle");
			break;
		case SOLVER_INIT:
			msg = QString("Initializing Frame %1").arg(lastFrame+1);
			break;
		case SOLVER_T_STEP:
			msg = QString("Optimizing Transformations");
			break;
		case SOLVER_W_STEP:
			msg = QString("Optimizing Weights");
			break;
		case SOLVER_OCCLUDED_ICP:
			msg = QString("Matching Occluded Parts");
			break;
		case SOLVER_UPDATE_SAMPLES:
			msg = QString("Update Sample Set");
			break;
		default:
			break;
	}

	return msg;
}

/*
 * Extrapolate weights on each frame, transform each frame to reference, save result to obj file
 */
void Deform::DeformGraphAll::writeTransformedFrames()
{
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);
	PointList vRef, nRef;
	allSampleGraph->getVertexNormal_Ref(allTfList, vRef, nRef);

	// Divide the points into groups corresponding to each bone
	const int numBones = defOpt.numBones;
	std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
	for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
		int label = findLabel(allSampleGraph->weightList[i]);
		vRef_bone[label].add(vRef[i]);
		nRef_bone[label].add(nRef[i]);
	}

	// Sort frame list
	idxList frameList;
	getSortedFrameList(frameSolution, frameList);

	for (int i = 0; i < frameList.size(); ++i) {
		index_type frameIdx = frameList[i];

		// Add this frame's samples to the mix, re-sampling previous ASG
		DeformGraphAll_Utility::FrameSolution* frameSoln = frameSolution.find(frameIdx)->second;
		std::vector<bool> isOccluded = frameSoln->isBoneOccludedOrExcluded(excludedLabels);

		// Transform all existing DSG samples to the new frame's pose
		PointList vertex_t, normal_t;
		allSampleGraph->getVertexNormal_Frame(frameIdx, allTfList, vertex_t, normal_t);

		// Transform each bone's reference samples to the new frame
		SparseTLArray::const_iterator it = allTfList.find(frameIdx);
		if (it == allTfList.end()) {
			cerr << "Error in resampleASG(): Cannot find transformation for target frame " << frameIdx << endl;
		}
		else
		{
			cout << "Transforming frame " << frameIdx << " to reference" << endl;

			const TransformList& tfList = *(it->second);
			const int numBones = tfList.size();
			const ShapeModel& shape = frameSoln->originalShape;

			// Extrapolate the label for all points in this frame
			idxList allPointIdx(shape.numVertices());
			for (int i = 0; i < allPointIdx.size(); ++i) {
				allPointIdx[i] = i;
			}

			// Extrapolate weights using the existing samples
			std::vector<int> newLabels;
			extrapolateWeights(vRef_bone, nRef_bone, isOccluded, shape.vertex, allPointIdx, tfList, defOpt.refSampleDist, newLabels);

			DeformAll_Utility::RefWeightList frameWeights(newLabels.size());
			for (int i = 0; i < newLabels.size(); ++i) {
				if (newLabels[i] >= 0) {
					frameWeights[i].bones.push_back(newLabels[i]);
					frameWeights[i].weights.push_back(1.0);
				}
			}

			// Deform to reference frame
			DeformAll_Utility::ShapeDeformed *shapeDef = new ShapeDeformed(shape);
			shapeDef->deform(frameWeights, tfList, false);

			// Store in frame solution
			frameSoln->setRefDeformed(shapeDef);

			// Save data
			if (!defOpt.writeFrameToRef.empty()) {
				// Save to a different file for every frame, if the string contains '#' symbols
				string saveFile = defOpt.writeFrameToRef;
				replaceField(saveFile, '#', frameIdx);

				shapeDef->writeGeometryOBJ(saveFile.c_str(), false);
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////
	//// TESTING CODE: deform ref to lastFrame
	////////////////////////////////////////////////////////////////////////////
	//// 1. Extrap weights on original ref frame
	//// Divide the points into groups corresponding to each bone
	//const int numBones = defOpt.numBones;
	//std::vector<PointList> vRef_bone(numBones), nRef_bone(numBones);
	//for (int i = 0; i < allSampleGraph->numSamples(); ++i) {
	//	int label = findLabel(allSampleGraph->weightList[i]);
	//	vRef_bone[label].add(vRef[i]);
	//	nRef_bone[label].add(nRef[i]);
	//}
	//DeformGraphAll_Utility::FrameSolution *refSoln = getFrameSoln(referenceFrameIndex);
	//const PointList &refPts = refSoln->originalShape.vertex;
	//idxList allPointIdx(refPts.size());
	//for (int i = 0; i < allPointIdx.size(); ++i) {
	//	allPointIdx[i] = i;
	//}
	//std::vector<bool> isOccluded = refSoln->isBoneOccludedOrExcluded(excludedLabels);
	//std::vector<int> newLabels;
	//extrapolateWeights(vRef_bone, nRef_bone, isOccluded, refPts, allPointIdx, refSoln->tfList, defOpt.refSampleDist, newLabels);
	//DeformAll_Utility::RefWeightList refWeights(refPts.size());
	//for (int i = 0; i < refWeights.size(); ++i) {
	//	if (newLabels[i] >= 0) {
	//		refWeights[i].bones.push_back(newLabels[i]);
	//		refWeights[i].weights.push_back(1.0);
	//	}
	//}

	//// 2. Deform reference frame to last solution's pose
	//DeformGraphAll_Utility::FrameSolution *lastSoln = getFrameSoln(lastFrame);
	//DeformAll_Utility::ShapeDeformed refDef(refSoln->originalShape);
	//refDef.deform(refWeights, lastSoln->tfList, true);

	//// 3. Save data
	//refDef.writeGeometryOBJ("J:\\temp.obj", false);
	////////////////////////////////////////////////////////////////////////////
	//// END TESTING CODE
	////////////////////////////////////////////////////////////////////////////
}

void Deform::DeformGraphAll::writeTransformedDSG()
{
	if (defOpt.writeRefToFrame.empty())
	{
		cerr << "Error in writeTransformedDSG(): no filename specified!" << endl;
		return;
	}

	// 1. Build an ASG with very small sample spacing, and prune ALL edges that stretch
	defOpt.refSampleDist = defOpt.finalRefSampleDist;
	rebuildASG(defOpt.finalDSGMaxEdgeStretchFactor, defOpt.finalDSGMaxEdgeLen, true, false);

	// Output the list of points in pts format
	const index_type numSamples = allSampleGraph->numSamples();
	PointList vList, nList;
	DeformAll_Utility::SparseTLArray allTfList;
	makeAllTFList(frameSolution, allTfList);

	// Sort frame list
	idxList frameList;
	getSortedFrameList(frameSolution, frameList);

	for (int i = 0; i < frameList.size(); ++i)
	{
		// Code in this loop mostly taken from AllSampleGraph::getVertexNormal_Frame
		index_type frameIndex = frameList[i];
		DeformGraphAll_Utility::FrameSolution* frameSoln = frameSolution.find(frameIndex)->second;

		vList.resize(numSamples);
		nList.resize(numSamples);

		// Find the transformations of the specific frame to transform to
		SparseTLArray::const_iterator it = allTfList.find(frameIndex);
		if (it == allTfList.end()) {
			cerr << "Error in getVertexNormal_Frame(): Cannot find transformation for target frame " << frameIndex << endl;
			return;
		}
		const TransformList& tfList = *(it->second);

		// Optimization: pre-multiplication of transform list
		SparseTLArray preMultTf;
		for (it = allTfList.begin(); it != allTfList.end(); ++it)
		{
			const int frameIndex = it->first;
			TransformList* newMultiple = new TransformList();
			*newMultiple = (-tfList) * *(it->second);
			preMultTf[frameIndex] = newMultiple;
		}

		// Compute the bounding box of the points
		real3_type box_min, box_max;
		for (int d = 0; d < 3; d++) {
			box_min[d] = REAL_TYPE_MAX;
			box_max[d] = -REAL_TYPE_MAX;
		}

		// Transform each sample
		for (int i = 0; i < numSamples; ++i) {
			if (!allSampleGraph->weightList[i].size()) {
				cerr << "Error in getVertexNormal_Frame(): No weights for sample " << i << endl;
				return;
			}

			if (frameIndex == allSampleGraph->fIndex[i]) {
				vList[i] = allSampleGraph->vertex[i];
				nList[i] = allSampleGraph->normal[i];
			}
			else {
				// Using pre-multiplication
				it = preMultTf.find(allSampleGraph->fIndex[i]);
				if (it == preMultTf.end()) {
					cerr << "Error in findClosestPoints(): Cannot find transformation for sample " << i << ", frame " << allSampleGraph->fIndex[i] << endl;
					break;
				}
				lbs(allSampleGraph->vertex[i], allSampleGraph->normal[i], vList[i], nList[i], allSampleGraph->weightList[i], *(it->second), false);
			}

			// Calculate the bounding box of the point set
			for(int d = 0; d < dim; d++){
				if (vList[i][d] < box_min[d]) {
					box_min[d] = vList[i][d];
				}
				if (vList[i][d] > box_max[d]) {
					box_max[d] = vList[i][d];
				}
			}
		}

		// Optimization: delete pre-multiplication of transform list
		for (it = preMultTf.begin(); it != preMultTf.end(); ++it)
		{
			delete it->second;
		}

		// Write points to file
		std::string ptsFile = defOpt.writeRefToFrame;
		replaceField(ptsFile, '#', frameIndex);
		writePtsFormat(ptsFile.c_str(), box_min, box_max, vList, nList);
	}
}

taucsType* Deform::DeformAll_Utility::solveLinear( TermList& terms, int numVars, taucsType& totalResidual, taucsType& stepResidual, const std::string& logFile, taucsType diagAddValue/* = 0.0*/ )
{
	taucs_ccs_matrix *A, *A_T, *lhs;
	ErrorTerm* all;


	//////////////////////////////////////////////////////////////////////////
	// Construct normal equations
	all = mergeTerms(terms);

	//A = all->lhs.writeTaucsMatrix(all->rows, numVars);
	//printMatrixInfo(A);

	A_T = all->lhs.writeTaucsMatrix(all->rows, numVars, true);
	A = MatrixTranspose(A_T);

	lhs = Mul2NonSymmMatSymmResult(A_T, A);


	// Optionally add values to the diagonal
	if (diagAddValue > 0.0) {
		addToDiagonal_Symmetric(lhs, diagAddValue);
	}

	// Allocate space for the RHS
	taucsType *b = new taucsType[all->rows];
	taucsType *rhs = new taucsType[numVars];
	taucsType *x = new taucsType[numVars];
	for (size_type i = 0; i < all->rows; i++) {
		b[i] = (taucsType) all->rhs[i];
	}

	// Calculate RHS
	MulNonSymmMatrixVector(A_T, b, rhs);

	//// Test residual calculation
	//for (size_type i = 0; i < numVars; i++) {
	//	x[i] = 0.0;
	//}
	//taucsType *Ax = new taucsType[all->rows];
	//totalResidual = 0;
	//MulNonSymmMatrixVector(A, x, Ax);
	//for (size_type i = 0; i < all->rows; i++) {
	//	taucsType temp = Ax[i] - b[i];
	//	totalResidual += temp*temp;
	//}
	//cout << "--->>> Current residual: " << totalResidual << endl;
	//delete[] Ax;


	//////////////////////////////////////////////////////////////////////////
	// Solve the linear system
	int result;
	char* options[] = { "taucs.factor.LLT=true", NULL };

	result = taucs_linsolve (lhs,NULL,1,x,rhs,options,NULL);

	QString message("Error");
	switch(result) {
		case TAUCS_SUCCESS:
			message = QString("Success!");
			break;
		case TAUCS_ERROR:
			message = QString("%1 -- Generic error.").arg(message);
			break;
		case TAUCS_ERROR_NOMEM:
			message = QString("%1 -- NOMEM error.").arg(message);
			break;
		case TAUCS_ERROR_BADARGS:
			message = QString("%1 -- BADARGS error.").arg(message);
			break;
		case TAUCS_ERROR_MAXDEPTH:
			message = QString("%1 -- MAXDEPTH error.").arg(message);
			break;
		case TAUCS_ERROR_INDEFINITE:
			message = QString("%1 -- INDEFINITE error.").arg(message);
			break;
		default:
			break;
	}
	if (result != TAUCS_SUCCESS) {
		cout << "--->>> Linear system solution result: " << qPrintable(message) << endl;

		// Reset solution to zeros before returning
		totalResidual = -1.0;
		stepResidual = -1.0;
		for (size_type i = 0; i < numVars; i++) {
			x[i] = 0.0;
		}

		// For further debugging, write solver information to file.
		if (!logFile.empty())
		{
			int fileIndex = 0;
			std::string saveFile, taucsLogFile;
			
			do {
				saveFile = logFile;
				replaceField(saveFile, '#', fileIndex);
				taucsLogFile = saveFile + "_log.txt";
				++fileIndex;
			} while (FileExists(taucsLogFile));

			// Log the merging of terms and result of linear system solve
			std::string AFile = saveFile + "_A.txt";
			std::string ATFile = saveFile + "_AT.txt";
			std::string ATAFile = saveFile + "_ATA.txt";
			std::string bFile = saveFile + "_b.txt";
			std::string xFile = saveFile + "_x.txt";
			taucs_logfile(const_cast<char*>(taucsLogFile.c_str()));

			ErrorTerm* temp = mergeTerms(terms);
			result = taucs_linsolve(lhs,NULL,1,x,rhs,options,NULL);

			// Write matrix system A * x = b, we will write A^T and b
			taucs_ccs_write_ijv(A, const_cast<char*>(AFile.c_str()));
			taucs_ccs_write_ijv(A_T, const_cast<char*>(ATFile.c_str()));
			taucs_ccs_write_ijv(lhs, const_cast<char*>(ATAFile.c_str()));
			// Write b (right hand side)
			{
				std::ofstream rhsOut;
				rhsOut.open(bFile.c_str());
				if (!rhsOut.fail())
				{
					for (size_type i = 0; i < all->rows; i++) {
						rhsOut << all->rhs[i] << std::endl;
					}
				}
				rhsOut.close();
			}
			// Write x (solution)
			{
				std::ofstream xOut;
				xOut.open(xFile.c_str());
				if (!xOut.fail())
				{
					for (size_type i = 0; i < numVars; i++) {
						xOut << x[i] << std::endl;
					}
				}
				xOut.close();
			}

			// Reset state
			taucs_logfile("none");
			delete temp;
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// Calculate residuals
	if (result == TAUCS_SUCCESS) {
		// Determine stop conditions
		taucsType *Ax = new taucsType[all->rows];
		//taucsType *ATAx = new taucsType[numVars];

		totalResidual = 0.0;
		stepResidual = 0.0;
		MulNonSymmMatrixVector(A, x, Ax);
		for (size_type i = 0; i < all->rows; i++) {
			taucsType temp = Ax[i] - b[i];
			totalResidual += temp*temp;
			stepResidual += Ax[i]*Ax[i];
		}

		//taucsType gradResidual = 0;
		//MulNonSymmMatrixVector(A_T, Ax, ATAx);
		//for (size_type i = 0; i < numVars; i++) {
		//	taucsType temp = 2.0*(ATAx[i] - rhs[i]);
		//	gradResidual += temp*temp;
		//}

		//cout << "--->>> Total Residual: " << totalResidual << ", Gradient: " << gradResidual << endl;

		// Free memory
		delete[] Ax;
		//delete[] ATAx;
	}


	//////////////////////////////////////////////////////////////////////////
	// Deallocate memory
	taucs_ccs_free(A);
	taucs_ccs_free(A_T);
	taucs_ccs_free(lhs);
	delete[] b;
	delete[] rhs;
	delete all;

	return x;
}

template <typename Real, int N_col>
void Deform::DeformAll_Utility::solveLinear_denseSVD( const std::vector< Geometry::Vec<N_col,Real> >& lhsRows, const std::vector<Real>& rhsEntries, Geometry::Vec<N_col,Real>& soln )
{
	// Allocate the storage space for the input/output arrays
	const int N_row = lhsRows.size();
	const int minMN = min(N_row,N_col);
	Real **A = new Real*[N_row];
	Real *pS = new Real[minMN];
	Real **pU = new Real*[N_row];
	Real **pVT = new Real*[minMN];
	for (int i = 0; i < N_row; ++i) {
		A[i] = new Real[N_col];
		pU[i] = new Real[minMN];
		for (int j = 0; j < N_col; j++) {
			A[i][j] = lhsRows[i][j];
		}
	}
	for (int i = 0; i < minMN; ++i) {
		pVT[i] = new Real[N_col];
	}

	// Compute SVD using LAPACK
	LapackLib::dgesvd(A, N_row, N_col, pS, pU, pVT);

	// Compute the least squares solution
	// 1. Reset the solution vector
	for (int i = 0; i < N_col; ++i) {
		soln[i] = 0.0;
	}
	// 2. Compute the solution
	for (int i = 0; i < minMN; ++i) {
		const Real si = pS[i];

		if (si > std::numeric_limits<Real>::epsilon()) {
			// Compute dot product of (ith column of U) with (rhsEntries)
			Real dp = 0.0;
			for (int r = 0; r < N_row; ++r) {
				dp += pU[r][i] * rhsEntries[r];
			}

			// Divide by singular value
			dp = dp / si;

			// Add to result
			for (int c = 0; c < N_col; ++c) {
				soln[c] += dp * pVT[i][c];
			}
		}
	}

	// Delete the vector and matrices
	for (int i = 0; i < N_row; ++i) {
		delete[] A[i];
		delete[] pU[i];
	}
	for (int i = 0; i < minMN; ++i) {
		delete[] pVT[i];
	}
	delete[] A;
	delete[] pS;
	delete[] pU;
	delete[] pVT;
}

// Declaration for the compiler
template void Deform::DeformAll_Utility::solveLinear_denseSVD<double,3>( const std::vector< Geometry::Vec<3,double> >& lhsRows, const std::vector<double>& rhsEntries, Geometry::Vec<3,double>& soln );

void Deform::DeformAll_Utility::helperJointJacPtSA( const real3_type& jointDefPos_i, const real3_type& jointDefPos_j, std::vector<data_type>& R )
{
	static real_type x_i1, x_i2, x_i3, x_j1, x_j2, x_j3;
	x_i1 = jointDefPos_i[0];       x_i2 = jointDefPos_i[1];       x_i3 = jointDefPos_i[2];
	x_j1 = jointDefPos_j[0];       x_j2 = jointDefPos_j[1];       x_j3 = jointDefPos_j[2];

	R.clear();
	R.resize(3, data_type(12, 0.0));
	R[0].val[1] = x_i3;
	R[0].val[2] = -x_i2;
	R[0].val[3] = 1.0;
	R[0].val[7] = -x_j3;
	R[0].val[8] = x_j2;
	R[0].val[9] = -1.0;
	R[1].val[0] = -x_i3;
	R[1].val[2] = x_i1;
	R[1].val[4] = 1.0;
	R[1].val[6] = x_j3;
	R[1].val[8] = -x_j1;
	R[1].val[10] = -1.0;
	R[2].val[0] = x_i2;
	R[2].val[1] = -x_i1;
	R[2].val[5] = 1.0;
	R[2].val[6] = -x_j2;
	R[2].val[7] = x_j1;
	R[2].val[11] = -1.0;
}

void Deform::DeformAll_Utility::checkEdgeLengths( const PointList& vertex, const ShapeModel& shape, real_type edgeThresholdSq, real_type cosAngleThreshold, std::vector<bool> &valid )
{
	if ((shape.shapeType != SHAPE_MESH) || (edgeThresholdSq == 0.0)) {
		return;
	}
	const idx3List& triangle_vertex_index = ((Mesh3_real&)shape).triangle_vertex_index;

	//// Reset the valid bits
	//valid.resize(vertex.size());
	//for (int i = 0; i < valid.size(); ++i) {
	//	valid[i] = true;
	//}

	// Draw triangle mesh, with bone weights painted on
	for(size_type t = 0,t_end = triangle_vertex_index.size(); t < t_end; ++t) {
		const index3_type& tvi = triangle_vertex_index[t];
		const index_type i0 = tvi[0];
		const index_type i1 = tvi[1];
		const index_type i2 = tvi[2];
		real3_type d01 = vertex[i0] - vertex[i1];
		real3_type d02 = vertex[i0] - vertex[i2];
		real3_type d12 = vertex[i1] - vertex[i2];
		real_type l01 = d01.square_norm();
		real_type l02 = d02.square_norm();
		real_type l12 = d12.square_norm();

		// Check three edge lengths
		if (l01 > edgeThresholdSq) {
			valid[i0] = false;
			valid[i1] = false;
		}
		if (l02 > edgeThresholdSq) {
			valid[i0] = false;
			valid[i2] = false;
		}
		if (l12 > edgeThresholdSq) {
			valid[i1] = false;
			valid[i2] = false;
		}

		//// Check three angles and remove skinny triangles
		//// a. Normalize vectors
		//d01 /= sqrt(l01); d02 /= sqrt(l02); d12 /= sqrt(l12);
		//// b. Compute and check angles
		//if (abs(d01*d02) > cosAngleThreshold) {
		//	valid[i0] = false;
		//}
		//if (abs(-d01*d12) > cosAngleThreshold) {
		//	valid[i1] = false;
		//}
		//if (abs(d02*d12) > cosAngleThreshold) {
		//	valid[i2] = false;
		//}
	}
}
void Deform::DeformAll_Utility::saveTransformationList( const TransformList& tfList, std::ofstream& out )
{
	unsigned int n = tfList.size();
	out.write((char*) &n, sizeof(unsigned int));

	for (unsigned int i = 0; i < n; ++i) {
		const SE3_6DOF& tf = tfList[i];
		const real3_type& tf_rot = tf.rot();
		const Geometry::Matrix33f& tf_rotMatrix = tf.rotMatrix();
		const real3_type& tf_tr = tf.tr();

		// Write rotation
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &tf_rot[d], sizeof(real_type));
		}

		// Write rotation matrix
		for (int d = 0; d < DIMENSION; ++d) {
			for (int e = 0; e < DIMENSION; ++e) {
				float value = tf_rotMatrix(d,e);
				out.write((char*) &value, sizeof(float));
			}
		}

		// Write translation
		for (int d = 0; d < DIMENSION; ++d) {
			out.write((char*) &tf_tr[d], sizeof(real_type));
		}
	}
}

void Deform::DeformAll_Utility::loadTransformationList( TransformList& tfList, std::ifstream& in )
{
	unsigned int n;
	in.read((char*) &n, sizeof(unsigned int));

	// Resize transformation list
	tfList.resize(n);

	for (unsigned int i = 0; i < n; ++i) {
		real3_type tf_rot, tf_tr;
		Geometry::Matrix33f tf_rotMatrix;

		// Read rotation
		for (int d = 0; d < DIMENSION; ++d) {
			in.read((char*) &tf_rot[d], sizeof(real_type));
		}

		// Read rotation matrix
		for (int d = 0; d < DIMENSION; ++d) {
			for (int e = 0; e < DIMENSION; ++e) {
				float value;
				in.read((char*) &value, sizeof(float));
				tf_rotMatrix(d,e) = value;
			}
		}

		// Read translation
		for (int d = 0; d < DIMENSION; ++d) {
			in.read((char*) &tf_tr[d], sizeof(real_type));
		}

		// Assign to list
		tfList[i] = SE3_6DOF(tf_rot, tf_rotMatrix, tf_tr);
	}
}

template<typename T>
void Deform::DeformAll_Utility::saveVectorList( const std::vector<T>& vec_list, std::ofstream& out )
{
	// Write total occluded list size
	unsigned int listSize = vec_list.size();
	out.write((char*) &listSize, sizeof(unsigned int));

	for (index_type i = 0; i < listSize; ++i) {
		T temp = vec_list[i];
		out.write((char*) &temp, sizeof(T));
	}
}

template<typename T>
void Deform::DeformAll_Utility::loadVectorList( std::vector<T>& vec_list, std::ifstream& in )
{
	unsigned int listSize;
	in.read((char*) &listSize, sizeof(unsigned int));

	vec_list.clear();
	if (listSize > 0) {
		vec_list.reserve(listSize);
		for (index_type j = 0; j < listSize; ++j) {
			T temp;
			in.read((char*) &temp, sizeof(T));
			vec_list.push_back(temp);
		}
	}
}

template void Deform::DeformAll_Utility::saveVectorList<unsigned int>( const std::vector<unsigned int>& vec_list, std::ofstream& out );
template void Deform::DeformAll_Utility::loadVectorList<unsigned int>( std::vector<unsigned int>& vec_list, std::ifstream& in );


